#ifndef BITMAP_H
#define BITMAP_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CBitmap K-3D object, which encapsulates bitmap images
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3ddialog.h>
#include <k3dpersistobject.h>
#include <k3dvariable.h>
#include <k3dimagebuffer.h>

#include <k3dibitmap.h>
#include <k3ditexture.h>
#include <k3dieditortexture.h>
#include <k3diritexture.h>
#include <k3ditimebase.h>

#include <sdpgtkopengldrawingarea.h>

// Forward class declarations ...
class CBitmap;

/////////////////////////////////////////////////////////////////////////////
// CBitmapDialog

/// Provides the user interface for CBitmap objects
class CBitmapDialog : public k3dDialog
{
public:
	CBitmapDialog();

	bool Create(CBitmap* Object);

protected:
	CBitmap* m_Object;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void DrawRGB();
	void DrawAlpha();
	void OnLoad();
	void OnSave();
	void OnView();

	// Standard control overrides ...
	bool InitializeControls();
	void UpdateControls();
	void FinalizeControls();

	sdpGtkOpenGLDrawingArea m_RGBWidget;
	sdpGtkOpenGLDrawingArea m_AlphaWidget;

	k3dImageBuffer m_PreviewImage;
};

///////////////////////////////////////////////////////////////////////////////
// CBitmap

class CBitmap : public k3dPersistObject,
							public k3dIBitmap,
							public k3dITexture,
							public k3dIEditorTexture,
							public k3dIRiTexture
{
friend class CBitmapDialog;
public:
	CBitmap();
	~CBitmap();

	CBitmapDialog m_UserInterface;

	// k3dIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIBitmap overrides
	sdpResult Allocate(gulong Width, gulong Height);
	sdpResult Free();

	k3dPixel* GetData();
	gulong GetWidth();
	gulong GetHeight();	

	sdpResult LoadFile(SDPCSTRING FilePath, SDPCSTRING Format, k3dIBitmapReadFormat* FormatFilter);
	sdpResult SaveFile(SDPCSTRING FilePath, SDPCSTRING Format, k3dIBitmapWriteFormat* FormatFilter);

	sdpResult FlipVertical();
	sdpResult FlipHorizontal();
	sdpResult Scale(gulong NewWidth, gulong NewHeight);

	// k3dIObject overrides
	sdpResult EditObject();
	sdpResult ObjectRenamed(SDPCSTRING NewName);
	sdpResult ObjectDeleted();

	// k3dIEditorTexture overrides
	sdpResult EditorTexturePreRender(k3dIEditorEngine* Engine);
	sdpResult EditorTextureSetup(k3dIEditorEngine* Engine);

	// k3dIRiTexture overrides
	sdpResult RiTextureStartJob(k3dIRenderJob* Job, k3dIRiEngine* Engine);
	sdpResult RiTexturePreRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);
	sdpResult RiTextureFinishJob(k3dIRenderJob* Job, k3dIRiEngine* Engine);
	
	SDPCSTRING RiTexturePath(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);

	// k3dIAnimate overrides ...
	sdpResult PostAnimate();

	//k3dIPersist overrides
	sdpResult InitNew();
	sdpResult LoadComplete();

protected:
	/// Stores the bitmap image path
	k3dVariable<sdpString> m_ImagePath;
	/// Stores the loaded bitmap image
	k3dImageBuffer m_Image;
	/// Stores the bitmap image after making it suitable for use with OpenGL
	k3dImageBuffer m_OpenGLData;

	typedef enum
	{
		CLAMP = 1,
		TILE = 2,
	} RepeatType;

	class VariableRepeatType : public k3dCustomVariable<RepeatType>
	{
	protected:
		sdpString ConvertToString() { return sdpFormatString("%d", m_Value); }
		bool ConvertFromString(SDPCSTRING String) { m_Value = RepeatType(atoi(String)); return true; }
	};

	VariableRepeatType m_XRepeat;
	VariableRepeatType m_YRepeat;
	k3dVariable<bool> m_UseAlpha;
	k3dVariable<bool> m_UseLuma;
	k3dVariable<bool> m_FullSize;
	k3dVariable<bool> m_Perspective;
	k3dVariable<bool> m_Antialias;

	k3dScalarChannel m_X1;
	k3dScalarChannel m_X2;
	k3dScalarChannel m_Y1;
	k3dScalarChannel m_Y2;

	/// Stores the absolute path to the saved TIFF version of this bitmap
	sdpString m_RiTexturePath;

	/// Called whenever we need to dump our buffer and start over ...
	void UpdateBuffer();
};

#endif // BITMAP_H
