#ifndef CAMERA_H
#define CAMERA_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CCamera K-3D object, which controls the position, orientation, and display of a CCameraView within the 3D workspace
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dtransformobject.h>
#include <k3dlocationserver.h>

#include <k3ddialog.h>
#include <k3dchannelbutton.h>
#include <k3dmouseinput.h>
#include <k3dshaderselector.h>
#include <k3dshaderargumentspage.h>

#include <k3dicamera.h>
#include <k3ditimebase.h>

#include <sdpgtkopengldrawingarea.h>

#include "editorengine.h"
#include "riengine.h"

///////////////////////////////////////////////////////////
// CFrustum

class CFrustum
{
public:
	CFrustum();
	CFrustum(gulong ViewportWidth, gulong ViewportHeight, double FocalLength);
	void SetParameters(gulong ViewportWidth, gulong ViewportHeight, double FocalLength);

	void CameraToScreen(sdpVector3& Coords);

	gulong ViewportWidth() { return m_ViewportWidth; }
	gulong ViewportHeight() { return m_ViewportHeight; }
	double FocalLength() { return m_FocalLength; }
	double FrustumLeft() { return m_FrustumLeft; }
	double FrustumTop() { return m_FrustumTop; }
	double FrustumRight() { return m_FrustumRight; }
	double FrustumBottom() { return m_FrustumBottom; }
	double FrustumNear() { return m_FrustumNear; }
	double FrustumFar() { return m_FrustumFar; }

protected:
	gulong m_ViewportWidth;
	gulong m_ViewportHeight;
	double m_FocalLength;
	double m_FrustumLeft;
	double m_FrustumTop;
	double m_FrustumRight;
	double m_FrustumBottom;
	double m_FrustumNear;
	double m_FrustumFar;

	double m_HalfWidth;
	double m_HalfHeight;
	double m_InverseWidth;
	double m_InverseHeight;
};


/////////////////////////////////////////////////////////////////////////////
// CCameraDialog

class CCameraDialog : public k3dDialog
{
public:
	CCameraDialog();

	bool Create(CCamera* Object);

protected:
	CCamera* m_Object;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void OnEditorSolid();
	void OnEditorFog();
	void OnEditorTexture();
	void OnEditorTwoSided();
	void OnEditorOrientation();
	void OnEditorNormals();
	
	void OnRiEngine();
	void OnRiXSamples();
	void OnRiYSamples();
	void OnRiPixelFilterName();
	void OnRiPixelFilterWidth();
	void OnRiPixelFilterHeight();
	void OnRiAtmosphere();
	void OnRiDepthOfField();
	void OnRiRadiosity();
	void OnRiRadiositySteps();

	void OnShowViewport();
	void OnHideViewport();

	// Update controls ...
	bool InitializeControls();
	void UpdateControls();

	/// Provides the UI for selecting shaders
	k3dShaderSelector m_RiAtmosphereShaderSelector;
	/// Provides the UI for controlling shader arguments
	k3dShaderArgumentsPage m_RiAtmosphereShaderPage;
};

///////////////////////////////////////////////////////////
// CCameraFrame

class CCameraFrame : public sdpGtkObjectContainer,
										public k3dMouseInput
{
public:
	CCameraFrame();

	/// Called to make the frame visible
	bool Create(CCamera* Object);

	/// Called when the frame's owner is deleted
	void ObjectDeleted();

	// Accesses the viewport
	sdpGtkOpenGLDrawingArea& Viewport() { return m_Viewport; }

	// Update titlebar ...
	sdpString Title();
	void UpdateTitlebar();

protected:
	CCamera* m_Object;
	sdpGtkOpenGLDrawingArea m_Viewport;

	// Event handlers ...
	void OnEvent(sdpGtkEvent* Event);
	
	void OnDelete(sdpGtkEvent* Event);
	void OnDestroy();
	void OnClose();
	void OnSetFocus();
	void OnRender();
	void OnDefaultView();
	void OnAxisView(SDPCSTRING Axis);
	void OnConfigureEvent(sdpGtkEvent* Event);
	void OnExposeEvent(sdpGtkEvent* Event);

	// k3dMouseInput overrides ...
	void OnMouseMove(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	void OnMButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnMButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnMButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnMButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnMButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	void OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	void OnLRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
};

///////////////////////////////////////////////////////////
// CCamera

class CCamera :	public k3dICamera,
							public k3dTransformObject,
							public k3dLocationServer
{
friend class CEditorEngine;
friend class COpenGLEngine;
friend class CRiEngine;
friend class CCameraDialog;
public:
	CCamera();
	virtual ~CCamera();

	CCameraDialog m_UserInterface;
	CCameraFrame m_Frame;

	// Initialization ...
	bool Create();

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIObject overrides
	sdpResult SetObjectContext(k3dIObjectContext* Context);
	sdpResult EditObject();
	sdpResult QueryDeleteObject();
	sdpResult ObjectRenamed(SDPCSTRING NewName);
	sdpResult ObjectDeleted();

	// k3dIAnimate overrides ...
	sdpResult PostAnimate();

	// k3dICamera overrides ...
	sdpResult ShowViewport();
	sdpResult HideViewport();
	gulong Width();
	gulong Height();
	sdpResult Redraw(bool Synchronous);
	sdpResult StartJob(k3dIRenderJob* Job);
	sdpResult QueueFrame(k3dIRenderFrame* Frame);
	sdpResult FinishJob(k3dIRenderJob* Job);
	sdpResult SnapshotCameraCoordinates();
	sdpResult SnapshotScreenCoordinates();
	sdpResult GetWorldPosition(gdouble Coords[3]);
	sdpResult GetWorldOrientation(gdouble Angles[3]);
	sdpResult GetViewingFrustum(gdouble* Left, gdouble* Right, gdouble* Top, gdouble* Bottom, gdouble* Near, gdouble* Far);
	sdpResult GetWorldToCameraMatrix(gdouble Matrix[16]);

	// k3dIUserEventClient overrides
	sdpResult OnLButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonDoubleClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnRButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnRButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnRButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnLRButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnLRButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);

	// k3dIPersist overrides
	sdpResult InitNew();

	/// Redraws the viewport
	sdpResult Redraw();
	/// Returns the viewport
	sdpGtkOpenGLDrawingArea& Viewport();

protected:
	// Cache some document interfaces ...
	k3dITimeBase* m_TimeBase;

	// Rendering engines ...
	CEditorEngine m_EditorEngine;
	CRiEngine m_RiEngine;

	/// Enumerates available rendering engine types
	typedef enum
	{
		EDITOR,
		RI,
	} RenderEngine;

	class VariableRenderEngine : public k3dCustomVariable<RenderEngine>
	{
	protected:
		sdpString ConvertToString() { return sdpFormatString("%d", m_Value); }
		bool ConvertFromString(SDPCSTRING String) { m_Value = RenderEngine(atoi(String)); return true; }
	};

	/// Enumerates field-dominance values for interlaced rendering
	typedef enum
	{
		ODDFIELDFIRST,
		EVENFIELDFIRST,
	} Field;

	class VariableField : public k3dCustomVariable<Field>
	{
	protected:
		sdpString ConvertToString() { return sdpFormatString("%d", m_Value); }
		bool ConvertFromString(SDPCSTRING String) { m_Value = Field(atoi(String)); return true; }
	};
	
	VariableRenderEngine m_DisplayMode;
	VariableRenderEngine m_RenderMode;
	k3dVariable<bool> m_CameraAutoUpdate;
	k3dVariable<bool> m_CameraLockSize;
	k3dVariable<bool> m_CameraEnableInteractive;
	k3dVariable<bool> m_CameraIsometric;

	VariableField m_CameraField;
	k3dVariable<gdouble> m_PixelAspectRatio;
	k3dVariable<gdouble> m_CropWindowLeft;
	k3dVariable<gdouble> m_CropWindowTop;
	k3dVariable<gdouble> m_CropWindowRight;
	k3dVariable<gdouble> m_CropWindowBottom;

	// Editor render data ...
	k3dVariable<bool> m_EditorSolid;
	k3dVariable<bool> m_EditorFog;
	k3dVariable<bool> m_EditorSmoothing;
	k3dVariable<bool> m_EditorTexture;
	k3dVariable<bool> m_EditorTwoSided;
	k3dVariable<bool> m_EditorOrientation;
	k3dVariable<bool> m_EditorNormals;
	k3dVariable<gulong> m_EditorWireframePatchDivisions;
	k3dVariable<gulong> m_EditorSolidPatchDivisions;

	// RIB render data ...
	k3dVariable<sdpString> m_RiRenderEngine;
	k3dVariable<gulong> m_RiXSamples;
	k3dVariable<gulong> m_RiYSamples;
	k3dVariable<sdpString> m_RiPixelFilterName;
	k3dVariable<gdouble> m_RiPixelFilterWidth;
	k3dVariable<gdouble> m_RiPixelFilterHeight;
	k3dVariable<bool> m_RiAtmosphere;
	k3dShaderInstance m_RiAtmosphereShader;
	k3dVariable<bool> m_RiDepthOfField;
	k3dVariable<bool> m_RiMotionBlur;
	k3dVariable<bool> m_RiFieldRender;
	k3dVariable<bool> m_RiRadiosity;
	k3dVariable<gulong> m_RiRadiositySteps;
	k3dVariable<gulong> m_RiRadiosityPatchSamples;
	k3dVariable<gdouble> m_RiRadiosityPatchSize;
	k3dVariable<gdouble> m_RiRadiosityElementSize;
	k3dVariable<gdouble> m_RiRadiosityMinimumSize;

	// Owned objects ...
	k3dScalarChannel m_Focallength;
	k3dScalarChannel m_Focalplane;
	k3dScalarChannel m_Fstop;
	k3dScalarChannel m_Shutteropen;
	k3dScalarChannel m_Shutterclose;
	k3dScalarChannel m_Exposure;
	k3dScalarChannel m_Gamma;
	k3dScalarChannel m_Xdirection;
	k3dScalarChannel m_Ydirection;
	k3dScalarChannel m_Zdirection;
	k3dScalarChannel m_Farplane;
//	k3dObjectVariable<k3dIObjectSelector> m_Target;

	// Colors ...
	static gdouble m_EditorBackgroundColor[4];
	static gdouble m_EditorNormalColor[4];
	static gdouble m_EditorSelectorColor[4];
	static gdouble m_EditorSelectedColor[4];
	static gdouble m_EditorUnselectedColor[4];
	static gdouble m_EditorControlPointColor[4];
	static gdouble m_EditorFirstColor[4];
	static gdouble m_EditorActiveColor[4];

	// User preferences ...
	static GLfloat m_PointSize;
	static gulong m_ClickSensitivity;
	static gdouble m_PositionSensitivity;
	static gdouble m_OrientationSensitivity;
	static gdouble m_ZoomSign;
	static bool m_LassoSelectPartial;
	static bool m_BoxSelectPartial;

	// Camera state
	sdpVector3 m_CurrentPosition;							// Current position
	sdpAngleAxis m_CurrentOrientation;				// Current camera orientation
	gdouble m_CurrentFocalLength;							// Current camera focal length (equivalent to nearplane)
	gdouble m_CurrentFarPlane;								// Current farplane distance
	gdouble m_CurrentNearPlane;								// Current nearplane distance
	gdouble m_CurrentFocalPlane;							// Current focus plane distance
	gdouble m_CurrentFStop;										// Current f-stop for depth-of-field effects
	gdouble m_CurrentWidthPlane;							// Current nearplane width / 2
	gdouble m_CurrentHeightPlane;							// Current nearplane height / 2
	sdpMatrix4 m_CurrentWorldToCameraMatrix;	// Stored matrix for converting world coords to camera coords

	// Helper functions ...
	void SetupViewVolume();											// Calculate parameters for creating the view frustum ...

	// Selection related ...
	sdpVector2Array m_LassoPoints;
	sdpRectangle m_RubberBand;

	/// Creates a GC for selection
	GdkGC* SelectionGC(sdpGtkWidget& Widget);
	/// XORs a rubber-band box into a widget
	void RubberBand(sdpGtkWidget& Widget, sdpRectangle Rectangle);

	bool SelectLasso(GdkModifierType Modifiers);
	bool SelectLassoPoints(GdkModifierType Modifiers);
	bool SelectLassoPolygons(GdkModifierType Modifiers);

	bool SelectBox(GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 StartMouse);
	bool SelectBoxPoints(GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 StartMouse);
	bool SelectBoxPolygons(GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 StartMouse);

	bool SelectSinglePoint(GdkModifierType Modifiers, sdpVector2 CurrentMouse);
	bool SelectSinglePolygon(GdkModifierType Modifiers, sdpVector2 CurrentMouse);
	bool SelectSingleMesh(GdkModifierType Modifiers, sdpVector2 CurrentMouse);

	// Our OpenGL context ...
	sdpGtkOpenGLDrawingArea m_OpenGLWidget;
};


#endif // CAMERA_H
