#ifndef CHANNEL_H
#define CHANNEL_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CChannel class which implements a data object that supports k3dIChannel
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dpersistobject.h>
#include <k3ddialog.h>
#include <k3dvariable.h>
#include <k3dmouseinput.h>

#include <k3didependencies.h>
#include <k3dichannel.h>
#include <k3displinechannel.h>
#include <k3displinechannelsegment.h>
#include <k3displinechannelnode.h>

#include <sdpgtkopengldrawingarea.h>
#include <sdpxmlielement.h>

#include <vector>

// Forward class declarations ...
class CChannel;

//////////////////////////////////////////////////////////////////////////
// CChannelNode

/// Encapsulates a single node (a 2-dimensional location and 1, 2, or 3-dimensional value) within a CChannel
class CChannelNode : public k3dISplineChannelNode
{
public:
	CChannelNode();
	CChannelNode(const sdpVector3& Value, const sdpVector2& Coords);
	virtual ~CChannelNode();

	// sdpIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_FAIL; }
	gulong AddRef() { return 1; }
	gulong Release() { return 0; }

	// k3dISplineChannelNode overrides
	sdpResult ValueNode();
	sdpResult SetPosition(gdouble Position[2]);
	sdpResult SetValue1(gdouble Value);
	sdpResult SetValue2(gdouble Value[2]);
	sdpResult SetValue3(gdouble Value[3]);
	sdpResult GetPosition(gdouble Position[2]);
	sdpResult GetValue1(gdouble* Value);
	sdpResult GetValue2(gdouble Value[2]);
	sdpResult GetValue3(gdouble Value[3]);

	/// Returns an indexed portion of the node location
	inline double& operator[](int i) { return m_Coords[i]; };
	/// Overloads the assignment operator to only copy location, not value
	inline CChannelNode& operator=(sdpVector2 Coords) { m_Coords = Coords; return *this; };
	/// Returns the node location
	inline operator sdpVector2&() { return m_Coords; };

	/// Saves the node to a K-3D XML document stream
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Loads the node from a K-3D XML document
	sdpResult Load(sdpXmlIElement* Element);

	/// Stores the node value
	sdpVector3 m_Value;
	/// Stores the node location
	sdpVector2 m_Coords;
	
	/// Stores whether this node is a value node
	bool m_ValueNode;
	/// Stores the selected state of this node
	bool m_Selected;

	static int m_NodeSize;
	static int m_ValueNodeSize;
};

inline bool operator < (const CChannelNode& A, const CChannelNode& B) { return false; };
inline bool operator == (const CChannelNode& A, const CChannelNode& B) { return false; };

//////////////////////////////////////////////////////////////////////////
// CChannelSegment

#define NODESPERSEGMENT	3

/// Encapsulates a CChannel segment (3 nodes forming a Bezier curve endpoint and normal vectors to either side)
class CChannelSegment : public k3dISplineChannelSegment
{
public:
	CChannelSegment();
	CChannelSegment(const sdpVector3& AValue, const sdpVector2& ACoords,
									const sdpVector3& BValue, const sdpVector2& BCoords,
									const sdpVector3& CValue, const sdpVector2& CCoords);

	// sdpIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_FAIL; }
	gulong AddRef() { return 1; }
	gulong Release() { return 0; }

	/// Returns a node by index
	inline CChannelNode& Node(int i) { return m_Nodes[i]; }

	/// Saves the segment to a K-3D XML output stream
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Loads the segment from a K-3D document
	sdpResult Load(sdpXmlIElement* Element);

	/// Stores the individual channel nodes
	CChannelNode m_Nodes[NODESPERSEGMENT];
	/// Stores the selected state of this segment
	bool m_Selected;
};

typedef std::vector<CChannelSegment*> Segments;
typedef Segments::iterator SegmentIterator;
typedef Segments::reverse_iterator ReverseSegmentIterator;

/////////////////////////////////////////////////////////////////////////////
// CChannelDialog

/// The main user interface for a CChannel object
class CChannelDialog : public k3dDialog, public k3dMouseInput
{
public:
	CChannelDialog();

	/// Called to create the UI
	bool Create(CChannel* Object);

	/// Called to force a redraw of the UI
	void RedrawAll();

protected:
	/// Control initialization
	bool InitializeControls();
	/// Control update
	void UpdateControls();
	/// Control cleanup
	void FinalizeControls();
	
	/// Dispatches custom events
	void OnEvent(sdpGtkEvent* Event);

	// k3dMouseInput overrides ...
	void OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnRButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLRButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLRButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLRButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	/// Called when the user moves the vertical scale slider
	void OnVScale();
	/// Called when the user wants to reset the view to its default
	void OnResetView();
	/// Called to straighten the entire channel
	void OnChannelStraighten();
	/// Called to mirror the entire channel
	void OnChannelMirrorX();
	/// Called to mirror the entire channel
	void OnChannelMirrorY();
	
	/// Called to change the channel cycles
	void OnCycles();
	/// Called to change the channel phase
	void OnPhase();
	
	/// Called to straighten a single segment
	void OnSegmentStraighten();

	/// Called to edit the value of a node
	void EditNode(CChannelNode& Node);
	/// Called to reset the scale & origin of the display
	void ResetView();
	/// Called to deselect everything
	void DeselectAll();

	/// Called to draw the curve graph
	void DrawCurve();
	/// Called to draw the color preview
	void DrawColor();
	/// Called to draw a background grid
	void DrawGrid();
	/// Called to draw a preview of the final Bezier curve, including frequency, phase, and modulation effects
	void DrawPreview();
	/// Called to draw the Bezier control curve
	void DrawControlCurve();
	/// Called to draw tangents to the Bezier control curve
	void DrawTangents();
	/// Called to draw handles at all the control points
	void DrawHandles();
	void DrawControlPoint(CChannelNode& Node);
	void DrawTangentPoint(CChannelNode& Node);
	/// Called to draw the current cursor position
	void DrawCursor();
	/// Called to draw the current time position
	void DrawTime();

	/// Converts "screen" coordinates to "world" coordinates
	sdpVector2 World(sdpVector2 ScreenCoords);
	/// Converts "world" coordinates to "screen" coordinates
	sdpVector2 Screen(sdpVector2 WorldCoords);
	/// Returns the maximum extents of the curve
	sdpRectangle Extents();
	/// Returns the maximum extents of the curve nodes (not the normals)
	sdpRectangle NodeExtents();
	/// Tests to see if the given mouse coordinates hit near a channel node
	bool HitTest(sdpVector2 CurrentMouse, SegmentIterator& Segment, gulong& NodeIndex);

	/// Update the position of the grid cursor
	void SetCursorX(gdouble X);
	/// Update the position of the grid cursor
	void SetCursorXY(gdouble X, gdouble Y);

	/// Moves a segment node, handling all constraints recursively
	void MoveNode(const SegmentIterator& Segment, const int Node, gdouble X, gdouble Y);
	/// Moves a segment node, handling all constraints recursively
	void MoveNode(const SegmentIterator& Segment, const int Node, sdpVector2 Coords);
	/// Creates a new control curve segment
	void CreateSegment(sdpVector2 Coords, bool AutoSmooth, SegmentIterator& Result);

	/// Stores the owning object
	CChannel* m_Object;

	/// Stores the OpenGL context for the control curve graph
	sdpGtkOpenGLDrawingArea m_CurveWidget;
	/// Stores the OpenGL context for the color preview bar
	sdpGtkOpenGLDrawingArea m_ColorWidget;

	/// Stores the coordinates of the user interface cursor
	sdpVector2 m_Cursor;

	// Colors ...
	static sdpVector3 m_BackgroundColor;
	static sdpVector3 m_NormalColor;
	static sdpVector3 m_ValueNodeColor;
	static sdpVector3 m_SelectedColor;
	static sdpVector3 m_TangentColor;
	static sdpVector3 m_SelectorColor;
	static sdpVector3 m_CursorColor;
	static sdpVector3 m_TimeColor;
	
	// Frustum
	static sdpRectangle m_Frustum;

	// Zoom / pan parameters ...
	sdpVector2 m_Origin;
	sdpVector2 m_Scale;
	gdouble m_VScale;
};

//////////////////////////////////////////////////////////////////////
// CChannel

/// Encapsulates a data value that can change over time (supports k3dIChannel)
class CChannel : public k3dPersistObject,
								 public k3dIChannel,
								 public k3dISplineChannel
{
friend class CChannelDialog;
public:
	CChannel();
	~CChannel();

	/// An instance of the channel's user interface
	CChannelDialog m_UserInterface;

	// k3dIUnknown
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIChannel overrides
	sdpResult SetDatatype(const k3dIChannel::DataType Data);
	k3dIChannel::DataType GetDatatype();
	sdpResult GetCurveRange(gdouble* Min, gdouble* Max);
	gdouble GetValue1(const gdouble Time, const gdouble MaxError);
	sdpResult GetValue2(const gdouble Time, const gdouble MaxError, double Value[2]);
	sdpResult GetValue3(const gdouble Time, const gdouble MaxError, double Value[3]);

	// k3dISplineChannel overrides ...
	sdpResult SetRange1(const gdouble Min, const gdouble Max);
	sdpResult SetRange2(const gdouble Min[2], const gdouble Max[2]);
	sdpResult SetRange3(const gdouble Min[3], const gdouble Max[3]);
	sdpResult GetRange1(gdouble* Min, gdouble* Max);
	sdpResult GetRange2(gdouble Min[2], gdouble Max[2]);
	sdpResult GetRange3(gdouble Min[3], gdouble Max[3]);
	gulong GetSegmentCount();
	K3DSPLINECHANNELSEGMENTHANDLE GetFirstSegment();
	K3DSPLINECHANNELSEGMENTHANDLE GetNextSegment(const K3DSPLINECHANNELSEGMENTHANDLE Handle);
	k3dISplineChannelSegment* GetSegment(const K3DSPLINECHANNELSEGMENTHANDLE Handle);
	k3dISplineChannelSegment* CreateSegment();
	sdpResult DestroySegment(const k3dISplineChannelSegment* Segment);
	gulong GetNodeCount();
	K3DSPLINECHANNELNODEHANDLE GetFirstNode();
	K3DSPLINECHANNELNODEHANDLE GetNextNode(const K3DSPLINECHANNELNODEHANDLE Handle);
	k3dISplineChannelNode* GetNode(const K3DSPLINECHANNELNODEHANDLE Handle);

	// k3dIObject overrides
//	sdpResult SetObjectContext(k3dIObjectContext* Context);
	sdpResult EditObject();
	sdpResult ObjectRenamed(SDPCSTRING NewName);
	sdpResult ObjectDeleted();
 
	// k3dITimeServer overrides
	sdpResult SetAnimTime(double Time, double DeltaTime);

	// k3dIPersist overrides
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	sdpResult Load(sdpXmlIElement* Element);

	/// Frees existing segments
	void Clear();

	/// Returns the number of scalar values (1, 2, or 3) per node, based on the current data type
	inline gulong GetElementsPerNode();

	/// Called when the channel-as-a-whole has been altered
	void NotifyChanged();
	/// Called while the channel is being interactively altered
	void NotifyTracking();

	class VariableDataType : public k3dCustomVariable<k3dIChannel::DataType>
	{
	protected:
		sdpString ConvertToString() { return sdpFormatString("%d", m_Value); }
		bool ConvertFromString(SDPCSTRING String) { m_Value = k3dIChannel::DataType(atoi(String)); return true; }
	};

	/// Stores the channel's data type
	VariableDataType m_DataType;
	/// Stores the minimum allowable data value
	k3dVariable<sdpVector3> m_MinValue;
	/// Stores the maximum allowable data value
	k3dVariable<sdpVector3> m_MaxValue;
	/// Stores the minimum allowable data coordinates
	k3dVariable<sdpVector2> m_MinCoords;
	/// Stores the maximum allowable data coordinates
	k3dVariable<sdpVector2> m_MaxCoords;
	/// Stores the number of times the channel should repeat during the animation
	k3dVariable<double> m_Cycles;
	/// Stores the phase offset between the channel and the animation
	k3dVariable<double> m_Phase;
//	k3dVariable<sdpString> m_LabelX;
//	k3dVariable<sdpString> m_LabelY;

	/// Stores the channel segments
	Segments m_Segments;

	/// Controls the maximum number of iterations when calculating the current channel value, to prevent infinite loops
	static gulong m_MaxIterations;
	static bool m_AutoUpdate;
};

#endif // CHANNEL_H
