// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CAxes K-3D object, which provides a configurable set of axes to help in visualizing the 3D workspace
		\author Tim Shead (tshead@k-3d.com)
*/

#include "axes.h"

#include <sdpfile.h>
#include <sdpgtkevent.h>

#include <k3diobjectcontext.h>
#include <k3dicamera.h>
#include <k3dmisc.h>

//#define SDPGL_TRACE
#include <sdpgl.h>

double CAxes::m_XAxisColor[4] = {1.0, 0.0, 0.0, 1.0};
double CAxes::m_YAxisColor[4] = {1.0, 1.0, 0.0, 1.0};
double CAxes::m_ZAxisColor[4] = {0.0, 1.0, 0.0, 1.0};
double CAxes::m_GridColor[4] = {0.4, 0.4, 0.4, 1.0};

CAxes::CAxes()
{
	AddVariable(&m_Axes("axes", true));
	AddVariable(&m_XYPlane("xyplane", false));
	AddVariable(&m_YZPlane("yzplane", false));
	AddVariable(&m_XZPlane("xzplane", true));
	AddVariable(&m_Size("size", 20.0));
	AddVariable(&m_GridSize("gridsize", 10));
	AddVariable(&m_Translation("translation", 0.0));
	AddVariable(&m_Rotation("rotation", 0.0));
	AddVariable(&m_Sensitivity("sensitivity", 0.1));
}

CAxes::~CAxes()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CAxes::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITarget>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;											

	return SDP_E_NO_INTERFACE;
}

gulong CAxes::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CAxes::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CAxes::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CAxes::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CAxes::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CAxes::EditorPushRender(k3dIEditorEngine* Engine)
{
	sdpglTrace("CAxes::EditorPushRender()");
	
	k3dEditorRender::EditorPushRender(Engine);

	glMatrixMode(GL_MODELVIEW);
	
	glTranslated(m_Translation()[0], m_Translation()[1], m_Translation()[2]);
	glRotated(m_Rotation()[2], 0.0, 0.0, 1.0);
	glRotated(-m_Rotation()[0], 1.0, 0.0, 0.0);
	glRotated(-m_Rotation()[1], 0.0, 1.0, 0.0);

	return SDP_S_OK;
}

sdpResult CAxes::EditorRender(k3dIEditorEngine* Engine)
{
	sdpglTrace("CAxes::EditorRender()");
	sdpglPushTrace(false);
		
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);

	glEnable(GL_POLYGON_OFFSET_LINE);
	glPolygonOffset(2.0, 2.0);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glDisable(GL_CULL_FACE);

	glLineWidth(1.0f);
	glDisable(GL_LINE_STIPPLE);

	gdouble gridsize = m_Size() * 0.5;
	gdouble cellsize = m_Size() / (double)m_GridSize();

	// Setup grid color
	glColor3dv(m_GridColor);

	// Draw XY plane
	if(m_XYPlane())
		{
			for(gulong x = 0; x < m_GridSize(); x++)
				{
					glBegin(GL_QUAD_STRIP);

					for(gulong y = 0; y <= m_GridSize(); y++)
						{
							glVertex3d(x * cellsize - gridsize, y * cellsize - gridsize, 0.0);
							glVertex3d((x + 1) * cellsize - gridsize, y * cellsize - gridsize, 0.0);
						}

					glEnd();
				}
		}

	// Draw YZ plane
	if(m_YZPlane())
		{
			for(gulong y = 0; y < m_GridSize(); y++)
				{
					glBegin(GL_QUAD_STRIP);

					for(gulong z = 0; z <= m_GridSize(); z++)
						{
							glVertex3d(0.0, y * cellsize - gridsize, z * cellsize - gridsize);
							glVertex3d(0.0, (y + 1) * cellsize - gridsize, z * cellsize - gridsize);
						}

					glEnd();
				}
		}

	// Draw XZ plane
	if(m_XZPlane())
		{
			for(gulong x = 0; x < m_GridSize(); x++)
				{
					glBegin(GL_QUAD_STRIP);

					for(gulong z = 0; z <= m_GridSize(); z++)
						{
							glVertex3d(x * cellsize - gridsize, 0.0, z * cellsize - gridsize);
							glVertex3d((x + 1) * cellsize - gridsize, 0.0, z * cellsize - gridsize);
						}

					glEnd();
				}
		}

	glDisable(GL_POLYGON_OFFSET_LINE);

	// Draw axes and labels
	if(m_Axes())
		{
			glBegin(GL_LINES);

			// Draw X axis
			glColor3dv(m_XAxisColor);
			glVertex3d(-gridsize, 0.0, 0.0);
			glVertex3d(gridsize, 0.0, 0.0);

			// Draw Y axis
			glColor3dv(m_YAxisColor);
			glVertex3d(0.0, -gridsize, 0.0);
			glVertex3d(0.0, gridsize, 0.0);

			// Draw Z axis
			glColor3dv(m_ZAxisColor);
			glVertex3d(0.0, 0.0, -gridsize);
			glVertex3d(0.0, 0.0, gridsize);

			glEnd();

			// Draw axis labels ...
/*
			static const GLubyte space[] = 
			{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
*/
			static const GLubyte letters[][13] = {
			{0x00, 0x00, 0xc3, 0xc3, 0xc3, 0xc3, 0xff, 0xc3, 0xc3, 0xc3, 0x66, 0x3c, 0x18}, 
			{0x00, 0x00, 0xfe, 0xc7, 0xc3, 0xc3, 0xc7, 0xfe, 0xc7, 0xc3, 0xc3, 0xc7, 0xfe}, 
			{0x00, 0x00, 0x7e, 0xe7, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xe7, 0x7e}, 
			{0x00, 0x00, 0xfc, 0xce, 0xc7, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc7, 0xce, 0xfc}, 
			{0x00, 0x00, 0xff, 0xc0, 0xc0, 0xc0, 0xc0, 0xfc, 0xc0, 0xc0, 0xc0, 0xc0, 0xff}, 
			{0x00, 0x00, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xfc, 0xc0, 0xc0, 0xc0, 0xff}, 
			{0x00, 0x00, 0x7e, 0xe7, 0xc3, 0xc3, 0xcf, 0xc0, 0xc0, 0xc0, 0xc0, 0xe7, 0x7e}, 
			{0x00, 0x00, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xff, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3}, 
			{0x00, 0x00, 0x7e, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x7e}, 
			{0x00, 0x00, 0x7c, 0xee, 0xc6, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06}, 
			{0x00, 0x00, 0xc3, 0xc6, 0xcc, 0xd8, 0xf0, 0xe0, 0xf0, 0xd8, 0xcc, 0xc6, 0xc3}, 
			{0x00, 0x00, 0xff, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0}, 
			{0x00, 0x00, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xdb, 0xff, 0xff, 0xe7, 0xc3}, 
			{0x00, 0x00, 0xc7, 0xc7, 0xcf, 0xcf, 0xdf, 0xdb, 0xfb, 0xf3, 0xf3, 0xe3, 0xe3}, 
			{0x00, 0x00, 0x7e, 0xe7, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xe7, 0x7e}, 
			{0x00, 0x00, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xfe, 0xc7, 0xc3, 0xc3, 0xc7, 0xfe}, 
			{0x00, 0x00, 0x3f, 0x6e, 0xdf, 0xdb, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0x66, 0x3c}, 
			{0x00, 0x00, 0xc3, 0xc6, 0xcc, 0xd8, 0xf0, 0xfe, 0xc7, 0xc3, 0xc3, 0xc7, 0xfe}, 
			{0x00, 0x00, 0x7e, 0xe7, 0x03, 0x03, 0x07, 0x7e, 0xe0, 0xc0, 0xc0, 0xe7, 0x7e}, 
			{0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0xff}, 
			{0x00, 0x00, 0x7e, 0xe7, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3}, 
			{0x00, 0x00, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3}, 
			{0x00, 0x00, 0xc3, 0xe7, 0xff, 0xff, 0xdb, 0xdb, 0xc3, 0xc3, 0xc3, 0xc3, 0xc3}, 
			{0x00, 0x00, 0xc3, 0x66, 0x66, 0x3c, 0x3c, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3}, 
			{0x00, 0x00, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3c, 0x3c, 0x66, 0x66, 0xc3}, 
			{0x00, 0x00, 0xff, 0xc0, 0xc0, 0x60, 0x30, 0x7e, 0x0c, 0x06, 0x03, 0x03, 0xff}
			};

			static const GLsizei width = 8;
			static const GLsizei height = 13;
			static const GLfloat xorigin = width * 0.5;
			static const GLfloat yorigin = height * 0.5;

			gdouble labelposition = gridsize * 1.1;

			glColor3dv(m_GridColor);
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			glRasterPos3d(labelposition, 0, 0);
			glBitmap(width, height, xorigin, yorigin, 0, 0, letters['X' - 'A']);
			glRasterPos3d(0, labelposition, 0);
			glBitmap(width, height, xorigin, yorigin, 0, 0, letters['Y' - 'A']);
			glRasterPos3d(0, 0, labelposition);
			glBitmap(width, height, xorigin, yorigin, 0, 0, letters['Z' - 'A']);
		}

	sdpglPopTrace();
	
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dITarget overrides

sdpResult CAxes::GetTargetCoordinates(double Coords[3])
{
	m_Translation().CopyArray(Coords);
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

/////////////////////////////////////////////////////////////////////////////
// CAxesDialog

CAxesDialog::CAxesDialog()
{
	m_Object = 0;
}

bool CAxesDialog::Create(CAxes* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "axes.gtkml", "axes.html"), false);

	return true;
}

void CAxesDialog::UpdateControls()
{
	ToggleButton("axes").SetState(m_Object->m_Axes());
	ToggleButton("xyplane").SetState(m_Object->m_XYPlane());
	ToggleButton("yzplane").SetState(m_Object->m_YZPlane());
	ToggleButton("xzplane").SetState(m_Object->m_XZPlane());
}

void CAxesDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onaxes")
		OnAxes();
	else if(Event->Name() == "onxyplane")
		OnXYPlane();
	else if(Event->Name() == "onyzplane")
		OnYZPlane();
	else if(Event->Name() == "onxzplane")
		OnXZPlane();
	else
		k3dDialog::OnEvent(Event);
}

void CAxesDialog::OnAxes()
{
	m_Object->m_Axes() = ToggleButton("axes").GetState();
	m_Object->OnModify(true);
}

void CAxesDialog::OnXYPlane()
{
	m_Object->m_XYPlane() = ToggleButton("xyplane").GetState();
	m_Object->OnModify(true);
}

void CAxesDialog::OnYZPlane()
{
	m_Object->m_YZPlane() = ToggleButton("yzplane").GetState();
	m_Object->OnModify(true);
}

void CAxesDialog::OnXZPlane()
{
	m_Object->m_XZPlane() = ToggleButton("xzplane").GetState();
	m_Object->OnModify(true);
}

