// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CLight K-3D object, which encapsulates a light source within the 3D workspace
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "light.h"

#include <sdpgl.h>

#include <k3dichannel.h>
#include <k3dmisc.h>
#include <k3drishadersetup.h>

#include <sdpgtkevent.h>
#include <sdpfile.h>

CLight::CLight() : k3dTransformObject()
{
	AddVariable(&m_Type("type", POINT));
	AddVariable(&m_CastShadows("castshadows", false));

//	AddObject(&m_Target("target"));
//	AddObject(&m_Light("light"));
	AddObject(&m_Intensity("intensity"));
	AddObject(&m_Color("color"));

	ReserveLocations(1);
}

CLight::~CLight()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CLight::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEditorLight>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dILight>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dILocationServer>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIRiLight>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITarget>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dITransform>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;											

	return SDP_E_NO_INTERFACE;
}

gulong CLight::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CLight::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CLight::SetObjectContext(k3dIObjectContext* Context)
{
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::SetObjectContext(Context)), SDP_E_FAIL);
	
	// Initialize our shaders ...
	g_return_val_if_fail(m_Shader.Create(Document()), SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult CLight::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CLight::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dTransformObject::ObjectRenamed(NewName);
}

sdpResult CLight::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTransformObject::ObjectDeleted();
}

////////////////////////////////////////////////////////////////
// k3dIEventClient overrides

sdpResult CLight::OnSelectionChanged(const k3dID ObjectID)
{
g_assert_not_reached();
/*
	if(m_Light.ID() == SelectorID)
		{
			GetShaderLightType();
			GetShaderLightIntensity();
			GetShaderLightColor();

			OnModify();
		}
*/
	return SDP_S_OK;
}

sdpResult CLight::OnPosition(const k3dID ObjectID)
{
	OnModify(false);
	return SDP_S_OK;
}

sdpResult CLight::OnTrack(const k3dID ObjectID)
{
	OnModify(false);
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////
// k3dIEditorRender overrides ...

sdpResult CLight::EditorPushRender(k3dIEditorEngine* Engine)
{
	sdpglTrace("CLight::EditorPushRender()");

	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	k3dTransformObject::EditorPushRender(Engine);

	return SDP_S_OK;
}

sdpResult CLight::EditorRender(k3dIEditorEngine* Engine)
{
	sdpglTrace("CLight::EditorRender()");

	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	glColor3d(1.0, 1.0, 1.0);
	glLineWidth(1.0f);
	glDisable(GL_LINE_STIPPLE);

	switch(m_Type())
		{
			case AMBIENT:
				EditorRenderAmbient();
				break;
			case POINT:
				EditorRenderPoint();
				break;
			case DISTANT:
				EditorRenderDistant();
				break;
			case SPOTLIGHT:
				EditorRenderSpotlight();
				break;
			default:
				g_assert_not_reached();	// Unknown type!
		}

	DrawTransformationHandles(1.0);

	return SDP_S_OK;
}

void CLight::EditorRenderAmbient()
{
	sdpglTrace("CLight::EditorRenderAmbient()");

	EditorRenderPoint();
}

void CLight::EditorRenderPoint()
{
	sdpglTrace("CLight::EditorRenderPoint()");

	// Draw a nice "furball"
	glBegin(GL_LINES);

	sdpVector3 coords = m_Locations[0].m_AnimXYZ;

	glVertex3d(coords[0] + 1.0, coords[1], coords[2]);
	glVertex3d(coords[0] - 1.0, coords[1], coords[2]);
	glVertex3d(coords[0], coords[1] + 1.0, coords[2]);
	glVertex3d(coords[0], coords[1] - 1.0, coords[2]);
	glVertex3d(coords[0], coords[1], coords[2] + 1.0);
	glVertex3d(coords[0], coords[1], coords[2] - 1.0);

	glVertex3d(coords[0] + 0.4, coords[1] + 0.4, coords[2] + 0.4);
	glVertex3d(coords[0] - 0.4, coords[1] - 0.4, coords[2] - 0.4);
	glVertex3d(coords[0] - 0.4, coords[1] + 0.4, coords[2] + 0.4);
	glVertex3d(coords[0] + 0.4, coords[1] - 0.4, coords[2] - 0.4);
	glVertex3d(coords[0] + 0.4, coords[1] + 0.4, coords[2] - 0.4);
	glVertex3d(coords[0] - 0.4, coords[1] - 0.4, coords[2] + 0.4);
	glVertex3d(coords[0] - 0.4, coords[1] + 0.4, coords[2] - 0.4);
	glVertex3d(coords[0] + 0.4, coords[1] - 0.4, coords[2] + 0.4);

	glEnd();
}

void CLight::EditorRenderSpotlight()
{
	sdpglTrace("CLight::EditorRenderSpotlight()");

	GLUquadricObj* quadric = gluNewQuadric();
	g_assert(quadric);

	// Draw the cone of illumination ...
//			glScaled(m_CurrentWidthPlane, m_CurrentHeightPlane, m_CurrentNearPlane);
//			glRotated(45.0, 0.0, 0.0, 1.0);
	glScaled(1.0, 1.0, -1.0);
	glTranslated(0.0, 0.0, -1.0);

	gluQuadricDrawStyle(quadric, GLenum(GLU_LINE));
	gluCylinder(quadric, sqrt(2.0), 0.001, 1.0, 8, 1);
	gluDeleteQuadric(quadric);

	// Draw the axis ...
	glBegin(GL_LINES);

	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, 0.0, -5.0);

	glEnd();
}

void CLight::EditorRenderDistant()
{
	sdpglTrace("CLight::EditorRenderDistant()");

	// Draw the "tail" of our arrow ...
	glBegin(GL_LINES);

	glVertex3dv(m_Locations[0].m_AnimXYZ);
	glVertex3d(m_Locations[0].m_AnimXYZ[0], m_Locations[0].m_AnimXYZ[1], m_Locations[0].m_AnimXYZ[2] - 5.0);

	glEnd();

	// Draw the "head" of our arrow as a cone ...
	GLUquadricObj* quadric = gluNewQuadric();
	g_assert(quadric);

	gluQuadricDrawStyle(quadric, GLenum(GLU_LINE));
	gluCylinder(quadric, 0.4, 0.001, 0.8, 8, 1);
	gluDeleteQuadric(quadric);
}

///////////////////////////////////////////////////////////////
// k3dILight overrides


/////////////////////////////////////////////////////////////////
// k3dIEditorLight overrides
sdpResult CLight::EditorLightPreRender(k3dIEditorEngine* Engine)
{
	sdpglTrace("CLight::EditorLightPreRender()");

	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::EditorLightSetup(k3dIEditorEngine* Engine, gulong Light)
{
	sdpglTrace("CLight::EditorLightSetup()");

	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return Setup(Light);
}

sdpResult CLight::Setup(gulong Light)
{
	// We don't render ambient lights!
	if(AMBIENT == m_Type())
		return SDP_S_OK;

	// Setup our own transformation matrix ...
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	k3dPushOpenGLMatrix(k3dGetObjectToWorldMatrix(ObjectContext()));

	// Offset our light index ...
	GLenum light = GLenum(Light + GL_LIGHT0);

	// Enable the light ...
	glEnable(light);

	// Setup intensity / attenuation ...
	gdouble intensity = 1.0 / m_Intensity->GetValue1(m_CurrentTime, m_DeltaTime);
	glLightf(light, GL_CONSTANT_ATTENUATION, 0.0f);
	glLightf(light, GL_LINEAR_ATTENUATION, 0.0f);
	glLightf(light, GL_QUADRATIC_ATTENUATION, GLfloat(intensity));

	// Setup color ...
	sdpVector3 color;
	m_Color->GetValue3(m_CurrentTime, m_DeltaTime, color);

	GLfloat glcolor[4];
	glcolor[0] = (GLfloat)color[0];
	glcolor[1] = (GLfloat)color[1];
	glcolor[2] = (GLfloat)color[2];
	glcolor[3] = 1.0f;

	glLightfv(light, GL_AMBIENT, glcolor);
	glLightfv(light, GL_DIFFUSE, glcolor);
	glLightfv(light, GL_SPECULAR, glcolor);

	// Setup light based on type
	if(DISTANT == m_Type())
		{
			GLfloat gllightpos[] = { 0.0f, 0.0f, -1.0f, 0.0f };
			glLightfv(light, GL_POSITION, gllightpos);
		}
	else if(SPOTLIGHT == m_Type())
		{
			// Set spotlight position
			GLfloat gllightpos[4];
			gllightpos[0] = (GLfloat)m_Locations[0].m_AnimXYZ[0];
			gllightpos[1] = (GLfloat)m_Locations[0].m_AnimXYZ[1];
			gllightpos[2] = (GLfloat)m_Locations[0].m_AnimXYZ[2];
			gllightpos[3] = 1.0f;
			glLightfv(light, GL_POSITION, gllightpos);

			// Set spotlight angle & exponent
//			glLightf(Light, GL_SPOT_CUTOFF, (GLfloat)m_Coneangle->GetValue1(m_CurrentTime, m_DeltaTime));
//			glLightf(Light, GL_SPOT_EXPONENT, (GLfloat)m_Spotlightfalloff->GetValue1(m_CurrentTime, m_DeltaTime));

			const static GLfloat lightdirection[3] = { 0.0f, 0.0f, -1.0f };
			glLightfv(light, GL_SPOT_DIRECTION, lightdirection);
		}
	else if(POINT == m_Type())
		{
			// Set point light position
			GLfloat gllightpos[4];
			gllightpos[0] = (GLfloat)m_Locations[0].m_AnimXYZ[0];
			gllightpos[1] = (GLfloat)m_Locations[0].m_AnimXYZ[1];
			gllightpos[2] = (GLfloat)m_Locations[0].m_AnimXYZ[2];
			gllightpos[3] = 1.0f;
			glLightfv(light, GL_POSITION, gllightpos);

			// Make sure it doesn't become a spotlight
			glLightf(light, GL_SPOT_CUTOFF, 180.0f);
		}

	glPopMatrix();
	return SDP_S_OK;
}

/////////////////////////////////////////////////////
// k3dIRiLight overrides

sdpResult CLight::RiLightStartJob(k3dIRenderJob* Job, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::RiLightPreRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::RiLightSetup(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);
	
	// If the user hasn't selected a shader, we're done ...
	if(0 == m_Shader.ShaderHandle())
		return SDP_S_FALSE;

	// Setup our own transformation matrix ...
	Engine->RiTransformBegin();
	Engine->RiIdentity();
	k3dPushRiMatrix(k3dGetObjectToWorldMatrix(ObjectContext()), Engine);

	// Setup light attributes ...
	k3dRiParameterList parameters;
	parameters.AddString("shadows", m_CastShadows() ? "on" : "off");
	parameters.SetAttributes("light", Engine);

	// Create our light shader helper object ...
	k3dRiLightShader light;
	m_Shader.Initialize(Frame, Engine, light, m_CurrentTime, m_DeltaTime);

	// Setup light "from" position ...
	sdpVector3 lightfrom = m_Locations[0].m_AnimXYZ;
	light.AddPoint(k3dIRiEngine::RI_FROM, lightfrom);

	// Setup light "to" position ...
	sdpVector3 lightto = lightfrom + sdpVector3(0.0, 0.0, 1.0);
	light.AddPoint(k3dIRiEngine::RI_TO, lightto);

	// Fire it all off to the rendering engine ...
	light.SetupLight(Engine);

	// Pop our transformation matrix
	Engine->RiTransformEnd();

	return SDP_S_OK;
}

sdpResult CLight::RiLightFinishJob(k3dIRenderJob* Job, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

SDPCSTRING const CLight::Key_LightType = "LightType";
SDPCSTRING const CLight::Key_Ambient = "Ambient";
SDPCSTRING const CLight::Key_Point = "Point";
SDPCSTRING const CLight::Key_Distant = "Distant";
SDPCSTRING const CLight::Key_Spotlight = "Spotlight";
SDPCSTRING const CLight::Key_LightIntensityParameter = "LightIntensityParameter";
SDPCSTRING const CLight::Key_LightColorParameter = "LightColorParameter";

void CLight::GetShaderLightType()
{
	sdpString lighttype = Key_Point;
/*
	k3dIParameter* parameter = m_Light.Interface()->FindParameter(Key_LightType);
	if(parameter) lighttype = parameter->Value();

	if(lighttype == Key_Ambient)
		{
			m_Type() = AMBIENT;
		}
	else if(lighttype == Key_Point)
		{
			m_Type() = POINT;
		}
	else if(lighttype == Key_Distant)
		{
			m_Type() = DISTANT;
		}
	else if(lighttype == Key_Spotlight)
		{
			m_Type() = SPOTLIGHT;
		}
	else
		{
			g_assert(0);	// Unknown light type - default to "point"
			m_Type() = POINT;
		}
*/
}

void CLight::GetShaderLightIntensity()
{
/*
	// Release any existing references to intensity ...
	m_Intensity.Release();

	// See if the shader specifies which parameter controls intensity ...
	sdpString lightintensity;
	k3dIParameter* parameter = m_Light.Interface()->FindParameter(Key_LightIntensityParameter);

	// If not, look for a parameter named "intensity" ...
	lightintensity = parameter ? parameter->Value() : "intensity";

	// Find the parameter ...
	for(k3dIShaderSelectorArgument* argument = m_Light->GetFirstArgument(); argument; argument = m_Light->GetNextArgument(argument))
		{
			if(lightintensity == argument->Name())
				{
					m_Intensity.Attach((k3dIUnknown*)argument->GetValue());
					return;
				}
		}

	OnModify();
*/
}

void CLight::GetShaderLightColor()
{
/*
	// Release any existing references to color ...
	m_Color.Release();

	// See if the shader specifies which parameter controls color ...
	sdpString lightcolor;
	k3dIParameter* parameter = m_Light.Interface()->FindParameter(Key_LightColorParameter);

	// If not, look for a parameter named "lightcolor" ...
	lightcolor = parameter ? parameter->Value() : "lightcolor";

	// Find the parameter ...
	for(k3dIShaderSelectorArgument* argument = m_Light->GetFirstArgument(); argument; argument = m_Light->GetNextArgument(argument))
		{
			if(lightcolor == argument->Name())
				{
					m_Color.Attach((k3dIUnknown*)argument->GetValue());
					return;
				}
		}

	OnModify();
*/
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CLight::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::InitNew()), SDP_E_FAIL);

	m_Shader.SetShader("k3d_pointlight");
	
//	g_return_val_if_fail(m_Target.CreateObjectSelector(Document(), "Light target", IID_k3dITarget, true, true, true, true, this, m_ObjectContext->ObjectID())), SDP_E_FAIL);
//	g_return_val_if_fail(m_Light.CreateShaderSelector(Document(), "Light shader", "pointlight", k3dIShaderInfo::SHADER_LIGHT, KGUID_NULL, false, true, this, m_ObjectContext->ObjectID())), SDP_E_FAIL);
	g_return_val_if_fail(m_Intensity.Create(Document(), "Light intensity", 1200.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Intensity", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Color.Create(Document(), "Light color", sdpVector3(1.0), true, "Time", "Color", true, this), SDP_E_FAIL);

	GetShaderLightType();
	GetShaderLightIntensity();
	GetShaderLightColor();

	return SDP_S_OK;
}

sdpResult CLight::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	g_return_val_if_fail(sdpSucceeded(m_Shader.Save(Element, Dependencies, "shader")), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::Save(Element, Dependencies)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dLocationServer::Save(Element, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element, SDP_E_FAIL);

	g_return_val_if_fail(sdpSucceeded(m_Shader.Load(Element, "shader")), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::Load(Element)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dLocationServer::Load(Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CLight::LoadComplete()
{
	// Call the base-class version ...
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::LoadComplete()), SDP_E_FAIL);

	// Give the shader a chance to initialize itself ...
	g_return_val_if_fail(sdpSucceeded(m_Shader.LoadComplete()), SDP_E_FAIL);

	GetShaderLightType();
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CLightDialog

CLightDialog::CLightDialog()
{
	m_Object = 0;
}

bool CLightDialog::Create(CLight* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "light.gtkml", "light.html"), false);

	return true;
}

bool CLightDialog::InitializeControls()
{
	// Setup shader selector ...
	m_ShaderSelector.Create(m_Object->Application(), &m_Object->m_Shader, &m_ShaderPage, false, k3dIShaderInfo::LIGHT, Container("shaderselector"));
	// Setup the shader arguments ...
	m_ShaderPage.Create(m_Object->Document(), &m_Object->m_Shader, Container("shadercontainer"));

	return true;
}

void CLightDialog::UpdateControls()
{
	// Setup editor options ...
	ToggleButton("castshadows").SetState(m_Object->m_CastShadows());
}

void CLightDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "oncastshadows")
		OnCastShadows();
	else
		k3dDialog::OnEvent(Event);
}

void CLightDialog::OnCastShadows()
{
	m_Object->m_CastShadows() = ToggleButton("castshadows").GetState();
	m_Object->OnModify(false);
}


