// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the Material K-3D object, which controls the surface appearance of rendered geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "material.h"

#include <k3dichannel.h>
#include <k3dimaterialserver.h>
#include <k3drishadersetup.h>
#include <k3diobjectcontext.h>
#include <k3dmisc.h>

#include <sdpfile.h>
#include <sdpgl.h>


////////////////////////////////////////////////////////////////////////////////////
// CMaterial

GLfloat CMaterial::m_BlackData[4] = {0.0f, 0.0f, 0.0f, 1.0f};
GLfloat CMaterial::m_WhiteData[4] = {1.0f, 1.0f, 1.0f, 1.0f};


CMaterial::CMaterial() : k3dPersistObject()
{
	AddVariable(&m_ShadowType("shadowtype", ST_OS));
	AddVariable(&m_TrueDisplacement("truedisplacement", false));
	AddVariable(&m_Matte("matte", false));
	AddVariable(&m_CameraRays("camerarays", true));
	AddVariable(&m_ReflectionRays("reflectionrays", true));

	AddObject(&m_Color("color"));
	AddObject(&m_Opacity("opacity"));
	AddObject(&m_LuminousColor("luminouscolor"));
	AddObject(&m_Luminosity("luminosity"));

	m_EditorTexture = 0;
	m_RiTexture = 0;
}

CMaterial::~CMaterial()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CMaterial::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEditorMaterial>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIMaterial>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIRiMaterial>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CMaterial::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CMaterial::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CMaterial::SetObjectContext(k3dIObjectContext* Context)
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::SetObjectContext(Context)), SDP_E_FAIL);
	
	// Initialize our shaders ...
	g_return_val_if_fail(m_SurfaceShader.Create(Document()), SDP_E_FAIL);
	g_return_val_if_fail(m_DisplacementShader.Create(Document()), SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult CMaterial::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CMaterial::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CMaterial::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

////////////////////////////////////////////////////////////////////////////
// k3dISelectorClient overrides

sdpResult CMaterial::OnSelectionChanged(const k3dID ObjectID)
{
g_assert_not_reached();
/*
	// Release any existing textures ...
	m_TextureSelector.Release();

	m_OpenGLTexture = 0;
	m_RiTexture = 0;

	// Figure out whether we're using a texture or not ...
	for(k3dIShaderSelectorArgument* argument = m_Surface->GetFirstArgument(); argument; argument = m_Surface->GetNextArgument(argument))
		{
			if(k3dIShaderInfoArgument::ARGUMENT_STRING != argument->Type())
				continue;

			m_TextureSelector.Attach((k3dIUnknown*)argument->GetValue());
			break;
		}

	// Update the display
	OnModify();
*/
	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMaterial overrides

sdpResult CMaterial::SetAmbient(double Color[3])
{
g_assert_not_reached();
	return SDP_S_OK;
}

sdpResult CMaterial::SetDiffuse(double Color[3])
{
g_assert_not_reached();
	return SDP_S_OK;
}

sdpResult CMaterial::SetSpecular(double Color[3])
{
g_assert_not_reached();
	return SDP_S_OK;
}

sdpResult CMaterial::SetEmissive(double Color[3])
{
g_assert_not_reached();
	return SDP_S_OK;
}

sdpResult CMaterial::SetSpecularSize(double Size)
{
g_assert_not_reached();
	return SDP_S_OK;
}

sdpResult CMaterial::SetTransparency(double Transparency)
{
g_assert_not_reached();
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////
// k3dIEditorMaterial overrides

sdpResult CMaterial::EditorMaterialPreRender(k3dIEditorEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	if(SDP_S_OK != Engine->EditorQueryTextured())
		return SDP_S_OK;

	k3dITexture* texture = m_SurfaceShader.GetPrimaryTexture();
	m_EditorTexture = texture ? sdpQueryInterface<k3dIEditorTexture>(texture) : 0;

	return SDP_S_OK;
}

sdpResult CMaterial::EditorMaterialSetup(k3dIEditorEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);

	glDepthMask(GL_TRUE);
	glDisable(GL_BLEND);

	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, m_BlackData);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, m_DiffuseData);
	glColor4fv(m_DiffuseData);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, m_WhiteData);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, (GLfloat)64.0);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, m_EmissiveData);

	if((SDP_S_OK == Engine->EditorQueryTextured()) && m_EditorTexture)
		m_EditorTexture->EditorTextureSetup(Engine);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////
// k3dIRiMaterial overrides

sdpResult CMaterial::RiMaterialStartJob(k3dIRenderJob* Job, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialPreRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialSetup(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	// If the user hasn't selected a shader, use NULL ...
	if(0 == m_SurfaceShader.ShaderHandle())
		{
			Engine->RiSurfaceV("null", 0, 	0, 0);			
			return SDP_S_FALSE;
		}
		
	// Setup the surface shader ...
	RiMaterialDefaultsSetup(Engine);
	k3dRiSurfaceShader surface;
	m_SurfaceShader.Initialize(Frame, Engine, surface, m_CurrentTime, m_DeltaTime);
	surface.Setup(Engine);

	// Setup the displacement shader (only if the user has selected a shader) ...
	if(m_DisplacementShader.ShaderHandle())
		{
			k3dRiDisplacementShader displacement;
			m_DisplacementShader.Initialize(Frame, Engine, displacement, m_CurrentTime, m_DeltaTime);
			displacement.Setup(Engine);
		}
		
	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialFinishJob(k3dIRenderJob* Job, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Job, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::RiMaterialDefaultsSetup(k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	// Set "render" attributes ...
	k3dRiParameterList render;

	// Set visibility parameters
	int visibility = 0;
	if(m_CameraRays())
		visibility |= 1;

	if(m_ReflectionRays())
		visibility |= 2;

	if(ST_NONE != m_ShadowType())
		visibility |= 4;

	render.AddInteger("visibility", visibility);

	// Set shadow-casting parameters
	switch(m_ShadowType())
		{
			case ST_NONE:
				render.AddString("casts_shadows", "none");
				break;
			case ST_OPAQUE:
				render.AddString("casts_shadows", "opaque");
				break;
			case ST_SHADE:
				render.AddString("casts_shadows", "Os");
				break;
			case ST_OS:
				render.AddString("casts_shadows", "shade");
				break;
			default:
				g_assert(0);	// Should never happen
		}

	// Enable / disable true displacement
	gint truedisplacement = m_TrueDisplacement() ? 1 : 0;
	render.AddInteger("truedisplacement", truedisplacement);
	render.SetAttributes("render", Engine);

	// Set "radiosity" attributes ...
	k3dRiParameterList radiosity;

	// Set the radiant color
	sdpVector3 color;
	m_LuminousColor->GetValue3(m_CurrentTime, m_DeltaTime, color);
	color *= m_Luminosity->GetValue1(m_CurrentTime, m_DeltaTime);
	radiosity.AddColor("emissioncolor", color);

	radiosity.SetAttributes("radiosity", Engine);


	// Set the "truedisplacement" attributes
	if(truedisplacement)
		{
			k3dRiParameterList displacement;

/* We definitely shouldn't be hard-coding this, but what's a program to do!? */
			double displacementbound = 1.0;
			displacement.AddString("coordinatesystem", "world");
			displacement.AddScalar("sphere", displacementbound);

			displacement.SetAttributes("displacementbound", Engine);
		}

	// Set base color
	m_Color->GetValue3(m_CurrentTime, m_DeltaTime, color);
	Engine->RiColor(color);

	// Set opacity
	sdpVector3 opacity = m_Opacity->GetValue1(m_CurrentTime, m_DeltaTime);
	Engine->RiOpacity(opacity);

	// Set the matte attribute
	Engine->RiMatte(m_Matte() ? 1 : 0);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

sdpResult CMaterial::PostAnimate()
{
	// Cache some current values for speed ...
	sdpVector3 color;

	m_Color->GetValue3(m_CurrentTime, m_DeltaTime, color);
	m_DiffuseData[0] = (GLfloat)color[0];
	m_DiffuseData[1] = (GLfloat)color[1];
	m_DiffuseData[2] = (GLfloat)color[2];
	m_DiffuseData[3] = (GLfloat)m_Opacity->GetValue1(m_CurrentTime, m_DeltaTime);

	m_LuminousColor->GetValue3(m_CurrentTime, m_DeltaTime, color);
	m_EmissiveData[0] = (GLfloat)color[0];
	m_EmissiveData[1] = (GLfloat)color[1];
	m_EmissiveData[2] = (GLfloat)color[2];
	m_EmissiveData[3] = 1.0f;

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CMaterial::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::InitNew()), SDP_E_FAIL);

	m_SurfaceShader.SetShader("k3d_plastic");
//	m_DisplacementShader.SetShader("");

	g_return_val_if_fail(m_Color.Create(Document(), "Material diffuse color", sdpVector3(1.0), true, "Time", "Diffuse color", true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Opacity.Create(Document(), "Material opacity", 1.0, true, 0.0, 1.0, "Time", "Opacity", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_LuminousColor.Create(Document(), "Material luminous color", sdpVector3(0.0), true, "Time", "Luminous color", true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Luminosity.Create(Document(), "Material luminous intensity", 0.5, true, 0.0, 1.0, "Time", "Luminous intensity", k3dIChannel::DATA_SCALAR, true, this), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	g_return_val_if_fail(sdpSucceeded(m_SurfaceShader.Save(Element, Dependencies, "surface")), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_DisplacementShader.Save(Element, Dependencies, "displacement")), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::Save(Element, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element, SDP_E_FAIL);

	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::Load(Element)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_SurfaceShader.Load(Element, "surface")), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_DisplacementShader.Load(Element, "displacement")), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMaterial::LoadComplete()
{
	// Call the base-class version ...
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::LoadComplete()), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_SurfaceShader.LoadComplete()), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_DisplacementShader.LoadComplete()), SDP_E_FAIL);
	
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CMaterialDialog

CMaterialDialog::CMaterialDialog()
{
	m_Object = 0;
}

CMaterialDialog::~CMaterialDialog()
{
	if(Root())
		RootWidget().Destroy();
}

bool CMaterialDialog::Create(CMaterial* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "material.gtkml", "material.html"), false);

	return true;
}

bool CMaterialDialog::InitializeControls()
{
	// Setup shader selectors ...
	m_SurfaceSelector.Create(m_Object->Application(), &m_Object->m_SurfaceShader, &m_SurfacePage, true, k3dIShaderInfo::SURFACE, Container("surfaceselector"));
	m_DisplacementSelector.Create(m_Object->Application(), &m_Object->m_DisplacementShader, &m_DisplacementPage, true, k3dIShaderInfo::DISPLACEMENT, Container("displacementselector"));

	// Setup the shader interfaces ...
	m_SurfacePage.Create(m_Object->Document(), &m_Object->m_SurfaceShader, Container("surfacecontainer"));
	m_DisplacementPage.Create(m_Object->Document(), &m_Object->m_DisplacementShader, Container("displacementcontainer"));

	// Setup channel buttons ...
	if(ChannelButton("color"))
		ChannelButton("color")->Connect(m_Object->m_Color.Interface());

	if(ChannelButton("opacity"))
		ChannelButton("opacity")->Connect(m_Object->m_Opacity.Interface());

	if(ChannelButton("luminouscolor"))
		ChannelButton("luminouscolor")->Connect(m_Object->m_LuminousColor.Interface());

	if(ChannelButton("luminosity"))
		ChannelButton("luminosity")->Connect(m_Object->m_Luminosity.Interface());

	// Initialize the "shadowtype" combo box ...
	sdpStringArray shadowtypes;
	shadowtypes.push_back("None");		// ST_NONE
	shadowtypes.push_back("Opaque");	// ST_OPAQUE
	shadowtypes.push_back("Opacity");	// ST_OS
	shadowtypes.push_back("Shaded");	// ST_SHADE

	sdpGtkCombo shadowtype = Combo("shadowtype");
	shadowtype.SetList(shadowtypes);

	return true;
}

void CMaterialDialog::UpdateControls()
{
	ToggleButton("truedisplacement").SetState(m_Object->m_TrueDisplacement());
	ToggleButton("matte").SetState(m_Object->m_Matte());
	ToggleButton("camerarays").SetState(m_Object->m_CameraRays());
	ToggleButton("reflectionrays").SetState(m_Object->m_ReflectionRays());
}

void CMaterialDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onshadowtype")
		OnShadowType();
	else if(Event->Name() == "ontruedisplacement")
		OnTrueDisplacement();
	else if(Event->Name() == "onmatte")
		OnMatte();
	else if(Event->Name() == "oncamerarays")
		OnCameraRays();
	else if(Event->Name() == "onreflectionrays")
		OnReflectionRays();
	else if(Event->Name() == "onapply")
		OnApply();
	else if(Event->Name() == "onremove")
		OnRemove();
	else if(Event->Name() == "onselect")
		OnSelect();
	else
		k3dDialog::OnEvent(Event);
}

void CMaterialDialog::OnShadowType()
{
}

void CMaterialDialog::OnTrueDisplacement()
{
	m_Object->m_TrueDisplacement() = ToggleButton("truedisplacement").GetState();
	m_Object->OnModify(true);
}

void CMaterialDialog::OnMatte()
{
	m_Object->m_Matte() = ToggleButton("matte").GetState();
	m_Object->OnModify(true);
}

void CMaterialDialog::OnCameraRays()
{
	m_Object->m_CameraRays() = ToggleButton("camerarays").GetState();
	m_Object->OnModify(true);
}

void CMaterialDialog::OnReflectionRays()
{
	m_Object->m_ReflectionRays() = ToggleButton("reflectionrays").GetState();
	m_Object->OnModify(true);
}

/// Functor object for applying a material to polygons / objects
class CMaterialApplyMaterial
{
public:
	CMaterialApplyMaterial(k3dIMaterial* Material) { m_Material = Material; }

	bool HandlePolygon(k3dIPolygon* Polygon)
	{
		Polygon->Material()->SetMaterial(m_Material);
		return true;
	}
	
	bool HandleObject(k3dIObjectContext* ObjectContext)
	{
		k3dIMaterialServer* materialserver = sdpQueryInterface<k3dIMaterialServer>(ObjectContext->Object());
		if(materialserver)
			{
				materialserver->SetMaterial(m_Material);
				materialserver->Release();
			}
			
		return true;
	}

protected:
	k3dIMaterial* m_Material;
};

void CMaterialDialog::OnApply()
{
	// Apply the material to selected polygons / objects ...
	CMaterialApplyMaterial functor(m_Object);
	
	k3dForEachPolygon(m_Object->Document(), k3dIState::VISIBLESELECTED, functor);
	k3dForEachObject(m_Object->Document(), k3dIState::VISIBLESELECTED, functor);
	k3dRedrawAll(m_Object->Document(), false);
}

/// Functor object for resetting polygons / objects to the default material
class CMaterialRemoveMaterial
{
public:
	bool HandlePolygon(k3dIPolygon* Polygon)
	{
		Polygon->Material()->SetMaterial(0);
		return true;
	}
	
	bool HandleObject(k3dIObjectContext* ObjectContext)
	{
		k3dIMaterialServer* materialserver = sdpQueryInterface<k3dIMaterialServer>(ObjectContext->Object());
		if(materialserver)
			{
				materialserver->SetMaterial(0);
				materialserver->Release();
			}
			
		return true;
	}
};

void CMaterialDialog::OnRemove()
{
	// Reset selected polygons / objects to the default material ...
	CMaterialRemoveMaterial functor;
	
	k3dForEachPolygon(m_Object->Document(), k3dIState::VISIBLESELECTED, functor);
	k3dForEachObject(m_Object->Document(), k3dIState::VISIBLESELECTED, functor);
	k3dRedrawAll(m_Object->Document(), false);
}

/// Functor object for selecting polygons / objects that match the given material
class CMaterialSelectMaterial
{
public:
	CMaterialSelectMaterial(k3dIMaterial* Material) { m_Material = Material; }

	bool HandlePolygon(k3dIPolygon* Polygon)
	{
		// Skip polygons that don't match the material ...
		if(Polygon->Material()->GetMaterial() != m_Material)
			return true;

		// Select & show the polygon ...
		Polygon->State()->SetState(k3dIState::VISIBLESELECTED, true);

		return true;
	}

	bool HandleObject(k3dIObjectContext* ObjectContext)
	{
		k3dIMaterialServer* materialserver = sdpQueryInterface<k3dIMaterialServer>(ObjectContext->Object());
		if(materialserver)
			{
				if(materialserver->GetMaterial() == m_Material)
					ObjectContext->State()->SetState(k3dIState::VISIBLESELECTED, true);
					
				materialserver->Release();
			}
			
		return true;
	}

	k3dIMaterial* m_Material;
};

void CMaterialDialog::OnSelect()
{
	// Select polygons / objects that use this material ...
	CMaterialSelectMaterial functor(m_Object);
	
	k3dForEachPolygon(m_Object->Document(), 0, functor);
	k3dForEachObject(m_Object->Document(), 0, functor);
	k3dRedrawAll(m_Object->Document(), false);
}

