#ifndef SOFTOBJECT_H
#define SOFTOBJECT_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CSoftObject K-3D object, which generates implicit surfaces
		\author Romain Behar (romainbehar@yahoo.com)
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3ddialog.h>
#include <k3deditorrender.h>
#include <k3dtrianglemesh.h>
#include <k3dimaterial.h>
#include <k3ditarget.h>
#include <k3dvariable.h>

extern const sdpCLSID sdpSoftObject;

// Forward class references
class CSoftObject;

/////////////////////////////////////////////////////////////////////////////
// CSoftObjectDialog

/// Provides the user interface for the SoftObject tool
class CSoftObjectDialog : public k3dDialog
{
friend class CSoftObject;
public:
	CSoftObjectDialog();

	/// Called to create the UI
	bool Create(CSoftObject* Object);

protected:
	/// Stores the pointer to the owning SoftObject object
	CSoftObject* m_Object;

	/// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void OnVoxelSize();
	void OnMarchingCube();
	void OnTetrahedral();
	void OnRadius();
	void OnHardness();
	void OnPointSkeleton();
	void OnWireFrameSkeleton();
	void OnMeshSkeleton();
	void OnBlobPotential();
	void OnInfiniteSOPotential();
	void OnFiniteSOPotential();

	void OnUpdate();

	/// Update controls ...
	void UpdateControls();
};

/////////////////////////////////////////////////////////////////////////////
// CSoftObject

class CSoftObject : public k3dTriangleMesh
{
friend class CSoftObjectDialog;
public:
	CSoftObject();
	~CSoftObject();

	CSoftObjectDialog m_UserInterface;

	// k3dTool overrides ...
	void OnModify(bool SaveAction);

	static gdouble RawSurfaceFunction(const sdpVector3& Point, gpointer UserData);
	gdouble SurfaceFunction(const sdpVector3& Point);

	// k3dIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIObject overrides
	sdpResult EditObject();
	sdpResult ObjectRenamed(SDPCSTRING NewName);
	sdpResult ObjectDeleted();

	// k3dIPersist overrides ...
	sdpResult InitNew();
	sdpResult LoadComplete();

	// Soft object parameters ...

	typedef enum
	{
		MARCHINGCUBE,
		TETRAHEDRAL
	} PolygonizationType;

	gdouble m_VoxelSize;
	PolygonizationType m_Algorithm;

	typedef enum
	{
		SETOFPOINTS,
		WIREFRAME,
		MESH
	} SkeletonType;

	gdouble m_Radius;
	gdouble m_Hardness;
	SkeletonType m_SkeletonType;

	typedef enum
	{
		BLOB,
		INFINITESOFTOBJECT,
		FINITESOFTOBJECT
	} PotentialType;

	PotentialType m_PotentialType;

protected:
	sdpVector3Array m_ControlPoints;
	std::vector<sdpVector3Array> m_ControlSticks;
};


#endif // SOFTOBJECT_H
