#ifndef SOFTOBJECTFORMULAE_H
#define SOFTOBJECTFORMULAE_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares functions to compute and polygonize soft objects
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "polygonizer.h"

#include <math.h>

#include <iostream>
#include <vector>

#include <sdptypes.h>

/////////////////////////////////////////////////////////////////////////////
// Soft object potential formulae

// Blinn object -> 'Blob'
gdouble BlobPotential(gdouble d, gdouble p)
{
	return 0.5*exp(p*(1-4*d*d));
}

// Kacic-Alesic and Wyvill function behaviour, by Carole Blanc and Christophe Schlick
gdouble FiniteSoftObjectPotential(gdouble d, gdouble p)
{
	if(d >= 1.0)
		return 0;

	gdouble d2 = d*d;
	if(d < 0.5)
		return 1 - 9*d2*d2/(p+(4.5-4*p)*d2);

	return (1-d2)*(1-d2)/(0.75-p+(1.5+4*p)*d2);
}

// Kacic-Alesic and Wyvill function behaviour, by Carole Blanc and Christophe Schlick
gdouble InfiniteSoftObjectPotential(gdouble d, gdouble p)
{
	gdouble d2 = d*d;
	if(d < 0.5)
		return 1 - 1/(2+p*(1-4*d2));

	return 1/(2+p*(-1+4*d2));
}


/////////////////////////////////////////////////////////////////////////////
// Soft object potential field formulae

class PotentialField : public Polygonizer
{
public:
	PotentialField(const sdpVector3Array& ControlPoints, gdouble Radius, gdouble Hardness, gdouble (*PotentialFunction)(gdouble, gdouble))
		{
			m_ControlPoints = ControlPoints;
			m_Radius = Radius;
			m_Hardness = Hardness;
			m_PotentialFunction = PotentialFunction;
		}
	virtual ~PotentialField() {}

	virtual gdouble FieldFunction(const sdpVector3& Point) = 0;

protected:

	sdpVector3Array m_ControlPoints;
	gdouble m_Radius;
	gdouble m_Hardness;
	gdouble (*m_PotentialFunction)(gdouble, gdouble);
};

class PointSetPotentialField : public PotentialField
{
public:
	PointSetPotentialField(const sdpVector3Array& ControlPoints, gdouble Radius, gdouble Hardness, gdouble (*PotentialFunction)(gdouble, gdouble)) :
		PotentialField(ControlPoints, Radius, Hardness, PotentialFunction) {}

	gdouble FieldFunction(const sdpVector3& Point)
		{
			gdouble value = 0;

//std::cout << m_ControlPoints.size() << std::endl;

			for(gulong i = 0; i < m_ControlPoints.size(); i++)
				//if(Point != m_ControlPoints[i])
				value += m_PotentialFunction((Point - m_ControlPoints[i]).Length()/m_Radius, m_Hardness);

			// functions are normalized in the Wyvill's form : f(0.5) = 0.5
			return value - 0.5;
		}

};


/*
gdouble CSoftObject::SegmentDistance(const sdpVector3& s1, const sdpVector3& s2, const sdpVector3& p)
{
	sdpVector3 diff = s2 - s1;

	gdouble pos = diff*(p - s1);
	// check whether p is outside segment, then nearest point is s1
	if(pos <= 0.0)
		return (s1 - p).Length();
		//return DistanceFunction(s1, p);

	gdouble sLength2 = diff.Length2();
	// check whether p is outside segment, then nearest point is s2
	if(pos >= sLength2)
		return (s2 - p).Length();
		//return DistanceFunction(s2, p);

	pos /= sLength2;
	sdpVector3 nearest = s1 + diff*pos;
	return (p - nearest).Length();
	//return DistanceFunction(nearest, p);
}
*/

/*gdouble CSoftObject::SurfaceFunction2(const sdpVector3& Point)
{
	for each edge compute SegmentDistance(edge, Point);
	MeshDistance = Min(previous for each);
		value = PotentialFunction(MeshDistance/m_Radius, m_Hardness);

	return value - 0.5;
}
*/


#endif // SOFTOBJECTFORMULAE_H
