// A C++ Implicit Surface Polygonizer
// Copyright 2000, Romain Behar <romainbehar@yahoo.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements Polygonizer, an implicit surface polygonizer
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <sdpcompiler.h>
#include "polygonizer.h"
#include <algorithm>
#include <assert.h>

// Number of iterations (convergence)
const int RES = 10;


// Directions
const int L = 0;	// left direction:	-x, -i
const int R = 1;	// right direction:	+x, +i
const int B = 2;	// bottom direction:	-y, -j
const int T = 3;	// top direction:	+y, +j
const int N = 4;	// near direction:	-z, -k
const int F = 5;	// far direction:	+z, +k


// The LBN corner of cube (i, j, k), corresponds with location
// (start.x+(i-.5)*size, start.y+(j-.5)*size, start.z+(k-.5)*size)
typedef enum
{
	LBN,	// left bottom near
	LBF,	// left bottom far
	LTN,	// left top near
	LTF,	// left top far
	RBN,	// right bottom near
	RBF,	// right bottom far
	RTN,	// right top near
	RTF	// right top far
} CornerType;


void Polygonizer::Init(double size, int xmin, int xmax, int ymin, int ymax, int zmin, int zmax, bool tetrahedral)
{
	centers.reserve(HashSize);
	corners.reserve(HashSize);
	edges.reserve(HashSize*2);

	voxel_size = size;

	// Lattice bounds
	nxmin = xmin; nxmax = xmax;
	nymin = ymin; nymax = ymax;
	nzmin = zmin; nzmax = zmax;

	// Init polygonizer algorithm
	if(tetrahedral)
		{
			m_PolygonizerDecomposition = TETRAHEDRAL;
		}
	else
		{
			MakeCubeTable();
			m_PolygonizerDecomposition = MARCHING_CUBE;
		}
}


// Polygonize the implicit surface function
//
//	double Function(sdpVector3)
//	return negative for inside, positive for outside
//	double size
//		width of the partitioning cube
//	int bounds
//		max. range of cubes (+/- on the three axes) from first cube
//	sdpVector3 starting_point
//		starting point on or near the surface
//	bool tetrahedral
//		tetrahedral decomposition, else Marching Cube
//
//	returns false if no surface could be found
//
//	triangles :
//		vertices are CCW when viewed from the out (positive) side
//		in a left-handed coordinate system, vertex normals point outwards

bool Polygonizer::Polygonize(const sdpVector3& starting_point, std::vector<sdpVector3> *save_vertices, std::vector<triangle> *save_triangles)
{
	// Can we find first point ?
	if(!FirstPoint(starting_point))
		return false;

	vertices = save_vertices;
	triangles = save_triangles;

	// Reset hash tables
	for(int i = 0; i < HashSize; i++)
		centers[i].clear();
	for(int j = 0; j < HashSize; j++)
		corners[j].clear();
	for(int k = 0; k < HashSize*2; k++ )
		edges[k].clear();

	// Push initial cube on stack
	cube c;

	c.l = location(0, 0, 0);

	for(int n = 0; n < 8; n++)
		c.corners[n] = SetCorner(location(Bit(n, 2), Bit(n, 1), Bit(n, 0)));

	cubes.push(c);
	SetCenter(location(0, 0, 0));


	// Process active cubes till none left
	while(cubes.size() > 0)
		{
			cube c = cubes.top();
			cubes.pop();

			// Polygonize
			switch(m_PolygonizerDecomposition)
				{
					case TETRAHEDRAL: TetrahedralDecomposition(c); break;
					case MARCHING_CUBE: MarchingCube(c); break;
					default:
						g_assert_not_reached();
				}

			// Test six face directions, maybe add to stack
			TestFace(c.l.i-1, c.l.j, c.l.k, c, L, LBN, LBF, LTN, LTF);
			TestFace(c.l.i+1, c.l.j, c.l.k, c, R, RBN, RBF, RTN, RTF);
			TestFace(c.l.i, c.l.j-1, c.l.k, c, B, LBN, LBF, RBN, RBF);
			TestFace(c.l.i, c.l.j+1, c.l.k, c, T, LTN, LTF, RTN, RTF);
			TestFace(c.l.i, c.l.j, c.l.k-1, c, N, LBN, LTN, RBN, RTN);
			TestFace(c.l.i, c.l.j, c.l.k+1, c, F, LBF, LTF, RBF, RTF);
		}

	return true;
}



// Find our first surface point

bool Polygonizer::FirstPoint(const sdpVector3& skeletonPoint)
{
	// Assuming the first point is always a soft object's skeleton,
	// expanding outwards on a straight line will cross the surface
	const gdouble precision = 0.00001;

	sdpVector3 previousWalk;

	sdpVector3 walker = skeletonPoint;
	gdouble walkerValue = FieldFunction(walker);
	bool sign = (walkerValue > 0);

	for(gdouble i = 0; i < 10000; i++)	// In case my assumptions were false
		{
			previousWalk = walker;
			walker += sdpVector3(-voxel_size, -voxel_size, -voxel_size);

			// Have we crossed the surface ?
			walkerValue = FieldFunction(walker);
			if((walkerValue > 0) != sign)
				{
					// Now get closer
					if(walkerValue < 0)
						std::swap(previousWalk, walker);

					sdpVector3 middle;
					gdouble middleValue;
					do
						{
							middle = 0.5 * (previousWalk + walker);
							middleValue = FieldFunction(middle);
							if(middleValue < 0)
								previousWalk = middle;
							else
								walker = middle;
						}
					while(fabs(middleValue) > precision);

					start = middle;

					return true;
				}
		}

	// Couldnt find the surface ??
	return false;
}


// From two points of differing sign, converge to zero crossing

void Polygonizer::Converge(const sdpVector3& p1, const sdpVector3& p2, double v1, sdpVector3& p)
{
	sdpVector3 pos = p1;
	sdpVector3 neg = p2;

	if(v1 < 0)
		std::swap(pos, neg);

	p = 0.5 * (pos + neg);

	for(int iter = 0; iter < RES; iter++)
		{
			if(FieldFunction(p) > 0.0)
				pos = p;
			else
				neg = p;

			p = 0.5 * (pos + neg);
		}
}



//**** Tetrahedral Polygonization ****

void Polygonizer::TetrahedralDecomposition(const cube& c)
{
	// Decompose into tetrahedra and polygonize
	DoTet(c, LBN, LTN, RBN, LBF);
	DoTet(c, RTN, LTN, LBF, RBN);
	DoTet(c, RTN, LTN, LTF, LBF);
	DoTet(c, RTN, RBN, LBF, RBF);
	DoTet(c, RTN, LBF, LTF, RBF);
	DoTet(c, RTN, LTF, RTF, RBF);
}


// Triangulate the tetrahedron (b, c, d should appear clockwise when viewed from a)

void Polygonizer::DoTet(const cube& cube1, int c1, int c2, int c3, int c4)
{
	corner a = cube1.corners[c1];
	corner b = cube1.corners[c2];
	corner c = cube1.corners[c3];
	corner d = cube1.corners[c4];

	int index = 0;
	bool apos, bpos, cpos, dpos;
	if(apos = (a.value > 0.0))
		index += 8;
	if(bpos = (b.value > 0.0))
		index += 4;
	if(cpos = (c.value > 0.0))
		index += 2;
	if(dpos = (d.value > 0.0))
		index += 1;

	// Index is now 4-bit number representing one of the 16 possible cases
	int e1, e2, e3, e4, e5, e6;
	if(apos != bpos)
		e1 = VerticeId(a, b);
	if(apos != cpos)
		e2 = VerticeId(a, c);
	if(apos != dpos)
		e3 = VerticeId(a, d);
	if(bpos != cpos)
		e4 = VerticeId(b, c);
	if(bpos != dpos)
		e5 = VerticeId(b, d);
	if(cpos != dpos)
		e6 = VerticeId(c, d);

	// 14 productive tetrahedral cases (0000 and 1111 do not yield polygons)
	switch(index)
		{
			case 1: Triangle(e5, e6, e3); break;
			case 2: Triangle(e2, e6, e4); break;
			case 3:	Triangle(e3, e5, e4); Triangle(e3, e4, e2); break;
			case 4: Triangle(e1, e4, e5); break;
			case 5: Triangle(e3, e1, e4); Triangle(e3, e4, e6); break;
			case 6: Triangle(e1, e2, e6); Triangle(e1, e6, e5); break;
			case 7: Triangle(e1, e2, e3); break;
			case 8: Triangle(e1, e3, e2); break;
			case 9: Triangle(e1, e5, e6); Triangle(e1, e6, e2); break;
			case 10: Triangle(e1, e3, e6); Triangle(e1, e6, e4); break;
			case 11: Triangle(e1, e5, e4); break;
			case 12: Triangle(e3, e2, e4); Triangle(e3, e4, e5); break;
			case 13: Triangle(e6, e2, e4); break;
			case 14: Triangle(e5, e3, e6); break;
		}
}


//**** Cubical Polygonization ****


const int LB = 0;	// left bottom edge
const int LT = 1;	// left top edge
const int LN = 2;	// left near edge
const int LF = 3;	// left far edge
const int RB = 4;	// right bottom edge
const int RT = 5;	// right top edge
const int RN = 6;	// right near edge
const int RF = 7;	// right far edge
const int BN = 8;	// bottom near edge
const int BF = 9;	// bottom far edge
const int TN = 10;	// top near edge
const int TF = 11;	// top far edge

// edge:                 LB, LT, LN, LF, RB, RT, RN, RF, BN, BF, TN, TF
const int corner1[12] = {LBN,LTN,LBN,LBF,RBN,RTN,RBN,RBF,LBN,LBF,LTN,LTF};
const int corner2[12] = {LBF,LTF,LTN,LTF,RBF,RTF,RTN,RTF,RBN,RBF,RTN,RTF};
// face on left when going corner1 to corner2
const int leftface[12] = {B,  L,  L,  F,  R,  T,  N,  R,  N,  B,  T,  F};
// face on right when going corner1 to corner2
const int rightface[12] = {L,  T,  N,  L,  B,  R,  R,  F,  B,  F,  N,  T};


// Triangulate the cube directly, without decomposition

void Polygonizer::MarchingCube(const cube& cube1)
{
	int index = 0;
	for(int k = 0; k < 8; k++)
		if(cube1.corners[k].value > 0.0)
			index += 1 << k;

	for(unsigned int i = 0; i < cubetable[index].size(); i++)
		{
			int a = -1, b = -1, count = 0;

			for(unsigned int j = 0; j < cubetable[index][i].size(); j++)
				{
					corner c1 = cube1.corners[corner1[cubetable[index][i][j]]];
					corner c2 = cube1.corners[corner2[cubetable[index][i][j]]];
					int c = VerticeId(c1, c2);

					if(++count > 2)
						Triangle(a, b, c);

					if (count < 3)
						a = b;
					b = c;
				}
		}
}


// Return next clockwise edge from given edge around given face

inline int Polygonizer::NextCWedge(int edge, int face)
{
	switch(edge)
		{
			case LB: return (face == L) ? LF : BN;
			case LT: return (face == L) ? LN : TF;
			case LN: return (face == L) ? LB : TN;
			case LF: return (face == L) ? LT : BF;
			case RB: return (face == R) ? RN : BF;
			case RT: return (face == R) ? RF : TN;
			case RN: return (face == R) ? RT : BN;
			case RF: return (face == R) ? RB : TF;
			case BN: return (face == B) ? RB : LN;
			case BF: return (face == B) ? LB : RF;
			case TN: return (face == T) ? LT : RN;
			case TF: return (face == T) ? RT : LF;
			default: assert(false);
		}
	return -1;
}


// Return face adjoining edge that is not the given face

inline int Polygonizer::OtherFace(int edge, int face)
{
	int other = leftface[edge];
	return face == other ? rightface[edge] : other;
}


// Create the 256 entry table for cubical polygonization

void Polygonizer::MakeCubeTable()
{
	for(int i = 0; i < 256; i++)
		{
			int e, done[12], pos[8];
			std::vector< std::vector<int> > lists;

			for(int c = 0; c < 8; c++) pos[c] = Bit(i, c);

			for(e = 0; e < 12; e++) done[e] = 0;
			for(e = 0; e < 12; e++)
				if(!done[e] && (pos[corner1[e]] != pos[corner2[e]]))
					{
						std::vector<int> ints;
						int start = e, edge = e;

						// Get face that is to right of edge from pos to neg corner
						int face = pos[corner1[e]]? rightface[e] : leftface[e];

						for(;;)
							{
							edge = NextCWedge(edge, face);
							done[edge] = 1;

							if(pos[corner1[edge]] != pos[corner2[edge]])
								{
									ints.push_back(edge);
									if(edge == start) break;
									face = OtherFace(edge, face);
								}
							}

						lists.push_back(ints);
					}

			cubetable.push_back(lists);
		}

	for(unsigned int k = 0; k < cubetable.size(); k++)
		{
			std::reverse(cubetable[k].begin(), cubetable[k].end());

			for(unsigned int j = 0; j < cubetable[k].size(); j++)
				{
					std::reverse(cubetable[k][j].begin(), cubetable[k][j].end());
				}
		}
}


//**** Storage ****


// Given cube at lattice (i, j, k), and four corners of face,
// if surface crosses face, compute other four corners of adjacent cube
// and add new cube to cube stack

void Polygonizer::TestFace(int i, int j, int k, cube& old, int face, int c1, int c2, int c3, int c4)
{
	// Test if no surface crossing, cube out of bounds, or already visited

	bool pos = old.corners[c1].value > 0.0;
	if((old.corners[c2].value > 0) == pos && (old.corners[c3].value > 0) == pos && (old.corners[c4].value > 0) == pos)
		return;

	if(i < nxmin || i > nxmax || j < nymin || j > nymax || k < nzmin || k > nzmax)
		return;

	if(SetCenter(location(i, j, k)))
		return;


	// Create new cube and add it to top of stack

	const int facebit[6] = {2, 2, 1, 1, 0, 0};
	cube newc;

	newc.l = location(i, j, k);

	int bit = facebit[face];
	int fc1 = Flip(c1, bit);
	int fc2 = Flip(c2, bit);
	int fc3 = Flip(c3, bit);
	int fc4 = Flip(c4, bit);
	newc.corners[fc1] = old.corners[c1];
	newc.corners[fc2] = old.corners[c2];
	newc.corners[fc3] = old.corners[c3];
	newc.corners[fc4] = old.corners[c4];

	for(int n = 0; n < 8; n++)
		if(n != fc1 && n != fc2 && n != fc3 && n != fc4)
			newc.corners[n] = SetCorner(location(i + Bit(n, 2), j + Bit(n, 1), k + Bit(n, 0)));

	cubes.push(newc);
}


// Return corner with the given lattice location set (and cache) its function value

corner Polygonizer::SetCorner(const location& l)
{
	corner c;
	int index = HashFunc(l);

	c.l = l;
	c.p = start + voxel_size*sdpVector3((double)l.i - 0.5, (double)l.j - 0.5, (double)l.k - 0.5);

	for(unsigned int n = 0; n < corners[index].size(); n++)
		if(corners[index][n].l == l)
			{
				c.value = corners[index][n].value;
				return c;
			}

	c.value = FieldFunction(c.p);
	corners[index].push_back(c);

	return c;
}


// Return true if already set, otherwise set and return false

bool Polygonizer::SetCenter(const location& l)
{
	int index = HashFunc(l);

	for(unsigned int n = 0; n < centers[index].size(); n++)
		if(centers[index][n] == l)
			return true;

	centers[index].push_back(l);

	return false;
}


// Set vertex id for edge

void Polygonizer::SetEdge(const location& l1, const location& l2, int vid)
{
	int index = HashFunc(l1) + HashFunc(l2);

	edge newel;
	if(l1.i > l2.i || (l1.i == l2.i && (l1.j > l2.j || (l1.j == l2.j && l1.k > l2.k))))
		{
			newel.l1 = l2;
			newel.l2 = l1;
		}
	else
		{
			newel.l1 = l1;
			newel.l2 = l2;
		}
	newel.vid = vid;

	edges[index].push_back(newel);
}


// Return vertex id for edge, return -1 if not set

int Polygonizer::GetEdge(const location& l1, const location& l2)
{
	int index = HashFunc(l1) + HashFunc(l2);

	if(l1.i > l2.i || (l1.i == l2.i && (l1.j > l2.j || (l1.j == l2.j && l1.k > l2.k))))
		{
			for(unsigned int n = 0; n < edges[index].size(); n++)
				{
					if(edges[index][n].l1 == l2 && edges[index][n].l2 == l1)
						return edges[index][n].vid;
				}
		}
	else
		{
			for(unsigned int n = 0; n < edges[index].size(); n++)
				{
					if(edges[index][n].l1 == l1 && edges[index][n].l2 == l2)
						return edges[index][n].vid;
				}
		}

	return -1;
}



//**** triangles ****

void Polygonizer::Triangle(int u, int v, int w)
{
	triangles->push_back(triangle(u, v, w));
}


//**** Vertices ****


// return index for vertex on edge:
// c1->value and c2->value are presumed of different sign
// return saved index if any, else compute vertex and save it

int Polygonizer::VerticeId(const corner& c1, const corner& c2)
{
	int vid = GetEdge(c1.l, c2.l);
	if(vid != -1)
		return vid;	// Previously computed

	sdpVector3 p;
	Converge(c1.p, c2.p, c1.value, p);
	vertices->push_back(p);	// Save vertex

	vid = vertices->size() - 1;
	SetEdge(c1.l, c2.l, vid);

	return vid;
}


// Check whether implicit surface is included in current one, from one of its skeleton point

bool Polygonizer::IsNotPolygonized(const sdpVector3& test_point)
{
	if(FirstPoint(test_point))
		{
			double range = voxel_size * 0.6;	// little more than the half of the voxel size

			for(unsigned int i = 0; i < vertices->size(); i++)
				if(AlmostEqual((*vertices)[i], start, range))
					return false;

			return true;

		}

	// Not polygonizable anyway
	return false;
}



