// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the COBJReader K-3D object, which reads Wavefront .obj files
		\author Tim Shead (tshead@k-3d.com)
*/

#include "objreader.h"
#include <iostream>
#include <fstream>
#include <strstream>

#include <sdpmath.h>
#include <sdpstring.h>

#include <k3dilocation.h>
#include <k3dipoint.h>
#include <k3dipolygon.h>
#include <k3dipolypointserver.h>
#include <k3dmisc.h>

#include <k3dcom.h>

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult COBJReader::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIFileFormat>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIGeometryReadFormat>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong COBJReader::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong COBJReader::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIFileFormat overrides

K3DFILEFORMATHANDLE COBJReader::GetFirstFileFormat()
{
	return reinterpret_cast<K3DFILEFORMATHANDLE>(1);
}

K3DFILEFORMATHANDLE COBJReader::GetNextFileFormat(K3DFILEFORMATHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(1 == reinterpret_cast<gulong>(Handle), 0);

	return reinterpret_cast<K3DFILEFORMATHANDLE>(0);
}

SDPCSTRING COBJReader::GetFileFormat(K3DFILEFORMATHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(1 == reinterpret_cast<gulong>(Handle), "");

	return "Wavefront .obj";
}

sdpResult COBJReader::QueryCanHandle(SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	sdpString filepath = FilePath;

	filepath.makelower();
	if(".obj" == filepath.right(4))
		return SDP_S_OK;

	return SDP_S_FALSE;
}

///////////////////////////////////////////////////////////////
// k3dIGeometryReadFormat overrides

k3dIFileFormat* COBJReader::FileFormat()
{
	return this;
}

sdpResult COBJReader::PreRead(SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult COBJReader::ReadOptions(SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult COBJReader::ReadFile(SDPCSTRING FilePath, SDPCSTRING Format)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, SDP_E_FAIL);
	g_return_val_if_string_fail(Format, SDP_E_FAIL);

	// Try to open the input file ...
	std::ifstream file(FilePath);
	g_return_val_if_fail(file.good(), SDP_E_FAIL);

	// Setup a line buffer ...
	const gulong linebuffersize = 1024;
	char* linebuffer = new char[linebuffersize];
	g_return_val_if_fail(linebuffer, SDP_E_FAIL);

	// Setup our parent mesh ...
	k3dIObject* parentobject = Document()->CreateObject(sdpMesh, false, true, true);
	g_assert(parentobject);
	parentobject->ObjectContext()->State()->SetState(k3dIState::SELECTED, true);
	k3dSetName(parentobject, "Untitled Wavefront .obj Object");

	k3dIMesh* mesh = sdpQueryInterface<k3dIMesh>(parentobject);
	g_assert(mesh);

	std::vector<k3dIPoint*> points;
	std::vector<sdpVector2> texturecoordinates;

	// Travel through the stream ...
	while(!file.eof())
		{
			file.getline(linebuffer, linebuffersize);

			// Skip comments ...
			if(linebuffer[0] == '#')
				continue;

			// If it's a vertex record ...
			if(linebuffer[0] == 'v')
				{
					// Skip normal vertices ...
					if(linebuffer[1] == 'n')
						continue;

					// Skip parameter space vertices ...
					if(linebuffer[1] == 'p')
						continue;

					// If it's a texture vertex ...
					if(linebuffer[1] == 't')
						{
							sdpVector2 coords;
							std::istrstream stream(&linebuffer[2]);
							stream >> coords[0] >> coords[1];

							texturecoordinates.push_back(coords);
						}
					// It's a geometric vertex ...
					else
						{
							// Get the vertex coordinates ...
							sdpVector3 coords;
							std::istrstream stream(&linebuffer[1]);
							stream >> coords[0] >> coords[1] >> coords[2];

							// Create a point ...
							k3dIPoint* point = mesh->CreatePoint();
							g_assert(point);

							point->Location()->SetOldXYZ(coords);
							point->Location()->SetLocalXYZ(coords);

							// Store the point for later ...
							points.push_back(point);
						}
				}
			// If its' a face record ...
			else if(linebuffer[0] == 'f')
				{
					// Create a polygon ...
					k3dIPolygon* polygon = mesh->CreatePolygon();
					g_assert(polygon);

					// Attach points by index ...
					gulong index;
					std::istrstream stream(&linebuffer[1]);
					while(!stream.eof())
						{
							stream >> index;

							// Attach the point to our polygon ...
							polygon->Points()->AddPoint(points[index-1]);
						}
				}
			// If it's a mesh record ...
			else if(linebuffer[0] == 'g')
				{
/*
					// Create a mesh object ...
					k3dIObject* object = Document()->CreateObject(sdpMesh, false, true, true);
					g_assert(object);
					object->ObjectContext()->State()->SetState(k3dIState::SELECTED, true);
					
					mesh = sdpQueryInterface<k3dIMesh>(object);
					g_assert(mesh);
					object->Release();
					
					// Set the mesh name ...
					sdpString meshname(&linebuffer[2]);
					k3dSetName(mesh, meshname);

					// Set the parent object ...
					mesh->Object()->ObjectContext()->SetParent(parentobject->ObjectContext());
*/
				}
			// If it's a material record ...
			else if(0 == strncmp(linebuffer, "usemtl", 6))
				{
					continue;
				}
		}

/*
	CArray<COBJReaderMaterial, COBJReaderMaterial> materialarray;
	COBJReaderMaterial material;

	for(linelength = file.ReadLine(buffer, sizeof(buffer)); linelength; linelength = file.ReadLine(buffer, sizeof(buffer)))
		{
			if(0 == strncmp(&buffer[0], "usemtl", 6))	// Create a material
				{
					material.m_Name.Empty();
					material.m_Name = &buffer[7];

					for(i = 0; i < (gulong)materialarray.GetSize(); i++)
						{
							if(material.m_Name == materialarray.GetAt(i).m_Name)
								{
									material = materialarray.GetAt(i);
									goto endmaterial;
								}
						}

					Document()->CreateObject(sdpMaterial, IID_k3dIObject, FALSE, TRUE, TRUE, (void**)&object);
					hr = object->QueryInterface(IID_k3dIMaterial, (void**)&material.m_k3dIMaterial);
					s.Format("%s %s Material", dlg.m_Name, material.m_Name);
					object->ObjectContext()->SetName(s);
					object->Release();

					materialarray.Add(material);

					endmaterial:;
				}
		}
*/

	delete[] linebuffer;

	return SDP_S_OK;
}

