// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CJavaScriptEngine and related classes, which allow K-3D scripts to be written in the JavaScript language
		\author Tim Shead (tshead@k-3d.com)
*/

#include "javascriptengine.h"

#include <k3diapplication.h>
#include <k3direnderfarm.h>
#include <k3direnderjob.h>
#include <k3direnderframe.h>
#include <k3didocument.h>
#include <k3dicamera.h>
#include <k3diobject.h>
#include <k3diobjectinfo.h>
#include <k3ditransform.h>
#include <k3dimesh.h>
#include <k3dipoint.h>
#include <k3dipolygon.h>
#include <k3dipolypointserver.h>
#include <k3dilocation.h>

#include <k3dcom.h>
#include <k3dmisc.h>

#include <sdpstring.h>

#include <iostream>

///////////////////////////////////////////////////////////////
// Utility functions

JSObject* Convert(JSContext* Context, const sdpVector3& Source)
{
	jsval values[3];
	values[0] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source[0]));
	values[1] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source[1]));
	values[2] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source[2]));

	return JS_NewArrayObject(Context, 3, values);
}

JSObject* Convert(JSContext* Context, const sdpAngleAxis& Source)
{
	jsval values[4];
	values[0] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, sdpDegrees(Source.angle)));
	values[1] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source.axis[0]));
	values[2] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source.axis[1]));
	values[3] = DOUBLE_TO_JSVAL(JS_NewDouble(Context, Source.axis[2]));

	return JS_NewArrayObject(Context, 4, values);
}

sdpVector3 ConvertVector3(JSContext* Context, jsval Value)
{
	sdpVector3 result;
	
	if(JSVAL_IS_OBJECT(Value))
		{
			jsval element;
			
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 0, &element);	JS_ValueToNumber(Context, element, &result[0]);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 1, &element);	JS_ValueToNumber(Context, element, &result[1]);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 2, &element);	JS_ValueToNumber(Context, element, &result[2]);
		}
	
	return result;
}

sdpAngleAxis ConvertAngleAxis(JSContext* Context, jsval Value)
{
	sdpAngleAxis result;
	
	if(JSVAL_IS_OBJECT(Value))
		{
			jsval element;
			
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 0, &element);	JS_ValueToNumber(Context, element, &result.angle);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 1, &element);	JS_ValueToNumber(Context, element, &result.axis[0]);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 2, &element);	JS_ValueToNumber(Context, element, &result.axis[1]);
			JS_GetElement(Context, JSVAL_TO_OBJECT(Value), 3, &element);	JS_ValueToNumber(Context, element, &result.axis[2]);
			
			result.angle = sdpRadians(result.angle);
		}
	
	return result;
}

///////////////////////////////////////////////////////////////
// The embedded Location object

JSObject* CJavaScriptLocation::Create(k3dILocation* Location, JSContext* Context, JSObject* ParentObject)
{
	// Sanity checks ...
	g_assert(Location);
	g_assert(Context);
	g_assert(ParentObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, ParentObject);
	JS_SetPrivate(Context, result, Location);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptLocation::Class = 
{
	"Location", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptLocation::Properties[] =
{
	{ "Old", PROPERTY_OLD, JSPROP_PERMANENT, GetOld, SetOld },
	{ "Local", PROPERTY_OLD, JSPROP_PERMANENT, GetLocal, SetLocal },
	{ "Animated", PROPERTY_OLD, JSPROP_PERMANENT, GetAnimated, SetAnimated },
	{ "World", PROPERTY_OLD, JSPROP_PERMANENT, GetWorld, SetWorld },
	{ "Camera", PROPERTY_OLD, JSPROP_PERMANENT, GetCamera, SetCamera },
	{ "Screen", PROPERTY_OLD, JSPROP_PERMANENT, GetScreen, SetScreen },
	{ "Normal", PROPERTY_OLD, JSPROP_PERMANENT, GetNormal, SetNormal },
	{ "Texture", PROPERTY_OLD, JSPROP_PERMANENT, GetTexture, SetTexture },
	{ "ImplicitTexture", PROPERTY_OLD, JSPROP_PERMANENT, GetImplicitTexture, SetImplicitTexture },
	{0}
};

JSFunctionSpec CJavaScriptLocation::Methods[] = 
{
	{0}
};

k3dILocation* CJavaScriptLocation::Location(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dILocation*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptLocation::GetOld(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	sdpVector3 result;
	Location(Context, Object)->GetOldXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetOld(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetOldXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetLocal(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result;
	Location(Context, Object)->GetLocalXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetLocal(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetLocalXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetAnimated(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result;
	Location(Context, Object)->GetAnimXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetAnimated(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetAnimXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetWorld(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result;
	Location(Context, Object)->GetWorldXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetWorld(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetWorldXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetCamera(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result;
	Location(Context, Object)->GetCameraXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetCamera(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetCameraXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetScreen(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result;
	Location(Context, Object)->GetScreenXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetScreen(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetScreenXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetNormal(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result;
	Location(Context, Object)->GetNormalXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetNormal(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetNormalXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetTexture(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result;
	Location(Context, Object)->GetTextureXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetTexture(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetTextureXYZ(value);
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::GetImplicitTexture(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 result;
	Location(Context, Object)->GetImplicitTextureXYZ(result);
	*Value = OBJECT_TO_JSVAL(Convert(Context, result));
	
	return JS_TRUE;
}

JSBool CJavaScriptLocation::SetImplicitTexture(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)	
{
	sdpVector3 value = ConvertVector3(Context, *Value);
	Location(Context, Object)->SetImplicitTextureXYZ(value);
	
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// The embedded Point object

JSObject* CJavaScriptPoint::Create(k3dIPoint* Point, JSContext* Context, JSObject* MeshObject)
{
	// Sanity checks ...
	g_assert(Point);
	g_assert(Context);
	g_assert(MeshObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, MeshObject);
	JS_SetPrivate(Context, result, Point);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptPoint::Class = 
{
	"Point", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptPoint::Properties[] =
{
	{ "Mesh", PROPERTY_MESH, JSPROP_READONLY | JSPROP_PERMANENT, GetMesh, 0 },
	{ "Location", PROPERTY_LOCATION, JSPROP_READONLY | JSPROP_PERMANENT, GetLocation, 0 },
	{0}
};

JSFunctionSpec CJavaScriptPoint::Methods[] = 
{
//	{ "CreatePoint", CreatePoint, 0, 0, 0 },
	{0}
};

k3dIPoint* CJavaScriptPoint::Point(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIPoint*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptPoint::GetMesh(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptPoint::GetLocation(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(CJavaScriptLocation::Create(Point(Context, Object)->Location(), Context, Object));
	return JS_TRUE;
}

/*
JSBool CJavaScriptPoint::CreatePoint(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dIPoint* point = Mesh(Context, Object)->CreatePoint();
	g_return_val_if_fail(point, JS_FALSE);

	*Value = OBJECT_TO_JSVAL(CJavaScriptPoint::Create(point, Context, Object));
	return JS_TRUE;
}
*/

///////////////////////////////////////////////////////////////
// The embedded PolyPointServer object

JSObject* CJavaScriptPolyPointServer::Create(k3dIPolyPointServer* PolyPointServer, JSContext* Context, JSObject* ParentObject)
{
	// Sanity checks ...
	g_assert(PolyPointServer);
	g_assert(Context);
	g_assert(ParentObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, ParentObject);
	JS_SetPrivate(Context, result, PolyPointServer);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptPolyPointServer::Class = 
{
	"PolyPointServer", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptPolyPointServer::Properties[] =
{
//	{ "Mesh", PROPERTY_MESH, JSPROP_READONLY | JSPROP_PERMANENT, GetMesh, 0 },
//	{ "Points", PROPERTY_LOCATION, JSPROP_PERMANENT, GetPoints, 0 },
	{0}
};

JSFunctionSpec CJavaScriptPolyPointServer::Methods[] = 
{
	{ "AddPoint", AddPoint, 1, 0, 0 },
	{ "RemovePoint", RemovePoint, 1, 0, 0 },
	{0}
};

k3dIPolyPointServer* CJavaScriptPolyPointServer::PolyPointServer(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIPolyPointServer*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

/*
JSBool CJavaScriptPolyPointServer::GetMesh(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptPolyPointServer::GetPoints(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(CJavaScriptPolyPointServer::Create(Polygon(Context, Object)->Points(), Context, Object));
	return JS_TRUE;
}
*/

JSBool CJavaScriptPolyPointServer::AddPoint(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	g_return_val_if_fail(JSVAL_IS_OBJECT(argv[0]), JS_FALSE);

	k3dIPoint* point = CJavaScriptPoint::Point(Context, JSVAL_TO_OBJECT(argv[0]));
	g_return_val_if_fail(point, JS_FALSE);
	
	g_return_val_if_fail(PolyPointServer(Context, Object)->AddPoint(point), JS_FALSE);	
	return JS_TRUE;
}

JSBool CJavaScriptPolyPointServer::RemovePoint(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	g_return_val_if_fail(JSVAL_IS_OBJECT(argv[0]), JS_FALSE);

	k3dIPoint* point = CJavaScriptPoint::Point(Context, JSVAL_TO_OBJECT(argv[0]));
	g_return_val_if_fail(point, JS_FALSE);
	
	g_return_val_if_fail(sdpSucceeded(PolyPointServer(Context, Object)->RemovePoint(point)), JS_FALSE);	
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// The embedded Polygon object

JSObject* CJavaScriptPolygon::Create(k3dIPolygon* Polygon, JSContext* Context, JSObject* MeshObject)
{
	// Sanity checks ...
	g_assert(Polygon);
	g_assert(Context);
	g_assert(MeshObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, MeshObject);
	JS_SetPrivate(Context, result, Polygon);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptPolygon::Class = 
{
	"Polygon", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptPolygon::Properties[] =
{
	{ "Mesh", PROPERTY_MESH, JSPROP_READONLY | JSPROP_PERMANENT, GetMesh, 0 },
	{ "Points", PROPERTY_POINTS, JSPROP_READONLY | JSPROP_PERMANENT, GetPoints, 0 },
	{0}
};

JSFunctionSpec CJavaScriptPolygon::Methods[] = 
{
//	{ "CreatePoint", CreatePoint, 0, 0, 0 },
	{0}
};

k3dIPolygon* CJavaScriptPolygon::Polygon(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIPolygon*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptPolygon::GetMesh(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptPolygon::GetPoints(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(CJavaScriptPolyPointServer::Create(Polygon(Context, Object)->Points(), Context, Object));
	return JS_TRUE;
}

/*
JSBool CJavaScriptPolygon::CreatePoint(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dIPoint* point = Mesh(Context, Object)->CreatePoint();
	g_return_val_if_fail(point, JS_FALSE);

	*Value = OBJECT_TO_JSVAL(CJavaScriptPoint::Create(point, Context, Object));
	return JS_TRUE;
}
*/

///////////////////////////////////////////////////////////////
// The embedded Mesh object

JSObject* CJavaScriptMesh::Create(k3dIMesh* Mesh, JSContext* Context, JSObject* ObjectObject)
{
	// Sanity checks ...
	g_assert(Mesh);
	g_assert(Context);
	g_assert(ObjectObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, ObjectObject);
	JS_SetPrivate(Context, result, Mesh);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptMesh::Class = 
{
	"Mesh", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptMesh::Properties[] =
{
/*
	{ "Document", PROPERTY_DOCUMENT, JSPROP_READONLY | JSPROP_PERMANENT, GetDocument, 0 },
	{ "Name", PROPERTY_NAME, JSPROP_PERMANENT, GetName, SetName },
	{ "Position", PROPERTY_POSITION, JSPROP_PERMANENT, GetPosition, SetPosition },
	{ "Scale", PROPERTY_SCALE, JSPROP_PERMANENT, GetScale, SetScale },
	{ "Rotation", PROPERTY_ROTATION, JSPROP_PERMANENT, GetRotation, SetRotation },
	{ "Mesh", PROPERTY_MESH, JSPROP_READONLY | JSPROP_PERMANENT, GetMesh, 0 },
*/
	{0}
};

JSFunctionSpec CJavaScriptMesh::Methods[] = 
{
	{ "CreatePoint", CreatePoint, 0, 0, 0 },
	{ "CreatePolygon", CreatePolygon, 0, 0, 0 },
	{0}
};

k3dIMesh* CJavaScriptMesh::Mesh(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIMesh*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

/*
JSBool CJavaScriptMesh::GetDocument(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptMesh::GetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, CJavaScriptObject::Object(Context, Object)->ObjectContext()->Name()));
	return JS_TRUE;
}

JSBool CJavaScriptMesh::SetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	CJavaScriptObject::Object(Context, Object)->ObjectContext()->SetName(JS_GetStringBytes(JS_ValueToString(Context, *Value)));
	return JS_TRUE;
}

JSBool CJavaScriptMesh::GetPosition(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpVector3 position;
			transform->GetTranslation(position);
			transform->Release();
			
			*Value = OBJECT_TO_JSVAL(Convert(Context, position));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptMesh::SetPosition(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpVector3 position = ConvertVector3(Context, *Value);
			transform->SetTranslation(position);
			transform->Release();
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptMesh::GetScale(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpVector3 scale;
			transform->GetScale(scale);
			transform->Release();
			
			*Value = OBJECT_TO_JSVAL(Convert(Context, scale));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptMesh::SetScale(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpVector3 scale = ConvertVector3(Context, *Value);
			transform->SetScale(scale);
			transform->Release();
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptMesh::GetRotation(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpAngleAxis rotation;
			transform->GetRotation(rotation);
			transform->Release();
			
			*Value = OBJECT_TO_JSVAL(Convert(Context, rotation));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptMesh::SetRotation(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpAngleAxis rotation = ConvertAngleAxis(Context, *Value);
			transform->SetRotation(rotation);
			transform->Release();
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptMesh::GetMesh(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dIMesh* mesh = sdpQueryInterface<k3dIMesh>(CJavaScriptObject::Object(Context, Object));
	g_return_val_if_fail(mesh, JS_FALSE);
	
	*Value = OBJECT_TO_JSVAL(CJavaScriptMesh::Create(mesh, Context, Object));
	
	return JS_TRUE;
}
*/

JSBool CJavaScriptMesh::CreatePoint(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dIPoint* point = Mesh(Context, Object)->CreatePoint();
	g_return_val_if_fail(point, JS_FALSE);

	*Result = OBJECT_TO_JSVAL(CJavaScriptPoint::Create(point, Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptMesh::CreatePolygon(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dIPolygon* polygon = Mesh(Context, Object)->CreatePolygon();
	g_return_val_if_fail(polygon, JS_FALSE);

	*Result = OBJECT_TO_JSVAL(CJavaScriptPolygon::Create(polygon, Context, Object));
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// The embedded Object object

JSObject* CJavaScriptObject::Create(k3dIObject* Object, JSContext* Context, JSObject* DocumentObject)
{
	// Sanity checks ...
	g_assert(Object);
	g_assert(Context);
	g_assert(DocumentObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, DocumentObject);
	JS_SetPrivate(Context, result, Object);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptObject::Class = 
{
	"Object", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptObject::Properties[] =
{
	{ "Document", PROPERTY_DOCUMENT, JSPROP_READONLY | JSPROP_PERMANENT, GetDocument, 0 },
	{ "Name", PROPERTY_NAME, JSPROP_PERMANENT, GetName, SetName },
	{ "Position", PROPERTY_POSITION, JSPROP_PERMANENT, GetPosition, SetPosition },
	{ "Orientation", PROPERTY_ORIENTATION, JSPROP_PERMANENT, GetOrientation, SetOrientation },
	{ "Scale", PROPERTY_SCALE, JSPROP_PERMANENT, GetScale, SetScale },
	{ "Mesh", PROPERTY_MESH, JSPROP_READONLY | JSPROP_PERMANENT, GetMesh, 0 },
	{0}
};

JSFunctionSpec CJavaScriptObject::Methods[] = 
{
	{ "EditObject", EditObject, 0, 0, 0 },
	{0}
};

k3dIObject* CJavaScriptObject::Object(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIObject*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptObject::GetDocument(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, CJavaScriptObject::Object(Context, Object)->ObjectContext()->Name()));
	return JS_TRUE;
}

JSBool CJavaScriptObject::SetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	CJavaScriptObject::Object(Context, Object)->ObjectContext()->SetName(JS_GetStringBytes(JS_ValueToString(Context, *Value)));
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetPosition(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpVector3 position;
			transform->GetPosition(k3dITransform::CURRENT_FRAME, position);
			transform->Release();
			
			*Value = OBJECT_TO_JSVAL(Convert(Context, position));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::SetPosition(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpVector3 position = ConvertVector3(Context, *Value);
			transform->SetPosition(k3dITransform::CURRENT_FRAME, position);
			transform->Release();
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetOrientation(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpAngleAxis rotation;
			transform->GetOrientation(k3dITransform::CURRENT_FRAME, rotation);
			transform->Release();
			
			*Value = OBJECT_TO_JSVAL(Convert(Context, rotation));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::SetOrientation(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpAngleAxis rotation = ConvertAngleAxis(Context, *Value);
			transform->SetOrientation(k3dITransform::CURRENT_FRAME, rotation);
			transform->Release();
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetScale(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpVector3 scale;
			transform->GetScale(k3dITransform::CURRENT_FRAME, scale);
			transform->Release();
			
			*Value = OBJECT_TO_JSVAL(Convert(Context, scale));
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::SetScale(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(CJavaScriptObject::Object(Context, Object));
	if(transform)
		{
			sdpVector3 scale = ConvertVector3(Context, *Value);
			transform->SetScale(k3dITransform::CURRENT_FRAME, scale);
			transform->Release();
		}
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::GetMesh(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dIMesh* mesh = sdpQueryInterface<k3dIMesh>(CJavaScriptObject::Object(Context, Object));
	g_return_val_if_fail(mesh, JS_FALSE);
	
	*Value = OBJECT_TO_JSVAL(CJavaScriptMesh::Create(mesh, Context, Object));
	
	return JS_TRUE;
}

JSBool CJavaScriptObject::EditObject(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	CJavaScriptObject::Object(Context, Object)->EditObject();
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// The embedded Camera object

JSObject* CJavaScriptCamera::Create(k3dICamera* Camera, JSContext* Context, JSObject* DocumentObject)
{
	// Sanity checks ...
	g_assert(Camera);
	g_assert(Context);
	g_assert(DocumentObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, DocumentObject);
	JS_SetPrivate(Context, result, Camera);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptCamera::Class = 
{
	"Camera", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptCamera::Properties[] =
{
	{ "Document", PROPERTY_DOCUMENT, JSPROP_READONLY | JSPROP_PERMANENT, GetDocument, 0 },
	{0}
};

JSFunctionSpec CJavaScriptCamera::Methods[] = 
{
	{ "Redraw", Redraw, 0, 0, 0 },
	{ "StartJob", StartJob, 1, 0, 0 },
	{ "QueueFrame", QueueFrame, 1, 0, 0 },
	{ "FinishJob", FinishJob, 1, 0, 0 },
	{0}
};

k3dICamera* CJavaScriptCamera::Camera(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dICamera*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptCamera::GetDocument(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptCamera::Redraw(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	Camera(Context, Object)->Redraw(false);
	return JS_TRUE;
}

JSBool CJavaScriptCamera::StartJob(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	g_return_val_if_fail(JSVAL_IS_OBJECT(argv[0]), JS_FALSE);

	k3dIRenderJob* job = CJavaScriptRenderJob::Job(Context, JSVAL_TO_OBJECT(argv[0]));
	g_return_val_if_fail(job, JS_FALSE);

	Camera(Context, Object)->StartJob(job);
	return JS_TRUE;
}

JSBool CJavaScriptCamera::QueueFrame(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	g_return_val_if_fail(JSVAL_IS_OBJECT(argv[0]), JS_FALSE);

	k3dIRenderFrame* frame = CJavaScriptRenderFrame::Frame(Context, JSVAL_TO_OBJECT(argv[0]));
	g_return_val_if_fail(frame, JS_FALSE);

	Camera(Context, Object)->QueueFrame(frame);
	return JS_TRUE;
}

JSBool CJavaScriptCamera::FinishJob(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	g_return_val_if_fail(JSVAL_IS_OBJECT(argv[0]), JS_FALSE);

	k3dIRenderJob* job = CJavaScriptRenderJob::Job(Context, JSVAL_TO_OBJECT(argv[0]));
	g_return_val_if_fail(job, JS_FALSE);

	Camera(Context, Object)->FinishJob(job);
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// The embedded Document object

JSObject* CJavaScriptDocument::Create(k3dIDocument* Document, JSContext* Context, JSObject* ApplicationObject)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Context);
	g_assert(ApplicationObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, ApplicationObject);
	JS_SetPrivate(Context, result, Document);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptDocument::Class = 
{
	"Document", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptDocument::Properties[] =
{
	{ "Application", PROPERTY_APPLICATION, JSPROP_READONLY | JSPROP_PERMANENT, GetApplication, 0 },
	{ "ActiveCamera", PROPERTY_ACTIVECAMERA, JSPROP_READONLY | JSPROP_PERMANENT, GetActiveCamera, 0 },
	{ "EditMode", PROPERTY_EDITMODE, JSPROP_PERMANENT, GetEditMode, SetEditMode },
	{0}
};

JSFunctionSpec CJavaScriptDocument::Methods[] = 
{
	{ "Save", Save, 0, 0, 0 },
	{ "SaveAs", SaveAs, 1, 0, 0 },
	{ "CreateObject", CreateObject, 1, 0, 0 },
	{ "Object", Object, 1, 0, 0 },
	{ "DeleteObject", DeleteObject, 1, 0, 0 },
	{ "RedrawAll", RedrawAll, 0, 0, 0 },
	{0}
};

k3dIDocument* CJavaScriptDocument::Document(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIDocument*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptDocument::GetApplication(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptDocument::GetActiveCamera(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dICamera* camera = Document(Context, Object)->GetActiveCamera();
	if(camera)
		*Value = OBJECT_TO_JSVAL(CJavaScriptCamera::Create(camera, Context, Object));
		
	return JS_TRUE;
}

JSBool CJavaScriptDocument::GetEditMode(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	switch(Document(Context, Object)->GetActiveEditMode())
		{
			case k3dIDocument::POINT:
				*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, "point"));
				break;
			case k3dIDocument::POLYGON:
				*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, "polygon"));
				break;
			case k3dIDocument::MESH:
				*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, "mesh"));
				break;
			default:
				g_return_val_if_fail(0, JS_FALSE);
		}
		
	return JS_TRUE;
}

JSBool CJavaScriptDocument::SetEditMode(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	sdpString mode = JS_GetStringBytes(JS_ValueToString(Context, *Value));
	if(mode == "point")
		Document(Context, Object)->SetActiveEditMode(k3dIDocument::POINT);
	else if(mode == "polygon")
		Document(Context, Object)->SetActiveEditMode(k3dIDocument::POLYGON);
	else if(mode == "mesh")
		Document(Context, Object)->SetActiveEditMode(k3dIDocument::MESH);
	else
		g_return_val_if_fail(0, JS_FALSE);
		
	return JS_TRUE;
}

JSBool CJavaScriptDocument::Save(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	Document(Context, Object)->Save();
	return JS_TRUE;
}

JSBool CJavaScriptDocument::SaveAs(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	Document(Context, Object)->SaveAs(JS_GetStringBytes(JS_ValueToString(Context, argv[0])));
	return JS_TRUE;
}

JSBool CJavaScriptDocument::CreateObject(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dObjectInfoSet objects;

	// Look for object types by name ...
	if(JSVAL_IS_STRING(argv[0]))
		k3dGetObjectInfo(CJavaScriptApplication::Application(Context, JS_GetParent(Context, Object)), JS_GetStringBytes(JS_ValueToString(Context, argv[0])), objects);
	
	// If we got at least one object type, create an instance of it ...
	if(objects.size())
		{
			k3dObjectInfoSet::iterator objectinfo = objects.begin(); 
			k3dIObject* object = Document(Context, Object)->CreateObject((*objectinfo)->ClassID(), false, true, true);
			g_return_val_if_fail(object, JS_FALSE);
			
			*Result = OBJECT_TO_JSVAL(CJavaScriptObject::Create(object, Context, Object));
		}
		
	return JS_TRUE;
}

JSBool CJavaScriptDocument::Object(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dIObject* object = 0;

	// Look for objects by name ...
	if(JSVAL_IS_STRING(argv[0]))
		object = k3dGetObject(Document(Context, Object), JS_GetStringBytes(JS_ValueToString(Context, argv[0])));
	
	// If we found something return it ...
	if(object)
		*Result = OBJECT_TO_JSVAL(CJavaScriptObject::Create(object, Context, Object));
		
	return JS_TRUE;
}

JSBool CJavaScriptDocument::DeleteObject(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	if(JSVAL_IS_OBJECT(argv[0]))
		{
			k3dIObject* object = CJavaScriptObject::Object(Context, JSVAL_TO_OBJECT(argv[0]));
			if(object)
				Document(Context, Object)->DeleteObject(object);
		}

	return JS_TRUE;
}

JSBool CJavaScriptDocument::RedrawAll(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	k3dRedrawAll(Document(Context, Object), false);
	
	return JS_TRUE;
}


///////////////////////////////////////////////////////////////
// The embedded Render Frame object

JSObject* CJavaScriptRenderFrame::Create(k3dIRenderFrame* Frame, JSContext* Context, JSObject* JobObject)
{
	// Sanity checks ...
	g_assert(Frame);
	g_assert(Context);
	g_assert(JobObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, JobObject);
	JS_SetPrivate(Context, result, Frame);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptRenderFrame::Class = 
{
	"RenderFrame", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptRenderFrame::Properties[] =
{
	{ "Job", PROPERTY_JOB, JSPROP_READONLY | JSPROP_PERMANENT, GetJob, 0 },
	{ "Name", PROPERTY_NAME, JSPROP_READONLY | JSPROP_PERMANENT, GetName, 0 },
	{ "Path", PROPERTY_PATH, JSPROP_READONLY | JSPROP_PERMANENT, GetPath, 0 },
	{ "Priority", PROPERTY_PRIORITY, JSPROP_PERMANENT, GetPriority, SetPriority },
	{ "ViewDuringRender", PROPERTY_VIEWDURINGRENDER, JSPROP_PERMANENT, GetViewDuringRender, SetViewDuringRender },
	{ "ViewAfterRender", PROPERTY_VIEWAFTERRENDER, JSPROP_PERMANENT, GetViewAfterRender, SetViewAfterRender },
	{ "Engine", PROPERTY_ENGINE, JSPROP_PERMANENT, GetEngine, SetEngine },
	{0}
};

JSFunctionSpec CJavaScriptRenderFrame::Methods[] = 
{
	{0}
};

k3dIRenderFrame* CJavaScriptRenderFrame::Frame(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIRenderFrame*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptRenderFrame::GetJob(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Frame(Context, Object)->Name()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetPath(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Frame(Context, Object)->Path()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetPriority(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = DOUBLE_TO_JSVAL(JS_NewDouble(Context, gdouble(Frame(Context, Object)->GetPriority())));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::SetPriority(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	gdouble priority = k3dIRenderFrame::MODERATE_PRIORITY;
	
	JS_ValueToNumber(Context, *Value, &priority);
	Frame(Context, Object)->SetPriority(guchar(priority));
	
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetViewDuringRender(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = BOOLEAN_TO_JSVAL(Frame(Context, Object)->GetViewDuringRender());
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::SetViewDuringRender(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	Frame(Context, Object)->SetViewDuringRender(JSVAL_TO_BOOLEAN(*Value) ? true : false);
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetViewAfterRender(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = BOOLEAN_TO_JSVAL(Frame(Context, Object)->GetViewAfterRender());
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::SetViewAfterRender(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	Frame(Context, Object)->SetViewAfterRender(JSVAL_TO_BOOLEAN(*Value) ? true : false);
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::GetEngine(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Frame(Context, Object)->GetEngine()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderFrame::SetEngine(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	Frame(Context, Object)->SetEngine(JS_GetStringBytes(JS_ValueToString(Context, *Value)));
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// The embedded Render Job object

JSObject* CJavaScriptRenderJob::Create(k3dIRenderJob* Job, JSContext* Context, JSObject* ApplicationObject)
{
	// Sanity checks ...
	g_assert(Job);
	g_assert(Context);
	g_assert(ApplicationObject);

	JSObject* result = JS_NewObject(Context, &Class, 0, ApplicationObject);
	JS_SetPrivate(Context, result, Job);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptRenderJob::Class = 
{
	"RenderJob", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptRenderJob::Properties[] =
{
	{ "Application", PROPERTY_APPLICATION, JSPROP_READONLY | JSPROP_PERMANENT, GetApplication, 0 },
	{ "Name", PROPERTY_NAME, JSPROP_READONLY | JSPROP_PERMANENT, GetName, 0 },
	{ "Path", PROPERTY_PATH, JSPROP_READONLY | JSPROP_PERMANENT, GetPath, 0 },
	{0}
};

JSFunctionSpec CJavaScriptRenderJob::Methods[] = 
{
	{ "CreateFrame", CreateFrame, 1, 0, 0 },
	{0}
};

k3dIRenderJob* CJavaScriptRenderJob::Job(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIRenderJob*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptRenderJob::GetApplication(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = OBJECT_TO_JSVAL(JS_GetParent(Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptRenderJob::GetName(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Job(Context, Object)->Name()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderJob::GetPath(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Job(Context, Object)->Path()));
	return JS_TRUE;
}

JSBool CJavaScriptRenderJob::CreateFrame(JSContext* Context, JSObject* Object, uintN argc, jsval* argv, jsval* Result)
{
	sdpString framename;
	
	// Look for object types by name ...
	if(JSVAL_IS_STRING(argv[0]))
		framename = JS_GetStringBytes(JS_ValueToString(Context, argv[0]));
	
	// If we got a name ...
	if(framename.size())
		{
			k3dIRenderFrame* frame = Job(Context, Object)->CreateFrame(framename);
			g_return_val_if_fail(frame, JS_FALSE);
			
			*Result = OBJECT_TO_JSVAL(CJavaScriptRenderFrame::Create(frame, Context, Object));
		}
		
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// The embedded Application object

JSObject* CJavaScriptApplication::Create(k3dIApplication* Application, JSContext* Context, JSObject* GlobalObject)
{
	// Sanity checks ...
	g_assert(Application);
	g_assert(Context);
	g_assert(GlobalObject);

	JSObject* result = JS_DefineObject(Context, GlobalObject, "Application", &Class, 0, 0);
	JS_SetPrivate(Context, result, Application);
	JS_DefineProperties(Context, result, Properties);
	JS_DefineFunctions(Context, result, Methods);
	
	return result;
}

JSClass CJavaScriptApplication::Class = 
{
	"Application", JSCLASS_HAS_PRIVATE, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub
};

JSPropertySpec CJavaScriptApplication::Properties[] =
{
	{ "ActiveDocument", PROPERTY_ACTIVEDOCUMENT, JSPROP_READONLY | JSPROP_PERMANENT, GetActiveDocument, 0 },
	{ "BasePath", PROPERTY_APPLICATIONPATH, JSPROP_READONLY | JSPROP_PERMANENT, GetBasePath, 0 },
	{0}
};

JSFunctionSpec CJavaScriptApplication::Methods[] = 
{
	{ "Close", Close, 0, 0, 0 },

	{ "OpenDocument", OpenDocument, 1, 0, 0 },
	{ "CreateDocument", CreateDocument, 0, 0, 0 },
	{ "CloseDocument", CloseDocument, 1, 0, 0 },

	{ "CreateJob", CreateJob, 1, 0, 0 },
	{ "StartJob", StartJob, 1, 0, 0 },
			
	{ "BrowserNavigate", BrowserNavigate, 1, 0, 0, },
	{ "Message", Message, 2, 0, 0 },
	{ "Query", Query, 2, 0, 0 },
	{ "Error", Error, 2, 0, 0 },
	{0}
};

k3dIApplication* CJavaScriptApplication::Application(JSContext* Context, JSObject* Object)
{
	if(JS_GetClass(Object) == &Class)
		return reinterpret_cast<k3dIApplication*>(JS_GetPrivate(Context, Object));
	else
		return 0;
}

JSBool CJavaScriptApplication::GetActiveDocument(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	k3dIDocument* document = Application(Context, Object)->GetActiveDocument();
	if(document)
		*Value = OBJECT_TO_JSVAL(CJavaScriptDocument::Create(document, Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptApplication::GetBasePath(JSContext* Context, JSObject* Object, jsval ID, jsval* Value)
{
	*Value = STRING_TO_JSVAL(JS_NewStringCopyZ(Context, Application(Context, Object)->BasePath()));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::Close(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	Application(Context, Object)->Close();
	return JS_TRUE;
}

JSBool CJavaScriptApplication::OpenDocument(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	k3dIDocument* document = Application(Context, Object)->OpenDocument(JS_GetStringBytes(JS_ValueToString(Context, argv[0])));
	if(document)
		*Result = OBJECT_TO_JSVAL(CJavaScriptDocument::Create(document, Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptApplication::CreateDocument(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	k3dIDocument* document = Application(Context, Object)->CreateDocument();
	if(document)
		*Result = OBJECT_TO_JSVAL(CJavaScriptDocument::Create(document, Context, Object));

	return JS_TRUE;
}

JSBool CJavaScriptApplication::CloseDocument(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	if(JSVAL_IS_OBJECT(argv[0]))
		{
			k3dIDocument* document = CJavaScriptDocument::Document(Context, JSVAL_TO_OBJECT(argv[0]));
			if(document)
				Application(Context, Object)->CloseDocument(document);
		}

	return JS_TRUE;
}

JSBool CJavaScriptApplication::CreateJob(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	k3dIRenderFarm* renderfarm = sdpQueryInterface<k3dIRenderFarm>(Application(Context, Object));
	g_return_val_if_fail(renderfarm, JS_FALSE);

	k3dIRenderJob* renderjob = renderfarm->CreateJob(JS_GetStringBytes(JS_ValueToString(Context, argv[0])));
	g_return_val_if_fail(renderjob, JS_FALSE);

	*Result = OBJECT_TO_JSVAL(CJavaScriptRenderJob::Create(renderjob, Context, Object));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::StartJob(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval* Result)
{
	k3dIRenderFarm* renderfarm = sdpQueryInterface<k3dIRenderFarm>(Application(Context, Object));
	g_return_val_if_fail(renderfarm, JS_FALSE);

	if(JSVAL_IS_OBJECT(argv[0]))
		{
			k3dIRenderJob* job = CJavaScriptRenderJob::Job(Context, JSVAL_TO_OBJECT(argv[0]));
			g_return_val_if_fail(job, JS_FALSE);
			
			renderfarm->StartJob(job);
		}

	return JS_TRUE;
}

JSBool CJavaScriptApplication::BrowserNavigate(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	Application(Context, Object)->BrowserNavigate(JS_GetStringBytes(JS_ValueToString(Context, argv[0])));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::Message(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	Application(Context, Object)->Message(JS_GetStringBytes(JS_ValueToString(Context, argv[0])), JS_GetStringBytes(JS_ValueToString(Context, argv[1])));
	return JS_TRUE;
}

JSBool CJavaScriptApplication::Query(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	sdpResult result = Application(Context, Object)->Query(JS_GetStringBytes(JS_ValueToString(Context, argv[0])), JS_GetStringBytes(JS_ValueToString(Context, argv[1])));
	*Result = BOOLEAN_TO_JSVAL(result == SDP_S_OK ? true : false);

	return JS_TRUE;
}

JSBool CJavaScriptApplication::Error(JSContext* Context, JSObject* Object, uintN argc, jsval *argv, jsval *Result)
{
	Application(Context, Object)->Error(JS_GetStringBytes(JS_ValueToString(Context, argv[0])), JS_GetStringBytes(JS_ValueToString(Context, argv[1])));
	return JS_TRUE;
}

///////////////////////////////////////////////////////////////
// CJavaScriptEngine

CJavaScriptEngine::CJavaScriptEngine()
{
	m_Application = 0;
	m_ReferenceCount = 0;
	
	m_Runtime = 0;
	m_Context = 0;
	m_GlobalObject = 0;
}

CJavaScriptEngine::~CJavaScriptEngine()
{
	// Shut down the interpreter
	if(m_Context)
		JS_DestroyContext(m_Context);
		
	if(m_Runtime)
		JS_Finish(m_Runtime);
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CJavaScriptEngine::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIScriptEngine>(this, InterfaceID, Interface)) return SDP_S_OK;											
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIScriptEngine*>(this), InterfaceID, Interface)) return SDP_S_OK;											

	return SDP_E_NO_INTERFACE;
}

gulong CJavaScriptEngine::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CJavaScriptEngine::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIScriptEngine overrides
SDPCSTRING CJavaScriptEngine::Language()
{
	return "JavaScript";
}

sdpResult CJavaScriptEngine::SetOwner(k3dIApplication* Application)
{
	// Sanity checks ...
	g_return_val_if_fail(Application, SDP_E_FAIL);
	g_return_val_if_fail(!m_Application, SDP_E_FAIL);

	// Store a reference to the application ...	
	m_Application = Application;

	// Fire up the interpreter ...
	m_Runtime = JS_NewRuntime(0x100000);
	g_return_val_if_fail(m_Runtime, SDP_E_FAIL);
	
	m_Context = JS_NewContext(m_Runtime, 0x1000); 
	g_return_val_if_fail(m_Context, SDP_E_FAIL);

	// Store a reference to ourselves in the interpreter context ...
	JS_SetContextPrivate(m_Context, this);

	// Set the branch callback, so we can cancel scripts if necessary ...
	JS_SetBranchCallback(m_Context, RawBranchCallback);

	// Create the global object ...
	static JSClass global_class = { "global", 0, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_PropertyStub, JS_EnumerateStub, JS_ResolveStub, JS_ConvertStub, JS_FinalizeStub };
	m_GlobalObject = JS_NewObject(m_Context, &global_class, NULL, NULL); 
	g_return_val_if_fail(m_GlobalObject, SDP_E_FAIL);
	
	// Initialize core JS objects ...
	JS_InitStandardClasses(m_Context, m_GlobalObject);

	// Create the application object ...
	CJavaScriptApplication::Create(m_Application, m_Context, m_GlobalObject);

	return SDP_S_OK;
}

sdpResult CJavaScriptEngine::Execute(SDPCSTRING Script)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Script, SDP_E_FAIL);
	
	// Evaluate the script ...
	jsval result;
	JS_EvaluateScript(m_Context, m_GlobalObject, Script, strlen(Script), "test", 0, &result); 
	
	return SDP_S_OK;
}

JSBool CJavaScriptEngine::RawBranchCallback(JSContext* Context, JSScript* Script)
{
	return reinterpret_cast<CJavaScriptEngine*>(JS_GetContextPrivate(Context))->BranchCallback(Context, Script);
}

JSBool CJavaScriptEngine::BranchCallback(JSContext* Context, JSScript* Script)
{
//	std::cout << "CJavaScriptEngine::BranchCallback()" << std::endl;

	return JS_TRUE;
}


