#ifndef DOCUMENT_H
#define DOCUMENT_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CDocument class, which encapsulates an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3didocument.h>
#include <k3ditimebase.h>
#include <k3diactions.h>
#include <k3dactionclient.h>

#include "action.h"

// Forward class declarations
class CApplication;
class CDocumentFrame;
class CHierarchyFrame;
class CLayer;
class CObjectInfo;
class CObjectContext;

class k3dIDependencies;
class sdpXmlIElement;

///////////////////////////////////////////////////////////
// CHierarchyTreeRecord

/// Encapsulates the placement of a CObjectContext within the document hierarchy
class CHierarchyTreeRecord
{
public:
	CHierarchyTreeRecord() {}
	CHierarchyTreeRecord(k3dIDocument::TreeNodeType Type, CObjectContext* Context) { m_Type = Type; m_ObjectContext = Context; }

	/// Stores the record type: PUSH, NODE, or POP
	k3dIDocument::TreeNodeType m_Type;
	/// Stores the parent CObjectContext
	CObjectContext* m_ObjectContext;
};

inline bool operator== (const CHierarchyTreeRecord& A, const CHierarchyTreeRecord& B) { return (A.m_Type == B.m_Type) && (A.m_ObjectContext == B.m_ObjectContext); }
inline bool operator!= (const CHierarchyTreeRecord& A, const CHierarchyTreeRecord& B) { return (A.m_Type != B.m_Type) || (A.m_ObjectContext != B.m_ObjectContext); }

typedef std::vector<CHierarchyTreeRecord> HierarchyTreeList;
typedef HierarchyTreeList::iterator HierarchyTreeIterator;


/////////////////////////////////////////////////////////////////////
// CDocument

/// Stores a collection of document windows
typedef std::list<CDocumentFrame*> DocumentFrameList;
typedef DocumentFrameList::iterator DocumentFrameIterator;

/// Stores a collection of hierarchy windows
typedef std::list<CHierarchyFrame*> HierarchyFrameList;
typedef HierarchyFrameList::iterator HierarchyFrameIterator;

/// Encapsulates an open K-3D document
class CDocument : public k3dIDocument,
									public k3dITimeBase,
									public k3dIActions,
									public k3dActionClient
{
friend class CAnimationParametersDialog;
friend class CDocumentFrame;
friend class CHierarchyFrame;
friend class CHierarchyControl;
friend class CLayer;
friend class CObjectContext;
public:
	CDocument();
	virtual ~CDocument();

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIDocument overrides ...
	k3dIApplication* Application();
	
	sdpResult Import(SDPCSTRING FilePath, SDPCSTRING Format, k3dIGeometryReadFormat* FormatFilter);
	sdpResult Export(SDPCSTRING FilePath, SDPCSTRING Format, k3dIGeometryWriteFormat* FormatFilter);
	sdpResult Save();
	sdpResult SaveAs(SDPCSTRING FilePath);

	k3dIActions* Actions();

	k3dIObject* CreateObject(const sdpCLSID& ClassID, bool Edit, bool UserVisible, bool Saveable);
	sdpResult CreateObject(sdpXmlIElement* Element);	// This isn't part of k3dIDocument, but it fits nicely here :)
	sdpResult DeleteObject(k3dIObject* Object);

	k3dICamera* GetActiveCamera();
	sdpResult SetActiveCamera(k3dICamera* Camera);

	K3DLAYERHANDLE GetFirstLayer();
	K3DLAYERHANDLE GetNextLayer(K3DLAYERHANDLE Layer);
	k3dILayer* GetLayer(K3DLAYERHANDLE Layer);
	k3dILayer* CreateLayer();
	sdpResult DestroyLayer(k3dILayer* Layer);
	sdpResult SetActiveLayer(k3dILayer* Layer);
	k3dILayer* GetActiveLayer();

	K3DOBJECTCONTEXTHANDLE GetFirstObjectContext();
	K3DOBJECTCONTEXTHANDLE GetNextObjectContext(K3DOBJECTCONTEXTHANDLE ObjectContext);
	k3dIObjectContext* GetObjectContext(K3DOBJECTCONTEXTHANDLE ObjectContext);
	sdpResult SetActiveObjectContext(k3dIObjectContext* ObjectContext);
	k3dIObjectContext* GetActiveObjectContext();

	TreeNodeType GetFirstObjectContextTree(K3DTREEHANDLE* Handle);
	TreeNodeType GetNextObjectContextTree(K3DTREEHANDLE* Handle);
	TreeNodeType GetContextHandleTree(k3dIObjectContext* Context, K3DTREEHANDLE* Handle);
	k3dIObjectContext* GetObjectContextTree(K3DTREEHANDLE Handle);

	Axis GetActiveAxis();
	sdpResult SetActiveAxis(Axis ActiveAxis);

	TransformType GetActiveTransformType();
	sdpResult SetActiveTransformType(TransformType ActiveTransformType);

	CoordinateSystem GetActiveCoordinateSystem();
	sdpResult SetActiveCoordinateSystem(CoordinateSystem System);

	EditMode GetActiveEditMode();
	sdpResult SetActiveEditMode(EditMode Mode);

	sdpResult FreezeChanges();
	sdpResult SnapshotWorldCoordinates();
	sdpResult AnimateDocument();

	k3dIMaterial* GetDefaultMaterial();

	// k3dITimeBase overrides ...
	gdouble GetStartTime();
	gulong GetStartFrame();
	gdouble GetViewTime();
	gulong GetViewFrame();
	gdouble GetEndTime();
	gulong GetEndFrame();
	gdouble GetDuration();
	gulong GetTotalFrames();
	gdouble GetFrameRate();
	gdouble GetFrameDelta();
	gdouble GetRenderTime();
	gdouble GetRenderDeltaTime();

	sdpResult SetStartTime(gdouble Time);
	sdpResult SetStartFrame(gulong Frame);
	sdpResult SetViewTime(gdouble Time);
	sdpResult SetViewFrame(gulong Frame);
	sdpResult SetEndTime(gdouble Time);
	sdpResult SetEndFrame(gulong Frame);
	sdpResult SetDuration(gdouble Duration);
	sdpResult SetTotalFrames(gulong Frames);
	sdpResult SetFrameRate(gdouble Rate);
	sdpResult SetRenderTime(gdouble Time);
	sdpResult SetRenderDeltaTime(gdouble Delta);

	// k3dIActions overrides ...
	sdpResult StartAction(SDPCSTRING Name);
	sdpResult IsRecording();
	sdpResult RecordStep(k3dIActionClient* Client, gulong OldValueCode, gulong OldValue, gulong OldValueSize, gulong NewValueCode, gulong NewValue, gulong NewValueSize);
	sdpResult FinishAction();

	gulong ActionCount();
	gulong RedoCount();

	sdpResult UndoAction();
	sdpResult RedoAction();

	// k3dIActionClient overrides ...
	sdpResult HandleStep(gulong ValueCode, gulong Value, gulong ValueSize);

	/// Called to create a new document
	bool Create(CApplication* Application);
	/// Called to open an existing document from an XML file
	bool Open(CApplication* Application, SDPCSTRING DocumentPath);

	/// Returns "true" if the document is in the process of closing itself
	bool IsDying();

	/// Called when the parent application wants to activate this document
	void OnActivate(CApplication* Application);
	/// Called when a document frame gains the focus
	void OnActivate(CDocumentFrame* Frame);
	/// Called when a hierarchy frame gains the focus
	void OnActivate(CHierarchyFrame* Frame);
	/// Called whenever this document is made active
	void OnActivate();

	/// Creates a new document frame
	CDocumentFrame* CreateDocumentFrame();
	/// Creates a new hierarchy frame
	CHierarchyFrame* CreateHierarchyFrame();

	/// Loads a document frame from an XML document
	CDocumentFrame* LoadDocumentFrame(sdpXmlIElement* Element);
	/// Loads a hierarchy frame from an XML document
	CHierarchyFrame* LoadHierarchyFrame(sdpXmlIElement* Element);

	/// Deletes a document frame
	void DeleteFrame(CDocumentFrame* Frame);
	/// Deletes a document frame
	void DeleteFrame(DocumentFrameIterator Frame);
	/// Deletes a hierarchy frame
	void DeleteFrame(CHierarchyFrame* Frame);
	/// Deletes a hierarchy frame
	void DeleteFrame(HierarchyFrameIterator Frame);

	/// Deletes all frames (hierarchy, document)
	void DeleteAllFrames();

	/// Stores the list of layers
	typedef std::vector<CLayer*> LayerList;
	typedef LayerList::iterator LayerIterator;
	LayerList m_Layers;

	/// Creates a new layer
	CLayer* CreateLayer(CDocument* Document);
	/// Returns "true" if it is OK to delete a layer
	bool CanDeleteLayer(CLayer* Layer);
	/// Deletes a layer
	void DeleteLayer(CLayer* Layer);
	/// Deletes a layer
	void DeleteLayer(LayerIterator Layer);
	/// Deletes all layers
	void DeleteAllLayers();
	/// Sets the active layer
	void SetActiveLayer(CLayer* Layer);

	/// Returns a guaranteed-unique object name
	sdpString GetUniqueObjectName(sdpString Name);
	/// Returns a guaranteed-unique object ID
	k3dID GetNextObjectID();

	/// Stores the object hierarchy
	HierarchyTreeList m_Hierarchy;

	/// Returns the full filepath for this document
	SDPCSTRING Path() { return m_DocumentPath; }
	/// Returns the title of this document
	SDPCSTRING Title() { return m_DocumentTitle; }

	/// Stores a pointer to the owning application
	CApplication* m_Application;

	/// Stores the currently active layer
	CLayer* m_ActiveLayer;
	/// Stores the currently active object context
	CObjectContext* m_ActiveObjectContext;
	/// Stores the currently active camera
	k3dICamera* m_ActiveCamera;

	/// Stores the collection of hierarchy frames
	HierarchyFrameList m_HierarchyFrames;
	/// Stores the collection of document frames
	DocumentFrameList m_DocumentFrames;

	/// Enumerates the document's animation playback modes
	typedef enum
	{
		REVERSEPLAY,
		STOP,
		PLAY,
		LOOPPLAY,
	} PlaybackMode;

	/// Returns the current animation playback mode
	PlaybackMode GetPlaybackMode();
	/// Sets the current animation playback mode
	void SetPlaybackMode(PlaybackMode Mode);

protected:
	/// Stores the full document filepath
	sdpString m_DocumentPath;
	/// Stores the document title
	sdpString m_DocumentTitle;
	/// Stores a number for uniquely naming new documents
	static gulong m_DocumentNumber;

	/// Set to "true" if the document is in the process of going-away
	bool m_Dying;

	/// Stores the current animation playback mode
	PlaybackMode m_PlaybackMode;

	/// Stores the current animation start time, as a percentage
	gdouble m_StartTime;
	/// Stores the current animation view time, as a percentage
	gdouble m_ViewTime;
	/// Stores the current animation end time, as a percentage
	gdouble m_EndTime;
	/// Stores the current animation duration, in seconds
	gdouble m_Duration;
	/// Stores the current animation frame rate, in frames-per-second
	gdouble m_FrameRate;
	/// Stores the current animation render time, as a percentage
	gdouble m_RenderTime;
	/// Stores the length of the render time, as a percentage
	gdouble m_RenderDeltaTime;

	/// Called by the animation interval timer
	static gint RawAnimationTimer(gpointer Data);
	/// Called by the animation interval timer
	gint AnimationTimer();
	/// Stores the animation timer handler id
	guint m_AnimationTimer;
	/// Used to ignore every other timer signal (so the display can keep up)
	bool m_IgnoreAnimationTimer;

	/// Stores the current active axis
	k3dIDocument::Axis m_ActiveAxis;
	/// Stores the current active transform type
	k3dIDocument::TransformType m_ActiveTransformType;
	/// Stores the current active coordinate system
	k3dIDocument::CoordinateSystem m_ActiveCoordinateSystem;
	/// Stores the current active edit mode
	k3dIDocument::EditMode m_ActiveEditMode;

	/// Attaches a CObjectContext to an object instance
	sdpResult AttachObjectContext(CObjectInfo* Info, k3dIObject* Object, bool UserVisible, bool Saveable);
	/// Attaches a CObjectContext to an object being loaded from an XML document
	sdpResult AttachObjectContext(CObjectInfo* Info, k3dIObject* Object, sdpXmlIElement* Element);

	/// Dumps the object hierarchy to cout for debugging
	void DumpContextHierarchy();
	/// Performs sanity checks on the object hierarchy for debugging
	void ValidateContextHierarchy();

	/// Stores the current unique object ID
	k3dID m_CurrentObjectID;

	/// Stores a pointer to the default material object
	k3dIMaterial* m_DefaultMaterial;

	/// Stores actions performed by the user
	ActionStack m_ActionStack;
	/// Stores actions that have been undone by the user
	ActionStack m_RedoStack;
	/// Stores a pointer to the action being currently recorded
	CAction* m_CurrentAction;

	/// Saves timebase information to an XML document
	sdpResult SaveTimeBase(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves information about the default material to an XML document
	sdpResult SaveDefaultMaterial(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves layer information to an XML document
	sdpResult SaveLayers(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves frame information to an XML document
	sdpResult SaveFrames(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves document objects to an XML document
	sdpResult SaveObjects(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	/// Saves object hierarchy information to an XML document
	sdpResult SaveObjectHierarchy(sdpXmlIElement* Element, k3dIDependencies* Dependencies);

	/// Loads timebase information from an XML document
	sdpResult LoadTimeBase(sdpXmlIElement* Element);
	/// Loads information about the default material from an XML document
	sdpResult LoadDefaultMaterial(sdpXmlIElement* Element);
	/// Loads layer information from an XML document
	sdpResult LoadLayers(sdpXmlIElement* Element);
	/// Loads frame information from an XML document
	sdpResult LoadFrames(sdpXmlIElement* Element);
	/// Loads objects from an XML document
	sdpResult LoadObjects(sdpXmlIElement* Element);
	/// Loads object hierarchy information from an XML document
	sdpResult LoadObjectHierarchy(sdpXmlIElement* Element);
	/// Signals objects that the load has been completed
	sdpResult LoadComplete(sdpXmlIElement* Element);

	/// Lets the UI know that actions have been recorded/undone/redone, etc.
	gulong m_ActionRevision;
	/// Lets the UI know that the hierarchy has been altered
	gulong m_HierarchyRevision;
	/// Lets the UI know that the layers have been changed
	gulong m_LayerRevision;
	/// Lets the UI know that the active edit mode has been altered
	gulong m_EditModeRevision;
	/// Lets the UI know that the active axis has been altered
	gulong m_ActiveAxisRevision;
	/// Lets the UI know that the active transformation has been altered
	gulong m_ActiveTransformationRevision;
	/// Lets the UI know that the playmode has been altered
	gulong m_PlaybackModeRevision;
	/// Lets the UI know that the timebase has been altered
	gulong m_TimebaseRevision;
	
	/// Enumerates undo-able actions
	typedef enum
	{
		STEP_UNCREATE_OBJECT,
		STEP_CREATE_OBJECT,
		STEP_UNDELETE_OBJECT,
		STEP_DELETE_OBJECT,
	} StepType;
};

#endif DOCUMENT_H
