#ifndef HIERARCHYCONTROL_H
#define HIERARCHYCONTROL_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CHierarchyControl class, which encapsulates a hierarchical view of an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dmouseinput.h>

#include <sdpgtkobjectcontainer.h>
#include <sdpgtkopengldrawingarea.h>
#include <sdpmath.h>

#include <vector>
#include <list>
#include <map>

// Forward class declarations
class k3dIDocument;
class k3dIObjectContext;
class k3dIDependencies;
class sdpXmlIElement;

///////////////////////////////////////////////////////////////////
// CRenameObjectDialog

/// Provides a user interface for changing an object's name
class CRenameObjectDialog : public sdpGtkObjectContainer
{
public:
	CRenameObjectDialog(k3dIObjectContext* ObjectContext);

	/// Called to make everything happen ...
	bool DoModal();

protected:
	/// Called to handle our custom events
	void OnEvent(sdpGtkEvent* Event);

	void OnDestroy();
	void OnNewName();
	void OnOK();
	void OnCancel();

	/// Called to put formatted data in all dialog controls
	void UpdateControls();

	/// Stores the object context of the object to be renamed
	k3dIObjectContext* m_ObjectContext;
	sdpString m_Name;
};

///////////////////////////////////////////////////////////
// CHierarchyItem

/// Encapsulates a single visible item within the hierarchy control
class CHierarchyItem
{
public:
	CHierarchyItem(k3dIObjectContext* ObjectContext);

	/// Returns the K-3D object represented by this item
	k3dIObjectContext* ObjectContext() const { return m_ObjectContext; }

	/// Sets the size in screen coordinates of this item
	void SetSize(sdpVector2 Size) { m_Size = Size; }
	/// Returns the size in screen coordinates of this item
	sdpVector2 Size() { return m_Size; }

	/// Sets the position in screen coordinates of this item
	void SetPosition(sdpVector2 Position) { m_Position = Position; }
	/// Returns the position in screen coordinates of this item
	sdpVector2 Position() { return m_Position; }

	/// Returns this item's bounding box (in screen coordinates)
	sdpRectangle BoundingBox();
	/// Returns this item's visibility icon bounding box (in screen coordinates)
	sdpRectangle VisibilityBoundingBox();
	/// Returns this item's selection icon bounding box (in screen coordinates)
	sdpRectangle SelectionBoundingBox();

	/// Sets the item's highlighted state
	void SetHighlighted(bool Highlighted) { m_Highlighted = Highlighted; }
	/// Returns the item's highlighted state
	const bool Highlighted() const { return m_Highlighted; }

	/// Opens the UI of the K-3D object represented by this item
	void EditObject();
	/// Opens a dialog for renaming the K-3D object represented by this item
	void RenameObject();

protected:
	k3dIObjectContext* m_ObjectContext;

	bool m_Highlighted;

	sdpVector2 m_Size;
	sdpVector2 m_Position;
};

typedef std::list<CHierarchyItem> HierarchyItems;
typedef HierarchyItems::iterator HierarchyItemIterator;
typedef HierarchyItems::const_iterator ConstHierarchyItemIterator;

typedef std::pair<HierarchyItemIterator, CHierarchyItem> HierarchyDragItem;
typedef std::list<HierarchyDragItem> HierarchyDragItems;
typedef HierarchyDragItems::iterator HierarchyDragIterator;
typedef HierarchyDragItems::const_iterator ConstHierarchyDragIterator;

///////////////////////////////////////////////////////////
// CHierarchyControl

class CHierarchyControl : public sdpGtkObjectContainer, public k3dMouseInput
{
public:
	CHierarchyControl();
	virtual ~CHierarchyControl();
	
	// Initialization ...
	bool Create(k3dIDocument* Document, sdpGtkWidget Owner);
	bool Load(k3dIDocument* Document, sdpXmlIElement* Element, sdpGtkWidget Owner);
	void Destroy();
	
	// Layout types ...
	void SetTreeView();
	void SetListView();

	// Returns a collection of currently-highlighted items ...
	HierarchyItems HighlightedItems() const;

	// Persistence ...
	bool Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	
	/// Returns the parent document
	k3dIDocument* Document() const;
	
protected:
	// Event handlers ...
	void OnEvent(sdpGtkEvent* Event);
	
	void OnConfigure();
	void OnDraw();
	void OnDestroy();
	void OnTreeView();
	void OnListView();
	void OnIdle();

	/// Draws all regular items
	void DrawItems();
	/// Draws any items being dragged
	void DrawDragItems();

	/// Draws an item
	void DrawItem(CHierarchyItem& Item);
	/// Draws an item being dragged
	void DrawDragItem(HierarchyDragIterator DragItem);

	void DrawInsertionPoint();
	void DrawListInsertionPoint();
	void DrawTreeInsertionPoint();

	/// Called to cause a redraw of the entire control
	void RedrawControl();
	/// Called to update our data when the document hierarchy is altered
	void UpdateRecords();
	/// Called to position each object for each layout type
	void UpdatePositions();
	void UpdateListPositions();
	void UpdateTreePositions();
	/// Called to update the appearance of toolbar controls, etc.
	void UpdateControls();
	/// Called to update the size / position of scrollbars
	void UpdateScrollbars();

	// k3dMouseInput overrides ...
	void OnLButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse);
	void OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);
	void OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse);

	/// Returns the hierarchy control record at the given screen coordinates (if any)
	CHierarchyItem* HitTest(sdpVector2 ScreenCoordinates);
	/// Computes the size of an item
	sdpVector2 StandardItemSize();
	/// Returns the zero-based index into the list corresponding to the given screen Y coordinate
	gulong ListIndex(gdouble ScreenY);
	/// Returns the zero-based index of any items being dragged within the list
	gulong ListDragIndex();
	/// Returns the screen coordinates corresponding to the given zero-based list index
	sdpVector2 ListPosition(gulong Index);
	/// Returns the zero-based index into the tree corresponding to the given screen Y coordinate
	gulong TreeIndex(gdouble ScreenY);
	/// Returns the zero-based index, plus new parent item, of any items being dragged within the tree
	gulong TreeDragIndex(HierarchyItemIterator& Parent, HierarchyItemIterator& Sibling);
	/// Returns the screen coordinates corresponding to the given zero-based tree index
	sdpVector2 TreePosition(gulong Index);
	/// Returns the screen coordinates corresponding to the given zero-based tree index, and offset to show an item's generation
	sdpVector2 TreePosition(gulong Index, CHierarchyItem& Item);
	/// Returns the zero-based generation of the given item
	gulong TreeGeneration(CHierarchyItem& Item);
	/// Returns the total horizontal offset in screen coordinates for an item, based on its generation
	gdouble TreeGenerationOffset(CHierarchyItem& Item);

	/// Enumerates different layout styles/behaviour for the hierarchy control
	typedef enum
	{
		HIERARCHY_LIST,
		HIERARCHY_TREE,
	} HierarchyLayout;

	/// Stores the current layout scheme
	HierarchyLayout m_Layout;
	/// Stores a reference to the owning document
	k3dIDocument* m_Document;
	/// OpenGL context for drawing
	sdpGtkOpenGLDrawingArea m_Widget;
	/// Stores the GL drawing list for our first ASCII character
	GLuint m_FontBase;
	/// Stores the height in pixels of the widget's font
	gint m_FontAscent;
	gint m_FontDescent;
	/// Stores the last hierarchy revision, so we can keep track of when it changes 
	gulong m_LastHierarchyRevision;
	/// Stores the collection of items for the list view
	HierarchyItems m_ListItems;
	/// Stores the collection of items for the tree view
	HierarchyItems m_TreeItems;
	/// Stores the collection of items being dragged by the mouse
	HierarchyDragItems m_DragItems;

	/// Stores the space between interior text and the outside border of an item
	gdouble m_InteriorBorder;
	/// Background color
	sdpVector4 m_BackgroundColor;
	/// The normal item text color
	sdpVector4 m_TextColor;
	/// The color of selected items
	sdpVector4 m_HighlightedColor;
	/// The selected item text color
	sdpVector4 m_HighlightedTextColor;
	/// Parent <-> child link color
	sdpVector4 m_LinkColor;
};

#endif // !HIERARCHYCONTROL_H
