#ifndef OBJECTINFO_H
#define OBJECTINFO_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CObjectInfo class, which encapsulates information about a class of K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpcom.h>
#include <k3diobjectinfo.h>
#include <k3diobjectlibrary.h>

#include <vector>
#include <gmodule.h>

// Forward class declarations
class sdpXmlIElement;

//////////////////////////////////////////////////////////////////////////////////////
// CObjectInfo

class CObjectInfo : public k3dIObjectInfo
{
public:
	CObjectInfo(k3dIObjectLibrary* Library, sdpXmlIElement* Element, sdpComClassProcedureProc ClassProcedure);
	virtual ~CObjectInfo() {}

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_NO_INTERFACE; }
	gulong AddRef() { return 0; }
	gulong Release() { return 0; }

	// k3dIObjectInfo overrides ...
	k3dIObjectLibrary* Library();
	const sdpCLSID& ClassID();
	SDPCSTRING Name();
	sdpResult QueryObjectInterface(const sdpIID& Interface);
	K3DBOOKMARKHANDLE GetFirstBookmarkHandle();
	K3DBOOKMARKHANDLE GetNextBookmarkHandle(K3DBOOKMARKHANDLE Handle);
	SDPCSTRING GetBookmark(K3DBOOKMARKHANDLE Handle);

	/// Returns the sdpCom class-instantiation procedure for this class type
	sdpComClassProcedureProc ClassProcedure() { return m_ClassProcedure; }

	/// Called from sdpXmlForEachElement() to store interface IDs supported by this class type
	bool HandleElement(sdpXmlIElement* Element);

protected:
	/// The library that owns this object
	k3dIObjectLibrary* m_Library;
	/// The object Class ID
	sdpCLSID m_ClassID;
	/// The object name
	sdpString m_Name;
	/// The function for creating objects of this type
	sdpComClassProcedureProc m_ClassProcedure;

	/// Stores a list of interfaces supported by this object type
	typedef std::vector<sdpIID> InterfaceList;
	typedef InterfaceList::iterator InterfaceIterator;
	InterfaceList m_Interfaces;
	
	/// Stores a list of bookmarks for this object
	typedef std::vector<sdpString> Bookmarks;
	typedef Bookmarks::iterator BookmarkIterator;
	Bookmarks m_Bookmarks;
};

//////////////////////////////////////////////////////////////////////////////////////
// CObjectLibrary

typedef std::vector<CObjectInfo*> ObjectInfos;
typedef ObjectInfos::iterator ObjectInfoIterator;

class CObjectLibrary : public k3dIObjectLibrary
{
public:
	CObjectLibrary(SDPCSTRING FilePath, SDPCSTRING FileName, GModule* Module, sdpComLibraryInfoProc LibraryInfoProc, sdpComClassProcedureProc ClassProcedureProc);
	virtual ~CObjectLibrary();

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_NO_INTERFACE; }
	gulong AddRef() { return 0; }
	gulong Release() { return 0; }

	// k3dIObjectLibrary overrides ...
	SDPCSTRING Name();
	SDPCSTRING Path();

	// Data access ...
	GModule* Module() { return m_Module; }
	sdpComLibraryInfoProc LibraryInfoProc() { return m_LibraryInfoProc; }
	sdpComClassProcedureProc ClassProcedureProc() { return m_ClassProcedureProc; }

	ObjectInfos& Objects() { return m_Objects; }

protected:
	// Helper functions ...
	void LoadObjects(sdpComLibraryInfoProc LibraryInfoProc, sdpComClassProcedureProc ClassProcedureProc);
	void LoadObject(sdpXmlIElement* Element);

	// Data ...
	GModule* m_Module;
	sdpString m_Name;
	sdpString m_Path;
	sdpComLibraryInfoProc m_LibraryInfoProc;
	sdpComClassProcedureProc m_ClassProcedureProc;

	ObjectInfos m_Objects;
};

//////////////////////////////////////////////////////////////////////////////////////
// CObjectLibraries

typedef std::vector<CObjectLibrary*> ObjectLibraries;
typedef ObjectLibraries::iterator ObjectLibraryIterator;

typedef std::vector<CObjectInfo*> ObjectInfoPointers;
typedef ObjectInfoPointers::iterator ObjectInfoPointerIterator;

class CObjectLibraries
{
public:
	CObjectLibraries();
	~CObjectLibraries();

	// Initialization ...
	void LoadLibraries(SDPCSTRING Directory, bool Recursive);

	// Functor-type operator for use with sdpForEachFile() ...
	bool HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName);

	// Data access ...
	ObjectLibraries& Libraries() { return m_Libraries; }
	ObjectInfoPointers Objects();

	// Helper functions ...
	CObjectInfo* ObjectInfo(const sdpCLSID& ClassID);

protected:
	ObjectLibraries m_Libraries;
};

#endif OBJECTINFO_H
