// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CApplicationFrame class, which implements the per-application UI
		\author Tim Shead (tshead@k-3d.com)
*/

#include "applicationframe.h"
#include "application.h"
#include "document.h"
#include "options.h"
#include "scripteditor.h"

#include <k3dmisc.h>

#include <sdpgtkfileselector.h>
#include <sdpgtkevents.h>
#include <sdpgtkutility.h>

#include <sdpxmlutility.h>

#include <sdpfile.h>
#include <sdputility.h>

#include <algorithm>

#include "../../k3d_version.h"

#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

//////////////////////////////////////////////////////////////////////
// CAboutBox

bool CAboutBox::Create()
{
	// Load the dialog template ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(Application.DialogPath(), "about.gtkml")), false);

	// Set our version and copyright notices ...
	Label("version").SetText(sdpFormatString("K-3D Version %s", VERSION_STRING));
	
	// Center it ...
	gtk_window_set_position(RootWindow(), GTK_WIN_POS_CENTER);

	// Make it all visible ...
	RootWidget().Show();

	return true;
}

void CAboutBox::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onok")
		RootWidget().Destroy();
	else if(Event->Name() == "ondestroy")
		CancelModal();
	else		
		sdpGtkObjectContainer::OnEvent(Event);
}

//////////////////////////////////////////////////////////////////////
// CSplashBox

CSplashBox::~CSplashBox()
{
	// Kill any GTK+ widgets ...
	if(Root())
		RootWidget().Destroy();
}

bool CSplashBox::Create()
{
	// Load the dialog template ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(Application.DialogPath(), "splash.gtkml")), false);

	// Set our version and copyright notices ...
	Label("version").SetText(sdpFormatString("K-3D Version %s", VERSION_STRING));
	
	// Center it ...
	gtk_window_set_position(RootWindow(), GTK_WIN_POS_CENTER);

	// Make it all visible ...
	RootWidget().Show();

	// Give the dialog a chance to display itself ...
	for(int z = 0; z < 10; z++)
		{
			sdpSleep(10);
			sdpGtkHandlePendingEvents();
		}

	return true;
}

void CSplashBox::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondestroy")
		DisconnectAllEvents();
	else		
		sdpGtkObjectContainer::OnEvent(Event);
}

//////////////////////////////////////////////////////////////////////
// CApplicationFrame

CApplicationFrame::CApplicationFrame()
{
	m_LastActiveDocumentRevision = 0;
}

CApplicationFrame::~CApplicationFrame()
{
	if(m_Application->Verbose())
		std::cout << "CApplicationFrame::~CApplicationFrame()" << std::endl;

	// Get rid of leftover script editors ...
	for(ScriptEditorIterator scripteditor = m_ScriptEditors.begin(); scripteditor != m_ScriptEditors.end(); scripteditor++)
		(*scripteditor)->RootWidget().Destroy();
}

bool CApplicationFrame::Create(CApplication* Application)
{
	// Sanity checks ...
	g_assert(Application);

	// Store a pointer to the application ...
	m_Application = Application;

	// Load our GTKML template ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(m_Application->DialogPath(), "applicationframe.gtkml")), false);

	// Make it all visible ...
	RootWidget().Show();

	return true;
}

void CApplicationFrame::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "ondelete")
		OnDelete(Event);
	else if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onfilenew")
		OnFileNew();
	else if(Event->Name() == "onfileopen")
		OnFileOpen();
	else if(Event->Name() == "onfileexit")
		OnFileExit();
	else if(Event->Name() == "ontoolsclicked")
		OnToolsClicked(Event);
	else if(Event->Name() == "onobjectsclicked")
		OnObjectsClicked(Event);
	else if(Event->Name() == "ontoolsscripteditor")
		OnToolsScriptEditor();
	else if(Event->Name() == "ontoolsoptions")
		OnToolsOptions();
	else if(Event->Name() == "onhelpindex")
		OnHelpIndex();
	else if(Event->Name() == "onhelphowto")
		OnHelpHowTo();
	else if(Event->Name() == "onhelpsearch")
		OnHelpSearch();
	else if(Event->Name() == "onhelptipoftheday")
		OnHelpTipOfTheDay();
	else if(Event->Name() == "onhelponline")
		OnHelpOnline();
	else if(Event->Name() == "onhelpabout")
		OnHelpAbout();
	else if(Event->Name() == "onidle")
		OnIdle();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void CApplicationFrame::OnDelete(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(m_Application->Verbose())
		std::cout << "CApplicationFrame::OnDelete()" << std::endl;

	// Kill the application ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(!SafeToExit());
}

void CApplicationFrame::OnDestroy()
{
	if(m_Application->Verbose())
		std::cout << "CApplicationFrame::OnDestroy()" << std::endl;

	// No more events from this point forward ...
	DisconnectAllEvents();

	// No need to keep track of widgets anymore ...
	Clear();

	// Remove ourselves from the application ...
	m_Application->DeleteFrame(this);
}

// User wants to create a new document ...
void CApplicationFrame::OnFileNew()
{
	m_Application->CreateDocument();
}

// User wants to open an existing document ...
void CApplicationFrame::OnFileOpen()
{
	// Get our starting path - start with the user's home directory ...
	sdpString startpath = sdpGetHomeDirectory();
	
	// Next, override it with the user's options ...
	sdpXmlIElement* startpathelement = sdpXmlGetXPathElement(Application()->Options()->GetRootElement(), "application/paths/path[@type=\"document\"]");
	if(startpathelement)
		startpath = startpathelement->Text();
	
	// Append a directory separator (GtkFileSelector seems to want it) ...
	startpath += G_DIR_SEPARATOR;

	// Get a filepath from the user ...
	sdpGtkFileSelector selector("Open K-3D Document:", startpath);
	selector.DoModal(RootWindow());

	if(!selector.OK())
		return;
		
	// Store the new path ...
	if(startpathelement)
		startpathelement->SetText(sdpDirectory(selector.FilePath()));

	m_Application->OpenDocument(selector.FilePath());
}

// User wants to exit the application ...
void CApplicationFrame::OnFileExit()
{
//	std::cout << "CApplicationFrame::OnFileExit()" << std::endl;

	// Kill ourselves!
	if(SafeToExit())
		RootWidget().Destroy();
}

void CApplicationFrame::OnToolsScriptEditor()
{
	CScriptEditor* dialog = new CScriptEditor(this);
	dialog->Create();
	
	m_ScriptEditors.insert(dialog);
}

void CApplicationFrame::OnToolsOptions()
{
	COptionsDialog dialog(m_Application->m_Options);
	dialog.Create();
	dialog.DoModal(RootWindow());
}

void CApplicationFrame::OnHelpIndex()
{
//	m_Application->BrowserNavigate(sdpAppendPath(m_ApplicationDirectory, sdpAppendPath("help", "index.html")));
}

void CApplicationFrame::OnHelpHowTo()
{
//	m_Application->BrowserNavigate(sdpAppendPath(m_ApplicationDirectory, sdpAppendPath("help", sdpAppendPath("howto", "index.html"))));
}

void CApplicationFrame::OnHelpTutorials()
{
//	m_Application->BrowserNavigate(sdpAppendPath(m_ApplicationDirectory, sdpAppendPath("help", sdpAppendPath("tutorials", "index.html"))));
}

void CApplicationFrame::OnHelpTipOfTheDay()
{
//	ShowTipDialog();
}

void CApplicationFrame::OnHelpOnline()
{
	m_Application->BrowserNavigate("http://www.k-3d.com");
}

void CApplicationFrame::OnHelpSearch()
{
/*
	// If we haven't already created the dialog ...
	if(!m_HelpSearchDialog) m_HelpSearchDialog = new CHelpSearchDialog();

	// If the dialog isn't already open ...
	if(!m_HelpSearchDialog->m_hWnd) m_HelpSearchDialog->Create(CHelpSearchDialog::IDD);

	// Make sure the dialog isn't minimized ...
	m_HelpSearchDialog->ShowWindow(SW_RESTORE);
*/
}


void CApplicationFrame::OnHelpAbout()
{
	CAboutBox aboutbox;
	g_return_if_fail(aboutbox.Create());

	aboutbox.DoModal();
}

void CApplicationFrame::OnToolsClicked(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	sdpGtkEventWidgetButtonPressEvent* event = (sdpGtkEventWidgetButtonPressEvent*)Event;

	// If it isn't a double-click, skip it ...
	if(event->Event()->type != GDK_2BUTTON_PRESS)
		return;

	// Make sure we've got a document ...
	g_return_if_fail(m_Application->ActiveDocument());

	// Get our list widget ...
	sdpGtkCList list = ToolsTree();

	// Get the row we clicked on ...
	int row;
	int column;
	list.GetHitInfo(event->Event()->x, event->Event()->y, &row, &column);

	// Convert it into an object info ...
	CObjectInfo* info = (CObjectInfo*)list.GetRowData(row);
	g_assert(info);

	// Make this undo-able ...
	k3dIObject* object = m_Application->ActiveDocument()->CreateObject(info->ClassID(), true, true, false);
	
	g_return_if_fail(object);
}

void CApplicationFrame::OnObjectsClicked(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	sdpGtkEventWidgetButtonPressEvent* event = (sdpGtkEventWidgetButtonPressEvent*)Event;

	// If it isn't a double-click, skip it ...
	if(event->Event()->type != GDK_2BUTTON_PRESS)
		return;

	// Make sure we've got a document ...
	g_return_if_fail(m_Application->ActiveDocument());

	// Get our list widget ...
	sdpGtkCList list = ObjectsTree();

	// Get the row we clicked on ...
	int row;
	int column;
	list.GetHitInfo(event->Event()->x, event->Event()->y, &row, &column);

	// Convert it into an object info ...
	CObjectInfo* info = (CObjectInfo*)list.GetRowData(row);
	g_assert(info);

	// Create a new object ...
	m_Application->ActiveDocument()->Actions()->StartAction(info->Name());
	k3dIObject* object = m_Application->ActiveDocument()->CreateObject(info->ClassID(), true, true, true);
	m_Application->ActiveDocument()->Actions()->FinishAction();
	g_return_if_fail(object);
	
	// Update the display ...
	k3dRedrawAll(m_Application->ActiveDocument(), false);
}

void CApplicationFrame::OnIdle()
{
	UpdateTrees();

}

class CApplicationFrameAddTool
{
public:
	CApplicationFrameAddTool(GtkObject* Control) { m_Control = Control; }

	void operator()(CObjectInfo* Object)
	{
		// Sanity checks ...
		g_assert(Object);

		for(K3DBOOKMARKHANDLE handle = Object->GetFirstBookmarkHandle(); handle; handle = Object->GetNextBookmarkHandle(handle))
			{
				if(0 == sdpString(Object->GetBookmark(handle)).find("Tools/"))
					{
						const char* data[1];
						data[0] = Object->Name();

						// This bizarre typecast is brought to you by the letters 'M' and 'S' ...
						sdpGtkCList list(GTK_CLIST(m_Control));
						gint row = list.Append(data);

						list.SetRowData(row, gpointer(Object));
					}
			}
	}
	
protected:
	GtkObject* m_Control;
};


class CApplicationFrameAddObject
{
public:
	CApplicationFrameAddObject(GtkObject* Control) { m_Control = Control; }

	void operator()(CObjectInfo* Object)
	{
		// Sanity checks ...
		g_assert(Object);

		for(K3DBOOKMARKHANDLE handle = Object->GetFirstBookmarkHandle(); handle; handle = Object->GetNextBookmarkHandle(handle))
			{
				if(0 == sdpString(Object->GetBookmark(handle)).find("Objects/"))
					{
						const char* data[1];
						data[0] = Object->Name();

						// This bizarre typecast is brought to you by the letters 'M' and 'S' ...
						sdpGtkCList list(GTK_CLIST(m_Control));
						gint row = list.Append(data);

						list.SetRowData(row, gpointer(Object));
					}
			}
	}

protected:
	GtkObject* m_Control;
};


void CApplicationFrame::UpdateTrees()
{
	// If nothing's changed, we're done ...
	if(m_LastActiveDocumentRevision == m_Application->m_ActiveDocumentRevision)
		return;

	// Store the active document revision so we don't do any unnecessary updates ...
	m_LastActiveDocumentRevision = m_Application->m_ActiveDocumentRevision;

	// Clear out the old stuff ...
	ToolsTree().Clear();
	ObjectsTree().Clear();

	// Add list entries ...
	ObjectInfoPointers objects = m_Application->Libraries().Objects();
	std::for_each(objects.begin(), objects.end(), CApplicationFrameAddTool(ToolsTree()));
	std::for_each(objects.begin(), objects.end(), CApplicationFrameAddObject(ObjectsTree()));

	// Sort those babies!
	ToolsTree().Sort();
	ObjectsTree().Sort();

	// If there isn't an active document, disable the trees ...
	if(m_Application->m_ActiveDocument)
		{
			ToolsTree().SetSensitive(true);
			ObjectsTree().SetSensitive(true);
		}
	else
		{
			ToolsTree().SetSensitive(false);
			ObjectsTree().SetSensitive(false);
		}
}

void CApplicationFrame::DestroyScriptEditor(CScriptEditor* ScriptEditor)
{
	// Sanity checks ...
	g_assert(ScriptEditor);
	
	// Remove it from our list ...
	m_ScriptEditors.erase(ScriptEditor);
	
	// Zap it!
	delete ScriptEditor;
}

// Prompt the user to see if it's safe to exit ...
bool CApplicationFrame::SafeToExit()
{
	if(0 == m_Application->m_Documents.size())
		return true;

	if(SDP_S_OK == m_Application->Query("Exit K-3D?  Unsaved changes in all open documents will be lost (No Undo)", "Exit K-3D:"))
		return true;
		
	return false;
}


