// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CHierarchyControl class, which encapsulates a hierarchical view of an open K-3D document
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "hierarchycontrol.h"

#include <k3didependencies.h>
#include <k3diobject.h>
#include <k3dmisc.h>

#include <sdpgtkevents.h>
#include <sdpgtkloopevents.h>
#include <sdpxmlutility.h>
#include <sdpfile.h>

#include <algorithm>

// We have an unfortunate clash with X ...
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

///////////////////////////////////////////////////////////////////////
// CRenameObjectDialog

CRenameObjectDialog::CRenameObjectDialog(k3dIObjectContext* ObjectContext)
{
	// Sanity checks ...
	g_assert(ObjectContext);

	m_ObjectContext = ObjectContext;
	m_Name = ObjectContext->Name();
}

bool CRenameObjectDialog::DoModal()
{
	// Load the dialog template ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(Application.DialogPath(), "renameobject.gtkml")), false);

	// Update controls ...
	UpdateControls();

	// Make it all visible ...
	RootWidget().Show();
	
	sdpGtkObjectContainer::DoModal();

	return true;
}

void CRenameObjectDialog::UpdateControls()
{
	Entry("oldname").SetText(m_ObjectContext->Name());
	Widget("oldname").SetSensitive(false);
	
	Entry("newname").SetText(m_ObjectContext->Name());
}

void CRenameObjectDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onok")
		OnOK();
	else if(Event->Name() == "oncancel")
		OnCancel();
	else if(Event->Name() == "onnewname")
		OnNewName();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void CRenameObjectDialog::OnDestroy()
{
	// Stop the modal loop ...
	CancelModal();

	// No more events from this point forward ...
	DisconnectAllEvents();
	
	// Clear up memory ...
	Clear();
}

void CRenameObjectDialog::OnNewName()
{
	m_Name = Entry("newname").GetText();
}

void CRenameObjectDialog::OnOK()
{
	m_ObjectContext->SetName(m_Name);
	
	RootWidget().Destroy();
}

void CRenameObjectDialog::OnCancel()
{
	RootWidget().Destroy();
}

///////////////////////////////////////////////////////////////////////////////
// CHierarchyItem

CHierarchyItem::CHierarchyItem(k3dIObjectContext* ObjectContext)
{
	// Sanity checks ...
	g_assert(ObjectContext);
	
	m_ObjectContext = ObjectContext;
	m_Highlighted = false;
}

sdpRectangle CHierarchyItem::BoundingBox()
{
	return sdpRectangle(m_Position[0], m_Position[0] + m_Size[0], m_Position[1], m_Position[1] + m_Size[1]);
}

sdpRectangle CHierarchyItem::VisibilityBoundingBox()
{
	sdpRectangle result = BoundingBox();
	
	result.Shrink(3);
	result.right = result.left + result.Height();
	
	return result;
}

sdpRectangle CHierarchyItem::SelectionBoundingBox()
{
	sdpRectangle result = VisibilityBoundingBox();
	
	result.left = result.right + 3;
	result.right = result.left + result.Height();
	
	return result;
}

void CHierarchyItem::EditObject()
{
	m_ObjectContext->Object()->EditObject();
}

void CHierarchyItem::RenameObject()
{
	CRenameObjectDialog dialog(m_ObjectContext);
	dialog.DoModal();
}

///////////////////////////////////////////////////////////////////////////////
// CHierarchyControl

CHierarchyControl::CHierarchyControl()
{
	m_Layout = HIERARCHY_TREE;
	m_Document = 0;
	m_FontBase = 0;
	m_LastHierarchyRevision = 0;
	
	m_InteriorBorder = 2;

	m_BackgroundColor = sdpVector4(1, 1, 1, 1);
	m_TextColor = sdpVector4(0, 0, 0, 1);
	m_HighlightedColor = sdpVector4(0, 0, 0.5, 0.5);
	m_HighlightedTextColor = sdpVector4(1, 1, 1, 1);
	m_LinkColor = sdpVector4(0, 0, 0, 1);
}


CHierarchyControl::~CHierarchyControl()
{
	if(Application.Verbose())
		std::cout << "CHierarchyControl::~CHierarchyControl()" << std::endl;
}


// Called to initialize the frame ...
bool CHierarchyControl::Create(k3dIDocument* Document, sdpGtkWidget Owner)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Owner.Attached());
	
	m_Document = Document;

	// Load our template ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(k3dGlobalDialogTemplatePath(Document->Application()), "hierarchycontrol.gtkml")), false);

	// Attach ourselves to the parent widget ...
	sdpGtkContainer(GTK_CONTAINER(Owner.Object())).Attach(RootWidget());

	// Extract our various resources ...
	Resource("backgroundcolor").getvalue(m_BackgroundColor);
	Resource("textcolor").getvalue(m_TextColor);
	Resource("highlightedcolor").getvalue(m_HighlightedColor);
	Resource("highlightedtextcolor").getvalue(m_HighlightedTextColor);
	Resource("linkcolor").getvalue(m_LinkColor);

	// Setup OpenGL ...
	sdpGtkContainer container(Container("hierarchy"));
	m_Widget.Create(container, true, true, 5, 6, 5, 16);

	// Setup some events ...
	MapEvent("destroy", "ondestroy", m_Widget, true);
	MapEvent("configure-event", "onconfigure", m_Widget, true);
	MapEvent("expose-event", "onexpose", m_Widget, true);
	MapEvent("motion-notify-event", "onmousemove", m_Widget, true);
	MapEvent("button-press-event", "onbuttondown", m_Widget, true);
	MapEvent("button-release-event", "onbuttonup", m_Widget, true);

	sdpGtkEvent* idleevent = MapEvent("timeout", "onidle", m_Widget, false);
	static_cast<sdpGtkEventTimeout*>(idleevent)->SetDuration(100);
	idleevent->Connect();

	// Get some font metrics ...
	const static sdpString metrictext("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_^");
	gdk_string_extents(static_cast<GtkWidget*>(Widget("hierarchy"))->style->font, metrictext, 0, 0, 0, &m_FontAscent, &m_FontDescent);

/*
	UpdateRecords();
	UpdatePositions();
	UpdateScrollbars();
*/
	
	UpdateControls();
	
	return true;
}

// Called to load the frame from a document ...
bool CHierarchyControl::Load(k3dIDocument* Document, sdpXmlIElement* Element, sdpGtkWidget Owner)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Element);
	g_assert(Element->Name() == sdpString("hierarchycontrol"));

	return Create(Document, Owner);
}

void CHierarchyControl::SetTreeView()
{
	m_Layout = HIERARCHY_TREE;
	UpdateControls();
	RedrawControl();
}

void CHierarchyControl::SetListView()
{
	m_Layout = HIERARCHY_LIST;
	UpdateControls();
	RedrawControl();
}

HierarchyItems CHierarchyControl::HighlightedItems() const
{
	HierarchyItems results;

	switch(m_Layout)
		{
			case HIERARCHY_LIST:
				{
					for(ConstHierarchyItemIterator item = m_ListItems.begin(); item != m_ListItems.end(); item++)
						if(item->Highlighted())
							results.push_back(*item);
					break;
				}
			case HIERARCHY_TREE:
				{
					for(ConstHierarchyItemIterator item = m_TreeItems.begin(); item != m_TreeItems.end(); item++)
						if(item->Highlighted())
							results.push_back(*item);
					break;
				}
		}
	
	return results;
}

bool CHierarchyControl::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlCreateElement(Element, "hierarchycontrol");

	return true;
}

k3dIDocument* CHierarchyControl::Document() const
{
	return m_Document;
}

void CHierarchyControl::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "onconfigure")
		OnConfigure();
	else if(Event->Name() == "onexpose")
		OnDraw();
	else if(Event->Name() == "onscrollbar")
		OnDraw();
	else if(Event->Name() == "onmousemove")
		OnMouseMove(Event);
	else if(Event->Name() == "onbuttondown")
		OnButtonDown(Event);
	else if(Event->Name() == "onbuttonup")
		OnButtonUp(Event);
	else if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "ontreeview")
		OnTreeView();
	else if(Event->Name() == "onlistview")
		OnListView();
	else if(Event->Name() == "onidle")
		OnIdle();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

// Called when the display is resized ...
void CHierarchyControl::OnConfigure()
{
	// Setup some OpenGL fonts ...
	m_Widget.Begin();

	m_FontBase = glGenLists(256);
#ifdef SDPWIN32
	GtkWidget* widget = static_cast<GtkWidget*>(Widget("hierarchy"));
//			GdkGCWin32Data* win32data = GDK_GC_WIN32DATA(widget->style->text_gc[0]);
//			wglUseFontBitmaps(win32data->xgc, 0, 256, m_FontBase);
	HWND window = HWND(GDK_WINDOW_XWINDOW(widget->window));
	HDC devicecontext = GetDC(window);
	wglUseFontBitmaps(devicecontext, 0, 256, m_FontBase);
#else // SDPWIN32
	glXUseXFont(gdk_font_id(static_cast<GtkWidget*>(Widget("hierarchy"))->style->font), 0, 256, m_FontBase);
#endif // !SDPWIN32

	m_Widget.End();

	UpdateScrollbars();
	OnDraw();
}

// Called to refresh the display ...
void CHierarchyControl::OnDraw()
{
	sdpglPushTrace(false);
	
	// If we're not completely initialized, we're done ...
	if(!m_Widget.Initialized())
		return;

	m_Widget.Begin();

	// Get the widget sizes ...
	gulong width = m_Widget.Width();
	gulong height = m_Widget.Height();
	glViewport(0, 0, width, height);

	// Clear the background ...
	glClearColor(m_BackgroundColor[0], m_BackgroundColor[1], m_BackgroundColor[2], 0);
	glClear(GL_COLOR_BUFFER_BIT);

	// Setup the viewing projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, width, height, 0, -2, 1);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslated(-Adjustment("hscrollbar").Value(), -Adjustment("vscrollbar").Value(), 0);

	// Setup drawing options ...
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	glShadeModel(GL_FLAT);

	// Draw items ...
	DrawItems();

	// Draw our insertion point for any drag items ...
	DrawInsertionPoint();

	// Draw drag items ...
	DrawDragItems();

	// Cleanup ...
	m_Widget.SwapBuffers();
	m_Widget.End();
	
	sdpglPopTrace();
}

void DrawRectangle(const sdpRectangle& Rectangle)
{
	glBegin(GL_QUADS);
	glVertex2d(Rectangle.left, Rectangle.top);
	glVertex2d(Rectangle.right, Rectangle.top);
	glVertex2d(Rectangle.right, Rectangle.bottom);
	glVertex2d(Rectangle.left, Rectangle.bottom);
	glEnd();
}

void DrawHollowRectangle(const sdpRectangle& Rectangle)
{
	glBegin(GL_LINE_LOOP);
	glVertex2d(Rectangle.left, Rectangle.top);
	glVertex2d(Rectangle.right, Rectangle.top);
	glVertex2d(Rectangle.right, Rectangle.bottom);
	glVertex2d(Rectangle.left, Rectangle.bottom);
	glEnd();
}

void CHierarchyControl::DrawItems()
{
	switch(m_Layout)
		{
			case HIERARCHY_LIST:
				{
					for(HierarchyItemIterator item = m_ListItems.begin(); item != m_ListItems.end(); item++)
						DrawItem(*item);
					break;
				}
			case HIERARCHY_TREE:
				{
					for(HierarchyItemIterator item = m_TreeItems.begin(); item != m_TreeItems.end(); item++)
						DrawItem(*item);
					break;
				}
		}
}

void CHierarchyControl::DrawDragItems()
{
	for(HierarchyDragIterator item = m_DragItems.begin(); item != m_DragItems.end(); item++)
		DrawDragItem(item);
}

void CHierarchyControl::DrawItem(CHierarchyItem& Item)
{
	// Handle any highlighting around our border ...
	if(Item.Highlighted())
		{
			glColor4dv(m_HighlightedColor);
			DrawRectangle(Item.BoundingBox());
		}

	glLineWidth(1.0f);
	glDisable(GL_LINE_SMOOTH);
	glHint(GL_LINE_SMOOTH_HINT, GL_FASTEST);

	// Draw the visible state ...
	glColor4dv(Item.Highlighted() ? m_HighlightedTextColor : m_TextColor);
	if(SDP_S_OK == Item.ObjectContext()->State()->QueryState(k3dIState::VISIBLE))
		DrawRectangle(Item.VisibilityBoundingBox());
	else
		DrawHollowRectangle(Item.VisibilityBoundingBox());

	// Draw the selection state ...
	glColor4dv(Item.Highlighted() ? m_HighlightedTextColor : m_TextColor);
	if(SDP_S_OK == Item.ObjectContext()->State()->QueryState(k3dIState::SELECTED))
		DrawRectangle(Item.SelectionBoundingBox());
	else
		DrawHollowRectangle(Item.SelectionBoundingBox());

	// Draw the text ...
	sdpRectangle textrectangle = Item.BoundingBox();
	textrectangle.Shrink(m_InteriorBorder);
	textrectangle.left = Item.SelectionBoundingBox().Right() + m_InteriorBorder;
		
	glColor4dv(Item.Highlighted() ? m_HighlightedTextColor : m_TextColor);
	glRasterPos2d(textrectangle.left, textrectangle.top + m_FontAscent);
	glListBase(m_FontBase);
	sdpString name = Item.ObjectContext()->Name();
	glCallLists(name.size(), GL_UNSIGNED_BYTE, (GLubyte*)name.c_str());
}

void CHierarchyControl::DrawDragItem(HierarchyDragIterator DragItem)
{
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	DrawItem(DragItem->second);
	
	glDisable(GL_BLEND);
}

void CHierarchyControl::DrawInsertionPoint()
{
	if(0 == m_DragItems.size())
		return;
		
	switch(m_Layout)
		{
			case HIERARCHY_LIST:
				DrawListInsertionPoint();
				break;
			case HIERARCHY_TREE:
				DrawTreeInsertionPoint();
				break;
		}
}

void CHierarchyControl::DrawListInsertionPoint()
{
	// Sanity checks ...
	g_assert(m_DragItems.size());
	g_assert(HIERARCHY_LIST == m_Layout);
	
	gulong width = m_Widget.Width();
	sdpVector2 insertionpoint = ListPosition(ListDragIndex());
	sdpRectangle rectangle(0, width, insertionpoint[1] - 1, insertionpoint[1] + 1);
	
	glColor4dv(m_HighlightedColor);
	DrawRectangle(rectangle);
}

void CHierarchyControl::DrawTreeInsertionPoint()
{
	// Sanity checks ...
	g_assert(m_DragItems.size());
	g_assert(HIERARCHY_TREE == m_Layout);
	
	HierarchyItemIterator parent;
	HierarchyItemIterator sibling;
	gulong dragindex = TreeDragIndex(parent, sibling);

	glColor4dv(m_HighlightedColor);
	
	if(parent != m_TreeItems.end())
		{
			glLineWidth(2.0f);
			glEnable(GL_LINE_SMOOTH);
			glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	
			DrawHollowRectangle(parent->BoundingBox());
		}
		
	if(sibling != m_TreeItems.end())
		{
			gulong width = m_Widget.Width();
			sdpVector2 insertionpoint = TreePosition(dragindex);
			sdpRectangle rectangle(TreeGenerationOffset(*sibling), width, insertionpoint[1] - 1, insertionpoint[1] + 1);
	
			DrawRectangle(rectangle);
		}
}

gulong CHierarchyControl::ListIndex(gdouble ScreenY)
{
	sdpVector2 standardsize = StandardItemSize();
	return gulong(sdpClamp<gdouble>(0.0, sdpRound(ScreenY / standardsize[1]), m_ListItems.size()));
}

gulong CHierarchyControl::ListDragIndex()
{
	// Sanity checks ...
	g_assert(m_DragItems.size());
	g_assert(HIERARCHY_LIST == m_Layout);
	
	HierarchyDragIterator dragitem = m_DragItems.begin();
	sdpVector2 center = dragitem->second.BoundingBox().Center();
	return ListIndex(center[1]);
}

sdpVector2 CHierarchyControl::ListPosition(gulong Index)
{
	sdpVector2 standardsize = StandardItemSize();
	return sdpVector2(0, Index * standardsize[1]);
}

gulong CHierarchyControl::TreeIndex(gdouble ScreenY)
{
	sdpVector2 standardsize = StandardItemSize();
	return gulong(sdpClamp<gdouble>(0.0, sdpRound(ScreenY / standardsize[1]), m_TreeItems.size()));
}

gulong CHierarchyControl::TreeDragIndex(HierarchyItemIterator& Parent, HierarchyItemIterator& Sibling)
{
	// Sanity checks ...
	g_assert(m_DragItems.size());
	g_assert(HIERARCHY_TREE == m_Layout);
	
	HierarchyDragIterator dragitem = m_DragItems.begin();
	sdpVector2 center = dragitem->second.BoundingBox().Center();
	gulong dragindex = TreeIndex(center[1]);
	
	HierarchyItemIterator neighbor = m_TreeItems.begin();
	std::advance(neighbor, glong(dragindex)-1);
	
	gdouble deltay = center[1] - TreePosition(dragindex)[1];
	
	if(deltay < 0)
		{
			Parent = neighbor;
			Sibling = m_TreeItems.end();
		}
	else
		{
			Parent = m_TreeItems.end();
			Sibling = neighbor;
			
		}
	
	return dragindex;
}

sdpVector2 CHierarchyControl::TreePosition(gulong Index)
{
	sdpVector2 standardsize = StandardItemSize();
	return sdpVector2(0, Index * standardsize[1]);
}

sdpVector2 CHierarchyControl::TreePosition(gulong Index, CHierarchyItem& Item)
{
	sdpVector2 standardsize = StandardItemSize();
	return sdpVector2(TreeGenerationOffset(Item), Index * standardsize[1]);
}

gulong CHierarchyControl::TreeGeneration(CHierarchyItem& Item)
{
	gulong generation = 0;
	
	for(k3dIObjectContext* context = Item.ObjectContext()->Parent(); context; context = context->Parent())
		generation++;
	
	return generation;
}

gdouble CHierarchyControl::TreeGenerationOffset(CHierarchyItem& Item)
{
	return TreeGeneration(Item) * 1.5 * (m_FontAscent + m_FontDescent);
}

void CHierarchyControl::Destroy()
{
	if(Application.Verbose())
		std::cout << "CHierarchyControl::OnDestroy()" << std::endl;

	DisconnectAllEvents();

	Clear();

	m_Widget.Destroy();	
}

// Called right before we disappear ...
void CHierarchyControl::OnDestroy()
{
	if(Application.Verbose())
		std::cout << "CHierarchyControl::OnDestroy()" << std::endl;

	// Zap our OpenGL context ...
	if(m_Widget.Initialized())
		m_Widget.Destroy();
		
	// Block any subsequent events ...
	DisconnectAllEvents();

	// No need to keep track of events anymore ...
	Clear();
}

void CHierarchyControl::OnTreeView()
{
	if(ToggleButton("treeview").GetState())
		SetTreeView();
}

void CHierarchyControl::OnListView()
{
	if(ToggleButton("listview").GetState())
		SetListView();
}

// Nothing much going on, so let's update our UI ...
void CHierarchyControl::OnIdle()
{
	// If nothing's changed, we're done ...
	if(m_LastHierarchyRevision == static_cast<CDocument*>(m_Document)->m_HierarchyRevision)
		return;

	UpdateRecords();
	UpdatePositions();
	UpdateScrollbars();
	RedrawControl();

	// Record the last revision to prevent unnecessary updates ...
	m_LastHierarchyRevision = static_cast<CDocument*>(m_Document)->m_HierarchyRevision;
}

void CHierarchyControl::RedrawControl()
{
	Widget("hierarchy").QueueDraw();
}

void CHierarchyControl::UpdateRecords()
{
	m_ListItems.clear();
	m_TreeItems.clear();

	// For each node in the hierarchy tree ...
	K3DTREEHANDLE handle;
	k3dIObjectContext* context;
	for(k3dIDocument::TreeNodeType type = m_Document->GetFirstObjectContextTree(&handle); k3dIDocument::COMPLETE != type; type = m_Document->GetNextObjectContextTree(&handle))
		{
			context = m_Document->GetObjectContextTree(handle);
			
			// Skip hidden objects ...
			if(SDP_S_OK != context->State()->QueryState(k3dIState::USERVIEWABLE))
				continue;

			if(k3dIDocument::PUSH != type)
				continue;

			m_ListItems.push_back(CHierarchyItem(context));
			m_TreeItems.push_back(CHierarchyItem(context));
		}
}

void CHierarchyControl::UpdatePositions()
{
	UpdateListPositions();
	UpdateTreePositions();
}

void CHierarchyControl::UpdateListPositions()
{
	HierarchyItemIterator item;
	sdpVector2 size = StandardItemSize();
	for(item = m_ListItems.begin(); item != m_ListItems.end(); item++)
		item->SetSize(size);

	gulong count = 0;
	for(item = m_ListItems.begin(); item != m_ListItems.end(); item++)
		item->SetPosition(ListPosition(count++));
}

void CHierarchyControl::UpdateTreePositions()
{
	HierarchyItemIterator item;
	sdpVector2 size = StandardItemSize();
	for(item = m_TreeItems.begin(); item != m_TreeItems.end(); item++)
		item->SetSize(size);

	gulong count = 0;
	for(item = m_TreeItems.begin(); item != m_TreeItems.end(); item++)
		item->SetPosition(TreePosition(count++, *item));
}

void CHierarchyControl::UpdateControls()
{
	ToggleButton("treeview").SetState(HIERARCHY_TREE == m_Layout);
	ToggleButton("listview").SetState(HIERARCHY_LIST == m_Layout);
}

void CHierarchyControl::UpdateScrollbars()
{
	if(!m_Widget.Initialized())
		return;

	sdpRectangle boundingbox(DBL_MAX, -DBL_MAX, DBL_MAX, -DBL_MAX);
	
	switch(m_Layout)
		{
			case HIERARCHY_LIST:
				{
					for(HierarchyItemIterator item = m_ListItems.begin(); item != m_ListItems.end(); item++)
						{
							sdpRectangle itemboundingbox = item->BoundingBox();
							boundingbox.left = sdpMin(boundingbox.left, itemboundingbox.left);
							boundingbox.right = sdpMax(boundingbox.right, itemboundingbox.right);
							boundingbox.top = sdpMin(boundingbox.top, itemboundingbox.top);
							boundingbox.bottom = sdpMax(boundingbox.bottom, itemboundingbox.bottom);							
						}
					break;
				}
			case HIERARCHY_TREE:
				{
					for(HierarchyItemIterator item = m_TreeItems.begin(); item != m_TreeItems.end(); item++)
						{
							sdpRectangle itemboundingbox = item->BoundingBox();
							boundingbox.left = sdpMin(boundingbox.left, itemboundingbox.left);
							boundingbox.right = sdpMax(boundingbox.right, itemboundingbox.right);
							boundingbox.top = sdpMin(boundingbox.top, itemboundingbox.top);
							boundingbox.bottom = sdpMax(boundingbox.bottom, itemboundingbox.bottom);
						}
					break;
				}
		}
		
	gulong width = m_Widget.Width();
	gulong height = m_Widget.Height();

	sdpVector2 size = StandardItemSize();

	BlockAllEvents();

	sdpGtkAdjustment hadjustment = Adjustment("hscrollbar");
	sdpGtkAdjustment vadjustment = Adjustment("vscrollbar");

	hadjustment.SetUpper(boundingbox.Width());
	vadjustment.SetUpper(boundingbox.Height());
	
	hadjustment.SetPageSize(width);
	vadjustment.SetPageSize(height);

	hadjustment.SetStepIncrement(size[1] * 0.5);
	vadjustment.SetStepIncrement(size[1] * 0.5);

	hadjustment.Changed();
	vadjustment.Changed();
	
	UnblockAllEvents();
}

CHierarchyItem* CHierarchyControl::HitTest(sdpVector2 ScreenCoordinates)
{
	ScreenCoordinates[0] += Adjustment("hscrollbar").Value();
	ScreenCoordinates[1] += Adjustment("vscrollbar").Value();

	switch(m_Layout)
		{
			case HIERARCHY_LIST:
				{
					for(HierarchyItemIterator item = m_ListItems.begin(); item != m_ListItems.end(); item++)
						if(item->BoundingBox().Contains(ScreenCoordinates))
							return &(*item);
					break;
				}
			case HIERARCHY_TREE:
				{
					for(HierarchyItemIterator item = m_TreeItems.begin(); item != m_TreeItems.end(); item++)
						if(item->BoundingBox().Contains(ScreenCoordinates))
							return &(*item);
					break;
				}
		}			
	
	return 0;
}


void CHierarchyControl::OnLButtonDown(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CHierarchyItem* item = HitTest(CurrentMouse);
	if(0 == item)
		return;

	if(item->VisibilityBoundingBox().Contains(CurrentMouse))
		return;

	if(item->SelectionBoundingBox().Contains(CurrentMouse))
		return;

	if(!item->Highlighted())
		{
			if(Modifiers & GDK_CONTROL_MASK)
				{
					item->SetHighlighted(true);
				}
			else
				{
					switch(m_Layout)
						{
							case HIERARCHY_LIST:
								{
									for(HierarchyItemIterator i = m_ListItems.begin(); i != m_ListItems.end(); i++)
										i->SetHighlighted(i->ObjectContext() == item->ObjectContext());
									break;
								}
							case HIERARCHY_TREE:
								{
									for(HierarchyItemIterator i = m_TreeItems.begin(); i != m_TreeItems.end(); i++)
										i->SetHighlighted(i->ObjectContext() == item->ObjectContext());
									break;
								}
						}
				}
		}
		
	RedrawControl();
}

void CHierarchyControl::OnLButtonClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CHierarchyItem* item = HitTest(CurrentMouse);
	if(0 == item)
		{
			switch(m_Layout)
				{
					case HIERARCHY_LIST:
						{
							for(HierarchyItemIterator i = m_ListItems.begin(); i != m_ListItems.end(); i++)
								i->SetHighlighted(false);
							break;
						}
					case HIERARCHY_TREE:
						{
							for(HierarchyItemIterator i = m_TreeItems.begin(); i != m_TreeItems.end(); i++)
								i->SetHighlighted(false);
							break;
						}
				}
			RedrawControl();
		}
	else
		{
			if(item->VisibilityBoundingBox().Contains(CurrentMouse))
				{
					item->ObjectContext()->State()->SetState(k3dIState::VISIBLE, SDP_S_OK == item->ObjectContext()->State()->QueryState(k3dIState::VISIBLE) ? false : true);
					k3dRedrawAll(Document(), false);
				}
			else if(item->SelectionBoundingBox().Contains(CurrentMouse))
				{
					item->ObjectContext()->State()->SetState(k3dIState::SELECTED, SDP_S_OK == item->ObjectContext()->State()->QueryState(k3dIState::SELECTED) ? false : true);
					k3dRedrawAll(Document(), false);
				}
		}
}

void CHierarchyControl::OnLButtonDoubleClick(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	CHierarchyItem* item = HitTest(CurrentMouse);
	if(0 == item)
		return;

	switch(m_Layout)
		{
			case HIERARCHY_LIST:
				{
					for(HierarchyItemIterator i = m_ListItems.begin(); i != m_ListItems.end(); i++)
						i->SetHighlighted(i->ObjectContext() == item->ObjectContext());
					break;
				}
			case HIERARCHY_TREE:
				{
					for(HierarchyItemIterator i = m_TreeItems.begin(); i != m_TreeItems.end(); i++)
						i->SetHighlighted(i->ObjectContext() == item->ObjectContext());
					break;
				}
		}			

	RedrawControl();

	item->EditObject();
}

void CHierarchyControl::OnLButtonStartDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse)
{
	switch(m_Layout)
		{
			case HIERARCHY_LIST:
				{
					for(HierarchyItemIterator item = m_ListItems.begin(); item != m_ListItems.end(); item++)
						if(item->Highlighted())
							m_DragItems.push_back(HierarchyDragItem(item, *item));
					break;
				}
			case HIERARCHY_TREE:
				{
					for(HierarchyItemIterator item = m_TreeItems.begin(); item != m_TreeItems.end(); item++)
						if(item->Highlighted())
							m_DragItems.push_back(HierarchyDragItem(item, *item));
					break;
				}
		}

	RedrawControl();
}

void CHierarchyControl::OnLButtonDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	for(HierarchyDragIterator item = m_DragItems.begin(); item != m_DragItems.end(); item++)
		{
			sdpVector2 newposition = item->first->Position() + (CurrentMouse - StartMouse);
			item->second.SetPosition(newposition);
		}

	RedrawControl();
}

void CHierarchyControl::OnLButtonEndDrag(GdkModifierType Modifiers, const sdpVector2 CurrentMouse, const sdpVector2 LastMouse, const sdpVector2 StartMouse)
{
	if(0 == m_DragItems.size())
		return;
		
	switch(m_Layout)
		{
			case HIERARCHY_LIST:
				{
					// Get the new insert point for the drag items ...
					HierarchyItemIterator insertionpoint = m_ListItems.begin();
					std::advance(insertionpoint, ListDragIndex());
					
					// Shuffle things around ...
					HierarchyDragIterator dragitem;
					for(dragitem = m_DragItems.begin(); dragitem != m_DragItems.end(); dragitem++)
						m_ListItems.insert(insertionpoint, dragitem->second);

					// Zap the originals ...
					for(dragitem = m_DragItems.begin(); dragitem != m_DragItems.end(); dragitem++)
						m_ListItems.erase(dragitem->first);
					m_DragItems.clear();
					
					UpdateListPositions();
					RedrawControl();
					break;
				}
			case HIERARCHY_TREE:
				{
					// Get the new insert point for the drag items ...
					HierarchyItemIterator parent;
					HierarchyItemIterator sibling;
					TreeDragIndex(parent, sibling);

					if(parent != m_TreeItems.end())
						{
							for(HierarchyDragIterator dragitem = m_DragItems.begin(); dragitem != m_DragItems.end(); dragitem++)
								dragitem->second.ObjectContext()->SetParent(parent->ObjectContext());
						}
					else
						{
							for(HierarchyDragIterator dragitem = m_DragItems.begin(); dragitem != m_DragItems.end(); dragitem++)
								dragitem->second.ObjectContext()->SetSibling(sibling->ObjectContext());
						}

					m_DragItems.clear();
					k3dRedrawAll(Document(), false);
					break;
				}
		}
}

sdpVector2 CHierarchyControl::StandardItemSize()
{
	return sdpVector2(200, m_FontAscent + m_FontDescent + (2 * m_InteriorBorder));
}
