// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CHierarchyFrame class, which displays a hierarchical view of an open K-3D document, using CHierarchyControl
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "hierarchyframe.h"

#include <k3dmisc.h>
#include <sdpgtkevents.h>
#include <sdpxmlutility.h>
#include <sdpfile.h>

// We have an unfortunate clash with X
#ifdef RootWindow
#undef RootWindow
#endif // RootWindow

///////////////////////////////////////////////////////////////////////////////
// CHierarchyFrame

CHierarchyFrame::CHierarchyFrame()
{
	m_Document = 0;
}


CHierarchyFrame::~CHierarchyFrame()
{
	if(Application.Verbose())
		std::cout << "CHierarchyFrame::~CHierarchyFrame()" << std::endl;
}


// Called to initialize the frame ...
bool CHierarchyFrame::Create(CDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);

	m_Document = Document;

	// Load the frame ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(k3dGlobalDialogTemplatePath(Document->Application()), "hierarchyframe.gtkml")), false);

	// Create the hierarchy control ...
	g_return_val_if_fail(m_Control.Create(m_Document, Widget("hierarchy")), false);

	// Make ourselves visible ...
	RootWidget().Show();

	// Update our titlebar ...
	UpdateTitlebar();

	return true;
}

// Called to load the frame from a document ...
bool CHierarchyFrame::Load(CDocument* Document, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Element);
	g_assert(Element->Name() == sdpString("hierarchyframe"));

	return Create(Document);
}

bool CHierarchyFrame::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* hierarchyframe = sdpXmlCreateElement(Element, "hierarchyframe");
	return m_Control.Save(hierarchyframe, Dependencies);
}

void CHierarchyFrame::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "onclose")
		OnClose();
	else if(Event->Name() == "ondelete")
		OnDelete(Event);
	else if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onsetfocus")
		OnSetFocus();
	else if(Event->Name() == "onlist")
		m_Control.SetListView();
	else if(Event->Name() == "ontree")
		m_Control.SetTreeView();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

// Called when the user clicks the File/Close menu item ...
void CHierarchyFrame::OnClose()
{
	if(Application.Verbose())
		std::cout << "CHierarchyFrame::OnClose()" << std::endl;

	RootWidget().Destroy();
}

// Called when the user clicks the WM "close" button ...
void CHierarchyFrame::OnDelete(sdpGtkEvent* Event)
{
	if(Application.Verbose())
		std::cout << "CHierarchyFrame::OnDelete()" << std::endl;

	// Let it happen ...
	((sdpGtkEventWidgetDeleteEvent*)Event)->SetResult(false);
}

// Called right before we disappear ...
void CHierarchyFrame::OnDestroy()
{
	if(Application.Verbose())
		std::cout << "CHierarchyFrame::OnDestroy()" << std::endl;

	// Block any subsequent events ...
	DisconnectAllEvents();

	// No need to keep track of events anymore ...
	Clear();

	// Zap the control, so it has a chance to clean-up BEFORE we delete ourselves...
	m_Control.DisconnectAllEvents();

	// Remove ourselves from the document ...
	m_Document->DeleteFrame(this);
}

bool CHierarchyFrame::OnSetFocus()
{
	// If the document is in the process of being closed, do nothing ...
	if(m_Document->IsDying())
		return true;

	// Activate the document ...
	m_Document->OnActivate(this);

	return true;
}

// Reset the window titlebar ...
void CHierarchyFrame::UpdateTitlebar()
{
	// Get our title and stuff it in the titlebar ...
	RootWindow().SetTitle(Title());
}

// Our window title is the same as the document ...
sdpString CHierarchyFrame::Title()
{
	return sdpString(m_Document->Title());
}


