
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CLayer class, which encapsulates a layer within a CDocument
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "layer.h"
#include "objectinfo.h"
#include "objectcontext.h"

#include <k3diobject.h>

#include <algorithm>

//////////////////////////////////////////////////////////////////////////
// CLayer

gulong CLayer::m_LayerNumber = 1;

CLayer::CLayer()
{
	m_Document = 0;
}

CLayer::~CLayer()
{
	if(Application.Verbose())
		std::cout << "CLayer::~CLayer " << m_Name << std::endl;

	// Zap objects that still "officially" exist ...
	for(ObjectContextIterator i = m_Objects.begin(); i != m_Objects.end(); i++)
		{
			// Give the object a chance to shut down its UI ...
			(*i)->Object()->ObjectDeleted();

			delete *i;
		}

	// Zap objects that were already deleted ...
	for(ObjectContextIterator j = m_DeletedObjects.begin(); j != m_DeletedObjects.end(); j++)
		delete *j;
}

bool CLayer::Create(CDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);
	m_Document = Document;

	// Set our default title ...
	m_Name = sdpFormatString("Untitled Layer %u", m_LayerNumber++);

	return true;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CLayer::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dILayer>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dILayer*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CLayer::AddRef()
{
	return 1;
}

gulong CLayer::Release()
{
	return 0;
}

///////////////////////////////////////////////////////////////
// k3dILayer overrides

k3dIDocument* CLayer::Document()
{
	return m_Document;
}

// Create an object context and attach it to a new object ...
sdpResult CLayer::AttachObjectContext(CObjectInfo* Info, k3dIObject* Object, bool UserVisible, bool Saveable)
{
	// Sanity checks ...
	g_assert(Info);
	g_assert(Object);

	// Create an object context ...
	CObjectContext* context = new CObjectContext(this, Info, Object);
	g_return_val_if_fail(context, SDP_E_FAIL);

	// Add it to our list ...
	m_Objects.push_back(context);

	// Set properties for this object
	context->SetState(k3dIState::USERVIEWABLE, UserVisible);
	context->SetState(k3dIState::SAVEABLE, Saveable);

	// Connect the object context to the object
	g_return_val_if_fail(sdpSucceeded(Object->SetObjectContext(context)), SDP_E_FAIL);

	// Tell the object context to initialize the object as a new object ...
	g_return_val_if_fail(sdpSucceeded(context->InitNew()), SDP_E_FAIL);

	// Are we recording actions?
	if(SDP_S_OK == m_Document->Actions()->IsRecording())
		m_Document->Actions()->RecordStep(this, STEP_UNCREATE_OBJECT, reinterpret_cast<gulong>(context), 0,
																					STEP_CREATE_OBJECT, reinterpret_cast<gulong>(context), 0);
	return SDP_S_OK;
}

// Create an object context and attach it to an object just created from a document ...
sdpResult CLayer::AttachObjectContext(CObjectInfo* Info, k3dIObject* Object, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Info);
	g_assert(Object);
	g_assert(Element);

	// Create an object context ...
	CObjectContext* context = new CObjectContext(this, Info, Object);
	g_return_val_if_fail(context, SDP_E_FAIL);

	// Add it to our list ...
	m_Objects.push_back(context);

	// Connect the object context to the object
	g_return_val_if_fail(sdpSucceeded(Object->SetObjectContext(context)), SDP_E_FAIL);

	// Tell the object context to load the object from the document ...
	g_return_val_if_fail(sdpSucceeded(context->Load(Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

void CLayer::DeleteObjectContext(k3dIObjectContext* Context)
{
	// Sanity checks ...
	g_assert(Context);

	// Find the object ...
	ObjectContextIterator context = std::find(m_Objects.begin(), m_Objects.end(), static_cast<CObjectContext*>(Context));
	g_assert(context != m_Objects.end());

	// Are we recording actions?
	if(SDP_S_OK == m_Document->Actions()->IsRecording())
		// If so, record our undo data ...
		m_Document->Actions()->RecordStep(this, STEP_UNDELETE_OBJECT, reinterpret_cast<gulong>(*context), 0,
																					STEP_DELETE_OBJECT, reinterpret_cast<gulong>(*context), 0);

	// Move it to the deleted pile ...
	m_DeletedObjects.push_back(*context);
	m_Objects.erase(context);
}

sdpResult CLayer::HandleStep(gulong ValueCode, gulong Value, gulong ValueLength)
{
	switch(ValueCode)
		{
			case STEP_CREATE_OBJECT:
			case STEP_UNDELETE_OBJECT:
				{
					ObjectContextIterator iterator = std::find(m_DeletedObjects.begin(), m_DeletedObjects.end(), reinterpret_cast<CObjectContext*>(Value));
					g_assert(iterator != m_DeletedObjects.end());
			
					m_Objects.push_back(*iterator);
					m_DeletedObjects.erase(iterator);
			
					return SDP_S_OK;
				}
			case STEP_UNCREATE_OBJECT:
			case STEP_DELETE_OBJECT:
				{
					ObjectContextIterator iterator = std::find(m_Objects.begin(), m_Objects.end(), reinterpret_cast<CObjectContext*>(Value));
					g_assert(iterator != m_Objects.end());
			
					m_DeletedObjects.push_back(*iterator);
					m_Objects.erase(iterator);
			
					return SDP_S_OK;
				}
		}
		
	return k3dActionClient::HandleStep(ValueCode, Value, ValueLength);
}



