// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CObjectContext class, which encapsulates data common to all K-3D objects
		\author Tim Shead (tshead@k-3d.com)
*/

#include "application.h"
#include "document.h"
#include "layer.h"
#include "objectinfo.h"
#include "objectcontext.h"

#include <k3diobject.h>
#include <k3ditimeserver.h>
#include <k3dianimate.h>
#include <k3ditransform.h>
#include <k3dilocationserver.h>
#include <k3dipersist.h>
#include <k3diusereventclient.h>

#include <algorithm>

#include <sdpxmlutility.h>

#include <iostream>

/////////////////////////////////////////////////////////////////////////////
// CObjectContext

CObjectContext::CObjectContext()
{
	m_ReferenceCount = 0;

	m_Layer = 0;
	m_ObjectInfo = 0;
	m_Parent = 0;

	m_ObjectID = 0;

	m_Object = 0;
	m_UserEventClient = 0;
	m_TimeServer = 0;
	m_Animate = 0;
	m_Transform = 0;
	m_Mesh = 0;
	m_LocationServer = 0;
	m_Persist = 0;
}

CObjectContext::CObjectContext(CLayer* Layer, CObjectInfo* ObjectInfo, k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Layer);
	g_assert(ObjectInfo);
	g_assert(Object);

	// Initialize member variables ...
	m_ReferenceCount = 0;

	m_Layer = Layer;
	m_ObjectInfo = ObjectInfo;
	m_Parent = 0;

	m_Object = Object;
	m_UserEventClient = 0;
	m_TimeServer = 0;
	m_Animate = 0;
	m_Transform = 0;
	m_Mesh = 0;
	m_LocationServer = 0;
	m_Persist = 0;

	// Default state
	SetState(k3dIState::VISIBLE | k3dIState::SELECTABLE | k3dIState::SAVEABLE, true);

	// Determine available object interfaces
	m_UserEventClient = sdpQueryInterface<k3dIUserEventClient>(m_Object);
	m_TimeServer = sdpQueryInterface<k3dITimeServer>(m_Object);
	m_Animate = sdpQueryInterface<k3dIAnimate>(m_Object);
	m_Transform = sdpQueryInterface<k3dITransform>(m_Object);
	m_Mesh = sdpQueryInterface<k3dIMesh>(m_Object);
	m_LocationServer = sdpQueryInterface<k3dILocationServer>(m_Object);
	m_Persist = sdpQueryInterface<k3dIPersist>(m_Object);

	// Increment our reference count
	AddRef();

	// Create a new object name (we MUST call this before adding ourselves to the hierarchy list ) ...
	SetName(sdpFormatString("Untitled %s", ObjectInfo->Name()));

	// Add records to hierarchy list
	m_Layer->m_Document->m_Hierarchy.push_back(CHierarchyTreeRecord(k3dIDocument::PUSH, this));
	m_Layer->m_Document->m_Hierarchy.push_back(CHierarchyTreeRecord(k3dIDocument::NODE, this));
	m_Layer->m_Document->m_Hierarchy.push_back(CHierarchyTreeRecord(k3dIDocument::POP, this));

	// Get a new object ID ...
	m_ObjectID = m_Layer->m_Document->GetNextObjectID();

	// Encourage the UI to update itself ...
	m_Layer->m_Document->m_HierarchyRevision++;
}

/*
class CObjectContextNotifyMaterialDeparting
{
public:
	CObjectContextNotifyMaterialDeparting(k3dIMaterial* Material) { m_Material = Material; }

	bool HandlePolygon(k3dIPolygon* Polygon)
	{
		if(Polygon->Material()->GetMaterial() == m_Material)
			Polygon->Material()->SetMaterial(0);
	}

protected:
	k3dIMaterial* m_Material;
};
*/
CObjectContext::~CObjectContext()
{
	if(Application.Verbose())
		{
			std::cout << "CObjectContext::~CObjectContext " << ((k3dIObjectContext*)this);
			std::cout << " - " << Name() << " - " << sdpString(m_ObjectID) << std::endl;
		}

/*
	// Notify objects that this object is departing and make sure children get a new parent
	ForEachObject(m_Document, 0, FNotifyObjectDeparting(this));

	// If we AREN'T closing the document, notify polygons that a material object is departing
	if(!m_Document->m_ClosingDocument)
		{
			// Notify polygons that a material object is departing
			k3dIMaterial* material = sdpQueryInterface<k3dIMaterial>(m_Object);
			if(material)
				{
					CObjectContextNotifyMaterialDeparting functor;
					k3dForEachPolygon(m_Document, 0, functor);
					material->Release();
				}
		}
*/		

	// Call Release() on the object until it disappears ...
	while(m_Object->Release())
		{
		}

	// Encourage the UI to update itself ...
	m_Layer->m_Document->m_HierarchyRevision++;
}

HierarchyTreeIterator CObjectContext::GetPushRecord()
{
	return std::find(m_Layer->m_Document->m_Hierarchy.begin(), m_Layer->m_Document->m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::PUSH, this));
}

HierarchyTreeIterator CObjectContext::GetNodeRecord()
{
	return std::find(m_Layer->m_Document->m_Hierarchy.begin(), m_Layer->m_Document->m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::NODE, this));
}

HierarchyTreeIterator CObjectContext::GetPopRecord()
{
	return std::find(m_Layer->m_Document->m_Hierarchy.begin(), m_Layer->m_Document->m_Hierarchy.end(), CHierarchyTreeRecord(k3dIDocument::POP, this));
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CObjectContext::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIObjectContext>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIState>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObjectContext*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CObjectContext::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CObjectContext::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObjectContext overrides

SDPCSTRING CObjectContext::Name()
{
	return m_Name;
}

k3dIObject* CObjectContext::Object()
{
	g_assert(m_Object);
	return m_Object;
}

const k3dID CObjectContext::ObjectID()
{
	return m_ObjectID;
}

k3dIObjectContext* CObjectContext::Parent()
{
	return m_Parent;
}

k3dIObjectInfo* CObjectContext::ObjectInfo()
{
	// Sanity checks ...
	g_assert(m_Object);

	return m_ObjectInfo;
}

sdpResult CObjectContext::SetName(SDPCSTRING NewName)
{
	// Sanity checks ...
	g_assert(m_Object);

	// See if the object will allow its' name to change ...
	if(SDP_S_OK != m_Object->QueryRenameObject())
		return SDP_S_FALSE;

	// We don't allow leading or trailing whitespace ...
	sdpString name(NewName);
	name.trim();

	// We don't allow empty names ...
	if(0 == name.size())
		name = "Untitled Object";

	// Make sure the name is unique ...
	name = m_Layer->m_Document->GetUniqueObjectName(name);	

	// Set the new name ...
	m_Name = name;

	// Encourage the UI to update itself ...
	m_Layer->m_Document->m_HierarchyRevision++;
	
	return SDP_S_OK;
}

sdpResult CObjectContext::SetActiveObject(bool State)
{
	if(State)
		m_Layer->m_Document->SetActiveObjectContext(this);
	else
		if(m_Layer->m_Document->m_ActiveObjectContext == this)
			m_Layer->m_Document->SetActiveObjectContext((CObjectContext*)0);

	return SDP_S_OK;
}

k3dIState* CObjectContext::State()
{
	return this;
}

k3dILayer* CObjectContext::Layer()
{
	return m_Layer;
}

sdpResult CObjectContext::QueryAncestor(k3dIObjectContext* ObjectContext)
{
	// Note that this doesn't cause a crash, it's just stupid since EVERY
	// object has the root (NULL) object for an ancestor ...
	g_return_val_if_fail(ObjectContext, SDP_S_OK);

	CObjectContext* object;
	for(object = this; object; object = object->m_Parent)
		{
			if(object->m_Parent == (CObjectContext*)ObjectContext)
				break;
		}

	return object ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CObjectContext::QueryParent(k3dIObjectContext* ObjectContext)
{
	return (m_Parent == (CObjectContext*)ObjectContext) ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CObjectContext::QueryChild(k3dIObjectContext* ObjectContext)
{
	return (((CObjectContext*)ObjectContext)->m_Parent == this) ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CObjectContext::QueryDescendant(k3dIObjectContext* ObjectContext)
{
	CObjectContext* object;
	for(object = (CObjectContext*)ObjectContext; object; object = object->m_Parent)
		{
			if(object->m_Parent == this)
				break;
		}

	return object ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult CObjectContext::SetParent(k3dIObjectContext* ObjectContext)
{
	m_Layer->m_Document->ValidateContextHierarchy();
//	m_Layer->m_Document->DumpContextHierarchy();

	// Remove our old records from the hierarchy ...
	HierarchyTreeIterator push = GetPushRecord();
	HierarchyTreeIterator pop = GetPopRecord();
	pop++;

	HierarchyTreeList temp;
	for(HierarchyTreeIterator i = push; i != pop; i++)
		temp.push_back(*i);
	m_Layer->m_Document->m_Hierarchy.erase(push, pop);

	// Get the new parent's "POP" record ...
	HierarchyTreeIterator parentpop;
	if(ObjectContext)
		parentpop = static_cast<CObjectContext*>(ObjectContext)->GetPopRecord();
	else
		parentpop = m_Layer->m_Document->m_Hierarchy.end();

	// Copy the records into position ...
	m_Layer->m_Document->m_Hierarchy.insert(parentpop, temp.begin(), temp.end());

	// Set our new parent ...
	m_Parent = ObjectContext ? static_cast<CObjectContext*>(ObjectContext) : 0;

	m_Layer->m_Document->ValidateContextHierarchy();
//	m_Layer->m_Document->DumpContextHierarchy();

	// Encourage the UI to update itself ...
	m_Layer->m_Document->m_HierarchyRevision++;

	return SDP_S_OK;
}

sdpResult CObjectContext::SetSibling(k3dIObjectContext* ObjectContext)
{
	m_Layer->m_Document->ValidateContextHierarchy();
//	m_Layer->m_Document->DumpContextHierarchy();

	// Remove our old records from the hierarchy ...
	HierarchyTreeIterator push = GetPushRecord();
	HierarchyTreeIterator pop = GetPopRecord();
	pop++;

	HierarchyTreeList temp;
	for(HierarchyTreeIterator i = push; i != pop; i++)
		temp.push_back(*i);
	m_Layer->m_Document->m_Hierarchy.erase(push, pop);

	// Get the new sibling's "POP" record ...
	HierarchyTreeIterator siblingpop;
	if(ObjectContext)
		{
			siblingpop = static_cast<CObjectContext*>(ObjectContext)->GetPopRecord();
			siblingpop++;
		}
	else
		siblingpop = m_Layer->m_Document->m_Hierarchy.begin();

	// Insert the new records into position ...
	m_Layer->m_Document->m_Hierarchy.insert(siblingpop, temp.begin(), temp.end());

	// Set our new parent ...
	if(ObjectContext)
		m_Parent = ObjectContext->Parent() ? static_cast<CObjectContext*>(ObjectContext->Parent()) : 0;
	else
		m_Parent = 0;

	m_Layer->m_Document->ValidateContextHierarchy();
//	m_Layer->m_Document->DumpContextHierarchy();

	// Encourage the UI to update itself ...
	m_Layer->m_Document->m_HierarchyRevision++;

	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////
// k3dIPersist overrides ...

sdpResult CObjectContext::InitNew()
{
	// Initialize the new object ...
	if(m_Persist)
		g_return_val_if_fail(sdpSucceeded(m_Persist->InitNew()), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CObjectContext::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(m_Persist); // If we're being loaded, the object better support k3dIPersist ...

	// Get the object's name ...
	g_return_val_if_fail(sdpXmlGetAttribute(Element, "name", m_Name), SDP_E_FAIL);

	if(Application.Verbose())
		std::cout << "Loading object \"" << m_Name << "\"" << std::endl;

	// Get the object's id ...
	g_return_val_if_fail(sdpXmlGetAttribute(Element, "id", m_ObjectID), SDP_E_FAIL);
	
	// Handle object state ...
	g_return_val_if_fail(sdpSucceeded(k3dState::Load(Element)), SDP_E_FAIL);

	// Give the object a chance ...
	g_return_val_if_fail(sdpSucceeded(m_Persist->Load(Element)), SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult CObjectContext::LoadComplete()
{
	// If this object isn't saveable, we're done ...
	if(!m_Persist)
		return SDP_S_FALSE;

	// Let the object know we're done ...
	g_return_val_if_fail(sdpSucceeded(m_Persist->LoadComplete()), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CObjectContext::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	// If this object isn't saveable, we're done ...
	if(!m_Persist)
		return SDP_S_FALSE;

	if(SDP_S_OK != QueryState(k3dIState::SAVEABLE))
		return SDP_S_FALSE;

	if(Application.Verbose())
		std::cout << "Saving object \"" << m_Name << "\"" << std::endl;

	// Save our data ...
	sdpXmlIElement* object = sdpXmlCreateElement(Element, "object", 0, "name", m_Name, "class", m_ObjectInfo->ClassID(), "id", m_ObjectID);
	g_return_val_if_fail(sdpSucceeded(k3dState::Save(object, Dependencies)), SDP_E_FAIL);
	
	// Give the object a chance to save its stuff ...
	g_return_val_if_fail(sdpSucceeded(m_Persist->Save(object, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CObjectContext::SaveComplete()
{
	// If this object isn't saveable, we're done ...
	if(!m_Persist)
		return SDP_S_FALSE;

	if(SDP_S_OK != QueryState(k3dIState::SAVEABLE))
		return SDP_S_FALSE;

	// Let the object know we're done ...
	g_return_val_if_fail(sdpSucceeded(m_Persist->SaveComplete()), SDP_E_FAIL);

	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////////
// k3dIState overrides ...

sdpResult CObjectContext::SetState(gulong Mask, bool State)
{
	// Special handling for selection / deselection ...
	if(Mask & k3dIState::SELECTED)
		{
			// Give the object a chance to override ...
			sdpResult hr = m_Object->QuerySelectObject();
			if(SDP_S_OK != hr)
				return hr;

			k3dState::SetState(Mask, State);

			// Notify the object that its' selection state has changed ...
			m_Object->ObjectSelected(State);

			// Make ourselves the active object
			SetActiveObject(State);
		}
	else
		{
			k3dState::SetState(Mask, State);
		}
		
	m_Layer->m_Document->m_HierarchyRevision++;

	return SDP_S_OK;
}


