// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the COptionsDialog, which provides a UI for changing global user options
		\author Tim Shead (tshead@k-3d.com)
*/

#include "options.h"
#include "application.h"

#include <sdpgtkevent.h>
#include <sdpxmlutility.h>
#include <sdpfile.h>

///////////////////////////////////////////////////////////////////////
// COptionsDialog

COptionsDialog::COptionsDialog(k3dDocument& Options) : m_Options(Options)
{
}

bool COptionsDialog::Create()
{
	// Load the dialog template ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(Application.DialogPath(), "options.gtkml")), false);

	// Update controls ...
	UpdateControls();

	// Make it all visible ...
	RootWidget().Show();

	return true;
}

void COptionsDialog::UpdateControls()
{
	sdpXmlIElement* root = m_Options.GetRootElement();

	// Custom render commands ...
	Entry("customenginevisible").SetText(sdpXmlGetXPathElement(root, "application/helpers/command[@type=\"render\" and @enginetype=\"ri\" and @engine=\"custom\" and @viewduringrender=\"true\"]")->Text());
	Entry("customenginehidden").SetText(sdpXmlGetXPathElement(root, "application/helpers/command[@type=\"render\" and @enginetype=\"ri\" and @engine=\"custom\" and @viewduringrender=\"false\"]")->Text());

	// Default render engine ...
	sdpXmlElementArray defaultengines = sdpXmlXPathSelect(root, "application/defaultengine");
	g_assert(1 == defaultengines.size());
	
	sdpString defaultenginetype, defaultengine, defaultenginename;
	sdpXmlGetAttribute(defaultengines[0], "enginetype", defaultenginetype);
	sdpXmlGetAttribute(defaultengines[0], "engine", defaultengine);

	sdpXmlElementArray engines = sdpXmlXPathSelect(root, "application/helpers/command[@type=\"render\" and @enginetype=\"ri\" and @viewduringrender=\"false\"]");
	sdpStringArray enginelist;
	
	for(gulong i = 0; i < engines.size(); i++)
		{
			sdpString enginetype, engine, enginename;
			sdpXmlGetAttribute(engines[i], "enginetype", enginetype);
			sdpXmlGetAttribute(engines[i], "engine", engine);
			sdpXmlGetAttribute(engines[i], "name", enginename);

			enginelist.push_back(enginename);
			
			if((defaultenginetype == enginetype) && (defaultengine == engine))
				defaultenginename = enginename;
		}
	Combo("defaultengine").SetList(enginelist);
	Combo("defaultengine").SetEntryText(defaultenginename);

	// Helper applications ...
	Entry("htmlviewer").SetText(sdpXmlGetXPathElement(root, "application/helpers/command[@type=\"htmlviewer\"]")->Text());
	Entry("bitmapviewer").SetText(sdpXmlGetXPathElement(root, "application/helpers/command[@type=\"bitmapviewer\"]")->Text());
	Entry("slpreprocessor").SetText(sdpXmlGetXPathElement(root, "application/helpers/command[@type=\"slpreprocessor\"]")->Text());
	Entry("slparser").SetText(sdpXmlGetXPathElement(root, "application/helpers/command[@type=\"slparser\"]")->Text());

	// Paths ...
	Entry("renderfarmpath").SetText(sdpXmlGetXPathElement(root, "application/paths/path[@type=\"renderfarm\"]")->Text());
}

void COptionsDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onhtmlviewer")
		OnHTMLViewer();
	else if(Event->Name() == "onbitmapviewer")
		OnBitmapViewer();
	else if(Event->Name() == "onslpreprocessor")
		OnSLPreprocessor();
	else if(Event->Name() == "onslparser")
		OnSLParser();
	else if(Event->Name() == "oncustomenginevisible")
		OnCustomEngineVisible();
	else if(Event->Name() == "oncustomenginehidden")
		OnCustomEngineHidden();
	else if(Event->Name() == "onrenderfarmpath")
		OnRenderFarmPath();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void COptionsDialog::OnDestroy()
{
	// Default render engine ...
	sdpXmlIElement* root = m_Options.GetRootElement();

	sdpXmlElementArray defaultengines = sdpXmlXPathSelect(root, "application/defaultengine");
	g_assert(1 == defaultengines.size());
	sdpXmlIElement* defaultengine = defaultengines[0];

	sdpXmlElementArray engines = sdpXmlXPathSelect(root, "application/helpers/command[@type=\"render\" and @enginetype=\"ri\" and @viewduringrender=\"false\"]");

	sdpString defaultenginename = Combo("defaultengine").GetEntryText();
			
	for(gulong i = 0; i < engines.size(); i++)
		{
			sdpString enginetype, engine, enginename;
			sdpXmlGetAttribute(engines[i], "enginetype", enginetype);
			sdpXmlGetAttribute(engines[i], "engine", engine);
			sdpXmlGetAttribute(engines[i], "name", enginename);

			if(enginename == defaultenginename)
				{
					defaultengine->CreateAttribute("enginetype", enginetype);
					defaultengine->CreateAttribute("engine", engine);
				}
		}

	// No more events from this point forward ...
	DisconnectAllEvents();
	
	// Cancel modal operation ....
	CancelModal();
	
	// Zap leftover widget pointers ...
	Clear();
	
	// Save the results
	m_Options.SaveOptions(Application.OptionsPath());
}

void COptionsDialog::OnHTMLViewer()
{
	sdpXmlGetXPathElement(m_Options.GetRootElement(), "application/helpers/command[@type=\"htmlviewer\"]")->SetText(Entry("htmlviewer").GetText());
}

void COptionsDialog::OnBitmapViewer()
{
	sdpXmlGetXPathElement(m_Options.GetRootElement(), "application/helpers/command[@type=\"bitmapviewer\"]")->SetText(Entry("bitmapviewer").GetText());
}

void COptionsDialog::OnSLPreprocessor()
{
	sdpXmlGetXPathElement(m_Options.GetRootElement(), "application/helpers/command[@type=\"slpreprocessor\"]")->SetText(Entry("slpreprocessor").GetText());
}

void COptionsDialog::OnSLParser()
{
	sdpXmlGetXPathElement(m_Options.GetRootElement(), "application/helpers/command[@type=\"slparser\"]")->SetText(Entry("slparser").GetText());
}

void COptionsDialog::OnCustomEngineVisible()
{
	sdpXmlGetXPathElement(m_Options.GetRootElement(), "application/helpers/command[@type=\"render\" and @enginetype=\"ri\" and @engine=\"custom\" and @viewduringrender=\"true\"]")->SetText(Entry("customenginevisible").GetText());
}

void COptionsDialog::OnCustomEngineHidden()
{
	sdpXmlGetXPathElement(m_Options.GetRootElement(), "application/helpers/command[@type=\"render\" and @enginetype=\"ri\" and @engine=\"custom\" and @viewduringrender=\"false\"]")->SetText(Entry("customenginehidden").GetText());
}

void COptionsDialog::OnRenderFarmPath()
{
	sdpXmlGetXPathElement(m_Options.GetRootElement(), "application/paths/path[@type=\"renderfarm\"]")->SetText(Entry("renderfarmpath").GetText());
}


