// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRenderJob class, which encapsulates a job to be rendered by the render farm
		\author Tim Shead (tshead@k-3d.com)
*/

#include "renderjob.h"
#include "application.h"

#include <sdpxmldocument.h>
#include <sdpfile.h>

#include <iostream>
#include <fstream>

CRenderJob::CRenderJob(k3dIRenderFarm* Farm, SDPCSTRING JobName)
{
	// Sanity checks ...
	g_assert(Farm);
	g_assert_string(JobName);
	
	// Store our data ...
	m_Farm = Farm;
	
	// Ensure that we end up with a unique name ...
	for(gulong i = 0; ; i++)
		{
			if(i)
				m_Name = sdpFormatString("%s-%u", JobName, i);
			else
				m_Name = JobName;
			
			if(!sdpFileExists(sdpAppendPath(Farm->Path(), m_Name)))
				break;
		}
		
	// Store our new path ...
	m_Path = sdpAppendPath(Farm->Path(), m_Name);
	
	// Create the directory ...
	if(!sdpCreateDirectory(m_Path))
		std::cerr << "CRenderJob::CRenderJob(): Error creating job directory \"" << m_Path << "\"" << std::endl;
}

CRenderJob::~CRenderJob()
{
	// Cleanup frames ...
	for(RenderFrameIterator frame = m_Frames.begin(); frame != m_Frames.end(); frame++)
		delete (*frame);
}

////////////////////////////////////////////////////////////////////////////////////
// k3dIRenderJob overrides ...
			
k3dIRenderFarm* CRenderJob::Farm()
{
	return m_Farm;
}

SDPCSTRING CRenderJob::Name()
{
	return m_Name;
}

SDPCSTRING CRenderJob::Path()
{
	return m_Path;
}

sdpResult CRenderJob::GetUniqueFile(SDPCSTRING File, SDPSTRING Buffer, gulong* BufferLength)
{
	// Sanity checks ...
	g_return_val_if_string_fail(File, SDP_E_FAIL);
	g_return_val_if_fail(Buffer, SDP_E_FAIL);
	g_return_val_if_fail(BufferLength, SDP_E_FAIL);
	g_return_val_if_fail(*BufferLength, SDP_E_FAIL);

	// Get the current time ...
	for(gulong i = 0; ; i++)
		{
			// Generate the path ...
			sdpString path;
			if(i)
				path = sdpAppendPath(Path(), sdpFormatString("%s-%u", File, i));
			else
				path = sdpAppendPath(Path(), File);

			// Make sure the file doesn't already exist ...
			if(sdpFileExists(path))
				continue;

			// Make sure the given buffer is large enough ...
			g_return_val_if_fail(path.size() <= *BufferLength, SDP_E_FAIL);

			// Copy the resulting path into the buffer ...
			strcpy(Buffer, SDPCSTRING(path));
			*BufferLength = path.size();

			return SDP_S_OK;
		}

	return SDP_E_FAIL;
}

k3dIRenderFrame* CRenderJob::CreateFrame(SDPCSTRING FrameName)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FrameName, 0);
	
	CRenderFrame* newframe = new CRenderFrame(this, FrameName);
	g_assert(newframe);

	if(Application.Verbose())
		std::cout << "CRenderJob::CreateFrame(): Created frame " << newframe->Name() << std::endl;
			
	m_Frames.push_back(newframe);
	return newframe;
}

//////////////////////////////////////////////////////////////////////////////
// Helper methods ...

bool CRenderJob::WriteControlFiles()
{
	// Create a control file for each frame ...
	for(RenderFrameIterator frame = m_Frames.begin(); frame != m_Frames.end(); frame++)
		g_return_val_if_fail((*frame)->WriteControlFile(), false);

	// Create the empty control file ...
	std::ofstream file(sdpAppendPath(Path(), ControlFile));
	g_return_val_if_fail(file.good(), false);
	
	// Create the XML document ...
	sdpXmlDocument document;
	sdpXmlIElement* root = document.CreateRootElement("k3dml");
	g_assert(root);
	
	// Write the document ...
	g_return_val_if_fail(document.Save(file), false);
	
	return true;
}

bool CRenderJob::MarkReady()
{
	// Mark each frame as "ready" ...
	for(RenderFrameIterator frame = m_Frames.begin(); frame != m_Frames.end(); frame++)
		g_return_val_if_fail((*frame)->MarkReady(), false);

	// Create the empty "ready" status file ...
	std::ofstream file(sdpAppendPath(Path(), StatusReadyFile));
	g_return_val_if_fail(file.good(), false);
	
	return true;
}


