// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CScriptEditor class, which provides a UI for editing / executing scripts
		\author Tim Shead (tshead@k-3d.com)
*/

#include "scripteditor.h"
#include "application.h"
#include "applicationframe.h"

#include <k3dcom.h>
#include <k3discriptengine.h>

#include <sdpgtkfileselector.h>
#include <sdpgtkevent.h>

#include <sdpxmlutility.h>

#include <sdpfile.h>

#include <strstream>
#include <fstream>

//////////////////////////////////////////////////////////////////////
// CScriptEditor

CScriptEditor::CScriptEditor(CApplicationFrame* Frame)
{
	// Sanity checks ...
	g_assert(Frame);
	
	m_Frame = Frame;
	m_ScriptEngine = 0;
}

CScriptEditor::~CScriptEditor()
{
	if(m_ScriptEngine)
		{
			m_ScriptEngine->Release();
			m_ScriptEngine = 0;
		}
}

bool CScriptEditor::Create()
{
	// Sanity checks ...

	// Load our GTKML template ...
	g_return_val_if_fail(LoadFile(sdpAppendPath(Application.DialogPath(), "scripteditor.gtkml")), false);

	// Attach scrollbars to the text widget ...
	GtkText* text = Text("script");
	
//	Scrollbar("hscrollbar").SetAdjustment(text->hadj);
	Scrollbar("vscrollbar").SetAdjustment(text->vadj);

	// Make it all visible ...
	RootWidget().Show();

	return true;
}

void CScriptEditor::OnEvent(sdpGtkEvent* Event)
{
	if(Event->Name() == "ondestroy")
		OnDestroy();
	else if(Event->Name() == "onfilenew")
		OnFileNew();
	else if(Event->Name() == "onfileopen")
		OnFileOpen();
	else if(Event->Name() == "onfilesave")
		OnFileSave();
	else if(Event->Name() == "onfilesaveas")
		OnFileSaveAs();
	else if(Event->Name() == "onfileclose")
		OnFileClose();
	else if(Event->Name() == "onexecute")
		OnExecute();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void CScriptEditor::OnDestroy()
{
//	std::cout << "CScriptEditor::OnDestroy()" << std::endl;

	// No more events from this point forward ...
	DisconnectAllEvents();

	// No need to keep track of widgets anymore ...
	Clear();
	
	// Tell the frame to get rid of us ...
	m_Frame->DestroyScriptEditor(this);
}

void CScriptEditor::OnFileNew()
{
	Text("script").DeleteText(0, -1);
	m_FilePath.clear();
}

void CScriptEditor::OnFileOpen()
{
	// Get our starting path - start with the user's home directory ...
	sdpString startpath = sdpGetHomeDirectory();
	
	// Next, override it with the user's options ...
	sdpXmlIElement* startpathelement = sdpXmlGetXPathElement(m_Frame->Application()->Options()->GetRootElement(), "application/paths/path[@type=\"script\"]");
	if(startpathelement)
		startpath = startpathelement->Text();
	
	// Last-but-not-least, see if we already have a script path ...
	if(m_FilePath.size())
		startpath = sdpDirectory(m_FilePath);

	// Append a directory separator (GtkFileSelector seems to want it) ...
	startpath += G_DIR_SEPARATOR;

	// Put up a file selection dialog ...
	sdpGtkFileSelector selector("Open K-3D Script:", startpath);
	selector.DoModal(RootWindow());
	if(!selector.OK())
		return;

	// Store the new path ...
	if(startpathelement)
		startpathelement->SetText(sdpDirectory(selector.FilePath()));

	// Try opening the file ...
	std::ifstream file(selector.FilePath());
	if(!file.good())
		{
			Application.Error("Error opening script!", "Open K-3D Script:");
			return;
		}

	// Get rid of old text ...	
	Text("script").DeleteText(0, -1);
	
	// Load the file into our text widget ...
	std::strstream script;
	file.get(*script.rdbuf(), '\0');
	script << std::ends;
		
	gint position = 0;
	Text("script").InsertText(script.str(), strlen(script.str()), &position);
	
	script.freeze(false);
	
	// Save the path ...
	m_FilePath = selector.FilePath();	
}

void CScriptEditor::OnFileSave()
{
	// If we don't have a filepath, prompt the user ...
	if(m_FilePath.size())
		SaveScript();
	else
		OnFileSaveAs();
}

void CScriptEditor::OnFileSaveAs()
{
	// Get our starting path - start with the user's home directory ...
	sdpString startpath = sdpGetHomeDirectory();
	
	// Next, override it with the user's options ...
	sdpXmlIElement* startpathelement = sdpXmlGetXPathElement(m_Frame->Application()->Options()->GetRootElement(), "application/paths/path[@type=\"script\"]");
	if(startpathelement)
		startpath = startpathelement->Text();
	
	// Last-but-not-least, see if we already have a script path ...
	if(m_FilePath.size())
		startpath = sdpDirectory(m_FilePath);

	// Append a directory separator (GtkFileSelector seems to want it) ...
	startpath += G_DIR_SEPARATOR;

	// Prompt the user for a filepath ...
	sdpGtkFileSelector selector("Save K-3D Script As:", startpath);
	selector.DoModal(RootWindow());
	
	// If the user cancelled, we're done ...
	if(!selector.OK())
		return;

	// Store the path and save the script ...		
	m_FilePath = selector.FilePath();

	if(startpathelement)
		startpathelement->SetText(sdpDirectory(selector.FilePath()));

	SaveScript();
}

void CScriptEditor::SaveScript()
{
	// Sanity checks ...
	g_assert(m_FilePath.size());
	
	// Try opening the file ...
	std::ofstream file(m_FilePath);
	if(!file.good())
		{
			Application.Error("Error saving script!", "Save K-3D Script:");
			return;
		}	
	
	// Get the script text from our widget ...
	sdpString script = Text("script").GetText(0, -1);
	file.write(script, script.size());
}

void CScriptEditor::OnFileClose()
{
	// Shut ourselves down ...
	RootWidget().Destroy();
}

void CScriptEditor::OnExecute()
{
	// Get a scripting engine, if we don't already have one ...
	if(0 == m_ScriptEngine)
		m_ScriptEngine = Application.ScriptEngine(sdpJavaScriptEngine);
		
	// If we couldn't find anything, let the user know and exit ...
	if(0 == m_ScriptEngine)
		{
			Application.Error("Unable to initialize scripting engine!", "Execute K-3D Script:");
			g_return_if_fail(0);
		}
		
	// Evaluate a simple script ...
	sdpString script = Text("script").GetText(0, -1);
	m_ScriptEngine->Execute(script);
}


