// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements classes for encapsulating information about a Pixar Renderman Shading Language shader
		\author Tim Shead (tshead@k-3d.com)
*/

#include "shaderinfo.h"
#include "application.h"

#include <k3ddocument.h>
#include <sdpxmlutility.h>

#include <iostream>
#include <strstream>
#include <sdpfile.h>

////////////////////////////////////////////////////////////////
// CShaderInfoArgument

CShaderInfoArgument::CShaderInfoArgument(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Element->Name() == sdpString("argument"));

	// Get the argument name ...
	sdpXmlGetAttribute(Element, "name", m_Name);
	m_Properties["name"] = m_Name;
	
	// Get the argument type ...
	sdpString type;
	sdpXmlGetAttribute(Element, "type", type);
	
	if(type == "float")
		m_Type = k3dIShaderInfoArgument::SCALAR;
	else if(type == "color")
		m_Type = k3dIShaderInfoArgument::COLOR;		
	else if(type == "point")
		m_Type = k3dIShaderInfoArgument::POINT;		
	else if(type == "vector")
		m_Type = k3dIShaderInfoArgument::VECTOR;		
	else if(type == "normal")
		m_Type = k3dIShaderInfoArgument::NORMAL;		
	else if(type == "string")
//		m_Type = k3dIShaderInfoArgument::STRING;	
		m_Type = k3dIShaderInfoArgument::TEXTURE;	
	else
		m_Type = k3dIShaderInfoArgument::UNKNOWN;
		
	m_Properties["type"] = type;

	// Get the default value ...
	sdpXmlGetAttribute(Element, "default", m_Default);
	m_Properties["default"] = m_Default;
	
	// Get the reference frame of the default value ...
	sdpString defaultframe;
	sdpXmlGetAttribute(Element, "defaultframe", defaultframe);
	m_Properties["defaultframe"] = defaultframe;

	// Get the storage type ...
	sdpString storage;
	sdpXmlGetAttribute(Element, "storage", storage);
	m_Properties["storage"] = storage;
}

/////////////////////////////////////////////////////////////////////////////
// k3dIShaderInfoArgument overrides ...

SDPCSTRING CShaderInfoArgument::Name()
{
	return m_Name;
}

k3dIShaderInfoArgument::ArgumentType CShaderInfoArgument::Type()
{
	return m_Type;
}

SDPCSTRING CShaderInfoArgument::Default()
{
	return m_Default;
}

////////////////////////////////////////////////////////////////////
// CShaderInfoArgumentArray

CShaderInfoArgumentArray::~CShaderInfoArgumentArray()
{
	// Cleanup our shader arguments ...
	for(ShaderInfoArgumentIterator i = begin(); i != end(); i++)
		delete (*i);
}

////////////////////////////////////////////////////////////////
// CShaderInfo

CShaderInfo::CShaderInfo(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_assert_string(FilePath);

	// Store the filepath ...
	m_FilePath = FilePath;
	m_Properties["filepath"] = m_FilePath;

	// Get the user's preferred shading-language preprocessor ...
	sdpXmlIElement* preprocessor = sdpXmlGetXPathElement(Application.Options()->GetRootElement(), "application/helpers/command[@type=\"slpreprocessor\"]");
	g_return_if_fail(preprocessor);
	sdpString preprocessorcommand = preprocessor->Text();
	preprocessorcommand.formattedreplace('%', "p", FilePath);

	// Get the user's preferred shading-language-to-XML parser ...	
	sdpXmlIElement* parser = sdpXmlGetXPathElement(Application.Options()->GetRootElement(), "application/helpers/command[@type=\"slparser\"]");
	g_return_if_fail(parser);
	sdpString parsercommand = parser->Text();
	
	// Create a command line that pipes the shader source through the preprocessor to the parser, and captures the parser's output
	sdpString commandline = preprocessorcommand + " | " + sdpAppendPath(Application.BasePath(), parsercommand);

	// Make it happen ...
	std::strstream buffer;
	sdpRunProcess(commandline, buffer);
	buffer << std::ends;
	
	// Parse the shader description ...
	k3dDocument shaderinfo;	
	g_return_if_fail(shaderinfo.Load(buffer));
	
	// Get the first child, which should be a shader ...
	sdpXmlIElement* shader = shaderinfo.GetRootElement()->GetElement(shaderinfo.GetRootElement()->GetFirstElement());
	g_return_if_fail(shader->Name() == sdpString("shader"));

	// Store the shader name ...
	sdpXmlGetAttribute(shader, "name", m_Name);
	m_Properties["name"] = m_Name;

	// Store the shader type ...
	sdpString type;
	sdpXmlGetAttribute(shader, "type", type);
	m_Properties["type"] = type;
	
	if(type == "surface")
		m_Type = k3dIShaderInfo::SURFACE;
	else if(type == "light")
		m_Type = k3dIShaderInfo::LIGHT;
	else if(type == "displacement")
		m_Type = k3dIShaderInfo::DISPLACEMENT;
	else if(type == "volume")
		m_Type = k3dIShaderInfo::VOLUME;
	else if(type == "transformation")
		m_Type = k3dIShaderInfo::TRANSFORMATION;
	else if(type == "imager")
		m_Type = k3dIShaderInfo::IMAGER;
		
	// For each shader argument ...
	for(SDPXMLELEMENTHANDLE e = shader->GetFirstElement(); e; e = shader->GetNextElement(e))
		m_Arguments.push_back(new CShaderInfoArgument(shader->GetElement(e)));
}

/////////////////////////////////////////////////////////////////////////////////////
// k3dIShader overrides ...

SDPCSTRING CShaderInfo::Name()
{
	return m_Name;
}

k3dIShaderInfo::ShaderType CShaderInfo::Type()
{
	return m_Type;
}

K3DSHADERINFOARGUMENTHANDLE CShaderInfo::GetFirstArgument()
{
	return K3DSHADERINFOARGUMENTHANDLE(m_Arguments.size() ? 1 : 0);
}

K3DSHADERINFOARGUMENTHANDLE CShaderInfo::GetNextArgument(K3DSHADERINFOARGUMENTHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	
	return K3DSHADERINFOARGUMENTHANDLE(gulong(Handle) < m_Arguments.size() ? gulong(Handle)+1 : 0);
}

k3dIShaderInfoArgument* CShaderInfo::GetArgument(K3DSHADERINFOARGUMENTHANDLE Handle)
{
	// Sanity checks ...
	g_return_val_if_fail(Handle, 0);
	g_return_val_if_fail(gulong(Handle) <= m_Arguments.size(), 0);
	
	return m_Arguments[gulong(Handle)-1];
}

///////////////////////////////////////////////////////////////////////////////////
// CShaderInfoList

CShaderInfoList::~CShaderInfoList()
{
	// Cleanup shaders ...
	for(ShaderInfoIterator i = begin(); i != end(); i++)
		delete (*i);
}

// Load shaders from the given directory ...
bool CShaderInfoList::LoadShaders(SDPCSTRING Directory)
{
	// Sanity checks ...
	g_assert_string(Directory);

	if(Application.Verbose())
		std::cout << "Loading shaders from " << Directory << std::endl;
	
	return sdpForEachFile(Directory, *this, false);
}

// Called by sdpForEachFile ...
bool CShaderInfoList::HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName)
{
	sdpString path(FilePath);

	// Skip files that aren't shader source code ...
	if(path.right(3) != ".sl")
		return true;

	// Some basic tracing ...
	if(Application.Verbose())
		std::cout << "  Loading shader " << FilePath << std::endl;
	
	// Create a shader info object ...
	CShaderInfo* shaderinfo = new CShaderInfo(path);
	push_back(shaderinfo);

	return true;
}

// Load shaders from the user's SHADERS environment variable ...
bool CShaderInfoList::LoadShaders()
{
	// Get the user's SHADERS path ...
	sdpString shaderspath = Application.ShadersPath();

	// Parse all the paths within SHADERS ...
	sdpString shaderdirectory;
	
	sdpString::size_type index = shaderspath.find(G_SEARCHPATH_SEPARATOR);
	while(index != sdpString::npos)
		{
			shaderdirectory = shaderspath.left(index);
			shaderspath = shaderspath.right(shaderspath.size() - (index + 1));

			if(shaderdirectory.size())
				LoadShaders(shaderdirectory);

			index = shaderspath.find(G_SEARCHPATH_SEPARATOR);
		}

	shaderdirectory = shaderspath;
	if(shaderdirectory.size())
		LoadShaders(shaderdirectory);

	return true;
}

