#ifndef K3DIMAGEBUFFER_H
#define K3DIMAGEBUFFER_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dImageBuffer class, which can store and manipulate an in-memory, 32-bit RGBA bitmap image
		\author Tim Shead (tshead@k-3d.com)
*/

///////////////////////////////////////////////////////////////////////////////
// k3dImageBuffer

#include <k3dibitmap.h>

/// Encapsulates an in-memory 32-bit RGBA bitmap image
class k3dImageBuffer
{
public:
	typedef enum
	{
		NEAREST,
		BILINEAR,
		BICUBIC
	} ScaleType;

	/// Standard constructor
	k3dImageBuffer();
	/// Copy constructor, allocates a new buffer and makes a copy
	k3dImageBuffer(k3dImageBuffer& Buffer);
	/// Copy constructor with explicit arguments
	k3dImageBuffer(const k3dPixel* const Data, const gulong Width, const gulong Height);

	/// Standard destructor
	~k3dImageBuffer();

	/// Allocates memory for an image with the given width and height
	bool Allocate(const gulong Width, const gulong Height);
	/// Makes a copy of the given image
	bool Copy(k3dImageBuffer& Buffer);
	/// Makes a copy of the given image
	bool Copy(const k3dPixel* const Data, const gulong Width, const gulong Height);
	/// Frees all allocated memory
	void Free();

	/// Flips the contents of the buffer vertically
	bool FlipVertical();
	/// Flips the contents of the buffer horizontally
	bool FlipHorizontal();

	/// Scales the buffer proportionately by the given percentage
	bool Scale(const gdouble Percent, const ScaleType Type);
	/// Scales the buffer by the given percentages
	bool Scale(const gdouble WidthPercent, const gdouble HeightPercent, const ScaleType Type);
	/// Scales the buffer to the given dimensions
	bool Scale(const gulong NewWidth, const gulong NewHeight, const ScaleType Type);
	/// Alters the buffer for use with OpenGL
	bool ConvertForOpenGL(const ScaleType Type);
	/// Alters the buffer for use with OpenGL, and constrains the image dimensions
	bool ConvertForOpenGL(const gulong MaxWidth, const gulong MaxHeight, const ScaleType Type);

	/// Returns a pointer to the image buffer
	k3dPixel* Data() { return m_Data; }
	/// Returns the width of the image buffer in pixels
	gulong Width() { return m_Width; }
	/// Returns the height of the image buffer in pixels
	gulong Height() { return m_Height; }

protected:
	/// Stores the allocated image buffer
	k3dPixel* m_Data;
	/// Stores the width of the image buffer in pixels
	gulong m_Width;
	/// Stores the height of the image buffer in pixels
	gulong m_Height;
};

#endif // K3DIMAGEBUFFER_H
