#ifndef K3DIMESH_H
#define K3DIMESH_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dIMesh interface, which encapsulates a geometric object with points, polygons, and NURBS curves
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <sdpcom.h>
#include <vector>
#include <list>
#include <set>

// Forward class declarations
class k3dIObject;
class k3dIDocument;
class k3dITransform;
class k3dIPolygonServer;
class k3dINURBSCurveServer;
class k3dIPointServer;
class k3dIPolygon;
class k3dINURBSCurve;
class k3dIPoint;

SDP_DECLARE_HANDLE(K3DMESHHANDLE);

/// Encapsulates a geometric object with points, polygons, and a linear transformation
class k3dIMesh : public virtual sdpIUnknown
{
public:
	/// Stores the sdpCom interface ID
	static const sdpIID iid;

	/// Returns the parent document
	virtual k3dIDocument* Document() = 0;
	/// Returns an interface for accessing polygons
	virtual k3dIPolygonServer* Polygons() = 0;
	/// Creates a new polygon
	virtual k3dIPolygon* CreatePolygon() = 0;
	/// Deletes an existing polygon
	virtual sdpResult DeletePolygon(k3dIPolygon* Polygon, bool DeletePoints) = 0;
	/// Returns an interface for accessing NURBS curves
	virtual k3dINURBSCurveServer* NURBSCurves() = 0;
	/// Creates a new NURBS curve
	virtual k3dINURBSCurve* CreateNURBSCurve() = 0;
	/// Deletes an existing NURBS curve
	virtual sdpResult DeleteNURBSCurve(k3dINURBSCurve* Curve, bool DeletePoints) = 0;
	/// Returns an interface for accessing points
	virtual k3dIPointServer* Points() = 0;
	/// Creates a new point
	virtual k3dIPoint* CreatePoint() = 0;
	/// Deletes an existing point
	virtual sdpResult DeletePoint(k3dIPoint* Point) = 0;
	/// Sets smooth-shading for any polygons that meet the given state and angle criteria
	virtual sdpResult ShadePolygons(gulong Criteria, gdouble Angle) = 0;
	/// Removes smooth-shading from any polygons that meet the given state criteria
	virtual sdpResult UnshadePolygons(gulong Criteria) = 0;
	/// Forces a recalculation of polygon normals, including smooth-shading
	virtual sdpResult CalculateNormals() = 0;
	/// Returns the dimensions of all points that meet the given criteria
	virtual sdpResult GetDimensions(gulong Criteria, double Min[3], double Max[3]) = 0;
	/// Returns the mesh color (for use in editor views)
	virtual sdpResult GetColor(double Color[3]) = 0;
	/// Sets the mesh color (for use in editor views)
	virtual sdpResult SetColor(double Color[3]) = 0;
};

/// Contains an array of meshes
typedef std::vector<k3dIMesh*> k3dMeshArray;
/// Contains a linked-list of meshes
typedef std::list<k3dIMesh*> k3dMeshList;
/// Contains a set of meshes
typedef std::set<k3dIMesh*> k3dMeshSet;

#endif // K3DIMESH_H
