#ifndef K3DEDGE_H
#define K3DEDGE_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dEdge class, which encapsulates a polygon edge (defined by two instances of k3dIPoint)
		\author Tim Shead (tshead@k-3d.com)
*/

#include <set>
#include <vector>
#include <algorithm>

#include <sdpsimpletypes.h>

//  Forward class declarations ...
class k3dIDocument;
class k3dIMesh;
class k3dIPolygon;
class k3dIPoint;

/////////////////////////////////////////////////////////////////////////
// k3dEdge

/// Represents a polygon edge, comprised of two points
class k3dEdge
{
public:
	k3dEdge(k3dIPoint* Point1, k3dIPoint* Point2)
	{
		// Sanity checks ...
		g_assert(Point1);
		g_assert(Point2);
	
		m_Point1 = Point1;
		m_Point2 = Point2;
	
		if(Point1 > Point2)
			{
				std::swap(m_Point1, m_Point2);
				m_Flipped = true;
			}
		else
			{
				m_Flipped = false;
			}
	}

	const k3dIPoint* Point1() const { return m_Point1; }
	const k3dIPoint* Point2() const { return m_Point2; }
	const bool Flipped() const { return m_Flipped; }

protected:
	/// Stores the first edge point
	k3dIPoint* m_Point1;
	/// Stores the second edge point
	k3dIPoint* m_Point2;
	/// Stores whether the edge points were flipped from their original orientation
	bool m_Flipped;
};

/// Less-than operator for comparing edges
inline bool operator < (const k3dEdge& A, const k3dEdge& B)
{
	if(A.Point1() == B.Point1())
		return A.Point2() < B.Point2();
	else
		return A.Point1() < B.Point1();
}

/// Equality operator for comparing edges
inline bool operator == (const k3dEdge& A, const k3dEdge& B) { return (A.Point1() == B.Point1()) && (A.Point2() == B.Point2()); }

/// Stores a set of unique edges
typedef std::set<k3dEdge> k3dEdgeSet;
typedef k3dEdgeSet::iterator k3dEdgeSetIterator;

/// Stores an array of edges
typedef std::vector<k3dEdge> k3dEdgeArray;
typedef k3dEdgeArray::iterator k3dEdgeArrayIterator;

/// Converts a polygon to a set of edges
void k3dGetEdges(k3dIPolygon* Polygon, k3dEdgeSet& Results);
/// Converts a mesh to a set of edges
void k3dGetEdges(k3dIMesh* Mesh, k3dEdgeSet& Results);
/// Creates a set of edges from every polygon that meets the given criteria
void k3dGetEdges(k3dIDocument* Document, gulong Criteria, k3dEdgeSet& Results);


#endif // K3DEDGE_H
