#ifndef K3DPOLYGON_H
#define K3DPOLYGON_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dPolygon class, which encapsulates a geometric polygon
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dipolygon.h>
#include <k3dipolypointserver.h>
#include <k3dilocationserver.h>
#include <k3dimaterialserver.h>
#include <k3dstate.h>
#include <k3dactionclient.h>

#include <sdpmath.h>

// Forward class declarations
class k3dMesh;
class k3dPolyPoint;

///////////////////////////////////////////////////////////////////////////////////////////
// k3dPolygon

/// Encapsulates a closed polygon / Bezier spline / Bezier patch
class k3dPolygon : public k3dIPolygon,
								public k3dIPolyPointServer,
								public k3dILocationServer,
								public k3dIMaterialServer,
								public k3dState,
								public k3dActionClient
{
public:
	k3dPolygon(k3dMesh* Mesh, k3dID ID);
	virtual ~k3dPolygon();

	// k3dIUnknown overrides ...
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface) { return SDP_E_NO_INTERFACE; }
	gulong AddRef() { return 0; }
	gulong Release() { return 0; }

	// k3dIPolygon overrides ...
	k3dIMesh* Mesh();
	k3dIPolyPointServer* Points();
	k3dILocationServer* Locations();
	k3dIMaterialServer* Material();
	k3dIState* State();
	sdpResult GetPolygonNormal(double Normal[3]);
	sdpResult SetPolygonNormal(double Normal[3]);
	const k3dID ID();

	// k3dIPolyPointServer overrides ...
	gulong GetPointCount();
	K3DPOLYPOINTHANDLE GetFirstPolyPoint();
	K3DPOLYPOINTHANDLE GetNextPolyPoint(K3DPOLYPOINTHANDLE Point);
	k3dIPolyPoint* GetPolyPoint(K3DPOLYPOINTHANDLE Point);
	k3dIPolyPoint* AddPoint(k3dIPoint* Point);
	sdpResult RemovePoint(k3dIPoint* Point);
	sdpResult RemovePolyPoint(k3dIPolyPoint* PolyPoint);

	// k3dILocationServer overrides ...
	gulong GetLocationCount();
	K3DLOCATIONHANDLE GetFirstLocation();
	K3DLOCATIONHANDLE GetNextLocation(K3DLOCATIONHANDLE Location);
	k3dILocation* GetLocation(K3DLOCATIONHANDLE Location);

	// k3dIMaterialServer overrides ...
	k3dIMaterial* GetMaterial();
	sdpResult SetMaterial(k3dIMaterial* Material);

	// k3dIState overrides ...
	sdpResult SetState(gulong Mask, bool State);

	// k3dIActionClient overrides ...
	sdpResult HandleStep(gulong ValueCode, gulong Value, gulong ValueSize);

	// k3dIPersist overrides ...
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);
	sdpResult Load(sdpXmlIElement* Element, k3dMesh* Mesh);
	sdpResult LoadComplete();

	/// Draws a wireframe representation of the polygon
	void OpenGLWireframe();
	/// Draws a wireframe representation of the polygon with orientation
	void OpenGLWireframeOrientation(sdpVector3& FirstColor, sdpVector3& Color);
	/// Draws a solid representation of the polygon
	void OpenGLSolid();
	/// Draws a solid representation of the polygon
	void OpenGLSolidTextured();

	/// Forces recalculation of the polyon normals
	void CalculateNormal();

	/// Stores the material assigned to this polygon
	k3dIMaterial* m_Material;

	/// Stores the polygon points attached to this polygon
	typedef std::vector<k3dPolyPoint*> PolyPoints;
	typedef PolyPoints::iterator PolyPointIterator;
	PolyPoints m_PolyPoints;

	/// Stores the parent mesh that owns this polygon
	k3dMesh* m_Mesh;
	/// Stores the unique (within the parent mesh) polygon ID
	k3dID m_ID;
	/// Stores the polygon normal
	sdpVector3 m_Normal;
	
	/// Enumerates undo-able steps
	typedef enum
	{
		STEP_STATE,
	} StepType;
};

#endif K3DPOLYGON_H
