#ifndef K3DRISHADERSETUP_H
#define K3DRISHADERSETUP_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares convenience classes for sending rendering parameters to a k3dIRiEngine interface
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpstring.h>
#include <sdpmath.h>

#include <k3dicamera.h>
#include <k3driengine.h>

///////////////////////////////////////////////////////////////////////////
// k3dRiParameterList

/// Encapsulates a set of Pixar Renderman Interface parameters
class k3dRiParameterList
{
public:
	k3dRiParameterList();
	~k3dRiParameterList();

	/// Adds an integer value to the list
	void AddInteger(SDPCSTRING Token, const int Arg);
	/// Adds a scalar (floating point) value to the list
	void AddScalar(SDPCSTRING Token, const double Arg);
	/// Adds a 3D point value to the list
	void AddPoint(SDPCSTRING Token, const sdpVector3& Arg);
	/// Adds an RGB color value to the list
	void AddColor(SDPCSTRING Token, const sdpVector3& Arg);
	/// Adds a string to the list
	void AddString(SDPCSTRING Token, SDPCSTRING Arg);
	/// Adds an untyped pointer to the list
	void AddPointer(SDPCSTRING Token, const gpointer Arg);

	/// Declares the type of each parameter to Renderman
	void DeclareParameters(k3dIRiEngine* Engine);
	/// Passes the parameters to Renderman as options
	void SetOptions(SDPCSTRING Name, k3dIRiEngine* Engine);
	/// Passes the parameters to Renderman as attributes
	void SetAttributes(SDPCSTRING Name, k3dIRiEngine* Engine);

	/// Frees all storage
	void ClearAll();

	/// Returns a pointer to the array of tokens (parameter names)
	SDPCSTRING* Tokens() { return &m_Tokens.front(); }
	/// Returns a pointer to the array of parameters (values)
	gpointer* Parameters() { return &m_Parameters.front(); }
	/// Returns the number of parameters
	gulong ParameterCount() { return m_Parameters.size(); }

protected:
	/// Looks to see if a parameter is already defined
	bool FindToken(SDPCSTRING Token);

	/// Stores the array of tokens to be passed to Renderman
	std::vector<const char*> m_Tokens;
	/// Stores the array of parameter values to be passed to Renderman
	std::vector<gpointer> m_Parameters;
	/// Stores the array of parameter types used for declarations to Renderman
	std::vector<sdpString> m_Types;

	/// Stores any integers added to the list
	std::list<gint> m_Ints;
	/// Stores any floating-point values added to the list
	std::list<gdouble> m_Floats;
	/// Stores any 3D points added to the list
	std::list<sdpVector3> m_Points;
	/// Stores any RGB colors added to the list
	std::list<sdpVector3> m_Colors;
	/// Stores any strings added to the list
	std::list<sdpString> m_Strings;
	/// Stores pointers to any strings added to the list
	std::list<gpointer> m_StringPointers;
};

///////////////////////////////////////////////////////////////////////////
// k3dRiShader

/// Encapsulates a Renderman Interface shader: a shader name, and a collection of parameters
class k3dRiShader : public k3dRiParameterList
{
public:
	// Default constructor, normally used with a ShaderSelector ...
	k3dRiShader();

	// Constructor for use as a stand-alone object ...
	k3dRiShader(SDPCSTRING ShaderName);

	/// Sets the name of the shader
	void SetShader(SDPCSTRING ShaderName);
	/// Returns the shader name
	SDPCSTRING ShaderName() { return m_ShaderName; }

protected:
	sdpString m_ShaderName;
};

////////////////////////////////////////////////////////////////////////////
// k3dRiLightShader

/// Encapsulates a Renderman Interface light shader
class k3dRiLightShader : public k3dRiShader
{
public:
	k3dRiLightShader();
	k3dRiLightShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to Renderman as a normal light source
	gulong SetupLight(k3dIRiEngine* Engine);
	/// Sends the parameter list to Renderman as an area light source
	gulong SetupAreaLight(k3dIRiEngine* Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiVolumeShader

/// Encapsulates a Renderman Interface volume shader
class k3dRiVolumeShader : public k3dRiShader
{
public:
	k3dRiVolumeShader();
	k3dRiVolumeShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to Renderman as an interior volume
	sdpResult SetupInterior(k3dIRiEngine* Engine);
	/// Sends the parameter list to Renderman as an exterior volume
	sdpResult SetupExterior(k3dIRiEngine* Engine);
	/// Sends the parameter list to Renderman as an atmosphere shader
	sdpResult SetupAtmosphere(k3dIRiEngine* Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiTransformationShader

/// Encapsulates a Renderman Interface transformation shader
class k3dRiTransformationShader : public k3dRiShader
{
public:
	k3dRiTransformationShader();
	k3dRiTransformationShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to Renderman as a transformation shader
	sdpResult Setup(k3dIRiEngine* Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiDisplacementShader

/// Encapsulates a Renderman Interface displacement shader
class k3dRiDisplacementShader : public k3dRiShader
{
public:
	k3dRiDisplacementShader();
	k3dRiDisplacementShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to Renderman as a displacement shader
	sdpResult Setup(k3dIRiEngine* Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiSurfaceShader

/// Encapsulates a Renderman Interface surface shader
class k3dRiSurfaceShader : public k3dRiShader
{
public:
	k3dRiSurfaceShader();
	k3dRiSurfaceShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to Renderman as a surface shader
	sdpResult Setup(k3dIRiEngine* Engine);
};

////////////////////////////////////////////////////////////////////////////
// k3dRiImagerShader

/// Encapsulates a Renderman Interface imaging shader
class k3dRiImagerShader : public k3dRiShader
{
public:
	k3dRiImagerShader();
	k3dRiImagerShader(SDPCSTRING ShaderName);

	/// Sends the parameter list to Renderman as an imaging shader
	sdpResult Setup(k3dIRiEngine* Engine);
};

#endif // K3DRISHADERSETUP_H
