#ifndef K3DSHADERINSTANCE_H
#define K3DSHADERINSTANCE_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dShaderInstance class, which manages an instance of a Pixar Renderman shader and its arguments
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3diapplication.h>
#include <k3didocument.h>
#include <k3ditexture.h>
#include <k3dvariable.h>
#include <k3drishadersetup.h>
#include <list>


///////////////////////////////////////////////////////////////////////////
// k3dShaderInstance

/// Manages an instance of a shader and its arguments
class k3dShaderInstance
{
friend class k3dShaderInstanceLoadShaderInstanceArguments;
public:
	///////////////////////////////////////////////////////////////////////////
	// ArgumentBase

	/// Base class for shader instance arguments
	class ArgumentBase
	{
	public:
		/// Static function for creating arguments based on type
		static ArgumentBase* CreateArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default);
		/// Static function for loading arguments
		static ArgumentBase* LoadArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Value);

		ArgumentBase(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type);
		ArgumentBase(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default);
		virtual ~ArgumentBase();

		/// Initializes the argument and sets its default value
		virtual bool Create() = 0;
		
		/// Loads the argument
		virtual bool LoadComplete() = 0;

		/// Returns the argument's document
		k3dIDocument* Document() { return m_Document; }
		/// Returns the argument's name
		const sdpString& Name() { return m_Name; }
		/// Returns the argument's type
		k3dIShaderInfoArgument::ArgumentType Type() { return m_Type; }
		/// Returns the argument's default value
		const sdpString& Default() { return m_Default; }
		/// Returns the argument's current value as a string
		virtual sdpString ConvertToString(gdouble Time, gdouble DeltaTime) = 0;

	protected:
		/// Stores the owning document
		k3dIDocument* m_Document;
		/// Stores the argument name
		sdpString m_Name;
		/// Stores the argument type
		k3dIShaderInfoArgument::ArgumentType m_Type;
		/// Stores the argument default value
		sdpString m_Default;
	};

	///////////////////////////////////////////////////////////////////////////
	// Argument

	template<class ValueType>
	class Argument : public ArgumentBase
	{
	public:
		Argument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type)
			: ArgumentBase(Document, Name, Type)
		{
		}
	
		Argument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default)
			: ArgumentBase(Document, Name, Type, Default)
		{
		}

	bool Create()
	{
		m_Value.ConvertFromString(m_Default);
		return true;
	}
	
	bool LoadComplete()
	{
		m_Value.ConvertFromString(m_Default);
		return true;
	}

	/// Sets the argument's value
	void SetValue(const ValueType& Value) { m_Value.SetValue(Value); }
	/// Returns the argument's value
	ValueType Value() { return m_Value.Value(); }
	/// Accesses the argument's value by reference
	ValueType& operator()() { return m_Value(); }

	sdpString ConvertToString(gdouble Time, gdouble DeltaTime) { return m_Value.ConvertToString(); }

	protected:
		/// Stores the value of this argument
		k3dVariable<ValueType> m_Value;
	};

	///////////////////////////////////////////////////////////////////////////
	// ChannelArgument

	/// Encapsulates shader arguments that are controlled by channels
	class ChannelArgument : public ArgumentBase
	{
	public:
		ChannelArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type);
		ChannelArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default);

		bool Create();
		bool LoadComplete();

		/// Returns the argument's object container
		k3dChannelVariable& ObjectContainer() { return m_Channel; }
		/// Returns the argument's channel
		k3dIChannel* Channel() { return m_Channel.Interface(); }
		/// Returns the argument's channel ID
		k3dID ID() { return m_Channel.Object()->ObjectContext()->ObjectID(); }

		sdpString ConvertToString(gdouble Time, gdouble DeltaTime);

	protected:
		/// Stores the channel for this argument
		k3dChannelVariable m_Channel;
	};

	///////////////////////////////////////////////////////////////////////////
	// TextureArgument

	/// Encapsulates shader arguments that are represented by textures
	class TextureArgument : public ArgumentBase
	{
	public:
		TextureArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type);
		TextureArgument(k3dIDocument* Document, SDPCSTRING Name, k3dIShaderInfoArgument::ArgumentType Type, SDPCSTRING Default);

		bool Create();
		bool LoadComplete();

		/// Returns the argument's object container
		k3dObjectVariable<k3dITexture>& ObjectContainer() { return m_Texture; }
		/// Returns the argument's texture
		k3dITexture* Texture() { return m_Texture.Interface(); }
		/// Returns the argument's texture ID
		k3dID ID() { return m_Texture.Object()->ObjectContext()->ObjectID(); }

		sdpString ConvertToString(gdouble Time, gdouble DeltaTime);

	protected:
		/// Stores the texture for this argument
		k3dObjectVariable<k3dITexture> m_Texture;
	};

	k3dShaderInstance();
	~k3dShaderInstance();

	/// Sets the owning document
	bool Create(k3dIDocument* Document);
	/// Sets the current shader (releases old data, channels, etc.)
	bool SetShader(SDPCSTRING Shader);
	/// Sets the current shader (releases old data, channels, etc.)
	bool SetShaderHandle(K3DSHADERINFOHANDLE ShaderHandle);
	/// Returns the primary (for Editor & OpenGL rendering purposes) texture argument
	k3dITexture* GetPrimaryTexture();

	/// Called to poke our arguments into a parameter list
	void Initialize(k3dIRenderFrame* Frame, k3dIRiEngine* Engine, k3dRiShader& Shader, gdouble Time, gdouble DeltaTime);

	/// Returns the current shader handle
	K3DSHADERINFOHANDLE ShaderHandle() { return m_ShaderHandle; }
	/// Returns the current shader name
	SDPCSTRING ShaderName();
	/// Returns the total number of arguments managed by this instance
	gulong ArgumentCount() { return m_Arguments.size(); }
	/// Returns the given argument by index
	ArgumentBase* ArgumentAt(gulong Index);

	/// Stores the shader instance data
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies, SDPCSTRING Name);
	/// Loads the shader instance data
	sdpResult Load(sdpXmlIElement* Element, SDPCSTRING Name);
	/// Initializes the shader instance data after load
	sdpResult LoadComplete();

protected:
	/// Allocates storage, channels, etc. for a shader
	bool Allocate();
	/// Releases old storage, channels, etc.
	void Release();

	/// Stores the owning application
	k3dIApplication* m_Application;
	/// Stores the owning document
	k3dIDocument* m_Document;
	/// Stores the current shader handle
	K3DSHADERINFOHANDLE m_ShaderHandle;
	/// Store the primary texture (for use with Editor & OpenGL rendering) if any ...
	TextureArgument* m_PrimaryTexture;

	/// A list of arguments
	typedef std::list<ArgumentBase*> Arguments;
	typedef Arguments::iterator ArgumentIterator;
	/// Stores the arguments for this shader instance
	Arguments m_Arguments;

	// Some common argument types ...
	typedef Argument<sdpString> StringArgument;
	
	
};

#endif // K3DSHADERINSTANCE_H
