#ifndef K3DTRANSFORMOBJECT_H
#define K3DTRANSFORMOBJECT_H

// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the k3dTransformObject class, which serves as a base class for K-3D objects that have position, orientation, scale, and can render themselves
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dpersistobject.h>
#include <k3dtransform.h>
#include <k3deditorrender.h>
#include <k3drirender.h>
#include <k3ditarget.h>

// Forward class references ...
class k3dILocationServer;

/////////////////////////////////////////////////////////////////////////////////////////
// k3dTransformObject

class k3dTransformObject : public k3dPersistObject,
												public k3dTransform,
												public k3dEditorRender,
												public k3dRiRender,
												public k3dITarget
{
public:
	k3dTransformObject();
	virtual ~k3dTransformObject();

	// k3dIPersist overrides
	sdpResult Load(sdpXmlIElement* Element);
	sdpResult Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies);

	// k3dIEditorRender overrides
	sdpResult EditorPushRender(k3dIEditorEngine* Engine);
	sdpResult EditorPushChildRender(k3dIEditorEngine* Engine);

	// k3dIRiRender overrides
	sdpResult RiPushRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);
	sdpResult RiPushChildRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine);

	// k3dITarget overrides
	sdpResult GetTargetCoordinates(double Coords[3]);

	// k3dIActionClient overrides ...
	sdpResult HandleStep(gulong ValueCode, gulong Value, gulong ValueSize);

	// k3dITransform overrides ...
	sdpResult SetPosition(const k3dITransform::ReferenceFrame Frame, gdouble Offsets[3]);
	sdpResult SetOrientation(const k3dITransform::ReferenceFrame Frame, gdouble AngleAxis[4]);
	sdpResult SetScale(const k3dITransform::ReferenceFrame Frame, gdouble Scale[3]);

	// k3dIUserEventClient overrides ...
	sdpResult OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY);
	sdpResult OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);
	sdpResult OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY);

protected:
	/// Draws a standard set of "handles" with which the user can interactively transform the object
	void DrawTransformationHandles(gdouble Radius);
	/// Draws a standard set of "handles" with which the user can interactively transform the object
	void DrawTransformationHandles(k3dILocationServer* LocationServer, gdouble MinimumRadius);
	
	// Interactive transformations ...
	typedef enum
	{
		NONE,
		ROTATEX,
		ROTATEY,
		ROTATEZ,
		ROTATEFREE,
		TRANSLATEX,
		TRANSLATEY,
		TRANSLATEZ,
		TRANSLATEFREE,
		POSITIVESCALEX,
		NEGATIVESCALEX,
		POSITIVESCALEY,
		NEGATIVESCALEY,
		POSITIVESCALEZ,
		NEGATIVESCALEZ,
		SCALEFREE
	} TransformType;

	TransformType m_TransformType;

	void Rotate(k3dICamera* Camera, GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 LastMouse, sdpVector2 StartMouse);
	void Translate(k3dICamera* Camera, GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 LastMouse, sdpVector2 StartMouse);
	void Scale(k3dICamera* Camera, GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 LastMouse, sdpVector2 StartMouse);

	/// Enumerates undo-able actions
	typedef enum
	{
		STEP_POSITION = 2000000000,
		STEP_ORIENTATION,
		STEP_SCALE,
		STEP_CHILD_POSITION,
		STEP_CHILD_ORIENTATION,
		STEP_CHILD_SCALE
	} StepType;
};

#endif // K3DTRANSFORMOBJECT_H
