// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dEdge class, which encapsulates a polygon edge (defined by two instances of k3dIPoint)
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpcompiler.h>
#include <k3dedge.h>
#include <k3didocument.h>
#include <k3dimesh.h>
#include <k3dipolygonserver.h>
#include <k3dipolygon.h>
#include <k3dipolypointserver.h>
#include <k3dipoint.h>
#include <k3distate.h>
#include <k3dmisc.h>

/// Converts a polygon to a set of edges
void k3dGetEdges(k3dIPolygon* Polygon, k3dEdgeSet& Results)
{
	// Sanity checks ...
	g_assert(Polygon);
	
	k3dIPolyPointServer* polypoints = Polygon->Points();
	g_assert(polypoints);

	// For each point in the polygon ...	
	K3DPOLYPOINTHANDLE lastpolypointhandle = polypoints->GetFirstPolyPoint();
	K3DPOLYPOINTHANDLE polypointhandle = polypoints->GetNextPolyPoint(lastpolypointhandle);
	while(polypointhandle)
		{
			// Store an edge ...
			Results.insert(k3dEdge(polypoints->GetPolyPoint(lastpolypointhandle)->Point(), polypoints->GetPolyPoint(polypointhandle)->Point()));

			lastpolypointhandle = polypointhandle;
			polypointhandle = polypoints->GetNextPolyPoint(polypointhandle);
		}

	// Store the last edge, and close the polygon ...
	Results.insert(k3dEdge(polypoints->GetPolyPoint(lastpolypointhandle)->Point(), polypoints->GetPolyPoint(polypoints->GetFirstPolyPoint())->Point()));
}

/// Converts a mesh to a set of edges
void k3dGetEdges(k3dIMesh* Mesh, gulong Criteria, k3dEdgeSet& Results)
{
	// Sanity checks ...
	g_assert(Mesh);

	// For each polygon in the mesh ...
	k3dIPolygonServer* polygons = Mesh->Polygons();
	for(K3DPOLYGONHANDLE polyhandle = polygons->GetFirstPolygon(); polyhandle; polyhandle = polygons->GetNextPolygon(polyhandle))
		{
			k3dIPolygon* polygon = polygons->GetPolygon(polyhandle);

			// If the polygon isn't visible and selected, skip it ...
			if(SDP_S_OK != polygon->State()->QueryState(Criteria))
				continue;

			// Grab the edges ...
			k3dGetEdges(polygon, Results);
		}
}

class k3dGetEdgesAddMesh
{
public:
	k3dGetEdgesAddMesh(const gulong Criteria, k3dEdgeSet& Results) : m_Criteria(Criteria), m_Results(Results) { }
	
	bool HandleMesh(k3dIMesh* Mesh)
	{
		k3dGetEdges(Mesh, m_Criteria, m_Results);
		return true;
	}
	
protected:
	const gulong m_Criteria;
	k3dEdgeSet& m_Results;
};

/// Creates a set of edges from every polygon in the document that meets the given criteria
void k3dGetEdges(k3dIDocument* Document, gulong Criteria, k3dEdgeSet& Results)
{
	// Sanity checks ...
	g_assert(Document);

	k3dGetEdgesAddMesh functor(Criteria, Results);
	k3dForEachMesh(Document, Criteria, functor);
}



