// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dImageBuffer class, which can store and manipulate an in-memory, 32-bit RGBA bitmap image
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dimagebuffer.h>
#include <sdpgl.h>

#include <sdpmath.h>

///////////////////////////////////////////////////////////////////////////////
// k3dImageBuffer

k3dImageBuffer::k3dImageBuffer()
{
	m_Data = 0;
	m_Width = 0;
	m_Height = 0;
}

k3dImageBuffer::k3dImageBuffer(k3dImageBuffer& Buffer)
{
	m_Data = 0;
	m_Width = 0;
	m_Height = 0;

	Copy(Buffer.Data(), Buffer.Width(), Buffer.Height());
}

k3dImageBuffer::k3dImageBuffer(const k3dPixel* const Data, const gulong Width, const gulong Height)
{
	m_Data = 0;
	m_Width = 0;
	m_Height = 0;

	Copy(Data, Width, Height);
}

k3dImageBuffer::~k3dImageBuffer()
{
	Free();
}

bool k3dImageBuffer::Allocate(const gulong Width, const gulong Height)
{
	// Sanity checks ...
	g_return_val_if_fail(Width, false);
	g_return_val_if_fail(Height, false);

	// Free up existing data ...
	Free();

	// Allocate a new buffer ...
	m_Data = new k3dPixel[Width * Height];
	g_return_val_if_fail(m_Data, false);

	// Store the buffer size ...
	m_Width = Width;
	m_Height = Height;

	return true;
}

bool k3dImageBuffer::Copy(k3dImageBuffer& Buffer)
{
	return Copy(Buffer.Data(), Buffer.Width(), Buffer.Height());
}

bool k3dImageBuffer::Copy(const k3dPixel* const Data, const gulong Width, const gulong Height)
{
	// Get rid of old data ...
	Free();

	if(Data && Width && Height)
		{
			Allocate(Width, Height);
			memcpy(m_Data, Data, Width * Height * sizeof(k3dPixel));
		}

	return true;
}

void k3dImageBuffer::Free()
{
	// Free the data buffer ...
	if(m_Data)
		delete[] m_Data;

	m_Data = 0;
	m_Width = 0;
	m_Height = 0;
}

bool k3dImageBuffer::FlipVertical()
{
	// If the buffer is empty, we're done ...
	g_return_val_if_fail(m_Data && m_Width && m_Height, false);

	// Allocate a one-scanline temp buffer for use swapping scanlines ...
	k3dPixel* buffer = new k3dPixel[m_Width];
	g_return_val_if_fail(buffer, false);

	// Setup pointers to the first and last rows of the image ...
	k3dPixel* front = m_Data;
	k3dPixel* back = m_Data + ((m_Height-1) * m_Width);

	// Figure out how many bytes in one row ...
	gulong rowbytes = sizeof(k3dPixel) * m_Width;

	while(back > front)
		{
			// Swap the front and back rows ...
			memcpy(buffer, back, rowbytes);
			memcpy(back, front, rowbytes);
			memcpy(front, buffer, rowbytes);

			// Move to the next pair of rows ...
			front += m_Width;
			back -= m_Width;
		}

	// Free up our temp buffer ...
	delete[] buffer;

	return true;
}

bool k3dImageBuffer::FlipHorizontal()
{
	// If the buffer is empty, we're done ...
	g_return_val_if_fail(m_Data && m_Width && m_Height, false);

	// Allocate a one-pixel temp buffer for use swapping pixels ...
	k3dPixel buffer;

	// For each scanline in the buffer ...
	for(gulong i = 0; i < m_Height; i++)
		{
			k3dPixel* scanline = m_Data + (i * m_Width);

			// Setup pointers to the first and last rows of the image ...
			k3dPixel* front = scanline;
			k3dPixel* back = scanline + m_Width - 1;

			while(back > front)
				{
					// Swap the front and back pixels ...
					buffer = *back;
					*back = *front;
					*front = buffer;

					// Move to the next pair of rows ...
					front++;
					back--;
				}
		}

	return true;
}

bool k3dImageBuffer::Scale(const gdouble Percent, const ScaleType Type)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Data, false);
	g_return_val_if_fail(m_Width, false);
	g_return_val_if_fail(m_Height, false);
	g_return_val_if_fail(Percent, false);
	
	return Scale(Percent, Percent, Type);
}

bool k3dImageBuffer::Scale(const gdouble WidthPercent, const gdouble HeightPercent, const ScaleType Type)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Data, false);
	g_return_val_if_fail(m_Width, false);
	g_return_val_if_fail(m_Height, false);
	g_return_val_if_fail(WidthPercent, false);
	g_return_val_if_fail(HeightPercent, false);
	
	gulong newwidth = sdpMin<gulong>(gulong(m_Width * WidthPercent), 1);
	gulong newheight = sdpMin<gulong>(gulong(m_Height * HeightPercent), 1);

	return Scale(newwidth, newheight, Type);
}

bool k3dImageBuffer::Scale(const gulong NewWidth, const gulong NewHeight, const ScaleType Type)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Data, false);
	g_return_val_if_fail(m_Width, false);
	g_return_val_if_fail(m_Height, false);
	g_return_val_if_fail(NewWidth, false);
	g_return_val_if_fail(NewHeight, false);
	
	// Allocate the result buffer ...
	k3dPixel* buffer = new k3dPixel[NewWidth * NewHeight];
	g_return_val_if_fail(buffer, false);

	// Very simple scale with no interpolation / antialiasing
	if(Type == NEAREST)
		{
			k3dPixel* source;
			k3dPixel* destination = buffer;

			gulong x, y;
			gulong i, j;

			double sourcewidth = m_Width-1;
			double sourceheight = m_Height-1;
			double destwidth = NewWidth-1;
			double destheight = NewHeight-1;

			// For each row in the destination ...
			for(y = 0; y < NewHeight; y++)
				{
					// Find the closest row in the source ...
					j = gulong((y / destheight) * sourceheight);

					// For each pixel in the destination row ...
					for(x = 0; x < NewWidth; x++)
						{
							// Find the closest pixel in the source row ...
							i = gulong((x / destwidth) * sourcewidth);

							// Copy the source pixel to the destination pixel ...
							source = m_Data + (j*m_Width) + i;
							*destination = *source;

							destination++;
						}
				}
		}
	else
		g_assert_not_reached();	// I must not have implemented this type!

	// Free our old data and attach ourselves to the newly-scaled buffer ...
	Free();

	m_Data = buffer;
	m_Width = NewWidth;
	m_Height = NewHeight;

	return true;
}

bool k3dImageBuffer::ConvertForOpenGL(const ScaleType Type)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Data, false);
	g_return_val_if_fail(m_Width, false);
	g_return_val_if_fail(m_Height, false);

	// Get the maximum texture size for this OpenGL implementation ...
	GLint maxsize;
	glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxsize);

	return ConvertForOpenGL(maxsize, maxsize, Type);
}

bool k3dImageBuffer::ConvertForOpenGL(const gulong MaxWidth, const gulong MaxHeight, const ScaleType Type)
{
	// Sanity checks ...
	g_return_val_if_fail(m_Data, false);
	g_return_val_if_fail(m_Width, false);
	g_return_val_if_fail(m_Height, false);

	// Calculate the best power-of-two dimensions for the image ...
	gulong xpower = gulong(ceil((m_Width <= MaxWidth) ? log(m_Width) / log(2.0) : log(MaxWidth) / log(2.0)));
	gulong ypower = gulong(ceil((m_Height <= MaxHeight) ? log(m_Height) / log(2.0) : log(MaxHeight) / log(2.0)));

	gulong width = 1 << xpower;
	gulong height = 1 << ypower;

	Scale(width, height, Type);

	return true;
}


