// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dMesh class, which implements k3dIMesh, and encapsulates a collection of points, polygons, NURBS curves, material, and shading information
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dmesh.h>
#include <k3dpoint.h>
#include <k3dpolygon.h>
#include <k3dnurbscurve.h>
#include <k3dpolypoint.h>
#include <k3dshadingnode.h>

#include <k3drishadersetup.h>
#include <k3dboundingbox.h>
#include <k3dmisc.h>

#include <k3diactions.h>
#include <k3dieditorengine.h>
#include <k3dieditormaterial.h>
#include <k3dimaterial.h>
#include <k3dirimaterial.h>
#include <k3didocument.h>
#include <k3direnderjob.h>
#include <k3direnderframe.h>
#include <k3dicamera.h>

#include <sdpgl.h>
#include <sdpxmldocument.h>

#include <algorithm>
#include <iostream>
#include <strstream>

//////////////////////////////////////////////////////////////////////////
// k3dMesh

// Static member variables
short k3dMesh::m_RiBilinearPatchLookup[4] = {0, 1, 3, 2};

sdpVector3 k3dMesh::m_SelectedColor = 1.0;
sdpVector3 k3dMesh::m_DefaultColor = 0.0;

k3dMesh::k3dMesh()
{
	AddVariable(&m_Color("color", m_DefaultColor));

	m_CurrentPointID = 0;
	m_CurrentPolygonID = 0;
	m_CurrentNURBSCurveID = 0;
	m_CurrentPolyPointID = 0;
}

class k3dMeshDeleteNURBSCurve
{
public:
	void operator()(k3dNURBSCurve* Curve) { delete Curve; }
};

class k3dMeshDeletePolygon
{
public:
	void operator()(k3dPolygon* Polygon) { delete Polygon; }
};

class k3dMeshDeletePoint
{
public:
	void operator()(k3dPoint* Point) { delete Point; }
};

class k3dMeshDeleteShadingNode
{
public:
	void operator()(k3dShadingNode* Node) { delete Node; }
};

k3dMesh::~k3dMesh()
{
	// Get rid of shading nodes ...
	k3dMeshDeleteShadingNode deletenode;
	std::for_each(m_ShadingNodes.begin(), m_ShadingNodes.end(), deletenode);
	m_ShadingNodes.clear();

	// Get rid of NURBS curves ...
	k3dMeshDeleteNURBSCurve deletecurve;
	std::for_each(m_NURBSCurves.begin(), m_NURBSCurves.end(), deletecurve);
	std::for_each(m_DeletedNURBSCurves.begin(), m_DeletedNURBSCurves.end(), deletecurve);
	m_NURBSCurves.clear();
	m_DeletedNURBSCurves.clear();

	// Get rid of polygons ...
	k3dMeshDeletePolygon deletepoly;
	std::for_each(m_Polygons.begin(), m_Polygons.end(), deletepoly);
	std::for_each(m_DeletedPolygons.begin(), m_DeletedPolygons.end(), deletepoly);
	m_Polygons.clear();
	m_DeletedPolygons.clear();

	// Get rid of points ...
	k3dMeshDeletePoint deletepoint;
	std::for_each(m_Points.begin(), m_Points.end(), deletepoint);
	std::for_each(m_DeletedPoints.begin(), m_DeletedPoints.end(), deletepoint);
	m_Points.clear();
	m_DeletedPoints.clear();

/*
	// Notify objects that geometry is going away ...
	if(m_Layer)
		{
			m_Layer->Document()->OnPointsDeparting();
			m_Layer->Document()->OnPolygonsDeparting();
		}
*/
}

////////////////////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult k3dMesh::ObjectSelected(bool State)
{
	// Make this a deep select of all points/polys/curves ...

	NURBSCurveIterator curveend = m_NURBSCurves.end();
	for(NURBSCurveIterator curve = m_NURBSCurves.begin(); curve != curveend; curve++)
		(*curve)->State()->SetState(k3dIState::SELECTED, State);

	PolygonIterator polyend = m_Polygons.end();
	for(PolygonIterator poly = m_Polygons.begin(); poly != polyend; poly++)
		(*poly)->State()->SetState(k3dIState::SELECTED, State);

	PointIterator pointend = m_Points.end();
	for(PointIterator point = m_Points.begin(); point != pointend; point++)
		(*point)->State()->SetState(k3dIState::SELECTED, State);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIMesh overrides

k3dIDocument* k3dMesh::Document()
{
	return k3dTransformObject::Document();
}

k3dIPolygonServer* k3dMesh::Polygons()
{
	return this;
}

k3dIPolygon* k3dMesh::CreatePolygon()
{
	k3dPolygon* polygon = new k3dPolygon(this, GetNextPolygonID());
	g_return_val_if_fail(polygon, 0);

	m_Polygons.push_back(polygon);

	// Are we recording actions?
	if(SDP_S_OK == Document()->Actions()->IsRecording())
		Document()->Actions()->RecordStep(this, STEP_UNCREATE_POLYGON, reinterpret_cast<gulong>(polygon), 0,
																					STEP_CREATE_POLYGON, reinterpret_cast<gulong>(polygon), 0);

	return polygon;
}

sdpResult k3dMesh::DeletePolygon(k3dIPolygon* Polygon, bool DeletePoints)
{
	// Get the polygon ...
	PolygonIterator poly = std::find(m_Polygons.begin(), m_Polygons.end(), Polygon);
	g_return_val_if_fail(poly != m_Polygons.end(), SDP_E_FAIL);
	k3dPolygon* polygon = *poly;

	// Delete any points that are exclusively "owned" by this polygon ...
	if(DeletePoints)
		{
			// Mark this polygon's points ...
			for(k3dPolygon::PolyPointIterator polypoint = polygon->m_PolyPoints.begin(); polypoint != polygon->m_PolyPoints.end(); polypoint++)
				(*polypoint)->m_Point->m_StateFlags |= k3dIState::USER1;
				
			// Unmark points used in other polygons ...
			for(PolygonIterator otherpoly = m_Polygons.begin(); otherpoly != m_Polygons.end(); otherpoly++)
				{
					// Skip the current polygon ...
					if(otherpoly == poly)
						continue;

					for(k3dPolygon::PolyPointIterator polypoint = (*otherpoly)->m_PolyPoints.begin(); polypoint != (*otherpoly)->m_PolyPoints.end(); polypoint++)
						(*polypoint)->m_Point->m_StateFlags &= ~k3dIState::USER1;
				}
				
			// Cache whatever's left over ...
			std::vector<k3dPoint*> victims;
			for(k3dPolygon::PolyPointIterator victim = polygon->m_PolyPoints.begin(); victim != polygon->m_PolyPoints.end(); victim++)
				if((*victim)->m_Point->m_StateFlags & k3dIState::USER1)
					victims.push_back((*victim)->m_Point);
					
			// Zap 'em ...
			for(gulong i = 0; i < victims.size(); i++)
				DeletePoint(victims[i]);
		}

	// Are we recording actions?
	if(SDP_S_OK == Document()->Actions()->IsRecording())
		// If so, record what we're up to ...
		Document()->Actions()->RecordStep(this, STEP_UNDELETE_POLYGON, reinterpret_cast<gulong>(polygon), 0,
																					STEP_DELETE_POLYGON, reinterpret_cast<gulong>(polygon), 0);

	// Move it to the deleted pile ...
	m_DeletedPolygons.push_back(polygon);
	m_Polygons.erase(poly);

	return SDP_S_OK;
}

k3dINURBSCurveServer* k3dMesh::NURBSCurves()
{
	return this;
}

k3dINURBSCurve* k3dMesh::CreateNURBSCurve()
{
	g_return_val_if_fail(0, 0);
	return 0;
}

sdpResult k3dMesh::DeleteNURBSCurve(k3dINURBSCurve* Curve, bool DeletePoints)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

k3dIPointServer* k3dMesh::Points()
{
	return this;
}

k3dIPoint* k3dMesh::CreatePoint()
{
	k3dPoint* point = new k3dPoint(this, GetNextPointID());
	g_return_val_if_fail(point, 0);

	m_Points.push_back(point);

	// Are we recording actions?
	if(SDP_S_OK == Document()->Actions()->IsRecording())
		Document()->Actions()->RecordStep(this, STEP_UNCREATE_POINT, reinterpret_cast<gulong>(point), 0,
																					STEP_CREATE_POINT, reinterpret_cast<gulong>(point), 0);

	return point;
}

sdpResult k3dMesh::DeletePoint(k3dIPoint* Point)
{
	// Get the point iterator ...
	PointIterator point = std::find(m_Points.begin(), m_Points.end(), Point);
	g_return_val_if_fail(point != m_Points.end(), SDP_E_FAIL);

	// Are we recording actions?
	if(SDP_S_OK == Document()->Actions()->IsRecording())
		// If so, record our undo data ...
		Document()->Actions()->RecordStep(this, STEP_UNDELETE_POINT, reinterpret_cast<gulong>(*point), 0,
																					STEP_DELETE_POINT, reinterpret_cast<gulong>(*point), 0);

	// Move it to the deleted pile ...
	m_DeletedPoints.push_back(*point);
	m_Points.erase(point);

	return SDP_S_OK;
}

class k3dMeshCalculateNormal
{
public:
	class k3dMeshResetPolypoint
	{
	public:
		k3dMeshResetPolypoint(k3dPolygon* Polygon) : m_Polygon(Polygon) {}

		void operator()(k3dPolyPoint* PolyPoint)
		{
			PolyPoint->m_NormalXYZ = m_Polygon->m_Normal;
		}

		k3dPolygon* m_Polygon;
	};

	void operator()(k3dPolygon* Polygon)
	{
		// Calculate the polygon's normal ...
		Polygon->CalculateNormal();
		Polygon->m_Normal.Normalize();

		// Reset polypoint normals to match the polygon's (flat shading) ...
		k3dMeshResetPolypoint functor(Polygon);
		std::for_each(Polygon->m_PolyPoints.begin(), Polygon->m_PolyPoints.end(), functor);
	}
};

class k3dMeshSetState
{
public:
	k3dMeshSetState(gulong Mask, bool State) { m_Mask = Mask; m_State = State; }

	void operator()(k3dPoint* Point)
	{
		Point->SetState(m_Mask, m_State);
	}

	gulong m_Mask;
	bool m_State;
};

sdpResult k3dMesh::ShadePolygons(gulong State, gdouble Angle)
{
	// Start by getting rid of any existing shading ...
	UnshadePolygons(State);

	// Setup polygons by calculating normals ...
	k3dMeshCalculateNormal normalfunctor;
	std::for_each(m_Polygons.begin(), m_Polygons.end(), normalfunctor);

	// Setup points by clearing the USER1 state ...
	k3dMeshSetState statefunctor(k3dIState::USER1, false);
	std::for_each(m_Points.begin(), m_Points.end(), statefunctor);

	// Temporary storage ...
	std::vector<k3dPolyPoint*> polypointarray;
	k3dShadingNode* node;
	gulong i, polypointcount;

/*
	CLayer* layer;
	k3dMesh* mesh;
	k3dPolygon *poly, *poly2;
	k3dPolyPoint *polypoint, *polypoint2;
	k3dPoint3D* point;
	k3dShadingNode *node;

	gulong i, count;

	sdpVector3 normal1, normal2;


	if(Angle)
		{
			double angle = fabs(cos(KRadians(MaxAngle)));

			for(layer = m_Layers.front(); layer; layer = layer->m_Next)
				{
					for(mesh = layer->m_Meshes.front(); mesh; mesh = mesh->m_Next)
						{
							for(poly = mesh->m_Polygons.front(); poly; poly = poly->m_Next)
								{
									if(SDP_S_OK != poly->QueryState(State)) continue;

									normal1 = poly->m_Normal;
									normal1.Normalize();

									for(polypoint = poly->m_PolyPoints.front(); polypoint; polypoint = polypoint->m_Next)
										{
											if(SDP_S_OK == polypoint->m_Point->QueryState(k3dIState::USER1)) continue;

											count = 1;
											polypointarray.clear();

											for(poly2 = poly->m_Next; poly2; poly2 = poly2->m_Next)
												{
													if(SDP_S_OK != poly2->QueryState(State)) continue;

													normal2 = poly2->m_Normal;
													normal2.Normalize();

													if(fabs(normal1 * normal2) >= angle)	// Seems bass-ackward, but cos() is that way ...
														{
    													for(polypoint2 = poly2->m_PolyPoints.front(); polypoint2; polypoint2 = polypoint2->m_Next)
    														{
    															if(polypoint2->m_Point == polypoint->m_Point)
																		{
																			polypointarray.push_back(polypoint2);
    																	count++;
    																}
    														}
														}
												}

											if(count > 1)
												{
													node = mesh->CreateShadingNode();
													node->m_PointCount = count;
													node->m_PointArray = (k3dPolyPoint**)malloc(count * sizeof(k3dPolyPoint*));

													node->m_PointArray[0] = polypoint;

													for(i = 1; i < count; i++)
														{
															node->m_PointArray[i] = polypointarray[i - 1];
														}
												}
										}
								}
						}
				}
		}
	else
*/
		{
			// For each polygon ...
			PolygonIterator polyend = m_Polygons.end();
			for(PolygonIterator poly = m_Polygons.begin(); poly != polyend; poly++)
				{
					// Skip polygons that don't match our criteria ...
					if(SDP_S_OK != (*poly)->QueryState(State))
						continue;

					// For each polypoint ...
					k3dPolygon::PolyPointIterator polypointend = (*poly)->m_PolyPoints.end();
					for(k3dPolygon::PolyPointIterator polypoint = (*poly)->m_PolyPoints.begin(); polypoint != polypointend; polypoint++)
						{
							// If this point is already smooth-shaded, skip it ...
							if(SDP_S_OK == (*polypoint)->m_Point->QueryState(k3dIState::USER1))
								continue;

							polypointarray.clear();
							polypointarray.push_back(*polypoint);

							// For each polygon following the one we're looking at ...
							PolygonIterator poly2 = poly;
							for(poly2++; poly2 != polyend; poly2++)
								{
									// Skip polygons that don't match our criteria
									if(SDP_S_OK != (*poly2)->QueryState(State))
										continue;

									// For each polypoint ...
									k3dPolygon::PolyPointIterator polypointend2 = (*poly2)->m_PolyPoints.end();
    							for(k3dPolygon::PolyPointIterator polypoint2 = (*poly2)->m_PolyPoints.begin(); polypoint2 != polypointend2; polypoint2++)
    								{
											// If it matches the poly point we're looking at ...
    									if((*polypoint2)->m_Point == (*polypoint)->m_Point)
												{
													// Mark it ...
													(*polypoint2)->m_Point->SetState(k3dIState::USER1, true);
													// Store it ...
													polypointarray.push_back(*polypoint2);
    										}
    								}
								}

							// If we found some polypoints to shade together ...
							polypointcount = polypointarray.size();
							if(polypointcount > 1)
								{
									// Mark this polypoint ...
									(*polypoint)->m_Point->SetState(k3dIState::USER1, true);

									// Create a new shading node ...
									node = new k3dShadingNode();

									for(i = 0; i < polypointcount; i++)
										node->m_PolyPoints.push_back(polypointarray[i]);

									// Stash it ...
									m_ShadingNodes.push_back(node);
								}
						}
				}
		}

	return SDP_S_OK;
}

sdpResult k3dMesh::UnshadePolygons(gulong State)
{
	// For each shading node ...
	ShadingNodeIterator nextnode;
	ShadingNodeIterator end = m_ShadingNodes.end();
	for(ShadingNodeIterator node = m_ShadingNodes.begin(); node != end; node = nextnode)
		{
			nextnode = node;
			nextnode++;

			// Remove any polygons that match our criteria ...
			(*node)->RemovePolygons(State);

			// If the node is empty, get rid of it ...
			if(0 == (*node)->PolyPointCount())
				{
					delete (*node);
					m_ShadingNodes.erase(node);
				}
		}

	return SDP_S_OK;
}

sdpResult k3dMesh::CalculateNormals()
{
	// Calculate polygon normals ...
	k3dMeshCalculateNormal functor;
	std::for_each(m_Polygons.begin(), m_Polygons.end(), functor);

	// Temp storage ...
	sdpVector3 normal;
	k3dShadingNode::PolyPointIterator polypoint;
	k3dShadingNode::PolyPointIterator polypointend;

	// Integrate polypoint vectors for smooth shading ...
	ShadingNodeIterator nodeend = m_ShadingNodes.end();
	for(ShadingNodeIterator node = m_ShadingNodes.begin(); node != nodeend; node++)
		{
			normal = 0.0;

			polypointend = (*node)->m_PolyPoints.end();
			for(polypoint = (*node)->m_PolyPoints.begin(); polypoint != polypointend; polypoint++)
				normal += (*polypoint)->Polygon()->m_Normal;

			for(polypoint = (*node)->m_PolyPoints.begin(); polypoint != polypointend; polypoint++)
				(*polypoint)->m_NormalXYZ = normal;
		}

	return SDP_S_OK;
}

// Return the dimensions of every point within the mesh that fits the given criteria ...
sdpResult k3dMesh::GetDimensions(gulong Criteria, double Min[3], double Max[3])
{
	Max[0] = Max[1] = Max[2] = -DBL_MAX;
	Min[0] = Min[1] = Min[2] = DBL_MAX;

	gulong count = 0;
	gulong end = m_Points.size();
	for(gulong i = 0; i < end; i++)
		{
			if(SDP_S_OK != m_Points[i]->State()->QueryState(Criteria))
				continue;

			count++;

			if(m_Points[i]->m_OldXYZ[0] > Max[0])
				Max[0] = m_Points[i]->m_OldXYZ[0];
			if(m_Points[i]->m_OldXYZ[0] < Min[0])
				Min[0] = m_Points[i]->m_OldXYZ[0];
			if(m_Points[i]->m_OldXYZ[1] > Max[1])
				Max[1] = m_Points[i]->m_OldXYZ[1];
			if(m_Points[i]->m_OldXYZ[1] < Min[1])
				Min[1] = m_Points[i]->m_OldXYZ[1];
			if(m_Points[i]->m_OldXYZ[2] > Max[2])
				Max[2] = m_Points[i]->m_OldXYZ[2];
			if(m_Points[i]->m_OldXYZ[2] < Min[2])
				Min[2] = m_Points[i]->m_OldXYZ[2];
		}

	if(!count)
		{
			Min[0] = Min[1] = Min[2] = 0.0;
			Max[0] = Max[1] = Max[2] = 0.0;
		}

	return SDP_S_OK;
}

sdpResult k3dMesh::GetColor(double Color[3])
{
	m_Color().CopyArray(Color);

	return SDP_S_OK;
}

sdpResult k3dMesh::SetColor(double Color[3])
{
	m_Color() = Color;

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////////////
// k3dIActionClient overrides ...

sdpResult k3dMesh::HandleStep(gulong ValueCode, gulong Value, gulong ValueSize)
{
	switch(ValueCode)
		{
			case STEP_UNDELETE_NURBSCURVE:
			case STEP_CREATE_NURBSCURVE:
				{
					NURBSCurveIterator curve = std::find(m_DeletedNURBSCurves.begin(), m_DeletedNURBSCurves.end(), reinterpret_cast<k3dNURBSCurve*>(Value));
					g_assert(curve != m_DeletedNURBSCurves.end());

					m_NURBSCurves.push_back(*curve);
					m_DeletedNURBSCurves.erase(curve);

					return SDP_S_OK;
				}
			case STEP_DELETE_NURBSCURVE:
			case STEP_UNCREATE_NURBSCURVE:
				{
					NURBSCurveIterator curve = std::find(m_NURBSCurves.begin(), m_NURBSCurves.end(), reinterpret_cast<k3dNURBSCurve*>(Value));
					g_assert(curve != m_NURBSCurves.end());

					m_DeletedNURBSCurves.push_back(*curve);
					m_NURBSCurves.erase(curve);

					return SDP_S_OK;
				}
			case STEP_UNDELETE_POLYGON:
			case STEP_CREATE_POLYGON:
				{
					PolygonIterator iterator = std::find(m_DeletedPolygons.begin(), m_DeletedPolygons.end(), reinterpret_cast<k3dPolygon*>(Value));
					g_assert(iterator != m_DeletedPolygons.end());

					m_Polygons.push_back(*iterator);
					m_DeletedPolygons.erase(iterator);

					return SDP_S_OK;
				}
			case STEP_DELETE_POLYGON:
			case STEP_UNCREATE_POLYGON:
				{
					PolygonIterator poly = std::find(m_Polygons.begin(), m_Polygons.end(), reinterpret_cast<k3dPolygon*>(Value));
					g_assert(poly != m_Polygons.end());

					m_DeletedPolygons.push_back(*poly);
					m_Polygons.erase(poly);

					return SDP_S_OK;
				}
			case STEP_UNDELETE_POINT:
			case STEP_CREATE_POINT:
				{
					PointIterator iterator = std::find(m_DeletedPoints.begin(), m_DeletedPoints.end(), reinterpret_cast<k3dPoint*>(Value));
					g_assert(iterator != m_DeletedPoints.end());

					m_Points.push_back(*iterator);
					m_DeletedPoints.erase(iterator);

					return SDP_S_OK;
				}
			case STEP_DELETE_POINT:
			case STEP_UNCREATE_POINT:
				{
					PointIterator iterator = std::find(m_Points.begin(), m_Points.end(), reinterpret_cast<k3dPoint*>(Value));
					g_assert(iterator != m_Points.end());

					m_DeletedPoints.push_back(*iterator);
					m_Points.erase(iterator);

					return SDP_S_OK;
				}
		}

	return k3dTransformObject::HandleStep(ValueCode, Value, ValueSize);
}

///////////////////////////////////////////////////////////////
// k3dIPolygonServer overrides

gulong k3dMesh::GetPolygonCount()
{
	return m_Polygons.size();
}

K3DPOLYGONHANDLE k3dMesh::GetFirstPolygon()
{
	return K3DPOLYGONHANDLE(m_Polygons.size() ? 1 : 0);
}

K3DPOLYGONHANDLE k3dMesh::GetNextPolygon(K3DPOLYGONHANDLE Polygon)
{
	// Sanity checks ...
	g_return_val_if_fail(Polygon, 0);
	g_return_val_if_fail(gulong(Polygon) <= m_Polygons.size(), 0);

	return K3DPOLYGONHANDLE(gulong(Polygon) < m_Polygons.size() ? gulong(Polygon)+1 : 0);
}

k3dIPolygon* k3dMesh::GetPolygon(K3DPOLYGONHANDLE Polygon)
{
	// Sanity checks ...
	g_return_val_if_fail(Polygon, 0);
	g_return_val_if_fail(gulong(Polygon) <= m_Polygons.size(), 0);

	return m_Polygons[gulong(Polygon)-1];
}

sdpResult k3dMesh::AddPolygon(k3dIPolygon* Polygon)
{
g_assert_not_reached();
return 0;
/*
	g_return_val_if_fail(0, SDP_E_FAIL);	// Not implemented for meshes - use k3dIMesh::CreatePolygon() instead ...
*/
}

sdpResult k3dMesh::RemovePolygon(k3dIPolygon* Polygon)
{
g_assert_not_reached();
return 0;
/*
	g_return_val_if_fail(0, SDP_E_FAIL);	// Not implemented for meshes - use k3dIMesh::DestroyPolygon() instead ...
*/
}

///////////////////////////////////////////////////////////////
// k3dINURBSCurveServer overrides

gulong k3dMesh::GetNURBSCurveCount()
{
	return m_NURBSCurves.size();
}

K3DNURBSCURVEHANDLE k3dMesh::GetFirstNURBSCurve()
{
	return K3DNURBSCURVEHANDLE(m_NURBSCurves.size() ? 1 : 0);
}

K3DNURBSCURVEHANDLE k3dMesh::GetNextNURBSCurve(K3DNURBSCURVEHANDLE Curve)
{
	// Sanity checks ...
	g_return_val_if_fail(Curve, 0);
	g_return_val_if_fail(gulong(Curve) <= m_NURBSCurves.size(), 0);

	return K3DNURBSCURVEHANDLE(gulong(Curve) < m_NURBSCurves.size() ? gulong(Curve)+1 : 0);
}

k3dINURBSCurve* k3dMesh::GetNURBSCurve(K3DNURBSCURVEHANDLE Curve)
{
	// Sanity checks ...
	g_return_val_if_fail(Curve, 0);
	g_return_val_if_fail(gulong(Curve) <= m_NURBSCurves.size(), 0);

	return m_NURBSCurves[gulong(Curve)-1];
}

sdpResult k3dMesh::AddNURBSCurve(k3dINURBSCurve* Curve)
{
g_assert_not_reached();
return 0;
/*
	g_return_val_if_fail(0, SDP_E_FAIL);	// Not implemented for meshes - use k3dIMesh::CreatePolygon() instead ...
*/
}

sdpResult k3dMesh::RemoveNURBSCurve(k3dINURBSCurve* Curve)
{
g_assert_not_reached();
return 0;
/*
	g_return_val_if_fail(0, SDP_E_FAIL);	// Not implemented for meshes - use k3dIMesh::DestroyPolygon() instead ...
*/
}

/////////////////////////////////////////////////////////////////
// k3dIPointServer overrides

gulong k3dMesh::GetPointCount()
{
	return m_Points.size();
}

K3DPOINTHANDLE k3dMesh::GetFirstPoint()
{
	return K3DPOINTHANDLE(m_Points.size() ? 1 : 0);
}

K3DPOINTHANDLE k3dMesh::GetNextPoint(K3DPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_Points.size(), 0);

	return K3DPOINTHANDLE(gulong(Point) < m_Points.size() ? gulong(Point)+1 : 0);
}

k3dIPoint* k3dMesh::GetPoint(K3DPOINTHANDLE Point)
{
	// Sanity checks ...
	g_return_val_if_fail(Point, 0);
	g_return_val_if_fail(gulong(Point) <= m_Points.size(), 0);

	return m_Points[gulong(Point)-1];
}

sdpResult k3dMesh::AddPoint(k3dIPoint* Point)
{
g_assert_not_reached();
return 0;
/*
	g_return_val_if_fail(0, SDP_E_FAIL);	// Not implemented for meshes ... use k3dIMesh::CreatePoint() instead ...
*/
}

sdpResult k3dMesh::RemovePoint(k3dIPoint* Point)
{
g_assert_not_reached();
return 0;
/*
	g_return_val_if_fail(0, SDP_E_FAIL);	// Not implemented for meshes ... use k3dIMesh::DeletePoint() instead ...
*/
}


////////////////////////////////////////////////////////
// k3dILocationServer overrides ...

gulong k3dMesh::GetLocationCount()
{
	return GetPointCount();
}

K3DLOCATIONHANDLE k3dMesh::GetFirstLocation()
{
	return K3DLOCATIONHANDLE(m_Points.size() ? 1 : 0);
}

K3DLOCATIONHANDLE k3dMesh::GetNextLocation(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_Points.size(), 0);

	return K3DLOCATIONHANDLE(gulong(Location) < m_Points.size() ? gulong(Location)+1 : 0);
}

k3dILocation* k3dMesh::GetLocation(K3DLOCATIONHANDLE Location)
{
	// Sanity checks ...
	g_return_val_if_fail(Location, 0);
	g_return_val_if_fail(gulong(Location) <= m_Points.size(), 0);

	return m_Points[gulong(Location)-1];
}

////////////////////////////////////////////////////////
// k3dIEditorRender overrides ...

sdpResult k3dMesh::EditorRender(k3dIEditorEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	if(SDP_S_OK == Engine->EditorQuerySolid())
		EditorRenderSolid(Engine);
	else
		EditorRenderWireframe(Engine);

	// Draw points if in point mode ...
	if(k3dIDocument::POINT == Document()->GetActiveEditMode())
		DrawPoints();

	// Draw polygon normals ...
	if(SDP_S_OK == Engine->EditorQueryNormals())
		DrawSelectedNormals();

	DrawTransformationHandles(this, 1.0);

	return SDP_S_OK;
}

void k3dMesh::EditorRenderWireframe(k3dIEditorEngine* Engine)
{
	// Draw polygons ...
	DrawWireframePolygons(Engine, k3dIState::VISIBLE);
}

void k3dMesh::EditorRenderSolid(k3dIEditorEngine* Engine)
{
	// Draw solid shaded polygons ...
	glEnable(GL_LIGHTING);
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(1.0, 1.0);
	if(SDP_S_OK == Engine->EditorQueryTextured())
		EditorSolidTexturedPolygons(Engine);
	else
		EditorSolidPolygons(Engine);
	glDisable(GL_POLYGON_OFFSET_FILL);

	// Setup for wireframe polys ...
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);

	// Draw non-renderable polygons ...
//	DrawWireframeUnrenderablePolygons();

	// Draw selected polygons ...
	DrawWireframePolygons(Engine, k3dIState::VISIBLE);
}

///////////////////////////////////////////////////////////////////////////
// Rendering helpers ...

void k3dMesh::DrawWireframePolygons(k3dIEditorEngine* Engine, gulong Criteria)
{
	// Should we draw polygon orientation?
	if(SDP_S_OK == Engine->EditorQueryOrientation())
		{
			// Draw polygons ...
			for(gulong i = 0; i < m_Polygons.size(); i++)
				{
					if(SDP_S_OK != m_Polygons[i]->QueryState(Criteria))
						continue;

					// Is the poly selected?
					if(SDP_S_OK == m_Polygons[i]->QueryState(k3dIState::SELECTED))
						{
							m_Polygons[i]->OpenGLWireframeOrientation(m_Color(), m_SelectedColor);
						}
					else
						{
							glColor3dv(m_Color());
							m_Polygons[i]->OpenGLWireframe();
						}
				}
		}
	// Nope, just draw the dang polys!
	else
		{
			// Draw polygons ...
			for(gulong i = 0; i < m_Polygons.size(); i++)
				{
					if(SDP_S_OK != m_Polygons[i]->QueryState(Criteria))
						continue;

					// Set the polygon color ...
					if(SDP_S_OK == m_Polygons[i]->QueryState(k3dIState::SELECTED))
						glColor3dv(m_SelectedColor);
					else
						glColor3dv(m_Color());

					// Draw the polygon ...
					m_Polygons[i]->OpenGLWireframe();
				}
		}

}

void k3dMesh::DrawPoints()
{
	// Draw un-selected points ...
	glColor3dv(m_Color());

	glBegin(GL_POINTS);

	gulong i;
	for(i = 0; i < m_Points.size(); i++)
		{
			// Skip hidden and selected points ...
			if(SDP_S_OK != m_Points[i]->QueryState(k3dIState::VISIBLE))
				continue;
			if(SDP_S_OK == m_Points[i]->QueryState(k3dIState::SELECTED))
				continue;

			glVertex3dv(m_Points[i]->m_AnimXYZ.n);
		}
	glEnd();

	// Draw selected points ...
	glColor3dv(m_SelectedColor);

	glBegin(GL_POINTS);
	for(i = 0; i < m_Points.size(); i++)
		{
			// Skip hidden and unselected points ...
			if(SDP_S_OK != m_Points[i]->QueryState(k3dIState::VISIBLE))
				continue;
			if(SDP_S_OK != m_Points[i]->QueryState(k3dIState::SELECTED))
				continue;

			glVertex3dv(m_Points[i]->m_AnimXYZ.n);
		}
	glEnd();
}

void k3dMesh::DrawSelectedNormals()
{
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);

	double normallength, finallength;

	sdpVector3 center, coords;
	k3dBoundingBox3 boundingbox;

	// Setup line parameters ...
	glLineWidth(1.0f);
	// Setup color ...
	glColor3d(0.0, 0.8, 0.0);

	// For each polygon ...
	for(gulong i = 0; i < m_Polygons.size(); i++)
		{
			// Ignore Bezier-spline, hidden, or un-selected polygons ...
			if(SDP_S_OK != m_Polygons[i]->QueryState(k3dIState::VISIBLESELECTED))
				continue;
			// Ignore polygons with two-or-fewer points ...
			if(m_Polygons[i]->GetPointCount() <= 2)
				continue;

			// Calculate bounding box for this poly
			boundingbox.Init();
			for(gulong j = 0; j < m_Polygons[i]->m_PolyPoints.size(); j++)
				boundingbox.Expand(m_Polygons[i]->m_PolyPoints[j]->m_Point->m_AnimXYZ);

			finallength = 0.5 * boundingbox.Size().Length();

			normallength = m_Polygons[i]->m_Normal.Length();
			if(normallength > 0.0)
				{
					normallength = finallength / normallength;

					center = boundingbox.Center();
					coords = m_Polygons[i]->m_Normal * normallength;
					coords += center;

					glBegin(GL_LINES);
					glVertex3dv(center);
					glVertex3dv(coords);
					glEnd();
				}
		}

	// Restore original parameters
	glPopAttrib();
}

void k3dMesh::EditorSolidPolygons(k3dIEditorEngine* Engine)
{
	k3dIMaterial* material;
	k3dIMaterial* lastmaterial = (k3dIMaterial*)-1;
	k3dIMaterial* defaultmaterial = Document()->GetDefaultMaterial();
	k3dIEditorMaterial* editormaterial = 0;

	// Draw polygons ...
	for(gulong i = 0; i < m_Polygons.size(); i++)
		{
			if(SDP_S_OK != m_Polygons[i]->QueryState(k3dIState::VISIBLE))
				continue;

			// See if we need to set up a new material ...
			material = m_Polygons[i]->m_Material;
			if(!material)
				material = defaultmaterial;

			if(material)
				{
					if(material != lastmaterial)
						{
							if(editormaterial)
								editormaterial->Release();

							editormaterial = sdpQueryInterface<k3dIEditorMaterial>(material);
							if(0 == editormaterial)
								continue;

							lastmaterial = material;
							editormaterial->EditorMaterialSetup(Engine);
						}
				}

			// Draw the polygon ...
			m_Polygons[i]->OpenGLSolid();
		}
}

void k3dMesh::EditorSolidTexturedPolygons(k3dIEditorEngine* Engine)
{
	k3dIMaterial* material;
	k3dIMaterial* lastmaterial = (k3dIMaterial*)-1;
	k3dIMaterial* defaultmaterial = Document()->GetDefaultMaterial();
	k3dIEditorMaterial* editormaterial = 0;

	// Draw polygons ...
	for(gulong i = 0; i < m_Polygons.size(); i++)
		{
			if(SDP_S_OK != m_Polygons[i]->QueryState(k3dIState::VISIBLE))
				continue;

			// See if we need to set up a new material ...
			material = m_Polygons[i]->m_Material;
			if(!material)
				material = defaultmaterial;

			if(material)
				{
					if(material != lastmaterial)
						{
							if(editormaterial)
								editormaterial->Release();

							editormaterial = sdpQueryInterface<k3dIEditorMaterial>(material);
							if(0 == editormaterial)
								continue;

							lastmaterial = material;
							editormaterial->EditorMaterialSetup(Engine);
						}
				}

			// Draw the polygon ...
			m_Polygons[i]->OpenGLSolidTextured();
		}
}

/*
void k3dMesh::EditorRenderActivePolygon(k3dPolygon* Polygon)
{
g_assert_not_reached();
	g_assert(Polygon->m_PolyPoints.front());
	g_assert(Polygon->m_PolyPoints.front()->m_Next);

	if(!Polygon->m_PolyPoints.front()) return;
	if(!Polygon->m_PolyPoints.front()->m_Next) return;

	// Draw the first segment ...
	glColor3d(1.0, 0.0, 1.0);
	glBegin(GL_LINES);
	glVertex3dv(Polygon->m_PolyPoints.front()->m_Point->m_AnimXYZ.n);
	glVertex3dv(Polygon->m_PolyPoints.front()->m_Next->m_Point->m_AnimXYZ.n);
	glEnd();

	// Draw the first point ...
	glBegin(GL_POINTS);
	glVertex3dv(Polygon->m_PolyPoints.front()->m_Point->m_AnimXYZ.n);
	glEnd();

	// Draw points ...
	glColor3d(1.0, 1.0, 1.0);
	glBegin(GL_POINTS);
	for(k3dPolyPoint* polypoint = Polygon->m_PolyPoints.front()->m_Next; polypoint; polypoint = polypoint->m_Next)
		{
			glVertex3dv(polypoint->m_Point->m_AnimXYZ.n);
		}
	glEnd();
}
*/

///////////////////////////////////////////////////////////////
// k3dIRiRender overrides

sdpResult k3dMesh::RiPushRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	k3dTransformObject::RiPushRender(Frame, Engine);

	if(m_RiSolid)
		Engine->RiSolidBegin(k3dIRiEngine::RI_PRIMITIVE);

	return SDP_S_OK;
}

sdpResult k3dMesh::RiRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Setup our material interfaces ...
	k3dIMaterial* material = 0;
	k3dIMaterial* lastmaterial = (k3dIMaterial*)-1;
	k3dIMaterial* defaultmaterial = Document()->GetDefaultMaterial();
	k3dIRiMaterial* rimaterial = 0;

	// For each polygon ...
	k3dPolygon* poly;
	for(gulong i = 0; i < m_Polygons.size(); i++)
		{
			poly = m_Polygons[i];

			// Skip hidden or unrenderable polys
			if(SDP_S_OK != poly->QueryState(k3dIState::VISIBLE))
				continue;

			// Get the polygon material ...
			material = poly->m_Material;

			// If a material hasn't been assigned, assign the default ...
			if(0 == material)
				{
					material = defaultmaterial;
					poly->m_Material = defaultmaterial;
				}

			// If the material has changed ...
			if(material != lastmaterial)
				{
					// Release the old Renderman material interface as needed ...
					if(rimaterial)
						{
							rimaterial->Release();
							rimaterial = 0;
						}

					// If we can't find a new Renderman material interface, skip it ...
					rimaterial = sdpQueryInterface<k3dIRiMaterial>(material);
					if(0 == rimaterial)
						continue;

					lastmaterial = material;

					// Setup the new material ...
					rimaterial->RiMaterialSetup(Frame, Engine);
				}

			// Render the polygon ...
			RiPolygon(Engine, rimaterial, poly);
		}

	return SDP_S_OK;
}

// Render a polygon ...
void k3dMesh::RiPolygon(k3dIRiEngine* Engine, k3dIRiMaterial* Material, k3dPolygon* Polygon)
{
	switch(Polygon->m_PolyPoints.size())
		{
			case 1:
			case 2:
				return;

			case 3:
				RiPolygon3(Engine, Material, Polygon);
				return;

			case 4:
				RiPolygon4(Engine, Material, Polygon);
				return;

			default:
				RiPolygonN(Engine, Material, Polygon);
				return;
		}
}

// Render a three-sided polygon as a bilinear patch ...
void k3dMesh::RiPolygon3(k3dIRiEngine* Engine, k3dIRiMaterial* Material, k3dPolygon* Polygon)
{
	// Arrays of coordinates ...
	sdpVector3* points = new sdpVector3[4];
	sdpVector3* normals = new sdpVector3[4];
	gdouble* scoords = new gdouble[4];
	gdouble* tcoords = new gdouble[4];

	k3dPolyPoint* middlepoint;
	for(gulong i = 0; i < 3; i++)
		{
			k3dPolyPoint* polypoint = Polygon->m_PolyPoints[i];

			if(1 == i)
				middlepoint = polypoint;

			points[i] = polypoint->m_Point->m_AnimXYZ;
			normals[i] = -polypoint->m_NormalXYZ;
			scoords[i] = polypoint->m_TextureXYZ[0];
			tcoords[i] = polypoint->m_TextureXYZ[1];
		}

	points[3] = middlepoint->m_Point->m_AnimXYZ;
	normals[3] = -middlepoint->m_NormalXYZ;
	scoords[3] = middlepoint->m_TextureXYZ[0];
	tcoords[3] = middlepoint->m_TextureXYZ[1];

	k3dRiParameterList params;
	params.AddPointer(k3dIRiEngine::RI_P, gpointer(points));
	params.AddPointer(k3dIRiEngine::RI_N, gpointer(normals));
	params.AddPointer(k3dIRiEngine::RI_S, gpointer(scoords));
	params.AddPointer(k3dIRiEngine::RI_T, gpointer(tcoords));

	Engine->RiPatchV(k3dIRiEngine::RI_BILINEAR, params.ParameterCount(), params.Tokens(), params.Parameters());
	
	delete[] points;
	delete[] normals;
	delete[] scoords;
	delete[] tcoords;
}

// Render a four-sided polygon as a bilinear patch ...
void k3dMesh::RiPolygon4(k3dIRiEngine* Engine, k3dIRiMaterial* Material, k3dPolygon* Polygon)
{
	// Arrays of coordinates ...
	sdpVector3* points = new sdpVector3[4];
	sdpVector3* normals = new sdpVector3[4];
	gdouble* scoords = new gdouble[4];
	gdouble* tcoords = new gdouble[4];

	for(gulong i = 0; i < 4; i++)
		{
			k3dPolyPoint* polypoint = Polygon->m_PolyPoints[i];

			points[m_RiBilinearPatchLookup[i]] = polypoint->m_Point->m_AnimXYZ;
			normals[m_RiBilinearPatchLookup[i]] = -polypoint->m_NormalXYZ;
			scoords[m_RiBilinearPatchLookup[i]] = polypoint->m_TextureXYZ[0];
			tcoords[m_RiBilinearPatchLookup[i]] = polypoint->m_TextureXYZ[1];
		}

	k3dRiParameterList params;
	params.AddPointer(k3dIRiEngine::RI_P, gpointer(points));
	params.AddPointer(k3dIRiEngine::RI_N, gpointer(normals));
	params.AddPointer(k3dIRiEngine::RI_S, gpointer(scoords));
	params.AddPointer(k3dIRiEngine::RI_T, gpointer(tcoords));

	Engine->RiPatchV(k3dIRiEngine::RI_BILINEAR, params.ParameterCount(), params.Tokens(), params.Parameters());

	delete[] points;
	delete[] normals;
	delete[] scoords;
	delete[] tcoords;
}

// Render a greater-than-four-sided polygon ...
void k3dMesh::RiPolygonN(k3dIRiEngine* Engine, k3dIRiMaterial* Material, k3dPolygon* Polygon)
{
	// the number of points in the poly ...
	gulong pointcount = Polygon->m_PolyPoints.size();

	// Arrays of coordinates ...
	sdpVector3* points = new sdpVector3[pointcount];
	sdpVector3* normals = new sdpVector3[pointcount];
	gdouble* scoords = new gdouble[pointcount];
	gdouble* tcoords = new gdouble[pointcount];

	for(gulong i = 0; i < Polygon->m_PolyPoints.size(); i++)
		{
			k3dPolyPoint* polypoint = Polygon->m_PolyPoints[i];

			points[i] = polypoint->m_Point->m_AnimXYZ;
			normals[i] = -polypoint->m_NormalXYZ;
			scoords[i] = polypoint->m_TextureXYZ[0];
			tcoords[i] = polypoint->m_TextureXYZ[1];

		}

	k3dRiParameterList params;
	params.AddPointer(k3dIRiEngine::RI_P, gpointer(points));
	params.AddPointer(k3dIRiEngine::RI_S, gpointer(scoords));
	params.AddPointer(k3dIRiEngine::RI_T, gpointer(tcoords));
	params.AddPointer(k3dIRiEngine::RI_N, gpointer(normals));

	Engine->RiGeneralPolygonV(1, &pointcount, params.ParameterCount(), params.Tokens(), params.Parameters());
	
	delete[] points;
	delete[] normals;
	delete[] scoords;
	delete[] tcoords;
}

sdpResult k3dMesh::RiPopRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	if(m_RiSolid)
		Engine->RiSolidEnd();

	return k3dTransformObject::RiPopRender(Frame, Engine);
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult k3dMesh::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	// Call the base-class implementation ...
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::Save(Element, Dependencies)), SDP_E_FAIL);

	// Save our geometry ...
	g_return_val_if_fail(sdpSucceeded(SaveGeometry(Element, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dMesh::SaveGeometry(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	// Save points ...
	g_return_val_if_fail(sdpSucceeded(SavePoints(Element, Dependencies)), SDP_E_FAIL);
	// Save polygons ...
	g_return_val_if_fail(sdpSucceeded(SavePolygons(Element, Dependencies)), SDP_E_FAIL);
	// Save shading data ...
	g_return_val_if_fail(sdpSucceeded(SaveShading(Element, Dependencies)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dMesh::SavePoints(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* points = sdpXmlCreateElement(Element, "points");

	for(PointIterator point = m_Points.begin(); point != m_Points.end(); point++)
		(*point)->Save(points, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dMesh::SavePolygons(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* polygons = sdpXmlCreateElement(Element, "polygons");

	for(PolygonIterator polygon = m_Polygons.begin(); polygon != m_Polygons.end(); polygon++)
		(*polygon)->Save(polygons, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dMesh::SaveNURBSCurves(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* nurbscurves = sdpXmlCreateElement(Element, "nurbscurves");

	for(NURBSCurveIterator curve = m_NURBSCurves.begin(); curve != m_NURBSCurves.end(); curve++)
		(*curve)->Save(nurbscurves, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dMesh::SaveShading(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* shading = sdpXmlCreateElement(Element, "shading");

	for(ShadingNodeIterator node = m_ShadingNodes.begin(); node != m_ShadingNodes.end(); node++)
		(*node)->Save(shading, Dependencies);

	return SDP_S_OK;
}

sdpResult k3dMesh::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	// Call the base-class implementation ...
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::Load(Element)), SDP_E_FAIL);

	// Load geometry ...
	g_return_val_if_fail(sdpSucceeded(LoadGeometry(Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dMesh::LoadGeometry(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	g_return_val_if_fail(sdpSucceeded(LoadPoints(Element)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(LoadPolygons(Element)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(LoadShading(Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

/// Functor for loading points from an XML document
class k3dMeshLoadPoint
{
public:
	k3dMeshLoadPoint(k3dMesh* Mesh) { m_Mesh = Mesh; }

	bool HandleElement(sdpXmlIElement* Element)
	{
		k3dPoint* point = static_cast<k3dPoint*>(m_Mesh->CreatePoint());
		point->Load(Element);

		return true;
	}

	k3dMesh* m_Mesh;
};

sdpResult k3dMesh::LoadPoints(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlIElement* points = sdpXmlFindElement(Element, "points");
	if(!points)
		return SDP_S_FALSE;

	k3dMeshLoadPoint functor(this);
	sdpXmlForEachElement(points, "point", functor);

	return SDP_S_OK;
}

/// Functor for loading polygons from an XML document
class k3dMeshLoadPolygon
{
public:
	k3dMeshLoadPolygon(k3dMesh* Mesh) { m_Mesh = Mesh; }

	bool HandleElement(sdpXmlIElement* Element)
	{
		k3dPolygon* polygon = static_cast<k3dPolygon*>(m_Mesh->CreatePolygon());
		polygon->Load(Element, m_Mesh);

		return true;
	}

	k3dMesh* m_Mesh;
};

sdpResult k3dMesh::LoadPolygons(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlIElement* polygons = sdpXmlFindElement(Element, "polygons");
	if(!polygons)
		return SDP_S_FALSE;

	k3dMeshLoadPolygon functor(this);
	sdpXmlForEachElement(polygons, "polygon", functor);

	return SDP_S_OK;
}

/// Functor for loading NURBS curves from an XML document
class k3dMeshLoadNURBSCurve
{
public:
	k3dMeshLoadNURBSCurve(k3dMesh* Mesh) { m_Mesh = Mesh; }

	bool HandleElement(sdpXmlIElement* Element)
	{
		k3dNURBSCurve* curve = static_cast<k3dNURBSCurve*>(m_Mesh->CreateNURBSCurve());
		curve->Load(Element, m_Mesh);

		return true;
	}

	k3dMesh* m_Mesh;
};

sdpResult k3dMesh::LoadNURBSCurves(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlIElement* curves = sdpXmlFindElement(Element, "nurbscurves");
	if(!curves)
		return SDP_S_FALSE;

	k3dMeshLoadNURBSCurve functor(this);
	sdpXmlForEachElement(curves, "curve", functor);

	return SDP_S_OK;
}

/// Functor object used to load shading nodes
class k3dMeshLoadShadingNode
{
public:
	k3dMeshLoadShadingNode(k3dMesh* Mesh) { m_Mesh = Mesh; }

	bool HandleElement(sdpXmlIElement* Element)
	{
		// Create a new shading node ...
		k3dShadingNode* node = new k3dShadingNode();
		g_assert(node);
		m_Mesh->m_ShadingNodes.push_back(node);

		node->Load(Element, m_Mesh);

		return true;
	}

protected:	k3dMesh* m_Mesh;
};

sdpResult k3dMesh::LoadShading(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	sdpXmlIElement* shading = sdpXmlFindElement(Element, "shading");
	if(!shading)
		return SDP_S_FALSE;

	k3dMeshLoadShadingNode functor(this);
	sdpXmlForEachElement(shading, "node", functor);

	return SDP_S_OK;
}

sdpResult k3dMesh::LoadComplete()
{
	// Call the base-class version ...
	g_return_val_if_fail(sdpSucceeded(k3dTransformObject::LoadComplete()), SDP_E_FAIL);

	// For each polygon ...
	for(PolygonIterator polygon = m_Polygons.begin(); polygon != m_Polygons.end(); polygon++)
		(*polygon)->LoadComplete();

	return SDP_S_OK;
}

/*
k3dShadingNode* k3dMesh::CreateShadingNode()
{
	k3dShadingNode* newnode = m_ShadingNodes.create();
	g_return_val_if_fail(newnode, NULL);

	newnode->m_Mesh = this;

	return newnode;
}

void k3dMesh::DestroyShadingNode(k3dShadingNode* Node)
{
	g_assert(Node);
	m_ShadingNodes.erase(Node);
}

void k3dMesh::MoveSelectedToLast()
{
	k3dPolygon* nextpoly;
	k3dPolygon* lastpoly = m_Polygons.back();
	for(k3dPolygon* poly = m_Polygons.front(); poly != lastpoly; poly = nextpoly)
		{
			nextpoly = m_Polygons[i]->m_Next;

			if(SDP_S_OK != m_Polygons[i]->QueryState(k3dIState::VISIBLESELECTED)) continue;
			m_Polygons[i]->SetState(k3dIState::VISIBLESELECTED, FALSE);

			m_Polygons.moveback(poly);
		}
}

void k3dMesh::UpdateNumbers()
{
	if(m_Sheet)
		m_Sheet->m_MeshStatsPage->UpdateNumbers();
}

void k3dMesh::UnweldPoints()
{
	k3dPolygon* poly;
	k3dPolyPoint* polypoint;
	k3dPoint *point, *newpoint;
	gulong clientcount;

	// Eliminate shading ...
	m_ShadingNodes.clear();

	for(point = m_Points.front(); point; point = m_Points[i]->m_Next)
		{
			if(SDP_S_OK != m_Points[i]->QueryState(k3dIState::VISIBLESELECTED)) continue;

			clientcount = 0;

			for(poly = m_Polygons.front(); poly; poly = m_Polygons[i]->m_Next)
				{
					for(polypoint = m_Polygons[i]->m_PolyPoints.front(); polypoint; polypoint = polypoint->m_Next)
						{
							if(polypoint->m_Point != point) continue;

							clientcount++;
							if(clientcount < 2) continue;

							newpoint = (k3dPoint*)CreatePoint();
							newm_Points[i]->CopyFrom(point);

							polypoint->m_Point = newpoint;
						}
				}
		}

	UpdateNumbers();
}

void k3dMesh::WeldPoints()
{
	k3dPolygon* poly;
	k3dPolyPoint *polypoint, *firstpolypoint, *lastpolypoint, *nextpolypoint;
	k3dPoint *point, *point2, *nextpoint2;

	// Eliminate shading ...
	m_ShadingNodes.clear();

	// First phase - eliminate points with identical coordinates ...
	for(point = m_Points.front(); point; point = m_Points[i]->m_Next)
		{
			// If the point wasn't selected, skip it ...
			if(SDP_S_OK != m_Points[i]->QueryState(k3dIState::VISIBLESELECTED)) continue;

			// For each subsequent point ...
			for(point2 = m_Points[i]->m_Next; point2; point2 = nextpoint2)
				{
					nextpoint2 = point2->m_Next;

					// If the point wasn't selected, skip it ...
					if(SDP_S_OK != point2->QueryState(k3dIState::VISIBLESELECTED)) continue;

					// If the point doesn't have identical coordinates, skip it ...
					if(m_Points[i]->m_OldXYZ != point2->m_OldXYZ) continue;

					// Update polypoints before we get rid of the point ...
					for(poly = m_Polygons.front(); poly; poly = m_Polygons[i]->m_Next)
						{
							for(polypoint = m_Polygons[i]->m_PolyPoints.front(); polypoint; polypoint = polypoint->m_Next)
								{
									if(polypoint->m_Point != point2) continue;
									polypoint->m_Point = point;
								}
						}

					// Get rid of the extraneous point ...
					DeletePoint(point2);
				}
		}

	// Second phase - eliminate adjacent polypoints that share the same point
	for(poly = m_Polygons.front(); poly; poly = m_Polygons[i]->m_Next)
		{
			firstpolypoint = lastpolypoint = m_Polygons[i]->m_PolyPoints.front();
			if(!firstpolypoint) continue;

			for(polypoint = m_Polygons[i]->m_PolyPoints.front()->m_Next; polypoint; polypoint = nextpolypoint)
				{
					nextpolypoint = polypoint->m_Next;

					if(SDP_S_OK != polypoint->m_Point->QueryState(k3dIState::VISIBLESELECTED)) continue;

					if(polypoint->m_Point != lastpolypoint->m_Point) continue;
					m_Polygons[i]->RemovePolyPoint(polypoint);
				}
		}

	UpdateNumbers();
}

void k3dMesh::DeepSelect(bool State)
{
	m_ObjectContext->State()->SetState(k3dIState::SELECTED, State);

	ForEachPolygon(polygon->SetState(k3dIState::SELECTED, State));
	ForEachPoint(m_Points[i]->SetState(k3dIState::SELECTED, State));
}

void k3dMesh::DeleteIfEmpty()
{
	if(m_Points.size()) return;
	if(m_Polygons.size()) return;

	// If the document is being closed, all bets are off ...
	if(SDP_S_OK == Document()->DocumentClosing()) return;

	sdpString title;
	title.format("%s Mesh:", m_ObjectContext->Name());
	if(SDP_S_OK != m_IO->Query("The Mesh is empty.  Delete? (No undo)", title)) return;

	k3dDeleteObject(this);
}


#define ONETHIRD	0.333333333
#define TWOTHIRDS	0.666666667


void CDocument::OnEditPolysPatches()
{
	CLayer* layer;
	k3dMesh* mesh;
	k3dPolygon* poly;
	k3dPoint *point1, *point2, *point3, *point4;
	k3dPolyPoint *polypoint, *firstpolypoint, *nextpolypoint, *lastpolypoint;
	k3dPolyPoint *controlpoint1, *controlpoint2, *controlpoint3, *controlpoint4;

	k3dBoundingBox3 bbox, texturebbox;
	sdpVector3 coords[4], texturecoords[4];
	sdpVector3 center, texturecenter;

	for(layer = m_LayerList.front(); layer; layer = layer->m_Next)
		{
			for(mesh = layer->m_Meshes.front(); mesh; mesh = mesh->m_Next)
				{
					for(poly = mesh->m_Polygons.front(); poly; poly = m_Polygons[i]->m_Next)
						{
							if(SDP_S_OK != m_Polygons[i]->QueryState(k3dIState::VISIBLE)) continue;
							if(SDP_S_OK != m_Polygons[i]->QueryState(k3dIState::SELECTED)) continue;
							if(SDP_S_OK != m_Polygons[i]->QueryClosed()) continue;
							if(SDP_S_OK != m_Polygons[i]->QueryPolygon()) continue;
							if(4 != m_Polygons[i]->GetPointCount()) continue;

							m_Polygons[i]->SetClosed(TRUE);
							m_Polygons[i]->SetBezierPatch();

							gulong i = 0;
							bbox.Init();
							texturebbox.Init();
							firstpolypoint = m_Polygons[i]->m_PolyPoints.front();

							for(polypoint = m_Polygons[i]->m_PolyPoints.front(); polypoint;)
								{
									nextpolypoint = polypoint->m_Next;

									bbox.Expand(polypoint->m_Point->m_OldXYZ);
									texturebbox.Expand(polypoint->m_TextureXYZ);

									coords[i] = polypoint->m_Point->m_OldXYZ;
									texturecoords[i] = polypoint->m_TextureXYZ;

									controlpoint1 = m_Polygons[i]->CreatePolyPoint(NULL);
									controlpoint2 = m_Polygons[i]->CreatePolyPoint(NULL);

									point1 = mesh->CreatePoint();
									point2 = mesh->CreatePoint();

									controlpoint1->m_Point = point1;
									controlpoint2->m_Point = point2;


									if(nextpolypoint)
										{
											point1->m_OldXYZ = point1->m_LocalXYZ = polypoint->m_Point->m_OldXYZ + ONETHIRD *	(nextpolypoint->m_Point->m_OldXYZ - polypoint->m_Point->m_OldXYZ);
											point2->m_OldXYZ = point2->m_LocalXYZ = polypoint->m_Point->m_OldXYZ + TWOTHIRDS * (nextpolypoint->m_Point->m_OldXYZ - polypoint->m_Point->m_OldXYZ);
											controlpoint1->m_TextureXYZ = polypoint->m_TextureXYZ + ONETHIRD *	(nextpolypoint->m_TextureXYZ - polypoint->m_TextureXYZ);
											controlpoint2->m_TextureXYZ = polypoint->m_TextureXYZ + TWOTHIRDS * (nextpolypoint->m_TextureXYZ - polypoint->m_TextureXYZ);
										}
									else
										{
											point1->m_OldXYZ = point1->m_LocalXYZ = polypoint->m_Point->m_OldXYZ + ONETHIRD *	(firstpolypoint->m_Point->m_OldXYZ - polypoint->m_Point->m_OldXYZ);
											point2->m_OldXYZ = point2->m_LocalXYZ = polypoint->m_Point->m_OldXYZ + TWOTHIRDS * (firstpolypoint->m_Point->m_OldXYZ - polypoint->m_Point->m_OldXYZ);
											controlpoint1->m_TextureXYZ = polypoint->m_TextureXYZ + ONETHIRD *	(firstpolypoint->m_TextureXYZ - polypoint->m_TextureXYZ);
											controlpoint2->m_TextureXYZ = polypoint->m_TextureXYZ + TWOTHIRDS * (firstpolypoint->m_TextureXYZ - polypoint->m_TextureXYZ);
										}

									polypoint->m_Next = controlpoint1;
									controlpoint1->m_Next = controlpoint2;
									controlpoint2->m_Next = nextpolypoint;

									polypoint = nextpolypoint;
									lastpolypoint = controlpoint2;
									i++;
								}

							controlpoint1 = m_Polygons[i]->CreatePolyPoint(NULL);
							controlpoint2 = m_Polygons[i]->CreatePolyPoint(NULL);
							controlpoint3 = m_Polygons[i]->CreatePolyPoint(NULL);
							controlpoint4 = m_Polygons[i]->CreatePolyPoint(NULL);

							point1 = mesh->CreatePoint();
							point2 = mesh->CreatePoint();
							point3 = mesh->CreatePoint();
							point4 = mesh->CreatePoint();

							center = bbox.Center();
							texturecenter = texturebbox.Center();

							point1->m_OldXYZ = point1->m_LocalXYZ = ((coords[0] - center) * ONETHIRD) + center;
							point2->m_OldXYZ = point2->m_LocalXYZ = ((coords[1] - center) * ONETHIRD) + center;
							point3->m_OldXYZ = point3->m_LocalXYZ = ((coords[2] - center) * ONETHIRD) + center;
							point4->m_OldXYZ = point4->m_LocalXYZ = ((coords[3] - center) * ONETHIRD) + center;

							controlpoint1->m_TextureXYZ = ((texturecoords[0] - texturecenter) * ONETHIRD) + texturecenter;
							controlpoint2->m_TextureXYZ = ((texturecoords[1] - texturecenter) * ONETHIRD) + texturecenter;
							controlpoint3->m_TextureXYZ = ((texturecoords[2] - texturecenter) * ONETHIRD) + texturecenter;
							controlpoint4->m_TextureXYZ = ((texturecoords[3] - texturecenter) * ONETHIRD) + texturecenter;

							controlpoint1->m_Point = point1;
							controlpoint2->m_Point = point2;
							controlpoint3->m_Point = point3;
							controlpoint4->m_Point = point4;

							lastpolypoint->m_Next = controlpoint1;
							controlpoint1->m_Next = controlpoint2;
							controlpoint2->m_Next = controlpoint3;
							controlpoint3->m_Next = controlpoint4;
							controlpoint4->m_Next = NULL;
						}
				}
		}

	RedrawAllView(false);
}

void CDocument::OnEditPolysSplines()
{
	CLayer* layer;
	k3dMesh* mesh;
	k3dPolygon* poly;
	k3dPoint *point1, *point2;
	k3dPolyPoint *polypoint, *firstpolypoint, *nextpolypoint, *controlpoint1, *controlpoint2;

	for(layer = m_LayerList.front(); layer; layer = layer->m_Next)
		{
			for(mesh = layer->m_Meshes.front(); mesh; mesh = mesh->m_Next)
				{
					for(poly = mesh->m_Polygons.front(); poly; poly = m_Polygons[i]->m_Next)
						{
							if(SDP_S_OK != m_Polygons[i]->QueryState(k3dIState::VISIBLE)) continue;
							if(SDP_S_OK != m_Polygons[i]->QueryState(k3dIState::SELECTED)) continue;
							if(SDP_S_OK != m_Polygons[i]->QueryPolygon()) continue;

							m_Polygons[i]->SetBezierSpline();

							firstpolypoint = m_Polygons[i]->m_PolyPoints.front();

							for(polypoint = m_Polygons[i]->m_PolyPoints.front(); polypoint;)
								{
									nextpolypoint = polypoint->m_Next;

									controlpoint1 = m_Polygons[i]->CreatePolyPoint(NULL);
									controlpoint2 = m_Polygons[i]->CreatePolyPoint(NULL);

									controlpoint1->m_Polygon = poly;
									controlpoint2->m_Polygon = poly;

									point1 = mesh->CreatePoint();
									point2 = mesh->CreatePoint();

									controlpoint1->m_Point = point1;
									controlpoint2->m_Point = point2;

									if(nextpolypoint)
										{
											point1->m_OldXYZ.n[0] = point1->m_LocalXYZ.n[0] = polypoint->m_Point->m_OldXYZ.n[0] + 0.2 * (nextpolypoint->m_Point->m_OldXYZ.n[0] - polypoint->m_Point->m_OldXYZ.n[0]);
											point1->m_OldXYZ.n[1] = point1->m_LocalXYZ.n[1] = polypoint->m_Point->m_OldXYZ.n[1] + 0.2 * (nextpolypoint->m_Point->m_OldXYZ.n[1] - polypoint->m_Point->m_OldXYZ.n[1]);
											point1->m_OldXYZ.n[2] = point1->m_LocalXYZ.n[2] = polypoint->m_Point->m_OldXYZ.n[2] + 0.2 * (nextpolypoint->m_Point->m_OldXYZ.n[2] - polypoint->m_Point->m_OldXYZ.n[2]);
											point2->m_OldXYZ.n[0] = point2->m_LocalXYZ.n[0] = polypoint->m_Point->m_OldXYZ.n[0] + 0.8 * (nextpolypoint->m_Point->m_OldXYZ.n[0] - polypoint->m_Point->m_OldXYZ.n[0]);
											point2->m_OldXYZ.n[1] = point2->m_LocalXYZ.n[1] = polypoint->m_Point->m_OldXYZ.n[1] + 0.8 * (nextpolypoint->m_Point->m_OldXYZ.n[1] - polypoint->m_Point->m_OldXYZ.n[1]);
											point2->m_OldXYZ.n[2] = point2->m_LocalXYZ.n[2] = polypoint->m_Point->m_OldXYZ.n[2] + 0.8 * (nextpolypoint->m_Point->m_OldXYZ.n[2] - polypoint->m_Point->m_OldXYZ.n[2]);
										}
									else
										{
											point1->m_OldXYZ.n[0] = point1->m_LocalXYZ.n[0] = polypoint->m_Point->m_OldXYZ.n[0] + 0.2 * (firstpolypoint->m_Point->m_OldXYZ.n[0] - polypoint->m_Point->m_OldXYZ.n[0]);
											point1->m_OldXYZ.n[1] = point1->m_LocalXYZ.n[1] = polypoint->m_Point->m_OldXYZ.n[1] + 0.2 * (firstpolypoint->m_Point->m_OldXYZ.n[1] - polypoint->m_Point->m_OldXYZ.n[1]);
											point1->m_OldXYZ.n[2] = point1->m_LocalXYZ.n[2] = polypoint->m_Point->m_OldXYZ.n[2] + 0.2 * (firstpolypoint->m_Point->m_OldXYZ.n[2] - polypoint->m_Point->m_OldXYZ.n[2]);
											point2->m_OldXYZ.n[0] = point2->m_LocalXYZ.n[0] = polypoint->m_Point->m_OldXYZ.n[0] + 0.8 * (firstpolypoint->m_Point->m_OldXYZ.n[0] - polypoint->m_Point->m_OldXYZ.n[0]);
											point2->m_OldXYZ.n[1] = point2->m_LocalXYZ.n[1] = polypoint->m_Point->m_OldXYZ.n[1] + 0.8 * (firstpolypoint->m_Point->m_OldXYZ.n[1] - polypoint->m_Point->m_OldXYZ.n[1]);
											point2->m_OldXYZ.n[2] = point2->m_LocalXYZ.n[2] = polypoint->m_Point->m_OldXYZ.n[2] + 0.8 * (firstpolypoint->m_Point->m_OldXYZ.n[2] - polypoint->m_Point->m_OldXYZ.n[2]);
										}

									polypoint->m_Next = controlpoint1;
									controlpoint1->m_Next = controlpoint2;
									controlpoint2->m_Next = nextpolypoint;

									polypoint = nextpolypoint;
								}
						}
				}
		}

	RedrawAllViews(false);
}

void CDocument::OnEditSplinesPatches()
{
}

void CDocument::OnEditSplinesPolys()
{
}
*/

k3dID k3dMesh::GetNextPointID()
{
	// Increment the point ID ...
	k3dID result = ++m_CurrentPointID;

	// If this fires, we ran out of IDs!  Better switch to 64 bit architecture :)
	g_assert(result);

	return result;
}

k3dID k3dMesh::GetNextPolygonID()
{
	// Increment the polygon ID ...
	k3dID result = ++m_CurrentPolygonID;

	// If this fires, we ran out of IDs!  Better switch to 64 bit architecture :)
	g_assert(result);

	return result;
}

k3dID k3dMesh::GetNextNURBSCurveID()
{
	// Increment the curve ID ...
	k3dID result = ++m_CurrentNURBSCurveID;

	// If this fires, we ran out of IDs!  Better switch to 64 bit architecture :)
	g_assert(result);

	return result;
}

k3dID k3dMesh::GetNextPolyPointID()
{
	// Increment the polypoint ID ...
	k3dID result = ++m_CurrentPolyPointID;

	// If this fires, we ran out of IDs!  Better switch to 64 bit architecture :)
	g_assert(result);

	return result;
}

k3dPoint* k3dMesh::FindPoint(k3dID PointID)
{
	// Sanity checks ...
	g_assert(PointID);

	for(PointIterator point = m_Points.begin(); point != m_Points.end(); point++)
		if((*point)->ID() == PointID)
			return (*point);

	// Couldn't find it!!!
	g_return_val_if_fail(0, 0);
	return 0;
}

k3dPolyPoint* k3dMesh::FindPolyPoint(k3dID PolyPointID)
{
	// Sanity checks ...
	g_assert(PolyPointID);

	for(PolygonIterator polygon = m_Polygons.begin(); polygon != m_Polygons.end(); polygon++)
		for(k3dPolygon::PolyPointIterator polypoint = (*polygon)->m_PolyPoints.begin(); polypoint != (*polygon)->m_PolyPoints.end(); polypoint++)
			if((*polypoint)->ID() == PolyPointID)
				return (*polypoint);

	// Couldn't find it!!!
	g_return_val_if_fail(0, 0);
	return 0;
}


