
// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dRiEngine class, which provides an implementation of k3dIRiEngine
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3driengine.h>

k3dRiEngine::k3dRiEngine()
{
	// By default, we send RIB to stdout ...
	m_Stream = &std::cout;
	
	// Keep track of object and light handles ...
	m_ObjectHandle = 1;
	m_LightHandle = 1;
	
	// Setup our built-in parameters ...
	m_StandardDeclarations["P"] = "point";
	m_StandardDeclarations["N"] = "point";
	m_StandardDeclarations["s"] = "float";
	m_StandardDeclarations["t"] = "float";
	m_StandardDeclarations["visibility"] = "integer";
	m_StandardDeclarations["truedisplacement"] = "integer";
	m_StandardDeclarations["casts_shadows"] = "string";
	m_StandardDeclarations["emissioncolor"] = "color";
}

k3dRiEngine::~k3dRiEngine()
{
}

sdpResult k3dRiEngine::SetOutputStream(const std::ostream* Stream)
{
	// Sanity checks ...
	g_assert(Stream);
	g_return_val_if_fail(Stream->good(), SDP_E_FAIL);

	m_Stream = const_cast<std::ostream*>(Stream);
	
	return SDP_S_OK;
}

sdpResult k3dRiEngine::WriteRIBHeader()
{
	RiStructure("RenderMan RIB-Structure 1.0");
	(*m_Stream) << "version 3.03" << std::endl;
	
	return SDP_S_OK;
}

sdpString k3dRiEngine::FormatParameters(gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	sdpString result;
	
	for(gulong i = 0; i < ParameterCount; i++)
		{
			if(i)
				result += " ";
				
			result += FormatParameterArray(1, Tokens[i], Parameters[i]);
		}
		
	return result;
}

sdpString k3dRiEngine::FormatParameters(gulong ArraySize, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	// Sanity checks ...
	g_assert(ArraySize);

	sdpString result;
	
	for(gulong i = 0; i < ParameterCount; i++)
		{
			if(i)
				result += " ";
				
			result += FormatParameterArray(ArraySize, Tokens[i], Parameters[i]);
		}
	
	return result;
}

sdpString k3dRiEngine::FormatToken(SDPCSTRING Token)
{
	return sdpFormatString("\"%s\"", Token);
}

sdpString k3dRiEngine::FormatArray(SDPCSTRING Contents)
{
	return sdpFormatString("[%s]", Contents);
}

sdpString k3dRiEngine::FormatParameterArray(gulong ArraySize, SDPCSTRING Token, gpointer Parameters)
{
	// Sanity checks ...
	g_assert(ArraySize);
	g_assert_string(Token);
	g_assert(Parameters);

	sdpString result;

	// Format the token ...
	sdpString token = FormatToken(Token);

	// Lookup the token type ...
	sdpString tokentype;
	if(m_StandardDeclarations.find(Token) != m_StandardDeclarations.end())
		{
			tokentype = m_StandardDeclarations[Token];
		}
	else if(m_UserDeclarations.find(Token) != m_UserDeclarations.end())
		{
			tokentype = m_UserDeclarations[Token];
		}
	else
		{
			std::cerr << "*** k3dRiEngine::FormatParameters(): unknown token \"" << Token << "\"" << std::endl;
			return result;
		}

	// Get the array ...
	sdpString parameterarray;
	if(tokentype == "integer")
		{
			parameterarray = FormatIntegerArray(ArraySize, Parameters);
		}
	else if(tokentype == "float")
		{
			parameterarray = FormatFloatArray(ArraySize, Parameters);
		}
	else if(tokentype == "color")
		{
			parameterarray = FormatColorArray(ArraySize, Parameters);
		}
	else if(tokentype == "point")
		{
			parameterarray = FormatPointArray(ArraySize, Parameters);
		}
	else if(tokentype == "string")
		{
			parameterarray = FormatStringArray(ArraySize, Parameters);
		}
	else
		{
			std::cerr << "*** k3dRiEngine::FormatParameterArray(): " << token << " is an unknown type: " << tokentype << std::endl;
			return result;
		}

	result = token + " " + parameterarray;
		
	return result;
}

sdpString k3dRiEngine::FormatIntegerArray(gulong ArraySize, gpointer Parameters)
{
	// Sanity checks ...
	g_assert(ArraySize);
	g_assert(Parameters);
	
	sdpString result;
	
	gint* parameters = reinterpret_cast<gint*>(Parameters);
	for(gulong i = 0; i < ArraySize; i++)
		{
			if(i)
				result += " ";
				
			result += FormatInteger(parameters[i]);
		}
		
	return FormatArray(result);
}

sdpString k3dRiEngine::FormatFloatArray(gulong ArraySize, gpointer Parameters)
{
	// Sanity checks ...
	g_assert(ArraySize);
	g_assert(Parameters);
	
	sdpString result;
	
	gdouble* parameters = reinterpret_cast<gdouble*>(Parameters);
	
	for(gulong i = 0; i < ArraySize; i++)
		{
			if(i)
				result += " ";
				
			result += FormatFloat(parameters[i]);
		}
		
	return FormatArray(result);
}

sdpString k3dRiEngine::FormatColorArray(gulong ArraySize, gpointer Parameters)
{
	// Sanity checks ...
	g_assert(ArraySize);
	g_assert(Parameters);
	
	sdpString result;
	
	gdouble* parameters = reinterpret_cast<gdouble*>(Parameters);
	for(gulong i = 0; i < ArraySize; i++)
		{
			if(i)
				result += " ";
				
			result += FormatColor(parameters);
			
			parameters+=3;
		}
		
	return FormatArray(result);
}

sdpString k3dRiEngine::FormatPointArray(gulong ArraySize, gpointer Parameters)
{
	// Sanity checks ...
	g_assert(ArraySize);
	g_assert(Parameters);
	
	sdpString result;
	
	gdouble* parameters = reinterpret_cast<gdouble*>(Parameters);
	for(gulong i = 0; i < ArraySize; i++)
		{
			if(i)
				result += " ";
				
			result += FormatPoint(parameters);
			
			parameters+=3;
		}
		
	return FormatArray(result);
}

sdpString k3dRiEngine::FormatStringArray(gulong ArraySize, gpointer Parameters)
{
	// Sanity checks ...
	g_assert(ArraySize);
	g_assert(Parameters);
	
	sdpString result;
	
	SDPCSTRING* parameters = reinterpret_cast<SDPCSTRING*>(Parameters);
	for(gulong i = 0; i < ArraySize; i++)
		{
			if(i)
				result += " ";
				
			result += FormatString(*parameters);
			
			parameters++;
		}
		
	return FormatArray(result);
}

sdpString k3dRiEngine::FormatInteger(gint Integer)
{
	return sdpFormatString("%d", Integer);
}

sdpString k3dRiEngine::FormatFloat(gdouble Float)
{
	return sdpFormatString("%g", Float);
}

sdpString k3dRiEngine::FormatColor(gdouble Color[3])
{
	return sdpFormatString("%g %g %g", Color[0], Color[1], Color[2]);
}

sdpString k3dRiEngine::FormatPoint(gdouble Point[3])
{
	return sdpFormatString("%g %g %g", Point[0], Point[1], Point[2]);
}

sdpString k3dRiEngine::FormatString(SDPCSTRING String)
{
	return sdpFormatString("\"%s\"", String);
}

sdpString k3dRiEngine::FormatBound(gdouble Bound[6])
{
	return sdpFormatString("%g %g %g %g %g %g", Bound[0], Bound[1], Bound[2], Bound[3], Bound[4], Bound[5]);
}

sdpString k3dRiEngine::FormatTransform(gdouble Transform[16])
{
	return sdpFormatString("%g %g %g %g %g %g %g %g %g %g %g %g %g %g %g %g",
												Transform[0], Transform[1], Transform[2], Transform[3],
												Transform[4], Transform[5], Transform[6], Transform[7],
												Transform[8], Transform[9], Transform[10], Transform[11],
												Transform[12], Transform[13], Transform[14], Transform[15]);
}

sdpString k3dRiEngine::FormatBasis(gdouble Basis[16])
{
	return sdpFormatString("%g %g %g %g %g %g %g %g %g %g %g %g %g %g %g %g",
												Basis[0], Basis[1], Basis[2], Basis[3],
												Basis[4], Basis[5], Basis[6], Basis[7],
												Basis[8], Basis[9], Basis[10], Basis[11],
												Basis[12], Basis[13], Basis[14], Basis[15]);
}

sdpResult k3dRiEngine::RiDeclare(SDPCSTRING Name, SDPCSTRING Type)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	g_return_val_if_string_fail(Type, SDP_E_FAIL);

//std::cerr << "k3dRiEngine::RiDeclare " << Name << " " << Type << std::endl;

	// Check to make sure the user isn't trying to override a builtin ...
//	if(m_StandardDeclarations.find(Name) != m_StandardDeclarations.end())
//		{
//			std::cerr << "k3dRiEngine::RiDeclare() - cannot override built-in variable \"" << Name << "\"" << std::endl;
//			return SDP_E_FAIL;
//		}

	// Save the declaration for later ...
	m_UserDeclarations[Name] = Type;

	(*m_Stream) << "Declare " << FormatToken(Name) << " " << FormatToken(Type) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiFrameBegin(gint FrameNumber)
{
	(*m_Stream) << "FrameBegin " << FormatInteger(FrameNumber) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiFrameEnd()
{
	(*m_Stream) << "FrameEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiWorldBegin()
{
	(*m_Stream) << "WorldBegin" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiWorldEnd()
{
	(*m_Stream) << "WorldEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiFormat(gint XResolution, gint YResolution, gdouble AspectRatio)
{
	(*m_Stream) << "Format " << FormatInteger(XResolution) << " " << FormatInteger(YResolution) << " " << FormatFloat(AspectRatio) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiFrameAspectRatio(gdouble AspectRatio)
{
	(*m_Stream) << "FrameAspectRatio " << FormatFloat(AspectRatio) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiScreenWindow(gdouble Left, gdouble Right, gdouble Bottom, gdouble Top)
{
	(*m_Stream) << "ScreenWindow " << FormatFloat(Left) << " " << FormatFloat(Right) << " " << FormatFloat(Bottom) << " " << FormatFloat(Top) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCropWindow(gdouble XMin, gdouble XMax, gdouble YMin, gdouble YMax)
{
	(*m_Stream) << "CropWindow " << FormatFloat(XMin) << " " << FormatFloat(XMax) << " " << FormatFloat(YMin) << " " << FormatFloat(YMax) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiProjectionV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Projection " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiClipping(gdouble NearPlane, gdouble FarPlane)
{
	(*m_Stream) << "Clipping " << FormatFloat(NearPlane) << " " << FormatFloat(FarPlane) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDepthOfField(gdouble FStop, gdouble FocalLength, gdouble FocalDistance)
{
	(*m_Stream) << "DepthOfField " << FormatFloat(FStop) << " " << FormatFloat(FocalLength) << " " << FormatFloat(FocalDistance) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiShutter(gdouble OpenTime, gdouble CloseTime)
{
	(*m_Stream) << "Shutter " << FormatFloat(OpenTime) << " " << FormatFloat(CloseTime) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPixelFilter(SDPCSTRING FilterName, gdouble XWidth, gdouble YWidth)
{
	(*m_Stream) << "PixelFilter " << FormatToken(FilterName) << " " << FormatFloat(XWidth) << " " << FormatFloat(YWidth) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPixelVariance(gdouble Variation)
{
	(*m_Stream) << "PixelVariance " << FormatFloat(Variation) << std::endl;
	return SDP_S_OK;
}
sdpResult k3dRiEngine::RiPixelSamples(gdouble XSamples, gdouble YSamples)
{
	(*m_Stream) << "PixelSamples " << FormatFloat(XSamples) << " " << FormatFloat(YSamples) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiExposure(gdouble Gain, gdouble Gamma)
{
	(*m_Stream) << "Exposure " << FormatFloat(Gain) << " " << FormatFloat(Gamma) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiImagerV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Imager " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiQuantize(SDPCSTRING Type, gint One, gint QMin, gint QMax, gdouble Amplitude)
{
	(*m_Stream) << "Quantize " << FormatToken(Type) << " " << FormatInteger(One) << " " << FormatInteger(QMin) << " " << FormatInteger(QMax) << " " << FormatFloat(Amplitude) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDisplayV(SDPCSTRING Name, SDPCSTRING Type, SDPCSTRING Mode, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Display " << FormatToken(Name) << " " << FormatString(Type) << " " << FormatString(Mode) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiHiderV(SDPCSTRING Type, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Hider " << FormatToken(Type) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiColorSamples(gulong ParameterCount, gdouble nRGB[], gdouble RGBn[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiRelativeDetail(gdouble RelativeDetail)
{
	(*m_Stream) << "RelativeDetail " << FormatFloat(RelativeDetail) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiOptionV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Option " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiAttributeBegin()
{
	(*m_Stream) << "AttributeBegin" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiAttributeEnd()
{
	(*m_Stream) << "AttributeEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiColor(gdouble Color[3])
{
	(*m_Stream) << "Color " << FormatArray(FormatColor(Color)) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiOpacity(gdouble Color[3])
{
	(*m_Stream) << "Opacity " << FormatArray(FormatColor(Color)) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTextureCoordinates(gdouble S1, gdouble T1, gdouble S2, gdouble T2, gdouble S3, gdouble T3, gdouble S4, gdouble T4)
{
	(*m_Stream) << "TextureCoordinates " << FormatFloat(S1) << " " << FormatFloat(T1) << " " << FormatFloat(S2) << " " << FormatFloat(T2) << " " << FormatFloat(S3) << " " << FormatFloat(T3) << " " << FormatFloat(S4) << " " << FormatFloat(T4) << std::endl;
	return SDP_S_OK;
}

gulong k3dRiEngine::RiLightSourceV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	gulong lighthandle = m_LightHandle++;
	
	(*m_Stream) << "LightSource " << FormatToken(Name) << " " << lighthandle << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return lighthandle;
}

gulong k3dRiEngine::RiAreaLightSourceV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	gulong lighthandle = m_LightHandle++;
	
	(*m_Stream) << "AreaLightSource " << FormatToken(Name) << " " << lighthandle << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return lighthandle;
}

sdpResult k3dRiEngine::RiIlluminate(gulong LightHandle, bool OnOff)
{
	(*m_Stream) << "Illuminate " << LightHandle << " " << OnOff << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSurfaceV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Surface " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiAtmosphereV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Atmosphere " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiInteriorV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Interior " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiExteriorV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Exterior " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiShadingRate(gdouble Size)
{
	(*m_Stream) << "ShadingRate " << FormatFloat(Size) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiShadingInterpolation(SDPCSTRING Type)
{
	(*m_Stream) << "ShadingInterpolation " << FormatToken(Type) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiMatte(bool OnOff)
{
	(*m_Stream) << "Matte " << OnOff << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiBound(gdouble Bound[6])
{
	(*m_Stream) << "Bound " << FormatArray(FormatBound(Bound)) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDetail(gdouble Bound[6])
{
	(*m_Stream) << "Detail " << FormatArray(FormatBound(Bound)) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDetailRange(gdouble MinVis, gdouble LowTran, gdouble UpTran, gdouble MaxVis)
{
	(*m_Stream) << "DetailRange " << FormatFloat(MinVis) << " " << FormatFloat(LowTran) << " " << FormatFloat(UpTran) << " " << FormatFloat(MaxVis) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiGeometricApproximation(SDPCSTRING Type, gdouble Value)
{
	(*m_Stream) << "GeometricApproximation " << FormatToken(Type) << " " << FormatFloat(Value) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiGeometricRepresentation(SDPCSTRING Type)
{
	(*m_Stream) << "GeometricRepresentation " << FormatToken(Type) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiOrientation(SDPCSTRING Orientation)
{
	(*m_Stream) << "Orientation " << FormatToken(Orientation) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiReverseOrientation()
{
	(*m_Stream) << "ReverseOrientation" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSides(gint Sides)
{
	(*m_Stream) << "Sides " << FormatInteger(Sides) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiIdentity()
{
	(*m_Stream) << "Identity" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTransform(gdouble Transform[16])
{
	(*m_Stream) << "Transform " << FormatArray(FormatTransform(Transform)) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiConcatTransform(gdouble Transform[16])
{
	(*m_Stream) << "ConcatTransform " << FormatArray(FormatTransform(Transform)) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPerspective(gdouble FieldOfView)
{
	(*m_Stream) << "Perspective " << FormatFloat(FieldOfView) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTranslate(gdouble DX, gdouble DY, gdouble DZ)
{
	(*m_Stream) << "Translate " << FormatFloat(DX) << " " << FormatFloat(DY) << " " << FormatFloat(DZ) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiRotate(gdouble Angle, gdouble DX, gdouble DY, gdouble DZ)
{
	(*m_Stream) << "Rotate " << FormatFloat(Angle) << " " << FormatFloat(DX) << " " << FormatFloat(DY) << " " << FormatFloat(DZ) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiScale(gdouble DX, gdouble DY, gdouble DZ)
{
	(*m_Stream) << "Scale " << FormatFloat(DX) << " " << FormatFloat(DY) << " " << FormatFloat(DZ) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSkew(gdouble Angle, gdouble DX1, gdouble DY1, gdouble DZ1, gdouble DX2, gdouble DY2, gdouble DZ2)
{
	(*m_Stream) << "Skew " << FormatFloat(Angle) << " " << FormatFloat(DX1) << " " << FormatFloat(DY1) << " " << FormatFloat(DZ1) << " " << FormatFloat(DX2) << " " << FormatFloat(DY2) << " " << FormatFloat(DZ2) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDeformationV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Deformation " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDisplacementV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Displacement " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCoordinateSystem(SDPCSTRING Space)
{
	(*m_Stream) << "CoordinateSystem " << FormatToken(Space) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCoordSysTransform(SDPCSTRING Space)
{
	(*m_Stream) << "CoordSysTransform " << FormatToken(Space) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTransformBegin()
{
	(*m_Stream) << "TransformBegin" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTransformEnd()
{
	(*m_Stream) << "TransformEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiAttributeV(SDPCSTRING Name, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Attribute " << FormatToken(Name) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPolygonV(gulong VertexCount, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Polygon " << FormatParameters(VertexCount, ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiGeneralPolygonV(gulong LoopCount, gulong VertexCounts[], gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	// Calculate the total number of vertices ...
	gulong vertexcount = 0;
	for(gulong i = 0; i < LoopCount; i++)
		vertexcount += VertexCounts[i];

	(*m_Stream) << "GeneralPolygon " << FormatIntegerArray(LoopCount, VertexCounts) << " " << FormatParameters(vertexcount, ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPointsPolygonsV(gulong PolygonCount, gulong VertexCounts[], gulong Vertexes[], gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiPointsGeneralPolygonsV(gulong PolygonCount, gulong LoopCounts[], gulong VertexCounts[], gulong Vertices[], gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiBasis(gdouble UBasis[16], gint UStep, gdouble VBasis[16], gint VStep)
{
	(*m_Stream) << "Basis " << FormatArray(FormatBasis(UBasis)) << " " << FormatInteger(UStep) << " " << FormatArray(FormatBasis(VBasis)) << " " << FormatInteger(VStep) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPatchV(SDPCSTRING Type, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Patch " << FormatToken(Type) << " " << FormatParameters(4, ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiPatchMeshV(SDPCSTRING Type, gulong UCount, SDPCSTRING UWrap, gulong VCount,	SDPCSTRING VWrap, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiNuPatchV(gulong UCount, gint UOrder, gdouble UKnot[], gdouble UMin, gdouble UMax, gulong VCount, gint VOrder,	gdouble VKnot[], gdouble VMin, gdouble VMax, 	gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiTrimCurve(gulong LoopCount, gulong CurveCounts[], gulong Order[], gdouble Knot[], gdouble AMin[], gdouble AMax[], gulong ParameterCount, gdouble U[], gdouble V[], gdouble W[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiSphereV(gdouble Radius, gdouble ZMin, gdouble ZMax, gdouble ThetaMax, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Sphere " << FormatFloat(Radius) << " " << FormatFloat(ZMin) << " " << FormatFloat(ZMax) << " " << FormatFloat(ThetaMax) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiConeV(gdouble Height, gdouble Radius, gdouble ThetaMax, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Cone " << FormatFloat(Height) << " " << FormatFloat(Radius) << " " << FormatFloat(ThetaMax) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCylinderV(gdouble Radius, gdouble ZMin, gdouble ZMax, gdouble ThetaMax,	gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Cylinder " << FormatFloat(Radius) << " " << FormatFloat(ZMin) << " " << FormatFloat(ZMax) << " " << FormatFloat(ThetaMax) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiHyperboloidV(gdouble Point1[3], gdouble Point2[3], gdouble ThetaMax, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Hyperboloid " << FormatPoint(Point1) << " " << FormatPoint(Point2) << " " << FormatFloat(ThetaMax) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiParaboloidV(gdouble RMax, gdouble ZMin, gdouble ZMax, gdouble ThetaMax, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Paraboloid " << FormatFloat(RMax) << " " << FormatFloat(ZMin) << " " << FormatFloat(ZMax) << " " << FormatFloat(ThetaMax) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiDiskV(gdouble Height, gdouble Radius, gdouble ThetaMax, 	gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Disk " << FormatFloat(Height) << " " << FormatFloat(Radius) << " " << FormatFloat(ThetaMax) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiTorusV(gdouble MajorRadius, gdouble MinorRadius, gdouble PhiMin,	gdouble PhiMax, gdouble ThetaMax, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Torus " << FormatFloat(MajorRadius) << " " << FormatFloat(MinorRadius) << " " << FormatFloat(PhiMin) << " " << FormatFloat(PhiMax) << " " << FormatFloat(ThetaMax) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiCurvesV(SDPCSTRING Degree, gulong ParameterCountcurves, gulong ParameterCountverts[], SDPCSTRING Wrap, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiGeometryV(SDPCSTRING Type, gulong ParameterCount, SDPCSTRING Tokens[], gpointer Parameters[])
{
	(*m_Stream) << "Geometry " << FormatToken(Type) << " " << FormatParameters(ParameterCount, Tokens, Parameters) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSolidBegin(SDPCSTRING Type)
{
	(*m_Stream) << "SolidBegin " << FormatToken(Type) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiSolidEnd()
{
	(*m_Stream) << "SolidEnd" << std::endl;
	return SDP_S_OK;
}

gulong k3dRiEngine::RiObjectBegin()
{
	gulong objecthandle = m_ObjectHandle++;
	(*m_Stream) << "ObjectBegin " << objecthandle << std::endl;
	
	return objecthandle;
}

sdpResult k3dRiEngine::RiObjectEnd()
{
	(*m_Stream) << "ObjectEnd" << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiObjectInstance(gulong Handle)
{
	(*m_Stream) << "ObjectInstance " << Handle << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiMotionBeginV(gulong ParameterCount, gdouble Times[])
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiMotionEnd()
{
	(*m_Stream) << "MotionEnd" << std::endl; 
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiErrorIgnore(gint Code, gint Severity, SDPCSTRING Message)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiErrorPrint(gint Code, gint Severity, SDPCSTRING Message)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiErrorAbort(gint Code, gint Severity, SDPCSTRING Message)
{
	g_return_val_if_fail(0, SDP_E_FAIL);
	return SDP_E_FAIL;
}

sdpResult k3dRiEngine::RiComment(SDPCSTRING Comment)
{
	(*m_Stream) << "#" << Comment << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiNewline()
{
	(*m_Stream) << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiReadArchive(SDPCSTRING Archive)
{
	(*m_Stream) << "ReadArchive " << Archive << std::endl;
	return SDP_S_OK;
}

sdpResult k3dRiEngine::RiStructure(SDPCSTRING Structure)
{
	(*m_Stream) << "##" << Structure << std::endl;
	return SDP_S_OK;
}
