// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dShaderArgumentsPage class, which provides a UI for editing a set of shader arguments
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dshaderargumentspage.h>
#include <k3dcom.h>
#include <k3ditexture.h>

#include <sdpgtkevents.h>

#include <iostream>

//////////////////////////////////////////////////////////////////////
// ButtonBase

k3dShaderArgumentsPage::ButtonBase* k3dShaderArgumentsPage::ButtonBase::CreateButton(k3dIDocument* Document, k3dShaderInstance::ArgumentBase* Argument)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(Argument);

	switch(Argument->Type())
		{
			case k3dIShaderInfoArgument::UNKNOWN:
			case k3dIShaderInfoArgument::POINT:
			case k3dIShaderInfoArgument::VECTOR:
			case k3dIShaderInfoArgument::ORIENTATION:
			case k3dIShaderInfoArgument::SCALE:
			case k3dIShaderInfoArgument::NORMAL:
			case k3dIShaderInfoArgument::COORDINATESYSTEM:
			case k3dIShaderInfoArgument::BOOLEAN:
			case k3dIShaderInfoArgument::OPTIONLIST:
			case k3dIShaderInfoArgument::STRING:
				g_assert_not_reached();	// Unsupported argument type!!
			case k3dIShaderInfoArgument::TEXTURE:
				{
					k3dShaderInstance::TextureArgument* argument = static_cast<k3dShaderInstance::TextureArgument*>(Argument);
					return new TextureButton(Document, &argument->ObjectContainer());
				}
			case k3dIShaderInfoArgument::COLOR:
			case k3dIShaderInfoArgument::SCALAR:
			case k3dIShaderInfoArgument::TIME:
			case k3dIShaderInfoArgument::ANGLE:
			case k3dIShaderInfoArgument::DISTANCE:
			case k3dIShaderInfoArgument::AREA:
			case k3dIShaderInfoArgument::VOLUME:
			case k3dIShaderInfoArgument::MASS:
			case k3dIShaderInfoArgument::FORCE:
			case k3dIShaderInfoArgument::PRESSURE:
				{
					k3dShaderInstance::ChannelArgument* argument = static_cast<k3dShaderInstance::ChannelArgument*>(Argument);
					return new ChannelButton(Document, argument->Channel());
				}
		}

	g_assert_not_reached();	// Unknown argument type!!
	return 0;
}

k3dShaderArgumentsPage::ButtonBase::ButtonBase(k3dIDocument* Document)
{
	// Sanity checks ...
	g_assert(Document);
	
	m_Document = Document;
}

k3dShaderArgumentsPage::ButtonBase::~ButtonBase()
{
}

//////////////////////////////////////////////////////////////////////
// ChannelButton

k3dShaderArgumentsPage::ChannelButton::ChannelButton(k3dIDocument* Document, k3dIChannel* Channel)
#ifdef	SDPWIN32
	: ButtonBase(Document)
#else
	: k3dShaderArgumentsPage::ButtonBase::ButtonBase(Document)
#endif
{
	// Sanity checks ...
	g_assert(Channel);

	m_Channel = Channel;
}

k3dShaderArgumentsPage::ChannelButton::~ChannelButton()
{
}

bool k3dShaderArgumentsPage::ChannelButton::Create(GtkWidget* Widget)
{
	// Sanity checks ...
	g_assert(Widget);
	g_return_val_if_fail(m_Button.Create(m_Channel, false, GTK_CONTAINER(Widget)), false);

	return true;
}

//////////////////////////////////////////////////////////////////////
// TextureButton

k3dShaderArgumentsPage::TextureButton::TextureButton(k3dIDocument* Document, k3dObjectVariable<k3dITexture>* ObjectContainer)
#ifdef	SDPWIN32
	: ButtonBase(Document)
#else
	: k3dShaderArgumentsPage::ButtonBase::ButtonBase(Document)
#endif
{
	// Sanity checks ...
	g_assert(ObjectContainer);

	m_ObjectContainer = ObjectContainer;
}

k3dShaderArgumentsPage::TextureButton::~TextureButton()
{
}

bool k3dShaderArgumentsPage::TextureButton::Create(GtkWidget* Widget)
{
	// Sanity checks ...
	g_assert(Widget);
	
	g_return_val_if_fail(m_Selector. Create(m_Document, m_ObjectContainer, true, true, true, sdpCLSIDNull, k3dITexture::iid, GTK_CONTAINER(Widget)), false);
	return true;
}

//////////////////////////////////////////////////////////////////////
// k3dShaderArgumentsPage

k3dShaderArgumentsPage::k3dShaderArgumentsPage()
{
}

k3dShaderArgumentsPage::~k3dShaderArgumentsPage()
{
}

bool k3dShaderArgumentsPage::Create(k3dIDocument* Document, k3dShaderInstance* ShaderInstance, GtkContainer* ParentWidget)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert(ShaderInstance);
	g_assert(ParentWidget);

	// Store data ...
	m_Document = Document;
	m_ShaderInstance = ShaderInstance;
	m_ParentWidget = ParentWidget;

	Update();
	
	return true;
}

void k3dShaderArgumentsPage::Update()
{
	// Get rid of any old stuff ...
	Release();

	// Set the number of controls to create ...
	gulong controlcount = m_ShaderInstance->ArgumentCount();

	// If we don't have any controls, we're done ...
	if(0 == controlcount)
		return;

	// Create a table to organize the controls ...
	sdpGtkTable table;
	table.Create(controlcount, 2, false);

	// Attach the table to our parent widget ...
	sdpGtkContainer(m_ParentWidget).Attach(table);

	// For each control ...
	for(gulong i = 0; i < controlcount; i++)
		{
			// Get the argument ...
			k3dShaderInstance::ArgumentBase* argument = m_ShaderInstance->ArgumentAt(i);

			// Create something to attach our button to ...
			sdpGtkEventBox box;
			box.Create();
			table.Attach(box, 0, 1, i, i+1, GtkAttachOptions(GTK_EXPAND | GTK_FILL), GtkAttachOptions(GTK_EXPAND | GTK_FILL), 0, 0);

			// Create a button ...
			ButtonBase* button = ButtonBase::CreateButton(m_Document, argument);
			g_assert(button);
			button->Create(box);

			// Create a label ...
			sdpGtkLabel label;
			label.Create(argument->Name());
			table.Attach(label, 1, 2, i, i+1, GTK_FILL, GTK_FILL, 4, 0);
		}

	// Make everything visible ...
	table.ShowAll();

	// Set the table as our root object, so we can destroy everything in one fell swoop when done ...
	m_Object = table;
	
	// We want to be notified when the table widget (and children) are about to be destroyed ...
	MapEvent("destroy", "ondestroy", table, true);
}

void k3dShaderArgumentsPage::Release()
{
	// No more events from this point forward ...
	DisconnectAllEvents();
	
	// Get rid of our buttons ...
	for(ButtonIterator button = m_Buttons.begin(); button != m_Buttons.end(); button++)
		delete (*button);

	m_Buttons.clear();

	// Delete all of the GTK+ widgets ...
	if(Root())
		RootWidget().Destroy();

	// Clear out our storage ...
	Clear();
}

void k3dShaderArgumentsPage::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);
	
	if(Event->Name() == "ondestroy")
		OnDestroy();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void k3dShaderArgumentsPage::OnDestroy()
{
	// No more events from this point forward ...
	DisconnectAllEvents();
	
	// Get rid of our buttons ...
	for(ButtonIterator button = m_Buttons.begin(); button != m_Buttons.end(); button++)
		delete (*button);

	m_Buttons.clear();

	// Clear out our storage ...
	Clear();
}


