// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dState class, which provides a default implementation of k3dIState
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpxml.h>
#include <k3dstate.h>

k3dState::k3dState()
{
	m_StateFlags = 0;
}

sdpResult k3dState::QueryState(gulong State)
{
	return ((m_StateFlags & State) == State) ? SDP_S_OK : SDP_S_FALSE;
}

sdpResult k3dState::SetState(gulong Mask, bool State)
{
	if(State)
		m_StateFlags |= Mask;
	else
		m_StateFlags &= ~Mask;

	return SDP_S_OK;
}

sdpResult k3dState::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	sdpXmlIElement* state = sdpXmlCreateElement(Element, "state", 0);
	g_return_val_if_fail(state, SDP_E_FAIL);

	if(m_StateFlags & VISIBLE)
		state->CreateAttribute("visible", "true");
			
	if(m_StateFlags & SELECTED)
		state->CreateAttribute("selected", "true");
			
	if(m_StateFlags & SELECTABLE)
		state->CreateAttribute("selectable", "true");
			
	if(m_StateFlags & USERVIEWABLE)
		state->CreateAttribute("userviewable", "true");
			
	return SDP_S_OK;
}

sdpResult k3dState::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	// Reset our state ...
	m_StateFlags = 0;

	// Implied by the fact that we're being loaded ... duh!
	k3dState::SetState(SAVEABLE, true);

	// See what we find in the file ...
	sdpXmlIElement* state = sdpXmlFindElement(Element, "state");
	if(!state)
		return SDP_S_FALSE;

	bool visible = false;
	sdpXmlGetAttribute(state, "visible", visible);
	k3dState::SetState(VISIBLE, visible);

	bool selected = false;
	sdpXmlGetAttribute(state, "selected", selected);
	k3dState::SetState(SELECTED, selected);

	bool selectable = false;
	sdpXmlGetAttribute(state, "selectable", selectable);
	k3dState::SetState(SELECTABLE, selectable);

	bool userviewable = false;
	sdpXmlGetAttribute(state, "userviewable", userviewable);
	k3dState::SetState(USERVIEWABLE, userviewable);

	return SDP_S_OK;
}

