// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dTool class, which encapsulates a K-3D tool object (a K-3D object that isn't designed to be persistent or serializable)
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <k3dtool.h>
#include <k3distate.h>

k3dTool::k3dTool()
{
	m_ActiveAxis = k3dIDocument::ZAXIS;
}

sdpResult k3dTool::QueryRenameObject()
{
	return SDP_S_FALSE;
}

sdpResult k3dTool::SetObjectContext(k3dIObjectContext* ObjectContext)
{
	if(sdpFailed(k3dUserObject::SetObjectContext(ObjectContext)))
		return SDP_E_FAIL;

	// Make sure this object doesn't show up in the hierarchy ...
	ObjectContext->State()->SetState(k3dIState::USERVIEWABLE, false);

	return SDP_S_OK;
}

void k3dTool::OnSet()
{
//	Document()->SetDatabase(this, TRUE);
	k3dUserObject::OnSet();
}

void k3dTool::OnModify(bool SaveAction)
{
	// Get the active axis ...
	m_ActiveAxis = Document()->GetActiveAxis();

	// For each object in the document ...
	k3dIObjectContext* context;
	for(K3DOBJECTCONTEXTHANDLE handle = Document()->GetFirstObjectContext(); handle; handle = Document()->GetNextObjectContext(handle))
		{
			context = Document()->GetObjectContext(handle);

			// If the object isn't visible, skip it ...
			if(SDP_S_OK != context->State()->QueryState(k3dIState::VISIBLE))
				continue;

			// If the object itself is selected, modify it as a whole ...
			if(SDP_S_OK == context->State()->QueryState(k3dIState::SELECTED))
				{
					ModifyObject(context->Object());
				}
			// Otherwise, try modifying parts of the object using k3dILocationServer ...
			else
				{
					ModifyLocationServer(context->Object());
				}
		}

	// Update the display
	k3dUserObject::OnModify(SaveAction);
}

void k3dTool::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Default implementation is to fall back to a location-by-location approach ...
	ModifyLocationServer(Object);
}

void k3dTool::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// See if the object supports k3dILocationServer ...
	k3dILocationServer* locationserver = sdpQueryInterface<k3dILocationServer>(Object);
	if(0 == locationserver)
		return;

	// If it does, get the object's size / origin ...
	GetObjectInfo(locationserver);

	// ... And give each location a chance to be modified ...
	k3dILocation* location;
	for(K3DLOCATIONHANDLE locationhandle = locationserver->GetFirstLocation(); locationhandle; locationhandle = locationserver->GetNextLocation(locationhandle))
		{
			// See if the location is visible and selected ...
			location = locationserver->GetLocation(locationhandle);
			if(SDP_S_OK != location->State()->QueryState(k3dIState::VISIBLESELECTED))
				continue;

			// Modify the location ...
			ModifyLocation(location);
		}

	locationserver->Release();
}

void k3dTool::ModifyLocation(k3dILocation* Location)
{
}

void k3dTool::GetObjectInfo(k3dILocationServer* LocationServer)
{
	sdpVector3 coords;
	K3DLOCATIONHANDLE locationhandle;

	// See if it's a mesh ...
	k3dIMesh* mesh = sdpQueryInterface<k3dIMesh>(LocationServer);
	if(mesh)
		{
			mesh->GetDimensions(k3dIState::SELECTED, m_Min, m_Max);
			mesh->Release();
			goto finish;
		}

	// Otherwise, figure it all out manually ...
	m_Min = DBL_MAX;
	m_Max = -DBL_MAX;
	for(locationhandle = LocationServer->GetFirstLocation(); locationhandle; locationhandle = LocationServer->GetNextLocation(locationhandle))
		{
			LocationServer->GetLocation(locationhandle)->GetOldXYZ(coords);
			m_Max = sdpVectorMax(m_Max, coords);
			m_Min = sdpVectorMin(m_Min, coords);
		}

finish:
	m_Size = m_Max - m_Min;
	m_Center = m_Size * 0.5;
	m_OriginSize = sdpVectorMax(m_Max, -m_Min);
}



