// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dTransformObject class, which serves as a base class for K-3D objects that have position, orientation, scale, and can render themselves
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3dtransformobject.h>
#include <k3dmisc.h>

#include <k3didocument.h>
#include <k3dilocationserver.h>
#include <k3dicamera.h>

#include <sdpgtkwidget.h>
#include <sdpgl.h>

//////////////////////////////////////////////////////////////////////////////////
// k3dTransformObject ...

k3dTransformObject::k3dTransformObject()
{
	m_TransformType = NONE;
}


k3dTransformObject::~k3dTransformObject()
{
}

///////////////////////////////////////////////////////////////////
// k3dIPersist overrides ...

sdpResult k3dTransformObject::Save(sdpXmlIElement* Element, k3dIDependencies* Dependencies)
{
	// Sanity checks ...
	g_assert(Element);
	g_assert(Dependencies);

	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::Save(Element, Dependencies)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dTransform::Save(this, Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult k3dTransformObject::Load(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(Element);

	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::Load(Element)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(k3dTransform::Load(this, Element)), SDP_E_FAIL);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////
// k3dIEditorRender overrides ...

sdpResult k3dTransformObject::EditorPushRender(k3dIEditorEngine* Engine)
{
	sdpglTrace("k3dTransformObject::EditorPushRender()");

	k3dEditorRender::EditorPushRender(Engine);
	k3dPushOpenGLMatrix(k3dGetTransformMatrix(this, k3dITransform::ANIMATED_FRAME));

	return SDP_S_OK;
}

sdpResult k3dTransformObject::EditorPushChildRender(k3dIEditorEngine* Engine)
{
	sdpglTrace("k3dTransformObject::EditorPushChildRender()");
	
	k3dEditorRender::EditorPushChildRender(Engine);
	k3dPushOpenGLMatrix(k3dGetTransformMatrix(this, k3dITransform::ANIMATED_CHILD_FRAME));
	
	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////
// k3dIRiRender overrides ...

sdpResult k3dTransformObject::RiPushRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	k3dRiRender::RiPushRender(Frame, Engine);
	k3dPushRiMatrix(k3dGetTransformMatrix(this, k3dITransform::ANIMATED_FRAME), Engine);

	return SDP_S_OK;
}

sdpResult k3dTransformObject::RiPushChildRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Frame, SDP_E_FAIL);
	g_return_val_if_fail(Engine, SDP_E_FAIL);
	
	k3dRiRender::RiPushChildRender(Frame, Engine);
	k3dPushRiMatrix(k3dGetTransformMatrix(this, k3dITransform::ANIMATED_CHILD_FRAME), Engine);
	
	return SDP_S_OK;
}

//////////////////////////////////////////////////////////////////
// k3dITarget overrides ...

sdpResult k3dTransformObject::GetTargetCoordinates(double Coords[3])
{
	// Get our location in world coordinates ...
	sdpVector3 worldcoords(0.0);

g_assert_not_reached();
//	GetWorldCoordinate(m_ObjectContext, worldcoords);

	// Copy the results ...
	worldcoords.CopyArray(Coords);

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////////////
// k3dIActionClient overrides ...

sdpResult k3dTransformObject::HandleStep(gulong ValueCode, gulong Value, gulong ValueSize)
{
	if(ValueCode == STEP_POSITION)
		{
			m_Position[k3dITransform::CURRENT_FRAME] = *reinterpret_cast<sdpVector3*>(Value);
			return SDP_S_OK;
		}
	else if(ValueCode == STEP_ORIENTATION)
		{
			m_Orientation[k3dITransform::CURRENT_FRAME] = *reinterpret_cast<sdpAngleAxis*>(Value);
			return SDP_S_OK;
		}
	else if(ValueCode == STEP_SCALE)
		{
			m_Scale[k3dITransform::CURRENT_FRAME] = *reinterpret_cast<sdpVector3*>(Value);
			return SDP_S_OK;
		}
	else if(ValueCode == STEP_CHILD_POSITION)
		{
			m_Position[k3dITransform::CURRENT_CHILD_FRAME] = *reinterpret_cast<sdpVector3*>(Value);
			return SDP_S_OK;
		}
	else if(ValueCode == STEP_CHILD_ORIENTATION)
		{
			m_Orientation[k3dITransform::CURRENT_CHILD_FRAME] = *reinterpret_cast<sdpAngleAxis*>(Value);
			return SDP_S_OK;
		}
	else if(ValueCode == STEP_CHILD_SCALE)
		{
			m_Scale[k3dITransform::CURRENT_CHILD_FRAME] = *reinterpret_cast<sdpVector3*>(Value);
			return SDP_S_OK;
		}
		
	return k3dUserObject::HandleStep(ValueCode, Value, ValueSize);
}

/////////////////////////////////////////////////////////////////////////////
// k3dITransform overrides ...

sdpResult k3dTransformObject::SetPosition(const k3dITransform::ReferenceFrame Frame, gdouble Position[3])
{
	// Are we recording actions right now?
	if(SDP_S_OK == Document()->Actions()->IsRecording())
		{
			if(k3dITransform::CURRENT_FRAME == Frame)
				Document()->Actions()->RecordStep(this, STEP_POSITION, reinterpret_cast<gulong>(new sdpVector3(m_Position[k3dITransform::OLD_FRAME])), sizeof(sdpVector3), STEP_POSITION, reinterpret_cast<gulong>(new sdpVector3(Position)), sizeof(sdpVector3));
			else if(k3dITransform::CURRENT_CHILD_FRAME == Frame)
				Document()->Actions()->RecordStep(this, STEP_CHILD_POSITION, reinterpret_cast<gulong>(new sdpVector3(m_Position[k3dITransform::OLD_CHILD_FRAME])), sizeof(sdpVector3), STEP_CHILD_POSITION, reinterpret_cast<gulong>(new sdpVector3(Position)), sizeof(sdpVector3));
		}
		
	// Make the change ...
	k3dTransform::SetPosition(Frame,Position);

	return SDP_S_OK;
}

sdpResult k3dTransformObject::SetOrientation(const k3dITransform::ReferenceFrame Frame, gdouble AngleAxis[4])
{
	// Are we recording actions right now?
	if(SDP_S_OK == Document()->Actions()->IsRecording())
		{
			if(k3dITransform::CURRENT_FRAME == Frame)
				Document()->Actions()->RecordStep(this, STEP_ORIENTATION, reinterpret_cast<gulong>(new sdpAngleAxis(m_Orientation[k3dITransform::OLD_FRAME])), sizeof(sdpAngleAxis), STEP_ORIENTATION, reinterpret_cast<gulong>(new sdpAngleAxis(AngleAxis)), sizeof(sdpAngleAxis));
			else if(k3dITransform::CURRENT_CHILD_FRAME == Frame)
				Document()->Actions()->RecordStep(this, STEP_CHILD_ORIENTATION, reinterpret_cast<gulong>(new sdpAngleAxis(m_Orientation[k3dITransform::OLD_CHILD_FRAME])), sizeof(sdpAngleAxis), STEP_CHILD_ORIENTATION, reinterpret_cast<gulong>(new sdpAngleAxis(AngleAxis)), sizeof(sdpAngleAxis));
		}
		
	// Make the change ...
	k3dTransform::SetOrientation(Frame, AngleAxis);

	return SDP_S_OK;
}

sdpResult k3dTransformObject::SetScale(const k3dITransform::ReferenceFrame Frame, gdouble Scale[3])
{
	// Are we recording actions right now?
	if(SDP_S_OK == Document()->Actions()->IsRecording())
		{
			if(k3dITransform::CURRENT_FRAME == Frame)
				Document()->Actions()->RecordStep(this, STEP_SCALE, reinterpret_cast<gulong>(new sdpVector3(m_Scale[k3dITransform::OLD_FRAME])), sizeof(sdpVector3), STEP_SCALE, reinterpret_cast<gulong>(new sdpVector3(Scale)), sizeof(sdpVector3));
			else if(k3dITransform::CURRENT_CHILD_FRAME == Frame)
				Document()->Actions()->RecordStep(this, STEP_CHILD_SCALE, reinterpret_cast<gulong>(new sdpVector3(m_Scale[k3dITransform::OLD_CHILD_FRAME])), sizeof(sdpVector3), STEP_CHILD_SCALE, reinterpret_cast<gulong>(new sdpVector3(Scale)), sizeof(sdpVector3));
		}
		
	// Make the change ...
	k3dTransform::SetScale(Frame, Scale);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// k3dIUserEvent overrides ...

sdpResult k3dTransformObject::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	switch(Document()->GetActiveTransformType())
		{
			case k3dIDocument::POSITION:
				m_TransformType = TRANSLATEFREE;
				break;
			case k3dIDocument::ORIENTATION:
				m_TransformType = ROTATEFREE;
				break;
			case k3dIDocument::SCALE:
				m_TransformType = SCALEFREE;
				break;
			default:
				g_assert_not_reached();	// Did you add a new transform type without telling me?
		}

	return SDP_S_OK;
}

sdpResult k3dTransformObject::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	switch(m_TransformType)
		{
			case NONE:
				break;

			case ROTATEX:
			case ROTATEY:
			case ROTATEZ:
			case ROTATEFREE:
				Rotate(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			case TRANSLATEX:
			case TRANSLATEY:
			case TRANSLATEZ:
			case TRANSLATEFREE:
				Translate(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			case POSITIVESCALEX:
			case NEGATIVESCALEX:
			case POSITIVESCALEY:
			case NEGATIVESCALEY:
			case POSITIVESCALEZ:
			case NEGATIVESCALEZ:
			case SCALEFREE:
				Scale(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			default:
				g_assert_not_reached();	// Did you add a new transform type without telling me?
		}

	return SDP_S_OK;
}

sdpResult k3dTransformObject::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	switch(m_TransformType)
		{
			case NONE:
				break;

			case ROTATEX:
			case ROTATEY:
			case ROTATEZ:
			case ROTATEFREE:
				Document()->Actions()->StartAction("Rotate");
				Rotate(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			case TRANSLATEX:
			case TRANSLATEY:
			case TRANSLATEZ:
			case TRANSLATEFREE:
				Document()->Actions()->StartAction("Translate");
				Translate(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			case POSITIVESCALEX:
			case NEGATIVESCALEX:
			case POSITIVESCALEY:
			case NEGATIVESCALEY:
			case POSITIVESCALEZ:
			case NEGATIVESCALEZ:
			case SCALEFREE:
				Document()->Actions()->StartAction("Scale");
				Scale(Camera, Modifiers, sdpVector2(CurrentX, CurrentY), sdpVector2(LastX, LastY), sdpVector2(StartX, StartY));
				break;

			default:
				g_assert_not_reached();	// Did you add a new transform type without telling me?
		}

	Document()->Actions()->FinishAction();
	Document()->FreezeChanges();

	m_TransformType = NONE;

	return SDP_S_OK;
}

void k3dTransformObject::Translate(k3dICamera* Camera, GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 LastMouse, sdpVector2 StartMouse)
{
	// Sanity checks ...
	g_assert(Camera);
	g_return_if_fail(Camera->Width());
	g_return_if_fail(Camera->Height());

	// If the mouse didn't really move, just reset the same position (in case we're recording UNDOs)
	if(LastMouse == CurrentMouse)
		{
			sdpVector3 position;
			GetPosition(k3dITransform::CURRENT_FRAME, position);
			SetPosition(k3dITransform::CURRENT_FRAME, position);
			return;
		}

	// Turn mouse coordinates into a vector in world coordinates ...
	sdpVector3 startorigin, startdirection, currentorigin, currentdirection;
	k3dMouseToWorldRay(Camera, StartMouse, startorigin, startdirection);
	k3dMouseToWorldRay(Camera, CurrentMouse, currentorigin, currentdirection);

	// Get our position in world coordinates ...
	sdpVector3 worldposition = k3dGetWorldPosition(ObjectContext());

	// Calculate the mouse's intersection with a plane, based on the active axis ...
	sdpVector3 startintersection, currentintersection;
	
	if(ControlDown(Modifiers))
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), worldposition), startorigin, startdirection, startintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					case k3dIDocument::YAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), worldposition), startorigin, startdirection, startintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					case k3dIDocument::ZAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), worldposition), startorigin, startdirection, startintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					default:
						g_assert_not_reached();	// Did you add a new axis type without telling me?
				}
		}
	else
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), worldposition), startorigin, startdirection, startintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(1, 0, 0), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					case k3dIDocument::YAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), worldposition), startorigin, startdirection, startintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 1, 0), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					case k3dIDocument::ZAXIS:
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), worldposition), startorigin, startdirection, startintersection);
						sdpPlaneLineIntersection(sdpPlane(sdpVector3(0, 0, 1), worldposition), currentorigin, currentdirection, currentintersection);
						break;
					default:
						g_assert_not_reached();	// Did you add a new axis type without telling me?
				}
		}

	// Calculate our change in world coordinates ...
	sdpVector3 worlddelta = currentintersection - startintersection;

	// Constrain movement to the active axis ...
	if(ControlDown(Modifiers))
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						worlddelta[1] = worlddelta[2] = 0;
						break;
					case k3dIDocument::YAXIS:
						worlddelta[0] = worlddelta[2] = 0;
						break;
					case k3dIDocument::ZAXIS:
						worlddelta[0] = worlddelta[1] = 0;
						break;
					default:
						g_assert_not_reached();	// Did you add a new axis type without telling me?
				}
		}

	sdpVector3 position;
	GetPosition(k3dITransform::OLD_FRAME, position);
	SetPosition(k3dITransform::CURRENT_FRAME, position + worlddelta);
	k3dTransformObject::OnModify(false);
}

static sdpVector3 Trackball(sdpVector2 ScreenCoords)
{
	// Convert screen coordinates to trackball coordinates ...
	sdpVector3 trackball(ScreenCoords, 0);
	gdouble radius = trackball.Length();
	if(radius < 1)
		trackball[2] = -sqrt(1 - radius * radius);
	else
		trackball.Normalize();

	return trackball;
}

void k3dTransformObject::Rotate(k3dICamera* Camera, GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 LastMouse, sdpVector2 StartMouse)
{
	// Sanity checks ...
	g_assert(Camera);
	g_return_if_fail(Camera->Width());
	g_return_if_fail(Camera->Height());

	// If the mouse didn't really move, just reset the same rotation (in case we're recording UNDOs)
	if(LastMouse == CurrentMouse)
		{
			sdpAngleAxis rotation;
			GetOrientation(k3dITransform::CURRENT_FRAME, rotation);
			SetOrientation(k3dITransform::CURRENT_FRAME, rotation);
			return;
		}

	// Convert mouse coordinates to screen coordinates ...
	sdpVector2 lastmouse, currentmouse;
	gdouble width = Camera->Width();
	gdouble height = Camera->Height();
	if(width > height)
		{
			lastmouse = k3dTransformMouseCoords(Camera, LastMouse, -width / height, width / height, 1, -1);
			currentmouse = k3dTransformMouseCoords(Camera, CurrentMouse, -width / height, width / height, 1, -1);
		}
	else
		{
			lastmouse = k3dTransformMouseCoords(Camera, LastMouse, -1, 1, -height / width, height / width);
			currentmouse = k3dTransformMouseCoords(Camera, CurrentMouse, -1, 1, -height / width, height / width);
		}

	// Convert screen coordinates to "trackball" coordinates ...		
	static const gdouble trackballsize = 1 / 0.9;
	sdpVector3 last = Trackball(lastmouse * trackballsize);
	sdpVector3 current = Trackball(currentmouse * trackballsize);
	
	// Transform "trackball" coordinates to model coordinates ...
	sdpMatrix4 cameratoworld = ObjectContext()->Parent() ? k3dGetCameraToObjectMatrix(Camera, ObjectContext()->Parent()) : k3dGetCameraToWorldMatrix(Camera);
	last = (cameratoworld * last) - (cameratoworld * sdpVector3(0, 0, 0));
	current = (cameratoworld * current) - (cameratoworld * sdpVector3(0, 0, 0));
				
	// Convert the change in mouse position to a quaternion ...	
	sdpQuaternion delta(last * current, last ^ current);

	// "Add" the quaternion to our existing orientation ...
	sdpAngleAxis currentrotation;
	GetOrientation(k3dITransform::CURRENT_FRAME, currentrotation);
	SetOrientation(k3dITransform::CURRENT_FRAME, sdpAngleAxis(delta * sdpQuaternion(currentrotation)));

	k3dTransformObject::OnModify(false);
}

void k3dTransformObject::Scale(k3dICamera* Camera, GdkModifierType Modifiers, sdpVector2 CurrentMouse, sdpVector2 LastMouse, sdpVector2 StartMouse)
{
	// Sanity checks ...
	g_assert(Camera);
	g_return_if_fail(Camera->Width());
	g_return_if_fail(Camera->Height());

	// If the mouse didn't really move, just reset the same scale (in case we're recording UNDOs)
	if(LastMouse == CurrentMouse)
		{
			sdpVector3 scale;
			GetScale(k3dITransform::CURRENT_FRAME, scale);
			SetScale(k3dITransform::CURRENT_FRAME, scale);
			return;
		}

	// Convert mouse coordinates to screen coordinates ...
	sdpVector2 startmouse, currentmouse;
	gdouble width = Camera->Width();
	gdouble height = Camera->Height();
	if(width > height)
		{
			startmouse = k3dTransformMouseCoords(Camera, StartMouse, -width / height, width / height, 1, -1);
			currentmouse = k3dTransformMouseCoords(Camera, CurrentMouse, -width / height, width / height, 1, -1);
		}
	else
		{
			startmouse = k3dTransformMouseCoords(Camera, StartMouse, -1, 1, -height / width, height / width);
			currentmouse = k3dTransformMouseCoords(Camera, CurrentMouse, -1, 1, -height / width, height / width);
		}

	sdpVector2 deltamouse = currentmouse - startmouse;
	
	sdpVector3 delta;
	if(ControlDown(Modifiers))
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						delta = sdpVector3(deltamouse[0], 0, 0);
						break;
					case k3dIDocument::YAXIS:
						delta = sdpVector3(0, deltamouse[1], 0);
						break;
					case k3dIDocument::ZAXIS:
						delta = sdpVector3(0, 0, deltamouse[0]);
						break;
					default:
						g_assert_not_reached();	// Did you add a new axis type without telling me?
				}
		}
	else
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						delta = sdpVector3(0, deltamouse[1], deltamouse[0]);
						break;
					case k3dIDocument::YAXIS:
						delta = sdpVector3(deltamouse[0], 0, deltamouse[1]);
						break;
					case k3dIDocument::ZAXIS:
						delta = sdpVector3(deltamouse[0], deltamouse[1], 0);
						break;
					default:
						g_assert_not_reached();	// Did you add a new axis type without telling me?
				}
		}

	sdpVector3 scale;
	GetScale(k3dITransform::OLD_FRAME, scale);
	SetScale(k3dITransform::CURRENT_FRAME, scale + delta);
	k3dTransformObject::OnModify(false);
}

// Draw a single transformation handle
static void DrawTransformationHandle(GLUquadricObj* Quadric, double Radius)
{
	static GLint slices = 32;

	glDisable(GL_LIGHTING);

	// Draw the circle ...
	gluQuadricDrawStyle(Quadric, GLenum(GLU_SILHOUETTE));
	gluQuadricNormals(Quadric, GLenum(GLU_SMOOTH));
	gluDisk(Quadric, 0.0, Radius, slices, 1);

	// Draw the axis ...
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, -Radius);
	glVertex3d(0.0, 0.0, Radius);
	glEnd();

	// Draw the arrows ...
	glEnable(GL_LIGHTING);
	gluQuadricDrawStyle(Quadric, GLenum(GLU_FILL));

	glPushMatrix();
	glTranslated(0.0, 0.0, Radius);
	gluCylinder(Quadric, Radius * 0.05, Radius * 0.001, Radius * 0.15, 8, 1);
	glPopMatrix();

	glPushMatrix();
	glTranslated(0.0, 0.0, -Radius);
	glRotated(180.0, 1.0, 0.0, 0.0);
	gluCylinder(Quadric, Radius * 0.05, Radius * 0.001, Radius * 0.15, 8, 1);
	glPopMatrix();
}

// Draw all three transformation handles
void k3dTransformObject::DrawTransformationHandles(gdouble Radius)
{
	// If we aren't the active object, we're done ...
	if(Document()->GetActiveObjectContext() != ObjectContext())
		return;

	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glMatrixMode(GL_MODELVIEW);

	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);

	GLUquadricObj* quadric = gluNewQuadric();
	g_return_if_fail(quadric);

	static GLfloat xcolor[3] = { 1, 0, 0 };
	static GLfloat ycolor[3] = { 1, 1, 0 };
	static GLfloat zcolor[3] = { 0, 1, 0 };
	
	static GLfloat xemissioncolor[3] = { 0.5, 0, 0 };
	static GLfloat yemissioncolor[3] = { 0.5, 0.5, 0 };
	static GLfloat zemissioncolor[3] = { 0, 0.5, 0 };

	// Z axis
	glColor3fv(zcolor);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, zcolor);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, zemissioncolor);
	DrawTransformationHandle(quadric, Radius);

	// Y axis
	glColor3fv(ycolor);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, ycolor);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, yemissioncolor);
	glPushMatrix();
	glRotated(-90.0, 1.0, 0.0, 0.0);
	DrawTransformationHandle(quadric, Radius);
	glPopMatrix();

	// X axis
	glColor3fv(xcolor);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, xcolor);
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, xemissioncolor);
	glPushMatrix();
	glRotated(-90.0, 0.0, 1.0, 0.0);
	DrawTransformationHandle(quadric, Radius);
	glPopMatrix();

	gluDeleteQuadric(quadric);

	glPopAttrib();
}

void k3dTransformObject::DrawTransformationHandles(k3dILocationServer* LocationServer, gdouble MinimumRadius)
{
	// Sanity checks ...
	g_assert(LocationServer);

	// If we aren't the active object, we're done ...
	if(Document()->GetActiveObjectContext() != ObjectContext())
		return;

	if(0 == LocationServer->GetLocationCount())
		{
			DrawTransformationHandles(MinimumRadius);
			return;
		}

	sdpVector3 min;
	sdpVector3 max;
	k3dObjectDimensions(LocationServer, 0, min, max);

	min.Abs();
	max.Abs();

	sdpVector3 size(sdpMax(min[0], max[0]), sdpMax(min[1], max[1]), sdpMax(min[2], max[2]));

	DrawTransformationHandles(sdpMax(MinimumRadius, size.Length()));
}

