// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the k3dTriangleMesh class, which efficiently encapsulates an implicit/procedural surface composed of triangles
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "k3dtrianglemesh.h"

#include <sdpxmldocument.h>
#include <k3diactions.h>
#include <k3dieditorengine.h>
#include <k3dieditormaterial.h>
#include <k3dimaterial.h>
#include <k3dirimaterial.h>
#include <k3drishadersetup.h>
#include <k3didocument.h>
#include <k3dboundingbox.h>
#include <k3dmisc.h>

#include <algorithm>
#include <iostream>
#include <strstream>

//////////////////////////////////////////////////////////////////////////
// k3dTriangleMesh

// Static member variables
short k3dTriangleMesh::m_RiBilinearPatchLookup[4] = {0, 1, 3, 2};

sdpVector3 k3dTriangleMesh::m_SelectedColor = 1.0;
sdpVector3 k3dTriangleMesh::m_DefaultColor = 0.0;

k3dTriangleMesh::k3dTriangleMesh()
{
	AddVariable(&m_Color("color", m_DefaultColor));
	AddObject(&m_Material("material"));
}

k3dTriangleMesh::~k3dTriangleMesh()
{
}

void k3dTriangleMesh::ReserveTriangles(gulong TriangleCount)
{
	m_Triangles.resize(TriangleCount);
}

void k3dTriangleMesh::SetFlatShading()
{
	m_SmoothShading = false;
}

void k3dTriangleMesh::SetSmoothShading()
{
	m_SmoothShading = true;
}

void k3dTriangleMesh::CalculateTriangleNormals()
{
	sdpVector3 edge1, edge2;
	TriangleIterator triangleend = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
		{
			edge1 = Location(i->m_Points[1]).m_AnimXYZ - Location(i->m_Points[0]).m_AnimXYZ;
			edge2 = Location(i->m_Points[2]).m_AnimXYZ - Location(i->m_Points[0]).m_AnimXYZ;
			
			i->m_Normal = edge1 ^ edge2;
			i->m_Normal.Normalize();
		}
}

void k3dTriangleMesh::CalculateLocationNormals()
{
	// Start by calculating all of the triangle normals ...
	CalculateTriangleNormals();

	// Zero out each location's normal ...	
	LocationIterator locationend = m_Locations.end();
	for(LocationIterator l = m_Locations.begin(); l != locationend; l++)
		l->m_NormalXYZ = 0.0;
		
	// For each triangle ...
	TriangleIterator triangleend = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
		{
			// Add the triangle's normal to the location ...
			Location(i->m_Points[0]).m_NormalXYZ += i->m_Normal;
			Location(i->m_Points[1]).m_NormalXYZ += i->m_Normal;
			Location(i->m_Points[2]).m_NormalXYZ += i->m_Normal;
		}
}

////////////////////////////////////////////////////////
// k3dIEditorRender overrides ...

sdpResult k3dTriangleMesh::EditorRender(k3dIEditorEngine* Engine)
{
	// Sanity checks ...
	g_return_val_if_fail(Engine, SDP_E_FAIL);

	if(SDP_S_OK == Engine->EditorQuerySolid())
		{
			// Setup the material ...
			k3dIMaterial* material = m_Material.Interface();
			if(0 == material)
				material = Document()->GetDefaultMaterial();
		
			k3dIEditorMaterial* editormaterial = sdpQueryInterface<k3dIEditorMaterial>(material);
			if(editormaterial)
				{
					editormaterial->EditorMaterialSetup(Engine);
					editormaterial->Release();
		
					if(SDP_S_OK ==  Engine->EditorQueryTextured())
						EditorRenderSolidTextured(Engine);
					else
						EditorRenderSolid(Engine);
				}
		}
	else
		{
			EditorRenderWireframe(Engine);
		}
		
	if(k3dIDocument::POINT == Document()->GetActiveEditMode())
		EditorRenderPoints(Engine);

	DrawTransformationHandles(this, 1.0);

	return SDP_S_OK;
}

void k3dTriangleMesh::EditorRenderWireframe(k3dIEditorEngine* Engine)
{
	// Setup for wireframe polys ...
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_1D);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);

	// For each triangle ...
	TriangleIterator end = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != end; i++)
		{
			glBegin(GL_LINE_LOOP);
			glVertex3dv(Location(i->m_Points[0]).m_AnimXYZ);
			glVertex3dv(Location(i->m_Points[1]).m_AnimXYZ);
			glVertex3dv(Location(i->m_Points[2]).m_AnimXYZ);
			glEnd();
		}
}

void k3dTriangleMesh::EditorRenderSolid(k3dIEditorEngine* Engine)
{
	if(m_SmoothShading)
		{
			CalculateLocationNormals();
			EditorRenderSolidSmooth(Engine);
		}
	else
		{
			CalculateTriangleNormals();
			EditorRenderSolidFlat(Engine);
		}
}

void k3dTriangleMesh::EditorRenderSolidFlat(k3dIEditorEngine* Engine)
{
	// Triangles ...
	glBegin(GL_TRIANGLES);
	TriangleIterator triangleend = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
		{
			glNormal3dv(i->m_Normal);
		
			glVertex3dv(Location(i->m_Points[0]).m_AnimXYZ);
			glVertex3dv(Location(i->m_Points[1]).m_AnimXYZ);
			glVertex3dv(Location(i->m_Points[2]).m_AnimXYZ);
		}
	glEnd();
}

void k3dTriangleMesh::EditorRenderSolidSmooth(k3dIEditorEngine* Engine)
{
	// Triangles ...
	glBegin(GL_TRIANGLES);
	TriangleIterator triangleend = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
		{
			glNormal3dv(Location(i->m_Points[0]).m_NormalXYZ);
			glVertex3dv(Location(i->m_Points[0]).m_AnimXYZ);
			glNormal3dv(Location(i->m_Points[1]).m_NormalXYZ);
			glVertex3dv(Location(i->m_Points[1]).m_AnimXYZ);
			glNormal3dv(Location(i->m_Points[2]).m_NormalXYZ);
			glVertex3dv(Location(i->m_Points[2]).m_AnimXYZ);
		}
	glEnd();
}

void k3dTriangleMesh::EditorRenderSolidTextured(k3dIEditorEngine* Engine)
{
	if(m_SmoothShading)
		{
			CalculateLocationNormals();
			EditorRenderSolidTexturedSmooth(Engine);
		}
	else
		{
			CalculateTriangleNormals();
			EditorRenderSolidTexturedFlat(Engine);
		}
}

void k3dTriangleMesh::EditorRenderSolidTexturedFlat(k3dIEditorEngine* Engine)
{
	glBegin(GL_TRIANGLES);

	TriangleIterator end = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != end; i++)
		{
			glNormal3dv(i->m_Normal);
		
			glTexCoord3dv(Location(i->m_Points[0]).m_TextureXYZ);
			glVertex3dv(Location(i->m_Points[0]).m_AnimXYZ);
			glTexCoord3dv(Location(i->m_Points[1]).m_TextureXYZ);
			glVertex3dv(Location(i->m_Points[1]).m_AnimXYZ);
			glTexCoord3dv(Location(i->m_Points[2]).m_TextureXYZ);
			glVertex3dv(Location(i->m_Points[2]).m_AnimXYZ);
		}
		
	glEnd();
}

void k3dTriangleMesh::EditorRenderSolidTexturedSmooth(k3dIEditorEngine* Engine)
{
	glBegin(GL_TRIANGLES);

	TriangleIterator end = m_Triangles.end();
	for(TriangleIterator i = m_Triangles.begin(); i != end; i++)
		{
			glNormal3dv(Location(i->m_Points[0]).m_NormalXYZ);
			glTexCoord3dv(Location(i->m_Points[0]).m_TextureXYZ);
			glVertex3dv(Location(i->m_Points[0]).m_AnimXYZ);
			glNormal3dv(Location(i->m_Points[1]).m_NormalXYZ);
			glTexCoord3dv(Location(i->m_Points[1]).m_TextureXYZ);
			glVertex3dv(Location(i->m_Points[1]).m_AnimXYZ);
			glNormal3dv(Location(i->m_Points[2]).m_NormalXYZ);
			glTexCoord3dv(Location(i->m_Points[2]).m_TextureXYZ);
			glVertex3dv(Location(i->m_Points[2]).m_AnimXYZ);
		}
		
	glEnd();
}

void k3dTriangleMesh::EditorRenderPoints(k3dIEditorEngine* Engine)
{
	glBegin(GL_POINTS);
	
	LocationIterator end = m_Locations.end();
	for(LocationIterator i = m_Locations.begin(); i != end; i++)
		glVertex3dv(i->m_AnimXYZ);
		
	glEnd();
}

///////////////////////////////////////////////////////////////
// k3dIMaterialServer overrides ...

k3dIMaterial* k3dTriangleMesh::GetMaterial()
{
	return m_Material.Interface();
}

sdpResult k3dTriangleMesh::SetMaterial(k3dIMaterial* Material)
{
std::cout << "a" << std::endl;

	// We allow NULL here, as a synonym for the default material ...
	if(Material)
		m_Material.AttachObject(Material);
	else
		m_Material.ReleaseObject();

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIRiRender overrides

sdpResult k3dTriangleMesh::RiRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// Setup our material interfaces ...
	k3dIMaterial* material = m_Material.Interface();
	if(0 == material)
		material = Document()->GetDefaultMaterial();

	k3dIRiMaterial* rimaterial = sdpQueryInterface<k3dIRiMaterial>(material);
	if(rimaterial)
		{
			// Setup the material ...
			rimaterial->RiMaterialSetup(Frame, Engine);
		
			// Render each triangle ...
			TriangleIterator triangleend = m_Triangles.end();
			for(TriangleIterator i = m_Triangles.begin(); i != triangleend; i++)
				RiTriangle(Engine, *i);
		}

	return SDP_S_OK;
}

// Render a triangle as a bilinear patch ...
void k3dTriangleMesh::RiTriangle(k3dIRiEngine* Engine, k3dTriangle& Triangle)
{
	// Arrays of coordinates ...
	sdpVector3* points = new sdpVector3[4];
	sdpVector3* normals = new sdpVector3[4];
	gdouble* scoords = new gdouble[4];
	gdouble* tcoords = new gdouble[4];

	for(gulong i = 0; i < 3; i++)
		{
			points[i] = Location(Triangle.m_Points[i]).m_AnimXYZ;
			normals[i] = m_SmoothShading ? Location(Triangle.m_Points[i]).m_NormalXYZ : Triangle.m_Normal;
//			scoords[i] = polypoint->m_TextureXYZ[0];
//			tcoords[i] = polypoint->m_TextureXYZ[1];
		}

	points[3] = Location(Triangle.m_Points[1]).m_AnimXYZ;
	normals[3] = m_SmoothShading ? Location(Triangle.m_Points[1]).m_NormalXYZ : Triangle.m_Normal;
//	scoords[3] = middlepoint->m_TextureXYZ[0];
//	tcoords[3] = middlepoint->m_TextureXYZ[1];

	k3dRiParameterList params;
	params.AddPointer(k3dIRiEngine::RI_P, gpointer(points));
	params.AddPointer(k3dIRiEngine::RI_N, gpointer(normals));
//	params.AddPointer(k3dIRiEngine::RI_S, gpointer(scoords));
//	params.AddPointer(k3dIRiEngine::RI_T, gpointer(tcoords));

	Engine->RiPatchV(k3dIRiEngine::RI_BILINEAR, params.ParameterCount(), params.Tokens(), params.Parameters());
	
	delete[] points;
	delete[] normals;
	delete[] scoords;
	delete[] tcoords;
}

