// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the renderjob application, which runs jobs scheduled with the virtual render farm
		\author Tim Shead (tshead@k-3d.com)
*/

#include <k3direnderjob.h>
#include <k3direnderframe.h>
#include <k3ddocument.h>

#include <sdpxmldocument.h>
#include <sdpxmlutility.h>

#include <sdpfile.h>
#include <sdputility.h>

#include <iostream>
#include <fstream>

#include <stdio.h>

#ifdef SDPUNIX
#include <unistd.h>
#endif // SDPUNIX

#include <time.h>

sdpString Now()
{
	// Get the current time ...
	time_t now_t;
	time(&now_t);
	
	tm* now_tm = localtime(&now_t);
	
	// Format it ...
	sdpString result;
	strftime(result.buffer(64), 64, "%m/%d/%Y %H:%M:%S", now_tm);

	return result;
}

class RenderJobRenderFrame
{
public:
	RenderJobRenderFrame(sdpXmlDocument& GlobalOptions) : m_GlobalOptions(GlobalOptions)
	{
	}

	bool HandleDirectory(SDPCSTRING DirectoryPath, SDPCSTRING Directory)
	{
		// Load the control file for the frame ...
		const sdpString controlpath = sdpAppendPath(DirectoryPath, k3dIRenderFrame::ControlFile);
		k3dDocument framecontrol;
		if(!framecontrol.LoadFile(controlpath))
			{
				std::cerr << "Error loading " << controlpath << " configuration file" << std::endl;
				return true;
			}

		// Get the frame ...
		sdpXmlIElement* frame = sdpXmlGetXPathElement(framecontrol.GetRootElement(), "frame");
		if(!frame)
			{
				std::cerr << controlpath << " does not contain frame data" << std::endl;
				return true;
			}
		
		// Load frame data ...
		bool viewafterrender = false;
		sdpXmlGetAttribute(frame, "viewafterrender", viewafterrender);
		
		// Render the frame ...
		system(sdpFormatString("renderframe %s", DirectoryPath));
		
		// If the user wants to see each frame after it's rendered ...
		if(viewafterrender)
			{
				// Figure out their preferred bitmap viewer ...
				sdpXmlElementArray viewers = sdpXmlXPathSelect(m_GlobalOptions.GetRootElement(), "application/helpers/command[@type=\"bitmapviewer\"]");
				if(viewers.size())
					{
						// View the image ...
						sdpString viewcommand = viewers[0]->Text();
						viewcommand.formattedreplace('%', "p", sdpAppendPath(DirectoryPath, k3dIRenderFrame::OutputImageFile));
						viewcommand += " &";
						
						system(viewcommand);
					}
				else
					{
						std::cerr << "No bitmap viewer defined for viewing " << DirectoryPath << std::endl;
						return true;
					}
			}
		
		return true;
	}
	
	sdpXmlDocument& m_GlobalOptions;
};

bool RenderJob(sdpXmlDocument& GlobalOptions, SDPCSTRING JobDirectory)
{
	// Sanity checks ...
	g_assert_string(JobDirectory);

	// Setup some standard paths ...
	const sdpString readypath = sdpAppendPath(JobDirectory, k3dIRenderJob::StatusReadyFile);
	const sdpString runningpath = sdpAppendPath(JobDirectory, k3dIRenderJob::StatusRunningFile);
	const sdpString errorpath = sdpAppendPath(JobDirectory, k3dIRenderJob::StatusErrorFile);
	const sdpString completepath = sdpAppendPath(JobDirectory, k3dIRenderJob::StatusCompleteFile);

	// Make sure the job exists ...
	if(!sdpFileExists(JobDirectory))
		{
			std::cerr << Now() << " - " << JobDirectory << " not found" << std::endl;
			return false;
		}

	// Skip the job if it's complete ...
	std::ifstream completefile(completepath, std::ios::in);
	if(completefile.good())
		return true;

	// Skip the job if it errored out ...
	std::ifstream errorfile(errorpath, std::ios::in);
	if(errorfile.good())
		return true;

	// Skip the job if it's running ...
	std::ifstream runningfile(runningpath, std::ios::in);
	if(runningfile.good())
		return true;

	// Make sure the job is ready ...
	std::ifstream readyfile(sdpAppendPath(JobDirectory, "ready"), std::ios::in);
	if(!readyfile.good())
		{
			std::cerr << "renderjob: Job " << JobDirectory << " is not ready" << std::endl;
			return false;
		}

	// Standard logging ...
	std::cout << Now() << " - Running " << JobDirectory << std::endl;

	// Switch the job status to running ...
	rename(readypath, runningpath);
	
	// For each directory in the job directory (non-recursive) ...
	RenderJobRenderFrame functor(GlobalOptions);
	sdpForEachDirectory(JobDirectory, functor, false);

	// Switch the job status to complete ...
	rename(runningpath, completepath);
	
	// Standard logging ...
	std::cout << Now() << " - Complete" << std::endl;
	
	return true;
}

/// Prints usage info
void Usage()
{
	std::cout << "usage: renderjob [options] [directory ...]" << std::endl;
	std::cout << std::endl;
	std::cout << "  -h, --help               prints this help information and exits" << std::endl;
	std::cout << "      --optionspath [path] overrides the path for storing user options" << std::endl;
	std::cout << "      --version            prints program version information and exits" << std::endl;
	std::cout << std::endl;
}

/// Prints version info
void Version()
{
#include "../../k3d_version.h"

	std::cout << "K-3D RenderJob Version " << VERSION_STRING << std::endl;
	std::cout << "Copyright (c) 1995-2000, Timothy M. Shead.  See the AUTHORS file for contributors." << std::endl;
	std::cout << "Licensed by the GNU General Public License.  See the COPYING file for details." << std::endl;
	std::cout << "K-3D Home Page: http://www.k-3d.com" << std::endl;
	std::cout << "Bug reports to tshead@k-3d.com" << std::endl;
}

/// Program main
int main(int argc, char* argv[])
{
	sdpString optionspath;

	// Convert our arguments into an array of strings ...
	typedef std::list<sdpString> ArgumentList;
	typedef ArgumentList::iterator ArgumentIterator;

	ArgumentList arguments;
	for(int i = 1; i < argc; i++)
		arguments.push_back(argv[i]);

	// Test our arguments ...
	if(0 == arguments.size())
		{
			Usage();
			return 1;
		}

	for(ArgumentIterator argument = arguments.begin(); argument != arguments.end(); )
		{
			if(*argument == "--help" || *argument == "-h")
				{
					Usage();
					return 0;
				}
			else if(*argument == "--version")
				{
					Version();
					return 0;
				}
			else if(*argument == "--optionspath")
				{
					ArgumentIterator name = argument++;
					ArgumentIterator value = argument++;
					optionspath = *(value);
					arguments.erase(name);
					arguments.erase(value);
				}
			else
				{
					argument++;
				}
		}

	// Load our global options file ...
	k3dDocument options;
	if(!options.LoadOptions(optionspath))
		{
			std::cerr << "*** renderjob: error loading configuration file " << options.OptionsFilePath(optionspath) << std::endl;
			return 1;
		}

	// Each remaining argument should be a frame path to render ...						
	int result = 0;
	for(ArgumentIterator job = arguments.begin(); job != arguments.end(); job++)
		{
			if(!RenderJob(options, *job))
				result = 1;
		}

	return result;
}


