// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CImageLayer K-3D object, which renders 2D graphics either in front of or behind all 3D geometry.
		\author Tim Shead (tshead@k-3d.com)
*/

#include "imagelayer.h"

#include <k3dieditorengine.h>
#include <k3dieditortexture.h>
#include <k3ditexture.h>
#include <k3diriengine.h>
#include <k3diritexture.h>
#include <k3dobjectselector.h>
#include <k3dcom.h>

#include <k3dmisc.h>
#include <sdpfile.h>

#include <sdpgl.h>

#include <sdpgtkevent.h>

double CImageLayer::m_ForegroundCount = 0.0;
double CImageLayer::m_BackgroundCount = 0.0;

CImageLayer::CImageLayer()
{
	AddVariable(&m_Editor("editor", true));
	AddVariable(&m_Layer("layer", BACKGROUND));

	AddObject(&m_Color("color"));
	AddObject(&m_Strength("strength"));
}

CImageLayer::~CImageLayer()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CImageLayer::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIRiRender>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CImageLayer::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CImageLayer::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CImageLayer::SetObjectContext(k3dIObjectContext* Context)
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::SetObjectContext(Context)), SDP_E_FAIL);
	
	// Initialize our shaders ...
	g_return_val_if_fail(m_Shader.Create(Document()), SDP_E_FAIL);
	
	return SDP_S_OK;
}

sdpResult CImageLayer::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CImageLayer::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CImageLayer::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

////////////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CImageLayer::EditorPreRender(k3dIEditorEngine* Engine)
{
	if(BACKGROUND == m_Layer())	
		return DrawEditor(Engine);

	return SDP_S_FALSE;
}

sdpResult CImageLayer::EditorPostRender(k3dIEditorEngine* Engine)
{
	if(FOREGROUND == m_Layer())
		return DrawEditor(Engine);

	return SDP_S_FALSE;
}

sdpResult CImageLayer::DrawEditor(k3dIEditorEngine* Engine)
{
	// If we don't have a texture assigned, we're done ...
	k3dITexture* texture = m_Shader.GetPrimaryTexture();
	if(0 == texture)
		return SDP_S_FALSE;
		
	k3dIEditorTexture* editortexture = sdpQueryInterface<k3dIEditorTexture>(texture);
	g_return_val_if_fail(editortexture, SDP_E_FAIL);
	
	// Store the GL state - we have to do it ourselves because we're NOT being called between EditorPushRender() and EditorPopRender()
	glPushAttrib(GL_ALL_ATTRIB_BITS);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluOrtho2D(-1.0, 1.0, -1.0, 1.0);

	// Get our color and strength ...
	sdpVector3 color;
	m_Color->GetValue3(m_CurrentTime, m_DeltaTime, color);
	gdouble strength = 0.25 * m_Strength->GetValue1(m_CurrentTime, m_DeltaTime);

	// Set ourselves up to draw an unlit, textured polygon with blending ...
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glDepthMask(GL_FALSE);
	glDisable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4d(color[0], color[1], color[2], strength);

	editortexture->EditorTextureSetup(Engine);

	glBegin(GL_POLYGON);

	glTexCoord2d(0.0, 0.0);
	glVertex3d(-1.0, 1.0, 1.0);
	
	glTexCoord2d(1.0, 0.0);
	glVertex3d(1.0, 1.0, 1.0);
	
	glTexCoord2d(1.0, 1.0);
	glVertex3d(1.0, -1.0, 1.0);
	
	glTexCoord2d(0.0, 1.0);
	glVertex3d(-1.0, -1.0, 1.0);

	glEnd();

	// Restore the GL state - we have to do it ourselves because we're NOT being called between EditorPushRender() and EditorPopRender()
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glPopAttrib();

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////////////
// k3dIRiRender overrides

sdpResult CImageLayer::RiPreRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	m_ForegroundCount = 0.0;
	m_BackgroundCount = 0.0;

	return SDP_S_OK;
}

sdpResult CImageLayer::RiRender(k3dIRenderFrame* Frame, k3dIRiEngine* Engine)
{
	// If the user hasn't selected a shader, we're done ...
	if(0 == m_Shader.ShaderHandle())
		return SDP_S_FALSE;
		
	// Get data from the camera ...
	sdpVector3 cameracoords;
	gdouble left, right, top, bottom, nearplane, farplane;
	Engine->Camera()->GetViewingFrustum(&left, &right, &top, &bottom, &nearplane, &farplane);
	Engine->Camera()->GetWorldPosition(cameracoords);

	// Reset the current transformation ...
	Engine->RiIdentity();
	Engine->RiTranslate(cameracoords[0], cameracoords[1], cameracoords[2]);
	Engine->RiRotate(90.0f, 1.0f, 0.0f, 0.0f);

	// Disable shadows ...
	k3dRiParameterList parameters;
	parameters.AddString("casts_shadows", "none");
	parameters.SetAttributes("render", Engine);

	// Set base color ...
	sdpVector3 color;
	m_Color->GetValue3(m_CurrentTime, m_DeltaTime, color);
	Engine->RiColor(color);

	// Set opacity ...
	sdpVector3 strength = m_Strength->GetValue1(m_CurrentTime, m_DeltaTime);
	Engine->RiOpacity(strength);

	// Setup the surface shader ...
	k3dRiSurfaceShader shader;
	m_Shader.Initialize(Frame, Engine, shader, m_CurrentTime, m_DeltaTime);
	shader.Setup(Engine);

	// Draw a sphere
	gdouble radius;
	if(BACKGROUND == m_Layer())
		{
			radius = (0.99 * farplane) - m_BackgroundCount;
			m_BackgroundCount += 1.0;
		}
	else
		{
			sdpVector3 fradius(right, top, nearplane);
			radius = (2.0 * fradius.Length()) + m_ForegroundCount;
			m_ForegroundCount += 0.01;
		}

	Engine->RiSphereV(radius, -radius, radius, 360.0, 0, 0, 0);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CImageLayer::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(sdpSucceeded(m_Strength.Create(Document(), "ImageLayer Strength", 1.0, true, 0.0, 1.0, "Time", "Strength", k3dIChannel::DATA_SCALAR, true, this)), SDP_E_FAIL);
	g_return_val_if_fail(sdpSucceeded(m_Color.Create(Document(), "ImageLayer Color", sdpVector3(1.0), true, "Time", "Color", true, this)), SDP_E_FAIL);
//	g_return_val_if_fail(sdpSucceeded(m_Texture.CreateObjectSelector(Document(), "ImageLayer texture", IID_k3dITexture, true, true, true, true, this, m_ObjectContext->ObjectID())), SDP_E_FAIL);
//	g_return_val_if_fail(sdpSucceeded(m_Shader.CreateShaderSelector(Document(), "ImageLayer shader", "imagelayer", k3dIShaderInfo::SHADER_SURFACE, sdpImageLayer, false, true, this, m_ObjectContext->ObjectID())), SDP_E_FAIL);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CImageLayerDialog

CImageLayerDialog::CImageLayerDialog()
{
	m_Object = 0;
}

bool CImageLayerDialog::Create(CImageLayer* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "imagelayer.gtkml", "imagelayer.html"), false);

	return true;
}

bool CImageLayerDialog::InitializeControls()
{
	// Setup channel buttons ...
	if(ChannelButton("color"))
		ChannelButton("color")->Connect(m_Object->m_Color.Interface());

	if(ChannelButton("strength"))
		ChannelButton("strength")->Connect(m_Object->m_Strength.Interface());

	// Setup shader selector ...
	m_ShaderSelector.Create(m_Object->Application(), &m_Object->m_Shader, &m_ShaderPage, false, k3dIShaderInfo::SURFACE, Container("shaderselector"));
	// Setup the shader arguments ...
	m_ShaderPage.Create(m_Object->Document(), &m_Object->m_Shader, Container("shadercontainer"));

	return k3dDialog::InitializeControls();
}

void CImageLayerDialog::UpdateControls()
{
	// Setup editor options ...
	ToggleButton("editor").SetState(m_Object->m_Editor());
	k3dDialog::UpdateControls();
}

void CImageLayerDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "oneditor")
		OnEditor();
	else
		k3dDialog::OnEvent(Event);
}

void CImageLayerDialog::OnEditor()
{
	m_Object->m_Editor() = ToggleButton("editor").GetState();
	m_Object->OnModify(false);
}

