/*$Id: kaleido.h,v 1.4 2000/08/21 02:01:25 tshead Exp $*/
/**********************************************************
* kaleido
*
*	Kaleidoscopic construction of uniform polyhedra
*	Copyright  1991-1997 Dr. Zvi Har'El <rl@math.technion.ac.il>
*
*	This program is free software; you can redistribute it and/or modify
*	it under the terms of the GNU General Public License as published by
*	the Free Software Foundation; either version 2 of the License, or
*	(at your option) any later version.
*
*	This program is distributed in the hope that it will be useful,
*	but WITHOUT ANY WARRANTY; without even the implied warranty of
*	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*	GNU General Public License for more details.
*
*	You should have received a copy of the GNU General Public License
*	along with this program; if not, write to the Free Software
*	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
*	Author:
*		Dr. Zvi Har'El,
*		Technion, Israel Institue of Technology,
*		Haifa 32000, Israel.
*		E-Mail: rl@math.technion.ac.il
**********************************************************/

/** \file
		\brief Declares functions for  generating uniform polyhedra for CPrimitives
		\author Dr. Zvi Har'El <rl@math.technion.ac.il>
*/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>


#ifndef MAXLONG
#define MAXLONG 0x7FFFFFFF
#endif

#ifndef MAXDIGITS
#define MAXDIGITS 10 /* (int)log10((double)MAXLONG) + 1 */
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifndef DBL_EPSILON
#define DBL_EPSILON 2.2204460492503131e-16
#endif

#define BIG_EPSILON 3e-2
#define DEG (180/M_PI)
#define AZ M_PI/7 /* axis azimuth */
#define EL M_PI/17 /* axis elevation */

#define Err(x){error.message=x;error.line=__LINE__;error.file=__FILE__;return 0;}
#define Free(lvalue){if(lvalue){free((char*)lvalue); lvalue=0;}}
#define Matfree(lvalue,n){if(lvalue)matfree((char**)lvalue,n);lvalue=0;}
#define Malloc(lvalue,n,type){if(!(lvalue=(type*)malloc((n)*sizeof(type))))Err(0)}
#define Realloc(lvalue,n,type){if(!(lvalue=(type*)realloc(lvalue,(n)*sizeof(type))))Err(0)}
#define Calloc(lvalue,n,type){if(!(lvalue=(type*)calloc(n,sizeof(type))))Err(0)}
#define Matalloc(lvalue,n,m,type){if(!(lvalue=(type**)matalloc(n,(m)*sizeof(type))))Err(0)}
#define Sprintfrac(lvalue,x){if(!(lvalue=sprintfrac(x)))return 0;}
#define numerator(x)(frac(x),frax.n)
#define denominator(x)(frac(x),frax.d)
#define compl(x)(frac(x),(double)frax.n/(frax.n-frax.d))

typedef struct {double x,y,z;} Kvector;

typedef struct {
	/* NOTE: some of the int's can be replaced by short's, char's,
		or even bit fields, at the expense of readability!!!*/	
	int index; /* index to the standard list, the array uniform[] */
	int N; /* number of faces types (atmost 5)*/
	int M; /* vertex valency  (may be big for dihedral polyhedra) */
	int V; /* vertex count */
	int E; /* edge count */
	int F; /* face count */
	int D; /* density */
	int chi; /* Euler characteristic */
	int g; /* order of symmetry group */
	int K; /* symmetry type: D=2, T=3, O=4, I=5 */
	int hemi;/* flag hemi polyhedron */
	int onesided;/* flag onesided polyhedron */
	int even; /* removed face in pqr| */
	int *Fi; /* face counts by type (array N)*/
	int *rot; /* vertex configuration (array M of 0..N-1) */
	int *snub; /* snub triangle configuration (array M of 0..1) */
	int *firstrot; /* temporary for vertex generation (array V) */
	int *anti; /* temporary for direction of ideal vertices (array E) */
	int *ftype; /* face types (array F) */
	int **e; /* edges (matrix 2 x E of 0..V-1)*/
	int **dual_e; /* dual edges (matrix 2 x E of 0..F-1)*/
	int **incid; /* vertex-face incidence (matrix M x V of 0..F-1)*/
	int **adj; /* vertex-vertex adjacency (matrix M x V of 0..V-1)*/
	double p[4]; /* p, q and r; |=0 */
	double minr; /* smallest nonzero inradius */
	double gon; /* basis type for dihedral polyhedra */
	double *n; /* number of side of a face of each type (array N) */
	double *m; /* number of faces at a vertex of each type (array N) */
	double *gamma; /* fundamental angles in radians (array N) */
	char *polyform; /* printable Wythoff symbol */
	char *config; /* printable vertex configuration */
	char *name; /* name, standard or manifuctured */
	char *dual_name; /* dual name, standard or manifuctured */
	Kvector *v; /* vertex coordinates (array V) */
	Kvector *f; /* face coordinates (array F)*/
} Polyhedron;

typedef struct { /* See uniform.h for explanation of the fields */
	char *Wythoff, *name, *dual;
	short Coxeter, Wenninger;
} Uniform;

typedef struct {long n,d;} Fraction;

typedef struct {char *message;int line;char *file;} Error;

Polyhedron*  kaleido( char *sym, int need_coordinates,
        int need_edgelist, int need_approx, int just_list);
Polyhedron*  polyalloc();

Kvector  rotate( Kvector vertex, Kvector axis, double angle );
Kvector  sum3( Kvector a, Kvector b, Kvector c );
Kvector  scale( double k, Kvector a );
Kvector  sum( Kvector a, Kvector b );
Kvector  diff( Kvector a, Kvector b );
Kvector  pole( double r, Kvector a, Kvector b, Kvector c );
Kvector  cross( Kvector a, Kvector b );

double  dot( Kvector a, Kvector b );

char*  sprintfrac( double x );
char*  nextsym( int input, int index );
char**  matalloc( int rows, int row_size );
void  frac( double x );


extern Uniform uniform[];
extern Fraction frax;
extern Error error;
extern int last_uniform;


#ifndef __STDC__
extern char *malloc(), *realloc(), *getenv();
#endif


int  mod( int i, int j );
double  same( Kvector a, Kvector b, double epsilon );
int  dihedral( Polyhedron *P, char *name, char *dual_name );
void  matfree( char **mat, int rows );
int  edgelist( Polyhedron *P );
int  faces( Polyhedron *P );
int  vertices( Polyhedron *P );
int  configuration( Polyhedron *P );
int  count( Polyhedron *P );
int  exceptions( Polyhedron *P);
int  newton( Polyhedron *P, int need_approx );
int  guessname( Polyhedron *P );
int  decompose( Polyhedron *P );
int  moebius( Polyhedron *P );
int  unpacksym( char *sym, Polyhedron *P );

void  polyfree( Polyhedron *P );
void  rotframe( double azimuth, double elevation, double angle );
void  rotarray( Kvector *newobj, Kvector *old, int n );
