#ifndef PRIMITIVES_H
#define PRIMITIVES_H

// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Declares the CPrimitives K-3D tool, for creating simple geometric primitives
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <k3dprimtool.h>
#include <k3dprimtooldialog.h>

#include "kaleido.h"

// Forward declarations
class CPrimitives;


/////////////////////////////////////////////////////////////////////////////
// CPrimitivesDialog

class CPrimitivesDialog : public k3dPrimToolDialog
{
friend CPrimitives;
public:
	CPrimitivesDialog();

	bool Create(CPrimitives* Object);

protected:
	CPrimitives* m_Object;
	gint m_CurrentPage;

	// Event handling ...
	void OnEvent(sdpGtkEvent* Event);

	void OnPageSwitch(sdpGtkEvent* Event);

	void OnDiskType();
	void OnDiskRadius();
	void OnDiskSegments();
	void OnDiskPower();
	void OnConeLength();
	void OnConeRadius();
	void OnConeLengthPower();
	void OnConeRadialPower();
	void OnConeLengthSegments();
	void OnConeRadialSegments();
	void OnConeCreateBase();
	void OnCylinderLength();
	void OnCylinderRadius();
	void OnCylinderRadialPower();
	void OnCylinderLengthSegments();
	void OnCylinderRadialSegments();
	void OnCylinderCreateBase();
	void OnCubeXLength();
	void OnCubeYLength();
	void OnCubeZLength();
	void OnCubeXSegments();
	void OnCubeYSegments();
	void OnCubeZSegments();
	void OnGridXLength();
	void OnGridYLength();
	void OnGridXSegments();
	void OnGridYSegments();
	void OnSphereXLength();
	void OnSphereYLength();
	void OnSphereZLength();
	void OnSphereLengthPower();
	void OnSphereRadialPower();
	void OnSphereLengthSegments();
	void OnSphereRadialSegments();
	void OnTorusInnerRadius();
	void OnTorusOuterRadius();
	void OnTorusRadialPower();
	void OnTorusCrossSectionPower();
	void OnTorusRadialSegments();
	void OnTorusCrossSectionSegments();
	void OnPolyhedronLength();
	void OnPolyhedronChoice(sdpGtkEvent* Event);

	// Update controls ...
	bool InitializeControls();
	void UpdateControls();

	void UpdateDiskControls();
	void UpdateConeControls();
	void UpdateCylinderControls();
	void UpdateCubeControls();
	void UpdateGridControls();
	void UpdateSphereControls();
	void UpdateTorusControls();
	void UpdatePolyhedronControls();
};

/////////////////////////////////////////////////////////////////////////////
// CPrimitives

class CPrimitives : public k3dPrimTool
{
friend CPrimitivesDialog;
public:
	CPrimitives();

	CPrimitivesDialog m_UserInterface;

	// Where it all happens ...
	void OnModify(bool SaveAction);

	void OnModifyDisk(bool SaveAction);
	void OnModifyCone(bool SaveAction);
	void OnModifyCylinder(bool SaveAction);
	void OnModifyCube(bool SaveAction);
	void OnModifyGrid(bool SaveAction);
	void OnModifySphere(bool SaveAction);
	void OnModifyTorus(bool SaveAction);
	void OnModifyPolyhedron(bool SaveAction);

	// k3dIUnknown overrides
	sdpResult QueryInterface(const sdpIID& InterfaceID, void** Interface);
	gulong AddRef();
	gulong Release();

	// k3dIObject overrides
	sdpResult EditObject();
	sdpResult ObjectDeleted();

	// Disk data ...
	typedef enum
	{
		DISK,
		TRIANGLES
	} DiskType;

	DiskType m_DiskType;
	double m_DiskRadius;
	gulong m_DiskSegments;
	double m_DiskPower;

	// Cone data ...
	double m_ConeLength;
	double m_ConeRadius;
	double m_ConeLengthPower;
	double m_ConeRadialPower;
	gulong m_ConeLengthSegments;
	gulong m_ConeRadialSegments;
	bool m_ConeCreateBase;

	// Cylinder data ...
	double m_CylinderLength;
	double m_CylinderRadius;
	double m_CylinderRadialPower;
	gulong m_CylinderLengthSegments;
	gulong m_CylinderRadialSegments;
	bool m_CylinderCreateBase;

	// Cube data ...
	double m_CubeXLength;
	double m_CubeYLength;
	double m_CubeZLength;
	gulong m_CubeXSegments;
	gulong m_CubeYSegments;
	gulong m_CubeZSegments;

	// Grid data ...
	double m_GridXLength;
	double m_GridYLength;
	gulong m_GridXSegments;
	gulong m_GridYSegments;

	// Sphere data ...
	double m_SphereXLength;
	double m_SphereYLength;
	double m_SphereZLength;
	double m_SphereLengthPower;
	double m_SphereRadialPower;
	gulong m_SphereLengthSegments;
	gulong m_SphereRadialSegments;

	// Torus data ...
	double m_TorusInnerRadius;
	double m_TorusOuterRadius;
	double m_TorusRadialPower;
	double m_TorusCrossSectionPower;
	gulong m_TorusRadialSegments;
	gulong m_TorusCrossSectionSegments;

	gulong m_PolyhedronNumber;
	bool m_PolyhedronDualCheck;
	double m_PolyhedronLength;

protected:
	// Special functions for generating disk coordinates ...
	void DiskCoords(gulong Index, sdpVector3& Results);
	void DiskTextureCoords(gulong Index, sdpVector3& Results);

	// Our mapping of cubical coordinates to points, used to manufacture cubes ...
	typedef std::map<sdpString, k3dIPoint*> CubeMap;
	typedef CubeMap::iterator CubeMapIterator;
	CubeMap m_CubeMap;

	// Special methods used to generate cubes ...
	gulong CubePointCount();
	gulong CubePolyCount();

	sdpString CubePointString(gulong X, gulong Y, gulong Z);
	void SetCubePoint(gulong X, gulong Y, gulong Z, k3dIPoint* Point);
	k3dIPoint* GetCubePoint(gulong X, gulong Y, gulong Z);
	void AddCubePoint(gulong PolyIndex, gulong X, gulong Y, gulong Z, gdouble S, gdouble T);
	
	void CreateCubeGeometry(bool SaveAction);

	// Special variables and methods to generate polyhedra ...
	typedef std::vector<sdpVector3> PolyhedronPoints;
	PolyhedronPoints m_PolyhedronPoints;

	sdpString m_PolyhedronProperties;
	double m_PolyhedronScaling;

	typedef std::vector<int> PolyhedronPolygon;
	typedef std::vector<PolyhedronPolygon> PolyhedronPolygons;
	PolyhedronPolygons m_PolyhedronPolygons;

	void LoadPolyhedron(int PolyhedronIndex, bool Dual);
	void PolyhedronPoint(Kvector v);
	void PolyhedronFace();
	void PolyhedronPolyPoint(int PointIndex);
	void PolyhedronTriangle(int PointIndex1, int PointIndex2, int PointIndex3);
	void PolyhedronQuad(int PointIndex1, int PointIndex2, int PointIndex3, int PointIndex4);

	int vrmodel( Polyhedron *P, Kvector *v, int V, Kvector *f, int F,	char *name, char *star, double azimuth, double elevation, double freeze);
};


#endif // PRIMITIVES_H
