// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements CExtrude, a K-3D tool for creating ducted solids from existing geometry
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "extrude.h"

#include <k3diactions.h>
#include <k3dipolypointserver.h>
#include <k3dicamera.h>
#include <k3dmisc.h>
#include <k3dprimtool.h>
#include <k3dprimtooldialog.h>

#include <sdpgl.h>

#include <sdpfile.h>
#include <sdpgtkevents.h>

#include <algorithm>
#include <iterator>
#include <iostream>

///////////////////////////////////////////////////////////////
// CExtrudeEdge

CExtrudeEdge::CExtrudeEdge(k3dIPoint* Point1, k3dIPoint* Point2)
{
	// Sanity checks ...
	g_assert(Point1);
	g_assert(Point2);
	
	m_Point1 = Point1;
	m_Point2 = Point2;
	
	if(Point1 > Point2)
		{
			std::swap(m_Point1, m_Point2);
			m_Flipped = true;
		}
	else
		{
			m_Flipped = false;
		}
		
	m_ReferenceCount = 1;
}

///////////////////////////////////////////////////////////////
// CExtrude

CExtrude::CExtrude()
{
	m_Segments = 5;

	m_ConnectExterior = true;
	m_ConnectInterior = false;
	m_CreateLayers = false;
	m_CreateEndCaps = true;

	m_Sensitivity = 0.1;

	m_Offset = sdpVector3(0.0, 0.0, 10.0);
	m_Scale = 1.0;
	m_Rotation = 0.0;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CExtrude::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CExtrude::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CExtrude::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CExtrude::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CExtrude::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPrimTool::ObjectDeleted();
}

class CExtrudeCreateEdgeList
{
public:
	CExtrudeCreateEdgeList(ExtrudeEdgeArray& Edges) : m_Edges(Edges) {}
	
	bool HandleMesh(k3dIMesh* Mesh)
	{
		// For each polygon in the mesh ...
		k3dIPolygonServer* polygons = Mesh->Polygons();
		for(K3DPOLYGONHANDLE polyhandle = polygons->GetFirstPolygon(); polyhandle; polyhandle = polygons->GetNextPolygon(polyhandle))
			{
				k3dIPolygon* polygon = polygons->GetPolygon(polyhandle);

				// If the polygon isn't visible and selected, skip it ...
				if(SDP_S_OK != polygon->State()->QueryState(k3dIState::VISIBLESELECTED))
					continue;

				// Skip degenerate cases ...
				if(polygon->Points()->GetPointCount() < 3)
					continue;

				// For each polypoint in the polygon ...
				k3dIPolyPointServer* polypoints = polygon->Points();

				K3DPOLYPOINTHANDLE lastpolypointhandle = polypoints->GetFirstPolyPoint();
				K3DPOLYPOINTHANDLE polypointhandle = polypoints->GetNextPolyPoint(lastpolypointhandle);

				while(polypointhandle)
					{
						// Insert an edge record ...
						CExtrudeEdge edge(polypoints->GetPolyPoint(lastpolypointhandle)->Point(), polypoints->GetPolyPoint(polypointhandle)->Point());
						ExtrudeEdgeArrayIterator edgeiterator = std::find(m_Edges.begin(), m_Edges.end(), edge);
						if(edgeiterator != m_Edges.end())
							edgeiterator->m_ReferenceCount++;
						else
							m_Edges.push_back(edge);
						
						lastpolypointhandle = polypointhandle;
						polypointhandle = polypoints->GetNextPolyPoint(polypointhandle);
					}

				// Close the polygon ...
				CExtrudeEdge edge(polypoints->GetPolyPoint(lastpolypointhandle)->Point(), polypoints->GetPolyPoint(polypoints->GetFirstPolyPoint())->Point());
				ExtrudeEdgeArrayIterator edgeiterator = std::find(m_Edges.begin(), m_Edges.end(), edge);
				if(edgeiterator != m_Edges.end())
					edgeiterator->m_ReferenceCount++;
				else
					m_Edges.push_back(edge);
			}

		return true;
	}
	
protected:
	ExtrudeEdgeArray& m_Edges;
};

bool FindInteriorEdges(CExtrudeEdge Edge)
{
	return Edge.m_ReferenceCount > 1;
}

void CExtrude::OnModify(bool SaveAction)
{
	// Remove old geometry as necessary ...
	if(m_CreateGeometry)
		DeleteAll();

	if(SaveAction)
		Document()->Actions()->StartAction("Extrude");

	// Create new geometry if needed ...
	if(m_CreateGeometry)
		{
			// Clear out old data ...
			m_Polygons.clear();
			m_ExteriorEdges.clear();
			m_InteriorEdges.clear();
			m_Points.clear();
	
			// Build a mapping of selected polygons ...
			k3dPolygonArray polygons;
			k3dGetPolygons(Document(), k3dIState::VISIBLESELECTED, polygons);
			gulong i;
			for(i = 0; i < polygons.size(); i++)
				m_Polygons[polygons[i]] = k3dPolygonArray();

			// Build our array of selected edges ...
			ExtrudeEdgeArray edges;
			CExtrudeCreateEdgeList functor(edges);
			k3dForEachMesh(Document(), 0, functor);
			
			// Divvy-up interior and exterior edges ...
			for(i = 0; i < edges.size(); i++)
				{
					if(1 == edges[i].m_ReferenceCount)
						m_ExteriorEdges.push_back(edges[i]);
					else if(2 == edges[i].m_ReferenceCount)
						m_InteriorEdges.push_back(edges[i]);
					else
						g_assert_not_reached();
				}
				
			// Build our point mapping ...
			k3dPointArray points;
			k3dGetPoints(Document(), k3dIState::VISIBLESELECTED, points);
			for(i = 0; i < points.size(); i++)
				m_Points[points[i]] = k3dPointArray(guint(m_Segments), 0);

			// For each segment ...
			for(i = 1; i <= m_Segments; i++)
				{
					gulong currentlayer = i-1;
				
					// Create exterior points ...
					for(ExtrudeEdgeArrayIterator edge = m_ExteriorEdges.begin(); edge != m_ExteriorEdges.end(); edge++)
						{
							if(0 == m_Points[edge->m_Point1][currentlayer])
								m_Points[edge->m_Point1][currentlayer] = CreatePoint(edge->m_Point1->Mesh());
								
							if(0 == m_Points[edge->m_Point2][currentlayer])
								m_Points[edge->m_Point2][currentlayer] = CreatePoint(edge->m_Point2->Mesh());
						}
						
					// Conditionally create interior points ...
					if(m_ConnectInterior || (m_CreateLayers && (i < m_Segments)) || (m_CreateEndCaps && (i == m_Segments)))
						{
							for(ExtrudeEdgeArrayIterator edge = m_InteriorEdges.begin(); edge != m_InteriorEdges.end(); edge++)
								{
									if(0 == m_Points[edge->m_Point1][currentlayer])
										m_Points[edge->m_Point1][currentlayer] = CreatePoint(edge->m_Point1->Mesh());
								
									if(0 == m_Points[edge->m_Point2][currentlayer])
										m_Points[edge->m_Point2][currentlayer] = CreatePoint(edge->m_Point2->Mesh());
								}
						}

					// Conditionally connect layers ...
					if(m_ConnectExterior)
						{
							// For each exterior edge ...
							for(ExtrudeEdgeArrayIterator edge = m_ExteriorEdges.begin(); edge != m_ExteriorEdges.end(); edge++)
								{
									// Create a new polygon ...
									k3dIPolygon* newpolygon = CreatePolygon(edge->m_Point1->Mesh());
//									edge->m_NewPolygons.push_back(newpolygon);

									k3dIPoint* point1;
									k3dIPoint* point2;
									
									if(currentlayer)
										{
											point1 = m_Points[edge->m_Point2][currentlayer-1];
											point2 = m_Points[edge->m_Point1][currentlayer-1];
										}
									else
										{
											point1 = edge->m_Point2;
											point2 = edge->m_Point1;
										}
										
									k3dIPoint* point3 = m_Points[edge->m_Point1][currentlayer];
									k3dIPoint* point4 = m_Points[edge->m_Point2][currentlayer];
									
									if(edge->m_Flipped)
										{
											std::swap(point1, point2);
											std::swap(point3, point4);
										}
									
									newpolygon->Points()->AddPoint(point1);
									newpolygon->Points()->AddPoint(point2);
									newpolygon->Points()->AddPoint(point3);
									newpolygon->Points()->AddPoint(point4);
								}
						}

					// Conditionally connect layers ...
					if(m_ConnectInterior)
						{
							// For each interior edge ...
							for(ExtrudeEdgeArrayIterator edge = m_InteriorEdges.begin(); edge != m_InteriorEdges.end(); edge++)
								{
									// Create a new polygon ...
									k3dIPolygon* newpolygon = CreatePolygon(edge->m_Point1->Mesh());
//									edge->m_NewPolygons.push_back(newpolygon);

									k3dIPoint* point1;
									k3dIPoint* point2;
									
									if(currentlayer)
										{
											point1 = m_Points[edge->m_Point2][currentlayer-1];
											point2 = m_Points[edge->m_Point1][currentlayer-1];
										}
									else
										{
											point1 = edge->m_Point2;
											point2 = edge->m_Point1;
										}
										
									k3dIPoint* point3 = m_Points[edge->m_Point1][currentlayer];
									k3dIPoint* point4 = m_Points[edge->m_Point2][currentlayer];
									
									if(edge->m_Flipped)
										{
											std::swap(point1, point2);
											std::swap(point3, point4);
										}
									
									newpolygon->Points()->AddPoint(point1);
									newpolygon->Points()->AddPoint(point2);
									newpolygon->Points()->AddPoint(point3);
									newpolygon->Points()->AddPoint(point4);
								}
						}

					// Conditionally create layers / endcaps ...
					if((m_CreateLayers && (i < m_Segments)) || (m_CreateEndCaps && (i == m_Segments)))
						{
							// For each selected original polygon ...
							for(ExtrudePolygonMapIterator polygon = m_Polygons.begin(); polygon != m_Polygons.end(); polygon++)
								{
									// Create a new polygon ...
									k3dIPolygon* newpolygon = CreatePolygon(polygon->first->Mesh());
//									polygon->second.push_back(newpolygon);

									// For each existing polypoint ...
									k3dIPolyPointServer* polypoints = polygon->first->Points();
									for(K3DPOLYPOINTHANDLE polypointhandle = polypoints->GetFirstPolyPoint(); polypointhandle; polypointhandle = polypoints->GetNextPolyPoint(polypointhandle))
										{
											k3dIPolyPoint* oldpolypoint = polypoints->GetPolyPoint(polypointhandle);
											
											// Attach its newer counterpart to the new poly ...			
											k3dIPolyPoint* newpolypoint = newpolygon->Points()->AddPoint(m_Points[oldpolypoint->Point()][currentlayer]);
											
											// Copy texture coordinates ...
											sdpVector3 texturecoords;
											oldpolypoint->Location()->GetTextureXYZ(texturecoords);
											newpolypoint->Location()->SetTextureXYZ(texturecoords);
											
											oldpolypoint->Location()->GetImplicitTextureXYZ(texturecoords);
											newpolypoint->Location()->SetImplicitTextureXYZ(texturecoords);
										}
										
								}
						}						
				}
								
			m_CreateGeometry = false;
		}

	if(SaveAction)
		Document()->Actions()->FinishAction();

	// For each layer ...
	for(gulong i = 1; i <= m_Segments; i++)
		{
			gulong currentlayer = i-1;
		
			// Calculate how far we are along our "path" ...
			gdouble percent = gdouble(i) / gdouble(m_Segments);
			
			// Interpolate our transformation values ...
			sdpVector3 offset = percent * m_Offset;
			sdpVector3 scale = sdpLerp(sdpVector3(1, 1, 1), m_Scale, percent);
			sdpVector3 rotation = sdpLerp(sdpVector3(0, 0, 0), m_Rotation, percent);

			// Create a transformation matrix ...
			sdpMatrix4 matrix = sdpIdentity3D();
			matrix = matrix * sdpTranslation3D(offset);
			matrix = matrix * sdpRotation3D(rotation);
			matrix = matrix * sdpScaling3D(scale);

			// Calculate points ...
			for(ExtrudePointMapIterator point = m_Points.begin(); point != m_Points.end(); point++)
				{
					// Skip non-existent points ...
					if(0 == point->second[currentlayer])
						continue;
						
					// Get the original point location ...
					sdpVector3 coords;
					point->first->Location()->GetLocalXYZ(coords);

					// Transform it ...
					coords = matrix * coords;
					
					// Store the new coordinates ...
					point->second[currentlayer]->Location()->SetOldXYZ(coords);
					point->second[currentlayer]->Location()->SetLocalXYZ(coords);
				}
		}
	
	// Update statistics in our dialog ...
	DisplayStatistics(m_UserInterface.Entry("stats"));
	k3dRedrawAll(Document(), false);
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CExtrude::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	gdouble xfactor = (CurrentX - LastX) * 0.01;
	gdouble yfactor = (CurrentY - LastY) * 0.01;

	switch(m_UserInterface.m_CurrentPage)
		{
			case CExtrudeDialog::EXTRUDE:
				break;
			case CExtrudeDialog::OFFSETS:
				// If the user holds down either control key, only translate along the active axis ...
				if(ControlDown(Modifiers))
					{
						switch(Document()->GetActiveAxis())
							{
								case k3dIDocument::XAXIS:
									m_Offset[0] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								case k3dIDocument::YAXIS:
									m_Offset[1] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								case k3dIDocument::ZAXIS:
									m_Offset[2] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								default:
									g_assert(0);
							}
					}
				// Otherwise, translate along the two non-active axes ...
				else
					{
						switch(Document()->GetActiveAxis())
							{
								case k3dIDocument::XAXIS:
									m_Offset[2] -= xfactor;
									m_Offset[1] -= yfactor;
									break;
								case k3dIDocument::YAXIS:
									m_Offset[0] -= xfactor;
									m_Offset[2] -= yfactor;
									break;
								case k3dIDocument::ZAXIS:
									m_Offset[0] += xfactor;
									m_Offset[1] -= yfactor;
									break;
								default:
									g_assert(0);
							}
					}
				break;
			case CExtrudeDialog::SCALING:
				// If the user holds down either shift key, constrain scale factors to be equal ...
				if(ShiftDown(Modifiers))
					{
						m_Scale -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
					}
				// If the user holds down either control key, only scale along the active axis ...
				else if(ControlDown(Modifiers))
					{
						switch(Document()->GetActiveAxis())
							{
								case k3dIDocument::XAXIS:
									m_Scale[0]-=(fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								case k3dIDocument::YAXIS:
									m_Scale[1]-=(fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								case k3dIDocument::ZAXIS:
									m_Scale[2]-=(fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								default:
									g_assert_not_reached();
							}
					}
				// Otherwise, scale along the two non-active axes ...
				else
					{
						switch(Document()->GetActiveAxis())
							{
								case k3dIDocument::XAXIS:
									m_Scale[2] -= xfactor;
									m_Scale[1] -= yfactor;
									break;
								case k3dIDocument::YAXIS:
									m_Scale[0] -= xfactor;
									m_Scale[2] -= yfactor;
									break;
								case k3dIDocument::ZAXIS:
									m_Scale[0] -= xfactor;
									m_Scale[1] -= yfactor;
									break;
								default:
									g_assert(0);
							}
					}
				break;
			case CExtrudeDialog::ROTATION:
				// Compute scaling factors for each mouse axis ...
				xfactor = sdpRadians(gdouble(CurrentX - LastX) * 0.1);
				yfactor = sdpRadians(gdouble(CurrentY - LastY) * 0.1);

				// If the user holds down either control key, only rotate along the active axis ...
				if(ControlDown(Modifiers))
					{
						switch(Document()->GetActiveAxis())
							{
								case k3dIDocument::XAXIS:
									m_Rotation[1] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								case k3dIDocument::YAXIS:
									m_Rotation[2] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								case k3dIDocument::ZAXIS:
									m_Rotation[0] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
									break;
								default:
									g_assert_not_reached();	// Did somebody add an axis?
							}
					}
				// Otherwise, rotate along the two non-active axes ...
				else
					{
						switch(Document()->GetActiveAxis())
							{
								case k3dIDocument::XAXIS:
									m_Rotation[0] -= xfactor;
									m_Rotation[2] -= yfactor;
									break;
								case k3dIDocument::YAXIS:
									m_Rotation[1] -= xfactor;
									m_Rotation[0] -= yfactor;
									break;
								case k3dIDocument::ZAXIS:
									m_Rotation[1] += xfactor;
									m_Rotation[2] -= yfactor;
									break;
								default:
									g_assert_not_reached();	// Did somebody add an axis?
							}
					}
				break;
			default:
				g_assert_not_reached();
				break;
		}
		
	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}

////////////////////////////////////////////////////////////////////////////
// CExtrudeDialog

CExtrudeDialog::CExtrudeDialog()
{
	m_Object = 0;
	m_CurrentPage = EXTRUDE;
}

bool CExtrudeDialog::Create(CExtrude* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dPrimToolDialog::Create(Object, true, "extrude.gtkml", "extrude.html"), false);

	return true;
}

bool CExtrudeDialog::InitializeControls()
{
	Notebook("tabs").SetPage(m_CurrentPage);

	return true;
}

void CExtrudeDialog::UpdateControls()
{
	Entry("segments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Segments, 0));

	ToggleButton("connectexterior").SetState(m_Object->m_ConnectExterior);
	ToggleButton("connectinterior").SetState(m_Object->m_ConnectInterior);
	ToggleButton("createlayers").SetState(m_Object->m_CreateLayers);
	ToggleButton("createendcaps").SetState(m_Object->m_CreateEndCaps);
	
	Entry("offsetx").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Offset[0], 3));
	Entry("offsety").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Offset[1], 3));
	Entry("offsetz").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Offset[2], 3));
	
	Entry("scalex").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Scale[0], 3));
	Entry("scaley").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Scale[1], 3));
	Entry("scalez").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Scale[2], 3));
	
	Entry("rotationx").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Rotation[0], 3));
	Entry("rotationy").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Rotation[1], 3));
	Entry("rotationz").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Rotation[2], 3));
}

void CExtrudeDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onsegments")
		OnSegments();
	else if(Event->Name() == "onconnectexterior")
		OnConnectExterior();
	else if(Event->Name() == "onconnectinterior")
		OnConnectInterior();
	else if(Event->Name() == "oncreatelayers")
		OnCreateLayers();
	else if(Event->Name() == "oncreateendcaps")
		OnCreateEndCaps();
	else if(Event->Name() == "onoffsetx")
		OnOffsetX();
	else if(Event->Name() == "onoffsety")
		OnOffsetY();
	else if(Event->Name() == "onoffsetz")
		OnOffsetZ();
	else if(Event->Name() == "onscalex")
		OnScaleX();
	else if(Event->Name() == "onscaley")
		OnScaleY();
	else if(Event->Name() == "onscalez")
		OnScaleZ();
	else if(Event->Name() == "onrotationx")
		OnRotationX();
	else if(Event->Name() == "onrotationy")
		OnRotationY();
	else if(Event->Name() == "onrotationz")
		OnRotationZ();
	else if(Event->Name() == "onpageswitch")
		OnPageSwitch(Event);
	else
		k3dPrimToolDialog::OnEvent(Event);
}

void CExtrudeDialog::OnSegments()
{
	if(UpdateControl(Entry("segments"), m_Object->m_Segments, k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CExtrudeDialog::OnConnectExterior()
{
	m_Object->m_ConnectExterior = ToggleButton("connectexterior").GetState();
	m_Object->m_CreateGeometry = true;
	m_Object->OnModify(false);
}

void CExtrudeDialog::OnConnectInterior()
{
	m_Object->m_ConnectInterior = ToggleButton("connectinterior").GetState();
	m_Object->m_CreateGeometry = true;
	m_Object->OnModify(false);
}

void CExtrudeDialog::OnCreateLayers()
{
	m_Object->m_CreateLayers = ToggleButton("createlayers").GetState();
	m_Object->m_CreateGeometry = true;
	m_Object->OnModify(false);
}

void CExtrudeDialog::OnCreateEndCaps()
{
	m_Object->m_CreateEndCaps = ToggleButton("createendcaps").GetState();
	m_Object->m_CreateGeometry = true;
	m_Object->OnModify(false);
}

void CExtrudeDialog::OnOffsetX()
{
	if(UpdateControl(Entry("offsetx"), m_Object->m_Offset[0], k3dIChannel::DATA_DISTANCE, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnOffsetY()
{
	if(UpdateControl(Entry("offsety"), m_Object->m_Offset[1], k3dIChannel::DATA_DISTANCE, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnOffsetZ()
{
	if(UpdateControl(Entry("offsetz"), m_Object->m_Offset[2], k3dIChannel::DATA_DISTANCE, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnScaleX()
{
	if(UpdateControl(Entry("scalex"), m_Object->m_Scale[0], k3dIChannel::DATA_SCALAR, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnScaleY()
{
	if(UpdateControl(Entry("scaley"), m_Object->m_Scale[1], k3dIChannel::DATA_SCALAR, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnScaleZ()
{
	if(UpdateControl(Entry("scalez"), m_Object->m_Scale[2], k3dIChannel::DATA_SCALAR, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnRotationX()
{
	if(UpdateControl(Entry("rotationx"), m_Object->m_Rotation[0], k3dIChannel::DATA_ANGLE, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnRotationY()
{
	if(UpdateControl(Entry("rotationy"), m_Object->m_Rotation[1], k3dIChannel::DATA_ANGLE, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnRotationZ()
{
	if(UpdateControl(Entry("rotationz"), m_Object->m_Rotation[2], k3dIChannel::DATA_ANGLE, -DBL_MAX, DBL_MAX, 3))
		m_Object->OnModify(false);
}

void CExtrudeDialog::OnPageSwitch(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Keep track of the current page ...
	sdpGtkEventNotebookSwitchPage* event = (sdpGtkEventNotebookSwitchPage*)Event;
	m_CurrentPage = NotebookPage(event->PageNumber());
}


