// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CLissaCurve K-3D tool, which generates geometry composed of Lissajous curves
		\author Tim Shead (tshead@k-3d.com)
*/

#include "lissacurve.h"

#include <k3dmisc.h>

#include <k3dipolypointserver.h>
#include <k3diactions.h>

#include <sdpfile.h>
#include <sdpgtkevent.h>

/////////////////////////////////////////////////////////////////////////////
// CLissaCurve

CLissaCurve::CLissaCurve()
{
	m_EdgeCount = 100;

	m_MAmp = 0.0;
	m_MFreq = 20.0;
	m_MPhase = 0.0;
	m_CustomM = false;
	m_XAmp = 5.0;
	m_XFreq = 1.0;
	m_XPhase = sdpRadians(90.0);
	m_CustomX = false;
	m_YAmp = 5.0;
	m_YFreq = 1.0;
	m_YPhase = 0.0;
	m_CustomY = false;
	m_ZAmp = 0.0;
	m_ZFreq = 1.0;
	m_ZPhase = 0.0;
	m_CustomZ = false;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CLissaCurve::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}


gulong CLissaCurve::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}


gulong CLissaCurve::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}





///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CLissaCurve::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;

/*
	if(sdpFailed(m_Xwaveform.CreateScalar(Document(), "LissaCurve X waveform", 0.0, TRUE, false, -1.0, 1.0, "Distance", "Amplitude", k3dIChannel::DATA_SCALAR, false, this, m_ObjectContext->ObjectID()))) KFAIL(SDP_E_FAIL);
	AddObject(&m_Xwaveform);
	m_Xwaveform->SetLinear1(-1.0, 1.0);

	if(sdpFailed(m_Ywaveform.CreateScalar(Document(), "LissaCurve Y waveform", 0.0, TRUE, false, -1.0, 1.0, "Distance", "Amplitude", k3dIChannel::DATA_SCALAR, false, this, m_ObjectContext->ObjectID()))) KFAIL(SDP_E_FAIL);
	AddObject(&m_Ywaveform);
	m_Ywaveform->SetLinear1(-1.0, 1.0);

	if(sdpFailed(m_Zwaveform.CreateScalar(Document(), "LissaCurve Z waveform", 0.0, TRUE, false, -1.0, 1.0, "Distance", "Amplitude", k3dIChannel::DATA_SCALAR, false, this, m_ObjectContext->ObjectID()))) KFAIL(SDP_E_FAIL);
	AddObject(&m_Zwaveform);
	m_Zwaveform->SetLinear1(-1.0, 1.0);

	if(sdpFailed(m_Mwaveform.CreateScalar(Document(), "LissaCurve M waveform", 0.0, TRUE, false, -1.0, 1.0, "Distance", "Amplitude", k3dIChannel::DATA_SCALAR, false, this, m_ObjectContext->ObjectID()))) KFAIL(SDP_E_FAIL);
	AddObject(&m_Mwaveform);
	m_Mwaveform->SetLinear1(-1.0, 1.0);

	if(!KOpenDialog(m_Sheet, this)) KFAIL(SDP_E_FAIL);
*/
}

sdpResult CLissaCurve::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPrimTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

/*
sdpResult CLissaCurve::OnLButtonStartDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	k3dPrimTool::OnLButtonStartDrag(Modifiers, MouseX, MouseY);
	return SDP_S_OK;
}

sdpResult CLissaCurve::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	k3dPrimTool::OnLButtonDrag(Modifiers, MouseX, MouseY);

	return SDP_S_OK;
}


sdpResult CLissaCurve::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	k3dPrimTool::OnLButtonEndDrag(Modifiers, MouseX, MouseY);

	return SDP_S_OK;
}
*/



/////////////////////////////////////////////////////////////////////////////
// CLissaCurve message handlers

void CLissaCurve::OnModify(bool SaveAction)
{

	// Calculate number of points / polys ...
	gulong numpoints = m_EdgeCount;
	gulong numpolys = 1;

	// If we need to recreate our geometry ...
	if(m_CreateGeometry)
		{
			gulong i;

			// Cleanup ...
			DeleteAll();

			if(SaveAction)
				Document()->Actions()->StartAction("LissaCurve");
				
			// Create the mesh ...
			CreateMesh("LissaCurve", numpoints, numpolys);
			
			if(SaveAction)
				Document()->Actions()->FinishAction();

			// Attach points to the polygon ...
			for(i = 0; i < numpoints; i++)
				m_Polygons[0][0]->Points()->AddPoint(m_Points[0][i]);
		}


	double mvalue;
	sdpVector3 coords;
	double percentdone, radiansdone, channeldone;

	for(gulong i = 0; i < numpoints; i++)
		{
			percentdone = (double)i / (double)m_EdgeCount;
			radiansdone = percentdone * sdpPiTimes2;

			if(m_CustomX)
				{
					channeldone = (percentdone * m_XFreq) + m_XPhase;
					if(channeldone > 1.0) channeldone -= floor(channeldone);

//					channeldone = m_Xwaveform->GetValue1(channeldone, 0.001);

					coords[0] = m_XAmp * channeldone;

				}
			else
				{
					coords[0] = m_XAmp * sin(m_XFreq * radiansdone + m_XPhase);
				}

			if(m_CustomY)
				{
					channeldone = (percentdone * m_YFreq) + m_YPhase;
					if(channeldone > 1.0) channeldone -= floor(channeldone);

//					channeldone = m_Ywaveform->GetValue1(channeldone, 0.001);

					coords[1] = m_YAmp * channeldone;
				}
			else
				{
					coords[1] = m_YAmp * sin(m_YFreq * radiansdone + m_YPhase);
				}

			if(m_CustomZ)
				{
					channeldone = (percentdone * m_ZFreq) + m_ZPhase;
					if(channeldone > 1.0) channeldone -= floor(channeldone);

//					channeldone = m_Zwaveform->GetValue1(channeldone, 0.001);

					coords[2] = m_ZAmp * channeldone;

				}
			else
				{
					coords[2] = m_ZAmp * sin(m_ZFreq * radiansdone + m_ZPhase);
				}

			if(m_CustomM)
				{
					channeldone = (percentdone * m_MFreq) + m_MPhase;
					if(channeldone > 1.0) channeldone -= floor(channeldone);

//					channeldone = m_Mwaveform->GetValue1(channeldone, 0.001);

					mvalue = 1 + (m_MAmp * channeldone);

				}
			else
				{
					mvalue = 1.0 + (m_MAmp * sin(m_MFreq * radiansdone + m_MPhase));
				}

			if(mvalue) coords *= mvalue;

			m_Points[0][i]->Location()->SetOldXYZ(coords);
			m_Points[0][i]->Location()->SetLocalXYZ(coords);
		}


	m_CreateGeometry = false;

//	m_Sheet->UpdateNumbers();

	k3dPrimTool::OnModify(SaveAction);
}


/////////////////////////////////////////////////////////////////////////////
// CLissaCurveDialog

CLissaCurveDialog::CLissaCurveDialog()
{
	m_Object = 0;
}

bool CLissaCurveDialog::Create(CLissaCurve* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dPrimToolDialog::Create(Object, true, "lissacurve.gtkml", "lissacurve.html"), false);

	return true;
}

void CLissaCurveDialog::UpdateControls()
{
	Entry("edgecount").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_EdgeCount, 0));

	Entry("xfreq").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_XFreq, 2));
	Entry("xamp").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_XAmp, 2));
	Entry("xphase").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_XPhase, 2));

	Entry("yfreq").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_YFreq, 2));
	Entry("yamp").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_YAmp, 2));
	Entry("yphase").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_YPhase, 2));

	Entry("zfreq").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_ZFreq, 2));
	Entry("zamp").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_ZAmp, 2));
	Entry("zphase").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_ZPhase, 2));

	Entry("mfreq").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_MFreq, 2));
	Entry("mamp").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_MAmp, 2));
	Entry("mphase").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_MPhase, 2));
}


void CLissaCurveDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onedgecount")
		OnEdgeCount();
	else if(Event->Name() == "onxfreq")
		OnXFreq();
	else if(Event->Name() == "onxamp")
		OnXAmp();
	else if(Event->Name() == "onxphase")
		OnXPhase();
	else if(Event->Name() == "onyfreq")
		OnYFreq();
	else if(Event->Name() == "onyamp")
		OnYAmp();
	else if(Event->Name() == "onyphase")
		OnYPhase();
	else if(Event->Name() == "onzfreq")
		OnZFreq();
	else if(Event->Name() == "onzamp")
		OnZAmp();
	else if(Event->Name() == "onzphase")
		OnZPhase();
	else if(Event->Name() == "onmfreq")
		OnMFreq();
	else if(Event->Name() == "onmamp")
		OnMAmp();
	else if(Event->Name() == "onmphase")
		OnMPhase();
	else
		k3dPrimToolDialog::OnEvent(Event);
}


void CLissaCurveDialog::OnEdgeCount()
{
	if(UpdateControl(Entry("edgecount"), m_Object->m_EdgeCount, k3dIChannel::DATA_SCALAR, 3, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}

}

void CLissaCurveDialog::OnXFreq()
{
	if(UpdateControl(Entry("xfreq"), m_Object->m_XFreq, k3dIChannel::DATA_SCALAR, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnXAmp()
{
	if(UpdateControl(Entry("xamp"), m_Object->m_XAmp, k3dIChannel::DATA_DISTANCE, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnXPhase()
{
	if(UpdateControl(Entry("xphase"), m_Object->m_XPhase, k3dIChannel::DATA_ANGLE, 2))
		m_Object->OnModify(false);

};

void CLissaCurveDialog::OnYFreq()
{
	if(UpdateControl(Entry("yfreq"), m_Object->m_YFreq, k3dIChannel::DATA_SCALAR, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnYAmp()
{
	if(UpdateControl(Entry("yamp"), m_Object->m_YAmp, k3dIChannel::DATA_DISTANCE, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnYPhase()
{
	if(UpdateControl(Entry("yphase"), m_Object->m_YPhase, k3dIChannel::DATA_ANGLE, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnZFreq()
{
	if(UpdateControl(Entry("zfreq"), m_Object->m_ZFreq, k3dIChannel::DATA_SCALAR, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnZAmp()
{
	if(UpdateControl(Entry("zamp"), m_Object->m_ZAmp, k3dIChannel::DATA_DISTANCE, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnZPhase()
{
	if(UpdateControl(Entry("zphase"), m_Object->m_ZPhase, k3dIChannel::DATA_ANGLE, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnMFreq()
{
	if(UpdateControl(Entry("mfreq"), m_Object->m_MFreq, k3dIChannel::DATA_SCALAR, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnMAmp()
{
	if(UpdateControl(Entry("mamp"), m_Object->m_MAmp, k3dIChannel::DATA_SCALAR, 2))
		m_Object->OnModify(false);

}

void CLissaCurveDialog::OnMPhase()
{
	if(UpdateControl(Entry("mphase"), m_Object->m_MPhase, k3dIChannel::DATA_ANGLE, 2))
		m_Object->OnModify(false);

}



/*

void CLissaCurvePage::OnCustomm()
{
	m_Sheet->m_Object->m_CustomM = IsDlgButtonChecked(IDC_CUSTOMM) ? TRUE : false;
	OnModify();
}

void CLissaCurvePage::OnCustomx()
{
	m_Sheet->m_Object->m_CustomX = IsDlgButtonChecked(IDC_CUSTOMX) ? TRUE : false;
	OnModify();
}

void CLissaCurvePage::OnCustomy()
{
	m_Sheet->m_Object->m_CustomY = IsDlgButtonChecked(IDC_CUSTOMY) ? TRUE : false;
	OnModify();
}

void CLissaCurvePage::OnCustomz()
{
	m_Sheet->m_Object->m_CustomZ = IsDlgButtonChecked(IDC_CUSTOMZ) ? TRUE : false;
	OnModify();
}





BOOL CLissaCurvePage::OnInitDialog()
{
	KPage::OnInitDialog();

	k3dSetLong(GetDlgItem(IDC_EDGES)->m_hWnd, m_Sheet->m_Object->m_Edges);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, GetDlgItem(IDC_MAMP)->m_hWnd, m_Sheet->m_Object->m_MAmp);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, GetDlgItem(IDC_MFREQ)->m_hWnd, m_Sheet->m_Object->m_MFreq);
	k3dSetDouble(k3dIChannel::DATA_ANGLE, GetDlgItem(IDC_MPHASE)->m_hWnd, m_Sheet->m_Object->m_MPhase);
	k3dSetDouble(k3dIChannel::DATA_DISTANCE, GetDlgItem(IDC_XAMP)->m_hWnd, m_Sheet->m_Object->m_XAmp);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, GetDlgItem(IDC_XFREQ)->m_hWnd, m_Sheet->m_Object->m_XFreq);
	k3dSetDouble(k3dIChannel::DATA_ANGLE, GetDlgItem(IDC_XPHASE)->m_hWnd, m_Sheet->m_Object->m_XPhase);
	k3dSetDouble(k3dIChannel::DATA_DISTANCE, GetDlgItem(IDC_YAMP)->m_hWnd, m_Sheet->m_Object->m_YAmp);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, GetDlgItem(IDC_YFREQ)->m_hWnd, m_Sheet->m_Object->m_YFreq);
	k3dSetDouble(k3dIChannel::DATA_ANGLE, GetDlgItem(IDC_YPHASE)->m_hWnd, m_Sheet->m_Object->m_YPhase);
	k3dSetDouble(k3dIChannel::DATA_DISTANCE, GetDlgItem(IDC_ZAMP)->m_hWnd, m_Sheet->m_Object->m_ZAmp);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, GetDlgItem(IDC_ZFREQ)->m_hWnd, m_Sheet->m_Object->m_ZFreq);
	k3dSetDouble(k3dIChannel::DATA_ANGLE, GetDlgItem(IDC_ZPHASE)->m_hWnd, m_Sheet->m_Object->m_ZPhase);

	CheckDlgButton(IDC_CUSTOMX, m_Sheet->m_Object->m_CustomX);
	CheckDlgButton(IDC_CUSTOMY, m_Sheet->m_Object->m_CustomY);
	CheckDlgButton(IDC_CUSTOMZ, m_Sheet->m_Object->m_CustomZ);
	CheckDlgButton(IDC_CUSTOMM, m_Sheet->m_Object->m_CustomM);

//	m_Sheet->m_Object->m_Xwaveform->Attach(GetDlgItem(IDC_XWAVEFORM)->m_hWnd);
//	m_Sheet->m_Object->m_Ywaveform->Attach(GetDlgItem(IDC_YWAVEFORM)->m_hWnd);
//	m_Sheet->m_Object->m_Zwaveform->Attach(GetDlgItem(IDC_ZWAVEFORM)->m_hWnd);
//	m_Sheet->m_Object->m_Mwaveform->Attach(GetDlgItem(IDC_MWAVEFORM)->m_hWnd);

	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}






void CLissaCurvePage::OnResetall()
{
	m_Sheet->m_Object->m_Edges = 100;

	m_Sheet->m_Object->m_MAmp = 0.0;
	m_Sheet->m_Object->m_MFreq = 20.0;
	m_Sheet->m_Object->m_MPhase = 0.0;
	m_Sheet->m_Object->m_CustomM = FALSE;
	m_Sheet->m_Object->m_XAmp = 5.0;
	m_Sheet->m_Object->m_XFreq = 1.0;
	m_Sheet->m_Object->m_XPhase = 90.0;
	m_Sheet->m_Object->m_CustomX = FALSE;
	m_Sheet->m_Object->m_YAmp = 5.0;
	m_Sheet->m_Object->m_YFreq = 1.0;
	m_Sheet->m_Object->m_YPhase = 0.0;
	m_Sheet->m_Object->m_CustomY = FALSE;
	m_Sheet->m_Object->m_ZAmp = 0.0;
	m_Sheet->m_Object->m_ZFreq = 1.0;
	m_Sheet->m_Object->m_ZPhase = 0.0;
	m_Sheet->m_Object->m_CustomZ = FALSE;

	OnModify();
}

*/
