// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CMirror K-3D tool, which reflects geometry about an axis
		\author Tim Shead (tshead@k-3d.com)
*/

#include "mirror.h"

#include <k3diactions.h>
#include <k3dimesh.h>
#include <k3dipolygon.h>
#include <k3dipolypointserver.h>
#include <k3dipolypoint.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>

#include <sdpfile.h>

/////////////////////////////////////////////////////////////////////////////
// CMirror

CMirror::CMirror()
{
	m_Axis = X;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CMirror::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CMirror::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CMirror::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CMirror::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	return SDP_S_OK;
}

sdpResult CMirror::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

class CMirrorMirror
{
public:
	CMirrorMirror(CMirror::AxisType Axis) { m_Axis = Axis; }
	
	bool HandlePoint(k3dIPoint* Point)
	{
		sdpVector3 coords;
		
		Point->Location()->GetLocalXYZ(coords);
		coords[m_Axis] = -coords[m_Axis];
		Point->Location()->SetLocalXYZ(coords);
	
		return true;
	}

protected:
	CMirror::AxisType m_Axis;
};

void CMirror::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Mirror");

	CMirrorMirror functor(m_Axis);
	k3dForEachPoint(Document(), k3dIState::VISIBLESELECTED, functor);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}

	k3dRedrawAll(Document(), false);		
}

//////////////////////////////////////////////////////////////////////////////////////
// CMirrorDialog

CMirrorDialog::CMirrorDialog()
{
	m_Object = 0;
}

bool CMirrorDialog::Create(CMirror* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "mirror.gtkml", "mirror.html"), false);

	return true;
}

void CMirrorDialog::UpdateControls()
{
	ToggleButton("x").SetState(m_Object->m_Axis == CMirror::X);
	ToggleButton("y").SetState(m_Object->m_Axis == CMirror::Y);
	ToggleButton("z").SetState(m_Object->m_Axis == CMirror::Z);
}

void CMirrorDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onx")
		OnX();
	else if(Event->Name() == "ony")
		OnY();
	else if(Event->Name() == "onz")
		OnZ();
	else if(Event->Name() == "onapply")
		OnApply();
	else
		k3dDialog::OnEvent(Event);
}

void CMirrorDialog::OnX()
{
	if(ToggleButton("x").GetState())
		m_Object->m_Axis = CMirror::X;
}

void CMirrorDialog::OnY()
{
	if(ToggleButton("y").GetState())
		m_Object->m_Axis = CMirror::Y;
}

void CMirrorDialog::OnZ()
{
	if(ToggleButton("z").GetState())
		m_Object->m_Axis = CMirror::Z;
}

void CMirrorDialog::OnApply()
{
	m_Object->OnModify(true);
}

