// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CPrimitives K-3D tool, for creating simple geometric primitives
		\author Timothy M. Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevents.h>
#include <k3dipolypointserver.h>

#include "primitives.h"

#include <algorithm>
#include <strstream>



/////////////////////////////////////////////////////////////////////////////
// CPrimitivesDialog

CPrimitivesDialog::CPrimitivesDialog()
{
	m_Object = 0;
	m_CurrentPage = 0;
}

bool CPrimitivesDialog::Create(CPrimitives* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dPrimToolDialog::Create(Object, true, "primitives.gtkml", "primitives.html"), false);

	return true;
}

bool CPrimitivesDialog::InitializeControls()
{
	// Initialize our page ...
	Notebook("tabs").SetPage(0);

	// Build polyhedra clist
	sdpGtkCList polyhedron_clist = CList("polyhedronchoice");

	polyhedron_clist.SetColumnTitle(0, "Nb");
	polyhedron_clist.SetColumnTitle(1, "Name");
	polyhedron_clist.ShowColumnTitles();

	sdpString number;
	const char* clistdata[2];

	for(gint j = 0; j < 80; j++)
		{
			number = sdpFormatString("%d", j+1);
			clistdata[0] = number;
			clistdata[1] = uniform[j].name;
			polyhedron_clist.Append(clistdata);

			number=sdpFormatString("%d*", j+1);
			clistdata[0] = number;
			clistdata[1] = uniform[j].dual;
			polyhedron_clist.Append(clistdata);
		}

	CList("polyhedronchoice").SetColumnWidth(0, 25);

	return true;
}

void CPrimitivesDialog::UpdateControls()
{
	UpdateDiskControls();
	UpdateConeControls();
	UpdateCylinderControls();
	UpdateCubeControls();
	UpdateGridControls();
	UpdateSphereControls();
	UpdateTorusControls();
	UpdatePolyhedronControls();
}

void CPrimitivesDialog::UpdateDiskControls()
{
	Entry("diskradius").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_DiskRadius, 3));
	Entry("disksegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_DiskSegments, 0));
	Entry("diskpower").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_DiskPower, 3));
}

void CPrimitivesDialog::UpdateConeControls()
{
	Entry("conelength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_ConeLength, 3));
	Entry("coneradius").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_ConeRadius, 3));
	Entry("conelengthsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_ConeLengthSegments, 0));
	Entry("coneradialsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_ConeRadialSegments, 0));
	Entry("conelengthpower").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_ConeLengthPower, 3));
	Entry("coneradialpower").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_ConeRadialPower, 3));
}

void CPrimitivesDialog::UpdateCylinderControls()
{
	Entry("cylinderlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_CylinderLength, 3));
	Entry("cylinderradius").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_CylinderRadius, 3));
	Entry("cylinderlengthsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_CylinderLengthSegments, 0));
	Entry("cylinderradialsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_CylinderRadialSegments, 0));
	Entry("cylinderradialpower").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_CylinderRadialPower, 3));
}

void CPrimitivesDialog::UpdateCubeControls()
{
	Entry("cubexlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_CubeXLength, 3));
	Entry("cubeylength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_CubeYLength, 3));
	Entry("cubezlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_CubeZLength, 3));
	Entry("cubexsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_CubeXSegments, 0));
	Entry("cubeysegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_CubeYSegments, 0));
	Entry("cubezsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_CubeZSegments, 0));
}

void CPrimitivesDialog::UpdateGridControls()
{
	Entry("gridxlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_GridXLength, 3));
	Entry("gridylength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_GridYLength, 3));
	Entry("gridxsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_GridXSegments, 0));
	Entry("gridysegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_GridYSegments, 0));
}

void CPrimitivesDialog::UpdateSphereControls()
{
	Entry("spherexlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_SphereXLength, 3));
	Entry("sphereylength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_SphereYLength, 3));
	Entry("spherezlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_SphereZLength, 3));
	Entry("spherelengthsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_SphereLengthSegments, 0));
	Entry("sphereradialsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_SphereRadialSegments, 0));
	Entry("spherelengthpower").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_SphereLengthPower, 3));
	Entry("sphereradialpower").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_SphereRadialPower, 3));
}

void CPrimitivesDialog::UpdateTorusControls()
{
	Entry("torusinnerradius").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_TorusInnerRadius, 3));
	Entry("torusouterradius").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_TorusOuterRadius, 3));
	Entry("toruscrosssectionsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_TorusCrossSectionSegments, 0));
	Entry("torusradialsegments").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_TorusRadialSegments, 0));
	Entry("toruscrosssectionpower").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_TorusCrossSectionPower, 3));
	Entry("torusradialpower").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_TorusRadialPower, 3));
}

void CPrimitivesDialog::UpdatePolyhedronControls()
{
	Entry("polyhedronlength").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_PolyhedronLength, 3));
}

void CPrimitivesDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onpageswitch")
		OnPageSwitch(Event);
	else if(Event->Name() == "ondisktype")
		OnDiskType();
	else if(Event->Name() == "ondiskradius")
		OnDiskRadius();
	else if(Event->Name() == "ondisksegments")
		OnDiskSegments();
	else if(Event->Name() == "ondiskpower")
		OnDiskPower();
	else if(Event->Name() == "onconelength")
		OnConeLength();
	else if(Event->Name() == "onconeradius")
		OnConeRadius();
	else if(Event->Name() == "onconelengthpower")
		OnConeLengthPower();
	else if(Event->Name() == "onconeradialpower")
		OnConeRadialPower();
	else if(Event->Name() == "onconelengthsegments")
		OnConeLengthSegments();
	else if(Event->Name() == "onconeradialsegments")
		OnConeRadialSegments();
	else if(Event->Name() == "onconecreatebase")
		OnConeCreateBase();
	else if(Event->Name() == "oncylinderlength")
		OnCylinderLength();
	else if(Event->Name() == "oncylinderradius")
		OnCylinderRadius();
	else if(Event->Name() == "oncylinderradialpower")
		OnCylinderRadialPower();
	else if(Event->Name() == "oncylinderlengthsegments")
		OnCylinderLengthSegments();
	else if(Event->Name() == "oncylinderradialsegments")
		OnCylinderRadialSegments();
	else if(Event->Name() == "oncylindercreatebase")
		OnCylinderCreateBase();
	else if(Event->Name() == "oncubexlength")
		OnCubeXLength();
	else if(Event->Name() == "oncubeylength")
		OnCubeYLength();
	else if(Event->Name() == "oncubezlength")
		OnCubeZLength();
	else if(Event->Name() == "oncubexsegments")
		OnCubeXSegments();
	else if(Event->Name() == "oncubeysegments")
		OnCubeYSegments();
	else if(Event->Name() == "oncubezsegments")
		OnCubeZSegments();
	else if(Event->Name() == "ongridxlength")
		OnGridXLength();
	else if(Event->Name() == "ongridylength")
		OnGridYLength();
	else if(Event->Name() == "ongridxsegments")
		OnGridXSegments();
	else if(Event->Name() == "ongridysegments")
		OnGridYSegments();
	else if(Event->Name() == "onspherexlength")
		OnSphereXLength();
	else if(Event->Name() == "onsphereylength")
		OnSphereYLength();
	else if(Event->Name() == "onspherezlength")
		OnSphereZLength();
	else if(Event->Name() == "onspherelengthpower")
		OnSphereLengthPower();
	else if(Event->Name() == "onsphereradialpower")
		OnSphereRadialPower();
	else if(Event->Name() == "onspherelengthsegments")
		OnSphereLengthSegments();
	else if(Event->Name() == "onsphereradialsegments")
		OnSphereRadialSegments();
	else if(Event->Name() == "ontorusinnerradius")
		OnTorusInnerRadius();
	else if(Event->Name() == "ontorusouterradius")
		OnTorusOuterRadius();
	else if(Event->Name() == "ontorusradialpower")
		OnTorusRadialPower();
	else if(Event->Name() == "ontoruscrosssectionpower")
		OnTorusCrossSectionPower();
	else if(Event->Name() == "ontorusradialsegments")
		OnTorusRadialSegments();
	else if(Event->Name() == "ontoruscrosssectionsegments")
		OnTorusCrossSectionSegments();
	else if(Event->Name() == "onpolyhedronlength")
		OnPolyhedronLength();
	else if(Event->Name() == "onpolyhedronchoice")
		OnPolyhedronChoice(Event);
	else
		k3dPrimToolDialog::OnEvent(Event);
}

// User switched from one tab to another ...
void CPrimitivesDialog::OnPageSwitch(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	// Keep track of the current page ...
	sdpGtkEventNotebookSwitchPage* event = (sdpGtkEventNotebookSwitchPage*)Event;
	m_CurrentPage = event->PageNumber();

	// Modify the geometry ...
	m_Object->m_CreateGeometry = true;
	m_Object->OnModify(false);

}

void CPrimitivesDialog::OnDiskType()
{
}

void CPrimitivesDialog::OnDiskRadius()
{
	if(UpdateControl(Entry("diskradius"), m_Object->m_DiskRadius, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnDiskSegments()
{
	if(UpdateControl(Entry("disksegments"), m_Object->m_DiskSegments, k3dIChannel::DATA_SCALAR, 3, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnDiskPower()
{
	if(UpdateControl(Entry("diskpower"), m_Object->m_DiskPower, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnConeLength()
{
	if(UpdateControl(Entry("conelength"), m_Object->m_ConeLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnConeRadius()
{
	if(UpdateControl(Entry("coneradius"), m_Object->m_ConeRadius, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnConeLengthPower()
{
	if(UpdateControl(Entry("conelengthpower"), m_Object->m_ConeLengthPower, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnConeRadialPower()
{
	if(UpdateControl(Entry("coneradialpower"), m_Object->m_ConeRadialPower, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnConeLengthSegments()
{
	if(UpdateControl(Entry("conelengthsegments"), m_Object->m_ConeLengthSegments, k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnConeRadialSegments()
{
	if(UpdateControl(Entry("coneradialsegments"), m_Object->m_ConeRadialSegments, k3dIChannel::DATA_SCALAR, 3, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnConeCreateBase()
{
//	UpdateControl(Entry("conecreatebase"), m_Object->m_ConeCreateBase, k3dIChannel::DATA_DISTANCE, 3))
}

void CPrimitivesDialog::OnCylinderLength()
{
	if(UpdateControl(Entry("cylinderlength"), m_Object->m_CylinderLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnCylinderRadius()
{
	if(UpdateControl(Entry("cylinderradius"), m_Object->m_CylinderRadius, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnCylinderRadialPower()
{
	if(UpdateControl(Entry("cylinderradialpower"), m_Object->m_CylinderRadialPower, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnCylinderLengthSegments()
{
	if(UpdateControl(Entry("cylinderlengthsegments"), m_Object->m_CylinderLengthSegments, k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnCylinderRadialSegments()
{
	if(UpdateControl(Entry("cylinderradialsegments"), m_Object->m_CylinderRadialSegments, k3dIChannel::DATA_SCALAR, 3, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnCylinderCreateBase()
{
//	UpdateControl(Entry("cylindercreatebase"), m_Object->m_CylinderCreateBase, k3dIChannel::DATA_DISTANCE, 3))
//	m_Object->OnModify(false);
}

void CPrimitivesDialog::OnCubeXLength()
{
	if(UpdateControl(Entry("cubexlength"), m_Object->m_CubeXLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnCubeYLength()
{
	if(UpdateControl(Entry("cubeylength"), m_Object->m_CubeYLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnCubeZLength()
{
	if(UpdateControl(Entry("cubezlength"), m_Object->m_CubeZLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnCubeXSegments()
{
	if(UpdateControl(Entry("cubexsegments"), m_Object->m_CubeXSegments, k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnCubeYSegments()
{
	if(UpdateControl(Entry("cubeysegments"), m_Object->m_CubeYSegments, k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnCubeZSegments()
{
	if(UpdateControl(Entry("cubezsegments"), m_Object->m_CubeZSegments, k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnGridXLength()
{
	if(UpdateControl(Entry("gridxlength"), m_Object->m_GridXLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnGridYLength()
{
	if(UpdateControl(Entry("gridylength"), m_Object->m_GridYLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnGridXSegments()
{
	if(UpdateControl(Entry("gridxsegments"), m_Object->m_GridXSegments, k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnGridYSegments()
{
	if(UpdateControl(Entry("gridysegments"), m_Object->m_GridYSegments, k3dIChannel::DATA_SCALAR, 1, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnSphereXLength()
{
	if(UpdateControl(Entry("spherexlength"), m_Object->m_SphereXLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnSphereYLength()
{
	if(UpdateControl(Entry("sphereylength"), m_Object->m_SphereYLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnSphereZLength()
{
	if(UpdateControl(Entry("spherezlength"), m_Object->m_SphereZLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnSphereLengthPower()
{
	if(UpdateControl(Entry("spherelengthpower"), m_Object->m_SphereLengthPower, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnSphereRadialPower()
{
	if(UpdateControl(Entry("sphereradialpower"), m_Object->m_SphereRadialPower, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnSphereLengthSegments()
{
	if(UpdateControl(Entry("spherelengthsegments"), m_Object->m_SphereLengthSegments, k3dIChannel::DATA_SCALAR, 2, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnSphereRadialSegments()
{
	if(UpdateControl(Entry("sphereradialsegments"), m_Object->m_SphereRadialSegments, k3dIChannel::DATA_SCALAR, 3, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnTorusInnerRadius()
{
	if(UpdateControl(Entry("torusinnerradius"), m_Object->m_TorusInnerRadius, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnTorusOuterRadius()
{
	if(UpdateControl(Entry("torusouterradius"), m_Object->m_TorusOuterRadius, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnTorusRadialPower()
{
	if(UpdateControl(Entry("torusradialpower"), m_Object->m_TorusRadialPower, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnTorusCrossSectionPower()
{
	if(UpdateControl(Entry("toruscrosssectionpower"), m_Object->m_TorusCrossSectionPower, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnTorusRadialSegments()
{
	if(UpdateControl(Entry("torusradialsegments"), m_Object->m_TorusRadialSegments, k3dIChannel::DATA_SCALAR, 3, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnTorusCrossSectionSegments()
{
	if(UpdateControl(Entry("toruscrosssectionsegments"), m_Object->m_TorusCrossSectionSegments, k3dIChannel::DATA_SCALAR, 3, ULONG_MAX))
		{
			m_Object->m_CreateGeometry = true;
			m_Object->OnModify(false);
		}
}

void CPrimitivesDialog::OnPolyhedronLength()
{
	if(UpdateControl(Entry("polyhedronlength"), m_Object->m_PolyhedronLength, k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(false);
}

void CPrimitivesDialog::OnPolyhedronChoice(sdpGtkEvent* Event)
{
	sdpGtkEventCListSelectRow* event = static_cast<sdpGtkEventCListSelectRow*>(Event);

	guint row = event->Row();

	m_Object->m_PolyhedronNumber = row / 2 + 1;
	m_Object->m_PolyhedronDualCheck = (row % 2 == 1);
	m_Object->m_CreateGeometry = true;
	m_Object->OnModify(false);
}

/////////////////////////////////////////////////////////////////////////////
// CPrimitives

CPrimitives::CPrimitives()
{
	// Disk data ...
	m_DiskType = DISK;
	m_DiskRadius = 5.0;
	m_DiskSegments = 32;
	m_DiskPower = 1.0;

	// Cone data ...
	m_ConeLength = 10.0;
	m_ConeRadius = 5.0;
	m_ConeLengthPower = 1.0;
	m_ConeRadialPower = 1.0;
	m_ConeLengthSegments = 1;
	m_ConeRadialSegments = 32;
	m_ConeCreateBase = true;

	// Cylinder data ...
	m_CylinderLength = 10.0;
	m_CylinderRadius = 2.5;
	m_CylinderRadialPower = 1.0;
	m_CylinderLengthSegments = 1;
	m_CylinderRadialSegments = 32;
	m_CylinderCreateBase = true;

	// Cube data ...
	m_CubeXLength = 10.0;
	m_CubeYLength = 10.0;
	m_CubeZLength = 10.0;
	m_CubeXSegments = 5;
	m_CubeYSegments = 5;
	m_CubeZSegments = 5;

	// Grid data ...
	m_GridXLength = 10.0;
	m_GridYLength = 10.0;
	m_GridXSegments = 5;
	m_GridYSegments = 5;

	// Sphere data ...
	m_SphereXLength = 10.0;
	m_SphereYLength = 10.0;
	m_SphereZLength = 10.0;
	m_SphereLengthPower = 1.0;
	m_SphereRadialPower = 1.0;
	m_SphereLengthSegments = 12;
	m_SphereRadialSegments = 32;

	// Torus data ...
	m_TorusInnerRadius = 2.5;
	m_TorusOuterRadius = 5.0;
	m_TorusRadialPower = 1.0;
	m_TorusCrossSectionPower = 1.0;
	m_TorusRadialSegments = 32;
	m_TorusCrossSectionSegments = 12;

	// Polyhedron data ...
	m_PolyhedronNumber = 1;
	m_PolyhedronDualCheck = false;
	m_PolyhedronLength = 10.0;
}

void CPrimitives::OnModify(bool SaveAction)
{
	// Remove old geometry as necessary ...
	if(m_CreateGeometry)
		DeleteAll();

	// Update, based on which tab is active ...
	switch(m_UserInterface.m_CurrentPage)
		{
			case 0:
				OnModifyDisk(SaveAction);
				break;
			case 1:
				OnModifyCone(SaveAction);
				break;
			case 2:
				OnModifyCylinder(SaveAction);
				break;
			case 3:
				OnModifyCube(SaveAction);
				break;
			case 4:
				OnModifyGrid(SaveAction);
				break;
			case 5:
				OnModifySphere(SaveAction);
				break;
			case 6:
				OnModifyTorus(SaveAction);
				break;
			case 7:
				OnModifyPolyhedron(SaveAction);
				break;
			default:
				g_assert_not_reached();	// Did you add a new tab without updating this code?
		}

	// Reset geometry flag
	m_CreateGeometry = false;

	// Update statistics in our dialog ...
	DisplayStatistics(m_UserInterface.Entry("stats"));

	// Update the display ...
	k3dPrimTool::OnModify(SaveAction);
}


// Calculate coordinates of a single point on a disk ...
void CPrimitives::DiskCoords(gulong Index, sdpVector3& Results)
{
	// Calculate point positions ...
	double theta = double(Index) * sdpPiTimes2 / double(m_DiskSegments);

	g_assert(m_DiskPower);
	double power = 1.0 / m_DiskPower;

	Results[0] = cos(theta);
	Results[1] = sin(theta);

	Results[0] = (sdpSign(Results[0]) * pow(fabs(Results[0]), power));
	Results[1] = (sdpSign(Results[1]) * pow(fabs(Results[1]), power));
	
	Results[2] = 0.0;
}

// Calculate texture coordinates of a single point on a disk ...
void CPrimitives::DiskTextureCoords(gulong Index, sdpVector3& Results)
{
	DiskCoords(Index, Results);
	Results[0] = (Results[0] * 0.5) + 0.5;
	Results[1] = 0.5 - (Results[1] * 0.5);
}

// Create a disk ...
void CPrimitives::OnModifyDisk(bool SaveAction)
{
	// Calculate number of points / polys
	g_assert(m_DiskSegments > 2);

	gulong numpoints;
	gulong numpolys;

	if(TRIANGLES == m_DiskType)
		{
			numpoints = m_DiskSegments + 1;
			numpolys = m_DiskSegments;
		}
	else if(DISK == m_DiskType)
		{
			numpoints = m_DiskSegments;
			numpolys = 1;
		}
	else
		g_assert_not_reached();	// Did you add a new disk type?

	// Create geometry as needed ...
	if(m_CreateGeometry)
		{
			if(SaveAction)
				Document()->Actions()->StartAction("Disk");

			// Create a mesh ...
			CreateMesh("Disk", numpoints, numpolys);

			if(SaveAction)
				Document()->Actions()->FinishAction();

			// Attach points to polys & generate texture coordinates ...
			sdpVector3 texturecoords;
			k3dIPolyPoint* polypoint;
			if(TRIANGLES == m_DiskType)
				{
					for(gulong i = 0; i < m_DiskSegments; i++)
						{
							polypoint = m_Polygons[0][i]->Points()->AddPoint(m_Points[0][i]);
							DiskTextureCoords(i, texturecoords);
							polypoint->Location()->SetTextureXYZ(texturecoords);
							polypoint->Location()->SetImplicitTextureXYZ(texturecoords);
							
							polypoint = m_Polygons[0][i]->Points()->AddPoint(m_Points[0][(i+1) % m_DiskSegments]);
							DiskTextureCoords((i+1)%m_DiskSegments, texturecoords);
							polypoint->Location()->SetTextureXYZ(texturecoords);
							polypoint->Location()->SetImplicitTextureXYZ(texturecoords);
							
							polypoint = m_Polygons[0][i]->Points()->AddPoint(m_Points[0][m_DiskSegments]);
							texturecoords = sdpVector3(0.5, 0.5, 0.0);
							polypoint->Location()->SetTextureXYZ(texturecoords);
							polypoint->Location()->SetImplicitTextureXYZ(texturecoords);
						}
				}
			else if(DISK == m_DiskType)
				{
					for(gulong i = 0; i < m_DiskSegments; i++)
						{
							polypoint = m_Polygons[0][0]->Points()->AddPoint(m_Points[0][i]);
							DiskTextureCoords(i, texturecoords);
							polypoint->Location()->SetTextureXYZ(texturecoords);
							polypoint->Location()->SetImplicitTextureXYZ(texturecoords);
						}
				}
			else
				g_assert_not_reached();	// Did you add a new disk type?
		}

	// Calculate point positions ...
	sdpVector3 coords;
	for(gulong i = 0; i < m_DiskSegments; i++)
		{
			DiskCoords(i, coords);
			coords *= m_DiskRadius;

			m_Points[0][i]->Location()->SetOldXYZ(coords);
			m_Points[0][i]->Location()->SetLocalXYZ(coords);
			
		}
}


// Create a cone ...
void CPrimitives::OnModifyCone(bool SaveAction)
{
	// Sanity checks ...
	g_assert(m_ConeRadialSegments > 2);
	g_assert(m_ConeLengthSegments > 0);

	// Create geometry as needed ...
	if(m_CreateGeometry)
		{
			if(SaveAction)
				Document()->Actions()->StartAction("Cone");

			// Create a mesh ...
			CreateMesh("Cone", m_ConeLengthSegments, m_ConeRadialSegments, false, true);

			if(SaveAction)
				Document()->Actions()->FinishAction();

/*
			// Create the base as needed ...
			if(m_CreateBase)
				{
					// Create a polygon ...
					k3dIPolygon* poly = CreatePolygon(mesh);
					g_assert(poly);

					// Attach it to the base of the cone ...
					sdpVector3 texturecoords(0.0);
					for(gulong i = 0; i <= m_RadialSegments; i++)
						{
							int index = GridPointIndex(i, m_LengthSegments + 1, m_RadialSegments + 1, m_LengthSegments + 1, FALSE, TRUE);
							k3dIPolyhedronPoint* polypoint = poly->Points()->GetPolyhedronPoint(poly->Points()->AddPoint(m_Points[index]));
							polypoint->Location()->SetTextureXYZ(texturecoords);
							polypoint->Location()->SetImplicitTextureXYZ(texturecoords);
						}
				}
*/
		}

	// Calculate point positions ...
	double dtheta = sdpPiTimes2 / double(m_ConeRadialSegments);
	double dlength = m_ConeLength / double(m_ConeLengthSegments);
	double lengthoffset = m_ConeLength * 0.5;

	double radius;
	double theta;
	sdpVector3 coords;
	sdpMatrix4 matrix;

	g_assert(m_ConeLengthPower);
	double lengthpower = 1.0 / m_ConeLengthPower;

	g_assert(m_ConeRadialPower);
	double radialpower = 1.0 / m_ConeRadialPower;

	for(gulong row = 0; row <= m_ConeLengthSegments; row++)
		{
			radius = m_ConeRadius * pow(double(row) / (double)m_ConeLengthSegments, lengthpower);

			for(gulong col = 0; col < m_ConeRadialSegments; col++)
				{
					theta = sdpPiTimes2 - (double(col) * dtheta);
					matrix = sdpIdentity3D() * sdpRotation3D(theta, sdpVector3(0.0, 1.0, 0.0));
					coords = matrix * sdpVector3(0.0, 0.0, 1.0);
					coords[0] = radius * sdpSign(coords[0]) * pow(fabs(coords[0]), radialpower);
					coords[1] = lengthoffset - (double(row) * dlength);
					coords[2] = radius * sdpSign(coords[2]) * pow(fabs(coords[2]), radialpower);

					m_Points[row][col]->Location()->SetOldXYZ(coords);
					m_Points[row][col]->Location()->SetLocalXYZ(coords);
				}
		}
}


// Create a cylinder ...
void CPrimitives::OnModifyCylinder(bool SaveAction)
{
	// Sanity checks ...
	g_assert(m_CylinderRadialSegments > 2);
	g_assert(m_CylinderLengthSegments > 0);

	// Create geometry as needed ...
	if(m_CreateGeometry)
		{
			if(SaveAction)
				Document()->Actions()->StartAction("Cylinder");

			// Create a mesh ...
			CreateMesh("Cylinder", m_CylinderLengthSegments, m_CylinderRadialSegments, false, true);

			if(SaveAction)
				Document()->Actions()->FinishAction();

/*
			// Create endcaps as needed ...
			if(m_CreateEndcaps)
				{
					for(int j = 0; j < 2; j++)
						{
							// Create a polygon ...
							k3dIPolygon* poly = CreatePolygon(mesh);
							g_assert(poly);

							// Attach it to the base of the cylinder ...
							sdpVector3 texturecoords(0.0);
							for(gulong i = 0; i <= m_RadialSegments; i++)
								{
									int index = GridPointIndex(i, j * (m_LengthSegments + 1), m_RadialSegments + 1, m_LengthSegments + 1, FALSE, TRUE);
									k3dIPolyhedronPoint* polypoint = poly->Points()->GetPolyhedronPoint(poly->Points()->AddPoint(m_Points[index]));
									polypoint->Location()->SetTextureXYZ(texturecoords);
									polypoint->Location()->SetImplicitTextureXYZ(texturecoords);
								}
						}
				}
*/
		}

	// Calculate point positions ...
	double dtheta = sdpPiTimes2 / double(m_CylinderRadialSegments);
	double dlength = m_CylinderLength / double(m_CylinderLengthSegments);
	double lengthoffset = m_CylinderLength * 0.5;

	sdpVector3 coords;
	sdpMatrix4 matrix;
	double theta;

	g_assert(m_CylinderRadialPower);
	double radialpower = 1.0 / m_CylinderRadialPower;

	double radius = m_CylinderRadius;

	for(gulong row = 0; row <= m_CylinderLengthSegments; row++)
		{
			for(gulong col = 0; col < m_CylinderRadialSegments; col++)
				{
					theta = sdpPiTimes2 - (double(col) * dtheta);
					matrix = sdpIdentity3D() * sdpRotation3D(theta, sdpVector3(0.0, 1.0, 0.0));
					coords = matrix * sdpVector3(0.0, 0.0, 1.0);
					coords[0] = radius * sdpSign(coords[0]) * pow(fabs(coords[0]), radialpower);
					coords[1] = lengthoffset - (double(row) * dlength);
					coords[2] = radius * sdpSign(coords[2]) * pow(fabs(coords[2]), radialpower);

					m_Points[row][col]->Location()->SetOldXYZ(coords);
					m_Points[row][col]->Location()->SetLocalXYZ(coords);
				}
		}
}

// Convert an ordered triple into a string for use with our cube point map ...
sdpString CPrimitives::CubePointString(gulong X, gulong Y, gulong Z)
{
	return sdpString(sdpVector3(X, Y, Z));
}

// Store the point at a given map location ...
void CPrimitives::SetCubePoint(gulong X, gulong Y, gulong Z, k3dIPoint* Point)
{
	// Sanity checks ...
	g_assert(Point);

	m_CubeMap[CubePointString(X, Y, Z)] = Point;
}

// Get the point from the given map location ...
k3dIPoint* CPrimitives::GetCubePoint(gulong X, gulong Y, gulong Z)
{
	// Make sure its been initialized ...
	CubeMapIterator i = m_CubeMap.find(CubePointString(X, Y, Z));
	g_assert(i != m_CubeMap.end());

	return i->second;
}

void CPrimitives::AddCubePoint(gulong PolyIndex, gulong X, gulong Y, gulong Z, gdouble S, gdouble T)
{
	// Attach the point to the polygon ...
	k3dIPolyPoint* polypoint = m_Polygons[0][PolyIndex]->Points()->AddPoint(GetCubePoint(X, Y, Z));
	g_assert(polypoint);
	
	// Store texture coordinates
	sdpVector3 texturecoords(S, T, 0.0);
	polypoint->Location()->SetTextureXYZ(texturecoords);
	polypoint->Location()->SetImplicitTextureXYZ(texturecoords);
}

// Return the number of points in the cube ...
gulong CPrimitives::CubePointCount()
{
	// Initialize the count ...
	gulong count = 0;

	for(gulong x = 0; x <= m_CubeXSegments; x++)
		{
			for(gulong y = 0; y <= m_CubeYSegments; y++)
				{
					for(gulong z = 0; z <= m_CubeZSegments; z++)
						{
							// Skip interior coordinates ...
							bool xface = (0 == x) || (m_CubeXSegments == x);
							bool yface = (0 == y) || (m_CubeYSegments == y);
							bool zface = (0 == z) || (m_CubeZSegments == z);

							if(!xface && !yface && !zface)
								continue;

							// Increment the count ...
							count++;
						}
				}
		}

	return count;
}

// Return the number of polygons in the cube ...
gulong CPrimitives::CubePolyCount()
{
	return (m_CubeXSegments * m_CubeYSegments * 2) + (m_CubeYSegments * m_CubeZSegments * 2) + (m_CubeXSegments * m_CubeZSegments * 2);
}

// Create cube geometry ...
void CPrimitives::CreateCubeGeometry(bool SaveAction)
{
	if(SaveAction)
		Document()->Actions()->StartAction("Cube");

	// Create the mesh ...
	CreateMesh("Cube", CubePointCount(), CubePolyCount());

	if(SaveAction)
		Document()->Actions()->FinishAction();

	// Initialize our mapping of point coordinates to points ...
	gulong pointindex = 0;
	gulong x, y, z;

	for(x = 0; x <= m_CubeXSegments; x++)
		{
			for(y = 0; y <= m_CubeYSegments; y++)
				{
					for(z = 0; z <= m_CubeZSegments; z++)
						{
							// Skip interior coordinates ...
							bool xface = (0 == x) || (m_CubeXSegments == x);
							bool yface = (0 == y) || (m_CubeYSegments == y);
							bool zface = (0 == z) || (m_CubeZSegments == z);

							if(!xface && !yface && !zface)
								continue;

							SetCubePoint(x, y, z, m_Points[0][pointindex++]);
						}
				}
		}

	// Attach points to polygons ...
	gulong polyindex = 0;

	// Front face ...
	z = 0;
	for(y = 0; y < m_CubeYSegments; y++)
		{
			for(x = 0; x < m_CubeXSegments; x++)
				{
					AddCubePoint(polyindex, x, y, z, gdouble(x) / gdouble(m_CubeXSegments), 1.0 - (gdouble(y) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x+1, y, z, gdouble(x+1) / gdouble(m_CubeXSegments), 1.0 - (gdouble(y) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x+1, y+1, z, gdouble(x+1) / gdouble(m_CubeXSegments), 1.0 - (gdouble(y+1) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y+1, z, gdouble(x) / gdouble(m_CubeXSegments), 1.0 - (gdouble(y+1) / gdouble(m_CubeYSegments)));
					
					polyindex++;
				}
		}

	// Back face ...
	z = m_CubeZSegments;
	for(y = 0; y < m_CubeYSegments; y++)
		{
			for(x = 0; x < m_CubeXSegments; x++)
				{
					AddCubePoint(polyindex, x+1, y, z, 1.0 - (gdouble(x+1) / gdouble(m_CubeXSegments)), 1.0 - (gdouble(y) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y, z, 1.0 - (gdouble(x) / gdouble(m_CubeXSegments)), 1.0 - (gdouble(y) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y+1, z, 1.0 - (gdouble(x) / gdouble(m_CubeXSegments)), 1.0 - (gdouble(y+1) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x+1, y+1, z, 1.0 - (gdouble(x+1) / gdouble(m_CubeXSegments)), 1.0 - (gdouble(y+1) / gdouble(m_CubeYSegments)));

					polyindex++;
				}
		}

	// Left face ...
	x = 0;
	for(y = 0; y < m_CubeYSegments; y++)
		{
			for(z = 0; z < m_CubeZSegments; z++)
				{
					AddCubePoint(polyindex, x, y, z+1, 1.0 - (gdouble(z+1) / gdouble(m_CubeZSegments)), 1.0 - (gdouble(y) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y, z, 1.0 - (gdouble(z) / gdouble(m_CubeZSegments)), 1.0 - (gdouble(y) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y+1, z, 1.0 - (gdouble(z) / gdouble(m_CubeZSegments)), 1.0 - (gdouble(y+1) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y+1, z+1, 1.0 - (gdouble(z+1) / gdouble(m_CubeZSegments)), 1.0 - (gdouble(y+1) / gdouble(m_CubeYSegments)));

					polyindex++;
				}
		}

	// Right face ...
	x = m_CubeXSegments;
	for(y = 0; y < m_CubeYSegments; y++)
		{
			for(z = 0; z < m_CubeZSegments; z++)
				{
					AddCubePoint(polyindex, x, y, z, gdouble(z) / gdouble(m_CubeZSegments), 1.0 - (gdouble(y) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y, z+1, gdouble(z+1) / gdouble(m_CubeZSegments), 1.0 - (gdouble(y) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y+1, z+1, gdouble(z+1) / gdouble(m_CubeZSegments), 1.0 - (gdouble(y+1) / gdouble(m_CubeYSegments)));
					AddCubePoint(polyindex, x, y+1, z, gdouble(z) / gdouble(m_CubeZSegments), 1.0 - (gdouble(y+1) / gdouble(m_CubeYSegments)));

					polyindex++;
				}
		}

	// Top face ...
	y = 0;
	for(z = 0; z < m_CubeZSegments; z++)
		{
			for(x = 0; x < m_CubeXSegments; x++)
				{
					AddCubePoint(polyindex, x+1, y, z, gdouble(x+1) / gdouble(m_CubeXSegments), gdouble(z) / gdouble(m_CubeZSegments));
					AddCubePoint(polyindex, x, y, z, gdouble(x) / gdouble(m_CubeXSegments), gdouble(z) / gdouble(m_CubeZSegments));
					AddCubePoint(polyindex, x, y, z+1, gdouble(x) / gdouble(m_CubeXSegments), gdouble(z+1) / gdouble(m_CubeZSegments));
					AddCubePoint(polyindex, x+1, y, z+1, gdouble(x+1) / gdouble(m_CubeXSegments), gdouble(z+1) / gdouble(m_CubeZSegments));

					polyindex++;
				}
		}

	// Bottom face ...
	y = m_CubeYSegments;
	for(z = 0; z < m_CubeZSegments; z++)
		{
			for(x = 0; x < m_CubeXSegments; x++)
				{
					AddCubePoint(polyindex, x, y, z, gdouble(x) / gdouble(m_CubeXSegments), 1.0 - (gdouble(z) / gdouble(m_CubeZSegments)));
					AddCubePoint(polyindex, x+1, y, z, gdouble(x+1) / gdouble(m_CubeXSegments), 1.0 - (gdouble(z) / gdouble(m_CubeZSegments)));
					AddCubePoint(polyindex, x+1, y, z+1, gdouble(x+1) / gdouble(m_CubeXSegments), 1.0 - (gdouble(z+1) / gdouble(m_CubeZSegments)));
					AddCubePoint(polyindex, x, y, z+1, gdouble(x) / gdouble(m_CubeXSegments), 1.0 - (gdouble(z+1) / gdouble(m_CubeZSegments)));

					polyindex++;
				}
		}
}

// Create a cube ...
void CPrimitives::OnModifyCube(bool SaveAction)
{
	// Sanity checks ...
	g_assert(m_CubeXSegments > 0);
	g_assert(m_CubeYSegments > 0);
	g_assert(m_CubeZSegments > 0);

	// Create geometry as needed ...
	if(m_CreateGeometry)
		CreateCubeGeometry(SaveAction);

	// Calculate point positions ...
	sdpVector3 coords;
	sdpVector3 offset = sdpVector3(m_CubeXLength, m_CubeYLength, m_CubeZLength) * -0.5;

	for(gulong x = 0; x <= m_CubeXSegments; x++)
		{
			for(gulong y = 0; y <= m_CubeYSegments; y++)
				{
					for(gulong z = 0; z <= m_CubeZSegments; z++)
						{
							// Skip interior coordinates ...
							bool xface = (0 == x) || (m_CubeXSegments == x);
							bool yface = (0 == y) || (m_CubeYSegments == y);
							bool zface = (0 == z) || (m_CubeZSegments == z);

							if(!xface && !yface && !zface)
								continue;

							// Calculate the point coordinates ...
							coords[0] = m_CubeXLength * double(x) / double(m_CubeXSegments);
							coords[1] =	m_CubeYLength * double(y) / double(m_CubeYSegments);
							coords[2] = m_CubeZLength * double(z) / double(m_CubeZSegments);
							coords += offset;

							// Assign 'em ...
							k3dIPoint* point = GetCubePoint(x, y, z);
							g_assert(point);
							point->Location()->SetOldXYZ(coords);
							point->Location()->SetLocalXYZ(coords);
						}
				}
		}
}

// Create a grid ...
void CPrimitives::OnModifyGrid(bool SaveAction)
{
	// Calculate number of points / polys
	g_assert(m_GridXSegments > 0);
	g_assert(m_GridYSegments > 0);

	// Create geometry as needed ...
	if(m_CreateGeometry)
		{
			if(SaveAction)
				Document()->Actions()->StartAction("Grid");

			// Create the mesh ...
			CreateMesh("Grid", m_GridYSegments, m_GridXSegments, false, false);

			if(SaveAction)
				Document()->Actions()->FinishAction();
		}

	// Calculate point positions ...
	sdpVector3 coords;
	sdpVector2 offset(m_GridXLength * 0.5, m_GridYLength * 0.5);

	for(gulong y = 0; y <= m_GridYSegments; y++)
		{
			for(gulong x = 0; x <= m_GridXSegments; x++)
				{
					coords[0] = (m_GridXLength * double(x) / double(m_GridXSegments)) - offset[0];
					coords[1] = offset[1] - (m_GridYLength * double(y) / double(m_GridYSegments));
					coords[2] = 0.0;

					m_Points[y][x]->Location()->SetOldXYZ(coords);
					m_Points[y][x]->Location()->SetLocalXYZ(coords);
				}
		}
}

// Create a sphere ...
void CPrimitives::OnModifySphere(bool SaveAction)
{
	// Calculate number of points / polys
	g_assert(m_SphereRadialSegments > 2);
	g_assert(m_SphereLengthSegments > 0);

	// Create geometry as needed ...
	if(m_CreateGeometry)
		{
			if(SaveAction)
				Document()->Actions()->StartAction("Sphere");

			// Create a mesh ...
			CreateMesh("Sphere", m_SphereLengthSegments, m_SphereRadialSegments, false, true);

			if(SaveAction)
				Document()->Actions()->FinishAction();
		}

	// Calculate point positions ...
	double dtheta = sdpPiTimes2 / double(m_SphereRadialSegments);
	double dphi = sdpPi / double(m_SphereLengthSegments);

	sdpVector3 coords1, coords2;
	sdpMatrix4 matrix1, matrix2;
	double theta, phi;

	g_assert(m_SphereLengthPower);
	double lengthpower = 1.0 / m_SphereLengthPower;

	g_assert(m_SphereRadialPower);
	double radialpower = 1.0 / m_SphereRadialPower;

	sdpVector3 scale = sdpVector3(m_SphereXLength, m_SphereYLength, m_SphereZLength) * 0.5;	// Convert diameters into radii

	// Calculate coordinates for the poles ...
	coords1 = sdpVector3(0.0, scale[1], 0.0);
	coords2 = sdpVector3(0.0, -scale[1], 0.0);
	{
		for(gulong col = 0; col < m_SphereRadialSegments; col++)
			{
				m_Points[0][MeshColumn(col, m_SphereRadialSegments, true)]->Location()->SetOldXYZ(coords1);
				m_Points[0][MeshColumn(col, m_SphereRadialSegments, true)]->Location()->SetLocalXYZ(coords1);

				m_Points[m_SphereLengthSegments][MeshColumn(col, m_SphereRadialSegments, true)]->Location()->SetOldXYZ(coords2);
				m_Points[m_SphereLengthSegments][MeshColumn(col, m_SphereRadialSegments, true)]->Location()->SetLocalXYZ(coords2);
			}
	}

	// Coordinates for everyone else ...
	for(gulong row = 1; row < m_SphereLengthSegments; row++)
		{
			phi = -sdpPiOver2 + (double(row) * dphi);
			matrix1 = sdpIdentity3D() * sdpRotation3D(phi, sdpVector3(1.0, 0.0, 0.0));
			coords1 = matrix1 * sdpVector3(0.0, 0.0, 1.0);
			coords1[2] = pow(coords1[2], lengthpower);

			for(gulong col = 0; col < m_SphereRadialSegments; col++)
				{
					theta = sdpPiTimes2 - (double(col) * dtheta);
					matrix2 = sdpIdentity3D() * sdpRotation3D(theta, sdpVector3(0.0, 1.0, 0.0));
					coords2 = matrix2 * coords1;
					coords2[0] = sdpSign(coords2[0]) * pow(fabs(coords2[0]), radialpower);
					coords2[2] = sdpSign(coords2[2]) * pow(fabs(coords2[2]), radialpower);

					coords2 = sdpProduct(coords2, scale);

					m_Points[row][MeshColumn(col, m_SphereRadialSegments, true)]->Location()->SetOldXYZ(coords2);
					m_Points[row][MeshColumn(col, m_SphereRadialSegments, true)]->Location()->SetLocalXYZ(coords2);
				}
		}
}

// Create a torus ...
void CPrimitives::OnModifyTorus(bool SaveAction)
{
	// Calculate number of points / polys
	g_assert(m_TorusRadialSegments > 2);
	g_assert(m_TorusCrossSectionSegments > 2);

	// Create geometry as needed ...
	if(m_CreateGeometry)
		{
			if(SaveAction)
				Document()->Actions()->StartAction("Torus");

			// Create a mesh ...
			CreateMesh("Torus", m_TorusCrossSectionSegments, m_TorusRadialSegments, true, true);

			if(SaveAction)
				Document()->Actions()->FinishAction();
		}

	// Calculate point positions ...
	double dtheta = sdpPiTimes2 / double(m_TorusRadialSegments);
	double dphi = sdpPiTimes2 / double(m_TorusCrossSectionSegments);

	double radius1 = (m_TorusInnerRadius + m_TorusOuterRadius) * 0.5;
	double radius2 = (m_TorusOuterRadius - m_TorusInnerRadius) * 0.5;

	sdpVector3 coords1, coords2;
	sdpMatrix4 matrix1, matrix2;
	double theta, phi;

	g_assert(m_TorusCrossSectionPower);
	double lengthpower = 1.0 / m_TorusCrossSectionPower;

	g_assert(m_TorusRadialPower);
	double radialpower = 1.0 / m_TorusRadialPower;

	for(gulong row = 0; row <= m_TorusCrossSectionSegments; row++)
		{
			phi = double(row) * dphi;
			matrix1 = sdpIdentity3D() * sdpRotation3D(phi, sdpVector3(1.0, 0.0, 0.0));
			coords1 = matrix1 * sdpVector3(0.0, 0.0, -1.0);
			coords1[2] = pow(coords1[2], lengthpower);
			coords1 *= radius2;

			for(gulong col = 0; col < m_TorusRadialSegments; col++)
				{
					theta = sdpPiTimes2 - (double(col) * dtheta);
					matrix2 = sdpIdentity3D() * sdpRotation3D(theta, sdpVector3(0.0, 1.0, 0.0));
					coords2 = matrix2 * sdpVector3(0.0, 0.0, 1.0);
					coords2[0] = sdpSign(coords2[0]) * pow(fabs(coords2[0]), radialpower);
					coords2[2] = sdpSign(coords2[2]) * pow(fabs(coords2[2]), radialpower);
					matrix2 = matrix2 * sdpTranslation3D(sdpVector3(0.0, 0.0, coords2.Length() * radius1));

					coords2 = matrix2 * coords1;

					m_Points[MeshRow(row, m_TorusCrossSectionSegments, true)][MeshColumn(col, m_TorusRadialSegments, true)]->Location()->SetOldXYZ(coords2);
					m_Points[MeshRow(row, m_TorusCrossSectionSegments, true)][MeshColumn(col, m_TorusRadialSegments, true)]->Location()->SetLocalXYZ(coords2);
				}
		}
}

void CPrimitives::LoadPolyhedron(int PolyhedronIndex, bool Dual )
{
	// Clear out old data ...
	m_PolyhedronPoints.clear();
	m_PolyhedronPolygons.clear();
	m_PolyhedronScaling = 0.5;

	// Solve polyhedron (need_coordinates = 1, need_edgelist = 0, need_approx = 0, just_list = 0)
	Polyhedron* polyhedron = kaleido(const_cast<char*>(SDPCSTRING(sdpFormatString("#%d", PolyhedronIndex))), 1, 0, 0, 0 );
	g_assert(polyhedron);

	double freeze = 0;
	double azimuth = 0;
	double elevation = 0;

	if(!Dual)
		{
			vrmodel(polyhedron, polyhedron->v, polyhedron->V, polyhedron->f, polyhedron->F, polyhedron->name, "", azimuth, elevation, freeze);
		}
	else
		{
			vrmodel(polyhedron, polyhedron->f, polyhedron->F, polyhedron->v, polyhedron->V, polyhedron->dual_name, "*", azimuth, elevation, freeze);
		}

	// Reverse the polygon orientations (kaleido must be a RH coordinate system!?)
	for(gulong i = 0; i < m_PolyhedronPolygons.size(); i++)
		std::reverse(m_PolyhedronPolygons[i].begin(), m_PolyhedronPolygons[i].end());

	// Write the polyhedron geometric properties string
	m_PolyhedronProperties = "";
	if(Dual && (polyhedron->hemi || polyhedron->index == last_uniform - 1))
		{
			m_PolyhedronProperties += "infinite dual";

			// Infinite duals have to scaled to a unit sphere
			double maxlength = 0.0;
			for(gulong i = 0; i < m_PolyhedronPoints.size(); i++)
				{
					double length = m_PolyhedronPoints[i].Length();
					if(length > maxlength)
						maxlength = length;
				}
			m_PolyhedronScaling /= maxlength;
		}


	// Free the polyhedron data ...
	polyfree(polyhedron);
}

void CPrimitives::PolyhedronPoint(Kvector v)
{
	m_PolyhedronPoints.push_back(sdpVector3(v.x, v.y, v.z));
}

void CPrimitives::PolyhedronFace()
{
	m_PolyhedronPolygons.push_back(PolyhedronPolygon());
}

void CPrimitives::PolyhedronPolyPoint(int PointIndex)
{
	m_PolyhedronPolygons.back().push_back(PointIndex);
}

void CPrimitives::PolyhedronTriangle(int PointIndex1, int PointIndex2, int PointIndex3)
{
	m_PolyhedronPolygons.push_back(PolyhedronPolygon());

	m_PolyhedronPolygons.back().push_back(PointIndex1);
	m_PolyhedronPolygons.back().push_back(PointIndex2);
	m_PolyhedronPolygons.back().push_back(PointIndex3);
}

void CPrimitives::PolyhedronQuad(int PointIndex1, int PointIndex2, int PointIndex3, int PointIndex4)
{
	m_PolyhedronPolygons.push_back(PolyhedronPolygon());

	m_PolyhedronPolygons.back().push_back(PointIndex1);
	m_PolyhedronPolygons.back().push_back(PointIndex2);
	m_PolyhedronPolygons.back().push_back(PointIndex3);
	m_PolyhedronPolygons.back().push_back(PointIndex4);
}

int  CPrimitives::vrmodel( Polyhedron *P, Kvector *v, int V, Kvector *f, int F,	char *name, char *star,	double azimuth, double elevation, double freeze)
{
	int i, j, k, l, ll, ii, *hit;
	Kvector *temp;

/*
* Rotate polyhedron
*/
	rotframe (azimuth, elevation, freeze);
	Malloc (temp, V, Kvector)
	rotarray (temp, v, V);
	v = temp;
	Malloc (temp, F, Kvector)
	rotarray (temp, f, F);
	f = temp;



	for (i = 0; i < V; i++)
		PolyhedronPoint( v[i] );
/*
* Auxiliary vertices (needed because current VRML browsers cannot handle
* non-simple polygons, i.e., polygons with self intersections):
* Each non-simple face is assigned an auxiliary vertex. By connecting it to
* the rest of the vertices the face is triangulated. The circum-center is used
* for the regular star faces of uniform polyhedra. The in-center is used for
* the pentagram (#79) and hexagram (#77) of the high-density snub duals, and
* for the pentagrams (#40, #58) and hexagram (#52) of the stellated duals with
* configuration (....)/2. Finally, the self-intersection of the crossed
* parallelogram is used for duals with form p q r| with an even denominator.
*
* This method do not work for the hemi-duals, whose faces are not star-shaped
* and have two self-intersections each.
* Thus, for each face we need six auxiliary vertices: The self intersections
* and the terminal points of the truncations of the infinite edges. The ideal
* vertices are listed, but are not used by the face-list.
* Note that the face of the last dual (#80) is octagonal, and constists of two
* quadrilaterals of the infinite type.
*/

	if (*star && P->even != -1)
		Malloc (hit, F, int);
	for (i = 0; i < F; i++)
		if (!*star && (frac (P->n[P->ftype[i]]),
			frax.d != 1 && frax.d != frax.n - 1) ||
			*star && (P->K == 5 && P->D > 30 ||
			denominator (P->m[0]) != 1)) {
			/* find the center of the face */
			double h;
			if (!*star && P->hemi && !P->ftype[i])
				h = 0;
			else
				h = P->minr / dot (f[i],f[i]);
			PolyhedronPoint( scale (h, f[i]) );
		} else if (*star && P->even != -1) {
			/* find the self-intersection of a crossed
			 * parallelogram. hit is set if v0v1 intersects v2v3*/
			Kvector v0, v1, v2, v3, c0, c1, p;
			double d0, d1;
			v0 = v[P->incid[0][i]];
			v1 = v[P->incid[1][i]];
			v2 = v[P->incid[2][i]];
			v3 = v[P->incid[3][i]];
			d0 = sqrt (dot (diff (v0, v2), diff (v0, v2)));
			d1 = sqrt (dot (diff (v1, v3), diff (v1, v3)));
			c0 = scale (d1, sum (v0, v2));
			c1 = scale (d0, sum (v1, v3));
			p = scale (0.5 / (d0 + d1), sum (c0, c1));
			PolyhedronPoint( p );
			p = cross (diff (p, v2), diff (p, v3));
			hit[i] = (dot (p, p) < 1e-6);
		} else if (*star && P->hemi && P->index != last_uniform - 1) {
			/* find the terminal points of the truncation and the
			* self-intersections.
			*  v23       v0       v21
			*  |  \     /  \     /  |
			*  |   v0123    v0321   |
			*  |  /     \  /     \  |
			*  v01       v2       v03
			*/
			Kvector v0, v1, v2, v3, v01, v03, v21, v23, v0123, v0321 ;
			Kvector u;
			double t = 1.5;/*truncation adjustment factor*/
			j = !P->ftype[P->incid[0][i]];
			v0 = v[P->incid[j][i]];/*real vertex*/
			v1 = v[P->incid[j+1][i]];/*ideal vertex (unit Kvector)*/
			v2 = v[P->incid[j+2][i]];/*real*/
			v3 = v[P->incid[(j+3)%4][i]];/*ideal*/
			/*compute intersections
				this uses the following linear algebra:
				v0123 = v0 + a v1 = v2 + b v3
				v0 x v3 + a (v1 x v3) = v2 x v3
				a (v1 x v3) = (v2 - v0) x v3
				a (v1 x v3) . (v1 x v3) =
					(v2 - v0) x v3 . (v1 x v3)
			*/
			u = cross (v1, v3);
			v0123 = sum (v0, scale (dot (cross (diff (v2, v0), v3),
				u) / dot (u,u), v1));
			v0321 = sum (v0, scale (dot (cross (diff (v0, v2), v1),
				u) / dot (u,u), v3));
			/*compute truncations*/
			v01 = sum (v0 , scale (t, diff (v0123, v0)));
			v23 = sum (v2 , scale (t, diff (v0123, v2)));
			v03 = sum (v0 , scale (t, diff (v0321, v0)));
			v21 = sum (v2 , scale (t, diff (v0321, v2)));
			PolyhedronPoint( v01 );
			PolyhedronPoint( v23 );
			PolyhedronPoint( v0123 );
			PolyhedronPoint( v03 );
			PolyhedronPoint( v21 );
			PolyhedronPoint( v0321 );
		} else if (*star && P->index == last_uniform - 1) {
			/* find the terminal points of the truncation and the
			* self-intersections.
			*  v23       v0       v21
			*  |  \     /  \     /  |
			*  |   v0123    v0721   |
			*  |  /     \  /     \  |
			*  v01       v2       v07
			*
			*  v65       v4       v67
			*  |  \     /  \     /  |
			*  |   v4365    v4567   |
			*  |  /     \  /     \  |
			*  v43       v6       v45
			*/
			Kvector v0, v1, v2, v3, v4, v5, v6, v7, v01, v07, v21, v23;
			Kvector v43, v45, v65, v67, v0123, v0721, v4365, v4567;
			double t = 1.5;/*truncation adjustment factor*/
			Kvector u;
			for (j = 0; j < 8; j++)
				if (P->ftype[P->incid[j][i]] == 3)
					break;
			v0 = v[P->incid[j][i]];/*real {5/3}*/
			v1 = v[P->incid[(j+1)%8][i]];/*ideal*/
			v2 = v[P->incid[(j+2)%8][i]];/*real {3}*/
			v3 = v[P->incid[(j+3)%8][i]];/*ideal*/
			v4 = v[P->incid[(j+4)%8][i]];/*real {5/2}*/
			v5 = v[P->incid[(j+5)%8][i]];/*ideal*/
			v6 = v[P->incid[(j+6)%8][i]];/*real {3/2}*/
			v7 = v[P->incid[(j+7)%8][i]];/*ideal*/
			/*compute intersections*/
			u = cross (v1, v3);
			v0123 = sum (v0, scale (dot (cross (diff (v2, v0), v3),
				u) / dot (u,u), v1));
			u = cross (v7, v1);
			v0721 = sum (v0, scale (dot (cross (diff (v2, v0), v1),
				u) / dot (u,u), v7));
			u = cross (v5, v7);
			v4567 = sum (v4, scale (dot (cross (diff (v6, v4), v7),
				u) / dot (u,u), v5));
			u = cross (v3, v5);
			v4365 = sum (v4, scale (dot (cross (diff (v6, v4), v5),
				u) / dot (u,u), v3));
			/*compute truncations*/
			v01 = sum (v0 , scale (t, diff (v0123, v0)));
			v23 = sum (v2 , scale (t, diff (v0123, v2)));
			v07 = sum (v0 , scale (t, diff (v0721, v0)));
			v21 = sum (v2 , scale (t, diff (v0721, v2)));
			v45 = sum (v4 , scale (t, diff (v4567, v4)));
			v67 = sum (v6 , scale (t, diff (v4567, v6)));
			v43 = sum (v4 , scale (t, diff (v4365, v4)));
			v65 = sum (v6 , scale (t, diff (v4365, v6)));
			PolyhedronPoint( v01 );
			PolyhedronPoint( v23 );
			PolyhedronPoint( v0123 );
			PolyhedronPoint( v07 );
			PolyhedronPoint( v21 );
			PolyhedronPoint( v0721 );
			PolyhedronPoint( v45 );
			PolyhedronPoint( v67 );
			PolyhedronPoint( v4567 );
			PolyhedronPoint( v43 );
			PolyhedronPoint( v65 );
			PolyhedronPoint( v4365 );
		}


/*
* Face list:
* Each face is printed in a separate line, by listing the indices of its
* vertices. In the non-simple case, the polygon is represented by the
* triangulation, each triangle consists of two polyhedron vertices and one
* auxiliary vertex.
*/

	ii = V;
	for (i = 0; i < F; i++) {

		PolyhedronFace();

		if (*star) {
			if (P->K == 5 && P->D > 30 || denominator (P->m[0]) != 1) {
				for (j = 0; j < P->M - 1; j++)
					PolyhedronTriangle( P->incid[j][i], P->incid[j+1][i], ii );

				PolyhedronTriangle( P->incid[j][i], P->incid[0][i], ii++ );

			} else if (P->even != -1) {
				if (hit[i])
				{
					PolyhedronTriangle( P->incid[3][i], P->incid[0][i], ii );
					PolyhedronTriangle( P->incid[1][i], P->incid[2][i], ii );
				}
				else
				{
					PolyhedronTriangle( P->incid[0][i], P->incid[1][i], ii );
					PolyhedronTriangle( P->incid[2][i], P->incid[3][i], ii );
				}
				ii++;
			} else if (P->hemi && P->index != last_uniform - 1) {
				j = !P->ftype[P->incid[0][i]];
				PolyhedronTriangle( ii, ii+1, ii+2 );
				PolyhedronQuad( P->incid[j][i], ii+2, P->incid[j+2][i], ii+5 );
				PolyhedronTriangle( ii+3, ii+4, ii+5 );
				ii += 6;
			} else if (P->index == last_uniform - 1) {
				for (j = 0; j < 8; j++)
					if (P->ftype[P->incid[j][i]] == 3)
						break;
				PolyhedronTriangle( ii, ii+1, ii+2 );
				PolyhedronQuad( P->incid[j][i], ii+2, P->incid[(j+2)%8][i], ii+5 );
				PolyhedronTriangle( ii+3, ii+4, ii+5 );

				PolyhedronTriangle( ii+6, ii+7, ii+8 );
				PolyhedronQuad( P->incid[(j+4)%8][i], ii+8, P->incid[(j+6)%8][i], ii+11 );
				PolyhedronTriangle( ii+9, ii+10, ii+11 );
				ii += 12;
			} else {
				for (j = 0; j < P->M; j++)
					PolyhedronPolyPoint( P->incid[j][i] );
			}
		} else {
			int split = (frac (P->n[P->ftype[i]]),
				 frax.d != 1 && frax.d != frax.n - 1);
			for (j = 0; j < V; j++) {
				for (k = 0; k < P->M; k++)
					if (P->incid[k][j] == i)
						break;
				if (k != P->M)
					break;
			}
			if (!split )
				PolyhedronPolyPoint( j );
			ll = j;
			for (l = P->adj[k][j]; l != j; l = P->adj[k][l]) {
				for (k = 0; k < P->M; k++)
					if (P->incid[k][l] == i)
						break;
				if (P->adj[k][l] == ll) k = mod (k + 1 , P->M);
				if (!split )
					PolyhedronPolyPoint( l );
				else
					PolyhedronTriangle( ll, l, ii );
				ll = l;
			}
			if ( split )
				PolyhedronTriangle( ll, j, ii++ );
		}

	}


/*
* Face color indices - for polyhedra with multiple face types
* For non-simple faces, the index is repeated as many times as needed by the
* triangulation.
*/
/*
	if (!*star && P->N != 1) {
		cout << "\tMaterialIndex[";
		for (i = 0; i < F; i++)
			if (frac (P->n[P->ftype[i]]),
				frax.d == 1 || frax.d == frax.n - 1)
			cout << P->ftype[i] << ",";
			else
				for (j = 0; j < frax.n; j++)
					cout << P->ftype[i] << ",";

		cout << "]" << endl << endl;
	}
	if (*star && P->even != -1)
		free (hit);
*/

	free (v);
	free (f);
	return 1;
}

// Create a polyhedron
void CPrimitives::OnModifyPolyhedron(bool SaveAction)
{
	// Create geometry as needed ...
	if(m_CreateGeometry)
		{
			// Make it
			LoadPolyhedron(m_PolyhedronNumber, m_PolyhedronDualCheck);

			// Name our mesh after the polyhedron type ...
			sdpString name = m_PolyhedronDualCheck ? uniform[m_PolyhedronNumber-1].dual : uniform[m_PolyhedronNumber-1].name;

			if(SaveAction)
				Document()->Actions()->StartAction("Polyhedron");

			// Create a mesh ...
			CreateMesh(name, m_PolyhedronPoints.size(), m_PolyhedronPolygons.size());

			if(SaveAction)
				Document()->Actions()->FinishAction();

			// For each polygon ...
			for(gulong i = 0; i < m_PolyhedronPolygons.size(); i++ )
				{
					// Attach points ...
					for(gulong j = 0; j < m_PolyhedronPolygons[i].size(); j++ )
						{
							m_Polygons[0][i]->Points()->AddPoint(m_Points[0][m_PolyhedronPolygons[i][j]]);
						}
				}
		}

	// Calculate point positions ...
	for(gulong i = 0; i < m_PolyhedronPoints.size(); i++)
		{
			sdpVector3 coords = m_PolyhedronPoints[i] * m_PolyhedronScaling * m_PolyhedronLength;

			m_Points[0][i]->Location()->SetOldXYZ(coords);
			m_Points[0][i]->Location()->SetLocalXYZ(coords);
		}

	// Show polyhedron geometry info
	m_UserInterface.Entry("polyhedronproperties").SetText(m_PolyhedronProperties);
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CPrimitives::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CPrimitives::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CPrimitives::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CPrimitives::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CPrimitives::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPrimTool::ObjectDeleted();
}

