// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CRotate K-3D tool, which alters the orientation of objects, or their geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "rotate.h"

#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>
#include <k3diactions.h>
#include <k3ditransform.h>

/////////////////////////////////////////////////////////////////////////////
// CRotate

CRotate::CRotate()
{
	m_Rotate = sdpEulerAngles(0.0, 0.0, 0.0, sdpEulerAngles::ZXYstatic);
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CRotate::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}


gulong CRotate::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}


gulong CRotate::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}



///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CRotate::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CRotate::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CRotate::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	gdouble xfactor = sdpRadians(gdouble(CurrentX - LastX) * 0.1);
	gdouble yfactor = sdpRadians(gdouble(CurrentY - LastY) * 0.1);

	// If the user holds down either control key, only rotate along the active axis ...
	if(ControlDown(Modifiers))
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Rotate[1] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::YAXIS:
						m_Rotate[2] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::ZAXIS:
						m_Rotate[0] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					default:
						g_assert_not_reached();	// Did somebody add an axis?
				}
		}
	// Otherwise, rotate along the two non-active axes ...
	else
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Rotate[0] -= xfactor;
						m_Rotate[2] -= yfactor;
						break;
					case k3dIDocument::YAXIS:
						m_Rotate[1] -= xfactor;
						m_Rotate[0] -= yfactor;
						break;
					case k3dIDocument::ZAXIS:
						m_Rotate[1] += xfactor;
						m_Rotate[2] -= yfactor;
						break;
					default:
						g_assert_not_reached();	// Did somebody add an axis?
				}
		}

	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}


sdpResult CRotate::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}


/*
// Called when the user rolls the mouse wheel on an Intellimouse-compatible mouse
sdpResult CRotate::OnViewMouseWheel(long Flags, long Delta, long X, long Y)
{
	double factor;

	k3dTool::OnViewMouseWheel(Flags, Delta, X, Y);

	factor = (double)Delta / (double)WHEEL_DELTA * 0.05;

	// Rotate along the active axis only
	switch(Document()->GetActiveAxis())
		{
			case k3dIDocument::XAXIS:
				m_Rotate[0] += factor;
				break;
			case k3dIDocument::YAXIS:
				m_Rotate[1] += factor;
				break;
			case k3dIDocument::ZAXIS:
				m_Rotate[2] += factor;
				break;
			default:
				g_assert(0);
		}


	// Update the scaling factor edit controls
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_RotatePage->GetDlgItem(IDC_SCALEX)->m_hWnd, m_Rotate[0]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_RotatePage->GetDlgItem(IDC_SCALEY)->m_hWnd, m_Rotate[1]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_RotatePage->GetDlgItem(IDC_SCALEZ)->m_hWnd, m_Rotate[2]);

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}
*/


void CRotate::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Rotate");

	// Call the default implementation ...
	k3dTool::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}


void CRotate::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// See if the object supports k3dITransform ...
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(Object);
	if(transform)
		{
			sdpAngleAxis angleaxis = sdpQuaternion(m_Rotate);
			transform->SetOrientation(k3dITransform::CURRENT_FRAME, angleaxis);
			transform->Release();
		}
	// It doesn't, so go to the default, location-by-location approach ...
	else
		{
			ModifyLocationServer(Object);
		}
}


void CRotate::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Setup our transformation matrix ...
	m_Matrix = sdpIdentity3D() * sdpRotation3D(sdpQuaternion(m_Rotate));

	// Call the default implementation ...
	k3dTool::ModifyLocationServer(Object);
}


void CRotate::ModifyLocation(k3dILocation* Location)
{
	sdpVector3 coords;
	Location->GetOldXYZ(coords);
	coords = m_Matrix * coords;
	Location->SetLocalXYZ(coords);
}


//////////////////////////////////////////////////////////////////////////////////////
// CRotateDialog

CRotateDialog::CRotateDialog()
{
	m_Object = 0;
}

bool CRotateDialog::Create(CRotate* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "rotate.gtkml", "rotate.html"), false);

	return true;
}

void CRotateDialog::UpdateControls()
{
	Entry("x").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Rotate[1], 3));
	Entry("y").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Rotate[2], 3));
	Entry("z").SetText(k3dConvertDoubleString(k3dIChannel::DATA_ANGLE, m_Object->m_Rotate[0], 3));
}

void CRotateDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onx")
		OnX();
	else if(Event->Name() == "ony")
		OnY();
	else if(Event->Name() == "onz")
		OnZ();
	else
		k3dDialog::OnEvent(Event);
}

void CRotateDialog::OnX()
{
	if(UpdateControl(Entry("x"), m_Object->m_Rotate[1], k3dIChannel::DATA_ANGLE, 3))
		m_Object->OnModify(true);
}

void CRotateDialog::OnY()
{
	if(UpdateControl(Entry("y"), m_Object->m_Rotate[2], k3dIChannel::DATA_ANGLE, 3))
		m_Object->OnModify(true);
}

void CRotateDialog::OnZ()
{
	if(UpdateControl(Entry("z"), m_Object->m_Rotate[0], k3dIChannel::DATA_ANGLE, 3))
		m_Object->OnModify(true);
};



