// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CScale K-3D tool, which alters the scale of objects or their geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "scale.h"

#include <k3dmisc.h>
#include <k3diactions.h>
#include <k3ditransform.h>

#include <sdpgtkevent.h>
#include <sdpfile.h>

/////////////////////////////////////////////////////////////////////////////
// CScale

CScale::CScale()
{
	m_Scale = 1.0;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CScale::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CScale::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CScale::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CScale::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CScale::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CScale::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.01;
	double yfactor = (CurrentY - LastY) * 0.01;

	// If the user holds down either shift key, constrain scale factors to be equal ...
	if(ShiftDown(Modifiers))
		{
			m_Scale[0] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
			m_Scale[1] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
			m_Scale[2] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
		}
	// If the user holds down either control key, only scale along the active axis ...
	else if(ControlDown(Modifiers))
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Scale[0]-=(fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::YAXIS:
						m_Scale[1]-=(fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::ZAXIS:
						m_Scale[2]-=(fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					default:
						g_assert(0);
				}
		}
	// Otherwise, scale along the two non-active axes ...
	else
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Scale[2] -= xfactor;
						m_Scale[1] -= yfactor;
						break;
					case k3dIDocument::YAXIS:
						m_Scale[0] -= xfactor;
						m_Scale[2] -= yfactor;
						break;
					case k3dIDocument::ZAXIS:
						m_Scale[0] -= xfactor;
						m_Scale[1] -= yfactor;
						break;
					default:
						g_assert(0);
				}
		}

	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CScale::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}

/*
// Called when the user rolls the mouse wheel on an Intellimouse-compatible mouse
sdpResult CScale::OnViewMouseWheel(long Flags, long Delta, long X, long Y)
{
	double factor;

	k3dTool::OnViewMouseWheel(Flags, Delta, X, Y);

	factor = (double)Delta / (double)WHEEL_DELTA * 0.05;

	// Scale along the active axis only
	switch(Document()->GetActiveAxis())
		{
			case k3dIDocument::XAXIS:
				m_Scale[0] += factor;
				break;
			case k3dIDocument::YAXIS:
				m_Scale[1] += factor;
				break;
			case k3dIDocument::ZAXIS:
				m_Scale[2] += factor;
				break;
			default:
				g_assert(0);
		}


	// Update the scaling factor edit controls
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_ScalePage->GetDlgItem(IDC_SCALEX)->m_hWnd, m_Scale[0]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_ScalePage->GetDlgItem(IDC_SCALEY)->m_hWnd, m_Scale[1]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_ScalePage->GetDlgItem(IDC_SCALEZ)->m_hWnd, m_Scale[2]);

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}
*/

void CScale::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Scale");

	// Call the default implementation ...
	k3dTool::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}

void CScale::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// See if the object supports k3dITransform ...
	k3dITransform* transform = sdpQueryInterface<k3dITransform>(Object);
	if(transform)
		{
			transform->SetScale(k3dITransform::CURRENT_FRAME, m_Scale);
			transform->Release();
		}
	// It doesn't, so go to the default, location-by-location approach ...
	else
		{
			ModifyLocationServer(Object);
		}
}

void CScale::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Setup our transformation matrix ...
	m_Matrix = sdpIdentity3D() * sdpScaling3D(m_Scale);

	// Call the default implementation ...
	k3dTool::ModifyLocationServer(Object);
}

void CScale::ModifyLocation(k3dILocation* Location)
{
	sdpVector3 coords;
	Location->GetOldXYZ(coords);
	coords = m_Matrix * coords;
	Location->SetLocalXYZ(coords);
}

//////////////////////////////////////////////////////////////////////////////////////
// CScaleDialog

CScaleDialog::CScaleDialog()
{
	m_Object = 0;
}

bool CScaleDialog::Create(CScale* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "scale.gtkml", "scale.html"), false);

	return true;
}

void CScaleDialog::UpdateControls()
{
	Entry("x").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Scale[0], 3));
	Entry("y").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Scale[1], 3));
	Entry("z").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_Scale[2], 3));
}

void CScaleDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onx")
		OnX();
	else if(Event->Name() == "ony")
		OnY();
	else if(Event->Name() == "onz")
		OnZ();
	else
		k3dDialog::OnEvent(Event);
}

void CScaleDialog::OnX()
{
	if(UpdateControl(Entry("x"), m_Object->m_Scale[0], k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(true);
}

void CScaleDialog::OnY()
{
	if(UpdateControl(Entry("y"), m_Object->m_Scale[1], k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(true);
}

void CScaleDialog::OnZ()
{
	if(UpdateControl(Entry("z"), m_Object->m_Scale[2], k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(true);
};



