// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSlant class, which applies a "slant" transformation to an object's geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "slant.h"

#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>
#include <k3diactions.h>

/////////////////////////////////////////////////////////////////////////////
// CSlant

CSlant::CSlant()
{
	m_Slant = 0.0;
	m_Sensitivity = 0.1;
	m_Mirror = false;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CSlant::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CSlant::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CSlant::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CSlant::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CSlant::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CSlant::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * m_Sensitivity;
	double yfactor = (CurrentY - LastY) * m_Sensitivity;

	// If the user holds down either control key, only translate along the active axis ...
	if(ControlDown(Modifiers))
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Slant[0] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::YAXIS:
						m_Slant[1] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					case k3dIDocument::ZAXIS:
						m_Slant[2] -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);
						break;
					default:
						g_assert(0);
				}
		}
	// Otherwise, translate along the two non-active axes ...
	else
		{
			switch(Document()->GetActiveAxis())
				{
					case k3dIDocument::XAXIS:
						m_Slant[2] -= xfactor;
						m_Slant[1] -= yfactor;
						break;
					case k3dIDocument::YAXIS:
						m_Slant[0] -= xfactor;
						m_Slant[2] -= yfactor;
						break;
					case k3dIDocument::ZAXIS:
						m_Slant[0] += xfactor;
						m_Slant[1] -= yfactor;
						break;
					default:
						g_assert(0);
				}
		}

	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CSlant::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}

/*
// Called when the user rolls the mouse wheel on an Intellimouse-compatible mouse
sdpResult CSlant::OnViewMouseWheel(long Flags, long Delta, long X, long Y)
{
	double factor;

	k3dTool::OnViewMouseWheel(Flags, Delta, X, Y);

	factor = (double)Delta / (double)WHEEL_DELTA * 0.05;

	// Slant along the active axis only
	switch(Document()->GetActiveAxis())
		{
			case k3dIDocument::XAXIS:
				m_Slant[0] += factor;
				break;
			case k3dIDocument::YAXIS:
				m_Slant[1] += factor;
				break;
			case k3dIDocument::ZAXIS:
				m_Slant[2] += factor;
				break;
			default:
				g_assert(0);
		}


	// Update the scaling factor edit controls
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_SlantPage->GetDlgItem(IDC_SCALEX)->m_hWnd, m_Slant[0]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_SlantPage->GetDlgItem(IDC_SCALEY)->m_hWnd, m_Slant[1]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_SlantPage->GetDlgItem(IDC_SCALEZ)->m_hWnd, m_Slant[2]);

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}
*/

void CSlant::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Slant");

	// Call the default implementation ...
	k3dTool::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}

void CSlant::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Got to the default, location-by-location approach ...
	ModifyLocationServer(Object);
}

void CSlant::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Call the default implementation ...
	k3dTool::ModifyLocationServer(Object);
}

void CSlant::ModifyLocation(k3dILocation* Location)
{
	double dx, dy, dz;

	sdpVector3 coords;
	Location->GetOldXYZ(coords);
//	coords -= m_Origin;

	switch(m_ActiveAxis)
		{
			case k3dIDocument::XAXIS:
				dx = 0.0;
				dy = m_Size[2] ? coords[2] / m_Size[2] : coords[2];
				dz =  m_Size[1] ? coords[1] / m_Size[1] : coords[1];
				break;
			case k3dIDocument::YAXIS:
				dx = m_Size[2] ? coords[2] / m_Size[2] : coords[2];
				dy = 0.0;
				dz = m_Size[0] ? coords[0] / m_Size[0] : coords[0];
				break;
			case k3dIDocument::ZAXIS:
				dx = m_Size[1] ? coords[1] / m_Size[1] : coords[1];
				dy = m_Size[0] ? coords[0] / m_Size[0] : coords[0];
				dz = 0.0;
				break;
			default:
				g_assert_not_reached();	// Unknown axis!
		}

/*
	if(m_Mirror)
		{
			dx = m_Xfalloff->GetValue1(fabs(dx), .0001);
			dy = m_Yfalloff->GetValue1(fabs(dy), .0001);
			dz = m_Zfalloff->GetValue1(fabs(dz), .0001);
		}
	else
		{
			dx = KSign<double>(dx) * m_Xfalloff->GetValue1(fabs(dx), .0001);
			dy = KSign<double>(dy) * m_Yfalloff->GetValue1(fabs(dy), .0001);
			dz = KSign<double>(dz) * m_Zfalloff->GetValue1(fabs(dz), .0001);
		}
*/

	coords[0] += dx * m_Slant[0];
	coords[1] += dy * m_Slant[1];
	coords[2] += dz * m_Slant[2];

//	coords += m_Origin;

	Location->SetLocalXYZ(coords);
}

//////////////////////////////////////////////////////////////////////////////////////
// CSlantDialog

CSlantDialog::CSlantDialog()
{
	m_Object = 0;
}

bool CSlantDialog::Create(CSlant* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "slant.gtkml", "slant.html"), false);

	return true;
}

void CSlantDialog::UpdateControls()
{
	Entry("x").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Slant[0], 3));
	Entry("y").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Slant[1], 3));
	Entry("z").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Slant[2], 3));
	Entry("sensitivity").SetText(k3dConvertDoubleString(k3dIChannel::DATA_DISTANCE, m_Object->m_Sensitivity, 3));
}

void CSlantDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onx")
		OnX();
	else if(Event->Name() == "ony")
		OnY();
	else if(Event->Name() == "onz")
		OnZ();
	else if(Event->Name() == "onsensitivity")
		OnSensitivity();
	else
		k3dDialog::OnEvent(Event);
}

void CSlantDialog::OnX()
{
	if(UpdateControl(Entry("x"), m_Object->m_Slant[0], k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
}

void CSlantDialog::OnY()
{
	if(UpdateControl(Entry("y"), m_Object->m_Slant[1], k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
}

void CSlantDialog::OnZ()
{
	if(UpdateControl(Entry("z"), m_Object->m_Slant[2], k3dIChannel::DATA_DISTANCE, 3))
		m_Object->OnModify(true);
};

void CSlantDialog::OnSensitivity()
{
	UpdateControl(Entry("sensitivity"), m_Object->m_Sensitivity, k3dIChannel::DATA_DISTANCE, 3);
};

