// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSmooth K-3D tool, which "smooths" geometry by averaging point locations
		\author Tim Shead (tshead@k-3d.com)
*/

#include "smooth.h"

#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>
#include <k3diactions.h>
#include <k3dipolypointserver.h>
#include <k3distate.h>

/////////////////////////////////////////////////////////////////////////////
// CSmooth

CSmooth::CSmooth()
{
	m_SmoothingFactor = 0.0;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CSmooth::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CSmooth::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CSmooth::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CSmooth::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CSmooth::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CSmooth::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = (CurrentX - LastX) * 0.01;
	double yfactor = (CurrentY - LastY) * 0.01;

	m_SmoothingFactor += fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor;

	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CSmooth::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}


/*
// Called when the user rolls the mouse wheel on an Intellimouse-compatible mouse
sdpResult CSmooth::OnViewMouseWheel(long Flags, long Delta, long X, long Y)
{
	double factor;

	k3dTool::OnViewMouseWheel(Flags, Delta, X, Y);

	factor = (double)Delta / (double)WHEEL_DELTA * 0.05;

	// Smooth along the active axis only
	switch(Document()->GetActiveAxis())
		{
			case k3dIDocument::XAXIS:
				m_Smooth[0] += factor;
				break;
			case k3dIDocument::YAXIS:
				m_Smooth[1] += factor;
				break;
			case k3dIDocument::ZAXIS:
				m_Smooth[2] += factor;
				break;
			default:
				g_assert(0);
		}


	// Update the scaling factor edit controls
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_SmoothPage->GetDlgItem(IDC_SCALEX)->m_hWnd, m_Smooth[0]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_SmoothPage->GetDlgItem(IDC_SCALEY)->m_hWnd, m_Smooth[1]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_SmoothPage->GetDlgItem(IDC_SCALEZ)->m_hWnd, m_Smooth[2]);

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}
*/

class CSmoothSmooth
{
public:
	CSmoothSmooth(gdouble SmoothingFactor) { m_SmoothingFactor = SmoothingFactor; }

	bool HandleMesh(k3dIMesh* Mesh)
	{
		return true;
	}

	bool HandleMeshPoint(k3dIMesh* Mesh, k3dIPoint* Point)
	{
		k3dIPolygonServer* polygons;
		k3dIPolygon* polygon;
		k3dIPolyPointServer* polypoints;
		K3DPOLYPOINTHANDLE polypointhandle;
		K3DPOLYPOINTHANDLE polypointhandle2;
		gulong count = 0;
		sdpVector3 oldcoords;
		sdpVector3 coords;
		sdpVector3 averagecoords;
		sdpVector3 newcoords;

		// Get the original point coordinates ...
		Point->Location()->GetOldXYZ(oldcoords);

		// For each polygon in the mesh ...
		polygons = Mesh->Polygons();
		for(K3DPOLYGONHANDLE polyhandle = polygons->GetFirstPolygon(); polyhandle; polyhandle = polygons->GetNextPolygon(polyhandle))
			{
				// Skip polygons that aren't visible and selected ...
				polygon = polygons->GetPolygon(polyhandle);
				if(SDP_S_OK != polygon->State()->QueryState(k3dIState::VISIBLESELECTED))
					continue;

				// For each polypoint in the polygon ...
				polypoints = polygon->Points();
				for(polypointhandle = polypoints->GetFirstPolyPoint(); polypointhandle; polypointhandle = polypoints->GetNextPolyPoint(polypointhandle))
					{
						// If it refers to our test point ...
						if(polypoints->GetPolyPoint(polypointhandle)->Point() == Point)
							{
								// Add the sum of the polygon's points ...
								for(polypointhandle2 = polypoints->GetFirstPolyPoint(); polypointhandle2; polypointhandle2 = polypoints->GetNextPolyPoint(polypointhandle2))
									{
										polypoints->GetPolyPoint(polypointhandle2)->Location()->GetOldXYZ(coords);
										averagecoords += coords;
										count++;
									}
							}
					}
			}

		if(count)
			{
				averagecoords /= count;
				newcoords = sdpLerp(oldcoords, averagecoords, m_SmoothingFactor);
				Point->Location()->SetLocalXYZ(newcoords);
			}

		return true;
	}

	gdouble m_SmoothingFactor;
};

void CSmooth::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Smooth");

	// Modify the geometry ...
	CSmoothSmooth functor(m_SmoothingFactor);
	k3dForEachMeshPoint(Document(), k3dIState::VISIBLESELECTED, functor);

	// Update the display
	k3dUserObject::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}


//////////////////////////////////////////////////////////////////////////////////////
// CSmoothDialog

CSmoothDialog::CSmoothDialog()
{
	m_Object = 0;
}

bool CSmoothDialog::Create(CSmooth* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "smooth.gtkml", "smooth.html"), false);

	return true;
}

void CSmoothDialog::UpdateControls()
{
	Entry("smoothingfactor").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_SmoothingFactor, 2));
}

void CSmoothDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "onsmoothingfactor")
		OnSmoothingFactor();
	else
		k3dDialog::OnEvent(Event);
}

void CSmoothDialog::OnSmoothingFactor()
{
	if(UpdateControl(Entry("smoothingfactor"), m_Object->m_SmoothingFactor, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(true);
}



