// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CSnap K-3D tool, which "snaps" a point to another point's location
		\author Tim Shead (tshead@k-3d.com)
		\author Romain Behar (romainbehar@yahoo.com)
*/

#include "snap.h"

#include <k3dmisc.h>

#include <k3diactions.h>
#include <k3dieditorengine.h>
#include <k3dicamera.h>

#include <sdpgl.h>

#include <sdpgtkevent.h>
#include <sdpfile.h>

#include <iostream>

/////////////////////////////////////////////////////////////////////////////
// CSnap

CSnap::CSnap()
{
	m_Point1 = 0;
	m_Point2 = 0;

	m_MovePolys = false;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CSnap::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIEditorRender>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CSnap::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CSnap::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CSnap::SetObjectContext(k3dIObjectContext* ObjectContext)
{
	if(sdpFailed(k3dTool::SetObjectContext(ObjectContext)))
		return SDP_E_FAIL;

	// Make sure that we get called by the rendering engine ...
	ObjectContext->State()->SetState(k3dIState::VISIBLE, true);

	return SDP_S_OK;
}

sdpResult CSnap::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CSnap::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIEditorRender overrides

sdpResult CSnap::EditorPostRender(k3dIEditorEngine* Engine)
{
	// Sanity checks ...
	g_assert(Engine);
	
	// If nothing's selected, we're done!
	if(!m_Point1 && !m_Point2)
		return SDP_S_FALSE;

	// We need to work in screen coordinates ...
	Engine->Camera()->SnapshotScreenCoordinates();

	// Get window dimensions ...
	gdouble width = Engine->Camera()->Width();
	gdouble height = Engine->Camera()->Height();
	gdouble size = sdpMin(width, height);

	// Make sure they're non-zero (minimized window?) ...
	if(0.0 == size)
		return SDP_S_FALSE;
	
	// Draw selected points
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	// Push a new modelview matrix onto the stack ...
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, width, height, 0, -1, 1);
	glViewport(0, 0, GLint(width), GLint(height));

	glColor3d(1.0, 0.0, 0.0);
	glPointSize(6.0f);

	glBegin(GL_POINTS);
	
	sdpVector3 coords;
	if(m_Point1)
		{
			m_Point1->Location()->GetScreenXYZ(coords);
			glVertex2d(coords[0], coords[1]);
		}
	if(m_Point2)
		{
			m_Point2->Location()->GetScreenXYZ(coords);
			glVertex2d(coords[0], coords[1]);
		}

	glEnd();

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glPopAttrib();

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

/// Functor for choosing points
class CSnapSelectPoint
{
public:
	CSnapSelectPoint(sdpVector2 Mouse)
	{
		m_Mouse = Mouse;
		m_Point = 0;
	}

	k3dIPoint* Point() { return m_Point; }

	bool HandlePoint(k3dIPoint* Point)
	{
		sdpVector3 screencoords;
		Point->Location()->GetScreenXYZ(screencoords);

		if((fabs(screencoords[0] - m_Mouse[0]) < 4.0) &&
			 (fabs(screencoords[1] - m_Mouse[1]) < 4.0))
			{
				m_Point = Point;
				return false;
			}
		return true;
	}

protected:
	sdpVector2 m_Mouse;
	k3dIPoint* m_Point;
};

// Called while the user clicks the left mouse button
sdpResult CSnap::OnLButtonClick(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY)
{
	Camera->SnapshotScreenCoordinates();

	CSnapSelectPoint functor(sdpVector2(CurrentX, CurrentY));
	k3dForEachPoint(Document(), k3dIState::VISIBLESELECTABLE, functor);

	if(!functor.Point())
		return SDP_S_OK;

	if(m_Point1)
		{
			if(functor.Point() == m_Point1)
				{
					Document()->Application()->Error("You can't select the same point twice!", "Snap:");
					return SDP_S_OK;
				}
				
			m_Point2 = functor.Point();
			OnModify(true);
		}
	else
		{
			m_Point1 = functor.Point();
			OnModify(false);
		}
		
	return SDP_S_OK;
}

void CSnap::OnModify(bool SaveAction)
{
	// Call the default implementation, to update the display ...
	k3dUserObject::OnModify(SaveAction);

	// If the user hasn't selected both points yet, we're done!
	if(!m_Point1 || !m_Point2)
		return;

	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Snap");

/*
	// Move the entire containing polygon ...
	sdpVector3 coords, coords2;
	if(m_MovePolys)
		{
			sdpVector3 start;
			m_Point1->Location()->GetOldXYZ(start);

			sdpVector3 end;
			m_Point2->Location()->GetOldXYZ(end);

			sdpVector3 delta = end - start;

			for(hr = m_PointMatchList->SetFirstMatch(); SDP_S_OK == hr; hr = m_PointMatchList->SetNextMatch())
				{
					m_PointLocation->GetOldXYZ(coords);
					coords[0] += dx;
					coords[1] += dy;
					coords[2] += dz;
					m_PointLocation->SetLocalXYZ(coords);
				}
		}
	// Move the solitary point ...
	else
*/
		{
			sdpVector3 coords;
			m_Point2->Location()->GetOldXYZ(coords);
			m_Point1->Location()->SetLocalXYZ(coords);
		}

	m_Point1 = 0;
	m_Point2 = 0;
	
	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}

////////////////////////////////////////////////////////////////////////////
// CSnapDialog

CSnapDialog::CSnapDialog()
{
	m_Object = 0;
}

bool CSnapDialog::Create(CSnap* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "snap.gtkml", "snap.html"), false);

	return true;
}


