// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTaper K-3D tool, which applies a "taper" transformation to geometry
		\author Tim Shead (tshead@k-3d.com)
*/

#include "taper.h"

#include <sdpfile.h>
#include <k3dmisc.h>

#include <sdpgtkevent.h>
#include <k3diactions.h>

/////////////////////////////////////////////////////////////////////////////
// CTaper

CTaper::CTaper()
{
	m_TaperFactor = 1.0;
	m_DisplaceX = true;
	m_DisplaceY = true;
	m_DisplaceZ = true;
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CTaper::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIUserEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;

	return SDP_E_NO_INTERFACE;
}

gulong CTaper::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CTaper::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

// Called to open an object's UI
sdpResult CTaper::EditObject()
{
	// Create a user interface ...
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);

	// Update the display ...
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CTaper::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dTool::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIUserEventClient overrides

// Called while the user drags the left mouse button
sdpResult CTaper::OnLButtonDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	// Compute scaling factors for each mouse axis ...
	double xfactor = ((double)(CurrentX - LastX)) * 0.01;
	double yfactor = ((double)(CurrentY - LastY)) * 0.01;

	m_TaperFactor -= (fabs(xfactor) >= fabs(yfactor) ? xfactor : yfactor);

	// Update our dialog ...
	m_UserInterface.UpdateControls();

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}

sdpResult CTaper::OnLButtonEndDrag(k3dICamera* Camera, GdkModifierType Modifiers, gdouble CurrentX, gdouble CurrentY, gdouble LastX, gdouble LastY, gdouble StartX, gdouble StartY)
{
	OnModify(true);

	return SDP_S_OK;
}

/*
// Called when the user rolls the mouse wheel on an Intellimouse-compatible mouse
sdpResult CTaper::OnViewMouseWheel(long Flags, long Delta, long X, long Y)
{
	double factor;

	k3dTool::OnViewMouseWheel(Flags, Delta, X, Y);

	factor = (double)Delta / (double)WHEEL_DELTA * 0.05;

	// Taper along the active axis only
	switch(Document()->GetActiveAxis())
		{
			case k3dIDocument::XAXIS:
				m_Taper[0] += factor;
				break;
			case k3dIDocument::YAXIS:
				m_Taper[1] += factor;
				break;
			case k3dIDocument::ZAXIS:
				m_Taper[2] += factor;
				break;
			default:
				g_assert(0);
		}


	// Update the scaling factor edit controls
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_TaperPage->GetDlgItem(IDC_SCALEX)->m_hWnd, m_Taper[0]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_TaperPage->GetDlgItem(IDC_SCALEY)->m_hWnd, m_Taper[1]);
	k3dSetDouble(k3dIChannel::DATA_SCALAR, m_Sheet->m_TaperPage->GetDlgItem(IDC_SCALEZ)->m_hWnd, m_Taper[2]);

	// Update the display
	OnModify(false);

	return SDP_S_OK;
}
*/

void CTaper::OnModify(bool SaveAction)
{
	// Record the action if requested ...
	if(SaveAction)
		Document()->Actions()->StartAction("Taper");

	// Call the default implementation ...
	k3dTool::OnModify(SaveAction);

	// Finish recording the action if requested ...
	if(SaveAction)
		{
			Document()->Actions()->FinishAction();
			Document()->FreezeChanges();
		}
}

void CTaper::ModifyObject(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Go to the default, location-by-location approach ...
	ModifyLocationServer(Object);
}

void CTaper::ModifyLocationServer(k3dIObject* Object)
{
	// Sanity checks ...
	g_assert(Object);

	// Call the default implementation ...
	k3dTool::ModifyLocationServer(Object);
}

void CTaper::ModifyLocation(k3dILocation* Location)
{
	sdpVector3 coords;
	Location->GetOldXYZ(coords);
//	coords -= m_Origin;

	double delta, scale;

	switch(m_ActiveAxis)
		{
			case k3dIDocument::XAXIS:
				if(m_OriginSize[0])
					delta = fabs(coords[0] / m_OriginSize[0]);
				break;
			case k3dIDocument::YAXIS:
				if(m_OriginSize[1])
					delta = fabs(coords[1] / m_OriginSize[1]);
				break;
			case k3dIDocument::ZAXIS:
				if(m_OriginSize[2])
					delta = fabs(coords[2] / m_OriginSize[2]);
				break;
			default:
				g_assert_not_reached();	// Unknown axis!!!
				break;
		}


//	delta = m_Falloff->GetValue1(delta, .001);
	scale = 1.0 + (delta * (m_TaperFactor - 1.0));

	if(m_DisplaceX && (k3dIDocument::XAXIS != m_ActiveAxis))
		coords[0] *= scale;

	if(m_DisplaceY && (k3dIDocument::YAXIS != m_ActiveAxis))
		coords[1] *= scale;

	if(m_DisplaceZ && (k3dIDocument::ZAXIS != m_ActiveAxis))
		coords[2] *= scale;

//	coords += m_Origin;

	Location->SetLocalXYZ(coords);
}

//////////////////////////////////////////////////////////////////////////////////////
// CTaperDialog

CTaperDialog::CTaperDialog()
{
	m_Object = 0;
}

bool CTaperDialog::Create(CTaper* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, true, false, "taper.gtkml", "taper.html"), false);

	return true;
}

void CTaperDialog::UpdateControls()
{
	Entry("taperfactor").SetText(k3dConvertDoubleString(k3dIChannel::DATA_SCALAR, m_Object->m_TaperFactor, 3));
	CheckButton("displacex").SetState(m_Object->m_DisplaceX);
	CheckButton("displacey").SetState(m_Object->m_DisplaceY);
	CheckButton("displacez").SetState(m_Object->m_DisplaceZ);
}

void CTaperDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ontaperfactor")
		OnTaperFactor();
	else if(Event->Name() == "ondisplacex")
		OnDisplaceX();
	else if(Event->Name() == "ondisplacey")
		OnDisplaceY();
	else if(Event->Name() == "ondisplacez")
		OnDisplaceZ();
	else
		k3dDialog::OnEvent(Event);
}

void CTaperDialog::OnTaperFactor()
{
	if(UpdateControl(Entry("taperfactor"), m_Object->m_TaperFactor, k3dIChannel::DATA_SCALAR, 3))
		m_Object->OnModify(true);
}

void CTaperDialog::OnDisplaceX()
{
	m_Object->m_DisplaceX = CheckButton("displacex").GetState();
	m_Object->OnModify(false);

}

void CTaperDialog::OnDisplaceY()
{
	m_Object->m_DisplaceY = CheckButton("displacey").GetState();
	m_Object->OnModify(false);

}

void CTaperDialog::OnDisplaceZ()
{
	m_Object->m_DisplaceZ = CheckButton("displacez").GetState();
	m_Object->OnModify(false);

}


