// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the functions necessary to export K-3D objects from the transformerlib module
		\author Timothy M. Shead (tshead@k-3d.com)
*/
		
#include <gmodule.h>

#include <strstream>
#include <sdpxml.h>

#include <k3dcom.h>
#include <k3dmisc.h>

#include "slanter.h"
#include "sphereizer.h"
#include "taperer.h"
#include "transformer.h"
#include "twister.h"
#include "waveizer.h"

#include "../../k3d_version.h"

extern "C" {

/// Stores information in XML format about the types of objects exported by this module
static std::strstream ClassInfo;

/// Called to create an XML stream object containing information on the types of objects exported by this module
static void BuildClassInfo()
{
	// Generate an XML document ...
	sdpXmlDocument document;
	sdpXmlIElement* object;
	sdpXmlIElement*  library = k3dStartLibraryInfo(&document, "transformerlib", VERSION_STRING);

	// Slanter object ...
	object = k3dAddObject(library, sdpSlanter, "Slanter");
	k3dAddBookmark(object, "Objects/Slanter");
	
	// Sphereizer object ...
	object = k3dAddObject(library, sdpSphereizer, "Sphereizer");
	k3dAddBookmark(object, "Objects/Sphereizer");
	
	// Taperer object ...
	object = k3dAddObject(library, sdpTaperer, "Taperer");
	k3dAddBookmark(object, "Objects/Taperer");
	
	// Transformer object ...
	object = k3dAddObject(library, sdpTransformer, "Transformer");
	k3dAddBookmark(object, "Objects/Transformer");
	
	// Twister object ...
	object = k3dAddObject(library, sdpTwister, "Twister");
	k3dAddBookmark(object, "Objects/Twister");
	
	// Waveizer object ...
	object = k3dAddObject(library, sdpWaveizer, "Waveizer");
	k3dAddBookmark(object, "Objects/Waveizer");
	
	ClassInfo << document << std::ends;
}

/// Called when this shared library is loaded
G_MODULE_EXPORT const gchar* g_module_check_init(GModule *module)
{
	// Sanity checks ...
	g_assert(module);

	// Build our XML sdpComClassInfo document ...
	BuildClassInfo();

	return 0;
}

/// Called when this shared library is about to be unloaded
G_MODULE_EXPORT void g_module_unload(GModule *module)
{
	// Sanity checks ...
	g_assert(module);

	// Unlock memory on our XML sdpComClassInfo document ...
	ClassInfo.rdbuf()->freeze(false);
}

/// Called to obtain an XML document describing the objects supported by this library
G_MODULE_EXPORT SDPCSTRING sdpComLibraryInfo()
{
	return ClassInfo.str();
}

/// Called to create an instance of an object class and return an interface of the requested type
G_MODULE_EXPORT sdpResult sdpComClassProcedure(const sdpCLSID& ClassID, const sdpIID& InterfaceID, void** Interface)
{
	if(ClassID == sdpSlanter)
		return sdpCreateClassInstance<CSlanter>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpSphereizer)
		return sdpCreateClassInstance<CSphereizer>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpTaperer)
		return sdpCreateClassInstance<CTaperer>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpTransformer)
		return sdpCreateClassInstance<CTransformer>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpTwister)
		return sdpCreateClassInstance<CTwister>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;
	if(ClassID == sdpWaveizer)
		return sdpCreateClassInstance<CWaveizer>(InterfaceID, Interface) ? SDP_S_OK : SDP_E_FAIL;

	return SDP_E_UNKNOWN_OBJECT;
}

}	// extern "C"

