// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CTwister K-3D object, which applies an animated "twist" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "twister.h"

#include <sdpfile.h>
#include <k3dmisc.h>
#include <k3dilocation.h>

CTwister::CTwister() : k3dPersistObject()
{
//	m_XFalloff() = FALSE;
//	m_YFalloff() = FALSE;
//	m_ZFalloff() = FALSE;
//	AddVariable(&m_OriginServer, "originclient", V_ORIGINCLIENT);
//	AddVariable(&m_XFalloff, "xfalloff", V_XFALLOFF);
//	AddVariable(&m_YFalloff, "yfalloff", V_YFALLOFF);
//	AddVariable(&m_ZFalloff, "zfalloff", V_ZFALLOFF);

	AddObject(&m_X("x"));
	AddObject(&m_Y("y"));
	AddObject(&m_Z("z"));
	
//	AddObject(&m_Xfalloff, "xfalloff", OC_XFALLOFF);
//	AddObject(&m_Yfalloff, "yfalloff", OC_YFALLOFF);
//	AddObject(&m_Zfalloff, "zfalloff", OC_ZFALLOFF);
}

CTwister::~CTwister()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CTwister::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CTwister::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CTwister::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CTwister::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CTwister::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CTwister::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CTwisterTwistObject
{
public:
	CTwisterTwistObject(gdouble X, gdouble Y, gdouble Z)
	{
		m_X = X;
		m_Y = Y;
		m_Z = Z;
	}

	bool HandleDescendantObject(k3dIObjectContext* Context)
	{
		// Get the object ...
		k3dIObject* object = Context->Object();
		
		// Get the object's dimensions ... if we can't, skip it ...
		sdpVector3 min;
		sdpVector3 max;
		if(!k3dObjectDimensions(object, 0, min, max))
			return true;

		// If the object doesn't support k3dILocationServer, we're done ...
		k3dILocationServer* locationserver = sdpQueryInterface<k3dILocationServer>(object);
		g_return_val_if_fail(locationserver, true);

		// Calculate the object origin ...
		sdpVector3 origin = (min + max) * 0.5;		

		// Calculate helper variables
		gdouble maxx = (max[0] - origin[0]) > (origin[0] - min[0]) ? max[0] - origin[0] : origin[0] - min[0];
		gdouble maxy = (max[1] - origin[1]) > (origin[1] - min[1]) ? max[1] - origin[1] : origin[1] - min[1];
		gdouble maxz = (max[2] - origin[2]) > (origin[2] - min[2]) ? max[2] - origin[2] : origin[2] - min[2];

		// For each location ...
		sdpVector3 coords;
		k3dILocation* location;
		gdouble delta, px, py;
		gdouble cosx, cosy, cosz;
		gdouble sinx, siny, sinz;
		for(K3DLOCATIONHANDLE handle = locationserver->GetFirstLocation(); handle; handle = locationserver->GetNextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->GetLocation(handle);
				location->GetAnimXYZ(coords);

				// Offset the origin ...
				coords -= origin;

				// Twist around X-Axis
				if(m_X && maxx)
					{
						delta = coords[0] / maxx;
//						if(m_XFalloff()) delta = KSign<double>(delta) * m_Xfalloff->GetValue1(fabs(delta), .0001);

						cosx = cos(m_X * delta);
						sinx = sin(m_X * delta);

						py = coords[1];
						coords[1] = (cosx * coords[1]) - (sinx * coords[2]);
						coords[2] = (sinx * py) + (cosx * coords[2]);
					}

				// Twist around Y-Axis
				if(m_Y && maxy)
					{
						delta = coords[1] / maxy;
//						if(m_YFalloff()) delta = KSign<double>(delta) * m_Yfalloff->GetValue1(fabs(delta), .0001);

						cosy = cos(m_Y * delta);
						siny = sin(m_Y * delta);

						px = coords[0];
						coords[0] = (cosy * coords[0]) + (siny * coords[2]);
						coords[2] = (cosy * coords[2]) - (siny * px);
					}

				// Twist around Z-Axis
				if(m_Z && maxz)
					{
						delta = coords[2] / maxz;
//						if(m_ZFalloff()) delta = KSign<double>(delta) * m_Zfalloff->GetValue1(fabs(delta), .0001);

						cosz = cos(m_Z * delta);
						sinz = sin(m_Z * delta);

						px = coords[0];
						coords[0] = (cosz * coords[0]) - (sinz * coords[1]);
						coords[1] = (sinz * px) + (cosz * coords[1]);
					}

				// Restore the origin ...
				coords += origin;
				location->SetAnimXYZ(coords);
			}
			
		locationserver->Release();									
						
		return true;
	}
	
protected:
	gdouble m_X;
	gdouble m_Y;
	gdouble m_Z;
};

sdpResult CTwister::PopAnimate()
{
	// Calculate Rotation values ...
	double x = m_X->GetValue1(m_CurrentTime, m_DeltaTime);
	double y = m_Y->GetValue1(m_CurrentTime, m_DeltaTime);
	double z = m_Z->GetValue1(m_CurrentTime, m_DeltaTime);
	
	// Setup our twist functor ...
	CTwisterTwistObject functor(x, y, z);
	
	// Apply it to each of our descendants ...
	k3dForEachDescendant(Document(), m_ObjectContext, functor);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CTwister::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(m_X.Create(Document(), "Twister X Rotation", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "X Rotation", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Y.Create(Document(), "Twister Y Rotation", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Y Rotation", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Z.Create(Document(), "Twister Z Rotation", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Z Rotation", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
/*
	if(sdpFailed(m_Xfalloff.CreateScalar(Document(), "Twister X Falloff", 0.0, TRUE, FALSE, 0.0, 1.0, "Distance", "Angle X", k3dIChannel::DATA_SCALAR, TRUE, this, m_ObjectContext->ObjectID()))) KFAIL(SDP_E_FAIL);
	m_Xfalloff->SetLinear1(0.0, 1.0);
	if(sdpFailed(m_Yfalloff.CreateScalar(Document(), "Twister Y Falloff", 0.0, TRUE, FALSE, 0.0, 1.0, "Distance", "Angle Y", k3dIChannel::DATA_SCALAR, TRUE, this, m_ObjectContext->ObjectID()))) KFAIL(SDP_E_FAIL);
	m_Yfalloff->SetLinear1(0.0, 1.0);
	if(sdpFailed(m_Zfalloff.CreateScalar(Document(), "Twister Z Falloff", 0.0, TRUE, FALSE, 0.0, 1.0, "Distance", "Angle Z", k3dIChannel::DATA_SCALAR, TRUE, this, m_ObjectContext->ObjectID()))) KFAIL(SDP_E_FAIL);
	m_Zfalloff->SetLinear1(0.0, 1.0);
*/

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CTwisterDialog

CTwisterDialog::CTwisterDialog()
{
	m_Object = 0;
}

bool CTwisterDialog::Create(CTwister* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "twister.gtkml", "twister.html"), false);

	return true;
}

bool CTwisterDialog::InitializeControls()
{
	// Setup channel buttons ...
	ConnectChannelButton(m_Object->m_X.Interface(), "x");
	ConnectChannelButton(m_Object->m_Y.Interface(), "y");
	ConnectChannelButton(m_Object->m_Z.Interface(), "z");

	return true;
}

void CTwisterDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

//	if(Event->Name() == "oncastshadows")
//		OnCastShadows();
//	else
		k3dDialog::OnEvent(Event);
}


