// K-3D
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the CWaveizer K-3D object, which applies an animated "wave" transformation to geometry
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include "waveizer.h"

#include <sdpfile.h>
#include <sdpgtkevent.h>
#include <k3dmisc.h>
#include <k3dilocation.h>

CWaveizer::CWaveizer() : k3dPersistObject()
{
	AddVariable(&m_DisplaceX("displacex", false));
	AddVariable(&m_DisplaceY("displacey", true));
	AddVariable(&m_DisplaceZ("displacez", false));
	AddVariable(&m_Along("along", X));

	AddObject(&m_Amplitude("amplitude"));
	AddObject(&m_Wavelength("wavelength"));
	AddObject(&m_Phase("phase"));
}

CWaveizer::~CWaveizer()
{
}

///////////////////////////////////////////////////////////////
// k3dIUnknown overrides

sdpResult CWaveizer::QueryInterface(const sdpIID& InterfaceID, void** Interface)
{
	if(sdpSupportInterface<k3dIAnimate>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIEventClient>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIObject>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dIPersist>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<k3dITimeServer>(this, InterfaceID, Interface)) return SDP_S_OK;										
	if(sdpSupportInterface<sdpIUnknown>(static_cast<k3dIObject*>(this), InterfaceID, Interface)) return SDP_S_OK;										

	return SDP_E_NO_INTERFACE;
}

gulong CWaveizer::AddRef()
{
	return sdpComAddRef(m_ReferenceCount);
}

gulong CWaveizer::Release()
{
	return sdpComRelease(this, m_ReferenceCount);
}

///////////////////////////////////////////////////////////////
// k3dIObject overrides

sdpResult CWaveizer::EditObject()
{
	g_return_val_if_fail(m_UserInterface.Create(this), SDP_E_FAIL);
	return SDP_S_OK;
}

sdpResult CWaveizer::ObjectRenamed(SDPCSTRING NewName)
{
	m_UserInterface.ObjectRenamed(NewName);
	return k3dPersistObject::ObjectRenamed(NewName);
}

sdpResult CWaveizer::ObjectDeleted()
{
	m_UserInterface.ObjectDeleted();
	return k3dPersistObject::ObjectDeleted();
}

///////////////////////////////////////////////////////////////
// k3dIAnimate overrides

class CWaveizerWaveizeObject
{
public:
	CWaveizerWaveizeObject(bool DisplaceX, bool DisplaceY, bool DisplaceZ, CWaveizer::Axis Along, gdouble Amplitude, gdouble Wavelength, gdouble Phase)
	{
		m_DisplaceX = DisplaceX;
		m_DisplaceY = DisplaceY;
		m_DisplaceZ = DisplaceZ;
		m_Along = Along;
		m_Amplitude = Amplitude;
		m_Wavelength = Wavelength;
		m_Phase = Phase;
	}

	bool ModifyObject(k3dIObjectContext* Context)
	{
		// Handle the possibility of a zero wavelength up front ...
		if(0 == m_Wavelength)
			return false;
	
		// Get the object ...
		k3dIObject* object = Context->Object();

		// If the object doesn't support k3dILocationServer, we're done ...
		k3dILocationServer* locationserver = sdpQueryInterface<k3dILocationServer>(object);
		if(!locationserver)
			return true;

		// For each location ...
		sdpVector3 coords;
		k3dILocation* location;
		gdouble wavefactor;
		for(K3DLOCATIONHANDLE handle = locationserver->GetFirstLocation(); handle; handle = locationserver->GetNextLocation(handle))
			{
				// Get location coordinates ...
				location = locationserver->GetLocation(handle);
				location->GetAnimXYZ(coords);

				switch(m_Along)
					{
						case CWaveizer::X:
							wavefactor = m_Amplitude * sin(sdpPiTimes2 * (coords[0] / m_Wavelength) + m_Phase);
							break;
						case CWaveizer::Y:
							wavefactor = m_Amplitude * sin(sdpPiTimes2 * (coords[1] / m_Wavelength) + m_Phase);
							break;
						case CWaveizer::Z:
							wavefactor = m_Amplitude * sin(sdpPiTimes2 * (coords[2] / m_Wavelength) + m_Phase);
							break;
						default:
							g_assert_not_reached();	// Unknown axis!!!
					}

				// Spherical waves
				if(m_DisplaceX && m_DisplaceY && m_DisplaceZ)
					{
						coords += sdpVector3(coords).Normalize() * wavefactor;
					}
				// Cylindrical waves
				else if(m_DisplaceX && m_DisplaceY)
					{
						coords += sdpVector3(coords[0], coords[1], 0).Normalize() * wavefactor;
					}
				else if(m_DisplaceY && m_DisplaceZ)
					{
						coords += sdpVector3(0, coords[1], coords[2]).Normalize() * wavefactor;
					}
				else if(m_DisplaceX && m_DisplaceZ)
					{
						coords += sdpVector3(coords[0], 0, coords[2]).Normalize() * wavefactor;
					}
				// Linear waves
				else if(m_DisplaceX)
					{
						coords[0] += wavefactor;
					}
				else if(m_DisplaceY)
					{
						coords[1] += wavefactor;
					}
				else if(m_DisplaceZ)
					{
						coords[2] += wavefactor;
					}

				location->SetAnimXYZ(coords);
			}
			
		locationserver->Release();									
						
		return true;
	}

	bool HandleChildObject(k3dIObjectContext* Context)
	{
		return ModifyObject(Context);
	}
	
	bool HandleDescendantObject(k3dIObjectContext* Context)
	{
		return ModifyObject(Context);
	}
	
protected:
	bool m_DisplaceX;
	bool m_DisplaceY;
	bool m_DisplaceZ;

	CWaveizer::Axis m_Along;
	
	gdouble m_Amplitude;
	gdouble m_Wavelength;
	gdouble m_Phase;
};

sdpResult CWaveizer::PopAnimate()
{
	// Calculate Rotation values ...
	gdouble amplitude = m_Amplitude->GetValue1(m_CurrentTime, m_DeltaTime);
	gdouble wavelength = m_Wavelength->GetValue1(m_CurrentTime, m_DeltaTime);
	gdouble phase = m_Phase->GetValue1(m_CurrentTime, m_DeltaTime);
	
	// Setup our waveize functor ...
	CWaveizerWaveizeObject functor(m_DisplaceX(), m_DisplaceY(), m_DisplaceZ(), m_Along(), amplitude, wavelength, phase);
	
	// Apply it to each of our descendants ...
//	k3dForEachChild(Document(), m_ObjectContext, functor);
	k3dForEachDescendant(Document(), m_ObjectContext, functor);

	return SDP_S_OK;
}

///////////////////////////////////////////////////////////////
// k3dIPersist overrides

sdpResult CWaveizer::InitNew()
{
	g_return_val_if_fail(sdpSucceeded(k3dPersistObject::InitNew()), SDP_E_FAIL);

	g_return_val_if_fail(m_Amplitude.Create(Document(), "Amplitude", 3.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Amplitude", k3dIChannel::DATA_DISTANCE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Wavelength.Create(Document(), "Wavelength", 5.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Wavelength", k3dIChannel::DATA_DISTANCE, true, this), SDP_E_FAIL);
	g_return_val_if_fail(m_Phase.Create(Document(), "Phase", 0.0, true, -1.0 * ULONG_MAX, ULONG_MAX, "Time", "Phase", k3dIChannel::DATA_ANGLE, true, this), SDP_E_FAIL);
	k3dChannelSetLinearValue(m_Phase.Interface(), 0.0, sdpPiTimes2);

	return SDP_S_OK;
}

/////////////////////////////////////////////////////////////////////////////
// CWaveizerDialog

CWaveizerDialog::CWaveizerDialog()
{
	m_Object = 0;
}

bool CWaveizerDialog::Create(CWaveizer* Object)
{
	// Store our parent object ...
	m_Object = Object;

	// Load our template ...
	g_return_val_if_fail(k3dDialog::Create(Object, false, true, "waveizer.gtkml", "waveizer.html"), false);

	return true;
}

bool CWaveizerDialog::InitializeControls()
{
	// Setup channel buttons ...
	ConnectChannelButton(m_Object->m_Amplitude.Interface(), "amplitude");
	ConnectChannelButton(m_Object->m_Wavelength.Interface(), "wavelength");
	ConnectChannelButton(m_Object->m_Phase.Interface(), "phase");

	return true;
}

void CWaveizerDialog::UpdateControls()
{
	CheckButton("displacex").SetState(m_Object->m_DisplaceX());
	CheckButton("displacey").SetState(m_Object->m_DisplaceY());
	CheckButton("displacez").SetState(m_Object->m_DisplaceZ());

	switch(m_Object->m_Along())
		{
			case CWaveizer::X:
				ToggleButton("alongx").SetState(true);
				break;
			case CWaveizer::Y:
				ToggleButton("alongy").SetState(true);
				break;
			case CWaveizer::Z:
				ToggleButton("alongz").SetState(true);
				break;
			default:
				g_assert_not_reached();
		}
}

void CWaveizerDialog::OnEvent(sdpGtkEvent* Event)
{
	// Sanity checks ...
	g_assert(Event);

	if(Event->Name() == "ondisplacex")
		OnDisplaceX();
	else if(Event->Name() == "ondisplacey")
		OnDisplaceY();
	else if(Event->Name() == "ondisplacez")
		OnDisplaceZ();
	else if(Event->Name() == "onalongx")
		OnAlongX();
	else if(Event->Name() == "onalongy")
		OnAlongY();
	else if(Event->Name() == "onalongz")
		OnAlongZ();
	else
		k3dDialog::OnEvent(Event);
}

void CWaveizerDialog::OnDisplaceX()
{
	m_Object->m_DisplaceX() = ToggleButton("displacex").GetState();
	m_Object->OnModify(false);
}

void CWaveizerDialog::OnDisplaceY()
{
	m_Object->m_DisplaceY() = ToggleButton("displacey").GetState();
	m_Object->OnModify(false);
}

void CWaveizerDialog::OnDisplaceZ()
{
	m_Object->m_DisplaceZ() = ToggleButton("displacez").GetState();
	m_Object->OnModify(false);
}

void CWaveizerDialog::OnAlongX()
{
	if(ToggleButton("alongx").GetState())
		{
			m_Object->m_Along() = CWaveizer::X;
			m_Object->OnModify(false);
		}
}

void CWaveizerDialog::OnAlongY()
{
	if(ToggleButton("alongy").GetState())
		{
			m_Object->m_Along() = CWaveizer::Y;
			m_Object->OnModify(false);
		}
}

void CWaveizerDialog::OnAlongZ()
{
	if(ToggleButton("alongz").GetState())
		{
			m_Object->m_Along() = CWaveizer::Z;
			m_Object->OnModify(false);
		}
}



