// SDPGL
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements all SDPGL functions
		\author Timothy M. Shead (tshead@k-3d.com)
*/

#include <sdpstring.h>

#if defined SDPWIN32
#include <windows.h>
#else // SDPWIN32
#include <GL/glx.h>
#endif // !SDPWIN32

#include <GL/gl.h>
#include <GL/glu.h>

#include <iostream>
#include <strstream>

#include "sdpglprototypes.h"
#include "sdpgltraceprototypes.h"

//////////////////////////////////////////////////////////////////////////////////
// Diagnostics ...

#if defined SDPUNIX

class sdpglVisualList
{
public:
	sdpglVisualList(Display* XDisplay, int XScreen)
	{
		// Sanity checks ...
		g_assert(XDisplay);
		
		m_Display = XDisplay;
		m_Screen = XScreen;

		XVisualInfo match;
		match.screen = XScreen;
		
		m_VisualList = XGetVisualInfo(XDisplay, VisualScreenMask, &match, &m_VisualCount);
	}
	
	~sdpglVisualList()
	{
		if(m_VisualList)
			XFree(m_VisualList);
			
		m_Display = 0;
		m_Screen = 0;
		m_VisualList = 0;
		m_VisualCount = 0;
	}
	
	gulong Count()
	{
		// Sanity checks ...
		g_assert(m_VisualCount >= 0);
		return gulong(m_VisualCount);
	}

	XVisualInfo* operator[](gulong Index)
	{
		// Sanity checks ...
		g_assert(Index < Count());
		return m_VisualList+Index;
	}
			
	XVisualInfo* GetGLXCapableVisual()
	{
		// Sanity checks ...
		g_assert(m_Display);
		g_assert(m_VisualList);
		g_assert(m_VisualCount);

		for(XVisualInfo* visual = m_VisualList; visual < m_VisualList + m_VisualCount; visual++)
			{
				int glxcapable = 0;
				glXGetConfig(m_Display, visual, GLX_USE_GL, &glxcapable);
			
				if(glxcapable)
					return visual;
			}
	
		return 0;
	}
	
protected:
	Display* m_Display;
	int m_Screen;
	XVisualInfo* m_VisualList;
	int m_VisualCount;
};

bool sdpglExists(std::ostream& ErrorOutput)
{
	// Get the X display ...
	Display* display = XOpenDisplay(NULL);
	if(!display)
		{
			ErrorOutput << "*** Could not connect to " << XDisplayName(NULL) << std::endl;
			return false;
		}

	// Get the X screen ...
	int screen = DefaultScreen(display);

	// Make sure the X server supports GLX ...
	if(!glXQueryExtension(display, NULL, NULL))
		{
			ErrorOutput << "*** GLX extensions not supported by X server" << std::endl;
			return false;
		}

	// Try loading the available visuals for this display & screen ...
	sdpglVisualList visuallist(display, screen);
	if(0 == visuallist.Count())
		{
			ErrorOutput << "*** Unable to load any X visuals" << std::endl;
			return false;
		}
		
	// See if any are GLX capable ...
	if(0 == visuallist.GetGLXCapableVisual())
		{
			ErrorOutput << "No GLX capable visual found" << std::endl;
			return false;
		}
		
	return true;
}

void PutList(std::ostream& Stream, const gchar* String)
{
	// Sanity checks ...
	g_assert(Stream.good());

	// If the string is empty, we're done ...
	if(!String)
		return;
		
	if(!strlen(String))
		return;

	// Parse the string into pieces ...
	std::istrstream string(String);
	std::string substring;
	for(string >> substring; !string.eof(); string >> substring)
		Stream << "\t" << substring << std::endl;
}

void PutList(std::ostream& Stream, const GLubyte* String)
{
	PutList(Stream, reinterpret_cast<const gchar*>(String));
}

static SDPCSTRING ClassOf(int XVisualClass)
{
	switch(XVisualClass)
		{
			case StaticGray:
				return "sg";
			case GrayScale:
				return "gs";
			case StaticColor:
				return "sc";
			case PseudoColor:
				return "pc";
			case TrueColor:
				return "tc";
			case DirectColor:
				return "dc";
			default:
				return "??";
		}
}

void sdpglConfiguration(std::ostream& Stream)
{
	// Get the X display ...
	Display* display = XOpenDisplay(NULL);
	g_return_if_fail(display);

	// Get the X screen ...
	int screen = DefaultScreen(display);

	// Make sure the X server supports GLX ...
	g_return_if_fail(glXQueryExtension(display, NULL, NULL));

	// GLX server details ...
	Stream << "GLX Server Vendor: " << glXQueryServerString(display, screen, GLX_VENDOR) << std::endl;
	Stream << "GLX Server Version: " << glXQueryServerString(display, screen, GLX_VERSION) << std::endl;
	Stream << "GLX Server Extensions: " << std::endl;
	PutList(Stream, glXQueryServerString(display, screen, GLX_EXTENSIONS));
	Stream << std::endl;

	// GLX client details ...
	Stream << "GLX Client Vendor: " << glXGetClientString(display, GLX_VENDOR) << std::endl;
	Stream << "GLX Client Version: " << glXGetClientString(display, GLX_VERSION) << std::endl;
	Stream << "GLX Client Extensions: " << std::endl;
	PutList(Stream, glXGetClientString(display, GLX_EXTENSIONS));
	Stream << std::endl;

	// Load available visuals ...
	sdpglVisualList visuallist(display, screen);
	g_return_if_fail(visuallist.Count());

	// Get a GLX capable visual ...
	XVisualInfo* glxcapablevisual = visuallist.GetGLXCapableVisual();
	g_return_if_fail(glxcapablevisual);

	// Create an OpenGL context, so we can grab some info about the implementation ...
	GLXContext context = glXCreateContext(display, glxcapablevisual, 0, GL_TRUE);
	Colormap colormap = XCreateColormap(display, RootWindow(display, glxcapablevisual->screen), glxcapablevisual->visual, AllocNone);
	XSetWindowAttributes attributes;
	attributes.colormap = colormap;
	attributes.border_pixel = 0;
	Window window = XCreateWindow(display, RootWindow(display, glxcapablevisual->screen), 0, 0, 100, 100, 0, glxcapablevisual->depth, InputOutput, glxcapablevisual->visual, CWBorderPixel | CWColormap, &attributes);
	glXMakeCurrent(display, window, context);
	
	// OpenGL details ...
	Stream << "OpenGL Vendor: " << glGetString(GL_VENDOR) << std::endl;
	Stream << "OpenGL Renderer: " << glGetString(GL_RENDERER) << std::endl;
	Stream << "OpenGL Version: " << glGetString(GL_VERSION) << std::endl;
	Stream << "OpenGL Extensions: " << std::endl;
	PutList(Stream, glGetString(GL_EXTENSIONS));
	Stream << std::endl;
	
	// GLU details ...
	Stream << "GLU Version: " << gluGetString(GLenum(GLU_VERSION)) << std::endl;
	Stream << "GLU Extensions: " << std::endl;
	PutList(Stream, gluGetString(GLenum(GLU_EXTENSIONS)));
	Stream << std::endl;
	
	// DRI details ...
	if(glXIsDirect(display, context))
		Stream << "Direct Rendering: supported" << std::endl;
	else
		Stream << "Direct Rendering: not supported" << std::endl;
	Stream << std::endl;

	// Create a listing of available GLX visuals ...
	Stream << "--------------------------------------------------------------------------------------------------------------" << std::endl;
	Stream << "id depth class buffer level rgba double stereo aux red green blue alpha depth stencil ared agreen ablue aalpha" << std::endl;
	Stream << "--------------------------------------------------------------------------------------------------------------" << std::endl;
	
	// For each available visual ...
	for(gulong i = 0; i < visuallist.Count(); i++)
		{
			XVisualInfo* visual = visuallist[i];
			
			int glxcapable;
			glXGetConfig(display, visual, GLX_USE_GL, &glxcapable);
	
			// Skip non-GLX-capable visuals ...
			if(!glxcapable)
				continue;
	
			// Display visual information ...
			Stream << visual->visualid << "   " << visual->depth << "    " << ClassOf(visual->c_class) << "    ";
	
			int value;

			glXGetConfig(display, visual, GLX_BUFFER_SIZE, &value);
			Stream << value << "     ";

			glXGetConfig(display, visual, GLX_LEVEL, &value);
			Stream << value << "    ";
	
			glXGetConfig(display, visual, GLX_RGBA, &value);
			Stream << (value ? "Y" : " ") << "     ";

			glXGetConfig(display, visual, GLX_DOUBLEBUFFER, &value);
			Stream << (value ? "Y" : " ") << "      ";

			glXGetConfig(display, visual, GLX_STEREO, &value);
			Stream << (value ? "Y" : " ") << "     ";

			glXGetConfig(display, visual, GLX_AUX_BUFFERS, &value);
			Stream << value << "   ";

			glXGetConfig(display, visual, GLX_RED_SIZE, &value);
			Stream << value << "    ";

			glXGetConfig(display, visual, GLX_GREEN_SIZE, &value);
			Stream << value << "    ";

			glXGetConfig(display, visual, GLX_BLUE_SIZE, &value);
			Stream << value << "     ";

			glXGetConfig(display, visual, GLX_ALPHA_SIZE, &value);
			Stream << value << "    ";

			glXGetConfig(display, visual, GLX_DEPTH_SIZE, &value);
			Stream << value << "      ";

			glXGetConfig(display, visual, GLX_STENCIL_SIZE, &value);
			Stream << value << "     ";

			glXGetConfig(display, visual, GLX_ACCUM_RED_SIZE, &value);
			Stream << value << "    ";

			glXGetConfig(display, visual, GLX_ACCUM_GREEN_SIZE, &value);
			Stream << value << "     ";

			glXGetConfig(display, visual, GLX_ACCUM_BLUE_SIZE, &value);
			Stream << value << "     ";

			glXGetConfig(display, visual, GLX_ACCUM_ALPHA_SIZE, &value);
			Stream << value << std::endl;
		}

	Stream << std::endl;
}

#else // SDPUNIX

bool sdpglExists(std::ostream& ErrorOutput)
{
	ErrorOutput << "*** sdpglExists hasn't been implemented for this architecture!!!" << std::endl;
	return true;
}

void sdpglConfiguration(std::ostream& Stream)
{
	Stream << "*** sdpglConfiguration hasn't been implemented for this architecture!!!" << std::endl;
}

#endif // !SDPUNIX

//////////////////////////////////////////////////////////////////////////////////
// Trace output ...

/// Stores a pointer to the desired output stream (defaults to cout)
static std::ostream* sdpgl_output_stream = &std::cout;

/// Maintains the current tracing status (on/off) on a stack
static std::stack<bool> sdpgl_output_stack;

/// Performs the actual work of sending text to the trace output
void sdpglRawTrace(SDPCSTRING Text)
{
	// See if tracing has been explicitly disabled ...
	if(sdpgl_output_stack.size())
		if(!sdpgl_output_stack.top())
			return;

	// Otherwise, trace the output ...
	(*sdpgl_output_stream) << Text;
	sdpgl_output_stream->flush();
}

void sdpglBeginFunctionTrace(SDPCSTRING Function, SDPCSTRING Arguments)
{
	sdpglRawTrace(sdpString(Function) + "(" + sdpString(Arguments));
}

void sdpglResultTrace(SDPCSTRING Text)
{
	sdpglRawTrace("=" + sdpString(Text));
}

void sdpglEndFunctionTrace()
{
	sdpglRawTrace(");\n");
}

void sdpglTrace(SDPCSTRING Text)
{
	sdpglRawTrace("//*** " + sdpString(Text) + "\n");
}

void sdpglPushTrace(bool State)
{
	if(State)
		sdpglTrace("SDPGL tracing enabled ...");
	else
		sdpglTrace("SDPGL tracing disabled ...");

	sdpgl_output_stack.push(State);
}

void sdpglPopTrace()
{
	bool current = sdpgl_output_stack.top();

	if(current)
		sdpglTrace("SDPGL tracing disabled ...");
	else
		sdpglTrace("SDPGL tracing re-enabled ...");

	sdpgl_output_stack.pop();
}

void sdpglSetStream(const std::ostream* Stream)
{
	// Sanity checks ...
	g_assert(Stream);
	
	sdpgl_output_stream = const_cast<std::ostream*>(Stream);
}

//////////////////////////////////////////////////////////////////////////////////
// Argument formatting ...

sdpString sdpGLvoid(const GLvoid* Arg)
{
	return sdpFormatString("%x", Arg);
}

sdpString sdpGLsizei(const GLsizei Arg)
{
	return sdpFormatString("%d", Arg);
}

sdpString sdpGLint(const GLint Arg)
{
	return sdpFormatString("%d", Arg);
}

sdpString sdpGLuint(const GLuint Arg)
{
	return sdpFormatString("%u", Arg);
}

sdpString sdpGLbyte(const GLbyte Arg)
{
	return sdpFormatString("%d", Arg);
}

sdpString sdpGLubyte(const GLubyte Arg)
{
	return sdpFormatString("%u", Arg);
}

sdpString sdpGLshort(const GLshort Arg)
{
	return sdpFormatString("%d", Arg);
}

sdpString sdpGLushort(const GLushort Arg)
{
	return sdpFormatString("%u", Arg);
}

sdpString sdpGLdouble(const GLdouble Arg)
{
	return sdpFormatString("%f", Arg);
}

sdpString sdpGLfloat(const GLfloat Arg)
{
	return sdpFormatString("%ff", Arg);
}

sdpString sdpGLclampf(const GLclampf Arg)
{
	return sdpFormatString("%ff", Arg);
}

sdpString sdpGLclampd(const GLclampd Arg)
{
	return sdpFormatString("%f", Arg);
}

sdpString sdpGLbitfield(const GLbitfield Arg)
{
	return sdpFormatString("0x%x", Arg);
}

sdpString sdpGLboolean(const GLboolean Arg)
{
	return sdpString(Arg ? "GL_TRUE" : "GL_FALSE");
}

sdpString sdpGLMatrixd(const GLdouble* Arg)
{
	sdpString result;
	result += "[" + sdpGLdouble(Arg[0]) + "," + sdpGLdouble(Arg[1]) + "," + sdpGLdouble(Arg[2]) + "," + sdpGLdouble(Arg[3]) + "]";
	result += "[" + sdpGLdouble(Arg[4]) + "," + sdpGLdouble(Arg[5]) + "," + sdpGLdouble(Arg[6]) + "," + sdpGLdouble(Arg[7]) + "]";
	result += "[" + sdpGLdouble(Arg[8]) + "," + sdpGLdouble(Arg[9]) + "," + sdpGLdouble(Arg[10]) + "," + sdpGLdouble(Arg[11]) + "]";
	result += "[" + sdpGLdouble(Arg[12]) + "," + sdpGLdouble(Arg[13]) + "," + sdpGLdouble(Arg[14]) + "," + sdpGLdouble(Arg[15]) + "]";

	return result;
}

sdpString sdpGLenum(const GLenum Arg)
{
	switch(Arg)
		{
			case 0x0000:
				return "GL_POINTS";
			case 0x0001:
				return "GL_LINES";
			case 0x0002:
				return "GL_LINE_LOOP";
			case 0x0003:
				return "GL_LINE_STRIP";
			case 0x0004:
				return "GL_TRIANGLES";
			case 0x0005:
				return "GL_TRIANGLE_STRIP";
			case 0x0006:
				return "GL_TRIANGLE_FAN";
			case 0x0007:
				return "GL_QUADS";
			case 0x0008:
				return "GL_QUAD_STRIP";
			case 0x0009:
				return "GL_POLYGON";
			case 0x0100:
				return "GL_ACCUM";
			case 0x0101:
				return "GL_LOAD";
			case 0x0102:
				return "GL_RETURN";
			case 0x0103:
				return "GL_MULT";
			case 0x0104:
				return "GL_ADD";
			case 0x0200:
				return "GL_NEVER";
			case 0x0201:
				return "GL_LESS";
			case 0x0202:
				return "GL_EQUAL";
			case 0x0203:
				return "GL_LEQUAL";
			case 0x0204:
				return "GL_GREATER";
			case 0x0205:
				return "GL_NOTEQUAL";
			case 0x0206:
				return "GL_GEQUAL";
			case 0x0207:
				return "GL_ALWAYS";
			case 0x0300:
				return "GL_SRC_COLOR";
			case 0x0301:
				return "GL_ONE_MINUS_SRC_COLOR";
			case 0x0302:
				return "GL_SRC_ALPHA";
			case 0x0303:
				return "GL_ONE_MINUS_SRC_ALPHA";
			case 0x0304:
				return "GL_DST_ALPHA";
			case 0x0305:
				return "GL_ONE_MINUS_DST_ALPHA";
			case 0x0306:
				return "GL_DST_COLOR";
			case 0x0307:
				return "GL_ONE_MINUS_DST_COLOR";
			case 0x0308:
				return "GL_SRC_ALPHA_SATURATE";
			case 0x0400:
				return "GL_FRONT_LEFT";
			case 0x0401:
				return "GL_FRONT_RIGHT";
			case 0x0402:
				return "GL_BACK_LEFT";
			case 0x0403:
				return "GL_BACK_RIGHT";
			case 0x0404:
				return "GL_FRONT";
			case 0x0405:
				return "GL_BACK";
			case 0x0406:
				return "GL_LEFT";
			case 0x0407:
				return "GL_RIGHT";
			case 0x0408:
				return "GL_FRONT_AND_BACK";
			case 0x0409:
				return "GL_AUX0";
			case 0x040A:
				return "GL_AUX1";
			case 0x040B:
				return "GL_AUX2";
			case 0x040C:
				return "GL_AUX3";
			case 0x0500:
				return "GL_INVALID_ENUM";
			case 0x0501:
				return "GL_INVALID_VALUE";
			case 0x0502:
				return "GL_INVALID_OPERATION";
			case 0x0503:
				return "GL_STACK_OVERFLOW";
			case 0x0504:
				return "GL_STACK_UNDERFLOW";
			case 0x0505:
				return "GL_OUT_OF_MEMORY";
			case 0x0600:
				return "GL_2D";
			case 0x0601:
				return "GL_3D";
			case 0x0602:
				return "GL_3D_COLOR";
			case 0x0603:
				return "GL_3D_COLOR_TEXTURE";
			case 0x0604:
				return "GL_4D_COLOR_TEXTURE";
			case 0x0700:
				return "GL_PASS_THROUGH_TOKEN";
			case 0x0701:
				return "GL_POINT_TOKEN";
			case 0x0702:
				return "GL_LINE_TOKEN";
			case 0x0703:
				return "GL_POLYGON_TOKEN";
			case 0x0704:
				return "GL_BITMAP_TOKEN";
			case 0x0705:
				return "GL_DRAW_PIXEL_TOKEN";
			case 0x0706:
				return "GL_COPY_PIXEL_TOKEN";
			case 0x0707:
				return "GL_LINE_RESET_TOKEN";
			case 0x0800:
				return "GL_EXP";
			case 0x0801:
				return "GL_EXP2";
			case 0x0900:
				return "GL_CW";
			case 0x0901:
				return "GL_CCW";
			case 0x0A00:
				return "GL_COEFF";
			case 0x0A01:
				return "GL_ORDER";
			case 0x0A02:
				return "GL_DOMAIN";
			case 0x0B00:
				return "GL_CURRENT_COLOR";
			case 0x0B01:
				return "GL_CURRENT_INDEX";
			case 0x0B02:
				return "GL_CURRENT_NORMAL";
			case 0x0B03:
				return "GL_CURRENT_TEXTURE_COORDS";
			case 0x0B04:
				return "GL_CURRENT_RASTER_COLOR";
			case 0x0B05:
				return "GL_CURRENT_RASTER_INDEX";
			case 0x0B06:
				return "GL_CURRENT_RASTER_TEXTURE_COORDS";
			case 0x0B07:
				return "GL_CURRENT_RASTER_POSITION";
			case 0x0B08:
				return "GL_CURRENT_RASTER_POSITION_VALID";
			case 0x0B09:
				return "GL_CURRENT_RASTER_DISTANCE";
			case 0x0B10:
				return "GL_POINT_SMOOTH";
			case 0x0B11:
				return "GL_POINT_SIZE";
			case 0x0B12:
				return "GL_POINT_SIZE_RANGE";
			case 0x0B13:
				return "GL_POINT_SIZE_GRANULARITY";
			case 0x0B20:
				return "GL_LINE_SMOOTH";
			case 0x0B21:
				return "GL_LINE_WIDTH";
			case 0x0B22:
				return "GL_LINE_WIDTH_RANGE";
			case 0x0B23:
				return "GL_LINE_WIDTH_GRANULARITY";
			case 0x0B24:
				return "GL_LINE_STIPPLE";
			case 0x0B25:
				return "GL_LINE_STIPPLE_PATTERN";
			case 0x0B26:
				return "GL_LINE_STIPPLE_REPEAT";
			case 0x0B30:
				return "GL_LIST_MODE";
			case 0x0B31:
				return "GL_MAX_LIST_NESTING";
			case 0x0B32:
				return "GL_LIST_BASE";
			case 0x0B33:
				return "GL_LIST_INDEX";
			case 0x0B40:
				return "GL_POLYGON_MODE";
			case 0x0B41:
				return "GL_POLYGON_SMOOTH";
			case 0x0B42:
				return "GL_POLYGON_STIPPLE";
			case 0x0B43:
				return "GL_EDGE_FLAG";
			case 0x0B44:
				return "GL_CULL_FACE";
			case 0x0B45:
				return "GL_CULL_FACE_MODE";
			case 0x0B46:
				return "GL_FRONT_FACE";
			case 0x0B50:
				return "GL_LIGHTING";
			case 0x0B51:
				return "GL_LIGHT_MODEL_LOCAL_VIEWER";
			case 0x0B52:
				return "GL_LIGHT_MODEL_TWO_SIDE";
			case 0x0B53:
				return "GL_LIGHT_MODEL_AMBIENT";
			case 0x0B54:
				return "GL_SHADE_MODEL";
			case 0x0B55:
				return "GL_COLOR_MATERIAL_FACE";
			case 0x0B56:
				return "GL_COLOR_MATERIAL_PARAMETER";
			case 0x0B57:
				return "GL_COLOR_MATERIAL";
			case 0x0B60:
				return "GL_FOG";
			case 0x0B61:
				return "GL_FOG_INDEX";
			case 0x0B62:
				return "GL_FOG_DENSITY";
			case 0x0B63:
				return "GL_FOG_START";
			case 0x0B64:
				return "GL_FOG_END";
			case 0x0B65:
				return "GL_FOG_MODE";
			case 0x0B66:
				return "GL_FOG_COLOR";
			case 0x0B70:
				return "GL_DEPTH_RANGE";
			case 0x0B71:
				return "GL_DEPTH_TEST";
			case 0x0B72:
				return "GL_DEPTH_WRITEMASK";
			case 0x0B73:
				return "GL_DEPTH_CLEAR_VALUE";
			case 0x0B74:
				return "GL_DEPTH_FUNC";
			case 0x0B80:
				return "GL_ACCUM_CLEAR_VALUE";
			case 0x0B90:
				return "GL_STENCIL_TEST";
			case 0x0B91:
				return "GL_STENCIL_CLEAR_VALUE";
			case 0x0B92:
				return "GL_STENCIL_FUNC";
			case 0x0B93:
				return "GL_STENCIL_VALUE_MASK";
			case 0x0B94:
				return "GL_STENCIL_FAIL";
			case 0x0B95:
				return "GL_STENCIL_PASS_DEPTH_FAIL";
			case 0x0B96:
				return "GL_STENCIL_PASS_DEPTH_PASS";
			case 0x0B97:
				return "GL_STENCIL_REF";
			case 0x0B98:
				return "GL_STENCIL_WRITEMASK";
			case 0x0BA0:
				return "GL_MATRIX_MODE";
			case 0x0BA1:
				return "GL_NORMALIZE";
			case 0x0BA2:
				return "GL_VIEWPORT";
			case 0x0BA3:
				return "GL_MODELVIEW_STACK_DEPTH";
			case 0x0BA4:
				return "GL_PROJECTION_STACK_DEPTH";
			case 0x0BA5:
				return "GL_TEXTURE_STACK_DEPTH";
			case 0x0BA6:
				return "GL_MODELVIEW_MATRIX";
			case 0x0BA7:
				return "GL_PROJECTION_MATRIX";
			case 0x0BA8:
				return "GL_TEXTURE_MATRIX";
			case 0x0BB0:
				return "GL_ATTRIB_STACK_DEPTH";
			case 0x0BB1:
				return "GL_CLIENT_ATTRIB_STACK_DEPTH";
			case 0x0BC0:
				return "GL_ALPHA_TEST";
			case 0x0BC1:
				return "GL_ALPHA_TEST_FUNC";
			case 0x0BC2:
				return "GL_ALPHA_TEST_REF";
			case 0x0BD0:
				return "GL_DITHER";
			case 0x0BE0:
				return "GL_BLEND_DST";
			case 0x0BE1:
				return "GL_BLEND_SRC";
			case 0x0BE2:
				return "GL_BLEND";
			case 0x0BF0:
				return "GL_LOGIC_OP_MODE";
			case 0x0BF1:
				return "GL_INDEX_LOGIC_OP";
			case 0x0BF2:
				return "GL_COLOR_LOGIC_OP";
			case 0x0C00:
				return "GL_AUX_BUFFERS";
			case 0x0C01:
				return "GL_DRAW_BUFFER";
			case 0x0C02:
				return "GL_READ_BUFFER";
			case 0x0C10:
				return "GL_SCISSOR_BOX";
			case 0x0C11:
				return "GL_SCISSOR_TEST";
			case 0x0C20:
				return "GL_INDEX_CLEAR_VALUE";
			case 0x0C21:
				return "GL_INDEX_WRITEMASK";
			case 0x0C22:
				return "GL_COLOR_CLEAR_VALUE";
			case 0x0C23:
				return "GL_COLOR_WRITEMASK";
			case 0x0C30:
				return "GL_INDEX_MODE";
			case 0x0C31:
				return "GL_RGBA_MODE";
			case 0x0C32:
				return "GL_DOUBLEBUFFER";
			case 0x0C33:
				return "GL_STEREO";
			case 0x0C40:
				return "GL_RENDER_MODE";
			case 0x0C50:
				return "GL_PERSPECTIVE_CORRECTION_HINT";
			case 0x0C51:
				return "GL_POINT_SMOOTH_HINT";
			case 0x0C52:
				return "GL_LINE_SMOOTH_HINT";
			case 0x0C53:
				return "GL_POLYGON_SMOOTH_HINT";
			case 0x0C54:
				return "GL_FOG_HINT";
			case 0x0C60:
				return "GL_TEXTURE_GEN_S";
			case 0x0C61:
				return "GL_TEXTURE_GEN_T";
			case 0x0C62:
				return "GL_TEXTURE_GEN_R";
			case 0x0C63:
				return "GL_TEXTURE_GEN_Q";
			case 0x0C70:
				return "GL_PIXEL_MAP_I_TO_I";
			case 0x0C71:
				return "GL_PIXEL_MAP_S_TO_S";
			case 0x0C72:
				return "GL_PIXEL_MAP_I_TO_R";
			case 0x0C73:
				return "GL_PIXEL_MAP_I_TO_G";
			case 0x0C74:
				return "GL_PIXEL_MAP_I_TO_B";
			case 0x0C75:
				return "GL_PIXEL_MAP_I_TO_A";
			case 0x0C76:
				return "GL_PIXEL_MAP_R_TO_R";
			case 0x0C77:
				return "GL_PIXEL_MAP_G_TO_G";
			case 0x0C78:
				return "GL_PIXEL_MAP_B_TO_B";
			case 0x0C79:
				return "GL_PIXEL_MAP_A_TO_A";
			case 0x0CB0:
				return "GL_PIXEL_MAP_I_TO_I_SIZE";
			case 0x0CB1:
				return "GL_PIXEL_MAP_S_TO_S_SIZE";
			case 0x0CB2:
				return "GL_PIXEL_MAP_I_TO_R_SIZE";
			case 0x0CB3:
				return "GL_PIXEL_MAP_I_TO_G_SIZE";
			case 0x0CB4:
				return "GL_PIXEL_MAP_I_TO_B_SIZE";
			case 0x0CB5:
				return "GL_PIXEL_MAP_I_TO_A_SIZE";
			case 0x0CB6:
				return "GL_PIXEL_MAP_R_TO_R_SIZE";
			case 0x0CB7:
				return "GL_PIXEL_MAP_G_TO_G_SIZE";
			case 0x0CB8:
				return "GL_PIXEL_MAP_B_TO_B_SIZE";
			case 0x0CB9:
				return "GL_PIXEL_MAP_A_TO_A_SIZE";
			case 0x0CF0:
				return "GL_UNPACK_SWAP_BYTES";
			case 0x0CF1:
				return "GL_UNPACK_LSB_FIRST";
			case 0x0CF2:
				return "GL_UNPACK_ROW_LENGTH";
			case 0x0CF3:
				return "GL_UNPACK_SKIP_ROWS";
			case 0x0CF4:
				return "GL_UNPACK_SKIP_PIXELS";
			case 0x0CF5:
				return "GL_UNPACK_ALIGNMENT";
			case 0x0D00:
				return "GL_PACK_SWAP_BYTES";
			case 0x0D01:
				return "GL_PACK_LSB_FIRST";
			case 0x0D02:
				return "GL_PACK_ROW_LENGTH";
			case 0x0D03:
				return "GL_PACK_SKIP_ROWS";
			case 0x0D04:
				return "GL_PACK_SKIP_PIXELS";
			case 0x0D05:
				return "GL_PACK_ALIGNMENT";
			case 0x0D10:
				return "GL_MAP_COLOR";
			case 0x0D11:
				return "GL_MAP_STENCIL";
			case 0x0D12:
				return "GL_INDEX_SHIFT";
			case 0x0D13:
				return "GL_INDEX_OFFSET";
			case 0x0D14:
				return "GL_RED_SCALE";
			case 0x0D15:
				return "GL_RED_BIAS";
			case 0x0D16:
				return "GL_ZOOM_X";
			case 0x0D17:
				return "GL_ZOOM_Y";
			case 0x0D18:
				return "GL_GREEN_SCALE";
			case 0x0D19:
				return "GL_GREEN_BIAS";
			case 0x0D1A:
				return "GL_BLUE_SCALE";
			case 0x0D1B:
				return "GL_BLUE_BIAS";
			case 0x0D1C:
				return "GL_ALPHA_SCALE";
			case 0x0D1D:
				return "GL_ALPHA_BIAS";
			case 0x0D1E:
				return "GL_DEPTH_SCALE";
			case 0x0D1F:
				return "GL_DEPTH_BIAS";
			case 0x0D30:
				return "GL_MAX_EVAL_ORDER";
			case 0x0D31:
				return "GL_MAX_LIGHTS";
			case 0x0D32:
				return "GL_MAX_CLIP_PLANES";
			case 0x0D33:
				return "GL_MAX_TEXTURE_SIZE";
			case 0x0D34:
				return "GL_MAX_PIXEL_MAP_TABLE";
			case 0x0D35:
				return "GL_MAX_ATTRIB_STACK_DEPTH";
			case 0x0D36:
				return "GL_MAX_MODELVIEW_STACK_DEPTH";
			case 0x0D37:
				return "GL_MAX_NAME_STACK_DEPTH";
			case 0x0D38:
				return "GL_MAX_PROJECTION_STACK_DEPTH";
			case 0x0D39:
				return "GL_MAX_TEXTURE_STACK_DEPTH";
			case 0x0D3A:
				return "GL_MAX_VIEWPORT_DIMS";
			case 0x0D3B:
				return "GL_MAX_CLIENT_ATTRIB_STACK_DEPTH";
			case 0x0D50:
				return "GL_SUBPIXEL_BITS";
			case 0x0D51:
				return "GL_INDEX_BITS";
			case 0x0D52:
				return "GL_RED_BITS";
			case 0x0D53:
				return "GL_GREEN_BITS";
			case 0x0D54:
				return "GL_BLUE_BITS";
			case 0x0D55:
				return "GL_ALPHA_BITS";
			case 0x0D56:
				return "GL_DEPTH_BITS";
			case 0x0D57:
				return "GL_STENCIL_BITS";
			case 0x0D58:
				return "GL_ACCUM_RED_BITS";
			case 0x0D59:
				return "GL_ACCUM_GREEN_BITS";
			case 0x0D5A:
				return "GL_ACCUM_BLUE_BITS";
			case 0x0D5B:
				return "GL_ACCUM_ALPHA_BITS";
			case 0x0D70:
				return "GL_NAME_STACK_DEPTH";
			case 0x0D80:
				return "GL_AUTO_NORMAL";
			case 0x0D90:
				return "GL_MAP1_COLOR_4";
			case 0x0D91:
				return "GL_MAP1_INDEX";
			case 0x0D92:
				return "GL_MAP1_NORMAL";
			case 0x0D93:
				return "GL_MAP1_TEXTURE_COORD_1";
			case 0x0D94:
				return "GL_MAP1_TEXTURE_COORD_2";
			case 0x0D95:
				return "GL_MAP1_TEXTURE_COORD_3";
			case 0x0D96:
				return "GL_MAP1_TEXTURE_COORD_4";
			case 0x0D97:
				return "GL_MAP1_VERTEX_3";
			case 0x0D98:
				return "GL_MAP1_VERTEX_4";
			case 0x0DB0:
				return "GL_MAP2_COLOR_4";
			case 0x0DB1:
				return "GL_MAP2_INDEX";
			case 0x0DB2:
				return "GL_MAP2_NORMAL";
			case 0x0DB3:
				return "GL_MAP2_TEXTURE_COORD_1";
			case 0x0DB4:
				return "GL_MAP2_TEXTURE_COORD_2";
			case 0x0DB5:
				return "GL_MAP2_TEXTURE_COORD_3";
			case 0x0DB6:
				return "GL_MAP2_TEXTURE_COORD_4";
			case 0x0DB7:
				return "GL_MAP2_VERTEX_3";
			case 0x0DB8:
				return "GL_MAP2_VERTEX_4";
			case 0x0DD0:
				return "GL_MAP1_GRID_DOMAIN";
			case 0x0DD1:
				return "GL_MAP1_GRID_SEGMENTS";
			case 0x0DD2:
				return "GL_MAP2_GRID_DOMAIN";
			case 0x0DD3:
				return "GL_MAP2_GRID_SEGMENTS";
			case 0x0DE0:
				return "GL_TEXTURE_1D";
			case 0x0DE1:
				return "GL_TEXTURE_2D";
			case 0x0DF0:
				return "GL_FEEDBACK_BUFFER_POINTER";
			case 0x0DF1:
				return "GL_FEEDBACK_BUFFER_SIZE";
			case 0x0DF2:
				return "GL_FEEDBACK_BUFFER_TYPE";
			case 0x0DF3:
				return "GL_SELECTION_BUFFER_POINTER";
			case 0x0DF4:
				return "GL_SELECTION_BUFFER_SIZE";
			case 0x1000:
				return "GL_TEXTURE_WIDTH";
			case 0x1001:
				return "GL_TEXTURE_HEIGHT";
			case 0x1003:
				return "GL_TEXTURE_INTERNAL_FORMAT";
			case 0x1004:
				return "GL_TEXTURE_BORDER_COLOR";
			case 0x1005:
				return "GL_TEXTURE_BORDER";
			case 0x1100:
				return "GL_DONT_CARE";
			case 0x1101:
				return "GL_FASTEST";
			case 0x1102:
				return "GL_NICEST";
			case 0x1200:
				return "GL_AMBIENT";
			case 0x1201:
				return "GL_DIFFUSE";
			case 0x1202:
				return "GL_SPECULAR";
			case 0x1203:
				return "GL_POSITION";
			case 0x1204:
				return "GL_SPOT_DIRECTION";
			case 0x1205:
				return "GL_SPOT_EXPONENT";
			case 0x1206:
				return "GL_SPOT_CUTOFF";
			case 0x1207:
				return "GL_CONSTANT_ATTENUATION";
			case 0x1208:
				return "GL_LINEAR_ATTENUATION";
			case 0x1209:
				return "GL_QUADRATIC_ATTENUATION";
			case 0x1300:
				return "GL_COMPILE";
			case 0x1301:
				return "GL_COMPILE_AND_EXECUTE";
			case 0x1400:
				return "GL_BYTE";
			case 0x1401:
				return "GL_UNSIGNED_BYTE";
			case 0x1402:
				return "GL_SHORT";
			case 0x1403:
				return "GL_UNSIGNED_SHORT";
			case 0x1404:
				return "GL_INT";
			case 0x1405:
				return "GL_UNSIGNED_INT";
			case 0x1406:
				return "GL_FLOAT";
			case 0x1407:
				return "GL_2_BYTES";
			case 0x1408:
				return "GL_3_BYTES";
			case 0x1409:
				return "GL_4_BYTES";
			case 0x140A:
				return "GL_DOUBLE";
			case 0x1500:
				return "GL_CLEAR";
			case 0x1501:
				return "GL_AND";
			case 0x1502:
				return "GL_AND_REVERSE";
			case 0x1503:
				return "GL_COPY";
			case 0x1504:
				return "GL_AND_INVERTED";
			case 0x1505:
				return "GL_NOOP";
			case 0x1506:
				return "GL_XOR";
			case 0x1507:
				return "GL_OR";
			case 0x1508:
				return "GL_NOR";
			case 0x1509:
				return "GL_EQUIV";
			case 0x150A:
				return "GL_INVERT";
			case 0x150B:
				return "GL_OR_REVERSE";
			case 0x150C:
				return "GL_COPY_INVERTED";
			case 0x150D:
				return "GL_OR_INVERTED";
			case 0x150E:
				return "GL_NAND";
			case 0x150F:
				return "GL_SET";
			case 0x1600:
				return "GL_EMISSION";
			case 0x1601:
				return "GL_SHININESS";
			case 0x1602:
				return "GL_AMBIENT_AND_DIFFUSE";
			case 0x1603:
				return "GL_COLOR_INDEXES";
			case 0x1700:
				return "GL_MODELVIEW";
			case 0x1701:
				return "GL_PROJECTION";
			case 0x1702:
				return "GL_TEXTURE";
			case 0x1800:
				return "GL_COLOR";
			case 0x1801:
				return "GL_DEPTH";
			case 0x1802:
				return "GL_STENCIL";
			case 0x1900:
				return "GL_COLOR_INDEX";
			case 0x1901:
				return "GL_STENCIL_INDEX";
			case 0x1902:
				return "GL_DEPTH_COMPONENT";
			case 0x1903:
				return "GL_RED";
			case 0x1904:
				return "GL_GREEN";
			case 0x1905:
				return "GL_BLUE";
			case 0x1906:
				return "GL_ALPHA";
			case 0x1907:
				return "GL_RGB";
			case 0x1908:
				return "GL_RGBA";
			case 0x1909:
				return "GL_LUMINANCE";
			case 0x190A:
				return "GL_LUMINANCE_ALPHA";
			case 0x1A00:
				return "GL_BITMAP";
			case 0x1B00:
				return "GL_POINT";
			case 0x1B01:
				return "GL_LINE";
			case 0x1B02:
				return "GL_FILL";
			case 0x1C00:
				return "GL_RENDER";
			case 0x1C01:
				return "GL_FEEDBACK";
			case 0x1C02:
				return "GL_SELECT";
			case 0x1D00:
				return "GL_FLAT";
			case 0x1D01:
				return "GL_SMOOTH";
			case 0x1E00:
				return "GL_KEEP";
			case 0x1E01:
				return "GL_REPLACE";
			case 0x1E02:
				return "GL_INCR";
			case 0x1E03:
				return "GL_DECR";
			case 0x1F00:
				return "GL_VENDOR";
			case 0x1F01:
				return "GL_RENDERER";
			case 0x1F02:
				return "GL_VERSION";
			case 0x1F03:
				return "GL_EXTENSIONS";
			case 0x2000:
				return "GL_S";
			case 0x2001:
				return "GL_T";
			case 0x2002:
				return "GL_R";
			case 0x2003:
				return "GL_Q";
			case 0x2100:
				return "GL_MODULATE";
			case 0x2101:
				return "GL_DECAL";
			case 0x2200:
				return "GL_TEXTURE_ENV_MODE";
			case 0x2201:
				return "GL_TEXTURE_ENV_COLOR";
			case 0x2300:
				return "GL_TEXTURE_ENV";
			case 0x2400:
				return "GL_EYE_LINEAR";
			case 0x2401:
				return "GL_OBJECT_LINEAR";
			case 0x2402:
				return "GL_SPHERE_MAP";
			case 0x2500:
				return "GL_TEXTURE_GEN_MODE";
			case 0x2501:
				return "GL_OBJECT_PLANE";
			case 0x2502:
				return "GL_EYE_PLANE";
			case 0x2600:
				return "GL_NEAREST";
			case 0x2601:
				return "GL_LINEAR";
			case 0x2700:
				return "GL_NEAREST_MIPMAP_NEAREST";
			case 0x2701:
				return "GL_LINEAR_MIPMAP_NEAREST";
			case 0x2702:
				return "GL_NEAREST_MIPMAP_LINEAR";
			case 0x2703:
				return "GL_LINEAR_MIPMAP_LINEAR";
			case 0x2800:
				return "GL_TEXTURE_MAG_FILTER";
			case 0x2801:
				return "GL_TEXTURE_MIN_FILTER";
			case 0x2802:
				return "GL_TEXTURE_WRAP_S";
			case 0x2803:
				return "GL_TEXTURE_WRAP_T";
			case 0x2900:
				return "GL_CLAMP";
			case 0x2901:
				return "GL_REPEAT";
			case 0x2A00:
				return "GL_POLYGON_OFFSET_UNITS";
			case 0x2A01:
				return "GL_POLYGON_OFFSET_POINT";
			case 0x2A02:
				return "GL_POLYGON_OFFSET_LINE";
			case 0x2A10:
				return "GL_R3_G3_B2";
			case 0x2A20:
				return "GL_V2F";
			case 0x2A21:
				return "GL_V3F";
			case 0x2A22:
				return "GL_C4UB_V2F";
			case 0x2A23:
				return "GL_C4UB_V3F";
			case 0x2A24:
				return "GL_C3F_V3F";
			case 0x2A25:
				return "GL_N3F_V3F";
			case 0x2A26:
				return "GL_C4F_N3F_V3F";
			case 0x2A27:
				return "GL_T2F_V3F";
			case 0x2A28:
				return "GL_T4F_V4F";
			case 0x2A29:
				return "GL_T2F_C4UB_V3F";
			case 0x2A2A:
				return "GL_T2F_C3F_V3F";
			case 0x2A2B:
				return "GL_T2F_N3F_V3F";
			case 0x2A2C:
				return "GL_T2F_C4F_N3F_V3F";
			case 0x2A2D:
				return "GL_T4F_C4F_N3F_V4F";
			case 0x3000:
				return "GL_CLIP_PLANE0";
			case 0x3001:
				return "GL_CLIP_PLANE1";
			case 0x3002:
				return "GL_CLIP_PLANE2";
			case 0x3003:
				return "GL_CLIP_PLANE3";
			case 0x3004:
				return "GL_CLIP_PLANE4";
			case 0x3005:
				return "GL_CLIP_PLANE5";
			case 0x4000:
				return "GL_LIGHT0";
			case 0x4001:
				return "GL_LIGHT1";
			case 0x4002:
				return "GL_LIGHT2";
			case 0x4003:
				return "GL_LIGHT3";
			case 0x4004:
				return "GL_LIGHT4";
			case 0x4005:
				return "GL_LIGHT5";
			case 0x4006:
				return "GL_LIGHT6";
			case 0x4007:
				return "GL_LIGHT7";
			case 0x8037:
				return "GL_POLYGON_OFFSET_FILL";
			case 0x8038:
				return "GL_POLYGON_OFFSET_FACTOR";
			case 0x803B:
				return "GL_ALPHA4";
			case 0x803C:
				return "GL_ALPHA8";
			case 0x803D:
				return "GL_ALPHA12";
			case 0x803E:
				return "GL_ALPHA16";
			case 0x803F:
				return "GL_LUMINANCE4";
			case 0x8040:
				return "GL_LUMINANCE8";
			case 0x8041:
				return "GL_LUMINANCE12";
			case 0x8042:
				return "GL_LUMINANCE16";
			case 0x8043:
				return "GL_LUMINANCE4_ALPHA4";
			case 0x8044:
				return "GL_LUMINANCE6_ALPHA2";
			case 0x8045:
				return "GL_LUMINANCE8_ALPHA8";
			case 0x8046:
				return "GL_LUMINANCE12_ALPHA4";
			case 0x8047:
				return "GL_LUMINANCE12_ALPHA12";
			case 0x8048:
				return "GL_LUMINANCE16_ALPHA16";
			case 0x8049:
				return "GL_INTENSITY";
			case 0x804A:
				return "GL_INTENSITY4";
			case 0x804B:
				return "GL_INTENSITY8";
			case 0x804C:
				return "GL_INTENSITY12";
			case 0x804D:
				return "GL_INTENSITY16";
			case 0x804F:
				return "GL_RGB4";
			case 0x8050:
				return "GL_RGB5";
			case 0x8051:
				return "GL_RGB8";
			case 0x8052:
				return "GL_RGB10";
			case 0x8053:
				return "GL_RGB12";
			case 0x8054:
				return "GL_RGB16";
			case 0x8055:
				return "GL_RGBA2";
			case 0x8056:
				return "GL_RGBA4";
			case 0x8057:
				return "GL_RGB5_A1";
			case 0x8058:
				return "GL_RGBA8";
			case 0x8059:
				return "GL_RGB10_A2";
			case 0x805A:
				return "GL_RGBA12";
			case 0x805B:
				return "GL_RGBA16";
			case 0x805C:
				return "GL_TEXTURE_RED_SIZE";
			case 0x805D:
				return "GL_TEXTURE_GREEN_SIZE";
			case 0x805E:
				return "GL_TEXTURE_BLUE_SIZE";
			case 0x805F:
				return "GL_TEXTURE_ALPHA_SIZE";
			case 0x8060:
				return "GL_TEXTURE_LUMINANCE_SIZE";
			case 0x8061:
				return "GL_TEXTURE_INTENSITY_SIZE";
			case 0x8063:
				return "GL_PROXY_TEXTURE_1D";
			case 0x8064:
				return "GL_PROXY_TEXTURE_2D";
			case 0x8066:
				return "GL_TEXTURE_PRIORITY";
			case 0x8067:
				return "GL_TEXTURE_RESIDENT";
			case 0x8068:
				return "GL_TEXTURE_BINDING_1D";
			case 0x8069:
				return "GL_TEXTURE_BINDING_2D";
			case 0x8074:
				return "GL_VERTEX_ARRAY";
			case 0x8075:
				return "GL_NORMAL_ARRAY";
			case 0x8076:
				return "GL_COLOR_ARRAY";
			case 0x8077:
				return "GL_INDEX_ARRAY";
			case 0x8078:
				return "GL_TEXTURE_COORD_ARRAY";
			case 0x8079:
				return "GL_EDGE_FLAG_ARRAY";
			case 0x807A:
				return "GL_VERTEX_ARRAY_SIZE";
			case 0x807B:
				return "GL_VERTEX_ARRAY_TYPE";
			case 0x807C:
				return "GL_VERTEX_ARRAY_STRIDE";
			case 0x807E:
				return "GL_NORMAL_ARRAY_TYPE";
			case 0x807F:
				return "GL_NORMAL_ARRAY_STRIDE";
			case 0x8081:
				return "GL_COLOR_ARRAY_SIZE";
			case 0x8082:
				return "GL_COLOR_ARRAY_TYPE";
			case 0x8083:
				return "GL_COLOR_ARRAY_STRIDE";
			case 0x8085:
				return "GL_INDEX_ARRAY_TYPE";
			case 0x8086:
				return "GL_INDEX_ARRAY_STRIDE";
			case 0x8088:
				return "GL_TEXTURE_COORD_ARRAY_SIZE";
			case 0x8089:
				return "GL_TEXTURE_COORD_ARRAY_TYPE";
			case 0x808A:
				return "GL_TEXTURE_COORD_ARRAY_STRIDE";
			case 0x808C:
				return "GL_EDGE_FLAG_ARRAY_STRIDE";
			case 0x808E:
				return "GL_VERTEX_ARRAY_POINTER";
			case 0x808F:
				return "GL_NORMAL_ARRAY_POINTER";
			case 0x8090:
				return "GL_COLOR_ARRAY_POINTER";
			case 0x8091:
				return "GL_INDEX_ARRAY_POINTER";
			case 0x8092:
				return "GL_TEXTURE_COORD_ARRAY_POINTER";
			case 0x8093:
				return "GL_EDGE_FLAG_ARRAY_POINTER";
			case 0x80D8:
				return "GL_COLOR_TABLE_FORMAT_EXT";
			case 0x80D9:
				return "GL_COLOR_TABLE_WIDTH_EXT";
			case 0x80DA:
				return "GL_COLOR_TABLE_RED_SIZE_EXT";
			case 0x80DB:
				return "GL_COLOR_TABLE_GREEN_SIZE_EXT";
			case 0x80DC:
				return "GL_COLOR_TABLE_BLUE_SIZE_EXT";
			case 0x80DD:
				return "GL_COLOR_TABLE_ALPHA_SIZE_EXT";
			case 0x80DE:
				return "GL_COLOR_TABLE_LUMINANCE_SIZE_EXT";
			case 0x80DF:
				return "GL_COLOR_TABLE_INTENSITY_SIZE_EXT";
			case 0x80E0:
				return "GL_BGR_EXT";
			case 0x80E1:
				return "GL_BGRA_EXT";
			case 0x80E2:
				return "GL_COLOR_INDEX1_EXT";
			case 0x80E3:
				return "GL_COLOR_INDEX2_EXT";
			case 0x80E4:
				return "GL_COLOR_INDEX4_EXT";
			case 0x80E5:
				return "GL_COLOR_INDEX8_EXT";
			case 0x80E6:
				return "GL_COLOR_INDEX12_EXT";
			case 0x80E7:
				return "GL_COLOR_INDEX16_EXT";
			default:
				return "*** Unknown OpenGL enum!! ***";
		}
}




//////////////////////////////////////////////////////////////////////////////////
// Implementation ...

void sdpglAccum(GLenum op,GLfloat value)
{
	sdpglBeginFunctionTrace("glAccum", sdpGLenum(op) + "," + sdpGLfloat(value));
	glAccum(op,value);
	sdpglEndFunctionTrace();
}

void sdpglAlphaFunc(GLenum func,GLclampf ref)
{
	sdpglBeginFunctionTrace("glAlphaFunc", sdpGLenum(func) + "," + sdpGLclampf(ref));
	glAlphaFunc(func,ref);
	sdpglEndFunctionTrace();
}

GLboolean sdpglAreTexturesResident(GLsizei n,const GLuint *textures,GLboolean *residences)
{
	sdpglBeginFunctionTrace("glAreTexturesResident", sdpGLsizei(n));//,const GLuint *textures,GLboolean *residenc);
	GLboolean result = glAreTexturesResident(n,textures,residences);
	sdpglResultTrace(sdpGLboolean(result));
	sdpglEndFunctionTrace();
	return result;
}

void sdpglArrayElement(GLint i)
{
	sdpglBeginFunctionTrace("glArrayElement", sdpGLint(i));
	glArrayElement(i);
	sdpglEndFunctionTrace();
}

void sdpglBegin(GLenum mode)
{
	sdpglBeginFunctionTrace("glBegin", sdpGLenum(mode));
	glBegin(mode);
	sdpglEndFunctionTrace();
}

void sdpglBindTexture(GLenum target,GLuint texture)
{
	sdpglBeginFunctionTrace("glBindTexture", sdpGLenum(target) + "," + sdpGLuint(texture));
	glBindTexture(target,texture);
	sdpglEndFunctionTrace();
}

void sdpglBitmap(GLsizei width,GLsizei height,GLfloat xorig,GLfloat yorig,GLfloat xmove,GLfloat ymove,const GLubyte *bitmap)
{
	sdpglBeginFunctionTrace("glBitmap", sdpGLsizei(width) + "," + sdpGLsizei(height) + "," + sdpGLfloat(xorig) + "," + sdpGLfloat(yorig) + "," + sdpGLfloat(xmove) + "," + sdpGLfloat(ymove));//,const GLubyte *bitmap",);
	glBitmap(width,height,xorig,yorig,xmove,ymove,bitmap);
	sdpglEndFunctionTrace();
}

void sdpglBlendFunc(GLenum sfactor,GLenum dfactor)
{
	sdpglBeginFunctionTrace("glBlendFunc", sdpGLenum(sfactor) + "," + sdpGLenum(dfactor));
	glBlendFunc(sfactor,dfactor);
	sdpglEndFunctionTrace();
}

void sdpglCallList(GLuint list)
{
	sdpglBeginFunctionTrace("glCallList", sdpGLuint(list));
	glCallList(list);
	sdpglEndFunctionTrace();
}

void sdpglCallLists(GLsizei n,GLenum type,const GLvoid *lists)
{
	sdpglBeginFunctionTrace("glCallLists", sdpGLsizei(n) + "," + sdpGLenum(type) + "," + sdpGLvoid(lists));
	glCallLists(n,type,lists);
	sdpglEndFunctionTrace();
}

void sdpglClear(GLbitfield mask)
{
	sdpglBeginFunctionTrace("glClear", sdpGLbitfield(mask));
	glClear(mask);
	sdpglEndFunctionTrace();
}

void sdpglClearAccum(GLfloat red,GLfloat green,GLfloat blue,GLfloat alpha)
{
	sdpglBeginFunctionTrace("glClearAccum", sdpGLfloat(red) + "," + sdpGLfloat(green) + "," + sdpGLfloat(blue) + "," + sdpGLfloat(alpha));
	glClearAccum(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglClearColor(GLclampf red,GLclampf green,GLclampf blue,GLclampf alpha)
{
	sdpglBeginFunctionTrace("glClearColor", sdpGLclampf(red) + "," + sdpGLclampf(green) + "," + sdpGLclampf(blue) + "," + sdpGLclampf(alpha));
	glClearColor(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglClearDepth(GLclampd depth)
{
	sdpglBeginFunctionTrace("glClearDepth", sdpGLclampd(depth));
	glClearDepth(depth);
	sdpglEndFunctionTrace();
}

void sdpglClearIndex(GLfloat c)
{
	sdpglBeginFunctionTrace("glClearIndex", sdpGLfloat(c));
	glClearIndex(c);
	sdpglEndFunctionTrace();
}

void sdpglClearStencil(GLint s)
{
	sdpglBeginFunctionTrace("glClearStencil", sdpGLint(s));
	glClearStencil(s);
	sdpglEndFunctionTrace();
}

void sdpglClipPlane(GLenum plane,const GLdouble *equation)
{
	sdpglBeginFunctionTrace("glClipPlane", sdpGLenum(plane));
	glClipPlane(plane,equation);
	sdpglEndFunctionTrace();
}

void sdpglColor3b(GLbyte red,GLbyte green,GLbyte blue)
{
	sdpglBeginFunctionTrace("glColor3b", sdpGLbyte(red) + "," + sdpGLbyte(green) + "," + sdpGLbyte(blue));
	glColor3b(red,green,blue);
	sdpglEndFunctionTrace();
}

void sdpglColor3bv(const GLbyte *v)
{
	sdpglBeginFunctionTrace("glColor3bv", sdpGLbyte(v[0]) + "," + sdpGLbyte(v[1]) + "," + sdpGLbyte(v[2]));
	glColor3bv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor3d(GLdouble red,GLdouble green,GLdouble blue)
{
	sdpglBeginFunctionTrace("glColor3d", sdpGLdouble(red) + "," + sdpGLdouble(green) + "," + sdpGLdouble(blue));
	glColor3d(red,green,blue);
	sdpglEndFunctionTrace();
}

void sdpglColor3dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glColor3dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]));
	glColor3dv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor3f(GLfloat red,GLfloat green,GLfloat blue)
{
	sdpglBeginFunctionTrace("glColor3f", sdpGLfloat(red) + "," + sdpGLfloat(green) + "," + sdpGLfloat(blue));
	glColor3f(red,green,blue);
	sdpglEndFunctionTrace();
}

void sdpglColor3fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glColor3fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[2]) + "," + sdpGLfloat(v[3]));
	glColor3fv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor3i(GLint red,GLint green,GLint blue)
{
	sdpglBeginFunctionTrace("glColor3i", sdpGLint(red) + "," + sdpGLint(green) + "," + sdpGLint(blue));
	glColor3i(red,green,blue);
	sdpglEndFunctionTrace();
}

void sdpglColor3iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glColor3iv", ""); //,const GLint );
	glColor3iv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor3s(GLshort red,GLshort green,GLshort blue)
{
	sdpglBeginFunctionTrace("glColor3s", sdpGLshort(red) + "," + sdpGLshort(green) + "," + sdpGLshort(blue));
	glColor3s(red,green,blue);
	sdpglEndFunctionTrace();
}

void sdpglColor3sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glColor3sv", ""); //,const GLshort );
	glColor3sv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor3ub(GLubyte red,GLubyte green,GLubyte blue)
{
	sdpglBeginFunctionTrace("glColor3ub", sdpGLubyte(red) + "," + sdpGLubyte(green) + "," + sdpGLubyte(blue));
	glColor3ub(red,green,blue);
	sdpglEndFunctionTrace();
}

void sdpglColor3ubv(const GLubyte *v)
{
	sdpglBeginFunctionTrace("glColor3ubv", ""); //,const GLubyte );
	glColor3ubv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor3ui(GLuint red,GLuint green,GLuint blue)
{
	sdpglBeginFunctionTrace("glColor3ui", sdpGLuint(red) + "," + sdpGLuint(green) + "," + sdpGLuint(blue));
	glColor3ui(red,green,blue);
	sdpglEndFunctionTrace();
}

void sdpglColor3uiv(const GLuint *v)
{
	sdpglBeginFunctionTrace("glColor3uiv", ""); //,const GLuint );
	glColor3uiv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor3us(GLushort red,GLushort green,GLushort blue)
{
	sdpglBeginFunctionTrace("glColor3us", sdpGLushort(red) + "," + sdpGLushort(green) + "," + sdpGLushort(blue));
	glColor3us(red,green,blue);
	sdpglEndFunctionTrace();
}

void sdpglColor3usv(const GLushort *v)
{
	sdpglBeginFunctionTrace("glColor3usv", ""); //,const GLushort );
	glColor3usv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor4b(GLbyte red,GLbyte green,GLbyte blue,GLbyte alpha)
{
	sdpglBeginFunctionTrace("glColor4b", sdpGLbyte(red) + "," + sdpGLbyte(green) + "," + sdpGLbyte(blue) + "," + sdpGLbyte(alpha));
	glColor4b(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglColor4bv(const GLbyte *v)
{
	sdpglBeginFunctionTrace("glColor4bv", ""); //,const GLbyte );
	glColor4bv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor4d(GLdouble red,GLdouble green,GLdouble blue,GLdouble alpha)
{
	sdpglBeginFunctionTrace("glColor4d", sdpGLdouble(red) + "," + sdpGLdouble(green) + "," + sdpGLdouble(blue) + "," + sdpGLdouble(alpha));
	glColor4d(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglColor4dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glColor4dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]) + "," + sdpGLdouble(v[3]));
	glColor4dv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor4f(GLfloat red,GLfloat green,GLfloat blue,GLfloat alpha)
{
	sdpglBeginFunctionTrace("glColor4f", sdpGLfloat(red) + "," + sdpGLfloat(green) + "," + sdpGLfloat(blue) + "," + sdpGLfloat(alpha));
	glColor4f(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}
 
void sdpglColor4fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glColor4fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]) + "," + sdpGLfloat(v[2]) + "," + sdpGLfloat(v[3]));
	glColor4fv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor4i(GLint red,GLint green,GLint blue,GLint alpha)
{
	sdpglBeginFunctionTrace("glColor4i", sdpGLint(red) + "," + sdpGLint(green) + "," + sdpGLint(blue) + "," + sdpGLint(alpha));
	glColor4i(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglColor4iv(const GLint *v)
{ 
	sdpglBeginFunctionTrace("glColor4iv", ""); //,const GLint );
	glColor4iv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor4s(GLshort red,GLshort green,GLshort blue,GLshort alpha)
{
	sdpglBeginFunctionTrace("glColor4s", sdpGLshort(red) + "," + sdpGLshort(green) + "," + sdpGLshort(blue) + "," + sdpGLshort(alpha));
	glColor4s(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglColor4sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glColor4sv", ""); //,const GLshort );
	glColor4sv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor4ub(GLubyte red,GLubyte green,GLubyte blue,GLubyte alpha)
{
	sdpglBeginFunctionTrace("glColor4ub", ""); //,GLubyte red,GLubyte green,GLubyte blue,GLubyte alpha);
	glColor4ub(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglColor4ubv(const GLubyte *v)
{
	sdpglBeginFunctionTrace("glColor4ubv", ""); //,const GLubyte );
	glColor4ubv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor4ui(GLuint red,GLuint green,GLuint blue,GLuint alpha)
{
	sdpglBeginFunctionTrace("glColor4ui", sdpGLuint(red) + "," + sdpGLuint(green) + "," + sdpGLuint(blue) + "," + sdpGLuint(alpha));
	glColor4ui(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglColor4uiv(const GLuint *v)
{
	sdpglBeginFunctionTrace("glColor4uiv", ""); //,const GLuint );
	glColor4uiv(v);
	sdpglEndFunctionTrace();
}

void sdpglColor4us(GLushort red,GLushort green,GLushort blue,GLushort alpha)
{
	sdpglBeginFunctionTrace("glColor4us", ""); //,GLushort red,GLushort green,GLushort blue,GLushort alpha);
	glColor4us(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglColor4usv(const GLushort *v)
{
	sdpglBeginFunctionTrace("glColor4usv", ""); //,const GLushort );
	glColor4usv(v);
	sdpglEndFunctionTrace();
}

void sdpglColorMask(GLboolean red,GLboolean green,GLboolean blue,GLboolean alpha)
{
	sdpglBeginFunctionTrace("glColorMask", sdpGLboolean(red) + sdpGLboolean(green) + sdpGLboolean(blue) + sdpGLboolean(alpha));//);
	glColorMask(red,green,blue,alpha);
	sdpglEndFunctionTrace();
}

void sdpglColorMaterial(GLenum face,GLenum mode)
{
	sdpglBeginFunctionTrace("glColorMaterial", sdpGLenum(face) + "," + sdpGLenum(mode));
	glColorMaterial(face,mode);
	sdpglEndFunctionTrace();
}

void sdpglColorPointer(GLint size,GLenum type,GLsizei stride,const GLvoid *pointer)
{
	sdpglBeginFunctionTrace("glColorPointer", sdpGLint(size) + "," + sdpGLenum(type) + "," + sdpGLsizei(stride));//,const GLvoid *pointer");
	glColorPointer(size,type,stride,pointer);
	sdpglEndFunctionTrace();
}

void sdpglCopyPixels(GLint x,GLint y,GLsizei width,GLsizei height,GLenum type)
{
	sdpglBeginFunctionTrace("glCopyPixels", sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLsizei(width) + "," + sdpGLsizei(height) + "," + sdpGLenum(type));
	glCopyPixels(x,y,width,height,type);
	sdpglEndFunctionTrace();
}

void sdpglCopyTexImage1D(GLenum target,GLint level,GLenum internalFormat,GLint x,GLint y,GLsizei width,GLint border)
{
	sdpglBeginFunctionTrace("glCopyTexImage1D", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLenum(internalFormat) + "," + sdpGLint(x) + "," + sdpGLint(y));//,GLsizei width,GLint border",&);
	glCopyTexImage1D(target,level,internalFormat,x,y,width,border);
	sdpglEndFunctionTrace();
}

void sdpglCopyTexImage2D(GLenum target,GLint level,GLenum internalFormat,GLint x,GLint y,GLsizei width,GLsizei height,GLint border)
{
	sdpglBeginFunctionTrace("glCopyTexImage2D", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLenum(internalFormat) + "," + sdpGLint(x) + "," + sdpGLint(y));//,GLsizei width,GLsizei height,GLint border",&);
	glCopyTexImage2D(target,level,internalFormat,x,y,width,height,border);
	sdpglEndFunctionTrace();
}

void sdpglCopyTexSubImage1D(GLenum target,GLint level,GLint xoffset,GLint x,GLint y,GLsizei width)
{
	sdpglBeginFunctionTrace("glCopyTexSubImage1D", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLint(xoffset) + "," + sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLsizei(width));
	glCopyTexSubImage1D(target,level,xoffset,x,y,width);
	sdpglEndFunctionTrace();
}

void sdpglCopyTexSubImage2D(GLenum target,GLint level,GLint xoffset,GLint yoffset,GLint x,GLint y,GLsizei width,GLsizei height)
{
	sdpglBeginFunctionTrace("glCopyTexSubImage2D", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLint(xoffset) + "," + sdpGLint(yoffset) + "," + sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLsizei(width) + "," + sdpGLsizei(height));
	glCopyTexSubImage2D(target,level,xoffset,yoffset,x,y,width,height);
	sdpglEndFunctionTrace();
}

void sdpglCullFace(GLenum mode)
{
	sdpglBeginFunctionTrace("glCullFace", sdpGLenum(mode));
	glCullFace(mode);
	sdpglEndFunctionTrace();
}

void sdpglDeleteLists(GLuint list,GLsizei range)
{
	sdpglBeginFunctionTrace("glDeleteLists", sdpGLuint(list) + "," + sdpGLsizei(range));
	glDeleteLists(list,range);
	sdpglEndFunctionTrace();
}

void sdpglDeleteTextures(GLsizei n,const GLuint *textures)
{
	sdpglBeginFunctionTrace("glDeleteTextures", sdpGLsizei(n));//,const GLuint *textur);
	glDeleteTextures(n,textures);
	sdpglEndFunctionTrace();
}

void sdpglDepthFunc(GLenum func)
{
	sdpglBeginFunctionTrace("glDepthFunc", sdpGLenum(func));
	glDepthFunc(func);
	sdpglEndFunctionTrace();
}

void sdpglDepthMask(GLboolean flag)
{
	sdpglBeginFunctionTrace("glDepthMask", sdpGLboolean(flag));//");
	glDepthMask(flag);
	sdpglEndFunctionTrace();
}

void sdpglDepthRange(GLclampd zNear,GLclampd zFar)
{
	sdpglBeginFunctionTrace("glDepthRange", ""); //,GLclampd zNear,GLclampd zFar",);
	glDepthRange(zNear,zFar);
	sdpglEndFunctionTrace();
}

void sdpglDisable(GLenum cap)
{
	sdpglBeginFunctionTrace("glDisable", sdpGLenum(cap));
	glDisable(cap);
	sdpglEndFunctionTrace();
}

void sdpglDisableClientState(GLenum array)
{
	sdpglBeginFunctionTrace("glDisableClientState", sdpGLenum(array));
	glDisableClientState(array);
	sdpglEndFunctionTrace();
}

void sdpglDrawArrays(GLenum mode,GLint first,GLsizei count)
{
	sdpglBeginFunctionTrace("glDrawArrays", sdpGLenum(mode) + "," + sdpGLint(first) + "," + sdpGLsizei(count));
	glDrawArrays(mode,first,count);
	sdpglEndFunctionTrace();
}

void sdpglDrawBuffer(GLenum mode)
{
	sdpglBeginFunctionTrace("glDrawBuffer", sdpGLenum(mode));
	glDrawBuffer(mode);
	sdpglEndFunctionTrace();
}

void sdpglDrawElements(GLenum mode,GLsizei count,GLenum type,const GLvoid *indices)
{
	sdpglBeginFunctionTrace("glDrawElements", sdpGLenum(mode) + "," + sdpGLsizei(count) + "," + sdpGLenum(type));//,const GLvoid *indices");
	glDrawElements(mode,count,type,indices);
	sdpglEndFunctionTrace();
}

void sdpglDrawPixels(GLsizei width,GLsizei height,GLenum format,GLenum type,const GLvoid *pixels)
{
	sdpglBeginFunctionTrace("glDrawPixels", sdpGLsizei(width) + "," + sdpGLsizei(height) + "," + sdpGLenum(format) + "," + sdpGLenum(type));//,const GLvoid *pixels",);
	glDrawPixels(width,height,format,type,pixels);
	sdpglEndFunctionTrace();
}

void sdpglEdgeFlag(GLboolean flag)
{
	sdpglBeginFunctionTrace("glEdgeFlag", sdpGLboolean(flag));
	glEdgeFlag(flag);
	sdpglEndFunctionTrace();
}

void sdpglEdgeFlagPointer(GLsizei stride, const GLboolean *pointer)
{
	sdpglBeginFunctionTrace("glEdgeFlagPointer", sdpGLsizei(stride));//,const GLvoid *pointer",&);
	glEdgeFlagPointer(stride, pointer);
	sdpglEndFunctionTrace();
}

void sdpglEdgeFlagv(const GLboolean *flag)
{
	sdpglBeginFunctionTrace("glEdgeFlagv", ""); //,const GLboolean *flag");
	glEdgeFlagv(flag);
	sdpglEndFunctionTrace();
}

void sdpglEnable(GLenum cap)
{
	sdpglBeginFunctionTrace("glEnable", sdpGLenum(cap));
	glEnable(cap);
	sdpglEndFunctionTrace();
}

void sdpglEnableClientState(GLenum array)
{
	sdpglBeginFunctionTrace("glEnableClientState", sdpGLenum(array));
	glEnableClientState(array);
	sdpglEndFunctionTrace();
}

void sdpglEnd()
{
	sdpglBeginFunctionTrace("glEnd", "");
	glEnd();
	sdpglEndFunctionTrace();
}

void sdpglEndList()
{
	sdpglBeginFunctionTrace("glEndList", "");
	glEndList();
	sdpglEndFunctionTrace();
}

void sdpglEvalCoord1d(GLdouble u)
{
	sdpglBeginFunctionTrace("glEvalCoord1d", sdpGLdouble(u));
	glEvalCoord1d(u);
	sdpglEndFunctionTrace();
}

void sdpglEvalCoord1dv(const GLdouble *u)
{
	sdpglBeginFunctionTrace("glEvalCoord1dv", ""); //,const GLdouble );
	glEvalCoord1dv(u);
	sdpglEndFunctionTrace();
}

void sdpglEvalCoord1f(GLfloat u)
{
	sdpglBeginFunctionTrace("glEvalCoord1f", sdpGLfloat(u));
	glEvalCoord1f(u);
	sdpglEndFunctionTrace();
}

void sdpglEvalCoord1fv(const GLfloat *u)
{
	sdpglBeginFunctionTrace("glEvalCoord1fv", ""); //,const GLfloat );
	glEvalCoord1fv(u);
	sdpglEndFunctionTrace();
}

void sdpglEvalCoord2d(GLdouble u,GLdouble v)
{
	sdpglBeginFunctionTrace("glEvalCoord2d", sdpGLdouble(u) + "," + sdpGLdouble(v));
	glEvalCoord2d(u,v);
	sdpglEndFunctionTrace();
}

void sdpglEvalCoord2dv(const GLdouble *u)
{
	sdpglBeginFunctionTrace("glEvalCoord2dv", ""); //,const GLdouble );
	glEvalCoord2dv(u);
	sdpglEndFunctionTrace();
}

void sdpglEvalCoord2f(GLfloat u,GLfloat v)
{
	sdpglBeginFunctionTrace("glEvalCoord2f", sdpGLfloat(u) + "," + sdpGLfloat(v));
	glEvalCoord2f(u,v);
	sdpglEndFunctionTrace();
}

void sdpglEvalCoord2fv(const GLfloat *u)
{
	sdpglBeginFunctionTrace("glEvalCoord2fv", ""); //,const GLfloat );
	glEvalCoord2fv(u);
	sdpglEndFunctionTrace();
}

void sdpglEvalMesh1(GLenum mode,GLint i1,GLint i2)
{
	sdpglBeginFunctionTrace("glEvalMesh1", sdpGLenum(mode) + "," + sdpGLint(i1) + "," + sdpGLint(i2));
	glEvalMesh1(mode,i1,i2);
	sdpglEndFunctionTrace();
}

void sdpglEvalMesh2(GLenum mode,GLint i1,GLint i2,GLint j1,GLint j2)
{
	sdpglBeginFunctionTrace("glEvalMesh2", sdpGLenum(mode) + "," + sdpGLint(i1) + "," + sdpGLint(i2) + "," + sdpGLint(j1) + "," + sdpGLint(j2));
	glEvalMesh2(mode,i1,i2,j1,j2);
	sdpglEndFunctionTrace();
}

void sdpglEvalPoint1(GLint i)
{
	sdpglBeginFunctionTrace("glEvalPoint1", sdpGLint(i));
	glEvalPoint1(i);
	sdpglEndFunctionTrace();
}

void sdpglEvalPoint2(GLint i,GLint j)
{
	sdpglBeginFunctionTrace("glEvalPoint2", sdpGLint(i) + "," + sdpGLint(j));
	glEvalPoint2(i,j);
	sdpglEndFunctionTrace();
}

void sdpglFeedbackBuffer(GLsizei size,GLenum type,GLfloat *buffer)
{
	sdpglBeginFunctionTrace("glFeedbackBuffer", sdpGLsizei(size) + "," + sdpGLenum(type));//,GLfloat *buffer");
	glFeedbackBuffer(size,type,buffer);
	sdpglEndFunctionTrace();
}

void sdpglFinish()
{
	sdpglBeginFunctionTrace("glFinish", "");
	glFinish();
	sdpglEndFunctionTrace();
}

void sdpglFlush()
{
	sdpglBeginFunctionTrace("glFlush", "");
	glFlush();
	sdpglEndFunctionTrace();
}

void sdpglFogf(GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glFogf", sdpGLenum(pname) + "," + sdpGLfloat(param));
	glFogf(pname,param);
	sdpglEndFunctionTrace();
}

void sdpglFogfv(GLenum pname,const GLfloat *params)
{
	sdpglBeginFunctionTrace("glFogfv", sdpGLenum(pname));//,const GLfloat *params",);
	glFogfv(pname,params);
	sdpglEndFunctionTrace();
}

void sdpglFogi(GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glFogi", sdpGLenum(pname) + "," + sdpGLint(param));
	glFogi(pname,param);
	sdpglEndFunctionTrace();
}

void sdpglFogiv(GLenum pname,const GLint *params)
{
	sdpglBeginFunctionTrace("glFogiv", sdpGLenum(pname));//,const GLint *params",);
	glFogiv(pname,params);
	sdpglEndFunctionTrace();
}

void sdpglFrontFace(GLenum mode)
{
	sdpglBeginFunctionTrace("glFrontFace", sdpGLenum(mode));
	glFrontFace(mode);
	sdpglEndFunctionTrace();
}

void sdpglFrustum(GLdouble left,GLdouble right,GLdouble bottom,GLdouble top,GLdouble zNear,GLdouble zFar)
{
	sdpglBeginFunctionTrace("glFrustum", sdpGLdouble(left) + "," + sdpGLdouble(right) + "," + sdpGLdouble(bottom) + "," + sdpGLdouble(top) + "," + sdpGLdouble(zNear) + "," + sdpGLdouble(zFar));
	glFrustum(left,right,bottom,top,zNear,zFar);
	sdpglEndFunctionTrace();
}

GLuint sdpglGenLists(GLsizei range)
{
	sdpglBeginFunctionTrace("glGenLists", sdpGLsizei(range));
	GLuint result = glGenLists(range);
	sdpglResultTrace(sdpGLuint(result));
	sdpglEndFunctionTrace();
	return result;
}

void sdpglGenTextures(GLsizei n,GLuint *textures)
{
	sdpglBeginFunctionTrace("glGenTextures", sdpGLsizei(n));//,GLuint *textur);
	glGenTextures(n,textures);
	sdpglEndFunctionTrace();
}

void sdpglGetBooleanv(GLenum pname,GLboolean *params)
{
	sdpglBeginFunctionTrace("glGetBooleanv", sdpGLenum(pname));
	glGetBooleanv(pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetClipPlane(GLenum plane,GLdouble *equation)
{
	sdpglBeginFunctionTrace("glGetClipPlane", sdpGLenum(plane));
	glGetClipPlane(plane,equation);
	sdpglEndFunctionTrace();
}

void sdpglGetDoublev(GLenum pname,GLdouble *params)
{
	sdpglBeginFunctionTrace("glGetDoublev", sdpGLenum(pname));
	glGetDoublev(pname,params);
	sdpglEndFunctionTrace();
}

GLenum sdpglGetError()
{
	sdpglBeginFunctionTrace("glGetError", "");
	GLenum result = glGetError();
	sdpglResultTrace(sdpGLenum(result));
	sdpglEndFunctionTrace();
	return result;
}

void sdpglGetFloatv(GLenum pname,GLfloat *params)
{
	sdpglBeginFunctionTrace("glGetFloatv", sdpGLenum(pname));
	glGetFloatv(pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetIntegerv(GLenum pname,GLint *params)
{
	sdpglBeginFunctionTrace("glGetIntegerv", sdpGLenum(pname));
	glGetIntegerv(pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetLightfv(GLenum light,GLenum pname,GLfloat *params)
{
	sdpglBeginFunctionTrace("glGetLightfv", sdpGLenum(light) + "," + sdpGLenum(pname));
	glGetLightfv(light,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetLightiv(GLenum light,GLenum pname,GLint *params)
{
	sdpglBeginFunctionTrace("glGetLightiv", sdpGLenum(light) + "," + sdpGLenum(pname));
	glGetLightiv(light,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetMapdv(GLenum target,GLenum query,GLdouble *v)
{
	sdpglBeginFunctionTrace("glGetMapdv", sdpGLenum(target) + "," + sdpGLenum(query));
	glGetMapdv(target,query,v);
	sdpglEndFunctionTrace();
}

void sdpglGetMapfv(GLenum target,GLenum query,GLfloat *v)
{
	sdpglBeginFunctionTrace("glGetMapfv", sdpGLenum(target) + "," + sdpGLenum(query));
	glGetMapfv(target,query,v);
	sdpglEndFunctionTrace();
}

void sdpglGetMapiv(GLenum target,GLenum query,GLint *v)
{
	sdpglBeginFunctionTrace("glGetMapiv", sdpGLenum(target) + "," + sdpGLenum(query));
	glGetMapiv(target,query,v);
	sdpglEndFunctionTrace();
}

void sdpglGetMaterialfv(GLenum face,GLenum pname,GLfloat *params)
{
	sdpglBeginFunctionTrace("glGetMaterialfv", sdpGLenum(face) + "," + sdpGLenum(pname));
	glGetMaterialfv(face,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetMaterialiv(GLenum face,GLenum pname,GLint *params)
{
	sdpglBeginFunctionTrace("glGetMaterialiv", sdpGLenum(face) + "," + sdpGLenum(pname));
	glGetMaterialiv(face,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetPixelMapfv(GLenum map,GLfloat *values)
{
	sdpglBeginFunctionTrace("glGetPixelMapfv", sdpGLenum(map));
	glGetPixelMapfv(map,values);
	sdpglEndFunctionTrace();
}

void sdpglGetPixelMapuiv(GLenum map,GLuint *values)
{
	sdpglBeginFunctionTrace("glGetPixelMapuiv", sdpGLenum(map));
	glGetPixelMapuiv(map,values);
	sdpglEndFunctionTrace();
}

void sdpglGetPixelMapusv(GLenum map,GLushort *values)
{
	sdpglBeginFunctionTrace("glGetPixelMapusv", sdpGLenum(map));
	glGetPixelMapusv(map,values);
	sdpglEndFunctionTrace();
}

void sdpglGetPointerv(GLenum pname,GLvoid* *params)
{
	sdpglBeginFunctionTrace("glGetPointerv", sdpGLenum(pname));
	glGetPointerv(pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetPolygonStipple(GLubyte *mask)
{
	sdpglBeginFunctionTrace("glGetPolygonStipple", "");
	glGetPolygonStipple(mask);
	sdpglEndFunctionTrace();
}

const char* sdpglGetString(GLenum name)
{
	sdpglBeginFunctionTrace("glGetString", sdpGLenum(name));
	const char* result = (const char*)glGetString(name);
	sdpglResultTrace(result);
	sdpglEndFunctionTrace();
	return result;
}

void sdpglGetTexEnvfv(GLenum target,GLenum pname,GLfloat *params)
{
	sdpglBeginFunctionTrace("glGetTexEnvfv", sdpGLenum(target) + "," + sdpGLenum(pname));
	glGetTexEnvfv(target,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetTexEnviv(GLenum target,GLenum pname,GLint *params)
{
	sdpglBeginFunctionTrace("glGetTexEnviv", sdpGLenum(target) + "," + sdpGLenum(pname));
	glGetTexEnviv(target,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetTexGendv(GLenum coord,GLenum pname,GLdouble *params)
{
	sdpglBeginFunctionTrace("glGetTexGendv", sdpGLenum(coord) + "," + sdpGLenum(pname));
	glGetTexGendv(coord,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetTexGenfv(GLenum coord,GLenum pname,GLfloat *params)
{
	sdpglBeginFunctionTrace("glGetTexGenfv", sdpGLenum(coord) + "," + sdpGLenum(pname));
	glGetTexGenfv(coord,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetTexGeniv(GLenum coord,GLenum pname,GLint *params)
{
	sdpglBeginFunctionTrace("glGetTexGeniv", sdpGLenum(coord) + "," + sdpGLenum(pname));
	glGetTexGeniv(coord,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetTexImage(GLenum target,GLint level,GLenum format,GLenum type,GLvoid *pixels)
{
	sdpglBeginFunctionTrace("glGetTexImage", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLenum(format) + "," + sdpGLenum(type));
	glGetTexImage(target,level,format,type,pixels);
	sdpglEndFunctionTrace();
}

void sdpglGetTexLevelParameterfv(GLenum target,GLint level,GLenum pname,GLfloat *params)
{
	sdpglBeginFunctionTrace("glGetTexLevelParameterfv", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLenum(pname));
	glGetTexLevelParameterfv(target,level,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetTexLevelParameteriv(GLenum target,GLint level,GLenum pname,GLint *params)
{
	sdpglBeginFunctionTrace("glGetTexLevelParameteriv", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLenum(pname));
	glGetTexLevelParameteriv(target,level,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetTexParameterfv(GLenum target,GLenum pname,GLfloat *params)
{
	sdpglBeginFunctionTrace("glGetTexParameterfv", sdpGLenum(target) + "," + sdpGLenum(pname));
	glGetTexParameterfv(target,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglGetTexParameteriv(GLenum target,GLenum pname,GLint *params)
{
	sdpglBeginFunctionTrace("glGetTexParameteriv", sdpGLenum(target) + "," + sdpGLenum(pname));
	glGetTexParameteriv(target,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglHint(GLenum target,GLenum mode)
{
	sdpglBeginFunctionTrace("glHint", sdpGLenum(target) + "," + sdpGLenum(mode));
	glHint(target,mode);
	sdpglEndFunctionTrace();
}

void sdpglIndexMask(GLuint mask)
{
	sdpglBeginFunctionTrace("glIndexMask", sdpGLuint(mask));
	glIndexMask(mask);
	sdpglEndFunctionTrace();
}

void sdpglIndexPointer(GLenum type,GLsizei stride,const GLvoid *pointer)
{
	sdpglBeginFunctionTrace("glIndexPointer", sdpGLenum(type) + "," + sdpGLsizei(stride));//,const GLvoid *pointer");
	glIndexPointer(type,stride,pointer);
	sdpglEndFunctionTrace();
}

void sdpglIndexd(GLdouble c)
{
	sdpglBeginFunctionTrace("glIndexd", sdpGLdouble(c));
	glIndexd(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexdv(const GLdouble *c)
{
	sdpglBeginFunctionTrace("glIndexdv", ""); //,const GLdouble );
	glIndexdv(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexf(GLfloat c)
{
	sdpglBeginFunctionTrace("glIndexf", sdpGLfloat(c));
	glIndexf(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexfv(const GLfloat *c)
{
	sdpglBeginFunctionTrace("glIndexfv", ""); //,const GLfloat );
	glIndexfv(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexi(GLint c)
{
	sdpglBeginFunctionTrace("glIndexi", sdpGLint(c));
	glIndexi(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexiv(const GLint *c)
{
	sdpglBeginFunctionTrace("glIndexiv", ""); //,const GLint );
	glIndexiv(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexs(GLshort c)
{
	sdpglBeginFunctionTrace("glIndexs", sdpGLshort(c));
	glIndexs(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexsv(const GLshort *c)
{
	sdpglBeginFunctionTrace("glIndexsv", ""); //,const GLshort );
	glIndexsv(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexub(GLubyte c)
{
	sdpglBeginFunctionTrace("glIndexub", ""); //,GLubyte);
	glIndexub(c);
	sdpglEndFunctionTrace();
}

void sdpglIndexubv(const GLubyte *c)
{
	sdpglBeginFunctionTrace("glIndexubv", ""); //,const GLubyte );
	glIndexubv(c);
	sdpglEndFunctionTrace();
}

void sdpglInitNames()
{
	sdpglBeginFunctionTrace("glInitNames", "");
	glInitNames();
	sdpglEndFunctionTrace();
}

void sdpglInterleavedArrays(GLenum format,GLsizei stride,const GLvoid *pointer)
{
	sdpglBeginFunctionTrace("glInterleavedArrays", sdpGLenum(format) + "," + sdpGLsizei(stride));//,const GLvoid *pointer",&);
	glInterleavedArrays(format,stride,pointer);
	sdpglEndFunctionTrace();
}

GLboolean sdpglIsEnabled(GLenum cap)
{
	sdpglBeginFunctionTrace("glIsEnabled", sdpGLenum(cap));
	GLboolean result = glIsEnabled(cap);
	sdpglResultTrace(sdpGLboolean(result));
	sdpglEndFunctionTrace();
	return result;
}

GLboolean sdpglIsList(GLuint list)
{
	sdpglBeginFunctionTrace("glIsList", sdpGLuint(list));
	GLboolean result = glIsList(list);
	sdpglResultTrace(sdpGLboolean(result));
	sdpglEndFunctionTrace();
	return result;
}

GLboolean sdpglIsTexture(GLuint texture)
{
	sdpglBeginFunctionTrace("glIsTexture", sdpGLuint(texture));
	GLboolean result = glIsTexture(texture);
	sdpglResultTrace(sdpGLboolean(result));
	sdpglEndFunctionTrace();
	return result;
}

void sdpglLightModelf(GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glLightModelf", sdpGLenum(pname) + "," + sdpGLfloat(param));
	glLightModelf(pname,param);
	sdpglEndFunctionTrace();
}

void sdpglLightModelfv(GLenum pname,const GLfloat *params)
{
	sdpglBeginFunctionTrace("glLightModelfv", sdpGLenum(pname));//,const GLfloat *params",);
	glLightModelfv(pname,params);
	sdpglEndFunctionTrace();
}

void sdpglLightModeli(GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glLightModeli", sdpGLenum(pname) + "," + sdpGLint(param));
	glLightModeli(pname,param);
	sdpglEndFunctionTrace();
}

void sdpglLightModeliv(GLenum pname,const GLint *params)
{
	sdpglBeginFunctionTrace("glLightModeliv", sdpGLenum(pname));//,const GLint *params",);
	glLightModeliv(pname,params);
	sdpglEndFunctionTrace();
}

void sdpglLightf(GLenum light,GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glLightf", sdpGLenum(light) + "," + sdpGLenum(pname) + "," + sdpGLfloat(param));
	glLightf(light,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglLightfv(GLenum light,GLenum pname,const GLfloat *params)
{
	sdpglBeginFunctionTrace("glLightfv", sdpGLenum(light) + "," + sdpGLenum(pname) + ",");
	if(pname == GL_AMBIENT)
		sdpglRawTrace(sdpGLfloat(params[0]) + "," + sdpGLfloat(params[1]) + "," + sdpGLfloat(params[2]));
	else if(pname == GL_DIFFUSE)
		sdpglRawTrace(sdpGLfloat(params[0]) + "," + sdpGLfloat(params[1]) + "," + sdpGLfloat(params[2]));
	else if(pname == GL_SPECULAR)
		sdpglRawTrace(sdpGLfloat(params[0]) + "," + sdpGLfloat(params[1]) + "," + sdpGLfloat(params[2]));
	else if(pname == GL_POSITION)
		sdpglRawTrace(sdpGLfloat(params[0]) + "," + sdpGLfloat(params[1]) + "," + sdpGLfloat(params[2]));
	else
		sdpGLvoid((void*)params);

	glLightfv(light,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglLighti(GLenum light,GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glLighti", sdpGLenum(light) + "," + sdpGLenum(pname) + "," + sdpGLint(param));
	glLighti(light,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglLightiv(GLenum light,GLenum pname,const GLint *params)
{
	sdpglBeginFunctionTrace("glLightiv", sdpGLenum(light) + "," + sdpGLenum(pname));//,const GLint *params",);
	glLightiv(light,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglLineStipple(GLint factor,GLushort pattern)
{
	sdpglBeginFunctionTrace("glLineStipple", sdpGLint(factor) + "," + sdpGLushort(pattern));
	glLineStipple(factor,pattern);
	sdpglEndFunctionTrace();
}

void sdpglLineWidth(GLfloat width)
{
	sdpglBeginFunctionTrace("glLineWidth", sdpGLfloat(width));
	glLineWidth(width);
	sdpglEndFunctionTrace();
}

void sdpglListBase(GLuint base)
{
	sdpglBeginFunctionTrace("glListBase", sdpGLuint(base));
	glListBase(base);
	sdpglEndFunctionTrace();
}

void sdpglLoadIdentity()
{
	sdpglBeginFunctionTrace("glLoadIdentity", "");
	glLoadIdentity();
	sdpglEndFunctionTrace();
}

void sdpglLoadMatrixd(const GLdouble *m)
{
	sdpglBeginFunctionTrace("glLoadMatrixd", ""); //,const GLdouble );
	glLoadMatrixd(m);
	sdpglEndFunctionTrace();
}

void sdpglLoadMatrixf(const GLfloat *m)
{
	sdpglBeginFunctionTrace("glLoadMatrixf", ""); //,const GLfloat );
	glLoadMatrixf(m);
	sdpglEndFunctionTrace();
}

void sdpglLoadName(GLuint name)
{
	sdpglBeginFunctionTrace("glLoadName", sdpGLuint(name));
	glLoadName(name);
	sdpglEndFunctionTrace();
}

void sdpglLogicOp(GLenum opcode)
{
	sdpglBeginFunctionTrace("glLogicOp", sdpGLenum(opcode));
	glLogicOp(opcode);
	sdpglEndFunctionTrace();
}

void sdpglMap1d(GLenum target,GLdouble u1,GLdouble u2,GLint stride,GLint order,const GLdouble *points)
{
	sdpglBeginFunctionTrace("glMap1d", sdpGLenum(target) + "," + sdpGLdouble(u1) + "," + sdpGLdouble(u2) + "," + sdpGLint(stride) + "," + sdpGLint(order));//,const GLdouble *points",&);
	glMap1d(target,u1,u2,stride,order,points);
	sdpglEndFunctionTrace();
}

void sdpglMap1f(GLenum target,GLfloat u1,GLfloat u2,GLint stride,GLint order,const GLfloat *points)
{
	sdpglBeginFunctionTrace("glMap1f", sdpGLenum(target) + "," + sdpGLfloat(u1) + "," + sdpGLfloat(u2) + "," + sdpGLint(stride) + "," + sdpGLint(order));//,const GLfloat *points",&);
	glMap1f(target,u1,u2,stride,order,points);
	sdpglEndFunctionTrace();
}

void sdpglMap2d(GLenum target,GLdouble u1,GLdouble u2,GLint ustride,GLint uorder,GLdouble v1,GLdouble v2,GLint vstride,GLint vorder,const GLdouble *points)
{
	sdpglBeginFunctionTrace("glMap2d", sdpGLenum(target) + "," + sdpGLdouble(u1) + "," + sdpGLdouble(u2) + "," + sdpGLint(ustride) + "," + sdpGLint(uorder) + "," + sdpGLdouble(v1) + "," + sdpGLdouble(v2));//,GLint vstride,GLint vorder,const GLdouble *points",&);
	glMap2d(target,u1,u2,ustride,uorder,v1,v2,vstride,vorder,points);
	sdpglEndFunctionTrace();
}

void sdpglMap2f(GLenum target,GLfloat u1,GLfloat u2,GLint ustride,GLint uorder,GLfloat v1,GLfloat v2,GLint vstride,GLint vorder,const GLfloat *points)
{
	sdpglBeginFunctionTrace("glMap2f", sdpGLenum(target) + "," + sdpGLfloat(u1) + "," + sdpGLfloat(u2) + "," + sdpGLint(ustride) + "," + sdpGLint(uorder) + "," + sdpGLfloat(v1) + "," + sdpGLfloat(v2));//,GLint vstride,GLint vorder,const GLfloat *points",&);
	glMap2f(target,u1,u2,ustride,uorder,v1,v2,vstride,vorder,points);
	sdpglEndFunctionTrace();
}

void sdpglMapGrid1d(GLint un,GLdouble u1,GLdouble u2)
{
	sdpglBeginFunctionTrace("glMapGrid1d", sdpGLint(un) + "," + sdpGLdouble(u1) + "," + sdpGLdouble(u2));
	glMapGrid1d(un,u1,u2);
	sdpglEndFunctionTrace();
}

void sdpglMapGrid1f(GLint un,GLfloat u1,GLfloat u2)
{
	sdpglBeginFunctionTrace("glMapGrid1f", sdpGLint(un) + "," + sdpGLfloat(u1) + "," + sdpGLfloat(u2));
	glMapGrid1f(un,u1,u2);
	sdpglEndFunctionTrace();
}

void sdpglMapGrid2d(GLint un,GLdouble u1,GLdouble u2,GLint vn,GLdouble v1,GLdouble v2)
{
	sdpglBeginFunctionTrace("glMapGrid2d", sdpGLint(un) + "," + sdpGLdouble(u1) + "," + sdpGLdouble(u2));//,GLint vn,GLdouble v1,GLdouble v);
	glMapGrid2d(un,u1,u2,vn,v1,v2);
	sdpglEndFunctionTrace();
}

void sdpglMapGrid2f(GLint un,GLfloat u1,GLfloat u2,GLint vn,GLfloat v1,GLfloat v2)
{
	sdpglBeginFunctionTrace("glMapGrid2f", sdpGLint(un) + "," + sdpGLfloat(u1) + "," + sdpGLfloat(u2));//,GLint vn,GLfloat v1,GLfloat v);
	glMapGrid2f(un,u1,u2,vn,v1,v2);
	sdpglEndFunctionTrace();
}

void sdpglMaterialf(GLenum face,GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glMaterialf", sdpGLenum(face) + "," + sdpGLenum(pname) + "," + sdpGLfloat(param));
	glMaterialf(face,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglMaterialfv(GLenum face,GLenum pname,const GLfloat *params)
{
	sdpglBeginFunctionTrace("glMaterialfv", sdpGLenum(face) + "," + sdpGLenum(pname) + ",");
	if(pname == GL_AMBIENT)
		sdpglRawTrace(sdpGLfloat(params[0]) + "," + sdpGLfloat(params[1]) + "," + sdpGLfloat(params[2]));
	else if(pname == GL_DIFFUSE)
		sdpglRawTrace(sdpGLfloat(params[0]) + "," + sdpGLfloat(params[1]) + "," + sdpGLfloat(params[2]));
	else if(pname == GL_SPECULAR)
		sdpglRawTrace(sdpGLfloat(params[0]) + "," + sdpGLfloat(params[1]) + "," + sdpGLfloat(params[2]));
	else if(pname == GL_EMISSION)
		sdpglRawTrace(sdpGLfloat(params[0]) + "," + sdpGLfloat(params[1]) + "," + sdpGLfloat(params[2]));
	else
		sdpGLvoid((void*)params);

	glMaterialfv(face,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglMateriali(GLenum face,GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glMateriali", sdpGLenum(face) + "," + sdpGLenum(pname) + "," + sdpGLint(param));
	glMateriali(face,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglMaterialiv(GLenum face,GLenum pname,const GLint *params)
{
	sdpglBeginFunctionTrace("glMaterialiv", sdpGLenum(face) + "," + sdpGLenum(pname));//,const GLint *params");
	glMaterialiv(face,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglMatrixMode(GLenum mode)
{
	sdpglBeginFunctionTrace("glMatrixMode", sdpGLenum(mode));
	glMatrixMode(mode);
	sdpglEndFunctionTrace();
}

void sdpglMultMatrixd(const GLdouble *m)
{
	sdpglBeginFunctionTrace("glMultMatrixd", sdpGLMatrixd(m));
	glMultMatrixd(m);
	sdpglEndFunctionTrace();
}

void sdpglMultMatrixf(const GLfloat *m)
{
	sdpglBeginFunctionTrace("glMultMatrixf", ""); //,const GLfloat );
	glMultMatrixf(m);
	sdpglEndFunctionTrace();
}

void sdpglNewList(GLuint list,GLenum mode)
{
	sdpglBeginFunctionTrace("glNewList", sdpGLuint(list) + "," + sdpGLenum(mode));
	glNewList(list,mode);
	sdpglEndFunctionTrace();
}

void sdpglNormal3b(GLbyte nx,GLbyte ny,GLbyte nz)
{
	sdpglBeginFunctionTrace("glNormal3b", sdpGLbyte(nx) + "," + sdpGLbyte(ny) + "," + sdpGLbyte(nz));
	glNormal3b(nx,ny,nz);
	sdpglEndFunctionTrace();
}

void sdpglNormal3bv(const GLbyte *v)
{
	sdpglBeginFunctionTrace("glNormal3bv", ""); //,const GLbyte );
	glNormal3bv(v);
	sdpglEndFunctionTrace();
}

void sdpglNormal3d(GLdouble nx,GLdouble ny,GLdouble nz)
{
	sdpglBeginFunctionTrace("glNormal3d", sdpGLdouble(nx) + "," + sdpGLdouble(ny) + "," + sdpGLdouble(nz));
	glNormal3d(nx,ny,nz);
	sdpglEndFunctionTrace();
}

void sdpglNormal3dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glNormal3dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]));
	glNormal3dv(v);
	sdpglEndFunctionTrace();
}

void sdpglNormal3f(GLfloat nx,GLfloat ny,GLfloat nz)
{
	sdpglBeginFunctionTrace("glNormal3f", sdpGLfloat(nx) + "," + sdpGLfloat(ny) + "," + sdpGLfloat(nz));
	glNormal3f(nx,ny,nz);
	sdpglEndFunctionTrace();
}

void sdpglNormal3fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glNormal3fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]) + "," + sdpGLfloat(v[2]) + ",");
	glNormal3fv(v);
	sdpglEndFunctionTrace();
}

void sdpglNormal3i(GLint nx,GLint ny,GLint nz)
{
	sdpglBeginFunctionTrace("glNormal3i", sdpGLint(nx) + "," + sdpGLint(ny) + "," + sdpGLint(nz));
	glNormal3i(nx,ny,nz);
	sdpglEndFunctionTrace();
}

void sdpglNormal3iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glNormal3iv", ""); //,const GLint );
	glNormal3iv(v);
	sdpglEndFunctionTrace();
}

void sdpglNormal3s(GLshort nx,GLshort ny,GLshort nz)
{
	sdpglBeginFunctionTrace("glNormal3s", sdpGLshort(nx) + "," + sdpGLshort(ny) + "," + sdpGLshort(nz));
	glNormal3s(nx,ny,nz);
	sdpglEndFunctionTrace();
}

void sdpglNormal3sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glNormal3sv", ""); //,const GLshort );
	glNormal3sv(v);
	sdpglEndFunctionTrace();
}

void sdpglNormalPointer(GLenum type,GLsizei stride,const GLvoid *pointer)
{
	sdpglBeginFunctionTrace("glNormalPointer", sdpGLenum(type) + "," + sdpGLsizei(stride));//,const GLvoid *pointer");
	glNormalPointer(type,stride,pointer);
	sdpglEndFunctionTrace();
}

void sdpglOrtho(GLdouble left,GLdouble right,GLdouble bottom,GLdouble top,GLdouble zNear,GLdouble zFar)
{
	sdpglBeginFunctionTrace("glOrtho", sdpGLdouble(left) + "," + sdpGLdouble(right) + "," + sdpGLdouble(bottom) + "," + sdpGLdouble(top) + "," + sdpGLdouble(zNear) + "," + sdpGLdouble(zFar));
	glOrtho(left,right,bottom,top,zNear,zFar);
	sdpglEndFunctionTrace();
}

void sdpglPassThrough(GLfloat token)
{
	sdpglBeginFunctionTrace("glPassThrough", sdpGLfloat(token));
	glPassThrough(token);
	sdpglEndFunctionTrace();
}

void sdpglPixelMapfv(GLenum map,GLsizei mapsize,const GLfloat *values)
{
	sdpglBeginFunctionTrace("glPixelMapfv", sdpGLenum(map) + "," + sdpGLsizei(mapsize));//,const GLfloat *values);
	glPixelMapfv(map,mapsize,values);
	sdpglEndFunctionTrace();
}

void sdpglPixelMapuiv(GLenum map,GLsizei mapsize,const GLuint *values)
{
	sdpglBeginFunctionTrace("glPixelMapuiv", sdpGLenum(map) + "," + sdpGLsizei(mapsize));//,const GLuint *values);
	glPixelMapuiv(map,mapsize,values);
	sdpglEndFunctionTrace();
}

void sdpglPixelMapusv(GLenum map,GLsizei mapsize,const GLushort *values)
{
	sdpglBeginFunctionTrace("glPixelMapusv", sdpGLenum(map) + "," + sdpGLsizei(mapsize));//,const GLushort *values);
	glPixelMapusv(map,mapsize,values);
	sdpglEndFunctionTrace();
}

void sdpglPixelStoref(GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glPixelStoref", sdpGLenum(pname) + "," + sdpGLfloat(param));
	glPixelStoref(pname,param);
	sdpglEndFunctionTrace();
}

void sdpglPixelStorei(GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glPixelStorei", sdpGLenum(pname) + "," + sdpGLint(param));
	glPixelStorei(pname,param);
	sdpglEndFunctionTrace();
}

void sdpglPixelTransferf(GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glPixelTransferf", sdpGLenum(pname) + "," + sdpGLfloat(param));
	glPixelTransferf(pname,param);
	sdpglEndFunctionTrace();
}

void sdpglPixelTransferi(GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glPixelTransferi", sdpGLenum(pname) + "," + sdpGLint(param));
	glPixelTransferi(pname,param);
	sdpglEndFunctionTrace();
}

void sdpglPixelZoom(GLfloat xfactor,GLfloat yfactor)
{
	sdpglBeginFunctionTrace("glPixelZoom", sdpGLfloat(xfactor) + "," + sdpGLfloat(yfactor));
	glPixelZoom(xfactor,yfactor);
	sdpglEndFunctionTrace();
}

void sdpglPointSize(GLfloat size)
{
	sdpglBeginFunctionTrace("glPointSize", sdpGLfloat(size));
	glPointSize(size);
	sdpglEndFunctionTrace();
}

void sdpglPolygonMode(GLenum face,GLenum mode)
{
	sdpglBeginFunctionTrace("glPolygonMode", sdpGLenum(face) + "," + sdpGLenum(mode));
	glPolygonMode(face,mode);
	sdpglEndFunctionTrace();
}

void sdpglPolygonOffset(GLfloat factor,GLfloat units)
{
	sdpglBeginFunctionTrace("glPolygonOffset", sdpGLfloat(factor) + "," + sdpGLfloat(units));
	glPolygonOffset(factor,units);
	sdpglEndFunctionTrace();
}

void sdpglPolygonStipple(const GLubyte *mask)
{
	sdpglBeginFunctionTrace("glPolygonStipple", ""); //,const GLubyte *mask");
	glPolygonStipple(mask);
	sdpglEndFunctionTrace();
}

void sdpglPopAttrib()
{
	sdpglBeginFunctionTrace("glPopAttrib", "");
	glPopAttrib();
	sdpglEndFunctionTrace();
}

void sdpglPopClientAttrib()
{
	sdpglBeginFunctionTrace("glPopClientAttrib", "");
	glPopClientAttrib();
	sdpglEndFunctionTrace();
}

void sdpglPopMatrix()
{
	sdpglBeginFunctionTrace("glPopMatrix", "");
	glPopMatrix();
	sdpglEndFunctionTrace();
}

void sdpglPopName()
{
	sdpglBeginFunctionTrace("glPopName", "");
	glPopName();
	sdpglEndFunctionTrace();
}

void sdpglPrioritizeTextures(GLsizei n,const GLuint *textures,const GLclampf *priorities)
{
	sdpglBeginFunctionTrace("glPrioritizeTextures", sdpGLsizei(n));//,const GLuint *textures,const GLclampf *prioriti);
	glPrioritizeTextures(n,textures,priorities);
	sdpglEndFunctionTrace();
}

void sdpglPushAttrib(GLbitfield mask)
{
	sdpglBeginFunctionTrace("glPushAttrib", sdpGLbitfield(mask));
	glPushAttrib(mask);
	sdpglEndFunctionTrace();
}

void sdpglPushClientAttrib(GLbitfield mask)
{
	sdpglBeginFunctionTrace("glPushClientAttrib", sdpGLbitfield(mask));
	glPushClientAttrib(mask);
	sdpglEndFunctionTrace();
}

void sdpglPushMatrix()
{
	sdpglBeginFunctionTrace("glPushMatrix", "");
	glPushMatrix();
	sdpglEndFunctionTrace();
}

void sdpglPushName(GLuint name)
{
	sdpglBeginFunctionTrace("glPushName", sdpGLuint(name));
	glPushName(name);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos2d(GLdouble x,GLdouble y)
{
	sdpglBeginFunctionTrace("glRasterPos2d", sdpGLdouble(x) + "," + sdpGLdouble(y));
	glRasterPos2d(x,y);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos2dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glRasterPos2dv", ""); //,const GLdouble );
	glRasterPos2dv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos2f(GLfloat x,GLfloat y)
{
	sdpglBeginFunctionTrace("glRasterPos2f", sdpGLfloat(x) + "," + sdpGLfloat(y));
	glRasterPos2f(x,y);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos2fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glRasterPos2fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]));
	glRasterPos2fv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos2i(GLint x,GLint y)
{
	sdpglBeginFunctionTrace("glRasterPos2i", sdpGLint(x) + "," + sdpGLint(y));
	glRasterPos2i(x,y);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos2iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glRasterPos2iv", ""); //,const GLint );
	glRasterPos2iv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos2s(GLshort x,GLshort y)
{
	sdpglBeginFunctionTrace("glRasterPos2s", sdpGLshort(x) + "," + sdpGLshort(y));
	glRasterPos2s(x,y);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos2sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glRasterPos2sv", ""); //,const GLshort );
	glRasterPos2sv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos3d(GLdouble x,GLdouble y,GLdouble z)
{
	sdpglBeginFunctionTrace("glRasterPos3d", sdpGLdouble(x) + "," + sdpGLdouble(y) + "," + sdpGLdouble(z));
	glRasterPos3d(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos3dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glRasterPos3dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]));
	glRasterPos3dv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos3f(GLfloat x,GLfloat y,GLfloat z)
{
	sdpglBeginFunctionTrace("glRasterPos3f", sdpGLfloat(x) + "," + sdpGLfloat(y) + "," + sdpGLfloat(z));
	glRasterPos3f(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos3fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glRasterPos3fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]) + "," + sdpGLfloat(v[2]));
	glRasterPos3fv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos3i(GLint x,GLint y,GLint z)
{
	sdpglBeginFunctionTrace("glRasterPos3i", sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLint(z));
	glRasterPos3i(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos3iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glRasterPos3iv", ""); //,const GLint );
	glRasterPos3iv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos3s(GLshort x,GLshort y,GLshort z)
{
	sdpglBeginFunctionTrace("glRasterPos3s", sdpGLshort(x) + "," + sdpGLshort(y) + "," + sdpGLshort(z));
	glRasterPos3s(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos3sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glRasterPos3sv", ""); //,const GLshort );
	glRasterPos3sv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos4d(GLdouble x,GLdouble y,GLdouble z,GLdouble w)
{
	sdpglBeginFunctionTrace("glRasterPos4d", sdpGLdouble(x) + "," + sdpGLdouble(y) + "," + sdpGLdouble(z) + "," + sdpGLdouble(w));
	glRasterPos4d(x,y,z,w);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos4dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glRasterPos4dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]) + "," + sdpGLdouble(v[3]));
	glRasterPos4dv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos4f(GLfloat x,GLfloat y,GLfloat z,GLfloat w)
{
	sdpglBeginFunctionTrace("glRasterPos4f", sdpGLfloat(x) + "," + sdpGLfloat(y) + "," + sdpGLfloat(z) + "," + sdpGLfloat(w));
	glRasterPos4f(x,y,z,w);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos4fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glRasterPos4fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]) + "," + sdpGLfloat(v[2]) + "," + sdpGLfloat(v[3]));
	glRasterPos4fv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos4i(GLint x,GLint y,GLint z,GLint w)
{
	sdpglBeginFunctionTrace("glRasterPos4i", sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLint(z) + "," + sdpGLint(w));
	glRasterPos4i(x,y,z,w);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos4iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glRasterPos4iv", ""); //,const GLint );
	glRasterPos4iv(v);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos4s(GLshort x,GLshort y,GLshort z,GLshort w)
{
	sdpglBeginFunctionTrace("glRasterPos4s", sdpGLshort(x) + "," + sdpGLshort(y) + "," + sdpGLshort(z) + "," + sdpGLshort(w));
	glRasterPos4s(x,y,z,w);
	sdpglEndFunctionTrace();
}

void sdpglRasterPos4sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glRasterPos4sv", ""); //,const GLshort );
	glRasterPos4sv(v);
	sdpglEndFunctionTrace();
}

void sdpglReadBuffer(GLenum mode)
{
	sdpglBeginFunctionTrace("glReadBuffer", sdpGLenum(mode));
	glReadBuffer(mode);
	sdpglEndFunctionTrace();
}

void sdpglReadPixels(GLint x,GLint y,GLsizei width,GLsizei height,GLenum format,GLenum type,GLvoid *pixels)
{
	sdpglBeginFunctionTrace("glReadPixels", sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLsizei(width) + "," + sdpGLsizei(height) + "," + sdpGLenum(format) + "," + sdpGLenum(type));//,GLvoid *pixe);
	glReadPixels(x,y,width,height,format,type,pixels);
	sdpglEndFunctionTrace();
}

void sdpglRectd(GLdouble x1,GLdouble y1,GLdouble x2,GLdouble y2)
{
	sdpglBeginFunctionTrace("glRectd", sdpGLdouble(x1) + "," + sdpGLdouble(y1) + "," + sdpGLdouble(x2) + "," + sdpGLdouble(y2));
	glRectd(x1,y1,x2,y2);
	sdpglEndFunctionTrace();
}

void sdpglRectdv(const GLdouble *v1,const GLdouble *v2)
{
	sdpglBeginFunctionTrace("glRectdv", ""); //,const GLdouble *v1,const GLdouble *v);
	glRectdv(v1,v2);
	sdpglEndFunctionTrace();
}

void sdpglRectf(GLfloat x1,GLfloat y1,GLfloat x2,GLfloat y2)
{
	sdpglBeginFunctionTrace("glRectf", sdpGLfloat(x1) + "," + sdpGLfloat(y1) + "," + sdpGLfloat(x2) + "," + sdpGLfloat(y2));
	glRectf(x1,y1,x2,y2);
	sdpglEndFunctionTrace();
}

void sdpglRectfv(const GLfloat *v1,const GLfloat *v2)
{
	sdpglBeginFunctionTrace("glRectfv", ""); //,const GLfloat *v1,const GLfloat *v);
	glRectfv(v1,v2);
	sdpglEndFunctionTrace();
}

void sdpglRecti(GLint x1,GLint y1,GLint x2,GLint y2)
{
	sdpglBeginFunctionTrace("glRecti", sdpGLint(x1) + "," + sdpGLint(y1) + "," + sdpGLint(x2) + "," + sdpGLint(y2));
	glRecti(x1,y1,x2,y2);
	sdpglEndFunctionTrace();
}

void sdpglRectiv(const GLint *v1,const GLint *v2)
{
	sdpglBeginFunctionTrace("glRectiv", ""); //,const GLint *v1,const GLint *v);
	glRectiv(v1,v2);
	sdpglEndFunctionTrace();
}

void sdpglRects(GLshort x1,GLshort y1,GLshort x2,GLshort y2)
{
	sdpglBeginFunctionTrace("glRects", sdpGLshort(x1) + "," + sdpGLshort(y1) + "," + sdpGLshort(x2) + "," + sdpGLshort(y2));
	glRects(x1,y1,x2,y2);
	sdpglEndFunctionTrace();
}

void sdpglRectsv(const GLshort *v1,const GLshort *v2)
{
	sdpglBeginFunctionTrace("glRectsv", ""); //,const GLshort *v1,const GLshort *v);
	glRectsv(v1,v2);
	sdpglEndFunctionTrace();
}

GLint sdpglRenderMode(GLenum mode)
{
	sdpglBeginFunctionTrace("glRenderMode", sdpGLenum(mode));
	GLint result = glRenderMode(mode);
	sdpglResultTrace(sdpGLint(result));
	sdpglEndFunctionTrace();
	return result;
}

void sdpglRotated(GLdouble angle,GLdouble x,GLdouble y,GLdouble z)
{
	sdpglBeginFunctionTrace("glRotated", sdpGLdouble(angle) + "," + sdpGLdouble(x) + "," + sdpGLdouble(y) + "," + sdpGLdouble(z));
	glRotated(angle,x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglRotatef(GLfloat angle,GLfloat x,GLfloat y,GLfloat z)
{
	sdpglBeginFunctionTrace("glRotatef", sdpGLfloat(angle) + "," + sdpGLfloat(x) + "," + sdpGLfloat(y) + "," + sdpGLfloat(z));
	glRotatef(angle,x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglScaled(GLdouble x,GLdouble y,GLdouble z)
{
	sdpglBeginFunctionTrace("glScaled", sdpGLdouble(x) + "," + sdpGLdouble(y) + "," + sdpGLdouble(z));
	glScaled(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglScalef(GLfloat x,GLfloat y,GLfloat z)
{
	sdpglBeginFunctionTrace("glScalef", sdpGLfloat(x) + "," + sdpGLfloat(y) + "," + sdpGLfloat(z));
	glScalef(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglScissor(GLint x,GLint y,GLsizei width,GLsizei height)
{
	sdpglBeginFunctionTrace("glScissor", sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLsizei(width) + "," + sdpGLsizei(height));
	glScissor(x,y,width,height);
	sdpglEndFunctionTrace();
}

void sdpglSelectBuffer(GLsizei size,GLuint *buffer)
{
	sdpglBeginFunctionTrace("glSelectBuffer", sdpGLsizei(size));//,GLuint *buffer");
	glSelectBuffer(size,buffer);
	sdpglEndFunctionTrace();
}

void sdpglShadeModel(GLenum mode)
{
	sdpglBeginFunctionTrace("glShadeModel", sdpGLenum(mode));
	glShadeModel(mode);
	sdpglEndFunctionTrace();
}

void sdpglStencilFunc(GLenum func,GLint ref,GLuint mask)
{
	sdpglBeginFunctionTrace("glStencilFunc", sdpGLenum(func) + "," + sdpGLint(ref) + "," + sdpGLuint(mask));
	glStencilFunc(func,ref,mask);
	sdpglEndFunctionTrace();
}

void sdpglStencilMask(GLuint mask)
{
	sdpglBeginFunctionTrace("glStencilMask", sdpGLuint(mask));
	glStencilMask(mask);
	sdpglEndFunctionTrace();
}

void sdpglStencilOp(GLenum fail,GLenum zfail,GLenum zpass)
{
	sdpglBeginFunctionTrace("glStencilOp", sdpGLenum(fail) + "," + sdpGLenum(zfail) + "," + sdpGLenum(zpass));
	glStencilOp(fail,zfail,zpass);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord1d(GLdouble s)
{
	sdpglBeginFunctionTrace("glTexCoord1d", sdpGLdouble(s));
	glTexCoord1d(s);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord1dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glTexCoord1dv", ""); //,const GLdouble );
	glTexCoord1dv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord1f(GLfloat s)
{
	sdpglBeginFunctionTrace("glTexCoord1f", sdpGLfloat(s));
	glTexCoord1f(s);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord1fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glTexCoord1fv", sdpGLfloat(v[0]));
	glTexCoord1fv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord1i(GLint s)
{
	sdpglBeginFunctionTrace("glTexCoord1i", sdpGLint(s));
	glTexCoord1i(s);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord1iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glTexCoord1iv", ""); //,const GLint );
	glTexCoord1iv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord1s(GLshort s)
{
	sdpglBeginFunctionTrace("glTexCoord1s", sdpGLshort(s));
	glTexCoord1s(s);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord1sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glTexCoord1sv", ""); //,const GLshort );
	glTexCoord1sv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord2d(GLdouble s,GLdouble t)
{
	sdpglBeginFunctionTrace("glTexCoord2d", sdpGLdouble(s) + "," + sdpGLdouble(t));
	glTexCoord2d(s,t);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord2dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glTexCoord2dv", ""); //,const GLdouble );
	glTexCoord2dv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord2f(GLfloat s,GLfloat t)
{
	sdpglBeginFunctionTrace("glTexCoord2f", sdpGLfloat(s) + "," + sdpGLfloat(t));
	glTexCoord2f(s,t);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord2fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glTexCoord2fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]));
	glTexCoord2fv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord2i(GLint s,GLint t)
{
	sdpglBeginFunctionTrace("glTexCoord2i", sdpGLint(s) + "," + sdpGLint(t));
	glTexCoord2i(s,t);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord2iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glTexCoord2iv", ""); //,const GLint );
	glTexCoord2iv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord2s(GLshort s,GLshort t)
{
	sdpglBeginFunctionTrace("glTexCoord2s", sdpGLshort(s) + "," + sdpGLshort(t));
	glTexCoord2s(s,t);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord2sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glTexCoord2sv", ""); //,const GLshort );
	glTexCoord2sv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord3d(GLdouble s,GLdouble t,GLdouble r)
{
	sdpglBeginFunctionTrace("glTexCoord3d", sdpGLdouble(s) + "," + sdpGLdouble(t) + "," + sdpGLdouble(r));
	glTexCoord3d(s,t,r);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord3dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glTexCoord3dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]));
	glTexCoord3dv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord3f(GLfloat s,GLfloat t,GLfloat r)
{
	sdpglBeginFunctionTrace("glTexCoord3f", sdpGLfloat(s) + "," + sdpGLfloat(t) + "," + sdpGLfloat(r));
	glTexCoord3f(s,t,r);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord3fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glTexCoord3fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]) + "," + sdpGLfloat(v[2]));
	glTexCoord3fv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord3i(GLint s,GLint t,GLint r)
{
	sdpglBeginFunctionTrace("glTexCoord3i", sdpGLint(s) + "," + sdpGLint(t) + "," + sdpGLint(r));
	glTexCoord3i(s,t,r);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord3iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glTexCoord3iv", ""); //,const GLint );
	glTexCoord3iv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord3s(GLshort s,GLshort t,GLshort r)
{
	sdpglBeginFunctionTrace("glTexCoord3s", sdpGLshort(s) + "," + sdpGLshort(t) + "," + sdpGLshort(r));
	glTexCoord3s(s,t,r);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord3sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glTexCoord3sv", ""); //,const GLshort );
	glTexCoord3sv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord4d(GLdouble s,GLdouble t,GLdouble r,GLdouble q)
{
	sdpglBeginFunctionTrace("glTexCoord4d", sdpGLdouble(s) + "," + sdpGLdouble(t) + "," + sdpGLdouble(r) + "," + sdpGLdouble(q));
	glTexCoord4d(s,t,r,q);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord4dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glTexCoord4dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]) + "," + sdpGLdouble(v[3]));
	glTexCoord4dv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord4f(GLfloat s,GLfloat t,GLfloat r,GLfloat q)
{
	sdpglBeginFunctionTrace("glTexCoord4f", sdpGLfloat(s) + "," + sdpGLfloat(t) + "," + sdpGLfloat(r) + "," + sdpGLfloat(q));
	glTexCoord4f(s,t,r,q);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord4fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glTexCoord4fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]) + "," + sdpGLfloat(v[2]) + "," + sdpGLfloat(v[3]));
	glTexCoord4fv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord4i(GLint s,GLint t,GLint r,GLint q)
{
	sdpglBeginFunctionTrace("glTexCoord4i", sdpGLint(s) + "," + sdpGLint(t) + "," + sdpGLint(r) + "," + sdpGLint(q));
	glTexCoord4i(s,t,r,q);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord4iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glTexCoord4iv", ""); //,const GLint );
	glTexCoord4iv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord4s(GLshort s,GLshort t,GLshort r,GLshort q)
{
	sdpglBeginFunctionTrace("glTexCoord4s", sdpGLshort(s) + "," + sdpGLshort(t) + "," + sdpGLshort(r) + "," + sdpGLshort(q));
	glTexCoord4s(s,t,r,q);
	sdpglEndFunctionTrace();
}

void sdpglTexCoord4sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glTexCoord4sv", ""); //,const GLshort );
	glTexCoord4sv(v);
	sdpglEndFunctionTrace();
}

void sdpglTexCoordPointer(GLint size,GLenum type,GLsizei stride,const GLvoid *pointer)
{
	sdpglBeginFunctionTrace("glTexCoordPointer", sdpGLint(size) + "," + sdpGLenum(type) + "," + sdpGLsizei(stride));//,const GLvoid *pointer");
	glTexCoordPointer(size,type,stride,pointer);
	sdpglEndFunctionTrace();
}

void sdpglTexEnvf(GLenum target,GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glTexEnvf", sdpGLenum(target) + "," + sdpGLenum(pname) + "," + sdpGLfloat(param));
	glTexEnvf(target,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglTexEnvfv(GLenum target,GLenum pname,const GLfloat *params)
{
	sdpglBeginFunctionTrace("glTexEnvfv", sdpGLenum(target) + "," + sdpGLenum(pname));//,const GLfloat *params",&);
	glTexEnvfv(target,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglTexEnvi(GLenum target,GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glTexEnvi", sdpGLenum(target) + "," + sdpGLenum(pname) + "," + sdpGLint(param));
	glTexEnvi(target,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglTexEnviv(GLenum target,GLenum pname,const GLint *params)
{
	sdpglBeginFunctionTrace("glTexEnviv", sdpGLenum(target) + "," + sdpGLenum(pname));//,const GLint *params",&);
	glTexEnviv(target,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglTexGend(GLenum coord,GLenum pname,GLdouble param)
{
	sdpglBeginFunctionTrace("glTexGend", sdpGLenum(coord) + "," + sdpGLenum(pname) + "," + sdpGLdouble(param));
	glTexGend(coord,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglTexGendv(GLenum coord,GLenum pname,const GLdouble *params)
{
	sdpglBeginFunctionTrace("glTexGendv", sdpGLenum(coord) + "," + sdpGLenum(pname));//,const GLdouble *params",);
	glTexGendv(coord,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglTexGenf(GLenum coord,GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glTexGenf", sdpGLenum(coord) + "," + sdpGLenum(pname) + "," + sdpGLfloat(param));
	glTexGenf(coord,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglTexGenfv(GLenum coord,GLenum pname,const GLfloat *params)
{
	sdpglBeginFunctionTrace("glTexGenfv", sdpGLenum(coord) + "," + sdpGLenum(pname));//,const GLfloat *params",);
	glTexGenfv(coord,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglTexGeni(GLenum coord,GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glTexGeni", sdpGLenum(coord) + "," + sdpGLenum(pname) + "," + sdpGLint(param));
	glTexGeni(coord,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglTexGeniv(GLenum coord,GLenum pname,const GLint *params)
{
	sdpglBeginFunctionTrace("glTexGeniv", sdpGLenum(coord) + "," + sdpGLenum(pname));//,const GLint *params",);
	glTexGeniv(coord,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglTexImage1D(GLenum target,GLint level,GLint internalformat,GLsizei width,GLint border,GLenum format,GLenum type,const GLvoid *pixels)
{
	sdpglBeginFunctionTrace("glTexImage1D", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLint(internalformat) + "," + sdpGLsizei(width) + "," + sdpGLint(border) + "," + sdpGLenum(format) + "," + sdpGLenum(type));//,const GLvoid *pixels",&);
	glTexImage1D(target,level,internalformat,width,border,format,type,pixels);
	sdpglEndFunctionTrace();
}

void sdpglTexImage2D(GLenum target,GLint level,GLint internalformat,GLsizei width,GLsizei height,GLint border,GLenum format,GLenum type,const GLvoid *pixels)
{
	sdpglBeginFunctionTrace("glTexImage2D", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLint(internalformat) + "," + sdpGLsizei(width) + "," + sdpGLsizei(height) + "," + sdpGLint(border) + "," + sdpGLenum(format) + "," + sdpGLenum(type));//,const GLvoid *pixels",&);
	glTexImage2D(target,level,internalformat,width,height,border,format,type,pixels);
	sdpglEndFunctionTrace();
}

void sdpglTexParameterf(GLenum target,GLenum pname,GLfloat param)
{
	sdpglBeginFunctionTrace("glTexParameterf", sdpGLenum(target) + "," + sdpGLenum(pname) + "," + sdpGLfloat(param));
	glTexParameterf(target,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglTexParameterfv(GLenum target,GLenum pname,const GLfloat *params)
{
	sdpglBeginFunctionTrace("glTexParameterfv", sdpGLenum(target) + "," + sdpGLenum(pname));//,const GLfloat *params",&);
	glTexParameterfv(target,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglTexParameteri(GLenum target,GLenum pname,GLint param)
{
	sdpglBeginFunctionTrace("glTexParameteri", sdpGLenum(target) + "," + sdpGLenum(pname) + "," + sdpGLint(param));
	glTexParameteri(target,pname,param);
	sdpglEndFunctionTrace();
}

void sdpglTexParameteriv(GLenum target,GLenum pname,const GLint *params)
{
	sdpglBeginFunctionTrace("glTexParameteriv", sdpGLenum(target) + "," + sdpGLenum(pname));//,const GLint *params",&);
	glTexParameteriv(target,pname,params);
	sdpglEndFunctionTrace();
}

void sdpglTexSubImage1D(GLenum target,GLint level,GLint xoffset,GLsizei width,GLenum format,GLenum type,const GLvoid *pixels)
{
	sdpglBeginFunctionTrace("glTexSubImage1D", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLint(xoffset) + "," + sdpGLsizei(width) + "," + sdpGLsizei(format) + "," + sdpGLenum(type));//,const GLvoid *pixels",&);
	glTexSubImage1D(target,level,xoffset,width,format,type,pixels);
	sdpglEndFunctionTrace();
}

void sdpglTexSubImage2D(GLenum target,GLint level,GLint xoffset,GLint yoffset,GLsizei width,GLsizei height,GLenum format,GLenum type,const GLvoid *pixels)
{
	sdpglBeginFunctionTrace("glTexSubImage2D", sdpGLenum(target) + "," + sdpGLint(level) + "," + sdpGLint(xoffset) + "," + sdpGLint(yoffset) + "," + sdpGLsizei(width) + "," + sdpGLsizei(height) + "," + sdpGLenum(format) + "," + sdpGLenum(type));//,const GLvoid *pixels",&);
	glTexSubImage2D(target,level,xoffset,yoffset,width,height,format,type,pixels);
	sdpglEndFunctionTrace();
}

void sdpglTranslated(GLdouble x,GLdouble y,GLdouble z)
{
	sdpglBeginFunctionTrace("glTranslated", sdpGLdouble(x) + "," + sdpGLdouble(y) + "," + sdpGLdouble(z));
	glTranslated(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglTranslatef(GLfloat x,GLfloat y,GLfloat z)
{
	sdpglBeginFunctionTrace("glTranslatef", sdpGLfloat(x) + "," + sdpGLfloat(y) + "," + sdpGLfloat(z));
	glTranslatef(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglVertex2d(GLdouble x,GLdouble y)
{
	sdpglBeginFunctionTrace("glVertex2d", sdpGLdouble(x) + "," + sdpGLdouble(y));
	glVertex2d(x,y);
	sdpglEndFunctionTrace();
}

void sdpglVertex2dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glVertex2dv", ""); //,const GLdouble );
	glVertex2dv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex2f(GLfloat x,GLfloat y)
{
	sdpglBeginFunctionTrace("glVertex2f", sdpGLfloat(x) + "," + sdpGLfloat(y));
	glVertex2f(x,y);
	sdpglEndFunctionTrace();
}

void sdpglVertex2fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glVertex2fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]));
	glVertex2fv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex2i(GLint x,GLint y)
{
	sdpglBeginFunctionTrace("glVertex2i", sdpGLint(x) + "," + sdpGLint(y));
	glVertex2i(x,y);
	sdpglEndFunctionTrace();
}

void sdpglVertex2iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glVertex2iv", ""); //,const GLint );
	glVertex2iv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex2s(GLshort x,GLshort y)
{
	sdpglBeginFunctionTrace("glVertex2s", sdpGLshort(x) + "," + sdpGLshort(y));
	glVertex2s(x,y);
	sdpglEndFunctionTrace();
}

void sdpglVertex2sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glVertex2sv", ""); //,const GLshort );
	glVertex2sv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex3d(GLdouble x,GLdouble y,GLdouble z)
{
	sdpglBeginFunctionTrace("glVertex3d", sdpGLdouble(x) + "," + sdpGLdouble(y) + "," + sdpGLdouble(z));
	glVertex3d(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglVertex3dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glVertex3dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]));
	glVertex3dv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex3f(GLfloat x,GLfloat y,GLfloat z)
{
	sdpglBeginFunctionTrace("glVertex3f", sdpGLfloat(x) + "," + sdpGLfloat(y) + "," + sdpGLfloat(z));
	glVertex3f(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglVertex3fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glVertex3fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]) + "," + sdpGLfloat(v[2]));
	glVertex3fv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex3i(GLint x,GLint y,GLint z)
{
	sdpglBeginFunctionTrace("glVertex3i", sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLint(z));
	glVertex3i(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglVertex3iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glVertex3iv", ""); //,const GLint );
	glVertex3iv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex3s(GLshort x,GLshort y,GLshort z)
{
	sdpglBeginFunctionTrace("glVertex3s", sdpGLshort(x) + "," + sdpGLshort(y) + "," + sdpGLshort(z));
	glVertex3s(x,y,z);
	sdpglEndFunctionTrace();
}

void sdpglVertex3sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glVertex3sv", ""); //,const GLshort );
	glVertex3sv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex4d(GLdouble x,GLdouble y,GLdouble z,GLdouble w)
{
	sdpglBeginFunctionTrace("glVertex4d", sdpGLdouble(x) + "," + sdpGLdouble(y) + "," + sdpGLdouble(z) + "," + sdpGLdouble(w));
	glVertex4d(x,y,z,w);
	sdpglEndFunctionTrace();
}

void sdpglVertex4dv(const GLdouble *v)
{
	sdpglBeginFunctionTrace("glVertex4dv", sdpGLdouble(v[0]) + "," + sdpGLdouble(v[1]) + "," + sdpGLdouble(v[2]) + "," + sdpGLdouble(v[3]));
	glVertex4dv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex4f(GLfloat x,GLfloat y,GLfloat z,GLfloat w)
{
	sdpglBeginFunctionTrace("glVertex4f", sdpGLfloat(x) + "," + sdpGLfloat(y) + "," + sdpGLfloat(z) + "," + sdpGLfloat(w));
	glVertex4f(x,y,z,w);
	sdpglEndFunctionTrace();
}

void sdpglVertex4fv(const GLfloat *v)
{
	sdpglBeginFunctionTrace("glVertex4fv", sdpGLfloat(v[0]) + "," + sdpGLfloat(v[1]) + "," + sdpGLfloat(v[2]) + "," + sdpGLfloat(v[3]));
	glVertex4fv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex4i(GLint x,GLint y,GLint z,GLint w)
{
	sdpglBeginFunctionTrace("glVertex4i", sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLint(z) + "," + sdpGLint(w));
	glVertex4i(x,y,z,w);
	sdpglEndFunctionTrace();
}

void sdpglVertex4iv(const GLint *v)
{
	sdpglBeginFunctionTrace("glVertex4iv", ""); //,const GLint );
	glVertex4iv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertex4s(GLshort x,GLshort y,GLshort z,GLshort w)
{
	sdpglBeginFunctionTrace("glVertex4s", sdpGLshort(x) + "," + sdpGLshort(y) + "," + sdpGLshort(z) + "," + sdpGLshort(w));
	glVertex4s(x,y,z,w);
	sdpglEndFunctionTrace();
}

void sdpglVertex4sv(const GLshort *v)
{
	sdpglBeginFunctionTrace("glVertex4sv", ""); //,const GLshort );
	glVertex4sv(v);
	sdpglEndFunctionTrace();
}

void sdpglVertexPointer(GLint size,GLenum type,GLsizei stride,const GLvoid *pointer)
{
	sdpglBeginFunctionTrace("glVertexPointer", sdpGLint(size) + "," + sdpGLenum(type) + "," + sdpGLsizei(stride));//,const GLvoid *pointer");
	glVertexPointer(size,type,stride,pointer);
	sdpglEndFunctionTrace();
}

void sdpglViewport(GLint x,GLint y,GLsizei width,GLsizei height)
{
	sdpglBeginFunctionTrace("glViewport", sdpGLint(x) + "," + sdpGLint(y) + "," + sdpGLsizei(width) + "," + sdpGLsizei(height));
	glViewport(x,y,width,height);
	sdpglEndFunctionTrace();
}


