// K-3D
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the openglsample application, which demonstrates how to mix OpenGL graphics with an SDPGTK UI
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpgtkobjectcontainer.h>
#include <sdpgtkopengldrawingarea.h>
#include <sdpgtkevent.h>
#include <sdpmath.h>

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <vector>

// sdpGtkObjectContainer is the main SDPGTK class - without alteration,
// it encapsulates all the functionality required to load GTKML documents
// and instantiate all their widgets.
//
// We derive our own class from it so we can override OnEvent() and
// handle user events.

//////////////////////////////////////////////////////////////////////////
// SampleWindow

/// The main window for the application
class SampleWindow : public sdpGtkObjectContainer
{
public:
	SampleWindow();

	/// Loads the GTKML template
	void Load();

protected:
	/// Handles events for this template
	void OnEvent(sdpGtkEvent* Event);

	/// Called to exit the program
	void Exit();
	/// Called when the window resized
	void OnConfigureEvent();
	/// Called to redraw the contents of the window
	void OnDraw();
	/// The timer was signalled
	void OnTimer();

	/// OpenGL widget for the main window
	sdpGtkOpenGLDrawingArea m_OpenGLWidget;

	// We skip every other timer event so the refresh loop can keep up ...
	bool m_SkipTimer;
	// Stores our current angle
	gulong m_Angle;
	// Stores our range of colors
	sdpVector3 m_Color1;
	sdpVector3 m_Color2;
	// Stores the call-list offset for our text
	GLuint m_FontBase;
};

SampleWindow::SampleWindow()
{
	m_SkipTimer = false;
	m_Angle = 0;
	m_Color1 = sdpVector3(1, 1, 0);
	m_Color2 = sdpVector3(0, 0, 1);
	m_FontBase = 0;
}

void SampleWindow::Load()
{
	// Load the GTKML document ...
	sdpGtkObjectContainer::LoadFile("mainwindow.gtkml");

	// Create our OpenGL drawing area ...
	sdpGtkContainer container = Container("drawingarea");
	g_return_if_fail(m_OpenGLWidget.Create(container, true, true, 5, 6, 5, 16));

	MapEvent("configure-event", "onconfigureevent", m_OpenGLWidget, true);
	MapEvent("expose-event", "onexposeevent", m_OpenGLWidget, true);

	// Make everything visible ...
	RootWidget().Show();
}

void SampleWindow::OnEvent(sdpGtkEvent* Event)
{
	// Handle events ...
	if(Event->Name() == "ondestroy")
		Exit();
	else if(Event->Name() == "onexit")
		Exit();
	else if(Event->Name() == "onconfigureevent")
		OnConfigureEvent();
	else if(Event->Name() == "onexposeevent")
		OnDraw();
	else if(Event->Name() == "ontimer")
		OnTimer();
	else
		sdpGtkObjectContainer::OnEvent(Event);
}

void SampleWindow::Exit()
{
	DisconnectAllEvents();
	gtk_main_quit();
}

void SampleWindow::OnConfigureEvent()
{
	OnDraw();
}

void SampleWindow::OnDraw()
{
	// If we haven't been properly initialized, yet, we're done ...
	if(!m_OpenGLWidget.Initialized())
		return;

	// Start OpenGL rendering ...
	m_OpenGLWidget.Begin();
	
	if(0 == m_FontBase)
		{
			m_FontBase = glGenLists(256);
			GtkStyle* defaultstyle = gtk_widget_get_default_style();
//			glXUseXFont(gdk_font_id(defaultstyle->font), 0, 256, m_FontBase);
		}

	// Set the viewport to the full size size of the window ...
	GLint width = m_OpenGLWidget.Width();
	GLint height = m_OpenGLWidget.Height();
	glViewport(0, 0, width, height);

	// Clear the buffer ...
	glClearDepth(1.0);
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Now, set the viewport to the largest square that fits in the window, to match our square view volume ...
	GLint size = sdpMin(width, height);
	GLint left = (width - size) / 2;
	GLint top = (height - size) / 2;
	glViewport(left, top, size, size);

	// Setup an orthographic projection ...
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-8.0, 8.0, -8.0, 8.0, 0.0, 100.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Setup shading ...
	glShadeModel(GL_SMOOTH);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	// Enable depth testing for hidden line removal
	glEnable(GL_DEPTH_TEST);

	// Create a Directional Light Source
	GLfloat position [] = { 1.0, 1.0, 1.0, 0.0 };
	glLightfv(GL_LIGHT0, GL_POSITION, position);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);

	// Calculate our current color ...
	if(0 == m_Angle)
		{
			m_Color1 = m_Color2;
			m_Color2 = sdpVector3(rand() / gdouble(RAND_MAX), rand() / gdouble(RAND_MAX), rand() / gdouble(RAND_MAX));
		}
		
	sdpVector3 currentcolor = sdpLerp(m_Color1, m_Color2, gdouble(m_Angle) / 360.0);

	// Setup materials ...
	GLfloat glcolor[4] = { 0.0, 0.0, 0.0, 0.0 };
	currentcolor.CopyArray(glcolor);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, glcolor);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, glcolor);
	GLfloat specular [] = { 1, 1, 1, 1 };
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
	GLfloat shininess [] = { 100 };
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, shininess);

	// Draw the sphere ...
	glTranslated(0.0, 0.0, -8.0);
	glRotated(m_Angle, 1.0, 0.0, 0.0);
	glRotated(m_Angle, 0.0, 1.0, 0.0);
	glRotated(m_Angle, 0.0, 0.0, 1.0);
	glScaled(1.0, 1.0, 0.3);
	GLUquadricObj* sphere = gluNewQuadric();
	gluSphere(sphere, 5.0, 24, 16);
	gluDeleteQuadric(sphere);

	glFlush();

	// Draw text ...
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
		
	sdpString text("SDPGTK OpenGL Sample");

	glColor3f(1, 1, 1);
	glRasterPos2f(-7, 7); 
	glListBase(m_FontBase);
	glCallLists(text.size(), GL_UNSIGNED_BYTE, (GLubyte*)text.c_str());

	// Finish OpenGL rendering ...
	m_OpenGLWidget.SwapBuffers();
	m_OpenGLWidget.End();

	// Increment our angle for next time ...
	m_Angle = (m_Angle + 1) % 360;
}

void SampleWindow::OnTimer()
{
	// Skip every other tick of the timer (gives the UI a chance to respond to events) ...
	m_SkipTimer = !m_SkipTimer;
	if(m_SkipTimer)
		return;

	m_OpenGLWidget.QueueDraw();
}

//////////////////////////////////////////////////////////////////////////////////
// main

/// Initializes GTK+ and loads the main window
gint main(gint argc, gchar* argv[])
{
	if(!sdpglExists(std::cerr))
		return 1;
		
	gtk_init(&argc, &argv);

	SampleWindow window;
	window.Load();

	gtk_main();

	return 0;
}


