
// SDPGTK Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sdpgtktable.h>
#include <sdpgtkutility.h>
#include <sdpxmlutility.h>
#include <sdpmath.h>

////////////////////////////////////////////////////////////////////////////////////////
// sdpGtkTable

sdpGtkTable::sdpGtkTable() : sdpGtkContainer()
{
}

sdpGtkTable::sdpGtkTable(GtkTable* Table) : sdpGtkContainer(GTK_CONTAINER(Table))
{
}

bool sdpGtkTable::Create(gint Rows, gint Columns, bool Homogenous)
{
	// Sanity checks ...
	g_assert(Rows > 0);
	g_assert(Columns > 0);

	m_Object = GTK_OBJECT(gtk_table_new(Rows, Columns, Homogenous));
	return Attached();
}

bool sdpGtkTable::Create(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_assert(ObjectContainer);
	g_assert(Element);

	gint rows = 1;
	sdpXmlGetAttribute(Element, "rows", rows);
	sdpGtkMarkAttribute(Element, "rows");

	gint columns = 1;
	sdpXmlGetAttribute(Element, "columns", columns);
	sdpGtkMarkAttribute(Element, "columns");

	bool homogeneous = true;
	sdpXmlGetAttribute(Element, "homogeneous", homogeneous);
	sdpGtkMarkAttribute(Element, "homogeneous");

	return Create(rows, columns, homogeneous);
}

bool sdpGtkTable::AttachChild(sdpGtkIObjectContainer* ObjectContainer, sdpXmlIElement* ChildElement, sdpGtkObject* Child)
{
	// Sanity checks ...
	g_assert(ChildElement);
	g_assert(Child);

	// Table cells ...
	sdpVector4 position(0, 1, 0, 1);
	sdpXmlGetAttribute(ChildElement, "tableposition", position);
	sdpGtkMarkAttribute(ChildElement, "tableposition");

	// Horizontal options ...
	bool hexpand = true;
	sdpXmlGetAttribute(ChildElement, "hexpand", hexpand);
	sdpGtkMarkAttribute(ChildElement, "hexpand");
	
	bool hshrink = true;
	sdpXmlGetAttribute(ChildElement, "hshrink", hshrink);
	sdpGtkMarkAttribute(ChildElement, "hshrink");
	
	bool hfill = true;
	sdpXmlGetAttribute(ChildElement, "hfill", hfill);
	sdpGtkMarkAttribute(ChildElement, "hfill");

	gint hoptions = (hexpand ? GTK_EXPAND : 0) | (hshrink ? GTK_SHRINK : 0) | (hfill ? GTK_FILL : 0);
	
	// Vertical options ...
	bool vexpand = true;
	sdpXmlGetAttribute(ChildElement, "vexpand", vexpand);
	sdpGtkMarkAttribute(ChildElement, "vexpand");
	
	bool vshrink = true;
	sdpXmlGetAttribute(ChildElement, "vshrink", vshrink);
	sdpGtkMarkAttribute(ChildElement, "vshrink");
	
	bool vfill = true;
	sdpXmlGetAttribute(ChildElement, "vfill", vfill);
	sdpGtkMarkAttribute(ChildElement, "vfill");

	gint voptions = (vexpand ? GTK_EXPAND : 0) | (vshrink ? GTK_SHRINK : 0) | (vfill ? GTK_FILL : 0);

	// Padding ...
	gint xpadding = 0;
	sdpXmlGetAttribute(ChildElement, "xpadding", xpadding);
	sdpGtkMarkAttribute(ChildElement, "xpadding");
	
	gint ypadding = 0;
	sdpXmlGetAttribute(ChildElement, "ypadding", ypadding);
	sdpGtkMarkAttribute(ChildElement, "ypadding");

	// Attach that baby!
	Attach(GTK_WIDGET(Child->Object()), position[0], position[1], position[2], position[3], GtkAttachOptions(hoptions), GtkAttachOptions(voptions), xpadding, ypadding);

	return true;
}

void sdpGtkTable::Attach(GtkWidget* Widget, gint Left, gint Right, gint Top, gint Bottom)
{
	// Sanity checks ...
	g_assert(Widget);
	g_return_if_fail(Attached());
	g_assert(Left >= 0);
	g_assert(Top >= 0);
	g_assert(Right >= 0);
	g_assert(Bottom >= 0);

	gtk_table_attach_defaults(GTK_TABLE(m_Object), Widget, Left, Right, Top, Bottom);
}

void sdpGtkTable::Attach(GtkWidget* Widget, gint Left, gint Right, gint Top, gint Bottom, GtkAttachOptions XOptions, GtkAttachOptions YOptions, gint XPadding, gint YPadding)
{
	// Sanity checks ...
	g_assert(Widget);
	g_return_if_fail(Attached());
	g_assert(Left >= 0);
	g_assert(Top >= 0);
	g_assert(Right >= 0);
	g_assert(Bottom >= 0);

	gtk_table_attach(GTK_TABLE(m_Object), Widget, Left, Right, Top, Bottom, XOptions, YOptions, XPadding, YPadding);
}

void sdpGtkTable::Resize(gint Rows, gint Columns)
{
	// Sanity checks ...
	g_return_if_fail(Attached());

	gtk_table_resize(GTK_TABLE(m_Object), Rows, Columns);
}

void sdpGtkTable::SetRowSpacing(guint Row, guint Spacing)
{
	// Sanity checks ...
	g_return_if_fail(Attached());

	gtk_table_set_row_spacing(GTK_TABLE(m_Object), Row, Spacing);
}

void sdpGtkTable::SetColumnSpacing(guint Column, guint Spacing)
{
	// Sanity checks ...
	g_return_if_fail(Attached());

	gtk_table_set_col_spacing(GTK_TABLE(m_Object), Column, Spacing);
}

void sdpGtkTable::SetRowSpacings(guint Spacing)
{
	// Sanity checks ...
	g_return_if_fail(Attached());

	gtk_table_set_row_spacings(GTK_TABLE(m_Object), Spacing);
}

void sdpGtkTable::SetColumnSpacings(guint Spacing)
{
	// Sanity checks ...
	g_return_if_fail(Attached());

	gtk_table_set_col_spacings(GTK_TABLE(m_Object), Spacing);
}

void sdpGtkTable::SetHomogeneous(bool Homogeneous)
{
	// Sanity checks ...
	g_return_if_fail(Attached());

	gtk_table_set_homogeneous(GTK_TABLE(m_Object), Homogeneous);
}

