// SDPSLPARSE
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements the sdpslparse command-line application, which converts Pixar Renderman Shading Language on stdin to XML on stdout
		\author Tim Shead (tshead@k-3d.com)
*/

#include <iostream>
#include <strstream>
#include <fstream>
#include <iterator>

#include <string>
#include <vector>
#include <algorithm>

#include <sdpxmldocument.h>
#include <sdpmath.h>

#include "sdpslparse_version.h"

// Define some character constants for parsing ...
static const char singlequote = '\'';
static const char doublequote = '"';
static const char escape = '\\';
static const char openbrace = '{';
static const char closebrace = '}';
static const char openparen = '(';
static const char closeparen = ')';
static const char openbracket = '[';
static const char closebracket = ']';
static const char equals = '=';
static const char comma = ',';
static const char semicolon = ';';

/// A string token
typedef sdpString Token;

/// Stores a collection of tokens
typedef std::vector<Token> Tokens;

////////////////////////////////////////////////////////////////////////////////////////////////
// WriteToken

/// Stores a parsed token
void WriteToken(Tokens& Output, Token& OutputToken)
{
	if(OutputToken.size())
		Output.push_back(OutputToken);

	OutputToken.clear();
}

////////////////////////////////////////////////////////////////////////////////////////////////
// ParseStream

/// Parses a stream into a series of whitespace-delimited tokens.  Handles string constants and special characters.
void ParseStream(std::istream& Stream, Tokens& Output)
{
	// Our input stream ...
	std::strstream stream;

	Stream.get(*stream.rdbuf(), '\0');
	stream << '\0';

	// Convert it to a string buffer ...
	std::string buffer = stream.str();
	stream.freeze(false);

	// Loop through the buffer ...
	char stringliteral = 0;

	Token token;

	for(Token::iterator c = buffer.begin(); c != buffer.end(); c++)
		{
			if(stringliteral)
				{
					if((*c) == stringliteral && ((*(c-1)) != escape))
						{
							token += *c;
							WriteToken(Output, token);
							stringliteral = 0;
						}
					else if((*c) == stringliteral && ((*(c-1)) == escape))
						{
							token.erase(token.end()-1);
							token += *c;
						}
					else
						{
							token += *c;
						}
				}
			else if(isspace(*c))
				{
					if(stringliteral)
						token += *c;
					else
						WriteToken(Output, token);
				}
			else
				{
					switch(*c)
						{
							case singlequote:
							case doublequote:
								stringliteral = *c;
								token += *c;
								break;

							case openparen:
							case closeparen:
							case openbracket:
							case closebracket:
							case openbrace:
							case closebrace:
							case equals:
							case comma:
							case semicolon:
								WriteToken(Output, token);
								token = *c;
								WriteToken(Output, token);
								break;

							default:
								token += *c;
						}
				}
		}
}

////////////////////////////////////////////////////////////////////////////////////////////////
// IsOutput

/// Tests a token to see if it's the "output" token used for light -> shader message passing
bool IsOutput(const Token& InputToken)
{
	return InputToken == "output" ? true : false;
}

////////////////////////////////////////////////////////////////////////////////////////////////
// IsStorage

/// Tests a token to see if it's a storage type
bool IsStorage(const Token& InputToken)
{
	static const char* types[] = 
	{
		"uniform", "varying"
	};
	
	for(gulong i = 0; i < sizeof(types) / sizeof(const char*); i++)
		if(InputToken == types[i])
			return true;
		
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////
// IsArgumentType

/// Tests a token to see if it's an argument type
bool IsArgumentType(const Token& InputToken)
{
	static const char* types[] = 
	{
		"float", "color", "point", "vector", "normal", "matrix", "string"
	};
	
	for(gulong i = 0; i < sizeof(types) / sizeof(const char*); i++)
		if(InputToken == types[i])
			return true;
		
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////
// IsString

/// Tests a token to see if it's a string type
bool IsString(const Token& InputToken)
{
	if((InputToken[0] == singlequote) || (InputToken[0] == doublequote))
		return true;
		
	return false;
}

////////////////////////////////////////////////////////////////////////////////////////////////
// ParseArguments

/// Parses tokens to pick out all of the shader arguments ...

void ParseArguments(sdpXmlIElement* Shader, Tokens& Arguments)
{
	// Sanity checks ...
	g_assert(Shader);

//std::copy(Arguments.begin(), Arguments.end(), std::ostream_iterator<sdpString>(std::cout));
//std::cout << std::endl;

//for(Tokens::const_iterator i = Arguments.begin(); i != Arguments.end(); i++)
//	std::cout << "[" << *i << "]" << std::endl;

	bool output = false;
	sdpString storage = "uniform";
	sdpString type = "";

	for(Tokens::const_iterator token = Arguments.begin(); token != Arguments.end(); token++)
		{
//std::cout << *token << std::endl;

			// Look to see if this is an output variable, part of the light->shader message system ...
			if(IsOutput(*token))
				{
					output = true;
					token++;
				}
		
			// Check for a storage class ...
			if(IsStorage(*token))
				{
					storage = *token;
					token++;
				}
	
			// Get the (required) argument type ...
			if(IsArgumentType(*token))
				{
					type = *token;
					token++;
				}

			// Whatever's left over has to be the argument name ...
			sdpString name = "";
			name = *token;
			token++;

			// Check to see if this argument is a matrix ...
			sdpString arraysize = "";
			if(*token == "[")
				{
					token++;
					arraysize = *token;
					token++;
					g_return_if_fail(*token == "]");
					token++;
				}
				
			// Sanity check for the (required) default value ...
			g_return_if_fail(*token == "=");
			token++;

			// Some types syntactically use the type name when declaring constants - we can safely skip these ...
			if(IsArgumentType(*token))
				token++;
				
			// If the next token is a string, and the argument isn't of string type, it must be a color/point/vector space ...
			sdpString argumentspace = "";
			if(IsString(*token) && (type != "string"))
				{
					argumentspace = *token;
					token++;
				}

			// Next, we start accumulating the argument default ...
			bool sametype = false;
			gulong depth = 0;
			sdpString argumentdefault = "";
			for( ; token != Arguments.end(); token++)
				{
					if((*token == ")") && (0 == depth))
						{
							break;
						}
					else if((*token == "(") || (*token == "{"))
						{
							argumentdefault += *token;
							depth++;
						}
					else if((*token == "}") || (*token == ")"))
						{
							argumentdefault += *token;
							depth--;
						}
					else if(*token == ";")
						{
							break;
						}
					else if((*token == ",") && (0 == depth))
						{
							sametype = true;
							break;
						}
					else
						{
							argumentdefault += *token;
						}
				}

			// Sanity checks ...
			g_return_if_fail(type.size());
			g_return_if_fail(name.size());

			// Massage our "array" types for consistency ...
			if((type == "color") || (type == "point") || (type == "vector") || (type == "normal"))
				{
					sdpVector3 temp;
					argumentdefault.getvalue(temp);
					argumentdefault = sdpString(temp);
				}

			// Get rid of excess quote marks and whitespace in strings ...
			if(argumentdefault.size())
				{
					if(IsString(argumentdefault))
						argumentdefault = argumentdefault.mid(1, argumentdefault.size()-2);
				}

			if(argumentspace.size())
				{
					if(IsString(argumentspace))
						argumentspace = argumentspace.mid(1, argumentspace.size()-2);
				}

			// Create the argument element ...
			sdpXmlIElement* argument = Shader->CreateElement("argument");
			g_assert(argument);

			// Set argument attributes ...
			argument->CreateAttribute("storage", storage);
			argument->CreateAttribute("type", type);
			argument->CreateAttribute("name", name);
			argument->CreateAttribute("default", argumentdefault);
			
			if(arraysize.size())
				argument->CreateAttribute("arraysize", arraysize);

			if(argumentspace.size())
				argument->CreateAttribute("space", argumentspace);

			if(output)
				argument->CreateAttribute("output", "true");
				
			// Clear our type information, unless our arguments were separated by commas ...
			if(!sametype)
				{
					output = false;
					storage = "uniform";
					type = "";
				}
				
			// Make sure we don't overrun the end of tokens ...
			if(token == Arguments.end())
				break;
		}
}

////////////////////////////////////////////////////////////////////////////////////////////////
// ParseShader

/// Parses tokens looking for a shader, and generates an XML document
void ParseShader(sdpXmlIElement* Root, Tokens& InputTokens)
{
	// Sanity checks ...
	g_assert(Root);
	
	// Find the beginning of a shader (by looking for a shader type) ...
	static const char* shadertypes[] = 
	{
		"light", "surface", "volume", "displacement", "transformation", "imager"
	};
	
	static const gulong shadertypecount = sizeof(shadertypes) / sizeof(const char*);
	
	Tokens::const_iterator shadertype = std::find_first_of(InputTokens.begin(), InputTokens.end(), shadertypes, shadertypes + shadertypecount);
	g_return_if_fail(shadertype != InputTokens.end());
	
	// The shader name should be the next token ...
	Tokens::const_iterator shadername = shadertype + 1;
	g_return_if_fail(shadername != InputTokens.end());

	// We now have the beginning of our shader arguments ...
	Tokens::const_iterator argstart = shadername+1;
	g_return_if_fail(argstart != InputTokens.end());
	g_return_if_fail(*argstart == "(");
		
	// Look for the end of the shader arguments ... since some arguments use parentheses in their declarations,
	// we have to maintain a count of nested parenthesis to figure out when we're done ...
	Tokens::const_iterator argend = argstart;
	gulong depth = 1;
	do
		{
			argend++;
			
			if(*argend == "(")
				depth++;
			else if(*argend == ")")
				depth--;
		}
	while(depth);
		
	// We now have all the argument tokens ...
	Tokens arguments;
	std::copy(argstart+1, argend, std::back_inserter(arguments));
		
	// Create the shader XML element ...
	sdpXmlIElement* shaderelement = Root->CreateElement("shader");
	g_assert(shaderelement);
	
	// Set some XML attributes ...
	shaderelement->CreateAttribute("type", *shadertype);
	shaderelement->CreateAttribute("name", *shadername);
	
	// Parse the arguments (if any) ...
	if(arguments.size())
		ParseArguments(shaderelement, arguments); 
}

////////////////////////////////////////////////////////////////////////////////////////////////
// main

int main(int argc, char* argv[])
{
	// Keep track of command line arguments that aren't options ...
	std::vector<sdpString> files;

	// Look for command-line options ...
	for(gint i = 1; i < argc; i++)
		{
			const sdpString argument = argv[i];

			if(argument == "--help" || argument == "-h")
				{
					std::cout << "Usage: sdpslparse [options] [file ...]" << std::endl;
					std::cout << std::endl;
					std::cout << "  -h, --help         prints this help information and exits" << std::endl;
					std::cout << "      --version      prints program version information and exits" << std::endl;
					std::cout << std::endl;
					exit(0);
				}
			else if(argument == "--version")
				{
					std::cout << "SDPSLPARSE Version " << VERSION_STRING << std::endl;
					std::cout << "Copyright (c) 1995-2000, Timothy M. Shead.  See the AUTHORS file for contributors." << std::endl;
					std::cout << "Licensed under the GNU General Public License.  See the COPYING file for details." << std::endl;
					std::cout << "SDPSLPARSE Home Page: http://www.k-3d.com" << std::endl;
					std::cout << "Bug reports to tshead@k-3d.com" << std::endl;
					exit(0);
				}
			else
				{
					files.push_back(argument);
				}
		}

	// Begin an XML document ...
	sdpXmlDocument document;
	sdpXmlIElement* root = document.CreateRootElement("k3dml");
	g_assert(root);

	// If we have filenames supplied on the command-line, read from them ...
	if(files.size())
		{
			for(gulong i = 0; i < files.size(); i++)
				{
					std::ifstream file(files[i]);
					Tokens tokens;
					ParseStream(file, tokens);
					ParseShader(root, tokens);
				}
		}
	// Otherwise, read from cin ...
	else
		{
			Tokens tokens;
			ParseStream(std::cin, tokens);
			ParseShader(root, tokens);
		}

	// Write the document to cout ...
	std::cout << document;

	return 0;
}


