#ifndef SDPFILE_H
#define SDPFILE_H

// SDPTypes Library
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/** \file
		\brief Declares templates, functions, and classes for manipulating files in a platform-independant manner
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpstring.h>

/// Defines the maximum filepath length on the target platform
extern const gulong sdpMaxPathLength;

/// Returns the full path to the program executable
sdpString sdpGetProcessPath(SDPCSTRING ArgVZero);
/// Safely returns an environment variable
sdpString sdpGetEnv(SDPCSTRING EnvironmentVariable);
/// Safely returns the current working directory
sdpString sdpGetCurrentDirectory();
/// Safely returns the user's home directory
sdpString sdpGetHomeDirectory();
/// Safely returns the user's temp directory
sdpString sdpGetTempDirectory();

/// Cross-platform method for creating a directory
bool sdpCreateDirectory(SDPCSTRING Path);
/// Tests to see if the given file exists
bool sdpFileExists(SDPCSTRING Path);

/// Forces a filepath to the native filesystem's format (converts slashes, etc.)
sdpString sdpMakeNativeFilePath(SDPCSTRING FilePath);
/// Forces a filepath to the Posix format (forward slashes)
sdpString sdpMakePosixFilePath(SDPCSTRING FilePath);

/// Returns the directory components of a filepath
sdpString sdpDirectory(SDPCSTRING FilePath);
/// Returns the file component of a filepath
sdpString sdpFile(SDPCSTRING FilePath);

/// Appends two paths (handles a trailing slash on the first path, if present)
sdpString sdpAppendPath(SDPCSTRING Path1, SDPCSTRING Path2);

/// Runs a process using popen() and redirects its output to a stream, returning when the process completes
bool sdpRunProcess(SDPCSTRING CommandLine, std::ostream& Stream);

/// Defines an array of filepaths
typedef std::vector<sdpString> sdpFileContainer;
/// Defines an iterator for an array of filepaths
typedef sdpFileContainer::iterator sdpFileIterator;

/// Add the each file within a directory to a list of paths
void sdpGetFiles(SDPCSTRING StartDirectory, const bool Recursive, sdpFileContainer& Container);
/// Adds subdirectories within a directory to a list of paths
void sdpGetDirectories(SDPCSTRING StartDirectory, const bool Recursive, sdpFileContainer& Container);

//////////////////////////////////////////////////////////////////////////////////////////////////////
// The Win32 versions of sdpForEachFile() and sdpForEachDirectory() ...

#if defined SDPWIN32
#include <windows.h>

/// Applies a functor object to each file in the given directory
template<class Type>
bool sdpForEachFile(SDPCSTRING StartDirectory, Type& Functor, const bool Recursive)
{
	// Generate the full search path ...
	sdpString startpath = sdpString(StartDirectory) + "\\*.*";

	// See if we've got at least one file that fits ...
	WIN32_FIND_DATA data;
	HANDLE handle = FindFirstFile(startpath, &data);
	if(handle != INVALID_HANDLE_VALUE)
		{
			// For each file that fits the spec ...
			while(1)
				{
					// Get the filename & filepath ...
					sdpString filename = data.cFileName;
					sdpString filepath = sdpString(StartDirectory) + "\\" + filename;

					// Only the call the function for files ...
					if(!(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
						if(!Functor.HandleFile(filepath, filename))
							return false;

					// Is the file a directory ... ?
					if((data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) && Recursive)
						{
							// Don't recurse into the current and parent directories ...
							if((filename != ".") && (filename != ".."))
								sdpForEachFile(filepath, Functor, Recursive);
						}

					if(!FindNextFile(handle, &data))
						break;
				}

			FindClose(handle);
		}

	return true;
}


/// Applies a functor object to each subdirectory in the given directory
template<class Type>
bool sdpForEachDirectory(SDPCSTRING StartDirectory, Type& Functor, const bool Recursive)
{
	// Generate the full search path ...
	sdpString filepath = sdpString(StartDirectory) + "\\*.*";

	// See if we've got at least one file that fits ...
	WIN32_FIND_DATA data;
	HANDLE handle = FindFirstFile(filepath, &data);
	if(handle != INVALID_HANDLE_VALUE)
		{
			// For each file that fits the spec ...
			while(1)
				{
					// Only handle directories ...
					if(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
						{
							// Get the filename & filepath ...
							sdpString filename = data.cFileName;
							sdpString filepath = sdpString(StartDirectory) + "\\" + filename;

							// Don't call the function for current and parent directories ...
							if((filename != ".") && (filename != ".."))
								{
									if(!Functor.HandleDirectory(filepath, filename))
										return false;

									if(Recursive)
										sdpForEachDirectory(filepath, Functor, Recursive);
								}
						}

					if(!FindNextFile(handle, &data))
						break;
				}

			FindClose(handle);
		}

	return true;
}


/////////////////////////////////////////////////////////////////////////////////////////
// The Posix versions of sdpForEachFile() and sdpForEachDirectory() ...

#else	// SDPWIN32

#include <dirent.h>

// Applies a functor object to each file within a directory
template<class Type>
bool sdpForEachFile(SDPCSTRING StartDirectory, Type& Functor, const bool Recursive)
{
	// Open the start directory ...
	DIR* directory = opendir(StartDirectory);
	g_return_val_if_fail(directory, false);

	// For each file within the directory ...
	dirent* child;
	while(child = readdir(directory))
		{
			// Get the filename & filepath ...
			sdpString filename=child->d_name;
			sdpString filepath = sdpString(StartDirectory) + "/" + filename;

			// If it's a directory ...
			DIR* childdirectory = opendir(filepath);
			if(childdirectory)
				{
					if(Recursive)
						{
							// Don't recurse into the current and parent directories ...
							if((filename != ".") && (filename != ".."))
								sdpForEachFile(filepath, Functor, Recursive);
						}

					closedir(childdirectory);
				}
			else
				{
					// Call the user function for our file ...
					if(!Functor.HandleFile(filepath, filename))
						return false;
				}
		}

	closedir(directory);
	return true;
}

// Applies a functor object to each subdirectory within the given directory
template<class Type>
bool sdpForEachDirectory(SDPCSTRING StartDirectory, Type& Functor, const bool Recursive)
{
	// Open the start directory ...
	DIR* directory = opendir(StartDirectory);
	g_return_val_if_fail(directory, false);

	// For each file within the directory ...
	dirent* child;
	while(child = readdir(directory))
		{
			// Get the filename & filepath ...
			sdpString filename=child->d_name;
			sdpString filepath = sdpString(StartDirectory) + "/" + filename;

			// If it's a directory ...
			DIR* childdirectory = opendir(filepath);
			if(childdirectory)
				{
					// Don't call the function for current and parent directories ...
					if((filename != ".") && (filename != ".."))
						{
							if(!Functor.HandleDirectory(filepath, filename))
								return false;

							if(Recursive)
								sdpForEachDirectory(filepath, Functor, Recursive);
						}

					closedir(childdirectory);
				}
		}

	closedir(directory);
	return true;
}

#endif // !SDPWIN32



#if defined SDPWIN32
/// Functor for deleting files, for use with sdpForEachFile()
class sdpTypesDeleteFile
{
public:
	bool HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName)
	{
		DeleteFile(FilePath);
		return true;
	}
};

#else // SDPWIN32

#include <unistd.h>

/// Functor for deleting files, for use with sdpForEachFile()
class sdpTypesDeleteFile
{
public:
	bool HandleFile(SDPCSTRING FilePath, SDPCSTRING FileName)
	{
		unlink(FilePath);
		return true;
	}
};

#endif // !SDPWIN32



#endif // SDPFILE_H
