#ifndef SDPSTRING_H
#define SDPSTRING_H

// SDPTypes Library
// Copyright  1995 - 2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/** \file
		\brief Declares an extended string class, sdpString, derived from std::string
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpsimpletypes.h>
#include <sdpcom.h>

#include <functional>
#include <string>
#include <vector>
#include <list>
#include <stack>
#include <set>
#include <map>

// Forward class declarations
class sdpVector2;
class sdpVector3;
class sdpVector4;
class sdpAngleAxis;
class sdpQuaternion;
class sdpEulerAngles;

// Stop the unicode madness!
#if defined (_UNICODE) || defined (UNICODE)
	#error "sdpString does not support UNICODE builds!"
#endif

/// The sdpString base class
typedef std::string sdpStringBase;

////////////////////////////////////////////////////////////////////////////////////////////////
// sdpString

/// An extended string class, derived from std::string
class  sdpString : public sdpStringBase
{
public:
	// Constructors ...
	sdpString();
	sdpString(SDPCSTRING String);
	sdpString(SDPCSTRING String, gulong Length);
	sdpString(const sdpString& String);
	sdpString(const sdpStringBase& String);
	sdpString(size_type Size, value_type Character);
	sdpString(const_iterator First, const_iterator Last);

	// Value Constructors ...
	sdpString(const bool& Value);
	sdpString(const gshort& Value);
	sdpString(const gushort& Value);
	sdpString(const gint& Value);
	sdpString(const guint& Value);
	sdpString(const glong& Value);
	sdpString(const gulong& Value);
	sdpString(const gfloat& Value);
	sdpString(const gdouble& Value);
	sdpString(const sdpVector2& Value);
	sdpString(const sdpVector3& Value);
	sdpString(const sdpVector4& Value);
	sdpString(const sdpAngleAxis& Value);
	sdpString(const sdpQuaternion& Value);
	sdpString(const sdpEulerAngles& Value);
	sdpString(const sdpGUID& Value);
	sdpString(const void*& Value);

	// Assignment operators ...
	sdpString& operator=(const sdpString& String);
	sdpString& operator=(const sdpStringBase& String);
	sdpString& operator=(SDPCSTRING String);
	sdpString& operator=(gchar Character);

	// comparison (equality)
	bool operator == (const sdpString& String) const;
	bool operator == (const sdpStringBase& String) const;
	bool operator == (SDPCSTRING String) const;
	bool operator == (gchar* String) const;
	bool operator == (gchar Character) const;
	/// Performs a case (in)sensitive equality test
	bool equals(SDPCSTRING String, bool UseCase = false) const;
	/// Performs a case (in)sensitive comparison
	bool lessthan(SDPCSTRING String, bool UseCase = false) const;

	// comparison operators
	bool operator != (const sdpString& String) const;
	bool operator != (const sdpStringBase& String) const;
	bool operator != (SDPCSTRING String) const;
	bool operator != (gchar* String) const;
	bool operator != (gchar Character) const;
	bool operator < (const sdpString& String) const;
	bool operator < (SDPCSTRING String) const;

	// concatenation
	const sdpString& operator += (const sdpString& String);
	const sdpString& operator += (const sdpStringBase& String);
	const sdpString& operator += (SDPCSTRING String);
	const sdpString& operator += (gchar Character);

	// Concatenation operators -- global friend functions.
	friend sdpString operator + (const sdpString& String1, const sdpString& String2);
	friend sdpString operator + (const sdpString& String, gchar Character);
	friend sdpString operator + (gchar Character, const sdpString& String);
	friend sdpString operator + (const sdpString& String1, SDPCSTRING String2);
	friend sdpString operator + (SDPCSTRING String1, const sdpString& String2);

	/// Returns a reference to the character at the given index
	#ifdef	SDPWIN32
	inline gchar& operator[](gint Index) { return sdpStringBase::operator[](Index); }
	#endif
	inline gchar& operator[](size_type Index) { return sdpStringBase::operator[](Index); }

	/// Returns the leftmost N characters of the string
	sdpString left(size_type Count) const;
	/// Returns a substring with the given starting index and size
	sdpString mid(size_type Start, size_type Count) const;
	/// Returns the rightmost N characters of the string
	sdpString right(size_type Count) const;

	/// Trims whitespace from both ends of the string
	sdpString& trim();
	/// Trims beginning whitespace from the string
	sdpString& trimleft();
	/// Trims trailing whitespace from the string
	sdpString& trimright();

	/// Converts all characters in the string to upper case
	sdpString& makeupper();
	/// Converts all characters in the string to lower case
	sdpString& makelower();

	/// Casts to a C style string ...
	inline operator const gchar*() const { return c_str(); }
	/// Simplifies use of the string as a fixed-size buffer argument to c-style APIs
	SDPSTRING buffer(size_type BufferSize);

	/// Builds a string using printf() style formatting
	void format(SDPCSTRING FormatString, ...);
	/// Builds a string using printf() style formatting
	void formatv(SDPCSTRING FormatString, va_list argList);
	/// Substitutes text into printf() style format specifiers
	void formattedreplace(const char Delimiter, SDPCSTRING Specifier, SDPCSTRING Replacement);

	/// Replaces every instance of Target with Replacement
	void replaceall(SDPCSTRING Target, SDPCSTRING Replacement);
	/// Replaces every instance of Target with Replacement
	void replaceall(const char Target, const char Replacement);

	/// Reverses the characters of the string in place
	sdpString& makereverse();
	sdpString spanincluding(SDPCSTRING Set);
	sdpString spanexcluding(SDPCSTRING Set);

	// Test the string contents ...
	const bool isalphanumeric() const;
	const bool isalphabetic() const;
	const bool isnumeric() const;
	const bool islowercase() const;
	const bool isprintable() const;
	const bool ispunctuation() const;
	const bool iswhitespace() const;
	const bool isuppercase() const;

	// Safe parsing into other types ...
	const bool getvalue(bool& Result) const;
	const bool getvalue(gshort& Result) const;
	const bool getvalue(gushort& Result) const;
	const bool getvalue(gint& Result) const;
	const bool getvalue(guint& Result) const;
	const bool getvalue(glong& Result) const;
	const bool getvalue(gulong& Result) const;
	const bool getvalue(gfloat& Result) const;
	const bool getvalue(gdouble& Result) const;
	const bool getvalue(sdpVector2& Result) const;
	const bool getvalue(sdpVector3& Result) const;
	const bool getvalue(sdpVector4& Result) const;
	const bool getvalue(sdpAngleAxis& Result) const;
	const bool getvalue(sdpQuaternion& Result) const;
	const bool getvalue(sdpEulerAngles& Result) const;
	const bool getvalue(sdpGUID& Result) const;
	const bool getvalue(void*& Result) const;
	const bool getvalue(sdpString& Result) const;

	/// Clears the string contents
	sdpString& clear();

	/// Copies one C-style string to another ...
	static void copystring(SDPCSTRING Source, SDPSTRING Destination, gint MaxChars = 0);

protected:
	/// Removes parentheses and commas so vector values can be treated consistently
	sdpString vectorcleanup() const;
};

// Inline constructors ...
inline sdpString::sdpString() : sdpStringBase() {}
inline sdpString::sdpString(const sdpString& String) : sdpStringBase(String) {}
inline sdpString::sdpString(const sdpStringBase& String) : sdpStringBase(String) {}
inline sdpString::sdpString(SDPCSTRING String) : sdpStringBase(String) {}
inline sdpString::sdpString(SDPCSTRING String, gulong Length) : sdpStringBase(String, Length) {}
inline sdpString::sdpString(const_iterator First, const_iterator Last) : sdpStringBase(First, Last) {}
inline sdpString::sdpString(size_type Size, value_type Character) : sdpStringBase(Size, Character) {}


// sdpString inline assignment operators -- the erase() function call fixes the MSVC assignment bug (see knowledge base article Q172398)
inline sdpString& sdpString::operator=(const sdpString& String)
{
	erase();
	assign(String.c_str());
	return *this;
};

inline sdpString& sdpString::operator=(const sdpStringBase& String)
{
	erase();
	assign(String.c_str());
	return *this;
};

inline sdpString& sdpString::operator=(SDPCSTRING String)
{
	erase();
	if(String != 0)
		assign(String);
	return *this;
};

inline sdpString& sdpString::operator=(gchar Character)
{
	erase();
	assign(1, Character);

	return *this;
};


// sdpString inline comparison (equality)
inline bool sdpString::operator==(const sdpString& String) const
{
	return compare(String.c_str()) == 0;
}

inline bool sdpString::operator==(const sdpStringBase& String) const
{
	return compare(String.data()) == 0;
}

inline bool sdpString::operator==(SDPCSTRING String) const
{
	return compare(String) == 0;
}

inline bool sdpString::operator==(gchar* String) const
{
	return compare(String) == 0;
}

inline bool sdpString::operator==(gchar Character) const
{
	return operator==(sdpStringBase(1, Character));
}

inline bool sdpString::equals(SDPCSTRING String, bool UseCase) const
{
	if(UseCase)
		return compare(String) == 0;

	sdpString a = *this;
	sdpString b = String;

	a.makelower();
	b.makelower();
	return a.compare(b) == 0;
}

inline bool sdpString::lessthan(SDPCSTRING String, bool UseCase) const
{
	if(UseCase)
		return compare(String) < 0;

	sdpString a = *this;
	sdpString b = String;

	a.makelower();
	b.makelower();
	return a.compare(b) < 0;
}

// sdpString inline comparison operators (inequality)

inline bool sdpString::operator!=(const sdpString& String) const
{
	return compare(String.c_str()) != 0;
}

inline bool sdpString::operator!=(const sdpStringBase& String) const
{
	return compare(String.data()) != 0;
}

inline bool sdpString::operator!=(SDPCSTRING String) const
{
	return compare(String) != 0;
}

inline bool sdpString::operator!=(gchar* String) const
{
	return compare(String) != 0;
}

inline bool sdpString::operator!=(gchar Character) const
{
	return operator!=(sdpStringBase(1, Character));
}

// sdpString comparison operator (less than)

inline bool sdpString::operator<(const sdpString& String) const
{
	return compare(String) < 0;
}

inline bool sdpString::operator<(SDPCSTRING String) const
{
	return compare(String) < 0;
}


// sdpString inline concatenation.

inline const sdpString& sdpString::operator+=(const sdpString& String)
{
	append(String);
	return *this;
}

inline const sdpString& sdpString::operator+=(const sdpStringBase& String)
{
	append(String);
	return *this;
}

inline const sdpString& sdpString::operator+=(SDPCSTRING String)
{
	if(String != 0)
		append(String);
	
	return *this;
}

inline const sdpString& sdpString::operator+=(gchar Character)
{
	append(1, Character);
	return *this;
}

// sdpString friend addition functions defined as inline

inline sdpString operator+(const sdpString& String1, const sdpString& String2)
{
	sdpString result(String1);
	return result.append(String2);
}

inline sdpString operator+(const sdpString& String, gchar Character)
{
	sdpString result(String);
	return result.append(1, Character);
}

inline sdpString operator+(gchar Character, const sdpString& String)
{
	sdpString result(1, Character);
	return result.append(String);
}

inline sdpString operator+(const sdpString& String1, SDPCSTRING String2)
{
	sdpString result(String1);
	return String2 == 0 ? result : result.append(String2);
}

inline sdpString operator+(SDPCSTRING String1, const sdpString& String2)
{
	sdpString result(String1);
	return result.append(String2);
}


// Create and format a sdpString with one line of code instead of two
sdpString sdpFormatString(SDPCSTRING Format, ...);
sdpString sdpFormatStringV(SDPCSTRING Format, va_list ArgList);

// Handy STL collections ...
typedef std::vector<sdpString> sdpStringArray;
typedef std::list<sdpString> sdpStringList;
typedef std::stack<sdpString> sdpStringStack;
typedef std::set<sdpString> sdpStringSet;
typedef std::multiset<sdpString> sdpStringMultiSet;
typedef std::map<sdpString, sdpString> sdpStringMap;
typedef std::multimap<sdpString, sdpString> sdpStringMultiMap;

// Functional comparators ...
//
// Can be used in STL collections and algorithms to perform case-insensitive
// comparisons of sdpString objects - useful for maps where the key may be the
// proper string but the wrong case.

class sdpStringLessNoCase : std::binary_function<sdpString, sdpString, bool>
{
public:
	bool operator()(const sdpString& A, const sdpString& B) const
	{
		return A.lessthan(B, false);
	}
};

class sdpStringEqualsNoCase : std::binary_function<sdpString, sdpString, bool>
{
public:
	bool operator()(const sdpString& A, const sdpString& B) const
	{
		return A.equals(B, false);
	}
};

#endif SDPSTRING_H
