
// SDPTypes Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

/** \file
		\brief Implements templates, functions, and classes for manipulating files in a platform-independant manner
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpcompiler.h>
#include <sdpstring.h>
#include <sdpfile.h>

#include <stdio.h>
#include <algorithm>

#ifdef SDPWIN32
#include <direct.h>
#include <io.h>
#define popen _popen
#define read _read
#define PATH_MAX _MAX_PATH
#endif	// SDPWIN32

#ifdef SDPUNIX
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#endif // SDPUNIX

#include <sys/stat.h>

// Defines the maximum path length on the target platform
const gulong sdpMaxPathLength = PATH_MAX;

// Safely returns the full program path ...
sdpString sdpGetProcessPath(SDPCSTRING ArgVZero)
{
	sdpString result;

#if defined SDPWIN32
	::GetModuleFileName(0, result.buffer(1024), 1024);
	result = sdpDirectory(result);
#else // SDPWIN32
	result = sdpAppendPath(sdpGetCurrentDirectory(), sdpDirectory(ArgVZero));
#endif // !SDPWIN32

	return result;
}

// Safely returns an environment variable ...
sdpString sdpGetEnv(SDPCSTRING EnvironmentVariable)
{
	sdpString result;
	
	gchar* variable = g_getenv(EnvironmentVariable);
	if(variable)
		result = variable;
	
	return result;
}

// Safely returns the current working directory ...
sdpString sdpGetCurrentDirectory()
{
	sdpString result;

	gchar* buffer = g_get_current_dir();
	result = buffer;
	g_free(buffer);

	return result;
}

// Safely returns the user's home directory ...
sdpString sdpGetHomeDirectory()
{
	sdpString result = g_get_home_dir();
	return result;
}

// Safely returns the user's temp directory ...
sdpString sdpGetTempDirectory()
{
	sdpString result = g_get_tmp_dir();
	return result;
}

// Cross-platform method for creating a directory
bool sdpCreateDirectory(SDPCSTRING Path)
{
	sdpString path = sdpMakeNativeFilePath(Path);

#ifdef SDPMICROSOFT
	return CreateDirectory(path, 0) ? true : false;
#else // SDPMICROSOFT
	return 0 == mkdir(path, 0777) ? true : false;
#endif // !SDPMICROSOFT
}

bool sdpFileExists(SDPCSTRING Path)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Path, false);

	struct stat statistics;
	return (-1 == stat(Path, &statistics)) ? false : true;
}

// Force a filepath to the native filesystem's format (convert slashes as needed) ...
sdpString sdpMakeNativeFilePath(SDPCSTRING FilePath)
{
	static char foreign_dir_separator = (G_DIR_SEPARATOR == '/') ? '\\' : '/';
	static char native_dir_separator = G_DIR_SEPARATOR;

	sdpString filepath(FilePath);
	std::replace(filepath.begin(), filepath.end(), foreign_dir_separator, native_dir_separator);

	return filepath;
}

// Force a filepath to a POSIX format (slashes go *forward*) ...
sdpString sdpMakePosixFilePath(SDPCSTRING FilePath)
{
	static char non_posix_dir_separator = '\\';
	static char posix_dir_separator = '/';
	
	sdpString filepath(FilePath);
	std::replace(filepath.begin(), filepath.end(), non_posix_dir_separator, posix_dir_separator);
	
	return filepath;
}

// Returns the directory components of a path, or "." ...
sdpString sdpDirectory(SDPCSTRING FilePath)
{
	gchar* buffer = g_dirname(sdpMakeNativeFilePath(FilePath));
	sdpString result = buffer;
	g_free(buffer);
	
	return result;
}

// Returns the file components of a path ...
sdpString sdpFile(SDPCSTRING FilePath)
{
	return sdpString(g_basename(sdpMakeNativeFilePath(FilePath)));
}

// Append to a path, handle some special cases:
sdpString sdpAppendPath(SDPCSTRING Path1, SDPCSTRING Path2)
{
	// Convert pointers to string objects ...
	sdpString path1 = sdpMakeNativeFilePath(Path1);
	sdpString path2 = sdpMakeNativeFilePath(Path2);

	// If the first path is empty, ignore it ...
	if(!path1.size())
		return path2;

	// If the second path is empty, ignore it ...
	if(!path2.size())
		return path1;

	// The second path is absolute, so ignore the first ...
	if(g_path_is_absolute(path2))
		return path2;

	// The second path is ".", so ignore it ...
	if(path2 == ".")
		return path1;

	// Combine the two paths ...
	if(path1[path1.size()-1] == G_DIR_SEPARATOR)
		return path1 + path2;
	else
		return path1 + G_DIR_SEPARATOR + path2;
}

/// Runs a process using popen() and redirects its output to a stream, returning when the process completes
bool sdpRunProcess(SDPCSTRING CommandLine, std::ostream& Stream)
{
	// Sanity checks ...
	g_return_val_if_string_fail(CommandLine, false);

	// Start the child process ...
	FILE* file = popen(CommandLine, "r");
	g_return_val_if_fail(file, false);
	
	// Allocate a buffer for reading ...
	const gulong buffersize = 512;
	char* buffer = new char[buffersize+1];
	g_return_val_if_fail(buffer, false);
	
 	// Read data from the pipe ...
	for(int bytesread = read(fileno(file), buffer, buffersize); bytesread > 0; bytesread = read(fileno(file), buffer, buffersize))
		{
			buffer[bytesread] = '\0';
			Stream << buffer;
		}
		
	// Delete the buffer ...
	delete[] buffer;
	return true;
}

// Function object used with sdpGetFiles() ...
class sdpTypesGetFiles
{
public:
	sdpTypesGetFiles(sdpFileContainer& Container) : m_Container(Container) {}
	bool HandleFile(SDPCSTRING FilePath, SDPCSTRING File) { m_Container.push_back(FilePath); return true; }
protected:
	sdpFileContainer& m_Container;
};

// Add the contents of a directory to a list of file paths ...
void sdpGetFiles(SDPCSTRING StartDirectory, const bool Recursive, sdpFileContainer& Container)
{
	sdpTypesGetFiles functor(Container);
	sdpForEachFile(StartDirectory, functor, Recursive);
}

// Function object used with sdpGetDirectories() ...
class sdpTypesGetDirectories
{
public:
	sdpTypesGetDirectories(sdpFileContainer& Container) : m_Container(Container) {}
	bool HandleDirectory(SDPCSTRING DirectoryPath, SDPCSTRING Directory) { m_Container.push_back(DirectoryPath); return true; }
protected:
	sdpFileContainer& m_Container;
};

// Add the contents of a directory to a list of directory paths ...
void sdpGetDirectories(SDPCSTRING StartDirectory, const bool Recursive, sdpFileContainer& Container)
{
	sdpTypesGetDirectories functor(Container);
	sdpForEachDirectory(StartDirectory, functor, Recursive);
}


