
// SDPTypes Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/** \file
		\brief Implements math-related functions and constants
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpmath.h>

const gdouble sdpPi = 3.14159265358979323846;		// PI, per CRC handbook, 14th. ed.
const gdouble sdpPiOver2 = (sdpPi * 0.5);				// PI / 2
const gdouble sdpPiTimes2 = (sdpPi * 2.0);			// PI * 2

gdouble sdpRound(const gdouble Value)
{
	return (Value - floor(Value)) < 0.5 ? floor(Value) : ceil(Value);
}

gdouble sdpFactorial(const gulong N)
{
	gdouble result = 1;
	
	for(gulong i = 2; i <= N; i++)
		result *= i;
	
	return result;
}

gdouble sdpBernsteinBasis(const gulong Order, const gulong ControlPoint, const gdouble Parameter)
{
	// Sanity checks ...
	g_assert(Order > 1);
	g_assert(ControlPoint <= Order);
	
	gulong n = Order - 1;
	gulong i = ControlPoint;
	gdouble t = Parameter;
	
	gdouble ni = sdpFactorial(n) / (sdpFactorial(i) * sdpFactorial(n - i));
	
	return ni * pow(t, i) * pow((1 - t), (n - i));
}

bool sdpPlaneLineIntersection(const sdpPlane& Plane, const sdpVector3 LineOrigin, const sdpVector3 LineDirection, sdpVector3& Result)
{
	// Calculate the angle (dot product) between line and plane ...
	gdouble vd = Plane.Normal() * LineDirection;
	
	// Make sure the line & plane aren't parallel ...
	if(0.0 == vd)
		return false;

	gdouble v0 = -(Plane.Normal() * LineOrigin + Plane.Distance());
	gdouble t = v0 / vd;
	
	Result = LineOrigin + (t * LineDirection);

	return true;
}

bool sdpLineIntersection(const sdpVector2& A1, const sdpVector2& A2, const sdpVector2& B1, const sdpVector2& B2, sdpVector2& Result)
{
	sdpVector2 a = (A2 - A1).Normalize();
	sdpVector2 b = (B2 - B1).Normalize();
	sdpVector2 c = B1 - A1;
	sdpVector2 bperp = b.Perpendicular();

	gdouble denominator = bperp * a;
	if(0 == denominator)
		return false;

	gdouble t = (bperp * c) / denominator;
	
	Result = A1 + (t * a);

	return true;
}

sdpLineSegmentIntersectionType sdpLineSegmentIntersection(const sdpVector2& A1, const sdpVector2& A2, const sdpVector2& B1, const sdpVector2& B2)
{
	const gdouble sign1 = sdpSign(sdpArea(A1, A2, B1));
	const gdouble sign2 = sdpSign(sdpArea(A1, A2, B2));

	// The three most likely possibilities ...
	if(sign1 && sign2)
		{
			// No intersection ...
			if(sign1 == sign2)
				return SDP_NOINTERSECTION;
				
			sdpVector2 intersection;
			g_return_val_if_fail(sdpLineIntersection(A1, A2, B1, B2, intersection), SDP_ERROR);
			
			const gdouble sign3 = sdpSign(sdpArea(B1, intersection, A1));
			const gdouble sign4 = sdpSign(sdpArea(B1, intersection, A2));

			if(sign3 != sign4)
				return SDP_INTERIORINTERSECTION;

			return SDP_EXTERIORINTERSECTION;
		}
	// Two variations with single-point intersections ...
	else if(sign1 || sign2)
		{
			sdpVector2 intersection;
			g_return_val_if_fail(sdpLineIntersection(A1, A2, B1, B2, intersection), SDP_ERROR);
			
			const gdouble sign3 = sdpSign(sdpArea(B1, intersection, A1));
			const gdouble sign4 = sdpSign(sdpArea(B1, intersection, A2));

			if(sign3 != sign4)
				return SDP_INTERIORCOLINEARPOINT;
		
			return SDP_EXTERIORCOLINEARPOINT;
		}
	// The four colinear possibilities ...
	else
		{
			// This ray is getting crowded!  Find something somewhere else
			sdpVector2 start = B1 + (B2 - B1).Perpendicular();

			const gdouble sign3 = sdpSign(sdpArea(start, B1, A1));
			const gdouble sign4 = sdpSign(sdpArea(start, B1, A2));
			const gdouble sign5 = sdpSign(sdpArea(start, B2, A1));
			const gdouble sign6 = sdpSign(sdpArea(start, B2, A2));
			
			if((sign3 != sign4) && (sign5 != sign6))
				return SDP_INTERIORCOLINEARPOINTS;
				
			if((sign3 == sign4) && (sign4 == sign5) && (sign5 == sign6))
				return SDP_EXTERIORCOLINEARPOINTSNOINTERSECTION;
			
			if((sign3 == sign4) && (sign4 != sign5) && (sign5 == sign6))
				return SDP_EXTERIORCOLINEARPOINTS;
			
			return SDP_INTERIOREXTERIORCOLINEARPOINTS;
		}
}

gdouble sdpArea(const sdpVector2& Point1, const sdpVector2& Point2, const sdpVector2& Point3)
{
	gdouble result = 0;
	
	result += (Point1[0] * Point2[1]) - (Point2[0] * Point1[1]);
	result += (Point2[0] * Point3[1]) - (Point3[0] * Point2[1]);
	result += (Point3[0] * Point1[1]) - (Point1[0] * Point3[1]);
	
	return result * 0.5;
}


