
// SDPTypes Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
// 
// You should have received a copy of the GNU General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/** \file
		\brief Implements the sdpRegularExpression wrapper class for regular expressions
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpcompiler.h>

#ifdef SDPWIN32
#define __STDC__	1
extern "C"
{
#include <regex.h>
}
#else // SDPWIN32
#include <unistd.h>
#include <regex.h>
#endif // !SDPWIN32

#include <sdpregex.h>
#include <vector>

sdpRegEx::sdpRegEx()
{
	// Zero out the state structure ...
	memset(&m_State, 0, sizeof(m_State));
}

sdpRegEx::sdpRegEx(SDPCSTRING Expression)
{
	// Sanity checks ...
	g_assert_string(Expression);

	// Zero out the state structure ...
	memset(&m_State, 0, sizeof(m_State));
	
		// Compile the regular expression ...
	g_assert(Create(Expression));
}

sdpRegEx::sdpRegEx(SDPCSTRING Expression, const bool CaseSensitive)
{
	// Sanity checks ...
	g_assert_string(Expression);

	// Zero out the state structure ...
	memset(&m_State, 0, sizeof(m_State));
	
		// Compile the regular expression ...
	g_assert(Create(Expression, CaseSensitive));
}

sdpRegEx::~sdpRegEx()
{
	regfree(&m_State);
}

bool sdpRegEx::Create(SDPCSTRING Expression)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Expression, false);

	return Create(Expression, false);
}

bool sdpRegEx::Create(SDPCSTRING Expression, const bool CaseSensitive)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Expression, false);

	// Get rid of old state information ...
	regfree(&m_State);
	memset(&m_State, 0, sizeof(m_State));

	// Initialize our case-folding translation table ...
	if(!CaseSensitive)
		{
			m_State.translate = reinterpret_cast<char*>(malloc(256));
			g_return_val_if_fail(m_State.translate, false);
			
			for(gulong i = 0; i < 256; i++)
				m_State.translate[i] = i;
			for(gulong j = 'A'; j < 'Z'; j++)
				m_State.translate[j] = (j - 'A') + 'a';
		}

	// Compile the expression ...
	re_syntax_options = RE_SYNTAX_EGREP;
	g_return_val_if_fail(0 == re_compile_pattern(Expression, strlen(Expression), &m_State), false);

	return true;
}

gint sdpRegEx::Search(SDPCSTRING String) const
{
	// Sanity checks ...
	g_return_val_if_string_fail(String, -1);
	
	sdpRegExMatchArray subexpressions;
	return Search(String, subexpressions);
}

gint sdpRegEx::Search(SDPCSTRING String, sdpRegExMatchArray& Subexpressions) const
{
	// Sanity checks ...
	g_return_val_if_string_fail(String, -1);
	
	return Search(String, 0, strlen(String), Subexpressions);
}

gint sdpRegEx::Search(SDPCSTRING String, const gint Start, const gint Range) const
{
	// Sanity checks ...
	g_return_val_if_string_fail(String, -1);

	sdpRegExMatchArray subexpressions;	
	return Search(String, Start, Range, subexpressions);
}

gint sdpRegEx::Search(SDPCSTRING String, const gint Start, const gint Range, sdpRegExMatchArray& Subexpressions) const
{
	// Sanity checks ...
	g_return_val_if_string_fail(String, -1);

	// Setup a structure for returning match results ...	
	re_registers matches;
	memset(&matches, 0, sizeof(matches));
	
	// Perform the search ...
	int result = re_search(const_cast<re_pattern_buffer*>(&m_State), String, strlen(String), Start, Range, &matches);

	// Copy matches into our output array ...
	Subexpressions.clear();
	for(guint i = 0; i <matches.num_regs; i++)
		if(matches.start[i] != -1)
			Subexpressions.push_back(sdpRegExMatch(matches.start[i], matches.end[i]));

	// Cleanup memory ...
	free(matches.start);
	free(matches.end);
	memset(&matches, 0, sizeof(matches));
																											
	return result;
}


