
// SDPXML Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/** \file
		\brief Implements the sdpXmlDocument class, which implements sdpXmlIDocument
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpxmldocument.h>
#include <sdpxmlelement.h>
#include <sdpxmlutility.h>

#include <sdpfile.h>
#include <sdpuserdata.h>

#include <algorithm>
#include <strstream>
#include <fstream>
#include <stdarg.h>

////////////////////////////////////////////////////////////////////////////////
// sdpXmlDocument

sdpXmlDocument::sdpXmlDocument()
{
	m_RootElement = 0;
	m_Indentation = "\t";
}

sdpXmlDocument::~sdpXmlDocument()
{
	Clear();
}

void sdpXmlDocument::Clear()
{
	if(m_RootElement)
		DestroyRootElement();
}

bool sdpXmlDocument::LoadFile(SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	// Open the file as a stream ...
	std::ifstream file(FilePath);
	g_return_val_if_fail(file.good(), false);
	
	return Load(file, FilePath);
}

bool sdpXmlDocument::Load(sdpString& Buffer, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Buffer.size(), false);

	// Turn the string buffer into a stream ...
	std::strstream stream;
	stream << Buffer << std::ends;

	return Load(stream, FilePath);
}

bool sdpXmlDocument::Load(std::istream& Stream, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_return_val_if_fail(Stream.good(), false);

	// Store our file path ...
	m_FilePath = sdpAppendPath(sdpGetCurrentDirectory(), FilePath);

	return sdpXmlParse(this, Stream, m_FilePath);
}

bool sdpXmlDocument::SaveFile(SDPCSTRING FilePath) const
{
	// Sanity checks ...
	g_return_val_if_string_fail(FilePath, false);

	// Store the file path for posterity and make it absolute ...
//	m_FilePath = sdpAppendPath(sdpGetCurrentDirectory(), FilePath);

	// Open the file ...
	std::ofstream file(FilePath);
	g_return_val_if_fail(file.good(), false);
	
	return Save(file);
}

bool sdpXmlDocument::Save(std::ostream& Stream) const
{
	// Sanity checks ...
	g_return_val_if_fail(Stream.good(), false);

	if(GetRootElement())
		Save(Stream, 0, GetRootElement());
	
	return true;
}

void sdpXmlDocument::Save(std::ostream& Stream, gulong Level, sdpXmlIElement* Element) const
{
	// Sanity checks ...
	g_assert(Element);
	
	Stream << Indentation(Level) << "<" << Element->Name();

	for(SDPXMLATTRIBUTEHANDLE a = Element->GetFirstAttribute(); a; a = Element->GetNextAttribute(a))
		Save(Stream, Element->GetAttribute(a));

	if(Element->GetFirstElement() || strlen(Element->Text()))
		{
			Stream << ">" << Element->Text() << std::endl;
	
			for(SDPXMLELEMENTHANDLE e = Element->GetFirstElement(); e; e = Element->GetNextElement(e))
				Save(Stream, Level+1, Element->GetElement(e));
	
			Stream << Indentation(Level) << "</" << Element->Name() << ">" << std::endl;
		}
	else
		{
			Stream << "/>" << std::endl;
		}
}

void sdpXmlDocument::Save(std::ostream& Stream, sdpXmlIAttribute* Attribute) const
{
	// Sanity checks ...
	g_assert(Attribute);

	Stream << " " << Attribute->Name() << "=\"" << sdpXmlEncodeString(Attribute->Value()) << "\"";
}

sdpString sdpXmlDocument::Indentation(gulong Level) const
{
	// Setup indentation ...
	sdpString result;
	
	for(gulong indent = 0; indent < Level; indent++)
		result += m_Indentation;
		
	return result;
}

////////////////////////////////////////////////////////////
// sdpXmlIDocument overrides ...

SDPCSTRING sdpXmlDocument::FilePath() const
{
	return m_FilePath;
}

sdpXmlIElement* sdpXmlDocument::GetRootElement() const
{
	return m_RootElement;
}

sdpXmlIElement* sdpXmlDocument::CreateRootElement(SDPCSTRING Name)
{
	// Sanity checks ...
	g_return_val_if_fail(0 == m_RootElement, 0);
	g_return_val_if_string_fail(Name, 0);
	
	m_RootElement = new sdpXmlElement(this, 0, Name, sdpStringArray(), 0, 0);
	return m_RootElement;
}

sdpResult sdpXmlDocument::DestroyRootElement()
{
	// Sanity checks ...
	g_return_val_if_fail(m_RootElement, SDP_E_FAIL);

	// Get rid of the element ...
	delete m_RootElement;
	m_RootElement = 0;
	
	return SDP_S_OK;
}

sdpResult sdpXmlDocument::SetIndentation(SDPCSTRING Indentation)
{
	m_Indentation = Indentation;
	
	return SDP_S_OK;
}

SDPCSTRING sdpXmlDocument::GetIndentation() const
{
	return m_Indentation;
}

sdpIUserData* sdpXmlDocument::UserData()
{
	return this;
}

///////////////////////////////////////////////////////////////
// SAX event handlers ...

void sdpXmlDocument::StartDocument()
{
//std::cout << "sdpXmlDocument::StartDocument()" << std::endl;

	// Make sure we start out with an empty document ...
	if(m_RootElement)
		DestroyRootElement();
	
	// Make sure we start with an empty stack ...
	while(m_ElementStack.size())	
		m_ElementStack.pop();
}

void sdpXmlDocument::EndDocument()
{
//std::cout << "sdpXmlDocument::EndDocument()" << std::endl;

	// Make sure we end up with an empty stack ...
	g_return_if_fail(m_ElementStack.empty());
}

void sdpXmlDocument::ProcessingInstruction(SDPCSTRING Target, SDPCSTRING Data)
{
//std::cout << "sdpXmlDocument::ProcessingInstruction() " << Target << std::endl;
}

void sdpXmlDocument::StartElement(SDPCSTRING Name, const sdpStringArray& Attributes)
{
//std::cout << "sdpXmlDocument::StartElement() " << "<" << Name << ">" << std::endl;

	if(m_ElementStack.empty())
		{
			m_RootElement = new sdpXmlElement(this, 0, Name, Attributes, 0, 0);
			m_ElementStack.push(m_RootElement);
		}
	else
		{
			sdpXmlElement* newelement = new sdpXmlElement(this, m_ElementStack.top(), Name, Attributes, 0, 0);
			m_ElementStack.top()->m_Elements.push_back(newelement);
			m_ElementStack.push(newelement);
		}
}

void sdpXmlDocument::EndElement()
{
//std::cout << "sdpXmlDocument::EndElement()" << std::endl;

	g_return_if_fail(m_ElementStack.size());
	m_ElementStack.pop();
}

void sdpXmlDocument::Characters(SDPCSTRING Text)
{
//std::cout << "sdpXmlDocument::Characters() " << Text << std::endl;

	// Get rid of whitespace before we do anything ...
	sdpString text(Text);
	text.trim();

	// If we've still got something, assign it to our element ...
	if(text.size())
		if(m_ElementStack.top())
			m_ElementStack.top()->SetText(text);
}

void sdpXmlDocument::CDATABlock(SDPCSTRING Value)
{
//std::cout << "sdpXmlDocument::CDATABlock()" << std::endl;

	// Sanity checks ...
	g_assert(m_RootElement);
	g_assert(m_ElementStack.size());

	Characters(Value);

/*
	sdpXmlElement* newelement = new sdpXmlElement(this, m_ElementStack.top(), "<![CDATA[", 0, 0, 0);
	newelement->SetText(Value);
	m_ElementStack.top()->m_Elements.push_back(newelement);
*/
}

void sdpXmlDocument::Warning(SDPCSTRING Message)
{
//	std::cout << FilePath() << ": " << CurrentLine() << "." << CurrentColumn() << ": " << sdpFormatStringV(Message, ArgList) << std::endl;
	std::cerr << FilePath() << ": " << Message << std::endl;
}

void sdpXmlDocument::Error(SDPCSTRING Message)
{
//	std::cout << FilePath() << ": " << CurrentLine() << "." << CurrentColumn() << ": " << sdpFormatStringV(Message, ArgList) << std::endl;
	std::cerr << FilePath() << ": " << Message << std::endl;
}

void sdpXmlDocument::FatalError(SDPCSTRING Message)
{
//	std::cout << FilePath() << ": " << CurrentLine() << "." << CurrentColumn() << ": " << sdpFormatStringV(Message, ArgList) << std::endl;
	std::cerr << FilePath() << ": " << Message << std::endl;
}

/*
gulong sdpXmlDocument::CurrentLine() const
{
	return getLineNumber(m_ParserContext);
}

gulong sdpXmlDocument::CurrentColumn() const
{
	return getColumnNumber(m_ParserContext);
}
*/

//////////////////////////////////////////////////////////////////////////////////
// Odds-and-ends

std::ostream& operator << (std::ostream& Stream, const sdpXmlDocument& Document)
{
	Document.Save(Stream);
	return Stream;
}

std::ostream& operator << (std::ostream& Stream, const sdpXmlIElement* const Element)
{
	// Sanity checks ...
	g_assert(Element);
	
	Stream << "<" << Element->Name();

	for(SDPXMLATTRIBUTEHANDLE a = Element->GetFirstAttribute(); a; a = Element->GetNextAttribute(a))
		Stream << " " << Element->GetAttribute(a)->Name() << "=\"" << Element->GetAttribute(a)->Value() << "\"";

	if(Element->GetFirstElement() || strlen(Element->Text()))
		{
			Stream << ">" << Element->Text() << std::endl;
	
			for(SDPXMLELEMENTHANDLE e = Element->GetFirstElement(); e; e = Element->GetNextElement(e))
				Stream << Element->GetElement(e);
	
			Stream << "</" << Element->Name() << ">" << std::endl;
		}
	else
		{
			Stream << "/>" << std::endl;
		}
		
	return Stream;
}

