
// SDPXML Library
// Copyright  1995-2000, Timothy M. Shead
//
// Contact: tshead@k-3d.com
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


/** \file
		\brief Implements the sdpXmlElement class, which implements sdpXmlIElement
		\author Tim Shead (tshead@k-3d.com)
*/

#include <sdpxmlattribute.h>
#include <sdpxmlelement.h>
#include <sdpxmlutility.h>

#include <algorithm>

/////////////////////////////////////////////////////
// class sdpXmlElement

// Constructor that gets called by the SAX parsing engine in sdpXmlDocument ...
sdpXmlElement::sdpXmlElement(sdpXmlIDocument* Document, sdpXmlIElement* ParentElement, SDPCSTRING Name, const sdpStringArray& Attributes, gulong Line, gulong Column)
{
	// Sanity checks ...
	g_assert(Document);
	g_assert_string(Name);

	m_Document = Document;
	m_ParentElement = ParentElement;
	m_Name = Name;
	
	m_Line = Line;
	m_Column = Column;
	
	// Add all our attributes ...
	for(gulong index = 0; index < Attributes.size(); index += 2)
		m_Attributes.push_back(new sdpXmlAttribute(this, Attributes[index], sdpXmlDecodeString(Attributes[index+1]), 0, 0));
}

sdpXmlElement::~sdpXmlElement()
{
	// Cleanup child elements ...
	for(ElementIterator element = m_Elements.begin(); element != m_Elements.end(); element++)
		delete (*element);

	// Cleanup child attributes ...
	for(AttributeIterator attribute = m_Attributes.begin(); attribute != m_Attributes.end(); attribute++)
		delete (*attribute);
}

sdpXmlIDocument* sdpXmlElement::Document() const
{
	return m_Document;
}

sdpXmlIElement* sdpXmlElement::ParentElement() const
{
	return m_ParentElement;
}

SDPCSTRING sdpXmlElement::Name() const
{
	return m_Name;
}

sdpResult sdpXmlElement::SetName(SDPCSTRING Name)
{
	g_return_val_if_string_fail(Name, SDP_E_FAIL);
	
	m_Name = Name;
	return SDP_S_OK;
}

SDPCSTRING sdpXmlElement::Text() const
{
	return m_Text;
}

sdpResult sdpXmlElement::SetText(SDPCSTRING Text)
{
	m_Text = Text;
	return SDP_S_OK;
}

gulong sdpXmlElement::Line() const
{
	return m_Line;
}

gulong sdpXmlElement::Column() const
{
	return m_Column;
}

SDPXMLATTRIBUTEHANDLE sdpXmlElement::GetFirstAttribute() const
{
	ConstAttributeIterator i = m_Attributes.begin();
	return i != m_Attributes.end() ? SDPXMLATTRIBUTEHANDLE(i - m_Attributes.begin() + 1) : 0;
}

SDPXMLATTRIBUTEHANDLE sdpXmlElement::GetNextAttribute(SDPXMLATTRIBUTEHANDLE Attribute) const
{
	// Sanity check ...
	g_return_val_if_fail(Attribute, 0);

	ConstAttributeIterator i = m_Attributes.begin() + gint(Attribute) - 1;
	i++;

	return i != m_Attributes.end() ? SDPXMLATTRIBUTEHANDLE(i - m_Attributes.begin() + 1) : 0;
}

sdpXmlIAttribute* sdpXmlElement::GetAttribute(SDPXMLATTRIBUTEHANDLE Attribute) const
{
	// Sanity check ...
	g_return_val_if_fail(Attribute, 0);

	ConstAttributeIterator i = m_Attributes.begin() + gint(Attribute) - 1;

	return (*i);
}

sdpXmlIAttribute* sdpXmlElement::CreateAttribute(SDPCSTRING Name, SDPCSTRING Value)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, 0);
	
	sdpXmlIAttribute* attribute = sdpXmlFindAttribute(this, Name);
	if(attribute)
		{
			attribute->SetValue(Value);
			return attribute;
		}

	sdpXmlAttribute* newattribute = new sdpXmlAttribute(this, Name, Value, 0, 0);
	g_return_val_if_fail(newattribute, 0);
	m_Attributes.push_back(newattribute);
		
	return newattribute;
}

sdpResult sdpXmlElement::DestroyAttribute(sdpXmlIAttribute* Attribute)
{
	// Sanity checks ...
	g_return_val_if_fail(Attribute, SDP_E_FAIL);
	
	// Find the attribute ...
	sdpXmlAttribute* attribute = static_cast<sdpXmlAttribute*>(Attribute);
	AttributeIterator i = std::find(m_Attributes.begin(), m_Attributes.end(), attribute);
	g_return_val_if_fail(i != m_Attributes.end(), SDP_E_FAIL);
	
	// Zap it ...
	delete (*i);
	m_Attributes.erase(i);
	
	return SDP_S_OK;
}

SDPXMLELEMENTHANDLE sdpXmlElement::GetFirstElement() const
{
	ConstElementIterator i = m_Elements.begin();
	return i != m_Elements.end() ? SDPXMLELEMENTHANDLE(i - m_Elements.begin() + 1) : 0;
}

SDPXMLELEMENTHANDLE sdpXmlElement::GetNextElement(SDPXMLELEMENTHANDLE Element) const
{
	// Sanity check ...
	g_return_val_if_fail(Element, 0);

	ConstElementIterator i = m_Elements.begin() + gint(Element) - 1;
	i++;

	return i != m_Elements.end() ? SDPXMLELEMENTHANDLE(i - m_Elements.begin() + 1) : 0;
}

sdpXmlIElement* sdpXmlElement::GetElement(SDPXMLELEMENTHANDLE Element) const
{
	// Sanity check ...
	g_return_val_if_fail(Element, 0);

	ConstElementIterator i = m_Elements.begin() + gint(Element) - 1;

	return (*i);
}

sdpXmlIElement* sdpXmlElement::CreateElement(SDPCSTRING Name)
{
	// Sanity checks ...
	g_return_val_if_string_fail(Name, 0);
	
	sdpXmlElement* element = new sdpXmlElement(m_Document, this, Name, sdpStringArray(), 0, 0);
	g_return_val_if_fail(element, 0);
	
	m_Elements.push_back(element);
	
	return element;
}

sdpResult sdpXmlElement::DestroyElement(sdpXmlIElement* Element)
{
	// Sanity checks ...
	g_return_val_if_fail(Element, SDP_E_FAIL);
	
	// Find the element ...
	sdpXmlElement* element = static_cast<sdpXmlElement*>(Element);
	ElementIterator i = std::find(m_Elements.begin(), m_Elements.end(), element);
	g_return_val_if_fail(i != m_Elements.end(), SDP_E_FAIL);
	
	// Zap it ...
	delete (*i);
	m_Elements.erase(i);
	
	return SDP_S_OK;
}

sdpIUserData* sdpXmlElement::UserData()
{
	return this;
}

