#include "sdpxmlsax.h"

#include <sdpstring.h>

#include <deque>
#include <iostream>
#include <fstream>

bool sdpXmlParseFile(sdpXmlISAXHandler* Handler, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_assert(Handler);
	g_assert_string(FilePath);

	std::ifstream file(FilePath);
	g_return_val_if_fail(file.good(), false);

	return sdpXmlParse(Handler, file, FilePath);
}

static char sdpXmlGetChar(std::istream& Stream, gulong& Line, gulong& Column)
{
	// Sanity checks ...
	g_assert(Stream.good());

	char result = 0;

	Stream.get(result);

	if(result == '\n')
		{
			Line++;
			Column = 0;
		}
	else
		{
			Column++;
		}

	return result;
}

static void sdpXmlSkipWhitespace(std::istream& Stream, gulong& Line, gulong& Column)
{
	// Sanity checks ...
	g_assert(Stream.good());

	for(char c = Stream.peek(); isspace(c); c = Stream.peek())
		c = sdpXmlGetChar(Stream, Line, Column);
}

bool sdpXmlParse(sdpXmlISAXHandler* Handler, std::istream& Stream, SDPCSTRING FilePath)
{
	// Sanity checks ...
	g_assert(Handler);
	g_assert(Stream.good());

	// Setup our parser state vector ...
	typedef enum
	{
		STARTDOCUMENT,
		PROCESSINGINSTRUCTION,
		COMMENT,
		STARTELEMENT,
		ATTRIBUTENAME,
		ATTRIBUTEVALUE,
		ELEMENT,
		ENDELEMENT,
	} ParserState;

	ParserState state = STARTDOCUMENT;
	gulong line = 1;
	gulong column = 0;
	gulong depth = 0;

	char c;
	sdpString pitarget;
	sdpString elementname;
	sdpString attributename;
	sdpString attributevalue;
	sdpString characterdata;
	sdpStringArray attributes;

	// Start the document ...
	Handler->StartDocument();

	// Skip any initial whitespace ...
	sdpXmlSkipWhitespace(Stream, line, column);

	while(!Stream.eof())
		{
			switch(state)
				{
					case STARTDOCUMENT:
						c = sdpXmlGetChar(Stream, line, column);
						if(c == '<')
							{
								c = sdpXmlGetChar(Stream, line, column);
								if(c == '?')
									{
										state = PROCESSINGINSTRUCTION;
									}
								else if(c == '!')
									{
										state = COMMENT;
									}
								else
									{
										state = STARTELEMENT;
										elementname += c;
									}
							}
						else
							{
								sdpString message = sdpString(line) + "." + sdpString(column) + ": Unexpected character near start-of document";
								Handler->Error(message);
							}
						break;

					case PROCESSINGINSTRUCTION:
						c = sdpXmlGetChar(Stream, line, column);
						if(c == '?')
							{
								c = sdpXmlGetChar(Stream, line, column);
								if(c != '>')
									{
										sdpString message = sdpString(line) + "." + sdpString(column) + ": Unexpected character near end-of processing instruction";
										Handler->Error(message);
									}

								Handler->ProcessingInstruction(pitarget, 0);
								pitarget.clear();
								state = ELEMENT;
							}
						else
							{
								pitarget += c;
							}
						break;

					case COMMENT:
						c = sdpXmlGetChar(Stream, line, column);
						if(c == '-')
							{
								c = sdpXmlGetChar(Stream, line, column);
								if(c == '-')
									{
										c = sdpXmlGetChar(Stream, line, column);
										if(c == '>')
											{
												state = ELEMENT;
												break;
											}
									}
							}

						break;

					case STARTELEMENT:
						c = sdpXmlGetChar(Stream, line, column);
						if(c == '/')
							{
								c = sdpXmlGetChar(Stream, line, column);
								if(c != '>')
									{
										sdpString message = sdpString(line) + "." + sdpString(column) + ": Unexpected character near end-of element";
										Handler->Error(message);
									}

								state = ELEMENT;
								Handler->StartElement(elementname, attributes);
								depth++;
								
								Handler->EndElement();
								depth--;
								if(0 == depth)
									{
										Handler->EndDocument();
										return true;
									}
								
								elementname.clear();
								attributes.clear();
							}
						else if(c == '>')
							{
								state = ELEMENT;
								Handler->StartElement(elementname, attributes);
								depth++;
								elementname.clear();
								attributes.clear();
							}
						else if(isspace(c))
							{
								sdpXmlSkipWhitespace(Stream, line, column);

								c = sdpXmlGetChar(Stream, line, column);
								if(c == '/')
									{
										c = sdpXmlGetChar(Stream, line, column);
										if(c != '>')
											{
												sdpString message = sdpString(line) + "." + sdpString(column) + ": Unexpected character near end-of element";
												Handler->Error(message);
											}

										state = ELEMENT;
										Handler->StartElement(elementname, attributes);
										depth++;
										
										Handler->EndElement();
										depth--;
										
										if(0 == depth)
											{
												Handler->EndDocument();
												return true;
											}
										
										elementname.clear();
										attributes.clear();
									}
								else if(c == '>')
									{
										state = ELEMENT;
										Handler->StartElement(elementname, attributes);
										depth++;
										elementname.clear();
										attributes.clear();
									}
								else
									{
										state = ATTRIBUTENAME;
										attributename += c;
									}
							}
						else
							{
								elementname += c;
							}
						break;

					case ATTRIBUTENAME:
						c = sdpXmlGetChar(Stream, line, column);
						if(c == '=')
							{
								sdpXmlSkipWhitespace(Stream, line, column);

								c = sdpXmlGetChar(Stream, line, column);
								if(c != '\'' && c != '"')
									{
										sdpString message = sdpString(line) + "." + sdpString(column) + ": Unexpected character near begining-of attribute value";
										Handler->Error(message);
									}

								state = ATTRIBUTEVALUE;
								break;
							}
						else if(isspace(c))
							{
								sdpXmlSkipWhitespace(Stream, line, column);
							}
						else
							{
								attributename += c;
							}
						break;

					case ATTRIBUTEVALUE:
						c = sdpXmlGetChar(Stream, line, column);
						if(c == '\'' || c == '"')
							{
								attributes.push_back(attributename);
								attributes.push_back(attributevalue);
								attributename.clear();
								attributevalue.clear();

								state = STARTELEMENT;
							}
						else
							{
								attributevalue += c;
							}
						break;

					case ELEMENT:
						c = sdpXmlGetChar(Stream, line, column);
						if(c == '<')
							{
								characterdata.trim();
								if(characterdata.size())
									{
										Handler->Characters(characterdata);
										characterdata.clear();
									}

								c = Stream.peek();
								if(c == '/')
									{
										state = ENDELEMENT;
									}
								else if(c == '!')
									{
										state = COMMENT;
									}
								else
									{
										state = STARTELEMENT;
									}
							}
						else
							{
								characterdata += c;
							}
						break;

					case ENDELEMENT:
						c = sdpXmlGetChar(Stream, line, column);
						if(c == '>')
							{
								state = ELEMENT;
								Handler->EndElement();
								depth--;
								
								if(0 == depth)
									{
										Handler->EndDocument();
										return true;
									}
							}
						break;
				}
		}

	// If we got this far, we're missing a close-element somewhere ...
	sdpString message = sdpString(line) + "." + sdpString(column) + ": Premature end-of-document at depth " + sdpString(depth);
	Handler->Error(message);
	
	// Finish the document ...
	Handler->EndDocument();

	return true;
}

